% LOADSEGIMAGE  Load oversegmented image.
% Stephen Gould <sgould@stanford.edu>
%
% Returns a structure with fields:
%   img     :: image
%   seg     :: segmentation
%   labels  :: groundtruth labels (if provided)
%   ... TODO

function [segImage] = loadSegImage(imgFilename, segFilename, labelFilename);

% load data
segImage.img = imread(imgFilename);
[H, W, C] = size(segImage.img);
segImage.seg = load(segFilename);
segImage.seg = segImage.seg - min(segImage.seg(:));

if (nargin == 3),
    segImage.labels = load(labelFilename);
else
    segImage.labels = -1 * ones(H, W);
end;

% basic statistics
segImage.minSegId = min(segImage.seg(:));
segImage.maxSegId = max(segImage.seg(:));
segImage.edgeIndx = [find(segImage.seg(:, [1, 1:end - 1]) ~= segImage.seg); ...
    find(segImage.seg([1, 1:end - 1], :) ~= segImage.seg)];

% find 4-connected neighbors
segShiftRight = segImage.seg(:, [1, 1:end - 1]);
segShiftDown = segImage.seg([1, 1:end - 1], :);

indx = find(segShiftRight ~= segImage.seg);
a = segImage.seg(indx); b = segShiftRight(indx);

indx = find(segShiftDown ~= segImage.seg);
a = [a; segImage.seg(indx)]; b = [b; segShiftDown(indx)];

segImage.nbrs = unique(sortrows([a, b; b, a]), 'rows');
indx = find(segImage.nbrs(:, 1) < segImage.nbrs(:, 2));
segImage.nbrs = segImage.nbrs(indx, :);

% cache segment pixels indices
segImage.pixIndx = {};
for i = 0:segImage.maxSegId,
    indx = find(segImage.seg == i);
    segImage.pixIndx{i + 1} = indx;
end;

% cache edgels
segImage.edgels = {};
for i = 1:length(segImage.nbrs),
    indx1 = segImage.pixIndx{segImage.nbrs(i, 1) + 1}; %find(segImage.seg == segImage.nbrs(i, 1));
    indx2 = segImage.pixIndx{segImage.nbrs(i, 2) + 1}; %find(segImage.seg == segImage.nbrs(i, 2));
    indx = [indx1(or(segShiftRight(indx1) == segImage.nbrs(i, 2), ...
        segShiftDown(indx1) == segImage.nbrs(i, 2)));
        indx2(or(segShiftRight(indx2) == segImage.nbrs(i, 1), ...
        segShiftDown(indx2) == segImage.nbrs(i, 1))) ];
    [y, x] = ind2sub([H, W], unique(indx));
    segImage.edgels{i} = [y, x];
end;

% cache segment information
segImage.centroids = zeros(segImage.maxSegId + 1, 2);
segImage.segLabels = -1 * ones(segImage.maxSegId + 1, 1);
for i = 0:segImage.maxSegId,
    indx = segImage.pixIndx{i + 1}; %find(segImage.seg == i);
    if (isempty(indx)), continue; end;
    [y, x] = ind2sub([H, W], indx);
    segImage.centroids(i + 1, :) = [mean(x), mean(y)];
    segImage.segLabels(i + 1) = mode(segImage.labels(indx));
end;
