/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneAssignment.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"

using namespace std;

// SceneAssignment class ----------------------------------------------------

SceneAssignment::SceneAssignment() : pixels(NULL), horizon(-1)
{
    // do nothing
}

SceneAssignment::SceneAssignment(int height, int width)
{
    pixels = cvCreateMat(height, width, CV_32SC1);
    cvSet(pixels, cvScalar(-1));
    horizon = height / 2;
}

SceneAssignment::SceneAssignment(const SceneAssignment& assignment)
{
    // copy entities
    if (assignment.pixels != NULL) {
        pixels = cvCloneMat(assignment.pixels);
    } else {
        pixels = NULL;
    }

    regions.reserve(assignment.regions.size());
    for (unsigned i = 0; i < assignment.regions.size(); i++) {
        regions.push_back(new SceneRegion(*assignment.regions[i]));
    }
    objects.reserve(assignment.objects.size());
    for (unsigned i = 0; i < assignment.objects.size(); i++) {
        objects.push_back(new SceneObject(*assignment.objects[i]));
    }
    regionBoundaries.reserve(assignment.regionBoundaries.size());
    for (unsigned i = 0; i < assignment.regionBoundaries.size(); i++) {
        regionBoundaries.push_back(new SceneRegionBoundary(*assignment.regionBoundaries[i]));
    }
    objectBoundaries.reserve(assignment.objectBoundaries.size());
    for (unsigned i = 0; i < assignment.objectBoundaries.size(); i++) {
        objectBoundaries.push_back(new SceneObjRegBoundary(*assignment.objectBoundaries[i]));
    }
    horizon = assignment.horizon;

    // update pointers
    for (vector<SceneRegion *>::iterator it = regions.begin(); it != regions.end(); it++) {
        set<SceneRegionBoundary *> newBoundaries;
        for (set<SceneRegionBoundary *>::iterator jt = (*it)->boundaries.begin();
             jt != (*it)->boundaries.end(); jt++) {
            newBoundaries.insert(regionBoundaries[(*jt)->id()]);
        }
        (*it)->boundaries = newBoundaries;

        set<SceneObjRegBoundary *> newObjBoundaries;
        for (set<SceneObjRegBoundary *>::iterator jt = (*it)->objBoundaries.begin();
             jt != (*it)->objBoundaries.end(); jt++) {
            newObjBoundaries.insert(objectBoundaries[(*jt)->id()]);
        }
        (*it)->objBoundaries = newObjBoundaries;

        if ((*it)->object != NULL) {
            (*it)->object = objects[(*it)->object->id()];
        }
    }

    for (vector<SceneObject *>::iterator it = objects.begin(); it != objects.end(); it++) {
        set<SceneRegion *> newRegions;
        for (set<SceneRegion *>::iterator jt = (*it)->regions.begin();
             jt != (*it)->regions.end(); jt++) {
            newRegions.insert(regions[(*jt)->id()]);
        }
        (*it)->regions = newRegions;

        set<SceneObjRegBoundary *> newBoundaries;
        for (set<SceneObjRegBoundary *>::iterator jt = (*it)->boundaries.begin();
             jt != (*it)->boundaries.end(); jt++) {
            newBoundaries.insert(objectBoundaries[(*jt)->id()]);
        }
        (*it)->boundaries = newBoundaries;
    }

    for (unsigned i = 0; i < regionBoundaries.size(); i++) {
        regionBoundaries[i]->regionA = regions[regionBoundaries[i]->regionA->id()];
        regionBoundaries[i]->regionB = regions[regionBoundaries[i]->regionB->id()];
    }

    for (unsigned i = 0; i < objectBoundaries.size(); i++) {
        objectBoundaries[i]->object = objects[objectBoundaries[i]->object->id()];
        objectBoundaries[i]->region = regions[objectBoundaries[i]->region->id()];
    }

    //copy scene depth information (used by sceneReconstruct3D)
    minDepth = MatrixXd(assignment.minDepth);
    maxDepth = MatrixXd(assignment.maxDepth);
}

SceneAssignment::~SceneAssignment()
{
    clear();
    if (pixels != NULL)
        cvReleaseMat(&pixels);
}

// i/o
void SceneAssignment::read(const char *filename)
{
    SVL_ASSERT(filename != NULL);

    clear();
    ifstream ifs(filename, ios::binary);
    SVL_ASSERT(!ifs.fail());

    // read version
    int nVersion;
    ifs.read((char *)&nVersion, sizeof(int));
    SVL_ASSERT(nVersion == 1);

    // read size
    int nRows, nCols, nRegions, nObjects;
    ifs.read((char *)&nRows, sizeof(int));
    ifs.read((char *)&nCols, sizeof(int));
    SVL_ASSERT(!ifs.fail());

    if ((nRows == 0) || (nCols == 0)) {
        SVL_LOG(SVL_LOG_WARNING, "scene assignment " << filename << " is empty");
        ifs.close();
        return;
    }

    resize(nRows, nCols);

    ifs.read((char *)&nRegions, sizeof(int));
    ifs.read((char *)&nObjects, sizeof(int));
    SVL_ASSERT(!ifs.fail());

    allocateRegions(nRegions);
    allocateObjects(nObjects);

    // read pixel assignment
    ifs.read((char *)CV_MAT_ELEM_PTR(*pixels, 0, 0), pixels->rows * pixels->cols * sizeof(int));
    SVL_ASSERT(!ifs.fail());

    // read horizon
    ifs.read((char *)&horizon, sizeof(int));
    SVL_ASSERT(!ifs.fail());

    // read regions and objects
    for (vector<SceneRegion *>::iterator it = regions.begin(); it != regions.end(); it++) {
        int objectId;
        ifs.read((char *)&(*it)->Sk, sizeof(int));
        ifs.read((char *)&objectId, sizeof(int));
        if (objectId != -1) {
            SVL_ASSERT(objectId < this->numObjects());
            (*it)->object = objects[objectId];
        }
    }
    SVL_ASSERT(!ifs.fail());

    for (vector<SceneObject *>::iterator it = objects.begin(); it != objects.end(); it++) {
        ifs.read((char *)&(*it)->Ok, sizeof(int));
    }
    SVL_ASSERT(!ifs.fail());
    
    // set region pixels
    const int *regId = (const int *)CV_MAT_ELEM_PTR(*pixels, 0, 0);
    for (int y = 0; y < pixels->rows; y++) {
        for (int x = 0; x < pixels->cols; x++, regId++) {
            SVL_ASSERT((*regId >= 0) && (*regId < nRegions));
            regions[*regId]->pixels.push_back(cvPoint(x, y));
        }
    }

    // connectivity
    updateNeighborhoods();
    ifs.close();
}

void SceneAssignment::write(const char *filename) const
{
    SVL_ASSERT(filename != NULL);
    ofstream ofs(filename, ios::binary);
    SVL_ASSERT(!ofs.fail());

    // version (so can read back old assignments if things change)
    int nVersion = 1;
    ofs.write((char *)&nVersion, sizeof(int));

    int nZero = 0;
    if (pixels == NULL) {
        SVL_LOG(SVL_LOG_WARNING, "writing empty scene assignment to " << filename);
        ofs.write((char *)&nZero, sizeof(int));
        ofs.write((char *)&nZero, sizeof(int));
        ofs.close();
        return;
    }

    // size (pixels, regions, objects)
    ofs.write((char *)&pixels->rows, sizeof(int));
    ofs.write((char *)&pixels->cols, sizeof(int));

    int numRegions = this->numRegions();
    ofs.write((char *)&numRegions, sizeof(int));

    int numObjects = this->numObjects();
    ofs.write((char *)&numObjects, sizeof(int));

    // pixel assignment
    ofs.write((char *)CV_MAT_ELEM_PTR(*pixels, 0, 0), pixels->rows * pixels->cols * sizeof(int));

    // horizon
    ofs.write((char *)&horizon, sizeof(int));

    // regions/objects
    for (vector<SceneRegion *>::const_iterator it = regions.begin(); it != regions.end(); it++) {
        int objectId = ((*it)->object == NULL) ? -1 : (*it)->object->id();
        ofs.write((char *)&(*it)->Sk, sizeof(int));
        ofs.write((char *)&objectId, sizeof(int));
    }

    for (vector<SceneObject *>::const_iterator it = objects.begin(); it != objects.end(); it++) {
        ofs.write((char *)&(*it)->Ok, sizeof(int));
    }

    ofs.close();
}

void SceneAssignment::resize(int height, int width)
{
    clear();
    if (pixels != NULL)
        cvReleaseMat(&pixels);

    pixels = cvCreateMat(height, width, CV_32SC1);
    cvSet(pixels, cvScalar(-1));
    horizon = height / 2;
}

void SceneAssignment::clear()
{
    for (unsigned i = 0; i < regions.size(); i++) {
        delete regions[i];
    }
    regions.clear();

    for (unsigned i = 0; i < objects.size(); i++) {
        delete objects[i];
    }
    objects.clear();

    for (unsigned i = 0; i < regionBoundaries.size(); i++) {
        delete regionBoundaries[i];
    }
    regionBoundaries.clear();

    for (unsigned i = 0; i < objectBoundaries.size(); i++) {
        delete objectBoundaries[i];
    }
    objectBoundaries.clear();
}

void SceneAssignment::clearCachedModelData()
{
    for (vector<SceneRegion *>::iterator it = regions.begin(); it != regions.end(); it++) {
        (*it)->clear();
    }
    for (vector<SceneObject *>::iterator it = objects.begin(); it != objects.end(); it++) {
        (*it)->clear();
    }
    for (vector<SceneRegionBoundary *>::iterator it = regionBoundaries.begin();
         it != regionBoundaries.end(); it++) {
        (*it)->clear();
    }
    for (vector<SceneObjRegBoundary *>::iterator it = objectBoundaries.begin();
         it != objectBoundaries.end(); it++) {
        (*it)->clear();
    }
}

void SceneAssignment::allocateRegions(int numRegions)
{
    regions.reserve(numRegions);
    while ((int)regions.size() < numRegions) {
        regions.push_back(new SceneRegion((int)regions.size()));
    }
}

void SceneAssignment::allocateObjects(int numObjects)
{
    objects.reserve(numObjects);
    while ((int)objects.size() < numObjects) {
        objects.push_back(new SceneObject((int)objects.size()));
    }
}

void SceneAssignment::remap()
{
    // pixel assignment mapping
    vector<int> mapping;
    mapping.reserve(regions.size());

    // copy over all non-empty assignment and update indicies
    vector<SceneRegion *> newRegions;
    newRegions.reserve(regions.size());
    for (vector<SceneRegion *>::iterator it = regions.begin(); it != regions.end(); it++) {
        mapping.push_back(-1);
        if (*it == NULL) continue;
        if ((*it)->empty()) {
            delete *it;
        } else {
            (*it)->_id = (int)newRegions.size();
            newRegions.push_back(*it);
            mapping.back() = (*it)->_id;
        }
    }

    vector<SceneObject *> newObjects;
    newObjects.reserve(objects.size());
    for (vector<SceneObject *>::iterator it = objects.begin(); it != objects.end(); it++) {
        if (*it == NULL) continue;
        if ((*it)->empty()) {
            delete *it;
        } else {
            (*it)->_id = (int)newObjects.size();
            newObjects.push_back(*it);
        }
    }

    vector<SceneRegionBoundary *> newRegionBoundaries;
    newRegionBoundaries.reserve(regionBoundaries.size());
    for (vector<SceneRegionBoundary *>::iterator it = regionBoundaries.begin();
         it != regionBoundaries.end(); it++) {
        if (*it == NULL) continue;
        if (((*it)->regionA == NULL) || ((*it)->regionB == NULL)) {
            SVL_ASSERT(((*it)->regionA == NULL) && ((*it)->regionB == NULL));
            delete *it;
        } else {
            (*it)->_id = (int)newRegionBoundaries.size();
            newRegionBoundaries.push_back(*it);
        }
    }

    vector<SceneObjRegBoundary *> newObjectBoundaries;
    newObjectBoundaries.reserve(objectBoundaries.size());
    for (vector<SceneObjRegBoundary *>::iterator it = objectBoundaries.begin();
         it != objectBoundaries.end(); it++) {
        if (*it == NULL) continue;
        if (((*it)->region == NULL) || ((*it)->object == NULL)) {
            delete *it;
        } else {
            (*it)->_id = (int)newObjectBoundaries.size();
            newObjectBoundaries.push_back(*it);
        }
    }

    // swap to new entities
    regions = newRegions;
    objects = newObjects;
    regionBoundaries = newRegionBoundaries;
    objectBoundaries = newObjectBoundaries;

    // renumber pixels
    int *p = (int *)CV_MAT_ELEM_PTR(*pixels, 0, 0);
    for (int i = 0; i < pixels->rows * pixels->cols; i++, p++) {
        *p = mapping[*p];
    }
}

// update region and object neighborhoods
void SceneAssignment::updateNeighborhoods()
{
    SVL_ASSERT(pixels != NULL);
    SVL_FCN_TIC;

    // clear neighbor lists
    clearRegionAndObjectBoundaries();

    // find region neighbors and compute boundary lengths
    map<pair<int, int>, int> boundaryLengths;

    const int *p = (int *)CV_MAT_ELEM_PTR(*pixels, 0, 0);
    const int *q = (int *)CV_MAT_ELEM_PTR(*pixels, 1, 0);
    for (int y = 0; y < pixels->rows; y++) {
        for (int x = 0; x < pixels->cols; x++) {
            if ((y < pixels->rows-1) && (*p != *q)) {
                pair<int, int> key(std::min(*p, *q), std::max(*p, *q));
                map<pair<int, int>, int>::iterator it = boundaryLengths.find(key);
                if (it == boundaryLengths.end()) {
                    boundaryLengths[key] = 1;
                } else {
                    it->second += 1;
                }
            }

            //if ((*p != *(p - 1)) && (*p >= 0) && (*(p - 1) >= 0)) {
            if ((x < pixels->cols-1) && (*p != *(p + 1))) {
                // TODO: check this (also add *(p - 1) != *q ?)
                pair<int, int> key(std::min(*p, *(p + 1)), std::max(*p, *(p + 1)));
                map<pair<int, int>, int>::iterator it = boundaryLengths.find(key);
                if (it == boundaryLengths.end()) {
                    boundaryLengths[key] = 1;
                } else {
                    it->second += 1;
                }
            }

            p++; q++;
        }
    }

    // update region boundaries
    vector<SceneRegionBoundary *> newRegionBoundaries;
    for (vector<SceneRegionBoundary *>::iterator it = regionBoundaries.begin();
         it != regionBoundaries.end(); it++) {
        if (!(*it)->valid()) {
            delete *it;
            continue;
        }

        pair<int, int> key = make_pair(std::min((*it)->regionA->id(), (*it)->regionB->id()),
            std::max((*it)->regionA->id(), (*it)->regionB->id()));
        map<pair<int, int>, int>::const_iterator b = boundaryLengths.find(key);
        if (b == boundaryLengths.end()) {
            delete *it;
        } else {
            (*it)->_id = (int)newRegionBoundaries.size();
            (*it)->length = b->second;
            newRegionBoundaries.push_back(*it);
            boundaryLengths.erase(b->first);
        }
    }

    for (map<pair<int, int>, int>::const_iterator it = boundaryLengths.begin();
         it != boundaryLengths.end(); it++) {
        newRegionBoundaries.push_back(new SceneRegionBoundary(regions[it->first.first],
                regions[it->first.second], (int)newRegionBoundaries.size()));
        newRegionBoundaries.back()->length = it->second;
    }

    // find object neighbors
    set<pair<int, int> > objectRegionNeighbors;
    for (vector<SceneRegionBoundary *>::iterator it = newRegionBoundaries.begin();
         it != newRegionBoundaries.end(); it++) {
        if ((*it)->regionA->object != NULL) {
            if (!(*it)->regionA->object->hasRegion((*it)->regionB)) {
                objectRegionNeighbors.insert(make_pair((*it)->regionA->object->id(),
                        (*it)->regionB->id()));
            }
        }
        if ((*it)->regionB->object != NULL) {
            if (!(*it)->regionB->object->hasRegion((*it)->regionA)) {
                objectRegionNeighbors.insert(make_pair((*it)->regionB->object->id(),
                        (*it)->regionA->id()));
            }
        }
    }

    // update object boundaries
    vector<SceneObjRegBoundary *> newObjectBoundaries;
    for (vector<SceneObjRegBoundary *>::const_iterator it = objectBoundaries.begin();
         it != objectBoundaries.end(); it++) {
        if (!(*it)->valid()) {
            delete *it;
            continue;
        }

        pair<int, int> key = make_pair((*it)->object->id(), (*it)->region->id());
        set<pair<int, int> >::const_iterator b = objectRegionNeighbors.find(key);
        if (b == objectRegionNeighbors.end()) {
            delete *it;
        } else {
            (*it)->_id = (int)newObjectBoundaries.size();
            newObjectBoundaries.push_back(*it);
            objectRegionNeighbors.erase(b);
        }
    }

    for (set<pair<int, int> >::const_iterator it = objectRegionNeighbors.begin();
         it != objectRegionNeighbors.end(); it++) {
        newObjectBoundaries.push_back(new SceneObjRegBoundary(objects[it->first],
                regions[it->second], (int)newObjectBoundaries.size()));
    }

    // update neighbor lists
    regionBoundaries = newRegionBoundaries;
    objectBoundaries = newObjectBoundaries;
    updateRegionAndObjectBoundaries();

    SVL_FCN_TOC;
}

SceneAssignment *SceneAssignment::clone() const
{
    return new SceneAssignment(*this);
}

void SceneAssignment::swap(SceneAssignment& assignment)
{
    std::swap(pixels, assignment.pixels);
    std::swap(regions, assignment.regions);
    std::swap(objects, assignment.objects);
    std::swap(regionBoundaries, assignment.regionBoundaries);
    std::swap(objectBoundaries, assignment.objectBoundaries);
    std::swap(horizon, assignment.horizon);
    std::swap(minDepth, assignment.minDepth);
    std::swap(maxDepth, assignment.maxDepth);
}

SceneAssignment& SceneAssignment::operator=(const SceneAssignment& assignment)
{
    if (&assignment != this) {
        SceneAssignment tmp(assignment);
        swap(tmp);
    }

    return *this;
}

bool SceneAssignment::operator==(const SceneAssignment& assignment)
{
    SVL_ASSERT((pixels != NULL) && (assignment.pixels != NULL));
    if ((assignment.width() != width()) || (assignment.height() != height()))
        return false;

    // TODO: modify to handle different orderings, but same configuration
    // TODO: modify to compare other variable assignments
    if (cvNorm(assignment.pixels, pixels, CV_L1) == 0)
        return false;

    return true;
}

void SceneAssignment::clearRegionAndObjectBoundaries()
{
    // regions
    for (vector<SceneRegion *>::iterator it = regions.begin(); it != regions.end(); it++) {
        (*it)->boundaries.clear();
        (*it)->objBoundaries.clear();
    }

    // objects
    for (vector<SceneObject *>::iterator it = objects.begin(); it != objects.end(); it++) {
        (*it)->boundaries.clear();
    }
}

void SceneAssignment::updateRegionAndObjectBoundaries()
{
    // region-to-region
    for (vector<SceneRegionBoundary *>::const_iterator it = regionBoundaries.begin();
         it != regionBoundaries.end(); it++) {
        (*it)->regionA->boundaries.insert(*it);
        (*it)->regionB->boundaries.insert(*it);
    }

    // object-to-region
    for (vector<SceneObjRegBoundary *>::const_iterator it = objectBoundaries.begin();
         it != objectBoundaries.end(); it++) {
        (*it)->object->boundaries.insert(*it);
        (*it)->region->objBoundaries.insert(*it);
    }
}
