/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneAssignment.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**  The SceneAssignment class holds all the scene entities and is responsible
**  for maintaining data integrity between them, i.e. making sure they point
**  to the right place.
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <list>

#include "Eigen/Core"

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"

using namespace std;
USING_PART_OF_NAMESPACE_EIGEN

// SceneAssignment class ----------------------------------------------------

class SceneAssignment {
 public:
    CvMat *pixels;                              // region assignment for pixel i
    vector<SceneRegion *> regions;              // region meta-data and random variables
    vector<SceneObject *> objects;              // object meta-data and random variables
    vector<SceneRegionBoundary *> regionBoundaries;  // boundary between two regions
    vector<SceneObjRegBoundary *> objectBoundaries;  // boundary between object and region
    int horizon;                                // row corresponding to horizon

    // depth information used by sceneReconstruct3D
    MatrixXd minDepth;             // minimum possible (unnormalized) depth for each pixel
    MatrixXd maxDepth;             // maximum possible (unnormalized) depth for each pixel
                                   // (above the horizon, this is 1 - horizon)

 public:
    SceneAssignment();
    SceneAssignment(int height, int width);
    SceneAssignment(const SceneAssignment& assignment);
    ~SceneAssignment();

    // access functions
    inline int width() const { return pixels == NULL ? 0 : pixels->cols; }
    inline int height() const { return pixels == NULL ? 0 : pixels->rows; }
    inline int numPixels() const { return pixels == NULL ? 0 : pixels->rows * pixels->cols; }
    inline int numRegions() const { return (int)regions.size(); }
    inline int numObjects() const { return (int)objects.size(); }
    inline int regionId(int x, int y) const;
    inline int objectId(int regId) const;

    // i/o
    void read(const char *filename);
    void write(const char *filename) const;

    // resize the pixel assignments and clears all regions and objects
    // TODO: make this keep current assignments (only removing regions
    // if size decreases and pixels in the region disappear).
    void resize(int height, int width);

    // clears all regions and objects
    void clear();

    // clears all cached model data in all entities
    void clearCachedModelData();

    // allocates new (empty) regions and objects
    void allocateRegions(int numRegions);
    void allocateObjects(int numObjects);

    // remap region and object ids (and remove empty regions/objects)
    void remap();
    
    // update boundaries
    void updateNeighborhoods();

    // copy the assignment or swap with another assignment
    SceneAssignment *clone() const;
    void swap(SceneAssignment& assignment);    

    SceneAssignment& operator=(const SceneAssignment& assignment);
    bool operator==(const SceneAssignment& assignment);

 protected:
    // manage boundaries and pointers
    void clearRegionAndObjectBoundaries();
    void updateRegionAndObjectBoundaries();
};

// inline functions ----------------------------------------------------------

inline int SceneAssignment::regionId(int x, int y) const
{ 
    if ((pixels == NULL) || (x < 0) || (x >= pixels->cols) || (y < 0) || (y >= pixels->rows))
        return -1;

    return CV_MAT_ELEM(*pixels, int, y, x);
}

inline int SceneAssignment::objectId(int regId) const
{
    if ((regId < 0) || (regId >= (int)regions.size()))
        return -1;

    if (regions[regId]->object == NULL)
        return -1;

    return regions[regId]->object->id();
}

