/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneBuildHOGFeatureCache.cpp
** AUTHOR(S):   Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**  Build a cache of patch responses using HOG feature.
**  Modified from buildPatchResponseCache.cpp.
*****************************************************************************/


#define NUM_THREADS 8

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sys/types.h>
#include <sstream>

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#include "win32/dirent.h"
#else
#include <dirent.h>
#endif

#include "Eigen/Core"

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlVision.h"

#include "sceneHOGFeatures.h"

using namespace std;

// Thread args
struct ProcessFileArgs {
  // idx is the file number within the directory; sample_idx is the number of
  // the output file
  // loader is a vector of NUM_THREADS svlImageLoader objects. Each thread should only use loader[tid] to avoid race conditions on loader's edge extractor object
  ProcessFileArgs(  const char *outD,
		    const vector<SceneHOGFeatures *> & ext,  unsigned idx, const vector<svlImageLoader *> & loader,
                    unsigned sample_idx, bool bRegionBasedArg) :
      outputDir(outD),  
      extractor(ext), idx(idx), loader(loader), sample_idx(sample_idx), bRegionBased(bRegionBasedArg) {}



  const char *outputDir;
    
  const vector<SceneHOGFeatures *> & extractor;
  unsigned idx;
  const vector<svlImageLoader *> & loader;
  unsigned sample_idx;

  bool bRegionBased;
};

// Thread function
void *processFile(void *voidArgs, unsigned tid) {


  //Copy arguments out of the helper struct
  ProcessFileArgs *args = (ProcessFileArgs*) voidArgs;
  const char *outputDir = args->outputDir;
  const vector<SceneHOGFeatures *> &  extractor = args->extractor;
  unsigned idx = args->idx;
  unsigned sample_idx = args->sample_idx;
  const vector<svlImageLoader *> & loader = args->loader;
  bool bRegionBased = args->bRegionBased;

  vector<IplImage *> images;

  if (!loader[tid]->getFrame(idx, images))
    return NULL;
            

  //Discard nearly uniform examples
  /*
  for (unsigned i = 0; i < images.size(); i++)
    {
      CvScalar mean;
      CvScalar sdev;
      
      cvAvgSdv(images[i], &mean, &sdev);

      SVL_LOG(SVL_LOG_DEBUG, "standard deviation: "<<sdev.val[0]);

      if ( (images[i]->depth==IPL_DEPTH_8U && sdev.val[0] < 5.0) || (images[i]->depth==IPL_DEPTH_32F && sdev.val[0] < 0.01)) //TODO-- figure out why svlPatchDefinition::patchValueHelper sometimes has windowNorm come out <= 0 even though patch is not uniform. Is it just numerical error?
	{
	  //stringstream filepath;
	  //filepath << "output_" << rand() << ".jpg";
	  //cvSaveImage(filepath.str().c_str(), images[i]);
	  return NULL;
	}
	}*/

  // process image
  VectorXd f;
  extractor[tid]->load(images[0]);
  extractor[tid]->prePocessing();
  
  if (bRegionBased == false) {
      CvRect r;
      r.x = 0;
      r.y = 0;
      r.width = images[0]->width;
      r.height = images[0]->height;
      f = extractor[tid]->extractFeatures(r);
  } else { // extract region-based features
      // creat and read region mask
      CvMat* regionMask = NULL;
      regionMask = cvCreateMat(images[0]->height, images[0]->width, CV_32SC1);
      SVL_ASSERT(regionMask != NULL);
      string regionMaskFilename = loader[tid]->getFrameBasename(idx) + ".mask.txt";
      bool bSuccess = readMatrix(regionMask, regionMaskFilename.c_str());
      SVL_ASSERT(bSuccess == true);

      // extract feature
      f = extractor[tid]->extractFeatures(regionMask);
      cvReleaseMat(&regionMask);
  }

  vector<double> v(f.rows(), 0);
  Eigen::Map<VectorXd>(&v[0], v.size()) = f;

  for (unsigned i = 0; i < v.size(); i++)
    {
      if (v[i] > 10e30)
	{
	  cout << v[i] << endl;
	  cout << "sceneBuildHOGFeatureCache override" << endl;
	  v[i] = 0;
	}

      if (v[i] < -10e30)
	{
	  cout << v[i] << endl;
	  cout << "sceneBuildHOGFeatureCache override" << endl;
	  v[i] = 0;
	}
    }
 
  svlWriteExampleToCache(outputDir, sample_idx, v);
                       
  // Free memory
  for (unsigned i = 0; i < images.size(); i++)
    cvReleaseImage(&(images[i]));

  delete args;

  return NULL;
}

void writeSVMLightExampleToCache(std::ostream& ofs, vector<double>& v, bool bPos)
{
    if (bPos)
        ofs << "+1 ";
    else
        ofs << "-1 ";
    for (int i = 0; i < (int)v.size(); i++) {
        if (v[i] >= 1e-5) 
            ofs << i+1 << ":" << v[i] << " ";
    }
    ofs << endl;
}

void writeSVMLightExampleToBinaryCache(std::ostream& ofs, vector<double>& v, bool bPos)
{
    int label = bPos ? 1 : -1;
    ofs.write((const char*)&label, sizeof(int));
    for (unsigned i = 0; i < v.size(); i++) {
      ofs.write((const char*)&v[i], sizeof(double));
    }
}

void processSVMLightFile(std::ostream& ofs, SceneHOGFeatures * extractor,
    unsigned idx, svlImageLoader& loader, bool bRegionBased, bool bPos) {
    vector<IplImage *> images;
    if (!loader.getFrame(idx, images))
        return ;
    // process image
    VectorXd f;
    extractor->load(images[0]);
    extractor->prePocessing();
    if (bRegionBased == false) {
        CvRect r = cvRect(0, 0, images[0]->width, images[0]->height);
        f = extractor->extractFeatures(r);
    } else { // extract region-based features
      // creat and read region mask
        CvMat* regionMask = NULL;
        regionMask = cvCreateMat(images[0]->height, images[0]->width, CV_32SC1);
        SVL_ASSERT(regionMask != NULL);
        string regionMaskFilename = loader.getFrameBasename(idx) + ".mask.txt";
        bool bSuccess = readMatrix(regionMask, regionMaskFilename.c_str());
        SVL_ASSERT(bSuccess == true);
        
        // extract feature
        f = extractor->extractFeatures(regionMask);
        cvReleaseMat(&regionMask);
    }
    
    vector<double> v(f.rows(), 0);
    Eigen::Map<VectorXd>(&v[0], v.size()) = f;
    
    // write feature vector
    if (ofs.fail()) {
        SVL_LOG(SVL_LOG_WARNING, "...failed on index " << idx);
    } else {    
        writeSVMLightExampleToCache(ofs, v, bPos);
    }
    // Free memory
    svlReleaseImages(images);
}

void processSVMLightBinaryFile(std::ostream& ofs, SceneHOGFeatures * extractor,
    unsigned idx, svlImageLoader& loader, bool bRegionBased, bool bPos) {
    vector<IplImage *> images;
    if (!loader.getFrame(idx, images))
        return ;
    // process image
    VectorXd f;
    extractor->load(images[0]);
    extractor->prePocessing();
    if (bRegionBased == false) {
        CvRect r = cvRect(0, 0, images[0]->width, images[0]->height);
        f = extractor->extractFeatures(r);
    } else { // extract region-based features
      // creat and read region mask
        CvMat* regionMask = NULL;
        regionMask = cvCreateMat(images[0]->height, images[0]->width, CV_32SC1);
        SVL_ASSERT(regionMask != NULL);
        string regionMaskFilename = loader.getFrameBasename(idx) + ".mask.txt";
        bool bSuccess = readMatrix(regionMask, regionMaskFilename.c_str());
        SVL_ASSERT(bSuccess == true);
        
        // extract feature
        f = extractor->extractFeatures(regionMask);
        cvReleaseMat(&regionMask);
    }
    
    vector<double> v(f.rows(), 0);
    Eigen::Map<VectorXd>(&v[0], v.size()) = f;
    
    // write feature vector
    if (ofs.fail()) {
        SVL_LOG(SVL_LOG_WARNING, "...failed on index " << idx);
    } else {    
        writeSVMLightExampleToBinaryCache(ofs, v, bPos);
    }
    // Free memory
    svlReleaseImages(images);
}

void processBinaryFile(std::ostream& ofs, SceneHOGFeatures * extractor,
    unsigned idx, svlImageLoader& loader, bool bRegionBased) {
  vector<IplImage *> images;
  if (!loader.getFrame(idx, images))
    return ;
  // process image
  VectorXd f;
  extractor->load(images[0]);
  extractor->prePocessing();
  if (bRegionBased == false) {
      CvRect r;
      r.x = 0;
      r.y = 0;
      r.width = images[0]->width;
      r.height = images[0]->height;
      f = extractor->extractFeatures(r);
  } else { // extract region-based features
      // creat and read region mask
      CvMat* regionMask = NULL;
      regionMask = cvCreateMat(images[0]->height, images[0]->width, CV_32SC1);
      SVL_ASSERT(regionMask != NULL);
      string regionMaskFilename = loader.getFrameBasename(idx) + ".mask.txt";
      bool bSuccess = readMatrix(regionMask, regionMaskFilename.c_str());
      SVL_ASSERT(bSuccess == true);

      // extract feature
      f = extractor->extractFeatures(regionMask);
      cvReleaseMat(&regionMask);
  }
  
  vector<double> v(f.rows(), 0);
  Eigen::Map<VectorXd>(&v[0], v.size()) = f;

  // write feature vector
  if (ofs.fail()) {
    SVL_LOG(SVL_LOG_WARNING, "...failed on index " << idx);
  } else {    
    svlWriteExampleToBinaryCache(ofs, v);
  }
            
  // Free memory
  svlReleaseImages(images);
}


void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./sceneBuildHOGFeatureCache [OPTIONS] (<image dir> | <image sequence>) <output dir>" << endl;
    cerr << "OPTIONS:" << endl
         << "  -width <num>         :: window width (default: 64)" << endl
         << "  -height <num>        :: window height (default: 128)" << endl
         << "  -numCellRow <num>    :: # of cell rows (default: 8)" << endl
         << "  -numCellCol <num>    :: # of cell cols (default: 16)" << endl
	 << "  -maxImages <num>     :: maximum training images to cache (default: 10000)" << endl
	 << "  -binary              :: creates binary cache file named <output dir>.bin" << endl
         << "  -svmLight            :: use svm_light format (default: false) " << endl
         << "  -bPos                :: processing positive samples (default: false) " << endl
 	 << "  -regionBased         :: region-based features (default: false)" << endl
	 << "  -parallel <id> <ct>  :: used for distributed jobs on the cluster." << endl
	 << "  -startIndex          :: starts numbering from this index instead of from 0" << endl
	 << "  -augment             :: augments the existing cache dir; overrides -startIndex " << endl
	 << "                        (same as using -startIndex [curr # files in cache dir])" << endl
	 << SVL_STANDARD_OPTIONS_USAGE << endl
	 << "  Note: both -augment and -startIndex make sure the TOTAL number of files in the " << endl
	 << "        cache dir doesn't exceed -maxImages, not just the number of files added" << endl
	 << "        in the current run." << endl
	 << endl;
}


int main(int argc, char *argv[])
{
  // svlAutoregsvlFeatureExtractorFileCheckersvlFeatureExtractor.debug();
  captureOpenCVerrors();
  svlCodeProfiler::tic(svlCodeProfiler::getHandle("buildPatchResponseCache"));

  // read commandline parameters
  const int NUM_REQUIRED_PARAMETERS = 2;
  
  int width = 64;
  int height = 128;
  int numCellRow = 8;
  int numCellCol = 16;
  int maxImages = 10000;
  bool binary = false;
  bool bSVMLight = false;
  bool bPos = false;
  bool bRegionBased = false;
  int  parallelId = 0;
  int  parallelCount = 0;
  int startIndex = 0;
  bool augment = false;

  SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
    SVL_CMDLINE_INT_OPTION("-width", width)
    SVL_CMDLINE_INT_OPTION("-height", height)
    SVL_CMDLINE_INT_OPTION("-numCellRow", numCellRow)
    SVL_CMDLINE_INT_OPTION("-numCellCol", numCellCol)
    SVL_CMDLINE_INT_OPTION("-maxImages", maxImages)
    SVL_CMDLINE_BOOL_OPTION("-regionBased", bRegionBased)
    SVL_CMDLINE_BOOL_OPTION("-binary", binary)
    SVL_CMDLINE_BOOL_OPTION("-svmLight", bSVMLight)
    SVL_CMDLINE_BOOL_OPTION("-bPos", bPos)
    SVL_CMDLINE_OPTION_BEGIN("-parallel", ptr)
      parallelId = atoi(ptr[0]);
      parallelCount = atoi(ptr[1]);
      if (parallelId < 0 || parallelId >= parallelCount) {
        SVL_LOG(SVL_LOG_FATAL, "illegal id number " << parallelId);
      } else if (parallelCount < 0) {
        SVL_LOG(SVL_LOG_FATAL, "illegal job count " << parallelCount);
      }
    SVL_CMDLINE_OPTION_END(2)
    SVL_CMDLINE_INT_OPTION("startIndex", startIndex)
    SVL_CMDLINE_BOOL_OPTION("augment", augment)
  SVL_END_CMDLINE_PROCESSING(usage());
    
  if (SVL_CMDLINE_ARGC != NUM_REQUIRED_PARAMETERS) {
    usage();
    return -1;
  }

  const char *imageDir = SVL_CMDLINE_ARGV[0];
  const char *outputDir = SVL_CMDLINE_ARGV[1];

  svlImageLoader loader(false, true);
  //  svlImageLoader loader;

  // read dictionary
  //  const svlFeatureExtractor *  dictionary = svlFeatureExtractorFactory().load(extractorFilename);
  SceneHOGFeatures* extractor = new SceneHOGFeatures();
  
  extractor->setWinWidth(width);
  extractor->setWinHeight(height);
  extractor->setNumCellRows(numCellRow);
  extractor->setNumCellCols(numCellCol);

  // will be compared to the maxImages parameter
  int numImages = 0;
  if (startIndex > 0 || augment)
    numImages = svlDirectoryListing(outputDir).size();
  
  if (augment)
    startIndex = numImages;

  if (strExtension(imageDir).compare("xml") == 0) {  
    loader.readImageSeq(imageDir);
  } else {
    loader.readDir(imageDir);
  }

  if (bRegionBased == false) {
      loader.setResizeSize(width,height);
  }
  
  if (!binary) {
      // ------ Output ASCII cache files ------
      if (bSVMLight) {
          // erases the possible "/" at the end of the directory name
          string outputDirStr = string(outputDir);
          int pos = outputDirStr.find_last_not_of("/");
          outputDirStr = outputDirStr.substr(0, pos+1);
      
          std::string jobId = "";
          if (parallelCount > 0) {
              jobId = "." + toString(parallelId);
          }
          std::string fileName = outputDirStr+jobId+".dat";
          //std::fstream ofs(fileName.c_str(), std::ios::out|std::ios::binary);
          std::ofstream ofs(fileName.c_str());
          SVL_ASSERT_MSG(ofs.fail() == false, "cannot open: " << fileName);
          
          for (unsigned i = 0; i < loader.numFrames() && numImages < maxImages; i++, numImages++) {
              if (parallelCount > 0) {
                  if ((i % unsigned(parallelCount)) != unsigned(parallelId)) continue ;
              }
              processSVMLightFile(ofs, extractor, i, loader, bRegionBased, bPos);
          }        
          
          ofs.close();
      } else {
          // TODO: how many threads?
          svlThreadPool threadPool(NUM_THREADS);
          threadPool.start();
          
          vector<svlImageLoader *> loaderSet;
          vector<SceneHOGFeatures *> extractorSet;
          for (unsigned i = 0; i < NUM_THREADS; i++) {
              loaderSet.push_back(new svlImageLoader(loader));
              SceneHOGFeatures* clone = new SceneHOGFeatures();
              assert(clone);
              //clone->setWinWidth(width);
              //clone->setWinHeight(height);
              clone->setNumCellCols(numCellCol);
              clone->setNumCellRows(numCellRow);
              extractorSet.push_back(clone);
          }
          
          // i is the index of the image we're processing, numImages is the number of
          // images in the cache directory currently, i+startIndex is the name of the file
          // we're about to output
          for (unsigned i = 0; i < loader.numFrames() && numImages < maxImages; i++, numImages++) {
              threadPool.addJob(processFile,
                  new ProcessFileArgs(outputDir, extractorSet, i, loaderSet, i + startIndex, bRegionBased));
          }
          // TODO: make this contingent on success?
          threadPool.finish();
          for (unsigned i = 0; i < loaderSet.size(); i++)
              delete loaderSet[i];
          
          for (unsigned i = 0; i < extractorSet.size(); i++)
              delete extractorSet[i];
      }
  } else { // binary
      // ------ Output binary cache file ------
      if (augment) {
          SVL_LOG(SVL_LOG_FATAL, "-augment flag and -binary flag are incompatible at present.");
      }
      
      // erases the possible "/" at the end of the directory name
      string outputDirStr = string(outputDir);
      int pos = outputDirStr.find_last_not_of("/");
      outputDirStr = outputDirStr.substr(0, pos+1);
      
      std::string jobId = "";
      if (parallelCount > 0) {
          jobId = "." + toString(parallelId);
      }
      std::string fileName = outputDirStr+jobId+".bin";
      std::fstream ofs(fileName.c_str(), std::ios::out|std::ios::binary);
      SVL_ASSERT_MSG(ofs.fail() == false, "fileName: " << fileName);
      if (bSVMLight) {          
          int version = 200;
          int data_typeid = 5;   // feature value type: double
          int target_typeid = 3; // target label type: int
          int num_feature = std::min((int)loader.numFrames(), maxImages - numImages);
          // get the feature length
          vector<IplImage *> images;
          loader.getFrame(0, images);
          VectorXd f;
          extractor->load(images[0]);
          extractor->prePocessing();
          CvRect r = cvRect(0, 0, images[0]->width, images[0]->height);
          f = extractor->extractFeatures(r);
          int feature_length = f.rows();
          
          ofs.write((const char*)&version, sizeof(int));
          ofs.write((const char*)&data_typeid, sizeof(int));
          ofs.write((const char*)&target_typeid, sizeof(int));
          ofs.write((const char*)&num_feature, sizeof(int));
          ofs.write((const char*)&feature_length, sizeof(int));

          for (unsigned i = 0; i < loader.numFrames() && numImages < maxImages; i++, numImages++) {
              if (parallelCount > 0) {
                  if ((i % unsigned(parallelCount)) != unsigned(parallelId)) continue ;
              }
              processSVMLightBinaryFile(ofs, extractor, i, loader, bRegionBased, bPos);
          }
      } else {
          for (unsigned i = 0; i < loader.numFrames() && numImages < maxImages; i++, numImages++) {
              if (parallelCount > 0) {
                  if ((i % unsigned(parallelCount)) != unsigned(parallelId)) continue ;
              }
              processBinaryFile(ofs, extractor, i, loader, bRegionBased);
          }
      }
  }

  delete extractor;
  svlCodeProfiler::toc(svlCodeProfiler::getHandle("buildPatchResponseCache"));
  svlCodeProfiler::print(cerr);
  
  return 0;
}
