/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneCacheErodedObjectFeatures.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**   Outputs detection candidates file and corresponding features file from
**   object dictionary moves and object detector.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <limits>
#include <set>
#include <algorithm>
#include <iterator>
#include <sys/time.h>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneObjectModel.h"
#include "sceneVisualization.h"
#include "sceneSegmentDict.h"
#include "sceneObjDetDict.h"
#include "sceneRegionFeatures.h"

#define WINDOW_NAME "sceneCacheErodedObjectFeatures"

using namespace std;

void viewProgress(SceneInstance& instance, const SceneObjDetDictEntry& move);
vector<double> computeObjectFeatureVector(const SceneInstance *instance,
    const SceneObjDetDictEntry *move, const ScenePatchObjectDetector *detector);

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneCacheErodedObjectFeatures [OPTIONS] <objectName> <imgList> <groundtruth XML>\n";
    cerr << "OPTIONS:\n"
         << "  -modelDir <dir>   :: object detector model directory (loads <dir>/<objectName>.dict and <dir>/<objectName>.model)\n"
         << "  -inMoves <dir>    :: moves directory (default: moves)\n"
         << "  -o <stem>         :: output filestem (appends <objectName>.cachedFeatures.bin and <objectName>.cachedDetections.xml)\n"
         << "  -noFeatures       :: no features (only original score)\n"
         << "  -x                :: visualize output\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // read commandline parameters
    const char *modelDir = ".";
    const char *inMovesDir = "moves";
    const char *outStem = "";
    bool bNoFeatures = false;

    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-modelDir", modelDir)
        SVL_CMDLINE_STR_OPTION("-inMoves", inMovesDir)
        SVL_CMDLINE_STR_OPTION("-o", outStem)
        SVL_CMDLINE_BOOL_OPTION("-noFeatures", bNoFeatures)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 3) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    string objectName = string(SVL_CMDLINE_ARGV[0]);
    const char *imageList = SVL_CMDLINE_ARGV[1];
    const char *gtObjectXML = SVL_CMDLINE_ARGV[2];

    // read list of images
    vector<string> baseNames;
    SVL_LOG(SVL_LOG_MESSAGE, "Reading image list from " << imageList << "...");
    baseNames = svlReadFile(imageList);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");

    // read ground-truth objects
    SVL_LOG(SVL_LOG_MESSAGE, "Reading groundtruth objects from " << gtObjectXML << "...");
    svlObject2dSequence gtObjects;
    gtObjects.read(gtObjectXML);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << gtObjects.size() << " image frames");

    // read object detection models
    string dictFilename = string(modelDir) + string("/") + objectName + string(".dict");
    string modelFilename = string(modelDir) + string("/") + objectName + string(".model");
    ScenePatchObjectDetector detector(dictFilename.c_str(), modelFilename.c_str());
    detector.setObjectName(objectName.c_str());

    // feature extraction
    svlObject2dSequence candidateDetections;
    SceneFeatureCache candidateData;

    // process images
    SVL_LOG(SVL_LOG_MESSAGE, "Processing images...");
    int hProcessImage = svlCodeProfiler::getHandle("processImage");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        svlCodeProfiler::tic(hProcessImage);

        // find groundtruth objects for this image
        svlObject2dFrame gtFrame = gtObjects[baseNames[i]];
        removeNonMatchingObjects(gtFrame, objectName.c_str());
        SVL_LOG(SVL_LOG_VERBOSE, "..." << baseNames[i] << " has " << gtFrame.size()
            << " " << objectName << " objects");

        // load image
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        IplImage *img = cvLoadImage(imgFilename.c_str(), CV_LOAD_IMAGE_COLOR);
        SVL_ASSERT_MSG(img != NULL, "path: " << imgFilename);
        SVL_LOG(SVL_LOG_MESSAGE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*img));

        // initialize scene
        SceneInstance instance(img, baseNames[i].c_str());

        // load object candidate moves
        SceneObjDetDict objCandidateMoves;
        string movesFilename = string(inMovesDir) + string("/") + objectName + 
            string(".") + instance.name() + string(".moves.bin");
        objCandidateMoves.load(movesFilename.c_str());
        SVL_LOG(SVL_LOG_VERBOSE, "..." << objCandidateMoves.numMoves() << " object candidate moves loaded");

        // iterate through candidate moves
        svlObject2dFrame candidateFrame;
        SVL_LOG(SVL_LOG_MESSAGE, "attempting " << objCandidateMoves.numMoves() << " object candidate moves...");
        for (int j = 0; j < objCandidateMoves.numMoves(); j++) {
            // get pixel mask for the move
            if ((objCandidateMoves[j].extent.width == 0) || (objCandidateMoves[j].extent.height == 0)) {
                SVL_LOG(SVL_LOG_DEBUG, "...object candidate move " << j << " is empty");
                continue;
            }

            CvMat *mask = objCandidateMoves.mask(j);
            if (!svlIsConnectedComponent(mask, true, 1)) {
                SVL_LOG(SVL_LOG_DEBUG, "...object candidate move " << j << " is not connected");
                cvReleaseMat(&mask);
                continue;
            }
            cvReleaseMat(&mask);

            // visualize move
            if (bVisualize) {
                viewProgress(instance, objCandidateMoves[j]);
            }

            // extract features and class label
            if (bNoFeatures) {
                candidateData.features.push_back(vector<double>(1, log(objCandidateMoves[j].originalScore)));
            } else {
                candidateData.features.push_back(computeObjectFeatureVector(&instance, &objCandidateMoves[j], &detector));
            }
            SVL_LOG(SVL_LOG_DEBUG, toString(candidateData.features.back()));

            double ao = 0.0;
            for (svlObject2dFrame::const_iterator o = gtFrame.begin(); o != gtFrame.end(); o++) {
                ao = std::max(ao, o->areaOverlap(svlObject2d(objCandidateMoves[j].extent)));
            }
            candidateData.labels.push_back(ao >= 0.5 ? 1 : 0);

            // add object to frame
            svlObject2d o(objCandidateMoves[j].extent);
            o.name = objectName;
            o.pr = ao;
            o.index = candidateData.size() - 1;
            candidateFrame.push_back(o);
        }

        // add to candidate detections
        candidateDetections[baseNames[i]] = candidateFrame;

        // free memory
        cvReleaseImage(&img);

        SVL_LOG(SVL_LOG_VERBOSE, "..." << candidateData.size() << " feature vectors of size "
            << candidateData.numFeatures() << " accumulated ("
            << candidateData.count(1) << " positive)");
        svlCodeProfiler::toc(hProcessImage);
    }

    candidateData.weights.resize(candidateData.size(), 1.0);
    
    // write accumulated data
    string filename = string(outStem) + objectName + string(".cachedFeatures.bin");
    candidateData.write(filename.c_str());

    filename = string(outStem) + objectName + string(".cachedDetections.xml");
    candidateDetections.write(filename.c_str());

    // free memory
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

// visualization functions -------------------------------------------------------------

void viewProgress(SceneInstance& instance, const SceneObjDetDictEntry& move)
{
    IplImage *canvas = visualize(&instance);

    cvRectangle(canvas, cvPoint(move.originalExtent.x, move.originalExtent.y),
        cvPoint(move.originalExtent.x + move.originalExtent.width, move.originalExtent.y + move.originalExtent.height),
        CV_RGB(255, 0, 0), 2);

    cvRectangle(canvas, cvPoint(move.extent.x, move.extent.y),
        cvPoint(move.extent.x + move.extent.width, move.extent.y + move.extent.height),
        CV_RGB(0, 255, 0), 2);

    resizeInPlace(&canvas, 240, 240 * canvas->width / canvas->height);

    cvNamedWindow(WINDOW_NAME, 1);
    cvShowImage(WINDOW_NAME, canvas);
    int ch = cvWaitKey(100);
    if ((char)ch == 'q') {
        if (svlLogger::getLogLevel() >= SVL_LOG_DEBUG) {
            instance.printScene();
        }
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
        svlCodeProfiler::print(cerr);
        exit(0);
    }

    if ((char)ch == 'p') {
        cvWaitKey(-1);
    }
    cvReleaseImage(&canvas);
}

// feature extraction functions --------------------------------------------------------

vector<double> computeObjectFeatureVector(const SceneInstance *instance,
    const SceneObjDetDictEntry *move, const ScenePatchObjectDetector *detector)
{
    SVL_ASSERT((instance != NULL) && (move != NULL) && (detector != NULL));
    SVL_FCN_TIC;

    vector<double> f;
    f.reserve(2 + 34 + 3 + 1);

    CvRect r = move->extent;

    CvRect e = cvRect(instance->width(), instance->height(), 0, 0);
    list<CvPoint> p;
    CvMat *m = cvCreateMat(instance->height(), instance->width(), CV_32SC1);
    move->decode(m);
    const int *q = (const int *)CV_MAT_ELEM_PTR(*m, 0, 0);
    for (int y = 0; y < instance->height(); y++) {
        for (int x = 0; x < instance->width(); x++, q++) {
            if (*q != 0) {
                p.push_back(cvPoint(x, y));

                e.x = std::min(x, e.x);
                e.y = std::min(y, e.y);
                e.width = std::max(x, e.width);
                e.height = std::max(y, e.height);
            }
        }
    }
    convertInPlace(&m, CV_8UC1);

    e.width -= e.x - 1;
    e.height -= e.y - 1;

    SVL_LOG(SVL_LOG_DEBUG, "...features for candidate window " << toString(r) << "/" << toString(e)
        << " with " << p.size() << " pixels");

    if ((r.x != e.x) || (r.y != e.y) || (r.width != e.width) || (r.height != e.height)) {
        SVL_LOG(SVL_LOG_ERROR, "extent mismatch for instance " << instance->name() 
            << ": " << toString(r) << " != " << toString(e));
        r = e;
    }

    // object detector features
    double softWindowScore = detector->evaluateRegion(instance->greyImage(), p);
    double fullWindowScore = detector->evaluateWindow(instance->greyImage(), r);
    f.push_back(softWindowScore);
    f.push_back(fullWindowScore);

    // add appearance features
#if 0
    svlSufficientStats stats(instance->numFilters(), SVL_PSS_DIAG);
    for (list<CvPoint>::const_iterator ip = p.begin(); ip != p.end(); ip++) {
        stats.accumulate(instance->pixelFilterFeatures(ip->x, ip->y));
    }

    for (int i = 0; i < instance->numFilters(); i++) {
        f.push_back(stats.sum(i) / stats.count());
        f.push_back(stats.sum2(i, i) / stats.count() - f.back() * f.back());
    }
#else
    vector<double> v = instance->regionFilterMean(m);
    f.insert(f.end(), v.begin(), v.end());
    v = instance->regionFilterVariance(m);
    f.insert(f.end(), v.begin(), v.end());
#endif

    // appearance features for entire image
    v = instance->regionFilterMean();
    f.insert(f.end(), v.begin(), v.end());
    v = instance->regionFilterVariance();
    f.insert(f.end(), v.begin(), v.end());
    
    // add "shape" and location features
    f.push_back((double)r.width / instance->width());
    f.push_back((double)r.height / instance->height());
    f.push_back((double)p.size() / r.width * r.height);
    f.push_back((double)r.height / (double)r.width);
    f.push_back((double)(r.x + r.width) / (2.0 * instance->width()));
    f.push_back((double)(r.y + r.height) / (2.0 * instance->height()));

    int nTop = 0;
    int nBottom = 0;
    int nLeft = 0;
    int nRight = 0;
    for (list<CvPoint>::const_iterator ip = p.begin(); ip != p.end(); ip++) {
        if (ip->x == r.x) nLeft += 1;
        if (ip->x == r.x + r.width - 1) nRight += 1;
        if (ip->y == r.y) nTop += 1;
        if (ip->y == r.y + r.height - 1) nBottom += 1;        
    }
    f.push_back((double)nTop / r.height);
    f.push_back((double)nBottom / r.height);
    f.push_back((double)nLeft / r.width);
    f.push_back((double)nRight / r.width);

    // horizon?

    // add bias term
    f.push_back(1.0);

    cvReleaseMat(&m);

    SVL_FCN_TOC;
    return f;
}
