/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneCacheObjectMoves.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code caching object detection moves.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneVisualization.h"
#include "sceneSegmentDict.h"
#include "sceneObjDetDict.h"

#define WINDOW_NAME "sceneCacheObjectMoves"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneCacheObjectMoves [OPTIONS] <detectionFile> <imgList>\n";
    cerr << "OPTIONS:\n"
         << "  -d <dir>         :: output directory (must exist for writing)\n"
         << "  -o <stem>        :: output filestem\n"
         << "  -nonMax          :: apply non-maximal neighborhood suppression\n"
         << "  -maxMoves <n>    :: maximum candidate moves per image\n"
         << "  -x               :: visualize moves\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // read commandline parameters
    const char *outputDir = NULL;
    const char *outputStem = NULL;
    bool bNonMaxSuppression = false;
    int maxMovesPerImage = SVL_INT_MAX;
    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-d", outputDir)
        SVL_CMDLINE_STR_OPTION("-o", outputStem)
        SVL_CMDLINE_BOOL_OPTION("-nonMax", bNonMaxSuppression)
        SVL_CMDLINE_INT_OPTION("-maxMoves", maxMovesPerImage)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)        
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 2) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *detectionsFile = SVL_CMDLINE_ARGV[0];
    const char *imgList = SVL_CMDLINE_ARGV[1];

    // read detections and image list
    SVL_LOG(SVL_LOG_MESSAGE, "Reading detections from " << detectionsFile << "...");
    svlObject2dSequence detections;
    detections.read(detectionsFile);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << detections.size() << " detection frames");

    if (bNonMaxSuppression) {
        SVL_LOG(SVL_LOG_MESSAGE, "...applying non-maximal neighborhood suppression");
        int total = countObjects(detections);
        int removed = nonMaximalSuppression(detections, 0.25, 0.25, 0.64);
        SVL_LOG(SVL_LOG_VERBOSE, removed << " out of " << total << " non-maximal objects suppressed");        
    }

    vector<string> baseNames;
    if (svlFileExists(imgList)) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading image list from " << imgList << "...");
        baseNames = svlReadFile(imgList);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");
    } else {
        baseNames.push_back(string(imgList));
    }

    // iterate through images
    SVL_LOG(SVL_LOG_MESSAGE, "Creating moves...");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        // initialize scene
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        SceneInstance instance(imgFilename.c_str(), baseNames[i].c_str());
        //instance.loadSegmentations(true);
        SVL_LOG(SVL_LOG_VERBOSE, "..." << baseNames[i] << ": " << toString(*instance.image()));
        
        // create object moves
        SceneObjDetDict moveDict;
        svlObject2dFrame frame = detections[baseNames[i]];
        SVL_LOG(SVL_LOG_VERBOSE, "...image has " << frame.size() << " total candidates");
        removeOutsideObjects(frame, cvRect(0, 0, instance.width(), instance.height()));
        keepMostProbableObjects(frame, maxMovesPerImage);        
        moveDict.create(instance.width(), instance.height(), frame);

        SVL_LOG(SVL_LOG_VERBOSE, "...created " << moveDict.numMoves() 
            << " moves for " << baseNames[i]);

#if 0
        // debugging
        CvMat *mask = cvCreateMat(instance.height(), instance.width(), CV_32SC1); 
        for (int k = 0; k < moveDict.numMoves(); k++) {
            moveDict[k].decode(mask);
            const int *p = (const int *)CV_MAT_ELEM_PTR(*mask, 0, 0);
            for (int y = 0; y < instance.height(); y++) {
                for (int x = 0; x < instance.width(); x++, p++) {
                    bool b = (y >= moveDict[k].originalExtent.y) &&
                        (y < moveDict[k].originalExtent.y + moveDict[k].originalExtent.height) &&
                        (x >= moveDict[k].originalExtent.x) &&
                        (x < moveDict[k].originalExtent.x + moveDict[k].originalExtent.width);
                    SVL_ASSERT(b == (*p == 1));
                }
            }
        }
        cvReleaseMat(&mask);
#endif

        // visualize
        if (bVisualize) {
            cvNamedWindow(WINDOW_NAME, 1);
            for (int k = 0; k < moveDict.numMoves(); k++) {
                IplImage *canvas = cvCloneImage(instance.image());

                CvMat *mask = moveDict.mask(k);
                cvConvertScale(mask, mask, -1, 1);
                cvSet(canvas, cvScalar(255, 255, 255), mask);

                CvRect r = moveDict[k].extent;
                cvRectangle(canvas, cvPoint(r.x, r.y), cvPoint(r.x + r.width, r.y + r.height),
                    CV_RGB(255, 255, 255), 3);
                cvRectangle(canvas, cvPoint(r.x, r.y), cvPoint(r.x + r.width, r.y + r.height),
                    CV_RGB(0, 255, 0), 1);

                cvShowImage(WINDOW_NAME, canvas);
                cvWaitKey(10);
                cvReleaseImage(&canvas);
            }
        }

        // save
        if (outputDir != NULL) {
            string filestem = string(outputDir) + string("/");
            if (outputStem != NULL) {
                filestem += string(outputStem) + string(".");
            }
            filestem += baseNames[i] + string(".moves.bin");
            moveDict.save(filestem.c_str());
        }
    }
    
    cvDestroyAllWindows();

    // print profile
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

