/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneCalibrateEnergies.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <limits>
#include <set>
#include <algorithm>
#include <iterator>
#include <sys/time.h>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneObjectModel.h"
#include "sceneVisualization.h"
#include "sceneSegmentDict.h"
#include "sceneObjDetDict.h"
#include "sceneRegionFeatures.h"
#include "sceneDataLoader.h"

#define WINDOW_NAME "sceneCalibrateEnergies"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneCalibrateEnergies [OPTIONS] <energyData> <detectionCandidates> <groundtruthObjects>\n";
    cerr << "OPTIONS:\n"
         << "  -object <name>    :: object name\n"
         << "  -o <filename>     :: output detections\n"
         << "  -maxPerImage <n>  :: maximum number of detections per image\n" 
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    // read commandline parameters
    const char *objectName = NULL;
    const char *outputFilename = NULL;
    int maxDetectionsPerImage = -1;
    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-object", objectName)
        SVL_CMDLINE_STR_OPTION("-o", outputFilename)
        SVL_CMDLINE_INT_OPTION("-maxPerImage", maxDetectionsPerImage)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 3) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *energyDataFile = SVL_CMDLINE_ARGV[0];
    const char *detectionsFile = SVL_CMDLINE_ARGV[1];
    const char *gtObjectXML = SVL_CMDLINE_ARGV[2];

    // read data
    SVL_LOG(SVL_LOG_MESSAGE, "Reading energies from " << energyDataFile << "...");
    SceneFeatureCache data;
    data.read(energyDataFile);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << data.size() << " samples");

    SVL_LOG(SVL_LOG_MESSAGE, "Reading candidate objects from " << detectionsFile << "...");
    svlObject2dSequence detections;
    detections.read(detectionsFile);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << detections.size() << " image frames");

    SVL_LOG(SVL_LOG_MESSAGE, "Reading groundtruth objects from " << gtObjectXML << "...");
    svlObject2dSequence gtObjects;
    gtObjects.read(gtObjectXML);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << gtObjects.size() << " image frames");

    // assign labels
    for (svlObject2dSequence::const_iterator it = detections.begin();
         it != detections.end(); it++) {

        svlObject2dFrame gtFrame = gtObjects[it->first];
        removeNonMatchingObjects(gtFrame, objectName);

        for (svlObject2dFrame::const_iterator jt = it->second.begin(); jt != it->second.end(); jt++) {

            double ao = 0.0;
            for (svlObject2dFrame::const_iterator o = gtFrame.begin(); o != gtFrame.end(); o++) {
                ao = std::max(ao, o->areaOverlap(*jt));
            }

            data.labels[jt->index] = (ao >= 0.5 ? 1 : 0);
            if ((ao > 0.25) && (ao < 0.5))
                data.weights[jt->index] = 0.0;
        }
    }

#if 1
    // check for finite features
    for (int i = 0; i < data.size(); i++) {
        if (!isfinite(data.features[i][0])) {
            data.features[i][0] = 0.0;
            data.weights[i] = 0.0;
        }
    }
#endif

    // whiten features
    svlFeatureWhitener featureWhitener;
    featureWhitener.clear(data.numFeatures());
    featureWhitener.train(data.features);
    featureWhitener.evaluate(data.features);

    // append bias
    for (int i = 0; i < data.size(); i++) {
        data.features[i].push_back(1.0);
    }

    // re-calibrate
    double lambda = 1.0e-9;
    svlMultiClassLogistic classifier(data.numFeatures(), 2, lambda);
    classifier.train(data.features, data.labels, data.weights);

    // output
    if (outputFilename != NULL) {
        for (svlObject2dSequence::iterator it = detections.begin(); it != detections.end(); it++) {
            for (svlObject2dFrame::iterator jt = it->second.begin(); jt != it->second.end(); jt++) {
                vector<double> scores;
                classifier.getClassScores(data.features[jt->index], scores);
                jt->pr = -scores[0];
                if (objectName != NULL) {
                    jt->name = string(objectName);
                }
            }
        }

        SVL_LOG(SVL_LOG_MESSAGE, "Writing detections to " << outputFilename << "...");
        detections.write(outputFilename);
    }

    // free memory
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}
