/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    classifyImages.cpp
** AUTHOR(S):   Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**  Application for classifying images using HOG feature.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string.h>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"
#include "sceneHOGFeatures.h"
#include "sceneSlidingWindowDetector.h"

using namespace std;

#define WINDOW_NAME "classifyImages"

// main -----------------------------------------------------------------------

void usage()
{
	//assert(false);
	cerr << SVL_USAGE_HEADER << endl;
	cerr << "USAGE: ./sceneClassifyImages [OPTIONS] <model> (<image>* | <imageSequence>)" << endl;
	cerr << "OPTIONS:" << endl
             << "  -width <num>       :: window width (default: 64)" << endl
             << "  -height <num>      :: window height (default: 128)" << endl
	     << "  -b                :: only search at base scale" << endl
	     << "  -n <object name>  :: name of object" << endl
	     << "  -o <filename>     :: output to file" << endl
	     << "  -t <n>            :: threshold for outputting a detection" << endl
	     << "                       (note: svlSlidingWindow's threshold still determines" << endl
	     << "                        the min. probability for a detection to be found)" << endl
	     << "  -r <x> <y> <w> <h> :: scan subregion of image" << endl
	     << "  -R <filename>     :: only scan rectangles in file (<x> <y> <w> <h>)" <<endl
	     << "  -maxImages <n>    :: maximum number of images to analyze" <<endl
	     << "  -x                :: display image and matches" << endl
	     << "  -g <filename>     :: groundtruth file (to analyze on the spot)" << endl
	     << "  -pr <filename>    :: output a pr curve if groundtruth present (from summary scores)" << endl
	     << "  -s <filename>     :: output an evaluations summary if groundtruth present" << endl
	     << "  -cache <folder>   :: caches the sliding window results" << endl
	     << SVL_STANDARD_OPTIONS_USAGE 
	     << endl;
}


//[[TODO: classifier type should refer to logistic, boosted, etc.]]

int main(int argc, char *argv[])
{		
  captureOpenCVerrors();
  
  const unsigned MAX_RESULTS_TO_SHOW = 16;
  const unsigned MAX_WIDTH = 1024;
  
  const int NUM_PARAMETERS = 2;

  // read commandline parameters  
  int width = 64;
  int height = 128;
  bool bBaseScaleOnly = false;
  const char *objectName = "[unknown]";
  const char *outputFilename = NULL;
  CvRect subRegion = cvRect(0, 0, 0, 0);
  char *regionsFile = NULL;
  bool bDisplayImage = false;
  bool bUseNonmax = false;
  int maxImages = INT_MAX;
  const char *groundTruth = NULL;
  const char *prFilename = NULL;
  const char *summaryFilename = NULL;
  float outputThreshold = 0.0;
  const char *cacheFolder = NULL;
  
  SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
    SVL_CMDLINE_INT_OPTION("-width", width)
    SVL_CMDLINE_INT_OPTION("-height", height)
    SVL_CMDLINE_BOOL_OPTION("-b", bBaseScaleOnly)
    SVL_CMDLINE_STR_OPTION("-n", objectName)
    SVL_CMDLINE_STR_OPTION("-o", outputFilename)
    SVL_CMDLINE_OPTION_BEGIN("-r", ptr)
      subRegion.x = atoi(ptr[0]);
      subRegion.y = atoi(ptr[1]);
      subRegion.width = atoi(ptr[2]);
      subRegion.height = atoi(ptr[3]);
    SVL_CMDLINE_OPTION_END(4)
    SVL_CMDLINE_STR_OPTION("-R", regionsFile)
    SVL_CMDLINE_INT_OPTION("-maxImages", maxImages)
    SVL_CMDLINE_BOOL_OPTION("-x", bDisplayImage)
    SVL_CMDLINE_BOOL_OPTION("-nm", bUseNonmax)      
    SVL_CMDLINE_STR_OPTION("-g", groundTruth)
    SVL_CMDLINE_STR_OPTION("-pr", prFilename)
    SVL_CMDLINE_STR_OPTION("-s", summaryFilename)
    SVL_CMDLINE_STR_OPTION("-cache", cacheFolder)
    SVL_CMDLINE_REAL_OPTION("-t", outputThreshold)
  SVL_END_CMDLINE_PROCESSING(usage());
    
  if (SVL_CMDLINE_ARGC < NUM_PARAMETERS) {
    usage();
    return -1;
  }

  svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));
	
  const char *classifierFilename = *SVL_CMDLINE_ARGV++; SVL_CMDLINE_ARGC--;
  
  // create the HOG feature extractor
  SceneHOGFeatures* extractor = new SceneHOGFeatures();
  extractor->setWinWidth(width);
  extractor->setWinHeight(height);

  int classifyHandle = svlCodeProfiler::getHandle("classification");

  svlImageLoader loader(false, true);
  loader.readCommandLine(SVL_CMDLINE_ARGV, SVL_CMDLINE_ARGC);

  SceneSlidingWindowDetector detector(objectName);
  bool bSuccess = detector.load(classifierFilename, extractor);
  SVL_ASSERT_MSG(bSuccess, "detector failed to load the classifier");
  
  svlObjectDetectionAnalyzer analyzer(objectName);
  if (groundTruth) {
    analyzer.loadGt(groundTruth);
  }

  // run classifier on all supplied images
  unsigned numImages = loader.numFrames();
  if (maxImages < (int)numImages) numImages =  maxImages;

  vector<IplImage *> annotatedImages;
  annotatedImages.resize(((numImages < MAX_RESULTS_TO_SHOW) ?
			  numImages : MAX_RESULTS_TO_SHOW), NULL);
  if (sqrt((double)annotatedImages.size()) * loader.getResizeWidth() > MAX_WIDTH) {
    annotatedImages.resize(1);
  }
  IplImage *resultsImage = NULL;
  
  if (bDisplayImage) {
    cvNamedWindow(WINDOW_NAME, 0);
  }

  if (cacheFolder) {
    svlCreateDirectory(cacheFolder);
  }

  ofstream ofs;
  if (outputFilename) {
    ofs.open(outputFilename);
    assert(!ofs.fail());
    ofs << "<Object2dSequence>" << endl;
  }

  SVL_LOG(SVL_LOG_MESSAGE, "Running classifier on " << numImages << " images/sequences...");

  //double threshold= detector.getThreshold();

  // TODO: not hard-coded
  double threshold= 0.05;
 
  for (unsigned index = 0; index < numImages; index++) 
    {
      SVL_LOG(SVL_LOG_VERBOSE, "Processing " << loader.getFrameBasename(index) << "...");

      vector<IplImage *> images;
      bool success = loader.getFrame(index, images);
      if (!success) continue;

      // classify image
      svlObject2dFrame objects;
 
      svlCodeProfiler::tic(classifyHandle);
      if (regionsFile != NULL) {
          CvRect r;
          svlObject2dFrame windows;
          ifstream ifs(regionsFile);
          SVL_ASSERT(!ifs.fail());
          while (1) {
              ifs >> r.x >> r.y >> r.width >> r.height;
              if (ifs.fail()) break;
              windows.push_back(svlObject2d(r));
          }
          detector.classifyImage(images, &objects, NULL);
          ifs.close();
      } else if (bBaseScaleOnly) {
          svlObject2dFrame windows;
	  if ((subRegion.width > 0) && (subRegion.height > 0)) {
              detector.getWindowsAtBaseScale(subRegion.width, subRegion.height, windows);
              for (vector<svlObject2d>::iterator it = windows.begin(); it != windows.end(); ++it) {
                  it->x += subRegion.x;
                  it->y += subRegion.y;
              }
	  } 
	  else {
              detector.getWindowsAtBaseScale(images[0]->width, images[0]->height, windows);
	  }
	  detector.classifyImage(images, windows, &objects);
      } 
      else if ((subRegion.width > 0) && (subRegion.height > 0)) {
          detector.classifySubImage(images, subRegion, &objects);
      } 
      else {
          detector.classifyImage(images, &objects);
      }
      svlCodeProfiler::toc(classifyHandle);
      
      // non maximal suppression
      if (bUseNonmax) {
	int total = objects.size();
	int removed = nonMaximalSuppression(objects, 0.25, 0.25, 0.64);
	SVL_LOG(SVL_LOG_VERBOSE, removed << " out of " << total << " non-maximal objects suppressed");     
      }

      if (bDisplayImage) {
	IplImage *colourImage = cvCreateImage(cvSize(images[0]->width, images[0]->height),
					      IPL_DEPTH_8U, 3);
	if (images[0]->nChannels == 1) {
	  cvCvtColor(images[0], colourImage, CV_GRAY2RGB);
	} else if (images[0]->nChannels == 3) {
	  cvCopy(images[0], colourImage);
	}
	if (resultsImage != NULL) {	
	  cvReleaseImage(&resultsImage);
	}
       
	for (unsigned i = 0; i < objects.size(); i++) {
	  cvRectangle(colourImage, cvPoint((int)objects[i].x, (int)objects[i].y),
		      cvPoint((int)(objects[i].x + objects[i].w - 1), 
			      (int)(objects[i].y + objects[i].h - 1)),
		      CV_RGB(0, (unsigned char)(255 * (objects[i].pr - threshold) / (1.0 - threshold)), 0), 3, 8);
	  cvRectangle(colourImage, cvPoint((int)objects[i].x, (int)objects[i].y),
		      cvPoint((int)(objects[i].x + objects[i].w - 1), 
			      (int)(objects[i].y + objects[i].h - 1)),
		      CV_RGB(255, 255, 255), 1);
	}
       
	cvRectangle(colourImage, cvPoint(subRegion.x, subRegion.y),
		    cvPoint(subRegion.x + subRegion.width - 1, subRegion.y + subRegion.height - 1), 
		    CV_RGB(0, 0, 255), 1);
       
	if (annotatedImages[index % annotatedImages.size()] != NULL) {
	  cvReleaseImage(&annotatedImages[index % annotatedImages.size()]);
	}
	annotatedImages[index % annotatedImages.size()] = colourImage;
	resultsImage = combineImages(annotatedImages);
	cvShowImage(WINDOW_NAME, resultsImage);
	int c = cvWaitKey(index != numImages - 1 ? 100 : 0);
	//int c = cvWaitKey(0);
       
	if (c == (int)'s') {
	  stringstream filename;
	  filename << "classifyImage" << index << ".jpg";
	  cvSaveImage(filename.str().c_str(), resultsImage);
	}
      }

      // output objects
      if (groundTruth) {
	// will run non-max suppression on objects!
	analyzer.accumulateScoresForFrame(strFilename(loader.getFrameBasename(index)), objects);
      }
      if (outputFilename) {
	if (outputThreshold > threshold)
	  removeBelowProbability(objects, outputThreshold);
	writeObject2dFrame(ofs, objects, strFilename(loader.getFrameBasename(index)).c_str());
      }
     
      // free image
      for (unsigned i = 0; i < images.size(); i++) {
	cvReleaseImage(&images[i]);
      }
     
    }//image names in image sequence loop
 
  if (outputFilename) {
    ofs << "</Object2dSequence>" << endl;
    ofs.close();
  }

  if (groundTruth && prFilename) {
      analyzer.writePRCurves(prFilename);
  }

  if (groundTruth && summaryFilename) {
      analyzer.quantizeScores();
      analyzer.writeScores(summaryFilename);
  }

  //  if (groundTruth && summaryFilename) {
  //      analyzer.writeSummaryScores(summaryFilename);
  //}

  // free results
  if (resultsImage != NULL) {
    cvReleaseImage(&resultsImage);
  }
  for (unsigned i = 0; i < annotatedImages.size(); i++) {
    if (annotatedImages[i] != NULL) {
      cvReleaseImage(&annotatedImages[i]);
    }
  }
  
  // free extractor
  delete extractor;

  svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
  svlCodeProfiler::print(cerr);
 
  return 0;
}
