/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneConfiguration.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Geremy Heitz <gaheitz@stanford.edu>
**              Rick Fulton <rafulton@cs.stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"

using namespace std;

// globals ------------------------------------------------------------------

SceneConfiguration gSceneConfig;
SceneRegionDefinitions gSceneRegionDefs;
SceneObjectDefinitions gSceneObjectDefs;

// SceneConfiguration -------------------------------------------------------

int SceneConfiguration::numRegionLabels = 8;
int SceneConfiguration::numObjectLabels = 0;

SceneConfiguration::SceneConfiguration() : svlOptions(), svlConfigurableModule("sceneConfig")
{
    // declare default data options
    declareOption("baseDir", "");
    declareOption("imgDir", "data/images/");
    declareOption("segDir", "data/multisegmentations/");
    declareOption("lblDir", "data/labels/");
    declareOption("auxLblDir", "");
    declareOption("featuresDir", "features/");
    declareOption("modelsDir", "models/");
    declareOption("movesDir", "moves/");
    declareOption("outputDir", "output/");

    declareOption("imgExt", ".jpg");
    declareOption("segExt", ".seg.int");
    declareOption("combinedSegExt", ".combined.seg.int" );
    declareOption("multiSegExt", ".S11.R14.seg.int .S19.R14.seg.int .S31.R14.seg.int");
    declareOption("regionLblExt", ".regions.txt");
    declareOption("layerLblExt", ".layers.txt");
    declareOption("layerClassLblExt", ".layerLabels.txt");
    declareOption("closedLoopFeatureCache", "");
    declareOption("closedLoopFeatureCacheObject", "");
    declareOption("featuresExt", ".features.bin"); // TODO: rename to pixelFeaturesExt
    declareOption("regionFeaturesExt", ".region.features.bin");
    declareOption("objectFeaturesExt", ".object.features.bin");
    declareOption("pedestrianFeaturesExt", ".ped.features.bin");
    declareOption("objDetScoreCacheExt", ".objDetScore.bin");
    declareOption("auxFeatureExt", "");
    declareOption("movesExt", ".moves.bin");

    declareOption("horizons", "horizons.txt");

    declareOption("objDetMoveDir", "features/");
    declareOption("objDetMoveExt", ".obj.seg.move.txt");
    declareOption("objDetMoveScoreExt", ".obj.seg.move.score.bin");

    declareOption("useCache", true);
    declareOption("smoothImage", false);

    // declare pixel feature parameters
    declareOption("filterBandwidth", 1.0);
    declareOption("featureGridSpacing", 0);
    declareOption("featureGridScales", 1);
    declareOption("includeRowFeature", true);
    declareOption("whitenPixelFeatures", true);

    declareOption("useAuxiliaryPixelFeatures", false);

    // declare region feature parameters
    declareOption("includeMomentFeatures", false);
    declareOption("debug:boostedPixelFeaturesOnly", false);

    // declare default experiment/model options
    declareOption("useGeometry", false); // implies horizon models
    declareOption("useBoostedRegionFeatures", false);
    declareOption("addWeightedPixelScore", 0.0);
    declareOption("objSingletonWeight", 1.0);

    // declare object detection parameters
    declareOption("trainingObjectsDir", "");
    declareOption("patchBasedObjectDetector", false);
    declareOption("softBackgroundRegionBasedObjectDetector", true);
    declareOption("useWorldHeightFeature", true);
    declareOption("useRegionSingletonsInJointModel", true);

    declareOption("useRegionFeaturesAtObjectLevel", false);
    declareOption("usePedestrianFeaturesAtObjectLevel", false);
    declareOption("useBoostedClassifierAtObjectLevel", false);
    declareOption("useIndependentPedestrianClassifierAtObjectLevel", false);
    declareOption("usePedDetectorAfterObjectBoosting", false);
    declareOption("useCrossTermsAfterObjectBoosting", false);
    declareOption("useRegionFeaturesAfterObjectBoosting", false);
    declareOption("searchRegionsForPedestrianDetector", true);
    declareOption("multiplyObjectScoreBySize", true);
    declareOption("objectContextWeight", 1.0);

    // declare learning options
    declareOption("balancedSingletons", false);
    declareOption("subSampleAllPixelClass", true);
    declareOption("regPixelSingleton", 1.0e-9);
    declareOption("regRegionSingleton", 1.0e-9);
    declareOption("regObjectSingleton", 1.0e-9);
    declareOption("regRegionPairwise", 1.0e-3);
    declareOption("regObjectContext", 1.0e-3);
    declareOption("regHorizonRegionPair", 1.0e-3);
}

SceneConfiguration::~SceneConfiguration()
{
    // do nothing
}

// options i/o
void SceneConfiguration::writeConfiguration(const char *filename) const
{
    SVL_ASSERT(filename != NULL);

    ofstream ofs(filename);
    SVL_ASSERT(!ofs.fail());

    ofs << "<svl>\n";
    ofs << "  <sceneConfig numRegionLabels=\"" << numRegionLabels << "\"\n"
        << "               numObjectLabels=\"" << numObjectLabels << "\" >\n";

    vector<string> options = getOptionNames();
    for (int i = 0; i < (int)options.size(); i++) {
        ofs << "    <option name=\"" << options[i] << "\" value=\""
            << getOption(options[i]) << "\" />\n";
    }

    ofs << "  </sceneConfig>\n";
    ofs << "</svl>\n";
    ofs.close();
}

void SceneConfiguration::readConfiguration(XMLNode& node)
{
    // read sizes and method
    if (node.getAttribute("numRegionLabels") != NULL) {
        numRegionLabels = atoi(node.getAttribute("numRegionLabels"));
    }
    if (node.getAttribute("numObjectLabels") != NULL) {
        numObjectLabels = atoi(node.getAttribute("numObjectLabels"));
    }

    // read options
    SVL_LOG(SVL_LOG_VERBOSE, "Reading " << node.nChildNode("option")
        << " options...");
    setOptionsFromXML(node, "option");

    // read region definitions
    if (node.nChildNode("regionDefinitions") != 0) {
        XMLNode childNode = node.getChildNode("regionDefinitions");
        gSceneRegionDefs.read(childNode);
    }

    // read object definitions
    if (node.nChildNode("objectDefinitions") != 0) {
        XMLNode childNode = node.getChildNode("objectDefinitions");
        gSceneObjectDefs.read(childNode);
    }
}

// utility functions
string SceneConfiguration::filebase(const char *dirKey, const char *baseName)
{
    return getOption("baseDir") + getOption(dirKey) + string(baseName);
}

string SceneConfiguration::filebase(const char *dirKey, const string& baseName)
{
    return getOption("baseDir") + getOption(dirKey) + baseName;
}

string SceneConfiguration::filename(const char *dirKey,
    const char *baseName, const char *extKey)
{
    return (getOption("baseDir") + getOption(dirKey) +
        string(baseName) + getOption(extKey));
}

string SceneConfiguration::filename(const char *dirKey,
    const string &baseName, const char *extKey)
{
    return (getOption("baseDir") + getOption(dirKey) +
        baseName + getOption(extKey));
}

// SceneRegionDefinitions ---------------------------------------------------

SceneRegionDefinitions::SceneRegionDefinitions()
{
    initDefaultRegions();
}

SceneRegionDefinitions::SceneRegionDefinitions(const char *filename)
{
    read(filename);
}

SceneRegionDefinitions::~SceneRegionDefinitions()
{
    // do nothing
}

// i/o
void SceneRegionDefinitions::clear()
{
    _keys.clear();
    _colors.clear();
    _names.clear();
    _surfaces.clear();
}

void SceneRegionDefinitions::write(const char *filename)
{
    SVL_ASSERT_MSG(false, "not implemented yet");
}

void SceneRegionDefinitions::read(const char *filename)
{
    SVL_ASSERT(filename != NULL);
    clear();

    // if no filename is given then reset to default definitions
    if (strlen(filename) == 0) {
        initDefaultRegions();
        return;
    }

    XMLNode root = XMLNode::parseFile(filename, "regionDefinitions");
    read(root);
}

void SceneRegionDefinitions::read(XMLNode& root)
{
    SVL_ASSERT(!root.isEmpty());
    clear();

    unsigned char red, green, blue;

    for (int i = 0; i < root.nChildNode("region"); i++) {
        XMLNode node = root.getChildNode("region", i);

        _keys[atoi(node.getAttribute("id"))] = i;
        _names.push_back(string(node.getAttribute("name")));
        if (sscanf(node.getAttribute("color"), "%hhd %hhd %hhd",
                &red, &green, &blue) != 3) {
            SVL_LOG(SVL_LOG_FATAL, "could not parse color for \"" << _names.back() << "\"");
        }
        _colors.push_back((red << 16) | (green << 8) | blue);

        if (node.getAttribute("surface") != NULL) {
            int surfTypeId = atoi(node.getAttribute("surface"));
            switch (surfTypeId) {
            case 0: _surfaces.push_back(ST_SKY); break;
            case 1: _surfaces.push_back(ST_GROUNDPLANE); break;
            case 2: _surfaces.push_back(ST_VERTICAL); break;
            default:
                if (surfTypeId > 0) {
                    SVL_LOG(SVL_LOG_ERROR, "unrecognized surface type for \""
                        << _names.back() << "\"");
                }
                _surfaces.push_back(ST_UNKNOWN);
                break;
            }
        } else {
            SVL_LOG(SVL_LOG_WARNING, "unknown surface type for \"" << _names.back() << "\"");
            _surfaces.push_back(ST_UNKNOWN);
        }
    }
}

set<int> SceneRegionDefinitions::keys() const
{
    set<int> k;
    for (map<int, int>::const_iterator it = _keys.begin(); it != _keys.end(); ++it) {
        k.insert(it->first);
    }

    return k;
}

string SceneRegionDefinitions::name(int key) const
{
    map<int, int>::const_iterator it = _keys.find(key);
    if (it == _keys.end()) {
        return string("unknown");
    }

    SVL_ASSERT((unsigned)it->second < _names.size());
    return _names[it->second];
}

unsigned int SceneRegionDefinitions::color(int key) const
{
    map<int, int>::const_iterator it = _keys.find(key);
    if (it == _keys.end()) {
        return 0x00000000;
    }

    SVL_ASSERT((unsigned)it->second < _colors.size());
    return _colors[it->second];
}

TSceneSurfaceType SceneRegionDefinitions::surface(int key) const
{
    map<int, int>::const_iterator it = _keys.find(key);
    if (it == _keys.end()) {
        return ST_UNKNOWN;
    }

    SVL_ASSERT((unsigned)it->second < _surfaces.size());
    return _surfaces[it->second];
}

// operators
unsigned int& SceneRegionDefinitions::operator[](int key)
{
    return _colors[_keys[key]];
}

void SceneRegionDefinitions::initDefaultRegions()
{
    clear();

    _keys[0] = 0;
    _colors.push_back(0x00808080);
    _names.push_back("sky");
    _surfaces.push_back(ST_SKY);

    _keys[1] = 1;
    _colors.push_back(0x00808000);
    _names.push_back("tree");
    _surfaces.push_back(ST_VERTICAL);

    _keys[2] = 2;
    _colors.push_back(0x00804080);
    _names.push_back("road");
    _surfaces.push_back(ST_GROUNDPLANE);

    _keys[3] = 3;
    _colors.push_back(0x00008000);
    _names.push_back("grass");
    _surfaces.push_back(ST_GROUNDPLANE);

    _keys[4] = 4;
    _colors.push_back(0x00000080);
    _names.push_back("water");
    _surfaces.push_back(ST_GROUNDPLANE);

    _keys[5] = 5;
    _colors.push_back(0x00800000);
    _names.push_back("building");
    _surfaces.push_back(ST_VERTICAL);

    _keys[6] = 6;
    _colors.push_back(0x00803200);
    _names.push_back("mountain");
    _surfaces.push_back(ST_VERTICAL);

    _keys[7] = 7;
    _colors.push_back(0x00ff8000);
    _names.push_back("foreground");
    _surfaces.push_back(ST_VERTICAL);
}


// SceneObjectDefinitions ---------------------------------------------------

SceneObjectDefinitions::SceneObjectDefinitions()
{
    // do nothing
}

SceneObjectDefinitions::SceneObjectDefinitions(const char *filename)
{
    read(filename);
}

SceneObjectDefinitions::~SceneObjectDefinitions()
{
    // do nothing
}

// i/o
void SceneObjectDefinitions::clear()
{
    _keys.clear();
    _colors.clear();
    _names.clear();
    _detectors.clear();
    _cachedDetections.clear();
}

void SceneObjectDefinitions::write(const char *filename)
{
    SVL_ASSERT(false);
}

void SceneObjectDefinitions::read(const char *filename)
{
    SVL_ASSERT(filename != NULL);
    clear();
    XMLNode root = XMLNode::parseFile(filename, "objectDefinitions");
    read(root);
}

void SceneObjectDefinitions::read(XMLNode& root)
{
    SVL_ASSERT(!root.isEmpty());
    clear();

    unsigned char red, green, blue;

    for (int i = 0; i < root.nChildNode("object"); i++) {
        XMLNode node = root.getChildNode("object", i);

        int objId = atoi(node.getAttribute("id"));
        if (_keys.find(objId) != _keys.end()) {
            SVL_LOG(SVL_LOG_FATAL, "object with id " << objId << " ("
                << _names[_keys[objId]] << ") already exists");
        }
        _keys[objId] = i;
        _names.push_back(string(node.getAttribute("name")));

        SVL_ASSERT(node.getAttribute("extractor") != NULL);
        SVL_ASSERT(node.getAttribute("detector") != NULL);
        _detectors.push_back(make_pair(string(node.getAttribute("extractor")),
                string(node.getAttribute("detector"))));
        if (node.getAttribute("cachedDetections") == NULL) {
            _cachedDetections.push_back(string(""));
        } else {
            _cachedDetections.push_back(string(node.getAttribute("cachedDetections")));
        }

        if (node.getAttribute("color") == NULL) {
            SVL_LOG(SVL_LOG_WARNING, "could not parse color for \"" << _names.back() << "\"");
            _colors.push_back(0x00ffffff);
        } else if (sscanf(node.getAttribute("color"), "%hhd %hhd %hhd",
                &red, &green, &blue) != 3) {
            SVL_LOG(SVL_LOG_FATAL, "could not parse color for \"" << _names.back() << "\"");
        } else {
            _colors.push_back((red << 16) | (green << 8) | blue);
        }
    }
}

set<int> SceneObjectDefinitions::keys() const
{
    set<int> k;
    for (map<int, int>::const_iterator it = _keys.begin(); it != _keys.end(); ++it) {
        k.insert(it->first);
    }

    return k;
}

string SceneObjectDefinitions::name(int key) const
{
    map<int, int>::const_iterator it = _keys.find(key);
    if (it == _keys.end()) {
        return string("unknown");
    }

    SVL_ASSERT((unsigned)it->second < _names.size());
    return _names[it->second];
}

unsigned int SceneObjectDefinitions::color(int key) const
{
    map<int, int>::const_iterator it = _keys.find(key);
    if (it == _keys.end()) {
        return 0x00000000;
    }

    SVL_ASSERT((unsigned)it->second < _colors.size());
    return _colors[it->second];
}

SceneRegionObjectClassifier *SceneObjectDefinitions::createDetector(int key) const
{
    map<int, int>::const_iterator it = _keys.find(key);
    if (it == _keys.end()) {
        return NULL;
    }

    SVL_LOG(SVL_LOG_VERBOSE, "...reading detector for " << _names[it->second] << " from "
        << _detectors[it->second].first << " and " << _detectors[it->second].second);

    SceneRegionObjectClassifier *detector =
        new SceneRegionObjectClassifier(_names[it->second].c_str());
    SVL_ASSERT(detector != NULL);

    bool bPatchBased = gSceneConfig.getOptionAsBool("patchBasedObjectDetector");
    bool bSuccess = detector->load(_detectors[it->second].second.c_str(),
        _detectors[it->second].first.c_str(), bPatchBased);

    SVL_ASSERT_MSG(bSuccess, "failed to load detector for " << _names[it->second]);

    return detector;
}

ScenePatchObjectDetector *SceneObjectDefinitions::createPatchDetector(int key) const
{
    map<int, int>::const_iterator it = _keys.find(key);
    if (it == _keys.end()) {
        return NULL;
    }

    SVL_LOG(SVL_LOG_VERBOSE, "...reading detector for " << _names[it->second] << " from "
        << _detectors[it->second].first << " and " << _detectors[it->second].second);

    ScenePatchObjectDetector *detector =
        new ScenePatchObjectDetector(_detectors[it->second].first.c_str(), _detectors[it->second].second.c_str());
    SVL_ASSERT(detector != NULL);

    detector->setObjectName(_names[it->second].c_str());
    return detector;
}

svlObject2dSequence *SceneObjectDefinitions::loadCachedDetections(int key) const
{
    map<int, int>::const_iterator it = _keys.find(key);
    if (it == _keys.end()) {
        return NULL;
    }
    
    if (_cachedDetections[it->second].empty()) {
        SVL_LOG(SVL_LOG_VERBOSE, "...no cached detections for " << _names[it->second]);
        return NULL;
    }

    SVL_LOG(SVL_LOG_VERBOSE, "...reading cached detections for " << _names[it->second] 
        << " from " << _cachedDetections[it->second]);

    svlObject2dSequence *sequence = new svlObject2dSequence();
    sequence->read(_cachedDetections[it->second].c_str());

    return sequence;
}
