/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneConfiguration.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Geremy Heitz <gaheitz@stanford.edu>
**              Rick Fulton <rafulton@cs.stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
**
** DESCRIPTION:
**  This file holds code configuration of a scene understanding model.
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <map>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneRegionObjectClassifier.h"
#include "scenePatchObjectDetector.h"

using namespace std;

// SceneConfiguration -------------------------------------------------------

class SceneConfiguration : public svlOptions, svlConfigurableModule {
 public:
    static int numRegionLabels;
    static int numObjectLabels;

 public:
    SceneConfiguration();
    ~SceneConfiguration();

    // options i/o
    void writeConfiguration(const char *filename) const;
    void readConfiguration(XMLNode& node);    
    void setConfiguration(const char *name, const char *value) {
        svlOptions::setOption(name, value);
    }

    // utility functions
    string filebase(const char *dirKey, const char *baseName);
    string filebase(const char *dirKey, const string& baseName);
    string filename(const char *dirKey, const char *baseName,
        const char *extKey);
    string filename(const char *dirKey, const string& baseName,
        const char *extKey);
};

extern SceneConfiguration gSceneConfig;

// sceneRegionDefinitions ---------------------------------------------------

typedef enum _TSceneSurfaceType {
    ST_UNKNOWN, ST_SKY, ST_GROUNDPLANE, ST_VERTICAL
} TSceneSurfaceType;

class SceneRegionDefinitions {
 protected:
    map<int, int> _keys;
    vector<unsigned int> _colors;
    vector<string> _names;
    vector<TSceneSurfaceType> _surfaces;

 public:
    SceneRegionDefinitions();
    SceneRegionDefinitions(const char *filename);
    virtual ~SceneRegionDefinitions();
    
    // i/o
    void clear();
    void write(const char *filename);
    void read(const char *filename);
    void read(XMLNode& root);

    bool empty() const { return _colors.empty(); }
    int size() const { return (int)_colors.size(); }

    set<int> keys() const;
    string name(int key) const;
    unsigned int color(int key) const;
    TSceneSurfaceType surface(int key) const;

    static unsigned char red(unsigned int c) {
        return ((c >> 16) & 0xff);
    }
    static unsigned char green(unsigned int c) {
        return ((c >> 8) & 0xff);
    }
    static unsigned char blue(unsigned int c) {
        return (c & 0xff);
    }

    // operators
    unsigned int& operator[](int key);    

 protected:
    virtual void initDefaultRegions();
};

extern SceneRegionDefinitions gSceneRegionDefs;

// SceneObjectDefinitions ---------------------------------------------------

class SceneObjectDefinitions {
 protected:
    map<int, int> _keys;
    vector<unsigned int> _colors;
    vector<string> _names;
    vector<pair<string, string> > _detectors;
    vector<string> _cachedDetections;

 public:
    SceneObjectDefinitions();
    SceneObjectDefinitions(const char *filename);
    ~SceneObjectDefinitions();

    // i/o
    void clear();
    void write(const char *filename);
    void read(const char *filename);
    void read(XMLNode& root);

    bool empty() const { return _names.empty(); }
    int size() const { return (int)_names.size(); }

    set<int> keys() const;
    string name(int key) const;
    unsigned int color(int key) const;

    static unsigned char red(unsigned int c) {
        return ((c >> 16) & 0xff);
    }
    static unsigned char green(unsigned int c) {
        return ((c >> 8) & 0xff);
    }
    static unsigned char blue(unsigned int c) {
        return (c & 0xff);
    }

    SceneRegionObjectClassifier *createDetector(int key) const;
    ScenePatchObjectDetector *createPatchDetector(int key) const;
    svlObject2dSequence *loadCachedDetections(int key) const;
};

extern SceneObjectDefinitions gSceneObjectDefs;
