/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneDataLoader.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"

using namespace std;

// SceneLayersAndLabels class -----------------------------------------------

SceneLayersAndLabels::SceneLayersAndLabels(int width, int height) : layers(NULL)
{
    layers = cvCreateMat(height, width, CV_32SC1);
    clear();
}

SceneLayersAndLabels::SceneLayersAndLabels(const SceneLayersAndLabels& layerAndLabels) :
    layers(NULL), labels(layerAndLabels.labels)
{
    if (layerAndLabels.layers != NULL) {
        layers = cvCloneMat(layerAndLabels.layers);
    }
}

SceneLayersAndLabels::~SceneLayersAndLabels()
{
    if (layers != NULL)
        cvReleaseMat(&layers);
}

// i/o
void SceneLayersAndLabels::clear()
{
    cvZero(layers);
    labels.resize(1);
    labels[0] = -1;
}

void SceneLayersAndLabels::clear(int width, int height)
{
    if ((layers->cols != width) || (layers->rows != height)) {
        cvReleaseMat(&layers);
        layers = cvCreateMat(height, width, CV_32SC1);
    }

    clear();
}

void SceneLayersAndLabels::load(const char *filename, const char *classFilename)
{
    // load layers
    bool bSuccess = readMatrix(layers, filename);
    if (!bSuccess) {
        SVL_LOG(SVL_LOG_FATAL, "Could not read layers from " << filename);
    }

    int numLayers = countLayers(layers);
    labels.resize(numLayers);
   
    // load layer labels
    ifstream ifs(classFilename);
    SVL_ASSERT_MSG(!ifs.fail(), "could not read layer classes from " << classFilename);
    for (int i = 0; i < numLayers; i++) {
        ifs >> labels[i];
        SVL_ASSERT_MSG(!ifs.fail(), "Layer class labels file " << classFilename
            << " is too short. Should have " << numLayers << " entries.");
    }
    ifs.close();    
}

void SceneLayersAndLabels::save(const char *filename, const char *classFilename) const
{
    // save layers
    writeMatrix(layers, filename);
    
    // save layer labels
    ofstream ofs(classFilename);
    SVL_ASSERT_MSG(!ofs.fail(), "could not write layer classes to " << classFilename);
    
    for (int i = 0; i < (int)labels.size(); i++) {
        ofs << labels[i] << "\n";
    }
    ofs.close();
}

// re-segment and re-assign labels
void SceneLayersAndLabels::resegment(const char *segFilename)
{
    SVL_ASSERT(layers != NULL);
    CvMat *segmentation = cvCreateMat(layers->rows, layers->cols, CV_32SC1);
    bool b = readMatrix(segmentation, segFilename);
    SVL_ASSERT_MSG(b, "could not read segmentation from " << segFilename);
    this->resegment(segmentation);
    cvReleaseMat(&segmentation);
}

void SceneLayersAndLabels::resegment(const CvMat *segmentation)
{
    SVL_ASSERT((segmentation != NULL) && (segmentation->rows == layers->rows)
        && (segmentation->cols == layers->cols));

    int numSegs = countLayers(segmentation);

#if 0
    CvMat *m = cvCloneMat(segmentation);
    SVL_ASSERT(svlConnectedComponents(m) == numSegs);
    cvReleaseMat(&m);
#endif

    vector<vector<int> > layerCounts(numSegs, vector<int>(labels.size(), 0));
    vector<int> segmentSizes(numSegs, 0);
    vector<int> layerSizes(labels.size(), 0);

    // count overlap of segments with layers
    const int *p = (const int *)CV_MAT_ELEM_PTR(*segmentation, 0, 0);
    const int *q = (const int *)CV_MAT_ELEM_PTR(*layers, 0, 0);
    for (int i = 0; i < layers->rows * layers->cols; i++, p++, q++) {
        layerCounts[*p][*q] += 1;
        segmentSizes[*p] += 1;
        layerSizes[*q] += 1;
    }
    
    // determine segment labels
    vector<int> segLabels(numSegs, -1);
    for (int i = 0; i < numSegs; i++) {
        if (segmentSizes[i] == 0) continue;

        int modeLayer = argmax<int>(layerCounts[i]);
        SVL_ASSERT(modeLayer < (int)labels.size());
        //double areaOverlap = (double)layerCounts[i][modeLayer] / 
        //    (double)(segmentSizes[i] + layerSizes[modeLayer] - layerCounts[i][modeLayer]);
        double segmentOverlap = (double)layerCounts[i][modeLayer] / (double)segmentSizes[i];

        // invalid label if segment overlaps with too many layers
        if (segmentOverlap < 0.8) continue;
        
        segLabels[i] = labels[modeLayer];
    }

    // reassign layers
    cvCopy(segmentation, layers);
    labels = segLabels;
}

void SceneLayersAndLabels::getLayerSizes() 
{
    layerSizes.resize(labels.size(), 0);
    const int *q = (const int *)CV_MAT_ELEM_PTR(*layers, 0, 0);
    for (int i = 0; i < layers->rows * layers->cols; i++, q++) {
        layerSizes[*q] += 1;
    }
}

void SceneLayersAndLabels::getLayerExtents()
{
    //    layerExtents.resize(labels.size(), VectorXd::Zero(4));
    layerExtents.clear();
    for (int i = 0; i < (int)labels.size(); i++) {
        // Vector4i: minX, maxX, minY, maxY
        Vector4i layerExtent;
        layerExtent[0] = layers->width - 1;
        layerExtent[1] = 0;
        layerExtent[2] = layers->height - 1;
        layerExtent[3] = 0;

        layerExtents.push_back(layerExtent);
    }
    const int *q = (const int *)CV_MAT_ELEM_PTR(*layers, 0, 0);
    for (int y = 0; y < layers->rows; y++) {
        for (int x = 0; x < layers->cols; x++, q++) {
            if ( x < layerExtents[*q][0] ) {
                layerExtents[*q][0] = x;
            }
            if ( x > layerExtents[*q][1] ) {
                layerExtents[*q][1] = x;
            }

            if ( y < layerExtents[*q][2] ) {
                layerExtents[*q][2] = y;
            }
            if ( y > layerExtents[*q][3] ) {
                layerExtents[*q][3] = y;
            }
        }
    }
    
    for (int i = 0; i < (int)labels.size(); i++) {
        int width = layerExtents[i][1] - layerExtents[i][0] + 1;
        int height = layerExtents[i][3] - layerExtents[i][2] + 1;
        layerExtents[i][1] = layerExtents[i][2];
        layerExtents[i][2] = width;
        layerExtents[i][3] = height;
    }
}

// operators
SceneLayersAndLabels& SceneLayersAndLabels::operator=(const SceneLayersAndLabels& layerAndLabels)
{
    if (layers == layerAndLabels.layers) {
        return *this;
    }

    if (layers != NULL) {
        cvReleaseMat(&layers);
        layers = NULL;
    }

    if (layerAndLabels.layers != NULL) {
        layers = cvCloneMat(layerAndLabels.layers);
    }
    labels = layerAndLabels.labels;

    return *this;
}

// protected functions
int SceneLayersAndLabels::countLayers(const CvMat *seg)
{
    SVL_ASSERT(seg != NULL);

    int numLayers = 0;
    const int *p = (const int *)CV_MAT_ELEM_PTR(*seg, 0, 0);
    for (int i = 0; i < seg->rows * seg->cols; i++, p++) {
        if (*p >= numLayers) numLayers = *p + 1;
    }
    
    return numLayers;
}

// SceneFeatureCache class --------------------------------------------------

SceneFeatureCache::SceneFeatureCache()
{
    // do nothing
}

SceneFeatureCache::SceneFeatureCache(const char *filename)
{
    read(filename, false);
}

SceneFeatureCache::SceneFeatureCache(const SceneFeatureCache& cache) :
    features(cache.features), labels(cache.labels), weights(cache.weights)
{
    // do nothing
}

SceneFeatureCache::~SceneFeatureCache()
{
    // do nothing
}

void SceneFeatureCache::reserve(int reserveSize)
{
    if (reserveSize < size()) return;
    features.reserve(reserveSize);
    labels.reserve(reserveSize);
    weights.reserve(reserveSize);
}

int SceneFeatureCache::count(int classId) const
{
    int c = 0;
    
    for (vector<int>::const_iterator it = labels.begin(); it != labels.end(); it++) {
        if (*it == classId)
            c += 1;
    }

    return c;
}

int SceneFeatureCache::size(const char *filename)
{
    SVL_ASSERT(filename != NULL);

    ifstream ifs(filename, ifstream::in | ifstream::binary);
    if (ifs.fail()) {
        SVL_LOG(SVL_LOG_ERROR, "could not find file " << filename);
        return 0;
    }

    unsigned version;
    ifs.read((char *)&version, sizeof(unsigned));
    int nFeatures;
    ifs.read((char *)&nFeatures, sizeof(int));
    ifs.seekg(0, ios::end);
    int len = (int)ifs.tellg() - 2 * sizeof(int);
    ifs.close();

    int bytesPerRecord = (nFeatures + 1) * sizeof(double) + sizeof(int);
    SVL_ASSERT_MSG(len % bytesPerRecord == 0, "corrupt file " << filename 
        << " (len: " << len << ", bytes/record = " << bytesPerRecord << ")");
    return (int)(len / bytesPerRecord);
}

int SceneFeatureCache::numFeatures(const char *filename)
{
    SVL_ASSERT(filename != NULL);

    ifstream ifs(filename, ifstream::in | ifstream::binary);
    if (ifs.fail()) {
        SVL_LOG(SVL_LOG_ERROR, "could not find file " << filename);
        return 0;
    }

    unsigned version;
    ifs.read((char *)&version, sizeof(unsigned));
    int nFeatures;
    ifs.read((char *)&nFeatures, sizeof(int));
    ifs.close();

    return nFeatures;
}
    
// i/o
void SceneFeatureCache::clear()
{
    features.clear();
    labels.clear();
    weights.clear();
}

int SceneFeatureCache::write(const char *filename, bool bAppend) const
{
    if (empty()) return 0;
    return write(filename, 0, size() - 1, bAppend);
}

int SceneFeatureCache::write(const char *filename, int startIndx, int endIndx, bool bAppend) const
{
    SVL_ASSERT(filename != NULL);
    SVL_ASSERT(features.size() == labels.size());
    SVL_ASSERT(features.size() == weights.size());
    SVL_ASSERT_MSG((startIndx >= 0) && (endIndx < (int)features.size()) && (startIndx <= endIndx),
        "startIndx = " << startIndx << ", endIndx = " << endIndx << ", size() = " << features.size());

    // open file
    unsigned version = 0x0001;
    int nFeatures = numFeatures();
    fstream ofs;
    if (bAppend) {
        ofs.open(filename, ios::in | ios::out | ios::binary);
        ofs.seekg(0, ios::beg);
        ofs.read((char *)&version, sizeof(unsigned));
        ofs.read((char *)&nFeatures, sizeof(int));
        if (ofs.fail()) {
            version = 0x0001;
            nFeatures = numFeatures();
            ofs.close();
            ofs.open(filename, ios::out | ios::binary);
            ofs.write((char *)&version, sizeof(unsigned));
            ofs.write((char *)&nFeatures, sizeof(int));
        } else {
            ofs.seekp(0, ios::end);
        }
    } else {
        ofs.open(filename, ios::out | ios::binary);
        ofs.write((char *)&version, sizeof(unsigned));
        ofs.write((char *)&nFeatures, sizeof(int));
    }
    SVL_ASSERT_MSG(!ofs.fail(), filename);

    // write data
    for (int i = startIndx; i <= endIndx ; i++) {
        SVL_ASSERT_MSG(features[i].size() == (unsigned)nFeatures, 
            features[i].size() << " != " << nFeatures);
        ofs.write((char *)&labels[i], sizeof(int));
        ofs.write((char *)&features[i][0], nFeatures * sizeof(double));
        ofs.write((char *)&weights[i], sizeof(double));
    }
    
    int len = (int)ofs.tellp() - 2 * sizeof(int);
    ofs.close();

    int bytesPerRecord = (nFeatures + 1) * sizeof(double) + sizeof(int);
    SVL_ASSERT_MSG(len % bytesPerRecord == 0, "corrupt file " << filename 
        << " (len: " << len << ", bytes/record = " << bytesPerRecord << ")");
    return (int)(len / bytesPerRecord);
}

int SceneFeatureCache::read(const char *filename, bool bAppend)
{
    return read(filename, 0, numeric_limits<int>::max(), bAppend);
}

int SceneFeatureCache::read(const char *filename, int startIndx, int endIndx, bool bAppend)
{
    SVL_ASSERT(filename != NULL);
    SVL_ASSERT((startIndx >= 0) && (endIndx >= startIndx));
    if (!bAppend) clear();
    
    // open file
    ifstream ifs(filename, ifstream::in | ifstream::binary);
    if (ifs.fail()) {
        SVL_LOG(SVL_LOG_ERROR, "could not find file " << filename);
        return size();
    }

    if (ifs.eof()) {
        SVL_LOG(SVL_LOG_WARNING, "empty file " << filename);
        return size();
    }

    unsigned version;
    ifs.read((char *)&version, sizeof(unsigned));
    SVL_ASSERT_MSG(version == 0x0001, "unrecognized file version");
    int nFeatures;
    ifs.read((char *)&nFeatures, sizeof(int));
    SVL_ASSERT_MSG(empty() || (nFeatures == numFeatures()), nFeatures << " != " << numFeatures());

    // goto start index
    ifs.seekg(startIndx * ((nFeatures + 1) * sizeof(double) + sizeof(int)), ios::cur);
    if (ifs.fail()) {
        ifs.close();
        SVL_LOG(SVL_LOG_WARNING, "less than " << startIndx << " record in file " << filename);
        return size();
    }

    // read until end of file or end index    
    int y;
    vector<double> x(nFeatures);
    double w;
    int recordCount = startIndx;
    while (recordCount <= endIndx) {
        ifs.read((char *)&y, sizeof(int));
        ifs.read((char *)&x[0], nFeatures * sizeof(double));
        ifs.read((char *)&w, sizeof(double));
        if (ifs.fail()) break;
        labels.push_back(y);
        features.push_back(x);
        weights.push_back(w);
        recordCount += 1;
    }
    ifs.close();

#if 1
    // check feature values
    for (int i = 0; i < size(); i++) {
        for (int j = 0; j < nFeatures; j++) {
            if (!isfinite(features[i][j])) {
                SVL_LOG(SVL_LOG_ERROR, "sample " << i << " feature " << j << " is "
                    << features[i][j] << " (" << toString(features[i]) << ")");
                break;
            }
        }
    }
#endif
    
    return size();
}

int SceneFeatureCache::append(const SceneFeatureCache& cache)
{
    if (cache.empty()) return size();
    if (empty()) {
        *this = cache;
        return size();
    }

    SVL_ASSERT(cache.numFeatures() == numFeatures());
    features.insert(features.end(), cache.features.begin(), cache.features.end());
    labels.insert(labels.end(), cache.labels.begin(), cache.labels.end());
    weights.insert(weights.end(), cache.weights.begin(), cache.weights.end());

    return size();
}


int SceneFeatureCache::subSample(int sampleRate, bool bBalanced)
{
    SVL_ASSERT(labels.size() == features.size());
    SVL_ASSERT(weights.size() == features.size());
    if (sampleRate <= 1) return size();

    vector<vector<double> > newFeatures;
    vector<int> newLabels;
    vector<double> newWeights;

    newFeatures.reserve(features.size() / sampleRate);
    newLabels.reserve(newFeatures.size());
    newWeights.reserve(newFeatures.size());

    vector<int> classCounts(2, 0);
    int maxClassCount = 0;

    for (unsigned i = 0; i < features.size(); i++) {
        // label for the purpose of counting
        int countLabel = (labels[i] < 0) ? 0 : labels[i] + 1;
        if ((int)classCounts.size() <= countLabel) {
            classCounts.resize(countLabel + 1, 0);
        }

        // random skip
        if ((rand() % sampleRate != 0) && (classCounts[countLabel] > maxClassCount - 500))
            continue;

        // accumulate
        newFeatures.push_back(features[i]);
        newLabels.push_back(labels[i]);
        newWeights.push_back(weights[i]);

        // update counts
        classCounts[countLabel] += 1;
        if (bBalanced && (classCounts[countLabel] > maxClassCount)) {
            maxClassCount = classCounts[countLabel];
        }
    }

    features = newFeatures;
    labels = newLabels;
    weights = newWeights;
    SVL_LOG(SVL_LOG_VERBOSE, "dataset resampled to contain " << this->size()
        << " (" << toString(classCounts) << ") samples");

    return size();
}

// SceneDataLoader ----------------------------------------------------------

// load horizons
map<string, int> SceneDataLoader::loadHorizonTrainingData()
{
    map<string, int> horizons;

    string filename = gSceneConfig.filebase("lblDir", gSceneConfig.getOption("horizons"));
    ifstream ifs(filename.c_str());
    if (ifs.fail()) {
        SVL_LOG(SVL_LOG_ERROR, "could not read horizon groundtruth from " << filename);
        return horizons;
    }
    
    SVL_LOG(SVL_LOG_VERBOSE, "reading groundtruth horizons...");
    while (!ifs.fail()) {
        string baseName;
        int cols, rows;
        double h;
        ifs >> baseName >> cols >> rows >> h;
        if (!ifs.fail()) {
            horizons[baseName] = (int)h;
        }
    }
    ifs.close();
    SVL_LOG(SVL_LOG_VERBOSE, "..." << horizons.size() << " horizons read");

    return horizons;
}

// load gPb features (8 * <width> * <height> * <32-bit float>)
vector<CvMat *> SceneDataLoader::loadGlobalPbFeatures(const char *filename, int width, int height)
{
    SVL_ASSERT((filename != NULL) && (width > 0) && (height > 0));

    const int CHANNELS = 8;

    SVL_LOG(SVL_LOG_VERBOSE, "loading " << (CHANNELS * width * height * sizeof(float)) << " bytes from " << filename << "...");
    vector<CvMat *> gPbOrientations = svlCreateMatrices(CHANNELS, height, width, CV_32FC1);

    ifstream ifs(filename, ios::binary);
    SVL_ASSERT_MSG(!ifs.fail(), "could not open gPb file " << filename);
    CvMat *dummy = cvCreateMat(width, height, CV_32FC1);
    for (int c = 0; c < CHANNELS; c++) {
        float *p = (float *)CV_MAT_ELEM_PTR(*dummy, 0, 0);
        ifs.read((char *)p, width * height * sizeof(float));
        cvTranspose(dummy, gPbOrientations[c]);
        SVL_ASSERT_MSG(!ifs.fail(), "could not read gPb channel " << c << " in file " << filename);
    }   
    cvReleaseMat(&dummy);
    ifs.close();

#if 0
    if (svlLogger::getLogLevel() <= SVL_LOG_DEBUG) {
        cvNamedWindow("loadGlobalPbFeatures", 1);
        CvMat *gPbEnergy = cvCreateMat(height, width, CV_32FC1);
        for (int c = 0; c < CHANNELS; c++) {
            CvMat *sq = cvCloneMat(gPbOrientations[c]);
            cvMul(sq, sq, sq);
            cvAdd(gPbEnergy, sq, gPbEnergy);
            cvReleaseMat(&sq);
        }
        cvPow(gPbEnergy, gPbEnergy, 0.5);
        cvShowImage("loadGlobalPbFeatures", gPbEnergy);
        cvWaitKey(-1);
        cvDestroyWindow("loadGlobalPbFeatures");
        cvReleaseMat(&gPbEnergy);
    }
#endif

    return gPbOrientations;
}
