/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneDataLoader.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
**
** DESCRIPTION:
**  This file holds code for managing different data formats.
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <map>

#include "Eigen/Core"

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"

using namespace std;

// SceneLayersAndLabels class -----------------------------------------------

class SceneLayersAndLabels {
 public:
    CvMat *layers;
    vector<int> labels;
    vector<int> layerSizes;
    vector<Vector4i> layerExtents;

 public:
    SceneLayersAndLabels(int width, int height);
    SceneLayersAndLabels(const SceneLayersAndLabels& layerAndLabels);
    ~SceneLayersAndLabels();

    inline int size() const { return (int)labels.size(); }

    // i/o
    void clear();
    void clear(int width, int height);
    void load(const char *filename, const char *classFilename);
    void save(const char *filename, const char *classFilename) const;

    // re-segment and re-assign labels
    void resegment(const char *segFilename);
    void resegment(const CvMat *segmentation);

    // get the size for each layer
    void getLayerSizes();
    
    // get the extent for each layer
    void getLayerExtents();

    // operators
    SceneLayersAndLabels& operator=(const SceneLayersAndLabels& layerAndLabels);
    inline int& operator[](unsigned index) { return labels[index]; }
    inline int operator[](unsigned index) const { return labels[index]; }

 protected:
    static int countLayers(const CvMat *seg);

};

// SceneFeatureCache class --------------------------------------------------
// Data is stored as:
//   <version, flags (int32)> <num features (int32)> 
//   <label 1 (int32)> <features 1 (double)> ... <weight (double)> 

class SceneFeatureCache {
 public:
    vector<vector<double> > features;
    vector<int> labels;
    vector<double> weights;

 public:
    SceneFeatureCache();
    SceneFeatureCache(const char *filename);
    SceneFeatureCache(const SceneFeatureCache& cache);
    ~SceneFeatureCache();

    inline bool empty() const { return features.empty(); }
    inline int size() const { return (int)features.size(); }
    int count(int classId) const;
    void reserve(int reserveSize);
    inline int numFeatures() const { return features.empty() ? 0 : (int)features[0].size(); }
    inline int numClasses() const { return labels.empty() ? 0 : *max_element(labels.begin(), labels.end()) + 1; }

    static int size(const char *filename);
    static int numFeatures(const char *filename);
    
    // i/o
    void clear();
    int write(const char *filename, bool bAppend = false) const;
    int write(const char *filename, int startIndx, int endIndx, bool bAppend = false) const;
    int read(const char *filename, bool bAppend = false);
    int read(const char *filename, int startIndx, int endIndx, bool bAppend = false);
    int append(const SceneFeatureCache& cache);
    
    int subSample(int sampleRate, bool bBalanced = true);
};

// SceneDataLoader ----------------------------------------------------------

namespace SceneDataLoader {
    // load horizons
    map<string, int> loadHorizonTrainingData();

    // load gPb features (8 * <width> * <height> * <32-bit float>)
    vector<CvMat *> loadGlobalPbFeatures(const char *filename, int width, int height);
};
