/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneEntities.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tinashig@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <iomanip>
#include <list>
#include <set>

#include "Eigen/Core"

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"

using namespace std;
USING_PART_OF_NAMESPACE_EIGEN

// SceneRegion class -------------------------------------------------------

SceneRegion::~SceneRegion()
{
    // remove any references to this object
    for (set<SceneRegionBoundary *>::iterator it = boundaries.begin(); 
         it != boundaries.end(); it++) {
        if ((*it)->regionA == this) {
            (*it)->regionA = NULL;
        }
        if ((*it)->regionB == this) {
            (*it)->regionB = NULL;
        }
    }

    for (set<SceneObjRegBoundary *>::iterator it = objBoundaries.begin(); 
         it != objBoundaries.end(); it++) {
        if ((*it)->region == this) {
            (*it)->region = NULL;
        }
    }

    if (object != NULL) {
        object->regions.erase(this);
    }
}

CvMat *SceneRegion::mask(int w, int h) const
{
    CvMat *mask = cvCreateMat(h, w, CV_32SC1);
    cvZero(mask);

    int *p = (int *)CV_MAT_ELEM_PTR(*mask, 0, 0);
    for (list<CvPoint>::const_iterator jt = pixels.begin(); jt != pixels.end(); jt++) {
        p[jt->y * mask->cols + jt->x] = 1;
    }

    return mask;
}

VectorXd SceneRegion::appearanceXMoment() const
{
    return ((xAk.firstMoments() - (double)centroidSum.x() / (double)size() * 
        Ak.firstMoments()) / Ak.count());
}

VectorXd SceneRegion::appearanceYMoment() const
{
    return ((yAk.firstMoments() - (double)centroidSum.y() / (double)size() * 
        Ak.firstMoments()) / Ak.count());
}

void SceneRegion::clear()
{
    SceneEntity::clear();
    for (set<SceneRegionBoundary *>::iterator it = boundaries.begin(); 
         it != boundaries.end(); it++) {
        (*it)->clear();
    }
    for (set<SceneObjRegBoundary *>::iterator it = objBoundaries.begin(); 
         it != objBoundaries.end(); it++) {
        (*it)->clear();
    }

    if (object != NULL) {
        object->clear();
    }
}

// SceneObject class ------------------------------------------------------

SceneObject::~SceneObject()
{
    // remove any references to this object
    for (set<SceneObjRegBoundary *>::iterator it = boundaries.begin(); 
         it != boundaries.end(); it++) {
        if ((*it)->object == this) {
            (*it)->object = NULL;
        }
    }

    for (set<SceneRegion *>::iterator it = regions.begin(); it != regions.end(); it++) {
        (*it)->object = NULL;
    }
}

list<CvPoint> SceneObject::pixels() const
{
    list<CvPoint> pixels;
    for (set<SceneRegion *>::const_iterator it = regions.begin(); it != regions.end(); it++) {
        pixels.insert(pixels.end(), (*it)->pixels.begin(), (*it)->pixels.end());
    }

    return pixels;
}

CvMat *SceneObject::mask(int w, int h) const
{
    CvMat *mask = cvCreateMat(h, w, CV_32SC1);
    cvZero(mask);

    int *p = (int *)CV_MAT_ELEM_PTR(*mask, 0, 0);
    for (set<SceneRegion *>::const_iterator it = regions.begin(); it != regions.end(); it++) {
        for (list<CvPoint>::const_iterator jt = (*it)->pixels.begin();
             jt != (*it)->pixels.end(); jt++) {
            p[jt->y * mask->cols + jt->x] = 1;
        }
    }

    return mask;
}

Vector2d SceneObject::centroid() const
{
    Vector2i centroidSum = Vector2i::Zero();
    int numPixels = 0;

    for (set<SceneRegion *>::const_iterator it = regions.begin(); it != regions.end(); it++) {
        centroidSum += (*it)->centroidSum;
        numPixels += (*it)->size();
    }
    
    return centroidSum.cast<double>() / (double)numPixels;
}

Vector4i SceneObject::extent() const
{
    if (regions.empty()) return Vector4i::Zero();

    int minX = numeric_limits<int>::max(), maxX = 0, minY = numeric_limits<int>::max(), maxY = 0;
    for (set<SceneRegion *>::const_iterator it = regions.begin(); it != regions.end(); it++) {
        Vector4i e = (*it)->extent();
        minX = std::min(minX, e[0]);
        minY = std::min(minY, e[1]);
        maxX = std::max(maxX, e[2] + e[0] - 1);
        maxY = std::max(maxY, e[3] + e[1] - 1);
    }

    return Vector4i(minX, minY, maxX - minX + 1, maxY - minY + 1);
}
    
VectorXd SceneObject::appearance() const
{
    SVL_ASSERT(!regions.empty());

    set<SceneRegion *>::const_iterator it = regions.begin();
    svlSufficientStats Ak((*it)->Ak);

    while (++it != regions.end()) {
        Ak.accumulate((*it)->Ak);
    }

    return Ak.firstMoments() / Ak.count();    
}

VectorXd SceneObject::appearanceXMoment() const
{
    SVL_ASSERT(!regions.empty());

    set<SceneRegion *>::const_iterator it = regions.begin();
    svlSufficientStats Ak((*it)->Ak);
    svlSufficientStats xAk((*it)->xAk);

    while (++it != regions.end()) {
        Ak.accumulate((*it)->Ak);
        xAk.accumulate((*it)->xAk);
    }

    return ((xAk.firstMoments() - centroid().x() / (double)size() * 
        Ak.firstMoments()) / Ak.count());
}

VectorXd SceneObject::appearanceYMoment() const
{
    SVL_ASSERT(!regions.empty());

    set<SceneRegion *>::const_iterator it = regions.begin();
    svlSufficientStats Ak((*it)->Ak);
    svlSufficientStats yAk((*it)->yAk);

    while (++it != regions.end()) {
        Ak.accumulate((*it)->Ak);
        yAk.accumulate((*it)->yAk);
    }

    return ((yAk.firstMoments() - centroid().y() / (double)size() * 
        Ak.firstMoments()) / Ak.count());
}

void SceneObject::clear()
{ 
    SceneEntity::clear(); 
    prObject.resize(0); 
    objDetScore.resize(0);
    bComputeObjDetScores = true;

    for (set<SceneObjRegBoundary *>::iterator it = boundaries.begin(); 
         it != boundaries.end(); it++) {
        (*it)->clear();
    }
}

// SceneRegionBoundary class ----------------------------------------------

SceneRegionBoundary::~SceneRegionBoundary()
{
    // remove any references to this object
    if (regionA != NULL) {
        regionA->boundaries.erase(this);
    }

    if (regionB != NULL) {
        regionB->boundaries.erase(this);
    }
}

// SceneObjRegBoundary class ----------------------------------------------

SceneObjRegBoundary::~SceneObjRegBoundary()
{
    // remove any references to this object
    if (region != NULL) {
        region->objBoundaries.erase(this);
    }

    if (object != NULL) {
        object->boundaries.erase(this);
    }
}

int SceneObjRegBoundary::length() const
{
    if ((region == NULL) || (object == NULL))
        return 0;

    int boundaryLength = 0;
    for (set<SceneRegionBoundary *>::const_iterator it = region->boundaries.begin();
         it != region->boundaries.end(); it++) {
        SceneRegion *r = (*it)->neighborOf(region);
        if (object->regions.find(r) != object->regions.end()) {
            boundaryLength += (*it)->length;
        }
    }

    return boundaryLength;
}

