/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneEntities.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tinashig@stanford.edu>
** DESCRIPTION:
**  This file defines the various scene entities (including regions, objects
**  and boundaries between them).
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <list>
#include <set>

#include "Eigen/Core"

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

using namespace std;
USING_PART_OF_NAMESPACE_EIGEN

// Forward declarations ---------------------------------------------------

class SceneAssignment;
class SceneInstance;
class SceneRegionBoundary;
class SceneObjRegBoundary;
class SceneObject;

// SceneEntity class ------------------------------------------------------
// All high-level scene entities inherit from this class.

class SceneEntity {
 protected:
    friend class SceneAssignment;
    friend class SceneInstance;
    int _id;                         // index into parent list

 public:
    VectorXd phi;                    // cached unary potential for scene models
                                     // (meaning depends on actual entity)

 public:
    SceneEntity(int id = -1) : _id(id), phi(0) {
        // do nothing
    }
    SceneEntity(const SceneEntity& e) : _id(e._id), phi(e.phi) {
        // do nothing
    }
    virtual ~SceneEntity() { /* do nothing */ }

    inline int id() const { return _id; }

    // invalid objects can be removed
    inline bool valid() const { return true; }

    // clears cached information when entity changes
    virtual void clear() { phi.resize(0); }
    inline bool dirty() const { return (phi.size() == 0); }
};

// SceneRegion class ------------------------------------------------------

class SceneRegion : public SceneEntity {
 public:
    // scene structure
    list<CvPoint> pixels;            // list of pixels coordinates for this region
    set<SceneRegionBoundary *> boundaries;     // region and object boundaries
    set<SceneObjRegBoundary *> objBoundaries;
    SceneObject *object;             // object that this region belongs (or NULL)

    // region properties
    int Sk;                          // region class label (-1 : unknown/invalid)
    svlSufficientStats Ak;           // region appearance features
    svlSufficientStats xAk;          // x-moment of appearance
    svlSufficientStats yAk;          // y-moment of appearance
    Vector2i centroidSum;            // unnormalized centroid

    VectorXd pixelScores;            // sum of pixel class singleton scores

 public:
    SceneRegion(int id = -1) : SceneEntity(id), object(NULL), Sk(-1), pixelScores(0) {
        centroidSum.setZero();
    }
    SceneRegion(const SceneRegion& e) : SceneEntity(e), pixels(e.pixels),
        boundaries(e.boundaries), objBoundaries(e.objBoundaries), object(e.object), 
        Sk(e.Sk), Ak(e.Ak), xAk(e.xAk), yAk(e.yAk), centroidSum(e.centroidSum), 
        pixelScores(e.pixelScores) {
        // do nothing
    }
    ~SceneRegion();

    inline bool valid() const { return !pixels.empty(); }

    inline bool empty() const { return pixels.empty(); }
    inline int size() const { return (int)pixels.size(); }

    CvMat *mask(int w, int h) const;
    inline Vector2d centroid() const { return centroidSum.cast<double>() / (double)size(); }
    inline Vector4i extent() const;
    inline bool isNeighbor(const SceneRegion *r) const;
    inline VectorXd appearance() const { return Ak.firstMoments() / Ak.count(); }
    VectorXd appearanceXMoment() const;
    VectorXd appearanceYMoment() const;

    // clears cached information when region changes
    void clear();
};

// SceneObject class ------------------------------------------------------

class SceneObject : public SceneEntity {
 public:
    // scene structure
    set<SceneRegion *> regions;     // pointers to component regions
    set<SceneObjRegBoundary *> boundaries;  // object-to-region boundaries

    // object properties
    int Ok;                         // object label (-1 : unknown/invalid)

    VectorXd objDetScore;           // scores from object detectors
    VectorXd prObject;              // probability of object

    bool bComputeObjDetScores;      // false if the score is set outside

 public:
    SceneObject(int id = -1) : SceneEntity(id), Ok(-1), bComputeObjDetScores(true) {
        // do nothing
    }
    SceneObject(const SceneObject& e) : SceneEntity(e), regions(e.regions),
        boundaries(e.boundaries), Ok(e.Ok), objDetScore(e.objDetScore), prObject(e.prObject) {
        // do nothing
    }
    ~SceneObject();

    inline bool valid() const { return !regions.empty(); }

    inline bool empty() const;
    inline int size() const;

    list<CvPoint> pixels() const;
    CvMat *mask(int w, int h) const;
    Vector2d centroid() const;
    Vector4i extent() const;
    inline bool hasRegion(SceneRegion *r) const { return (regions.find(r) != regions.end()); }
    inline bool isNeighbor(const SceneRegion *r) const;

    void setObjDetScoreFromCache(VectorXd& objScores) {objDetScore = objScores; bComputeObjDetScores = false;};

    VectorXd appearance() const;
    VectorXd appearanceXMoment() const;
    VectorXd appearanceYMoment() const;

    // clears cached information when object changes
    void clear();
};

// SceneRegionBoundary class ----------------------------------------------

class SceneRegionBoundary : public SceneEntity {
 public:
    SceneRegion *regionA;           // first region on boundary
    SceneRegion *regionB;           // second region on boundary

    double contrast;                // contrast penalty
    int length;                     // number of pixels along boundary

 public:
    SceneRegionBoundary(int id = -1) : SceneEntity(id),
        regionA(NULL), regionB(NULL), contrast(0.0), length(-1) {
        // do nothing
    }
    SceneRegionBoundary(SceneRegion *rA, SceneRegion *rB, int id = -1) :
        SceneEntity(id), regionA(rA), regionB(rB), contrast(0.0), length(-1) {
        // do nothing
    }
    ~SceneRegionBoundary();

    inline bool valid() const { return ((regionA != NULL) && (regionB != NULL)); }

    inline pair<int, int> regions() const {
        int a = regionA == NULL ? -1 : regionA->id();
        int b = regionB == NULL ? -1 : regionB->id();
        return make_pair(std::min(a, b), std::max(a, b));
    }

    inline SceneRegion *neighborOf(SceneRegion *r) {
        if (r == regionA) return regionB;
        if (r == regionB) return regionA;
        SVL_LOG(SVL_LOG_FATAL, "region " << r->id() << " not on this boundary");
        return NULL;
    }

    // clears cached information when object changes
    void clear() { SceneEntity::clear(); contrast = 0.0; }
};

// SceneObjRegBoundary class ----------------------------------------------

class SceneObjRegBoundary : public SceneEntity {
 public:
    SceneObject *object;            // object entity
    SceneRegion *region;            // region entity

 public:
    SceneObjRegBoundary(int id = -1) : SceneEntity(id), object(NULL), region(NULL) {
        // do nothing
    }
    SceneObjRegBoundary(SceneObject *o, SceneRegion *r, int id = -1) : SceneEntity(id),
        object(o), region(r) {
        // do nothing
    }
    ~SceneObjRegBoundary();

    inline bool valid() const { return ((object != NULL) && (region != NULL)); }
    int length() const;
};

// inline functions -------------------------------------------------------

inline Vector4i SceneRegion::extent() const
{
    if (empty()) return Vector4i::Zero();

    Vector4i e(pixels.begin()->x, pixels.begin()->y, pixels.begin()->x, pixels.begin()->y);
    for (list<CvPoint>::const_iterator it = pixels.begin(); it != pixels.end(); it++) {
        e[0] = std::min(e[0], it->x);
        e[1] = std::min(e[1], it->y);
        e[2] = std::max(e[2], it->x);
        e[3] = std::max(e[3], it->y);
    }
    e[2] = e[2] - e[0] + 1;
    e[3] = e[3] - e[1] + 1;

    return e;
}

inline bool SceneRegion::isNeighbor(const SceneRegion *r) const
{
    // can't be a neighbor of yourself
    if (r == this) return false;

    // search boundaries
    for (set<SceneRegionBoundary *>::const_iterator it = boundaries.begin();
         it != boundaries.end(); it++) {
        if (((*it)->regionA == r) || ((*it)->regionB == r)) {
            SVL_ASSERT(((*it)->regionA == this) || ((*it)->regionB == this));
            return true;
        }
    }

    return false;
}

inline bool SceneObject::empty() const
{
    for (set<SceneRegion *>::const_iterator it = regions.begin(); it != regions.end(); it++) {
        if (!(*it)->empty())
            return false;
    }

    return true;
}

inline int SceneObject::size() const
{
    int s = 0;
    for (set<SceneRegion *>::const_iterator it = regions.begin(); it != regions.end(); it++) {
        s += (*it)->size();
    }

    return s;
}

inline bool SceneObject::isNeighbor(const SceneRegion *r) const
{
    for (set<SceneObjRegBoundary *>::const_iterator it = boundaries.begin();
         it != boundaries.end(); it++) {
        if ((*it)->region == r) {
            SVL_ASSERT((*it)->object == this);
            return true;
        }
    }

    return false;
}

