/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneEvalObjectDetectors.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for running object detectors on novel images.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "scenePatchObjectDetector.h"

#define WINDOW_NAME "sceneEvalObjectDetectors"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneEvalObjectDetectors [OPTIONS] <dict> <model> (<evalDir> | <evalSeq> | <evalImg>)\n";
    cerr << "OPTIONS:\n"
         << "  -n <name>        :: object name\n"
         << "  -i <detections>  :: input candidate detections\n"
         << "  -o <filename>    :: output detections\n"
         << "  -s <scale>       :: base scale, i.e. multiple of detector size (default: 1.0)\n"
         << "  -x               :: visualize results\n"
         << "  -viewDictionary  :: visualize patch dictionary\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // set defaults and read commandline parameters
    const char *objectName = "unknown";
    const char *inputDetections = NULL;
    const char *outFilename = NULL;
    double baseScale = 1.0;
    bool bVisualize = false;
    bool bViewDict = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-n", objectName)
        SVL_CMDLINE_STR_OPTION("-i", inputDetections)
        SVL_CMDLINE_STR_OPTION("-o", outFilename)
        SVL_CMDLINE_REAL_OPTION("-s", baseScale)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
        SVL_CMDLINE_BOOL_OPTION("-viewDictionary", bViewDict)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 3) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *dictName = SVL_CMDLINE_ARGV[0];
    const char *modelName = SVL_CMDLINE_ARGV[1];
    const char *evalList = SVL_CMDLINE_ARGV[2];

    // initialize detector
    SVL_LOG(SVL_LOG_MESSAGE, "loading object detector (dictionary and model)...");
    ScenePatchObjectDetector detector(dictName, modelName);
    SVL_LOG(SVL_LOG_MESSAGE, "...detector base scale is " << detector.baseWidth() 
        << "x" << detector.baseHeight() << "; image will be scaled by " << baseScale);
    detector.setObjectName(objectName);

    if (bViewDict) {
        IplImage *canvas = detector.visualizeDictionary();
        cvNamedWindow("patchDictionary", 1);
        cvShowImage("patchDictionary", canvas);
        cvWaitKey(-1);
        cvDestroyWindow("patchDictionary");
        cvReleaseImage(&canvas);
    }

    // load detection candidates
    svlObject2dSequence candidates;
    if (inputDetections != NULL) {
        svlCodeProfiler::tic(svlCodeProfiler::getHandle("loadDetections"));
        SVL_LOG(SVL_LOG_MESSAGE, "loading candidate detections...");
        bool success = candidates.read(inputDetections);
        SVL_ASSERT(success);
        SVL_LOG(SVL_LOG_MESSAGE, "..." << countObjects(candidates) << " objects from "
            << candidates.size() << " frames loaded...");
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("loadDetections"));
    }

    // read input images
    svlImageSequence imageSequence;
    if (!svlFileExists(evalList)) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading evaluation list from directory " << evalList << "...");
        imageSequence.dir(evalList);
    } else {
        IplImage *tmp = cvLoadImage(evalList);
        if (tmp == NULL) {
            SVL_LOG(SVL_LOG_MESSAGE, "Reading evaluation list from " << evalList << "...");
            imageSequence.load(evalList);
        } else {
            SVL_LOG(SVL_LOG_MESSAGE, "Instantiating evaluation list with " << evalList << "...");
            cvReleaseImage(&tmp);
            imageSequence.setDirectory(strDirectory(evalList));
            imageSequence.push_back(strFilename(evalList));
        }
    }

    if (bVisualize) {
        cvNamedWindow(WINDOW_NAME, 1);
    }

    // iterate over training images
    SVL_LOG(SVL_LOG_MESSAGE, "Processing " << imageSequence.size() << " images...");
    svlObject2dSequence detections;
    for (int i = 0; i < (int)imageSequence.numImages(); i++) {
        string baseName = strBaseName(imageSequence[i]);
        SVL_LOG(SVL_LOG_VERBOSE, "processing image " << baseName << "...");

        string filename = imageSequence.directory() + '/' + imageSequence[i];
        IplImage *img = cvLoadImage(filename.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
        if (img == NULL) {
            SVL_LOG(SVL_LOG_ERROR, "could not load image " << imageSequence[i]);
            continue;
        }

        if (inputDetections != NULL) {
            svlObject2dFrame frame = candidates[baseName];
            SVL_LOG(SVL_LOG_VERBOSE, "processing " << frame.size() << " windows...");
            vector<double> scores = detector.evaluateWindows(img, frame);

            for (int j = 0; j < (int)scores.size(); j++) {
                frame[j].pr = scores[j];
            }
            detections[baseName] = sortObjects(frame);
        } else {
            if (baseScale != 1.0) {
                resizeInPlace(&img, (int)(baseScale * img->height), (int)(baseScale * img->width));
                detections[baseName] = sortObjects(detector.evaluateImage(img));
                scaleObject2dFrame(detections[baseName], 1.0 / baseScale);
            } else {
                detections[baseName] = sortObjects(detector.evaluateImage(img));
            }
        }

        SVL_LOG(SVL_LOG_DEBUG, "...processed " << detections[baseName].size() << " windows");

        if (bVisualize) {
            const svlObject2dFrame& frame = detections[baseName];
            IplImage *canvas = colorImage(img);
            for (int j = (int)frame.size() - 1; j >= 0; j--) {
                double rank = (double)j / (double)frame.size();
                cvRectangle(canvas, cvPoint((int)frame[j].x, (int)frame[j].y),
                    cvPoint((int)(frame[j].x + frame[j].w - 1), 
                        (int)(frame[j].y + frame[j].h - 1)), 
                    CV_RGB(0, (unsigned char)(255 * (1.0 - rank)), 0), 1);
            }

            cvShowImage(WINDOW_NAME, canvas);
            int ch = cvWaitKey((i == 0) && (imageSequence.numImages() == 1) ? -1 : 10);
            cvReleaseImage(&canvas);
            if (ch == (int)'q') break;
        }

        cvReleaseImage(&img);
    }

    // write output
    if (outFilename != NULL) {
        SVL_LOG(SVL_LOG_MESSAGE, "writing results to " << outFilename << "...");
        detections.write(outFilename);
    }

    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

