/* 198 *****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneEvalPixelModel.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for evaluating the pixel-based scene understanding model.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneVisualization.h"

#define WINDOW_NAME "sceneModel"
#define SINGLETON_WND "sceneModelSingletons"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneEvalPixelModel [OPTIONS] (<evalList> | <baseName>)\n";
    cerr << "OPTIONS:\n"
         << "  -outRegions <ext> :: extension for region id output (default: none)\n"
         << "  -outClasses <ext> :: extension for region class output (default: none)\n"
         << "  -outHorizon <ext> :: extension for horizon output (default: none)\n"
         << "  -outImages <ext>  :: extension for visualization output (default: none)\n"
         << "  -outSingletons <ext> :: extension for singleton potentials (default: none)\n"
         << "  -outSingletonImages <ext> :: extension for singleton visualizations (default: none)\n"
         << "  -x                :: visualize output\n"
         << "  -xm               :: visualize singleton marginals\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    // read commandline parameters
    const char *outRegionsExt = NULL;
    const char *outClassesExt = NULL;
    const char *outHorizonExt = NULL;
    const char *outImagesExt = NULL;
    const char *outSingletonExt = NULL;
    const char *outSingletonImageExt = NULL;
    bool bVisualize = false;
    bool bVisualizeSingletons = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-outRegions", outRegionsExt)
        SVL_CMDLINE_STR_OPTION("-outClasses", outClassesExt)
        SVL_CMDLINE_STR_OPTION("-outHorizon", outHorizonExt)
        SVL_CMDLINE_STR_OPTION("-outImages", outImagesExt)
        SVL_CMDLINE_STR_OPTION("-outSingletons", outSingletonExt)
        SVL_CMDLINE_STR_OPTION("-outSingletonImages", outSingletonImageExt)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
        SVL_CMDLINE_BOOL_OPTION("-xm", bVisualizeSingletons)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *evalList = SVL_CMDLINE_ARGV[0];
    gSceneConfig.setOption("useGeometry", false);
    
    // read list of evaluation images
    vector<string> baseNames;
    if (svlFileExists(evalList)) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading evaluation list from " << evalList << "...");
        baseNames = svlReadFile(evalList);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");
    } else {
        baseNames.push_back(string(evalList));
    }

    // read models
    ScenePixelModel *model = new ScenePixelModel();
    SVL_ASSERT(model != NULL);

    model->loadModels(gSceneConfig.filebase("modelsDir", "").c_str());

    svlConfusionMatrix regionConfusion(gSceneConfig.numRegionLabels);

    // process images
    SVL_LOG(SVL_LOG_MESSAGE, "Processing image features...");
    int hProcessImage = svlCodeProfiler::getHandle("processImage");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        svlCodeProfiler::tic(hProcessImage);
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");

        // load image
        IplImage *img = cvLoadImage(imgFilename.c_str(), CV_LOAD_IMAGE_COLOR);
        SVL_ASSERT_MSG(img != NULL, "path: " << imgFilename);
        SVL_LOG(SVL_LOG_MESSAGE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*img));

        // load labels
        CvMat *classLabels = NULL;
        string lblFilename = gSceneConfig.filename("lblDir", baseNames[i], "regionLblExt");
        classLabels = cvCreateMat(img->height, img->width, CV_32SC1);
        SVL_ASSERT(classLabels != NULL);
        bool bSuccess = readMatrix(classLabels, lblFilename.c_str());
        if (!bSuccess) {
            cvReleaseMat(&classLabels);
            SVL_LOG(SVL_LOG_ERROR, "Could not read class labels for " << baseNames[i]);
        }

        // compute initial singletons
        SceneInstance instance(img, baseNames[i].c_str());
        SVL_LOG(SVL_LOG_MESSAGE, "inferring first stage (pixel crf model)...");
        model->inferPixels(&instance); // compared inference algorithm
        model->inferRegions(&instance);

        // show results
        if (bVisualize) {
            IplImage *canvas = visualize(&instance);
            cvNamedWindow(WINDOW_NAME, 1);
            cvShowImage(WINDOW_NAME, canvas);
            cvWaitKey(100);
            cvReleaseImage(&canvas);
        }

        if (bVisualizeSingletons) {
            //IplImage *canvas = visualizeRegionSingletons(&instance);
            IplImage *canvas = visualizePixelEntropy(&instance);
            cvNamedWindow(SINGLETON_WND, 1);
            cvShowImage(SINGLETON_WND, canvas);
            cvWaitKey(-1);
            cvReleaseImage(&canvas);
        }

        CvMat *predictedClasses = instance.getPixelClasses();

        // accumulate predicted classes
        if (classLabels != NULL) {
            for (int y = 0; y < instance.height(); y++) {
                for (int x = 0; x < instance.width(); x++) {
                    regionConfusion.accumulate(CV_MAT_ELEM(*classLabels, int, y, x),
                        CV_MAT_ELEM(*predictedClasses, int, y, x));
                }
            }
        }

        // output images
        if (outImagesExt) {
            IplImage *canvas = visualize(&instance);
            cvSaveImage((gSceneConfig.filebase("outputDir", baseNames[i]) +
                string(outImagesExt)).c_str(), canvas);
            cvReleaseImage(&canvas);
        }

        // output regions
        if (outRegionsExt) {
            writeMatrix(instance.assignment.pixels, 
                (gSceneConfig.filebase("outputDir", baseNames[i]) + 
                    string(outRegionsExt)).c_str());
        }

        // output classes
        if (outClassesExt) {
            writeMatrix(predictedClasses, (gSceneConfig.filebase("outputDir",
                baseNames[i]) + string(outClassesExt)).c_str());
        }

        // output horizon
        if (outHorizonExt) {
            string filename = gSceneConfig.filebase("outputDir", baseNames[i]) +
                string(outHorizonExt);
            ofstream ofs(filename.c_str());
            SVL_ASSERT(!ofs.fail());
            ofs << instance.assignment.horizon << "\n";
            ofs.close();
        }

        // output singletons
        if (outSingletonExt) {
            for (int j = 0; j < gSceneConfig.numRegionLabels; j++) {
                string filename = gSceneConfig.filebase("outputDir", baseNames[i]) +
                    string(".") + toString(j) + string(outSingletonExt);
                ofstream ofs(filename.c_str());
                SVL_ASSERT(!ofs.fail());
                int varIndx = 0;
                for (int y = 0; y < instance.height(); y++) {
                    for (int x = 0; x < instance.width(); x++, varIndx++) {
                        if (x != 0) ofs << " ";
                        ofs << instance.pixelSingletons[varIndx][j];
                    }
                    ofs << "\n";
                }
                ofs.close();
            }
        }

        // output singleton images
        if (outSingletonImageExt) {
            IplImage *canvas = visualizeRegionSingletons(&instance);
            cvSaveImage((gSceneConfig.filebase("outputDir", baseNames[i]) +
                string(outSingletonImageExt)).c_str(), canvas);
            cvReleaseImage(&canvas);            
        }

        // compute score for this image
        if (classLabels != NULL) {
            cvSub(classLabels, predictedClasses, predictedClasses);
            int errors = cvCountNonZero(predictedClasses);
            SVL_LOG(SVL_LOG_MESSAGE, "...region class score: " << (1.0 - (double)errors / (img->width * img->height)));
        }

        // free memory
        cvReleaseMat(&predictedClasses);
        if (classLabels != NULL)
            cvReleaseMat(&classLabels);
        cvReleaseImage(&img);
        svlCodeProfiler::toc(hProcessImage);
    }

    if (svlLogger::getLogLevel() <= SVL_LOG_VERBOSE) {
        regionConfusion.printRowNormalized(cout, "--- Class Confusion Matrix ---");
    }
    SVL_LOG(SVL_LOG_VERBOSE, "Overall class accuracy: " << regionConfusion.accuracy());

    if ((baseNames.size() == 1) && (bVisualize || bVisualizeSingletons)) {
        cvWaitKey(-1);
    }

    delete model;
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

