/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneExplorer.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <limits>
#include <iostream>
#include <fstream>
#include <deque>

#include "wx/wx.h"
#include "wx/utils.h"
#include "wx/wxprec.h"
#include "wx/cmdline.h"
#include "wx/aboutdlg.h"

#include "sceneExplorer.h"
#include "sceneVisualization.h"

using namespace std;

// set to zero for automatic sizing
#define CANVAS_WIDTH 640
#define CANVAS_HEIGHT 480

#define NOT_IMPLEMENTED_YET wxMessageBox(_T("Functionality not implementet yet."),\
	_T("Error"), wxOK | wxICON_EXCLAMATION, this);

// Global Variables and Tables -------------------------------------------------

MainWindow *gMainWindow = NULL;

// Logger Callbacks ------------------------------------------------------------

void messageCallback(const char *message)
{
    if (gMainWindow != NULL) {
        gMainWindow->_sceneText->AppendText(message);
        gMainWindow->_sceneText->AppendText("\n");
    } else {
        cout << "--- " << message << "\n";
    }
}

void warningCallback(const char *message)
{
    if (gMainWindow != NULL) {
        gMainWindow->_sceneText->SetDefaultStyle(wxTextAttr(*wxBLUE));
        gMainWindow->_sceneText->AppendText(message);
        gMainWindow->_sceneText->AppendText("\n");
        gMainWindow->_sceneText->SetDefaultStyle(wxTextAttr(*wxBLACK));
    } else {
        cerr << "-W- " << message << "\n";
    }
}

void errorCallback(const char *message)
{
    if (gMainWindow != NULL) {
        gMainWindow->_sceneText->SetDefaultStyle(wxTextAttr(*wxRED));
        gMainWindow->_sceneText->AppendText(message);
        gMainWindow->_sceneText->AppendText("\n");
        gMainWindow->_sceneText->SetDefaultStyle(wxTextAttr(*wxBLACK));
    } else {
        wxMessageBox(_T(message), _T("Error"), wxOK | wxICON_EXCLAMATION, NULL);
    }
}

// Event Tables ----------------------------------------------------------------

BEGIN_EVENT_TABLE(SegProposalMoveDialog, wxDialog)
    EVT_ERASE_BACKGROUND(SegProposalMoveDialog::on_erase_background)
    EVT_PAINT(SegProposalMoveDialog::on_paint)
    EVT_MOUSE_EVENTS(SegProposalMoveDialog::on_mouse)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(MainCanvas, wxWindow)
    EVT_ERASE_BACKGROUND(MainCanvas::on_erase_background)
    EVT_SIZE(MainCanvas::on_size)
    EVT_PAINT(MainCanvas::on_paint)
    EVT_CHAR(MainCanvas::on_key)
    EVT_MOUSE_EVENTS(MainCanvas::on_mouse)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(MainWindow, wxFrame)
    EVT_CLOSE(MainWindow::on_close)

    EVT_MENU(FILE_NEW, MainWindow::on_file_menu)
    EVT_MENU(FILE_OPEN, MainWindow::on_file_menu)
    EVT_MENU(FILE_OPEN_REGIONS, MainWindow::on_file_menu)
    EVT_MENU(FILE_CONFIGURATION, MainWindow::on_file_menu)
    EVT_MENU(FILE_EXIT, MainWindow::on_file_menu)

    EVT_MENU(EDIT_UNDO, MainWindow::on_edit_menu)
    EVT_MENU(EDIT_POLYGON, MainWindow::on_edit_menu)

    EVT_MENU(MODEL_INFER_REGIONS, MainWindow::on_model_menu)
    EVT_MENU(MODEL_INFER_HORIZON, MainWindow::on_model_menu)

    EVT_MENU(VIEW_IMAGE, MainWindow::on_view_menu)
    EVT_MENU(VIEW_REGIONS, MainWindow::on_view_menu)
    EVT_MENU(VIEW_OBJECTS, MainWindow::on_view_menu)
    EVT_MENU(VIEW_BOUNDARIES, MainWindow::on_view_menu)
    EVT_MENU(VIEW_HORIZON, MainWindow::on_view_menu)
    EVT_MENU(VIEW_SEGMOVES, MainWindow::on_view_menu)

    EVT_MENU(HELP_ABOUT, MainWindow::on_help_menu)
END_EVENT_TABLE()

// TableDialog Implementation ---------------------------------------------------

TableDialog::TableDialog(wxWindow *parent, const wxString& title, const wxPoint& pos,
    const wxSize& size, long style, const wxString& name) :
    wxDialog(parent, wxID_ANY, title, pos, size, style, name), _grid(NULL)
{
    _grid = new wxGrid(this, wxID_ANY);
    _grid->CreateGrid(gSceneConfig.numObjectLabels + 1,
        gSceneConfig.numRegionLabels);

    for (int i = 0; i < gSceneConfig.numRegionLabels; i++) {
        _grid->SetColLabelValue(i, gSceneRegionDefs.name(i).c_str());
    }

    _grid->SetRowLabelValue(0, "other");
    for (int i = 0; i < gSceneConfig.numObjectLabels; i++) {
        _grid->SetRowLabelValue(i + 1, gSceneObjectDefs.name(i).c_str());
    }
}

TableDialog::~TableDialog()
{
    delete _grid;
}

void TableDialog::setRowHeaders(const vector<string>& str)
{
    _grid->CreateGrid(str.size(), _grid->GetNumberCols());
    for (unsigned i = 0; i < str.size(); i++) {
        _grid->SetRowLabelValue(i, str[i].c_str());
    }
}

void TableDialog::setColHeaders(const vector<string>& str)
{
    _grid->CreateGrid(_grid->GetNumberRows(), str.size());
    for (unsigned i = 0; i < str.size(); i++) {
        _grid->SetColLabelValue(i, str[i].c_str());
    }
}

void TableDialog::setData(const vector<double>& data)
{
    double maxVal = maxElem(data);
    _grid->ClearGrid();

    int indx = 0;
    for (int r = 0; r < _grid->GetNumberRows(); r++) {
        for (int c = 0; c < _grid->GetNumberCols(); c++) {
            _grid->SetCellTextColour(r, c, (data[indx] == maxVal) ? *wxRED : *wxBLACK);
            _grid->SetCellValue(r, c, toString(data[indx++]));
        }
    }
}

void TableDialog::setData(const vector<double>& data, const vector<double>& auxData)
{
    double maxVal = maxElem(data);
    _grid->ClearGrid();

    int indx = 0;
    for (int r = 0; r < _grid->GetNumberRows(); r++) {
        for (int c = 0; c < _grid->GetNumberCols(); c++, indx++) {
            _grid->SetCellTextColour(r, c, (data[indx] == maxVal) ? *wxRED : *wxBLACK);
            if (isfinite(auxData[indx])) {
                _grid->SetCellValue(r, c, toString(data[indx]) + string(" (") + 
                    toString(auxData[indx]) + string(")"));
            } else {
                _grid->SetCellValue(r, c, toString(data[indx]));
            }
        }
    }
}

void TableDialog::clearData()
{
    _grid->ClearGrid();
}

// SegProposalMoveDialog Class ------------------------------------------------

SegProposalMoveDialog::SegProposalMoveDialog(wxWindow *parent) :
    wxDialog(parent, wxID_ANY, wxT("Proposal Moves"), wxDefaultPosition, 
        wxSize(960, 240), wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER, wxPanelNameStr),
    _segmentations(NULL), _imageBuffer(NULL)
{
    // do nothing
}
    
SegProposalMoveDialog::~SegProposalMoveDialog()
{
    if (_imageBuffer != NULL)
        delete _imageBuffer;
}

void SegProposalMoveDialog::setSegmentations(svlMultiSeg *pSegDict, const IplImage *pImg)
{
    SVL_ASSERT((pSegDict != NULL) && (pImg != NULL));

    _segmentations = pSegDict;
    vector<IplImage *> views;
    for (int i = 0; i < _segmentations->count(); i++) {
        views.push_back(cvCloneImage(pImg));
        cvConvertScale(views.back(), views.back(), 0.5, 127);
        cvRectangle(views.back(), cvPoint(0, 0), 
            cvPoint(views.back()->width, views.back()->height), CV_RGB(255, 0, 0));

        int *p = (int *)CV_MAT_ELEM_PTR(*_segmentations->segmentation(i), 0, 0);
        for (int y = 0; y < _segmentations->height(); y++) {
            for (int x = 0; x < _segmentations->width(); x++, p++) {
                if (((y > 0) && (*p != *(p - _segmentations->width()))) ||
                    ((x > 0) && (*p != *(p - 1)))) {
                    views.back()->imageData[y * views.back()->widthStep + 3 * x + 2] = 0xff;
                    views.back()->imageData[y * views.back()->widthStep + 3 * x + 1] = 0x00;
                    views.back()->imageData[y * views.back()->widthStep + 3 * x + 0] = 0x00;
                }
            }
        }
    }

    IplImage *canvas = combineImages(views, 1, views.size());
    _imageBuffer = new unsigned char[3 * canvas->width * canvas->height];
    
    for (int y = 0; y < canvas->height; y++) {
        for (int x = 0; x < canvas->width; x++) {
            _imageBuffer[3 * (y * canvas->width + x) + 0] =
                canvas->imageData[y * canvas->widthStep + 3 * x + 2];
            _imageBuffer[3 * (y * canvas->width + x) + 1] =
                canvas->imageData[y * canvas->widthStep + 3 * x + 1];
            _imageBuffer[3 * (y * canvas->width + x) + 2] =
                canvas->imageData[y * canvas->widthStep + 3 * x + 0];
        }
    }
    
    cvReleaseImage(&canvas);
    svlReleaseImages(views);
}

void SegProposalMoveDialog::on_erase_background(wxEraseEvent &event)
{
    // do nothing (and avoid flicker)
}

void SegProposalMoveDialog::on_paint(wxPaintEvent &event)
{
    int width, height;
    GetClientSize(&width, &height);

    wxPaintDC dc(this);
    if (_imageBuffer == NULL) {
        dc.Clear();
        dc.SetTextForeground(wxColor(255, 0, 0));
        wxSize s = dc.GetTextExtent("no segments");
        dc.DrawText("no segments", (int)(width - s.x)/2, (int)(height - s.y)/2);
        return;        
    }

    // copy frame to screen
    SVL_ASSERT(_imageBuffer != NULL);
    wxImage screenImage(_segmentations->width() * _segmentations->count(), 
        _segmentations->height());
    screenImage.SetData(_imageBuffer, true);
    dc.DrawBitmap(screenImage.Scale(width, height), 0, 0);
}

void SegProposalMoveDialog::on_mouse(wxMouseEvent &event)
{
    if (_segmentations == NULL)
        return;

    int width, height;
    GetClientSize(&width, &height);
    double scaleX = (double)(_segmentations->width() * _segmentations->count()) / (double)width;
    double scaleY = (double)_segmentations->height() / (double)height;

    int segId = (int)(scaleX * event.m_x / _segmentations->width());
    int x = (int)(scaleX * event.m_x) % _segmentations->width();
    int y = std::min((int)(scaleY * event.m_y), _segmentations->height() - 1);
        
    if (event.LeftDown()) {        
        ((MainCanvas *)this->GetParent())->newRegion(_segmentations->pixels(segId,
                CV_MAT_ELEM(*_segmentations->segmentation(segId), int, y, x)));
    }

    event.Skip();
}

// MainCanvas Implementation ---------------------------------------------------

MainCanvas::MainCanvas(wxWindow *parent, wxWindowID id,const wxPoint& pos,
    const wxSize& size, long style, const wxString& name) :
    wxWindow(parent, id, pos, size, style, name),
    _model(NULL), _instance(NULL), _energy(0.0), _undoAssignment(NULL), _imageBuffer(NULL),
    _lastScenePoint(-1, -1), _downScenePoint(-1, -1), _viewMode(VM_IMAGE), _mouseMode(MM_NONE),
    _regionSingletons(NULL), _proposalMovesDialog(NULL)
{
    SetMinSize(wxSize(320, 240));
}

MainCanvas::~MainCanvas()
{
    if (_model != NULL) {
        delete _model;
    }

    if (_instance != NULL) {
        delete _instance;
        delete[] _imageBuffer;
    }

    if (_undoAssignment != NULL) {
        delete _undoAssignment;
    }

    if (_regionSingletons != NULL)
        delete _regionSingletons;

    if (_proposalMovesDialog != NULL)
        delete _proposalMovesDialog;
}

void MainCanvas::on_erase_background(wxEraseEvent &event)
{
    // do nothing (and avoid flicker)
}

void MainCanvas::on_paint(wxPaintEvent &WXUNUSED(event))
{
    int width, height;
    GetClientSize(&width, &height);

    wxPaintDC dc(this);
    if (_instance == NULL) {
        dc.Clear();
        dc.SetTextForeground(wxColor(255, 0, 0));
        wxSize s = dc.GetTextExtent("no scene");
        dc.DrawText("no scene", (int)(width - s.x)/2, (int)(height - s.y)/2);
        return;
    }

    // viewing nothing
    if (_instance == NULL) {
        dc.Clear();
        dc.SetTextForeground(wxColor(0, 0, 255));
        wxSize s = dc.GetTextExtent(_instance->name().c_str());
        dc.DrawText(_instance->name().c_str(), (int)(width - s.x)/2, (int)(height - s.y)/2);
        return;
    }

    // copy frame to screen
    SVL_ASSERT(_imageBuffer != NULL);
#if ((CANVAS_WIDTH <= 0) || (CANVAS_HEIGHT <= 0))
    wxImage screenImage(_instance->width(), _instance->height());
#else
    wxImage screenImage(640, 480);
#endif
    screenImage.SetData(_imageBuffer, true);
    dc.DrawBitmap(screenImage.Scale(width, height), 0, 0);

    // draw polygon move
    if ((_mouseMode == MM_POLYGON) && !_polygon.empty()) {
    	double scaleX = (double)width / (double)_instance->width();
	double scaleY = (double)height / (double)_instance->height();

        dc.SetPen(wxPen(wxColor(255, 0, 0), 2));
	for (unsigned i = 1; i < _polygon.size(); i++) {
	    dc.DrawLine((int)(scaleX * _polygon[i - 1].x), (int)(scaleY * _polygon[i - 1].y),
		(int)(scaleX * _polygon[i].x), (int)(scaleY * _polygon[i].y));
	}
	dc.DrawLine((int)(scaleX * _polygon.back().x), (int)(scaleY * _polygon.back().y),
	    (int)(scaleX * _lastScenePoint.x), (int)(scaleY * _lastScenePoint.y));
    }   


}

void MainCanvas::on_size(wxSizeEvent &event)
{
    int width, height;

    GetClientSize(&width, &height);

    this->Refresh(false);
    this->Update();
}

void MainCanvas::on_key(wxKeyEvent &event)
{
    switch (event.m_keyCode) {
    case WXK_ESCAPE:
        _polygon.clear();
    	break;
    default:
    	event.Skip();
    }

    // refresh view
    updateStatusBar();
    this->Refresh(false);
    this->Update();
}

void MainCanvas::on_mouse(wxMouseEvent &event)
{
    if (_instance == NULL)
	return;

    int width, height;
    GetClientSize(&width, &height);
    double scaleX = (double)_instance->width() / (double)width;
    double scaleY = (double)_instance->height() / (double)height;

    wxPoint scenePoint = wxPoint((int)(scaleX * event.m_x), (int)(scaleY * event.m_y));

    if (event.LeftUp()) {
	if (_mouseMode == MM_POLYGON) {            
	    _polygon.push_back(scenePoint);
        } else if (validScenePoint(_downScenePoint) && validScenePoint(scenePoint)) {
            int regIdA = _instance->assignment.regionId(_downScenePoint.x, _downScenePoint.y);
            int regIdB = _instance->assignment.regionId(scenePoint.x, scenePoint.y);
            if (_instance->assignment.regions[regIdA]->isNeighbor(_instance->assignment.regions[regIdB])) {
                undoableAction();
                _instance->movePixels(regIdA, regIdB);
                _instance->splitOrphans();
                _instance->assignment.remap();
                updateRegions();
                updateTables(_instance->assignment.regionId(scenePoint.x, scenePoint.y));
            }
        }
    } if (event.LeftDown()) {
        _downScenePoint = scenePoint;
        if (_mouseMode == MM_NEW_POLYGON) {
            _mouseMode = MM_POLYGON;
        }
    } else if (event.LeftDClick()) {
        if ((_mouseMode == MM_POLYGON) && (_polygon.size() > 1)) {
            // design proposal move
            _polygon.push_back(scenePoint);
            //CvMat *mask = cvCreateMat(_instance->height(), _instance->width(), CV_8UC1);
            IplImage *mask = cvCreateImage(cvSize(_instance->width(), _instance->height()), IPL_DEPTH_8U, 1);
            cvZero(mask);
	    wxPoint mu = _polygon.front();
	    for (unsigned i = 1; i < _polygon.size(); i++) {
		mu.x += _polygon[i].x;
		mu.y += _polygon[i].y;
		cvLine(mask, cvPoint(_polygon[i - 1].x, _polygon[i - 1].y), 
                    cvPoint(_polygon[i].x, _polygon[i].y), CV_RGB(255, 255, 255), 2);
	    }
            cvLine(mask, cvPoint(_polygon.back().x, _polygon.back().y), 
                cvPoint(_polygon.front().x, _polygon.front().y), CV_RGB(255, 255, 255), 2);
            mu.x /= _polygon.size();
            mu.y /= _polygon.size();

#if 0
            cvNamedWindow("debug", 1);
            cvCircle(mask, cvPoint(mu.x, mu.y), 1, CV_RGB(255, 255, 255));
            cvShowImage("debug", mask);
            cvWaitKey(100);
            cvCircle(mask, cvPoint(mu.x, mu.y), 1, CV_RGB(0, 0, 0));
#endif            

            cvFloodFill(mask, cvPoint(mu.x, mu.y), cvScalar(1));

            CvMat *r = cvCloneMat(_instance->assignment.pixels);
            cvSet(r, cvScalar(-1), mask);

            // make move
	    undoableAction();
            _instance->movePixels(r);

            cvReleaseMat(&r);
	    //cvReleaseMat(&mask);
            cvReleaseImage(&mask);

            // clear points
            _polygon.clear();
            _mouseMode = MM_NEW_POLYGON;

            _instance->splitOrphans();
            _instance->assignment.remap();
            updateRegions();
            if (validScenePoint(mu)) {
                updateTables(_instance->assignment.regionId(mu.x, mu.y));
            } else {
                updateTables();
            }
        } else if (validScenePoint(scenePoint)) {
            int regionId = _instance->assignment.regionId(scenePoint.x, scenePoint.y);
            wxString title(string("Region ") + toString(regionId));
            if (_regionSingletons == NULL) {
                _regionSingletons = new TableDialog(NULL, title);
            } else {
                _regionSingletons->SetTitle(title);
            }
            updateTables(regionId);
            _regionSingletons->Show(true);
        }
    } else if (event.Dragging()) {
        // to do
	updateImageBuffer();
    } else if (event.Moving()) {
        // to do
    }

    _lastScenePoint = scenePoint;

    event.Skip();
    updateStatusBar();
    this->Refresh(false);
    this->Update();
}

void MainCanvas::undo()
{
    if (_undoAssignment == NULL)
        return;

    _instance->assignment.swap(*_undoAssignment);
    updateRegions();
    if (_regionSingletons != NULL)
        _regionSingletons->clearData();
}

void MainCanvas::loadModel(const char *modelDir)
{
    if (_model != NULL) {
        delete _model;
    }
    _model = new SceneRegionModel();
    _model->loadModels(modelDir);
    if (_regionSingletons != NULL)
        _regionSingletons->Show(false);

    if (_instance != NULL) {
        _model->initializeInstance(_instance);
        // first-stage inference
        ((MainWindow *)GetParent())->SetStatusText("inferring pixels...");
        _energy = _model->inferPixels(_instance);
        
        // second-stage inference
        ((MainWindow *)GetParent())->SetStatusText("inferring horizon...");
        _model->inferHorizon(_instance);
        
        // third-stage inference
        updateRegions();
    }
}

void MainCanvas::loadInstance(const char *imgFilename)
{
    if (_instance != NULL) {
        delete _instance;
        delete[] _imageBuffer;
        if (_undoAssignment != NULL) {
            delete _undoAssignment;
            _undoAssignment = NULL;
        }
    }
    if (_regionSingletons != NULL)
        _regionSingletons->Show(false);
    if (_proposalMovesDialog != NULL) {
        delete _proposalMovesDialog;
        _proposalMovesDialog = NULL;
    }

    _instance = new SceneInstance(imgFilename);
    _instance->loadSegmentations(_instance->name().c_str());
#if ((CANVAS_WIDTH <= 0) || (CANVAS_HEIGHT <= 0))
    _imageBuffer = new unsigned char[3 * _instance->width() * _instance->height()];
#else
    _imageBuffer = new unsigned char[3 * CANVAS_WIDTH * CANVAS_HEIGHT];
#endif

    // first-stage inference
    ((MainWindow *)GetParent())->SetStatusText("inferring pixels...");
    _energy = _model->inferPixels(_instance);

    // second-stage inference
    ((MainWindow *)GetParent())->SetStatusText("inferring horizon...");
    _model->inferHorizon(_instance);

    // third-stage inference
    updateRegions();
}

void MainCanvas::loadRegions(const char *regFilename)
{
    SVL_ASSERT(_instance != NULL);

    CvMat *regions = cvCreateMat(_instance->height(), _instance->width(), CV_32SC1);
    bool bSuccess = readMatrix(regions, regFilename);
    if (!bSuccess) {
        SVL_LOG(SVL_LOG_ERROR, "Could not load region file " << regFilename);
        cvReleaseMat(&regions);
        return;
    }

    if (_regionSingletons != NULL)
        _regionSingletons->Show(false);

    undoableAction();
    _instance->assignPixelsFromLabels(regions);
    cvReleaseMat(&regions);

    updateRegions();
}

void MainCanvas::clearRegions()
{
    SVL_ASSERT(_instance != NULL);

    undoableAction();
    CvMat *r = cvCreateMat(_instance->height(), _instance->width(), CV_32SC1);
    cvZero(r);
    _instance->movePixels(r);
    cvReleaseMat(&r);

    updateRegions();
    this->Refresh(false);
    this->Update();    
}

void MainCanvas::newRegion(const vector<CvPoint>& pixels)
{
    SVL_ASSERT(_instance != NULL);

    undoableAction();
    _instance->movePixels(pixels, -1);
    _instance->splitOrphans();
    updateRegions();
    this->Refresh(false);
    this->Update();
}

void MainCanvas::inferRegions()
{
    SVL_ASSERT(_instance != NULL);
    
    undoableAction();
    updateRegions();
    this->Refresh(false);
    this->Update();
}

void MainCanvas::inferHorizon()
{
    SVL_ASSERT(_instance != NULL);
    
    undoableAction();
    _model->inferHorizon(_instance);
    updateRegions();
    this->Refresh(false);
    this->Update();
}

void MainCanvas::setViewMode(unsigned v)
{
    _viewMode |= v;
    updateImageBuffer();
    this->Refresh(false);
    this->Update();
}

void MainCanvas::clearViewMode(unsigned v)
{
    _viewMode &= ~v;
    updateImageBuffer();
    this->Refresh(false);
    this->Update();
}

void MainCanvas::setMouseMode(TMouseMode m)
{
    _mouseMode = m;
    _polygon.clear();
}

void MainCanvas::clearMouseMode()
{
    setMouseMode(MM_NONE);
}

void MainCanvas::showProposalMoves(bool bShow)
{
    if (bShow && (_instance != NULL) && (_proposalMovesDialog == NULL)) {
        _proposalMovesDialog = new SegProposalMoveDialog(this);
        _proposalMovesDialog->setSegmentations(&_instance->segmentations,
            _instance->image());
    }

    if (_proposalMovesDialog != NULL) {
        _proposalMovesDialog->Show(bShow);
    }
}

void MainCanvas::undoableAction()
{
    if (_instance == NULL) return;
    if (_undoAssignment == NULL) {
        _undoAssignment = new SceneAssignment(_instance->assignment);
    } else {
        *_undoAssignment = _instance->assignment;
    }
}

void MainCanvas::updateRegions()
{
    ((MainWindow *)GetParent())->SetStatusText("inferring regions...");
    wxSetCursor(*wxHOURGLASS_CURSOR);
    double e = _model->inferRegions(_instance);
    *gMainWindow->_sceneText << "  energy: ";
    if (e < _energy) {
        gMainWindow->_sceneText->SetDefaultStyle(wxTextAttr(wxTheColourDatabase->Find("FOREST GREEN")));
    } else {
        gMainWindow->_sceneText->SetDefaultStyle(wxTextAttr(wxTheColourDatabase->Find("MAROON")));
    }
    *gMainWindow->_sceneText << e << "\n";
    gMainWindow->_sceneText->SetDefaultStyle(wxTextAttr(*wxBLACK));    
    _energy = e;
    updateImageBuffer();
    ((MainWindow *)GetParent())->SetStatusText("");
    wxSetCursor(*wxSTANDARD_CURSOR);
}

void MainCanvas::updateImageBuffer()
{
    if (_imageBuffer == NULL)
	return;

    IplImage *canvas = cvCloneImage(_instance->image());
#if ((CANVAS_WIDTH > 0) && (CANVAS_HEIGHT > 0))
    resizeInPlace(&canvas, CANVAS_HEIGHT, CANVAS_WIDTH);
#endif

    if (!(_viewMode & VM_IMAGE)) {
        cvZero(canvas);
    }
    if (_viewMode & VM_REGIONS) {
        visualizeClasses(_instance, canvas, (_viewMode & VM_IMAGE) ? 0.5 : 1.0);
    }
    if (_viewMode & VM_BOUNDARIES) {
        visualizeBoundaries(_instance, canvas);
    }
    if (_viewMode & VM_OBJECTS) {
        visualizeObjects(_instance, canvas);
    }
    if (_viewMode & VM_HORIZON) {
        visualizeHorizon(_instance, canvas);
    }

    for (int y = 0; y < canvas->height; y++) {
        for (int x = 0; x < canvas->width; x++) {
            _imageBuffer[3 * (y * canvas->width + x) + 0] =
                canvas->imageData[y * canvas->widthStep + 3 * x + 2];
            _imageBuffer[3 * (y * canvas->width + x) + 1] =
                canvas->imageData[y * canvas->widthStep + 3 * x + 1];
            _imageBuffer[3 * (y * canvas->width + x) + 2] =
                canvas->imageData[y * canvas->widthStep + 3 * x + 0];
        }
    }

    cvReleaseImage(&canvas);
}

void MainCanvas::updateStatusBar()
{
    wxString status;

    if (validScenePoint(_lastScenePoint)) {
        int regionId = _instance->assignment.regionId(_lastScenePoint.x, _lastScenePoint.y);
        status = wxString::Format("(%d, %d): region %d of %d",
            _lastScenePoint.x, _lastScenePoint.y, regionId, _instance->numRegions());
    }

    ((MainWindow *)GetParent())->SetStatusText(status);
}

void MainCanvas::updateTables(int regionId)
{    
    if (_regionSingletons != NULL) {
        vector<double> phi = eigenToSTL(_instance->assignment.regions[regionId]->phi);
//         vector<double> objScores(phi.size(), numeric_limits<double>::quiet_NaN());
//         for (int i = 0; i < gSceneConfig.numObjectLabels; i++) {
//             objScores[(i + 2) * gSceneConfig.numRegionLabels - 1] =
//                 _instance->regions[regionId].objDetScore[i];
//         }
        
        wxString title(string("Region ") + toString(regionId));
        _regionSingletons->SetTitle(title);
//        _regionSingletons->setData(phi, objScores);
        _regionSingletons->setData(phi);
        _regionSingletons->Show(true);
    }    
}

bool MainCanvas::validScenePoint(const wxPoint& p) const
{
    if (_instance == NULL) return false;

    return ((p.x >= 0) && (p.x < _instance->width()) &&
        (p.y >= 0) && (p.y < _instance->height()));
}


// MainWindow Implementation ---------------------------------------------------

MainWindow::MainWindow(wxWindow* parent, wxWindowID id, const wxString& title,
    const wxPoint& pos, const wxSize& size, long style) :
    wxFrame(parent, id, title, pos, size, style)
{
    wxMenu *file_menu = new wxMenu;
    wxMenu *edit_menu = new wxMenu;
    wxMenu *model_menu = new wxMenu;
    wxMenu *view_menu = new wxMenu;
    wxMenu *help_menu = new wxMenu;

    file_menu->Append(FILE_NEW, _T("&New\tCtrl-N"), _T("Clear scene instance"));
    file_menu->Append(FILE_OPEN, _T("&Open...\tCtrl-O"), _T("Open scene"));
    file_menu->Append(FILE_OPEN_REGIONS, _T("Open &Regions...\tCtrl-R"), _T("Open regions file"));
    file_menu->AppendSeparator();
    file_menu->Append(FILE_CONFIGURATION, _T("&Configuration...\tCtrl-C"), _T("Configure scene model"));
    file_menu->AppendSeparator();
    file_menu->Append(FILE_EXIT, _T("E&xit\tAlt-X"), _T("Exit this program"));

    edit_menu->Append(EDIT_UNDO, _T("&Undo\tCtrl-Z"), _T("Undo last change"));
    edit_menu->AppendSeparator();
    edit_menu->AppendCheckItem(EDIT_POLYGON, _T("&Polygon Mode"), _T("Toggle polygon move mode"));

    model_menu->Append(MODEL_INFER_REGIONS, _T("Infer &Regions"), _T("Infer region classes"));
    model_menu->Append(MODEL_INFER_HORIZON, _T("Infer &Horizon"), _T("Infer horizon and region classes"));

    view_menu->AppendCheckItem(VIEW_IMAGE, _T("View &image"));
    view_menu->AppendCheckItem(VIEW_REGIONS, _T("View &region classes"));
    view_menu->AppendCheckItem(VIEW_OBJECTS, _T("View &objects"));
    view_menu->AppendCheckItem(VIEW_BOUNDARIES, _T("View region &boundaries"));
    view_menu->AppendCheckItem(VIEW_HORIZON, _T("View &horizon"));
    view_menu->AppendSeparator();
    view_menu->AppendCheckItem(VIEW_SEGMOVES, _T("View proposal &moves"));

    help_menu->Append(HELP_ABOUT, _T("&About...\tF1"), _T("Show about dialog"));

    wxMenuBar *menu_bar = new wxMenuBar();
    menu_bar->Append(file_menu, _T("&File"));
    menu_bar->Append(edit_menu, _T("&Edit"));
    menu_bar->Append(model_menu, _T("&Model"));
    menu_bar->Append(view_menu, _T("&View"));
    menu_bar->Append(help_menu, _T("&Help"));
    SetMenuBar(menu_bar);

    view_menu->Check(VIEW_IMAGE, true);
    view_menu->Check(VIEW_REGIONS, true);

    // setup subwindows
    _subWindows = new wxGridSizer(1, 2, 1, 1);
    _subWindows->SetMinSize(wxSize(640, 240));
    _subWindows->Add(_sceneCanvas = new MainCanvas(this), 0, wxEXPAND | wxBORDER);
    _subWindows->Add(_sceneText = new wxTextCtrl(this, wxID_ANY, "", wxDefaultPosition,
            wxDefaultSize, wxTE_MULTILINE |wxTE_DONTWRAP), 0, wxEXPAND | wxBORDER); 
    SetSizer(_subWindows);

    this->CreateStatusBar();
    _sceneCanvas->setViewMode(VM_IMAGE | VM_REGIONS);
    _sceneCanvas->clearMouseMode();
}

MainWindow::~MainWindow()
{
    // do nothing
}

void MainWindow::on_file_menu(wxCommandEvent& event)
{
    if (event.GetId() == FILE_NEW) {
        wxMessageDialog dlg(this, _T("Clear scene?"),
            _T("New"), wxYES_NO | wxICON_QUESTION);
        if (dlg.ShowModal() == wxID_YES) {
	    //SetTitle("Scene Explorer");
            _sceneCanvas->clearRegions();
        }
    } else if (event.GetId() == FILE_OPEN) {
        wxFileDialog dlg(this, _T("Choose image to open"), _T(gSceneConfig.filebase("imgDir", "").c_str()),
	    _T(""), _T("Image files (*.jpg)|*.jpg|All files (*.*)|*.*"), wxOPEN | wxFD_CHANGE_DIR);
        dlg.SetDirectory(_T(gSceneConfig.filebase("imgDir", "").c_str()));
        if (dlg.ShowModal() == wxID_OK) {
            _sceneCanvas->loadInstance(dlg.GetPath().c_str());
	    SetTitle((string("Scene Explorer [") + strBaseName(dlg.GetFilename().c_str()) + string("]")).c_str());
        }
    } else if (event.GetId() == FILE_OPEN_REGIONS) {
        wxFileDialog dlg(this, _T("Choose regions to open"), _T(gSceneConfig.filebase("outputDir", "").c_str()),
	    _T(""), _T("Text files (*.txt)|*.txt|All files (*.*)|*.*"), wxOPEN | wxFD_CHANGE_DIR);
        dlg.SetDirectory(_T(gSceneConfig.filebase("outputDir", "").c_str()));
        if (dlg.ShowModal() == wxID_OK) {
            _sceneCanvas->loadRegions(dlg.GetPath().c_str());
        }
    } else if (event.GetId() == FILE_CONFIGURATION) {
        wxFileDialog dlg(this, _T("Choose configuration to open"), _T(""),
	    _T(""), _T("XML files (*.xml)|*.xml|All files (*.*)|*.*"), wxOPEN | wxFD_CHANGE_DIR);
        if (dlg.ShowModal() == wxID_OK) {
            svlConfigurationManager::get().configure(dlg.GetPath().c_str());
            _sceneCanvas->loadModel(gSceneConfig.filebase("modelsDir", "").c_str());
            SetStatusText("Scene configuration loaded");
        }
    } else if (event.GetId() == FILE_EXIT) {
        Close(true);
    }

    Refresh(false);
    Update();
}

void MainWindow::on_edit_menu(wxCommandEvent& event)
{
    if (event.GetId() == EDIT_UNDO) {
	_sceneCanvas->undo();
    } else if (event.GetId() == EDIT_POLYGON) {
        _sceneCanvas->setMouseMode(event.IsChecked() ? MM_NEW_POLYGON : MM_NONE);
    }

    Refresh(false);
    Update();
}

void MainWindow::on_model_menu(wxCommandEvent& event)
{
    if (event.GetId() == MODEL_INFER_REGIONS) {
	_sceneCanvas->inferRegions();
    } else if (event.GetId() == MODEL_INFER_HORIZON) {
        _sceneCanvas->inferHorizon();
    }

    Refresh(false);
    Update();
}

void MainWindow::on_view_menu(wxCommandEvent& event)
{
    if (event.GetId() == VIEW_IMAGE) {
        if (event.IsChecked())
            _sceneCanvas->setViewMode(VM_IMAGE);
        else _sceneCanvas->clearViewMode(VM_IMAGE);
    } else if (event.GetId() == VIEW_REGIONS) {
        if (event.IsChecked())
            _sceneCanvas->setViewMode(VM_REGIONS);
        else _sceneCanvas->clearViewMode(VM_REGIONS);
    } else if (event.GetId() == VIEW_OBJECTS) {
        if (event.IsChecked())
            _sceneCanvas->setViewMode(VM_OBJECTS);
        else _sceneCanvas->clearViewMode(VM_OBJECTS);
    } else if (event.GetId() == VIEW_BOUNDARIES) {
        if (event.IsChecked())
            _sceneCanvas->setViewMode(VM_BOUNDARIES);
        else _sceneCanvas->clearViewMode(VM_BOUNDARIES);
    } else if (event.GetId() == VIEW_HORIZON) {
        if (event.IsChecked())
            _sceneCanvas->setViewMode(VM_HORIZON);
        else _sceneCanvas->clearViewMode(VM_HORIZON);
    } else if (event.GetId() == VIEW_SEGMOVES) {
        _sceneCanvas->showProposalMoves(event.IsChecked());
    }

    Refresh(false);
    Update();
}

void MainWindow::on_help_menu(wxCommandEvent& event)
{
    if (event.GetId() == HELP_ABOUT) {
        wxAboutDialogInfo info;
        info.SetName(_T("Scene Explorer"));
        info.SetVersion(_T("0.3"));
        info.SetDescription(_T("This program allows you to explorer the energy function for a scene."));
        info.SetCopyright(_T("Copyright (C) 2009\nStephen Gould <sgould@stanford.edu>\nTianshi Gao <tianshig@stanford.edu>"));

        wxAboutBox(info);
    }
}

void MainWindow::on_close(wxCloseEvent& event)
{
    // not implemented yet
    event.Skip();
}

// SceneExplorerApp Implementation ------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneExplorer [OPTIONS] (<baseName>)?\n";
    cerr << "OPTIONS:\n"
         << "  -inRegions <ext>  :: initialize regions from file (default: none)\n"
         << "  -inDir <dir>      :: directory for region initialization (default: use configXML)\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

bool SceneExplorerApp::OnInit()
{
    // setup main window
    gMainWindow = new MainWindow(NULL, wxID_ANY, wxT("Scene Explorer"),
        wxDefaultPosition, wxSize(640, 240));
    SetTopWindow(gMainWindow);

    // parse command-line options
    const char *inRegionsExt = NULL;
    const char *inRegionDir = NULL;

    SVL_BEGIN_CMDLINE_PROCESSING(wxApp::argc, wxApp::argv)
        SVL_CMDLINE_STR_OPTION("-inRegions", inRegionsExt)
        SVL_CMDLINE_STR_OPTION("-inDir", inRegionDir)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC > 1) {
        usage();
        exit(-1);
    }

    // initialize svlLogger callbacks
    svlLogger::showMessageCallback = messageCallback;
    svlLogger::showWarningCallback = warningCallback;
    svlLogger::showErrorCallback = errorCallback;

    gMainWindow->Show();
    gMainWindow->SetFocus();
    gMainWindow->_sceneCanvas->loadModel(gSceneConfig.filebase("modelsDir", "").c_str());

    if (SVL_CMDLINE_ARGC == 1) {
        const char *baseName = SVL_CMDLINE_ARGV[0];
        gMainWindow->_sceneCanvas->loadInstance(gSceneConfig.filename("imgDir",
                baseName, "imgExt").c_str());
        if (inRegionsExt != NULL) {
            string filename;
            if (inRegionDir == NULL) {
                filename = gSceneConfig.filebase("outputDir", baseName) + string(inRegionsExt);
            } else {
                filename = string(inRegionDir) + string("/") + string(baseName) + string(inRegionsExt);
            }
            gMainWindow->_sceneCanvas->loadRegions(filename.c_str());
        }
        gMainWindow->SetTitle(wxT("Scene Explorer [") + wxString(baseName) + wxT("]"));
    }

    return true;
}

int SceneExplorerApp::OnExit()
{
    svlCodeProfiler::print(cerr);
    return 0;
}

IMPLEMENT_APP(SceneExplorerApp)

