/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneExplorer.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  GUI for displaying a scene and exploring its energy function.
**
*****************************************************************************/

#pragma once

#include <string>
#include <vector>
#include <map>

#include "wx/wx.h"
#include "wx/glcanvas.h"
#include "wx/utils.h"
#include "wx/grid.h"
#include "wx/textctrl.h"

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"

// wxWidgets Event Constants --------------------------------------------------

enum
{
    FILE_NEW = wxID_HIGHEST,
    FILE_OPEN = wxID_HIGHEST + 1,
    FILE_OPEN_REGIONS = wxID_HIGHEST + 20,
    FILE_CONFIGURATION = wxID_HIGHEST + 30,
    FILE_EXIT = wxID_EXIT,

    EDIT_UNDO = wxID_HIGHEST + 290,
    EDIT_POLYGON = wxID_HIGHEST + 295,

    MODEL_INFER_REGIONS = wxID_HIGHEST + 300,
    MODEL_INFER_HORIZON = wxID_HIGHEST + 310,

    VIEW_IMAGE = wxID_HIGHEST + 500,
    VIEW_REGIONS = wxID_HIGHEST + 510,
    VIEW_OBJECTS = wxID_HIGHEST + 515,
    VIEW_BOUNDARIES = wxID_HIGHEST + 520,
    VIEW_HORIZON = wxID_HIGHEST + 540,
    VIEW_SEGMOVES = wxID_HIGHEST + 550,

    HELP_ABOUT = wxID_ABOUT
};

// Mouse Modes ----------------------------------------------------------------

typedef enum {
    MM_NONE, MM_NEW_POLYGON, MM_POLYGON
} TMouseMode;

// View Modes -----------------------------------------------------------------

#define VM_IMAGE      0x0001
#define VM_REGIONS    0x0002
#define VM_OBJECTS    0x0004
#define VM_BOUNDARIES 0x0008
#define VM_HORIZON    0x0010

// TableDialog Class ----------------------------------------------------------

class TableDialog : public wxDialog
{
 public:
    TableDialog(wxWindow *parent, const wxString& title,
        const wxPoint& pos = wxDefaultPosition,
        const wxSize& size = wxDefaultSize,
        long style = wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER,
        const wxString& name = wxPanelNameStr);
    ~TableDialog();

    void setRowHeaders(const vector<string>& str);
    void setColHeaders(const vector<string>& str);
    void setData(const vector<double>& data);
    void setData(const vector<double>& data, const vector<double>& auxData);
    void clearData();

 protected:
    wxGrid *_grid;
};

// SegProposalMoveDialog Class ------------------------------------------------

class SegProposalMoveDialog : public wxDialog
{
 protected:
    svlMultiSeg *_segmentations;
    unsigned char *_imageBuffer;

 public:
    SegProposalMoveDialog(wxWindow *parent);
    ~SegProposalMoveDialog();

    void setSegmentations(svlMultiSeg *pSegDict, const IplImage *pImg);

    void on_erase_background(wxEraseEvent &event);
    void on_paint(wxPaintEvent &event);
    void on_mouse(wxMouseEvent &event);

    DECLARE_EVENT_TABLE()
};

// MainCanvas Class -----------------------------------------------------------
// This is required under Linux because wxFrame doesn't get keyboard focus
// without a child window.

class MainCanvas : public wxWindow
{
 public:
    MainCanvas(wxWindow *parent,
	       wxWindowID id = wxID_ANY,
	       const wxPoint& pos = wxDefaultPosition,
	       const wxSize& size = wxDefaultSize,
	       long style = wxDEFAULT_FRAME_STYLE | wxSUNKEN_BORDER | wxWANTS_CHARS,
	       const wxString& name = wxPanelNameStr);
    ~MainCanvas();

    void on_erase_background(wxEraseEvent &event);
    void on_paint(wxPaintEvent &event);
    void on_size(wxSizeEvent &event);
    void on_key(wxKeyEvent &event);
    void on_mouse(wxMouseEvent &event);

    void undo();

    void loadModel(const char *modelDir);
    void loadInstance(const char *imgFilename);
    void loadRegions(const char *regFilename);
    void clearRegions();
    void newRegion(const vector<CvPoint>& pixels);
    void inferRegions();
    void inferHorizon();

    void setViewMode(unsigned v);
    void clearViewMode(unsigned v);
    void setMouseMode(TMouseMode m);
    void clearMouseMode();

    void showProposalMoves(bool bShow);

 protected:
    void undoableAction();
    void updateImageBuffer();
    void updateStatusBar();
    void updateRegions();
    void updateTables(int regionId = -1);
    bool validScenePoint(const wxPoint& p) const;

 protected:
    SceneRegionModel *_model;
    SceneInstance *_instance;
    double _energy;
    SceneAssignment *_undoAssignment;

    unsigned char *_imageBuffer;
    wxPoint _lastScenePoint;
    wxPoint _downScenePoint;
    unsigned _viewMode;
    TMouseMode _mouseMode;
    vector<wxPoint> _polygon;

    TableDialog *_regionSingletons;
    SegProposalMoveDialog *_proposalMovesDialog;

    DECLARE_EVENT_TABLE()
};

// MainWindow Class -----------------------------------------------------------

class SceneExplorerApp;

class MainWindow : public wxFrame
{
    friend class SceneExplorerApp;
    friend class MainCanvas;
    friend void messageCallback(const char *message);
    friend void warningCallback(const char *message);
    friend void errorCallback(const char *message);

 public:
    MainWindow(wxWindow* parent,
	  wxWindowID id,
          const wxString& title,
	  const wxPoint& pos = wxDefaultPosition,
	  const wxSize& size = wxDefaultSize,
	  long style = wxDEFAULT_FRAME_STYLE | wxSUNKEN_BORDER | wxWANTS_CHARS);
    ~MainWindow();

    // event callbacks
    void on_file_menu(wxCommandEvent& event);
    void on_edit_menu(wxCommandEvent& event);
    void on_view_menu(wxCommandEvent& event);
    void on_model_menu(wxCommandEvent& event);
    void on_help_menu(wxCommandEvent& event);
    void on_close(wxCloseEvent& event);

 protected:
    MainCanvas *_sceneCanvas;
    wxTextCtrl *_sceneText;
    wxSizer *_subWindows;

    DECLARE_EVENT_TABLE()
};

// SceneExplorer Application --------------------------------------------------

class SceneExplorerApp : public wxApp
{
 public:
    bool OnInit();
    int OnExit();
};

// Global Variables -----------------------------------------------------------

extern MainWindow *gMainWindow;


