/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneExtractDetectionCandidates.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#define WINDOW_NAME "sceneExtractDetectionCandidates"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneExtractDetectionCandidates [OPTIONS] <w> <h> <detections>\n";
    cerr << "OPTIONS:\n"
         << "  -i <dir>         :: image directory\n"
         << "  -n <name>        :: object name\n"
         << "  -g <xml>         :: ground-truth annotations\n"
         << "  -po <dir>        :: output dir for positive samples (default: none)\n"
         << "  -no <dir>        :: output dir for negative samples (default: none)\n"
         << "  -pt <threshold>  :: positive area of overlap threshold (default: 0.5)\n"
         << "  -nt <threshold>  :: negative area of overlap threshold (default: 0.2)\n"
         << "  -robust          :: generates additional position examples\n"
         << "  -x               :: visualize results\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // set defaults and read commandline parameters
    const char *imgDir = ".";
    const char *objectName = NULL;
    const char *groundTruthFile = NULL;
    const char *outDirPos = NULL;
    const char *outDirNeg = NULL;
    double thresholdPos = 0.5;
    double thresholdNeg = 0.2;
    bool bRobust = false;
    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-i", imgDir)
        SVL_CMDLINE_STR_OPTION("-n", objectName)
        SVL_CMDLINE_STR_OPTION("-g", groundTruthFile)
        SVL_CMDLINE_STR_OPTION("-po", outDirPos)
        SVL_CMDLINE_STR_OPTION("-no", outDirNeg)
        SVL_CMDLINE_REAL_OPTION("-pt", thresholdPos)
        SVL_CMDLINE_REAL_OPTION("-nt", thresholdNeg)
        SVL_CMDLINE_BOOL_OPTION("-robust", bRobust)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 3) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const int width = atoi(SVL_CMDLINE_ARGV[0]);
    const int height = atoi(SVL_CMDLINE_ARGV[1]);
    const char *detectionsFile = SVL_CMDLINE_ARGV[2];

    // read detections
    svlObject2dSequence detectionCandidates;
    SVL_LOG(SVL_LOG_MESSAGE, "Reading detections from " << detectionsFile << "...");
    detectionCandidates.read(detectionsFile);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << detectionCandidates.size() << " image frames");

    SVL_LOG(SVL_LOG_MESSAGE, "Running non-maximal suppression...");
    nonMaximalSuppression(detectionCandidates);
    SVL_LOG(SVL_LOG_MESSAGE, "...done");

    if (objectName != NULL) {
        SVL_LOG(SVL_LOG_MESSAGE, "Removing non-" << objectName << " objects...");
        for (svlObject2dSequence::iterator it = detectionCandidates.begin(); it != detectionCandidates.end(); it++) {
            removeNonMatchingObjects(it->second, objectName);
        }
        SVL_LOG(SVL_LOG_MESSAGE, "...done");
    }

    // read groundtruth
    svlObject2dSequence groundTruthObjects;
    if (groundTruthFile != NULL) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading ground truth objects from " << groundTruthFile << "...");
        groundTruthObjects.read(groundTruthFile);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << groundTruthObjects.size() << " image frames");
    }

    // counters
    int posIndx = 0;
    int negIndx = 0;
    if (outDirPos != NULL) {
        vector<string> dummy = svlDirectoryListing(outDirPos, ".jpg");
        posIndx = (int)dummy.size() + 1;
        SVL_LOG(SVL_LOG_VERBOSE, "Starting from positive index " << posIndx);
    }
    if (outDirNeg != NULL) {
        vector<string> dummy = svlDirectoryListing(outDirNeg, ".jpg");
        negIndx = (int)dummy.size() + 1;
        SVL_LOG(SVL_LOG_VERBOSE, "Starting from negative index " << negIndx);
    }

    // iterate through frames
    SVL_LOG(SVL_LOG_MESSAGE, "Extracting patches...");
    for (svlObject2dSequence::const_iterator it = detectionCandidates.begin();
         it != detectionCandidates.end(); it++) {

        SVL_LOG(SVL_LOG_VERBOSE, "Processing " << it->second.size()
            << " detections for image " << it->first << "...");

        string filename = string(imgDir) + string("/") + it->first + string(".jpg");
        IplImage *image = cvLoadImage(filename.c_str(), CV_LOAD_IMAGE_GRAYSCALE);

        if (bVisualize) {
            cvNamedWindow(WINDOW_NAME, 1);
            cvShowImage(WINDOW_NAME, image);
            cvWaitKey(100);
        }

        //svlObject2dSequence::const_iterator gt = groundTruthObjects.find(it->first);
        svlObject2dSequence::iterator gt = groundTruthObjects.find(it->first);

        // positive examples
        if ((outDirPos != NULL) && (gt != groundTruthObjects.end())) {
            svlObject2dFrame frame = it->second;
            removeNonGroundTruthObjects(frame, gt->second, thresholdPos);

            SVL_LOG(SVL_LOG_VERBOSE, "..." << frame.size() << " positives");
            for (svlObject2dFrame::const_iterator jt = frame.begin(); jt != frame.end(); jt++) {
                CvRect r = jt->getRect();
                if (r.x < 0) continue;
                if (r.y < 0) continue;
                if (r.x + r.width >= image->width) continue;
                if (r.y + r.height >= image->height) continue;

                IplImage *clip = cvCreateImage(cvSize(r.width, r.height), IPL_DEPTH_8U, 1);
                cvSetImageROI(image, r);
                cvCopyImage(image, clip);
                cvResetImageROI(image);
                resizeInPlace(&clip, height, width);

                string outFilename = string(outDirPos) + string("/image") + 
                    padString(toString(posIndx), 5) + string(".jpg");
                cvSaveImage(outFilename.c_str(), clip);
                posIndx += 1;

                if (bRobust) {
                    // flipped version
                    cvFlip(clip, NULL, 1);
                    outFilename = string(outDirPos) + string("/image") + 
                        padString(toString(posIndx), 5) + string(".jpg");
                    cvSaveImage(outFilename.c_str(), clip);
                    posIndx += 1;

                    // expanded border
                    cvReleaseImage(&clip);
                    r.x = std::max(0, r.x - 2 * r.width / width);
                    r.y = std::max(0, r.y - 2 * r.height / height);
                    r.width += 4 * r.width / width;
                    r.height += 4 * r.height / height;
                    if (r.x + r.width > image->width) r.width = image->width - r.x;
                    if (r.y + r.height > image->height) r.height = image->height - r.y;

                    IplImage *clip = cvCreateImage(cvSize(r.width, r.height), IPL_DEPTH_8U, 1);
                    cvSetImageROI(image, r);
                    cvCopyImage(image, clip);
                    cvResetImageROI(image);
                    resizeInPlace(&clip, height, width);

                    outFilename = string(outDirPos) + string("/image") + 
                        padString(toString(posIndx), 5) + string(".jpg");
                    cvSaveImage(outFilename.c_str(), clip);
                    posIndx += 1;                    
                    
                    cvFlip(clip, NULL, 1);
                    outFilename = string(outDirPos) + string("/image") + 
                        padString(toString(posIndx), 5) + string(".jpg");
                    cvSaveImage(outFilename.c_str(), clip);
                    posIndx += 1;                    
                }


                cvReleaseImage(&clip);
            }
        }

        // negative examples
        if (outDirNeg != NULL) {
            svlObject2dFrame frame = it->second;
            if (gt != groundTruthObjects.end()) {
                removeGroundTruthObjects(frame, gt->second, thresholdNeg);
            }

            SVL_LOG(SVL_LOG_VERBOSE, "..." << frame.size() << " negatives");
            for (svlObject2dFrame::const_iterator jt = frame.begin(); jt != frame.end(); jt++) {
                CvRect r = jt->getRect();
                IplImage *clip = cvCreateImage(cvSize(r.width, r.height), IPL_DEPTH_8U, 1);
                cvSetImageROI(image, r);
                cvCopyImage(image, clip);
                cvResetImageROI(image);
                resizeInPlace(&clip, height, width);

                string outFilename = string(outDirNeg) + string("/image") + 
                    padString(toString(negIndx), 5) + string(".jpg");
                cvSaveImage(outFilename.c_str(), clip);
                negIndx += 1;

                cvReleaseImage(&clip);
            }
        }

        // free memory
        cvReleaseImage(&image);
    }

    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

