/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneHOGFeatures.h
** AUTHOR(S):   Tianshi Gao <tianshig@stanford.edu>
**
*****************************************************************************/

#include <vector>
#include <string>
#include <cmath>
#include <algorithm>
#include <iterator>
#include <fstream>

#include "Eigen/Core"

#include "svlBase.h"
#include "svlVision.h"

#include "sceneHOGFeatures.h"

using namespace std;

#define PI 3.14159265
const int CELL_SIZE = 8;
const int NUM_CELL_ROWS = 8;
const int NUM_CELL_COLS = 16;
const int BLOCK_SIZE = 2;
const double BLOCK_OVERLAP = 0.5;
const int ORI_BIN_SIZE = 9;
const int WIN_WIDTH = 64;
const int WIN_HEIGHT = 128;
const double EPSILON = 1e-20;

SceneHOGFeatures::SceneHOGFeatures() : 
    _image(NULL), _votingMap(NULL), _cellSize(CELL_SIZE), _numCellRows(NUM_CELL_ROWS), _numCellCols(NUM_CELL_COLS),
    _blockSize(BLOCK_SIZE), _blockOverlap(BLOCK_OVERLAP),
    _oriBinSize(ORI_BIN_SIZE), _winWidth(WIN_WIDTH), _winHeight(WIN_HEIGHT)
{
    // do nothing
}

SceneHOGFeatures::SceneHOGFeatures(const IplImage * img, const char* extractorFilename) :
    _votingMap(NULL), _cellSize(CELL_SIZE), _numCellRows(NUM_CELL_ROWS), _numCellCols(NUM_CELL_COLS),
    _blockSize(BLOCK_SIZE), _blockOverlap(BLOCK_OVERLAP),
    _oriBinSize(ORI_BIN_SIZE), _winWidth(WIN_WIDTH), _winHeight(WIN_HEIGHT)
{
    SVL_ASSERT(img != NULL);
    _image = cvCloneImage(img);
    if (extractorFilename != NULL)
        loadExtractor(extractorFilename);
    // prepocessing to compute the orientation integral images
    prePocessing();
}

SceneHOGFeatures::SceneHOGFeatures(const char* filename, const char* extractorFilename) :
    _votingMap(NULL), _cellSize(CELL_SIZE), _blockSize(BLOCK_SIZE), _blockOverlap(BLOCK_OVERLAP),
    _oriBinSize(ORI_BIN_SIZE), _winWidth(WIN_WIDTH), _winHeight(WIN_HEIGHT)
{
    SVL_ASSERT(filename != NULL);
    //declareOptions();
    load(filename);
    if (extractorFilename != NULL)
        loadExtractor(extractorFilename);
    // prepocessing to compute the orientation integral images
    prePocessing();
}

SceneHOGFeatures::~SceneHOGFeatures()
{
    freeImages();
}

void SceneHOGFeatures::extract(const vector<CvPoint> &locations, 
                               const vector<IplImage*> &channels, 
	                       vector<vector<double> > &output,
	                       bool sparse,
	                       int outputOffset)
{
    static int h = svlCodeProfiler::getHandle("extract");
    svlCodeProfiler::tic(h);
    
    if (channels.empty()) {
        SVL_LOG(SVL_LOG_WARNING, "channels are empty...");
        return;
    }

    // only uses the first channel, TODO: run on multiple channels
    SVL_ASSERT(channels[0] != NULL);

    // free previous images
    freeImages();
    
    // initialization
    _image = cvCloneImage(channels[0]);

    // prepocessing to compute the orientation integral images
    prePocessing();

    // cache for feature extraction
    vector<IplImage *> enlargedWins;
    for (int o = 0; o < _oriBinSize; o++) {
        IplImage* enlargedWin = cvCreateImage(cvSize(_cellSize+_winWidth, _cellSize+_winHeight), IPL_DEPTH_32F, 1);
        cvSetZero(enlargedWin);
        enlargedWins.push_back(enlargedWin);
    }
    
    // pre-allocate memeory for cell hist
    int numCellRows = _winHeight/_cellSize;
    int numCellCols = _winWidth/_cellSize;
    vector<vector<double> *> cellResponses;
    for (int i = 0; i < numCellCols*numCellRows; i++) {
        vector<double>* hist = new vector<double>(_oriBinSize, 0);
        cellResponses.push_back(hist);
    }

    for (int i = 0; i < (int)locations.size(); i++) {
        CvRect r;
        r.x = locations[i].x;
        r.y = locations[i].y;
        r.width = _winWidth;
        r.height = _winHeight;
//         if (r.x + _winWidth - 1 >= _image->width) {
//             cout << "width out of boundary: " << r.x << " + " << _winWidth << " - 1 >= " << _image->width;
//         }
//         if (r.y + _winHeight - 1 >= _image->height) {
//             cout << "height out of boundary: " << r.y << " + " << _winHeight << " - 1 >= " << _image->height;
//         }

//        VectorXd f = extractFeatures(r);
        VectorXd f = extractFeatures(r, enlargedWins, cellResponses);
        vector<double> f1(f.rows(), 0);
        Eigen::Map<VectorXd>(&f1[0], f1.size()) = f;
        output.push_back(f1);
    }

    // free memory
    svlReleaseImages(enlargedWins);

    svlCodeProfiler::toc(h);
}

// Given a rect region, extract HOG features for it
VectorXd SceneHOGFeatures::extractFeatures(const CvRect& r, vector<vector<double> *>* cellHist) const
{
    static int h = svlCodeProfiler::getHandle("extractFeatures");
    svlCodeProfiler::tic(h);
    SVL_ASSERT(r.x >=0 && r.x + r.width - 1 < _image->width && r.y >= 0 && r.y + r.height - 1 < _image->height);
    SVL_ASSERT_MSG((r.width - r.height)*(_winWidth - _winHeight) >= 0, 
        "r.width: " << r.width << " r.height: " << r.height << " winWidth: " << _winWidth << " winHeight: " << _winHeight);
    
    // TODO: make sure r has correct aspect ratio as win
    //int cellSize = (int)(_cellSize * (double)r.width/(double)_winWidth);
    int cellSize = _cellSize;
    int numCellRows = _winHeight/_cellSize;
    int numCellCols = _winWidth/_cellSize;

    int overlapCells = (int)(_blockSize * _blockOverlap);
    int blockCellStep = _blockSize - overlapCells;
    int numBlockCols = (numCellCols - _blockSize) / blockCellStep + 1;
    int numBlockRows = (numCellRows - _blockSize) / blockCellStep + 1;
    SVL_ASSERT(numCellRows >= _blockSize && numCellCols >= _blockSize);

    VectorXd descriptor;
    descriptor = VectorXd::Zero(numBlockRows*numBlockCols*_blockSize*_blockSize*_oriBinSize);
    
    vector<vector<double> *> cellResponses;
    // histogram for each cell
    
#if 0
    /* ----------- integral image version without spatial interpolation ------------ */
    for (int i = 0; i < numCellRows*numCellCols; i++) {
        vector<double>* hist = new vector<double>(_oriBinSize, 0);
        CvPoint p;
        p.x = r.x + (i%numCellCols) * cellSize;
        p.y = r.y + (i/numCellCols) * cellSize;

        SVL_ASSERT_MSG(p.x+cellSize-1 < _image->width && p.y+cellSize-1 < _image->height, 
            "p.x: " << p.x << " p.y: " << p.y << " cellSize: " << cellSize 
            << " imageWidth: " << _image->width << " imageHeight: " << _image->height);

        for (int j = 0; j < _oriBinSize; j++) {
            IplImage* oriImage = _oriIntegralImages[j];
            double mag = CV_IMAGE_ELEM(oriImage, double, p.y+cellSize, p.x+cellSize)
                - CV_IMAGE_ELEM(oriImage, double, p.y+cellSize, p.x)
                - CV_IMAGE_ELEM(oriImage, double, p.y, p.x+cellSize)
                + CV_IMAGE_ELEM(oriImage, double, p.y, p.x);
            (*hist)[j] = mag;
        }

        cellResponses.push_back(hist);

        if (cellHist != NULL) {
            vector<double>* hist1 = new vector<double>(_oriBinSize, 0);
            copy(hist->begin(), hist->end(), hist1->begin());
            cellHist->push_back(hist1);
        }
   }
#else
    /* ----- slower version without integral images but with spatial interpolation ---- */
    SVL_ASSERT_MSG(r.width == _winWidth && r.height == _winHeight,
        r.width << " != " << _winWidth << " || " << r.height << " != " << _winHeight);
    // enlarge the border of the window
    vector<IplImage *> enlargedWins;

    for (int o = 0; o < _oriBinSize; o++) {
        IplImage* enlargedWin = cvCreateImage(cvSize(_cellSize+_winWidth, _cellSize+_winHeight), IPL_DEPTH_32F, 1);
        CvRect rect = cvRect(_cellSize/2, _cellSize/2, r.width, r.height);
        cvSetZero(enlargedWin);
        cvSetImageROI(enlargedWin, rect);
        cvSetImageROI(_oriImages[o], r);
        cvCopy(_oriImages[o], enlargedWin);
        cvResetImageROI(_oriImages[o]);
        cvResetImageROI(enlargedWin);
        enlargedWins.push_back(enlargedWin);
    }
    
    for (int i = 0; i < numCellRows*numCellCols; i++) {
        vector<double>* hist = new vector<double>(_oriBinSize, 0);

        CvRect cellSupport = cvRect((i%numCellCols) * _cellSize,(i/numCellCols) * _cellSize, 2*_cellSize, 2*_cellSize);
        for (int o = 0; o < _oriBinSize; o++) {
            cvSetImageROI(enlargedWins[o], cellSupport);
            (*hist)[o] = cvDotProduct(enlargedWins[o], _votingMap);
            cvResetImageROI(enlargedWins[o]);
        }

        cellResponses.push_back(hist);

        if (cellHist != NULL) {
            vector<double>* hist1 = new vector<double>(_oriBinSize, 0);
            copy(hist->begin(), hist->end(), hist1->begin());
            cellHist->push_back(hist1);
        }
    }

    // free memeory
    svlReleaseImages(enlargedWins);

#endif
    // for each block, stack histograms from different cells and normalize
    for (int i = 0; i < numBlockRows; i++) {
        int startCellRowIndx = i*blockCellStep;
        for (int j = 0; j < numBlockCols; j++) {
            int startCellColIndx = j*blockCellStep;
            VectorXd blockDescriptor;
            blockDescriptor = VectorXd::Zero(_blockSize*_blockSize*_oriBinSize);

            // collect histogram from each cell 
            for (int ii = 0; ii < _blockSize; ii++) {
                for (int jj = 0; jj < _blockSize; jj++) {
                    int tmpIdx = (startCellRowIndx+ii)*numCellCols+startCellColIndx+jj;
                    SVL_ASSERT_MSG( tmpIdx < (int)cellResponses.size(),
                        endl <<
                        "startCellColIndx:    " << startCellColIndx << endl << 
                        "startCellRowIndx:    " << startCellRowIndx << endl << 
                        "ii:                  " << ii << endl <<
                        "jj:                  " << jj << endl <<
                        "num cell cols:       " << numCellCols << endl << 
                        "tmpIdx:              " << tmpIdx << endl <<
                        "cell response size:  " << cellResponses.size() << endl);

                    vector<double>* hist = cellResponses[tmpIdx];
                    SVL_ASSERT(hist->size() == _oriBinSize);
                    blockDescriptor.segment( (ii*_blockSize+jj)*_oriBinSize, _oriBinSize ) << Eigen::Map<VectorXd>(&(*hist)[0], hist->size());
                }
            }
            // normalization
            double normalizingConst = sqrt(blockDescriptor.squaredNorm() + EPSILON);
            blockDescriptor /= normalizingConst;
            // truncate and re-normalize
            vector<double> tmpElemMax(blockDescriptor.rows(), 0.2);
            blockDescriptor = blockDescriptor.cwise().min(Eigen::Map<VectorXd>(&tmpElemMax[0], tmpElemMax.size()));
            normalizingConst = sqrt(blockDescriptor.squaredNorm() + EPSILON);
            blockDescriptor /= normalizingConst;
            descriptor.segment((i*numBlockCols + j)*blockDescriptor.rows(), blockDescriptor.rows()) << blockDescriptor;
        }
    }

    // free histogram for each cell
    for (int i = 0; i < (int)cellResponses.size(); i++) {
        delete cellResponses[i];
    }

    svlCodeProfiler::toc(h);
    return descriptor; 
}


// Given a rect region, extract HOG features for it
VectorXd SceneHOGFeatures::extractFeatures(const CvRect& r, vector<IplImage *>& enlargedWins, vector<vector<double> *> cellResponses,
    vector<vector<double> *>* cellHist) const
{
    static int h = svlCodeProfiler::getHandle("extractFeatures");
    svlCodeProfiler::tic(h);
    SVL_ASSERT(r.x >=0 && r.x + r.width - 1 < _image->width && r.y >= 0 && r.y + r.height - 1 < _image->height);
    SVL_ASSERT_MSG((r.width - r.height)*(_winWidth - _winHeight) >= 0, 
        "r.width: " << r.width << " r.height: " << r.height << " winWidth: " << _winWidth << " winHeight: " << _winHeight);
    SVL_ASSERT(enlargedWins.empty() == false && enlargedWins.size() == _oriBinSize);
    for (int i = 0; i < (int)enlargedWins.size(); i++) {
        SVL_ASSERT(enlargedWins[i] != NULL);
    }


    
    // TODO: make sure r has correct aspect ratio as win
    //int cellSize = (int)(_cellSize * (double)r.width/(double)_winWidth);
    int cellSize = _cellSize;
    int numCellRows = _winHeight/_cellSize;
    int numCellCols = _winWidth/_cellSize;

    int overlapCells = (int)(_blockSize * _blockOverlap);
    int blockCellStep = _blockSize - overlapCells;
    int numBlockCols = (numCellCols - _blockSize) / blockCellStep + 1;
    int numBlockRows = (numCellRows - _blockSize) / blockCellStep + 1;
    SVL_ASSERT(numCellRows >= _blockSize && numCellCols >= _blockSize);
    
    SVL_ASSERT(cellResponses.size() == numCellCols*numCellRows);

    VectorXd descriptor;
    descriptor = VectorXd::Zero(numBlockRows*numBlockCols*_blockSize*_blockSize*_oriBinSize);
    
    // histogram for each cell
    /* ----- slower version without integral images but with spatial interpolation ---- */
    SVL_ASSERT(r.width ==_winWidth && r.height == _winHeight);
    CvRect rect = cvRect(_cellSize/2, _cellSize/2, _winWidth, _winHeight);
    for (int o = 0; o < _oriBinSize; o++) {
        cvSetZero(enlargedWins[o]);
        cvSetImageROI(enlargedWins[o], rect);

        cvSetImageROI(_oriImages[o], r);
        cvCopy(_oriImages[o], enlargedWins[o]);

        cvResetImageROI(_oriImages[o]);
        cvResetImageROI(enlargedWins[o]);
    }

    for (int i = 0; i < numCellRows*numCellCols; i++) {
        //vector<double>* hist = new vector<double>(_oriBinSize, 0);
        vector<double>* hist = cellResponses[i];
        SVL_ASSERT(hist != NULL);
        CvRect cellSupport = cvRect((i%numCellCols) * _cellSize, (i/numCellCols) * _cellSize, 2*_cellSize, 2*_cellSize);

        for (int o = 0; o < _oriBinSize; o++) {
            cvSetImageROI(enlargedWins[o], cellSupport);
            (*hist)[o] = cvDotProduct(enlargedWins[o], _votingMap);
            cvResetImageROI(enlargedWins[o]);
        }

        //cellResponses.push_back(hist);

        if (cellHist != NULL) {
            vector<double>* hist1 = new vector<double>(_oriBinSize, 0);
            copy(hist->begin(), hist->end(), hist1->begin());
            cellHist->push_back(hist1);
        }
    }

    // for each block, stack histograms from different cells and normalize
    for (int i = 0; i < numBlockRows; i++) {
        int startCellRowIndx = i*blockCellStep;
        for (int j = 0; j < numBlockCols; j++) {
            int startCellColIndx = j*blockCellStep;
            VectorXd blockDescriptor;
            blockDescriptor = VectorXd::Zero(_blockSize*_blockSize*_oriBinSize);

            // collect histogram from each cell 
            for (int ii = 0; ii < _blockSize; ii++) {
                for (int jj = 0; jj < _blockSize; jj++) {
                    int tmpIdx = (startCellRowIndx+ii)*numCellCols+startCellColIndx+jj;
                    SVL_ASSERT_MSG( tmpIdx < (int)cellResponses.size(),
                        endl <<
                        "startCellColIndx:    " << startCellColIndx << endl << 
                        "startCellRowIndx:    " << startCellRowIndx << endl << 
                        "ii:                  " << ii << endl <<
                        "jj:                  " << jj << endl <<
                        "num cell cols:       " << numCellCols << endl << 
                        "tmpIdx:              " << tmpIdx << endl <<
                        "cell response size:  " << cellResponses.size() << endl);

                    vector<double>* hist = cellResponses[tmpIdx];
                    SVL_ASSERT(hist->size() == _oriBinSize);
                    blockDescriptor.segment( (ii*_blockSize+jj)*_oriBinSize, _oriBinSize ) << Eigen::Map<VectorXd>(&(*hist)[0], hist->size());
                }
            }
            // normalization
            double normalizingConst = sqrt(blockDescriptor.squaredNorm() + EPSILON);
            blockDescriptor /= normalizingConst;

            // truncate and re-normalize
            vector<double> tmpElemMax(blockDescriptor.rows(), 0.2);
            blockDescriptor = blockDescriptor.cwise().min(Eigen::Map<VectorXd>(&tmpElemMax[0], tmpElemMax.size()));
            normalizingConst = sqrt(blockDescriptor.squaredNorm() + EPSILON);
            blockDescriptor /= normalizingConst;

            descriptor.segment((i*numBlockCols + j)*blockDescriptor.rows(), blockDescriptor.rows()) << blockDescriptor;
        }
    }

    // free histogram for each cell
//     for (int i = 0; i < (int)cellResponses.size(); i++) {
//         delete cellResponses[i];
//     }

    svlCodeProfiler::toc(h);
    return descriptor; 
}

// given the mask of the region, extract the feature
VectorXd SceneHOGFeatures::extractFeatures(const CvMat* regionMask, Vector4i* regionExtent) const
{
    static int h = svlCodeProfiler::getHandle("extractFeatures");
    svlCodeProfiler::tic(h);
    Vector4i tmpRegionExtent;
    if (regionExtent == NULL) {
        getRegionExtent(regionMask, tmpRegionExtent);
        regionExtent = &tmpRegionExtent;
    }
    SVL_ASSERT_MSG((*regionExtent)[0] >=0 && (*regionExtent)[0]+(*regionExtent)[2]-1 < _image->width
        && (*regionExtent)[1] >= 0 && (*regionExtent)[1] + (*regionExtent)[3] - 1 < _image->height, 
        "region extent: " << toString(*regionExtent) << endl << "image size: " << _image->width << "x" << _image->height);

    // cell related size is in pixel
    double cellSizeX = (double)(*regionExtent)[2] / (double)_numCellCols;
    double cellSizeY = (double)(*regionExtent)[3] / (double)_numCellRows;

    // block related size is in cell
    int overlapCells = (int)(_blockSize * _blockOverlap);
    int blockCellStep = _blockSize - overlapCells;
    int numBlockCols = (_numCellCols - _blockSize) / blockCellStep + 1;
    int numBlockRows = (_numCellRows - _blockSize) / blockCellStep + 1;
    SVL_ASSERT(_numCellRows >= _blockSize && _numCellCols >= _blockSize);

    VectorXd descriptor;
    descriptor = VectorXd::Zero(numBlockRows*numBlockCols*_blockSize*_blockSize*_oriBinSize);
    
#if 0
    cvNamedWindow("debug", 1);
    IplImage* tmpImage = cvCloneImage(_image);
    //cout << "real cell size: " << cellSize << endl;
#endif

    vector<vector<double> *> cellResponses;
    // histogram for each cell
    SVL_ASSERT_MSG((int)_oriIntegralImages.size() == _oriBinSize, 
        _oriIntegralImages.size() << " != " << _oriBinSize);
    for (int i = 0; i < _numCellRows*_numCellCols; i++) {
        vector<double>* hist = new vector<double>(_oriBinSize, 0);
        CvPoint p;
        p.x = (int)round((*regionExtent)[0] + (i%_numCellCols) * cellSizeX);
        p.y = (int)round((*regionExtent)[1] + (i/_numCellCols) * cellSizeY);
        
        int cellWidth = std::min((int)round((*regionExtent)[0] + ((i%_numCellCols)+1) * cellSizeX), _image->width-1) - p.x + 1;
        int cellHeight = std::min((int)round((*regionExtent)[1] + ((i/_numCellCols)+1) * cellSizeY), _image->height-1) - p.y + 1;
            
        SVL_ASSERT_MSG(p.x+cellWidth-1 < _image->width && p.y+cellHeight-1 < _image->height, 
            "p.x: " << p.x << " p.y: " << p.y << " cell width: " << cellWidth << " cell height: " << cellHeight
            << " imageWidth: " << _image->width << " imageHeight: " << _image->height);

        bool bSkip = false;
        int numNonZeroPixels = 0;
        for (int j = p.y; j < p.y + cellHeight; j++) {
            for (int k = p.x; k < p.x + cellWidth; k++) {
                if (CV_MAT_ELEM(*regionMask, int, j, k) == 1)
                    numNonZeroPixels++;
            }
        }
        if ((double)numNonZeroPixels / (double)(cellWidth*cellHeight) < 0.05) {
            bSkip = true;
        }

        if (bSkip == false) {
            for (int j = 0; j < _oriBinSize; j++) {
                IplImage* oriImage = _oriIntegralImages[j];
                double mag = CV_IMAGE_ELEM(oriImage, double, p.y+cellHeight, p.x+cellWidth)
                    - CV_IMAGE_ELEM(oriImage, double, p.y+cellHeight, p.x)
                    - CV_IMAGE_ELEM(oriImage, double, p.y, p.x+cellWidth)
                    + CV_IMAGE_ELEM(oriImage, double, p.y, p.x);
                (*hist)[j] = mag;
            }
        }
#if 0
        for (int j = 0; j < _oriBinSize; j++) {
            cout << (*hist)[j] << " ";
        }
        cout << endl;
        cvRectangle(tmpImage, cvPoint(p.x, p.y), cvPoint(p.x+cellWidth-1, p.y+cellHeight-1), cvScalar(255, 255, 255), 1);
        cvShowImage("debug", tmpImage);
        cvNamedWindow("mask", 1);
        IplImage* tmpRegionMask = cvCreateImage(cvSize(tmpImage->width, tmpImage->height), tmpImage->depth, tmpImage->nChannels);
        cvConvertScale(regionMask, tmpRegionMask, 250, 0);
        cvShowImage("mask", tmpRegionMask);
        cvWaitKey(-1);
        cvReleaseImage(&tmpRegionMask);
#endif
        cellResponses.push_back(hist);
    }
    
    // for each block, stack histograms from different cells and normalize
    for (int i = 0; i < numBlockRows; i++) {
        for (int j = 0; j < numBlockCols; j++) {
            vector<double> tmpBlockDescriptor;
            int startCellColIndx = j*blockCellStep;
            int startCellRowIndx = i*blockCellStep;
            
            // collect histogram from each cell 
            for (int ii = 0; ii < _blockSize; ii++) {
                for (int jj = 0; jj < _blockSize; jj++) {
                    int tmpIdx = (startCellRowIndx+ii)*_numCellCols+startCellColIndx+jj;
                    SVL_ASSERT_MSG( tmpIdx < (int)cellResponses.size(),
                        endl <<
                        "startCellColIndx:    " << startCellColIndx << endl << 
                        "startCellRowIndx:    " << startCellRowIndx << endl << 
                        "ii:                  " << ii << endl <<
                        "jj:                  " << jj << endl <<
                        "num cell cols:       " << _numCellCols << endl << 
                        "tmpIdx:              " << tmpIdx << endl <<
                        "cell response size:  " << cellResponses.size() << endl);
                    vector<double>* hist = cellResponses[tmpIdx];
                    copy(hist->begin(), hist->end(), back_inserter(tmpBlockDescriptor));
                }
            }

            // normalization
            VectorXd blockDescriptor = Eigen::Map<VectorXd>(&tmpBlockDescriptor[0], tmpBlockDescriptor.size());
            double normalizingConst = sqrt(blockDescriptor.squaredNorm() + EPSILON);
            blockDescriptor /= normalizingConst;

            // truncate and re-normalize
//              vector<double> tmpElemMax(blockDescriptor.rows(), 0.2);
//              VectorXd tmpElemMaxEign = Eigen::Map<VectorXd>(&tmpElemMax[0], tmpElemMax.size());
//              blockDescriptor = blockDescriptor.cwise().max(tmpElemMaxEign);
//              normalizingConst = sqrt(blockDescriptor.squaredNorm() + EPSILON);
//              blockDescriptor /= normalizingConst;

            descriptor.segment((i*numBlockCols + j)*blockDescriptor.rows(), blockDescriptor.rows()) << blockDescriptor;
        }
    }
 
    // free histogram for each cell
    for (int i = 0; i < (int)cellResponses.size(); i++) {
        delete cellResponses[i];
    }

#if 0
    cvDestroyWindow("debug");
    cvReleaseImage(&tmpImage);
#endif 
    
    svlCodeProfiler::toc(h);
    return descriptor; 
}

// extract features for all supplied windows
void SceneHOGFeatures::extractFeatures(const svlObject2dFrame& windows, vector<vector<double> >& features) const
{
    static int h = svlCodeProfiler::getHandle("extractFeatures");
    svlCodeProfiler::tic(h);

    for (int i = 0; i < (int)windows.size(); i++) {
        CvRect r = windows[i].getRect();
        VectorXd f = extractFeatures(r);
        vector<double> fd(f.rows(), 0);
        Eigen::Map<VectorXd>(&fd[0], fd.size()) = f;
        features.push_back(fd);
    }

    svlCodeProfiler::toc(h);
}

// load image from file
void SceneHOGFeatures::load(const char* filename)
{
    SVL_ASSERT(filename != NULL);
    
    // free previous images
    freeImages();
  
    _image = cvLoadImage(filename, CV_LOAD_IMAGE_UNCHANGED);
    SVL_ASSERT(_image != NULL);
}

// load image by copying from an existing image
void SceneHOGFeatures::load(const IplImage* img)
{
    SVL_ASSERT(img != NULL);
    
    // free previous images
    freeImages();

    _image = cvCloneImage(img);
}

bool SceneHOGFeatures::loadExtractor(const char* extractorFilename)
{
    if (extractorFilename != NULL) {
        //ifstream ifs(extractorFilename);
        //if (ifs.fail()) return false;
        //bool bSuccess = !ifs.fail();
        //SVL_ASSERT_MSG(bSuccess == true, "load extractor failed: " << extractorFilename);
        //         ifs >> _numCellRows;
        //         ifs >> _numCellCols;
        //         ifs >> _blockSize;
        //         ifs >> _blockOverlap;
        //         ifs >> _oriBinSize;
        //        ifs.close();
        XMLNode root = XMLNode::parseFile(extractorFilename, "SceneHOG");
        SVL_ASSERT(!root.isEmpty());
        char* tag = "option";
        for (int i = 0; i < root.nChildNode(tag); i++) {
            XMLNode node = root.getChildNode(tag, i);
            const char *name = node.getAttribute("name");
            const char *value = node.getAttribute("value");
            assert((name != NULL) && (value != NULL));
            
            if (strcmp(name, "cellSize") == 0) {
                _cellSize = atoi(value);
            } else if (strcmp(name, "blockSize") == 0) {
                _blockSize = atoi(value);
            } else if (strcmp(name, "blockOverlap") == 0) {
                _blockOverlap = atof(value);
            } else if (strcmp(name, "oriBinSize") == 0) {
                _oriBinSize = atoi(value);
            } else if (strcmp(name, "winWidth") == 0) {
                _winWidth = atoi(value);
            } else if (strcmp(name, "winHeight") == 0) {
                _winHeight = atoi(value);
            } else if (strcmp(name, "numCellRows") == 0) {
                _numCellRows = atoi(value);
            } else if (strcmp(name, "numCellCols") == 0) {
                _numCellCols = atoi(value);
            }
        }
        return true;
    } else {
        return false;
    }
}

// compute the gradient and create integral images for each orientation
void SceneHOGFeatures::prePocessing()
{
    SVL_ASSERT(_image != NULL);
    computeGradient();
    SVL_ASSERT(_gradientImages[0][0] != NULL && _gradientImages[0][1] != NULL);
    computeOriImages();
    //computeOriIntegralImages();
    computeVotingMap(); // for spatial interpolation
}

void SceneHOGFeatures::setOriBinSize(int oriBinSize)
{
    // if the orientation bin size is changed,
    // re-compute the integral image for each orientation
    if (_oriBinSize != oriBinSize) {
        _oriBinSize = oriBinSize;
        if (_gradientImages.empty()) {
            prePocessing();
        } else {
            computeOriImages();
            //computeOriIntegralImages();
        }
    }
}

void SceneHOGFeatures::summary()
{
    int numCellRows = _winHeight/_cellSize;
    int numCellCols = _winWidth/_cellSize;
    
    int overlapCells = (int)(_blockSize * _blockOverlap);
    int blockCellStep = _blockSize - overlapCells;
    int numBlockCols = (numCellCols - _blockSize) / blockCellStep + 1;
    int numBlockRows = (numCellRows - _blockSize) / blockCellStep + 1;
    SVL_ASSERT(numCellRows >= _blockSize && numCellCols >= _blockSize);
    
    cout << "Scene HOG feature summary:" << endl
         << "---------------------------" << endl
         << "cell size:            " << _cellSize << endl
         << "number of cell rows:  " << _numCellRows << endl
         << "number of cell cols:  " << _numCellCols << endl
         << "block size:           " << _blockSize << endl
         << "block overlap:        " << _blockOverlap << endl
         << "orienation bins:      " << _oriBinSize << endl
         << "window width:         " << _winWidth << endl
         << "window height:        " << _winHeight << endl
         << "num cell cols:        " << numCellCols << endl
         << "num cell rows:        " << numCellRows << endl
         << "num block cols:       " << numBlockCols << endl
         << "num block rows:       " << numBlockRows << endl
         << "block cell step:      " << blockCellStep << endl
         << "feature length:       " << numBlockRows*numBlockCols*_blockSize*_blockSize*_oriBinSize << endl;
}

int SceneHOGFeatures::getNumFeatures() const
{
    int numCellRows = _winHeight/_cellSize;
    int numCellCols = _winWidth/_cellSize;
    
    int overlapCells = (int)(_blockSize * _blockOverlap);
    int blockCellStep = _blockSize - overlapCells;
    int numBlockCols = (numCellCols - _blockSize) / blockCellStep + 1;
    int numBlockRows = (numCellRows - _blockSize) / blockCellStep + 1;

    return (numBlockRows * numBlockCols * _blockSize * _blockSize * _oriBinSize);    
}

// for debugging
void SceneHOGFeatures::visualize(const CvRect& r)
{
    if( _oriImages.empty() ) {
        SVL_ASSERT(_image != NULL);
        prePocessing();
    }
    
    int numCellRows = _winHeight/_cellSize;
    int numCellCols = _winWidth/_cellSize;

    vector<vector<double> *> cellHist;
    extractFeatures(r, &cellHist);

    IplImage* cellHistMap = cvCreateImage(cvSize(r.width, r.height), IPL_DEPTH_32F, 1);
    IplImage* windowImg = cvCreateImage(cvSize(r.width, r.height), _image->depth, _image->nChannels);
    cvSetImageROI(_image, r);
    cvCopyImage(_image, windowImg);
    cvResetImageROI(_image);

    // draw cells
    for (int i = 0; i < numCellRows; i++) {
        cvLine(windowImg, cvPoint(0, _cellSize*i), cvPoint(windowImg->width - 1,_cellSize*i), cvScalar(255,255,255), 1);            
    }

    for (int i = 0; i < numCellCols; i++) {
        cvLine(windowImg, cvPoint(_cellSize*i, 0), cvPoint(_cellSize*i, windowImg->height - 1), cvScalar(255,255,255), 1);
    }
    
    // normalization by block
    vector<double> normalConsts;
    int numBlockCols = numCellCols / _blockSize;
    int numBlockRows = numCellRows / _blockSize;

    for (int i = 0; i < numBlockRows; i++) {
        for (int j = 0; j < numBlockCols; j++) {
            vector<double> tmpBlockDescriptor;
            int startCellColIndx = j*_blockSize;
            int startCellRowIndx = i*_blockSize;

            for (int ii = 0; ii < _blockSize; ii++) {
                for (int jj = 0; jj < _blockSize; jj++) {
                    int tmpIndx = (startCellRowIndx+ii)*numCellCols+startCellColIndx+jj;
                    copy(cellHist[tmpIndx]->begin(), cellHist[tmpIndx]->end(), back_inserter(tmpBlockDescriptor));
                }
            }
            VectorXd blockDescriptor = Eigen::Map<VectorXd>(&tmpBlockDescriptor[0], tmpBlockDescriptor.size());
            double normalizingConst = sqrt(blockDescriptor.squaredNorm() + EPSILON);
            cout << "normalization const 1: " << normalizingConst << endl;
            blockDescriptor /= normalizingConst;
            normalConsts.push_back(normalizingConst);
            // truncate and re-normalize
//             vector<double> tmpElemMax(blockDescriptor.rows(), 0.2);
//             VectorXd tmpElemMaxEign = Eigen::Map<VectorXd>(&tmpElemMax[0], tmpElemMax.size());
//             blockDescriptor = blockDescriptor.cwise().max(tmpElemMaxEign);
//             normalizingConst = sqrt(blockDescriptor.squaredNorm() + EPSILON);
//             cout << "normalization const 2: " << normalizingConst << endl;
        }
    } 
 

    for (int i = 0; i < (int)cellHist.size(); i++) {
        vector<double>* hist = cellHist[i];
        VectorXd hist1 = Eigen::Map<VectorXd>(&(*hist)[0], hist->size());
        int blockCol = (i%numCellCols) / _blockSize;
        int blockRow = (i/numCellCols) / _blockSize;
        int blockID = blockRow * numBlockCols + blockCol;
        cout << "cell ID: " << i << " cell col: " << i%numCellCols << " cell row: " << i/numCellCols << " blockRow: " << blockRow << " blockCol: " << blockCol << " blockID: " << blockID << endl;
        hist1 /= normalConsts[blockID];
        int cx = (i%numCellCols) * _cellSize + _cellSize/2;
        int cy = i/numCellCols * _cellSize + _cellSize/2;
        int maxIndx = 0;
        double maxValue = 0;
        for (int j = 0; j < (int)hist1.rows(); j++) {
            if (hist1[j] > maxValue) {
                maxValue = hist1[j];
                maxIndx = j;
            }
            double degree = PI / _oriBinSize * j;
            int dx = (int)(hist1[j]*_cellSize*cos(degree));
            int dy = (int)(hist1[j]*_cellSize*sin(degree));
            int intensity = (int)(hist1[j]*255.0);
            cvLine(cellHistMap, cvPoint(cx, cy), cvPoint(cx+dx,cy+dy), cvScalar(intensity,intensity,intensity), 1);
            cvLine(windowImg, cvPoint(cx, cy), cvPoint(cx+dx,cy+dy), cvScalar(0,0,intensity), 1);
        }
        //double maxDegree = PI / _cellSize * maxIndx;
        //cout << "maxDegree: " << maxDegree << " dx: " << dx << " dy: " << dy << " maxValue: " << maxValue << endl;
    }

    cvNamedWindow("imageWithCells", 1);
    cvShowImage("imageWithCells", windowImg);
    cvNamedWindow("cellHists", 1);
    cvShowImage("cellHists", cellHistMap);
    cvWaitKey(-1);

    for (int i = 0; i < (int)cellHist.size(); i++) {
        delete cellHist[i];
    }

    cvDestroyWindow("cellHists");
    cvDestroyWindow("imageWithCells");
    cvReleaseImage(&cellHistMap);
    cvReleaseImage(&windowImg);
}


// void SceneHOGFeatures::declareOptions()
// {
//     declareOption("cellSize", CELL_SIZE);
//     declareOption("blockSize", BLOCK_SIZE);
//     declareOption("blockOverlap", BLOCK_OVERLAP);
//     declareOption("orientationBins", ORI_BIN_SIZE);
//     declareOption("windowWidth", WIN_WIDTH);
//     declareOption("windowHeight",WIN_HEIGHT);
// }


// compute gradient and store the dx and dy components in _gradientImages respectively
void SceneHOGFeatures::computeGradient()
{
    static int h = svlCodeProfiler::getHandle("computeGradient");
    svlCodeProfiler::tic(h);

    // free previous gradient images
    if (!_gradientImages.empty()) {
        for (int i = 0; i < (int)_gradientImages.size(); i++) {
            if (!_gradientImages[i].empty()) {
                svlReleaseImages(_gradientImages[i]);
            }
        }
        _gradientImages.clear();
    }


    IplImage* img = cvCreateImage(cvSize(_image->width, _image->height), IPL_DEPTH_32F, 1);    
    if (_image->nChannels > 1) {
        // convert to grayscale
        IplImage* imgLayers[3];
        imgLayers[0] = cvCreateImage(cvSize(_image->width, _image->height), _image->depth, 1);
        imgLayers[1] = cvCreateImage(cvSize(_image->width, _image->height), _image->depth, 1);
        imgLayers[2] = cvCreateImage(cvSize(_image->width, _image->height), _image->depth, 1);
        cvSplit(_image, imgLayers[0], imgLayers[1], imgLayers[2], NULL);

        for (int i = 0; i < 3; i++) {
            // convert to float and compute gradient
            cvConvertScale(imgLayers[i], img);

            // compute dx and dy
            IplImage* imgDx = cvCreateImage(cvSize(_image->width, _image->height), IPL_DEPTH_32F, 1);
            IplImage* imgDy = cvCreateImage(cvSize(_image->width, _image->height), IPL_DEPTH_32F, 1);
            cvSobel(img, imgDx, 1, 0, 1);
            cvSobel(img, imgDy, 0, 1, 1);
        
            vector<IplImage *> tmpGradient;
            tmpGradient.push_back(imgDx);
            tmpGradient.push_back(imgDy);
            _gradientImages.push_back(tmpGradient);
            
            // free memory
            cvReleaseImage(&imgLayers[i]);
        }
    } else {
        // convert to float
        cvConvertScale( _image, img);
        
        // compute dx and dy
        IplImage* imgDx = cvCreateImage(cvSize(_image->width, _image->height), IPL_DEPTH_32F, 1);
        IplImage* imgDy = cvCreateImage(cvSize(_image->width, _image->height), IPL_DEPTH_32F, 1);
        cvSobel(img, imgDx, 1, 0, 1);
        cvSobel(img, imgDy, 0, 1, 1);
        
        vector<IplImage *> tmpGradient;
        tmpGradient.push_back(imgDx);
        tmpGradient.push_back(imgDy);
        _gradientImages.push_back(tmpGradient);
    }
    
    // free image
    cvReleaseImage(&img);
    
    svlCodeProfiler::toc(h);
}

void SceneHOGFeatures::computeOriImages()
{
    static int h = svlCodeProfiler::getHandle("computeOriImages");
    svlCodeProfiler::tic(h);

    // free previous images
    if (!_oriImages.empty()) {
        svlReleaseImages(_oriImages);
        _oriImages.clear();
    }
    
    // allocate orientation channels
    for (int i = 0; i < _oriBinSize; i++) {
        IplImage* oriImage = cvCreateImage(cvSize(_image->width, _image->height), IPL_DEPTH_32F, 1);
        cvZero(oriImage);
        _oriImages.push_back(oriImage);
    }    
    
    // make sure the gradient has been computed
    SVL_ASSERT(_gradientImages.empty() == false);
    SVL_ASSERT(_gradientImages[0][0] != NULL && _gradientImages[0][1] != NULL);

    // loop through each pixel to populate orientation channels
    for (int i = 0; i < _image->width; i++) {
        for (int j = 0; j <_image->height; j++) {
            float dx = 0, dy = 0, mag = 0;
            // if it's a color image, then select the strongest mag among different color channels
            if (_image->nChannels > 1) {
                float dxMax = 0, dyMax = 0, magMax = 0;
                for (int k = 0; k < (int)_gradientImages.size(); k++) {
                    dx = CV_IMAGE_ELEM(_gradientImages[k][0], float, j, i);            
                    dy = CV_IMAGE_ELEM(_gradientImages[k][1], float, j, i);
                    mag = sqrt(dx*dx + dy*dy);
                    if (mag > magMax) {
                        magMax = mag;
                        dxMax = dx;
                        dyMax = dy;
                    }
                }
                mag = magMax;
                dx = dxMax;
                dy = dyMax;
            } else {
                dx = CV_IMAGE_ELEM(_gradientImages[0][0], float, j, i);            
                dy = CV_IMAGE_ELEM(_gradientImages[0][1], float, j, i);
                mag = sqrt(dx*dx + dy*dy);
            }

            int oriDegree = ((int)(atan(dy/(dx + 1e-20)) / PI * 180) + 180) % 180;
            int oriIndex = 0;
            float oriBinWidth = 180.0 / _oriBinSize;

            // without soft binning
//             oriIndex = oriDegree/ (180 / _oriBinSize);
//             oriIndex = oriIndex >= _oriBinSize ? oriIndex-1 : oriIndex;
//             CV_IMAGE_ELEM(_oriImages[oriIndex], float, j, i) = mag;
            if (oriDegree <= oriBinWidth/2) {
                // circular binning
                CV_IMAGE_ELEM(_oriImages[0], float, j, i) = mag * (float)oriDegree/oriBinWidth;
                CV_IMAGE_ELEM(_oriImages[_oriBinSize-1], float, j, i) = mag * (1.0 - (float)oriDegree/oriBinWidth);
                
            } else if (oriDegree >= 180 - 90/_oriBinSize) {
                // circular binning
                CV_IMAGE_ELEM(_oriImages[_oriBinSize - 1], float, j, i) = mag * (180.0 - oriDegree)/oriBinWidth;
                CV_IMAGE_ELEM(_oriImages[0], float, j, i) = mag * (1 - (180.0 - oriDegree)/oriBinWidth);
            } else {
                float binWidth = (float)(180 / _oriBinSize);
                int oriIndex1 = (int)((oriDegree - round(90/_oriBinSize)) / binWidth);
                int oriIndex2 = oriIndex1 + 1;
                oriIndex1 = oriIndex1 >= _oriBinSize ? oriIndex1-1 : oriIndex1;
                oriIndex2 = oriIndex2 >= _oriBinSize ? oriIndex2-1 : oriIndex2;
                SVL_ASSERT(oriDegree >= 90/_oriBinSize + oriIndex1*binWidth 
                           && oriDegree <= 90/_oriBinSize + oriIndex2*binWidth);
                CV_IMAGE_ELEM(_oriImages[oriIndex1], float, j, i) = mag * (oriDegree - 90/_oriBinSize - oriIndex1*binWidth) / binWidth;
                CV_IMAGE_ELEM(_oriImages[oriIndex2], float, j, i) = mag * (90/_oriBinSize + oriIndex2*binWidth - oriDegree) / binWidth;
            }

            //SVL_ASSERT_MSG(oriIndex >= 0 && oriIndex < _oriBinSize, 
            //    "oriIndex: " << oriIndex << " atan(dy/dx): " << atan(dy/dx)
            //    << " dx: " << dx << " dy: " << dy);
        }
    }

    svlCodeProfiler::toc(h);
}

/* for each orientation, create an integral image
 * of the gradient magnitute from that orientation
 */
void SceneHOGFeatures::computeOriIntegralImages()
{
    static int h = svlCodeProfiler::getHandle("computeOriIntegralImages");
    svlCodeProfiler::tic(h);

    SVL_ASSERT(_oriImages.size() == _oriBinSize);
    // free previous orientation integral images
    if (!_oriIntegralImages.empty()) {
        svlReleaseImages(_oriIntegralImages);
        _oriIntegralImages.clear();
    }   

    for (int i = 0; i < _oriBinSize; i++) { 
        IplImage* oriIntegralImage = cvCreateImage(cvSize(_image->width+1, _image->height+1), IPL_DEPTH_64F, 1);
        cvIntegral(_oriImages[i], oriIntegralImage);
        _oriIntegralImages.push_back(oriIntegralImage);
    }

    svlCodeProfiler::toc(h);
}

// compute the vector of pairs of cell ID and weight for each pixel
void SceneHOGFeatures::computeVotingMap()
{
    if(_votingMap == NULL) {
        _votingMap = cvCreateImage(cvSize(2*_cellSize, 2*_cellSize), IPL_DEPTH_32F, 1);
    } else if (_votingMap->width != 2 * _cellSize) {
        cvReleaseImage(&_votingMap);
        _votingMap = cvCreateImage(cvSize(2*_cellSize, 2*_cellSize), IPL_DEPTH_32F, 1);
    }
    
    float centerX = (float)_cellSize - 0.5;
    float centerY = (float)_cellSize - 0.5;
    for (int y = 0; y < _votingMap->height; y++) {
        for (int x = 0; x < _votingMap->width; x++) {
            CV_IMAGE_ELEM(_votingMap, float, y, x) = (1.0 - abs(x - centerX) / (float)_cellSize) * (1.0 - abs(y - centerY) / (float)_cellSize) ;
        }
    }
}

// free allocated images
void SceneHOGFeatures::freeImages()
{
    if (_image != NULL) {
        cvReleaseImage(&_image);
    }

    if (!_oriImages.empty()) {
        svlReleaseImages(_oriImages);
        _oriImages.clear();
    }

    if (!_oriIntegralImages.empty()) {
        svlReleaseImages(_oriIntegralImages);
        _oriIntegralImages.clear();
    }

    if (!_gradientImages.empty()) {
        for (int i = 0; i < (int)_gradientImages.size(); i++) {
            if (!_gradientImages[i].empty()) {
                svlReleaseImages(_gradientImages[i]);
            }
        }
        _gradientImages.clear();
    }    
    
    if (_votingMap != NULL) {
        cvReleaseImage(&_votingMap);
    }
}

void SceneHOGFeatures::getRegionExtent(const CvMat* regionMask, Vector4i& regionExtent) const
{
    int minX = regionMask->width;
    int maxX = 0;
    int minY = regionMask->height;
    int maxY = 0;

    for (int i = 0; i < regionMask->width; i++) {
        for (int j = 0; j < regionMask->height; j++) {
            if (CV_MAT_ELEM(*regionMask, int, j, i) != 0) {
                if (i < minX)
                    minX = i;
                if (i > maxX)
                    maxX = i;
                if (j < minY)
                    minY = j;
                if (j > maxY)
                    maxY = j;
            }
        }
    }
    regionExtent[0] = minX;
    regionExtent[1] = minY;
    regionExtent[2] = maxX - minX + 1;
    regionExtent[3] = maxY - minY + 1;
}

