/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneHOGFeatures.h
** AUTHOR(S):   Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**  This file contains common functionality for the scene HOG feature extraction.
**
*****************************************************************************/

#pragma once

#include <vector>
#include <string>

#include "Eigen/Core"

#include "svlBase.h"
#include "svlVision.h"

using namespace std;

typedef vector<pair<int, double> > votingT;

class SceneHOGFeatures {
 protected:
    IplImage *_image;
    vector<IplImage *> _oriImages;
    vector<IplImage *> _oriIntegralImages;
    vector<vector<IplImage *> > _gradientImages;
    IplImage *_votingMap; // for spatial interpolation

    int _cellSize;    // in pixel
    int _numCellRows; 
    int _numCellCols; 
    int _blockSize;   // in cell 
    double _blockOverlap; // in percentage of the block
    int _oriBinSize;  
    int _winWidth;
    int _winHeight;
 
 public:
    SceneHOGFeatures();
    SceneHOGFeatures(const IplImage * img, const char* extractorFilename = NULL);
    SceneHOGFeatures(const char* filename, const char* extractorFilename = NULL);
    ~SceneHOGFeatures();

    void load(const char* filename);
    void load(const IplImage* img);
    bool loadExtractor(const char* extractorFilename);
    void prePocessing();
    void extract(const vector<CvPoint> &locations, 
                 const vector<IplImage*> &channels, 
	         vector<vector<double> > &output,
	         bool sparse,
	         int outputOffset = 0); 
    VectorXd extractFeatures(const CvRect& r, vector<vector<double> *>* cellHist = NULL) const;
    VectorXd extractFeatures(const CvRect& r, vector<IplImage *>& enlargedWins, 
        vector<vector<double> *> cellResponses,vector<vector<double> *>* cellHist = NULL) const;
    VectorXd extractFeatures(const CvMat* regionMask, Vector4i* regionExtent = NULL) const;
    void extractFeatures(const svlObject2dFrame& windows, vector<vector<double> >& features) const;
    //void validateOption(const char *name);
    void summary();

    // set options
    void setCellSize(int cellSize) { _cellSize = cellSize; computeVotingMap();}
    void setNumCellRows(int numCellRows) { _numCellRows = numCellRows; }
    void setNumCellCols(int numCellCols) { _numCellCols = numCellCols; }
    void setBlockSize(int blockSize) { _blockSize = blockSize; }
    void setBlockOverlap(double blockOverlap) { _blockOverlap = blockOverlap; }
    void setOriBinSize(int oriBinSize);
    void setWinWidth(int winWidth) { _winWidth = winWidth; }
    void setWinHeight(int winHeight) { _winHeight = winHeight; }

    // get options
    int getCellSize() const { return _cellSize; }
    int getNumCellRows() const { return _numCellRows; }
    int getNumCellCols() const { return _numCellCols; }
    int getBlockSize() const { return _blockSize; }
    double getBlockOverlap() const { return _blockOverlap; }
    int getOriBinSize() const { return _oriBinSize; } 
    int getWinWidth() const { return _winWidth; }
    int getWinHeight() const { return  _winHeight; }

    int getNumFeatures() const;

    // for debugging
    void visualize(const CvRect& r);

 protected:
    //void declareOptions();
    void computeGradient();
    void computeOriImages();
    void computeOriIntegralImages();
    void computeVotingMap();
    void freeImages();
    void getRegionExtent(const CvMat* regionMask, Vector4i& regionExtent) const;
};
