/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneHarvestTrainingExamples.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Harvests "false-positive" training examples from an initial train/evaluate
**  run or from any over-segmentation of the image. Useful for active learning.
**
** TODO: clean up
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneVisualization.h"

#define WINDOW_NAME "sceneHarvestTrainingExamples"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneHarvestTrainingExamples [OPTIONS] (<trainList> | <baseName>)\n";
    cerr << "OPTIONS:\n"
         << "  -m <mode>         :: harvesting mode: REGIONS (default), FEATURES, CACHE\n"
         << "  -inDir <ext>      :: extension for input layers (default: <sceneConfig.outputDir>)\n"
         << "  -inRegions <ext>  :: extension for region id input (default: .predictedRegions.txt)\n"
        //<< "  -inLabels <ext>   :: extension for region labels (default: .predictedLabels.txt)\n"
         << "  -inCache <ext>    :: extension for feature cache input (default: .mistakes.bin)\n"
         << "  -outLabels <ext>  :: output for region training examples (default: .correctedLabels.txt)\n"
         << "  -outCache <file>  :: output filename for feature cache (default: none)\n"
         << "  -subSample <n>    :: subsample cache (default: none)\n"
         << "  -useRegIdAsLabel  :: use region id as the label instead of <sceneConfig.layerClassLblExt>\n"        
         << "  -x                :: visualize output\n"
         << SVL_STANDARD_OPTIONS_USAGE 
         << endl;
}

int main(int argc, char *argv[])
{
    // read commandline parameters
    captureOpenCVerrors();
    
    const char *mode = "REGIONS";
    const char *inDir = NULL;
    const char *inRegions = ".predictedRegions.txt";
    const char *inCache = ".mistakes.bin";
    //const char *inLabels = ".predictedLabels.txt";
    const char *outLabels = ".correctedLabels.txt";
    const char *outCache = NULL;
    int subSampleCache = 0;
    bool bUseRegIdAsLabel = false;
    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-m", mode)
        SVL_CMDLINE_STR_OPTION("-inDir", inDir)
        SVL_CMDLINE_STR_OPTION("-inRegions", inRegions)
        //SVL_CMDLINE_STR_OPTION("-inLabels", inLabels)        
        SVL_CMDLINE_STR_OPTION("-inCache", inCache)
        SVL_CMDLINE_STR_OPTION("-outLabels", outLabels)
        SVL_CMDLINE_STR_OPTION("-outCache", outCache)
        SVL_CMDLINE_INT_OPTION("-subSample", subSampleCache)
        SVL_CMDLINE_BOOL_OPTION("-useRegIdAsLabel", bUseRegIdAsLabel)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *trainList = SVL_CMDLINE_ARGV[0];

    // read list of images
    vector<string> baseNames;
    if (svlFileExists(trainList)) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading evaluation list from " << trainList << "...");
        baseNames = svlReadFile(trainList);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");
    } else {
        baseNames.push_back(string(trainList));
    }

    // process images
    if (!strcasecmp(mode, "REGIONS")) {
        vector<int> labelCounts(gSceneConfig.numRegionLabels + gSceneConfig.numObjectLabels, 0);
        SVL_LOG(SVL_LOG_MESSAGE, "Reading results...");
        for (int i = 0; i < (int)baseNames.size(); i++) {
            // load image
            string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
            IplImage *img = cvLoadImage(imgFilename.c_str(), CV_LOAD_IMAGE_COLOR);
            SVL_ASSERT(img != NULL);
            SVL_LOG(SVL_LOG_VERBOSE, baseNames[i] << " (" << (i + 1) << " of "
                << baseNames.size() << "): " << toString(*img));
            
            // load groundtruth regions and labels
            CvMat *regions = cvCreateMat(img->height, img->width, CV_32SC1);
            string lblFilename = gSceneConfig.filename("lblDir", baseNames[i], "layerLblExt");
            bool bSuccess = readMatrix(regions, lblFilename.c_str());
            SVL_ASSERT_MSG(bSuccess, "Could not read class groundtruth for " << baseNames[i]);
            
            int numRegions = 0;
            vector<int> regionSizes;
            int *p = (int *)CV_MAT_ELEM_PTR(*regions, 0, 0);
            for (int j = 0; j < regions->rows * regions->cols; j++, p++) {
                if (*p < 0) continue;
                SVL_ASSERT(*p >= 0);
                numRegions = std::max(numRegions, *p + 1);
                regionSizes.resize(numRegions, 0);
                regionSizes[*p] += 1;
            }
            
            vector<int> regionLabels(numRegions, -1);
            if (bUseRegIdAsLabel) {
                for (int k = 0; k < numRegions; k++) {
                    regionLabels[i] = i;
                }
            } else {
                lblFilename = gSceneConfig.filename("lblDir", baseNames[i], "layerClassLblExt");
                ifstream ifs(lblFilename.c_str());
                SVL_ASSERT_MSG(!ifs.fail(), "could not read layer classes from " << lblFilename);
                for (int k = 0; k < numRegions; k++) {
                    ifs >> regionLabels[k];
                    SVL_ASSERT_MSG(!ifs.fail(), "Layer class labels file " << lblFilename
                        << " is too short. Should have " << numRegions << " entries.");
                }
                ifs.close();
            }
            
            // load predicted regions
            CvMat *predictedRegions = cvCreateMat(img->height, img->width, CV_32SC1);
            if (inDir == NULL) {
                lblFilename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(inRegions);
            } else {
                lblFilename = string(inDir) + string("/") + baseNames[i] + string(inRegions);
            }
            bSuccess = readMatrix(predictedRegions, lblFilename.c_str());
            if (!bSuccess) {
                SVL_LOG(SVL_LOG_WARNING, "Could not read predicted regions for " << baseNames[i]);
                cvReleaseMat(&predictedRegions);
                cvReleaseMat(&regions);
                cvReleaseImage(&img);
                continue;
            }
            
            int numPredictedRegions = 0;
            vector<vector<int> > regionOverlapCounts;
            vector<int> predictedRegionSizes;
            p = (int *)CV_MAT_ELEM_PTR(*predictedRegions, 0, 0);
            int *q = (int *)CV_MAT_ELEM_PTR(*regions, 0, 0);
            for (int y = 0; y < predictedRegions->rows; y++) {
                for (int x = 0; x < predictedRegions->cols; x++, p++, q++) {
                    if (*q < 0) continue;
                    SVL_ASSERT((*p >= 0) && (*q < numRegions));
                    numPredictedRegions = std::max(numPredictedRegions, *p + 1);
                    regionOverlapCounts.resize(numPredictedRegions, vector<int>(numRegions, 0));
                    regionOverlapCounts[*p][*q] += 1;
                    predictedRegionSizes.resize(numPredictedRegions, 0);
                    predictedRegionSizes[*p] += 1;
                }
            }
            
            // label mistakes
            vector<int> trainingLabels(numPredictedRegions, -1);
            for (int j = 0; j < numPredictedRegions; j++) {
                int regId = argmax<int>(regionOverlapCounts[j]);
                if (regionLabels[regId] < 0)
                    continue;
                if (predictedRegionSizes[j] == 0.0) {
                    SVL_LOG(SVL_LOG_WARNING, "region " << j << " is empty");
                    continue;
                }
                SVL_ASSERT(regId < (int)regionLabels.size());
                double areaOverlap = (double)regionOverlapCounts[j][regId] /
                    (double)(regionSizes[regId] + predictedRegionSizes[j]);
                double areaInternal = (double)regionOverlapCounts[j][regId] /
                    (double)predictedRegionSizes[j];
                SVL_ASSERT_MSG(areaInternal <= 1.0, areaInternal);
                if (((areaInternal > 0.9) && (regionLabels[regId] < gSceneConfig.numRegionLabels - 1)) ||
                    ((areaOverlap > 0.2) && (regionLabels[regId] >= gSceneConfig.numRegionLabels - 1)) ||
                    (areaOverlap > 0.5)) {
                    trainingLabels[j] = regionLabels[regId];
                    SVL_ASSERT(trainingLabels[j] < (int)labelCounts.size());
                    labelCounts[trainingLabels[j]] += 1;
                }
            }
            
            SVL_LOG(SVL_LOG_DEBUG, "...groundtruth: " << toString(regionLabels));
            SVL_LOG(SVL_LOG_DEBUG, "...training labels: " << toString(trainingLabels));

            // write labels
            string outFilename;
            if (inDir == NULL) {
                outFilename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(outLabels);
            } else {
                outFilename = string(inDir) + string("/") + baseNames[i] + string(outLabels);
            }
            SVL_LOG(SVL_LOG_VERBOSE, "writing training labels to " << outFilename);
            ofstream ofs(outFilename.c_str());
            for (int j = 0; j < numPredictedRegions; j++) {
                ofs << trainingLabels[j] << "\n";
            }
            ofs.close();
            
            // free memory        
            cvReleaseMat(&predictedRegions);
            cvReleaseMat(&regions);
            cvReleaseImage(&img);
        }

        SVL_LOG(SVL_LOG_MESSAGE, "New training examples: " << toString(labelCounts));

    } else if (!strcasecmp(mode, "FEATURES")) {
        SceneFeatureCache cache;

        // accmulate features
        for (int i = 0; i < (int)baseNames.size(); i++) {
            string filename;
            if (inDir == NULL) {
                filename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(inCache);
            } else {
                filename = string(inDir) + string("/") + baseNames[i] + string(inCache);
            }

            SceneFeatureCache tmpCache(filename.c_str());
            if (subSampleCache > 1) {
                tmpCache.subSample(subSampleCache, true);
                SVL_LOG(SVL_LOG_VERBOSE, "subsampled to " << tmpCache.size() 
                    << " samples of length " << tmpCache.numFeatures());
            }
            
            cache.append(tmpCache);
            SVL_LOG(SVL_LOG_VERBOSE, baseNames[i] << " : accumulated " << cache.size() 
                << " samples of length " << cache.numFeatures());
        }

        if (outCache != NULL) {
            string filename = gSceneConfig.getOption("baseDir") + string(outCache);
            int n = cache.write(filename.c_str(), true);
            SVL_LOG(SVL_LOG_MESSAGE, "cache file " << filename << " has " << n
                << " samples of length " << cache.numFeatures());
        }
    } else {
        SVL_LOG(SVL_LOG_ERROR, "unknown mode " << mode);
    }

    // shutdown
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

