/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneInference.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**  Code for running inference on a SceneRegionModel or SceneObjectModel.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <limits>
#include <set>
#include <algorithm>
#include <iterator>
#include <sys/time.h>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneObjectModel.h"
#include "sceneVisualization.h"
#include "sceneSegmentDict.h"
#include "sceneObjDetDict.h"
#include "sceneRegionFeatures.h"

#define WINDOW_NAME "sceneInference"

using namespace std;

// datatypes ----------------------------------------------------------------

struct objCandT {
    Vector4i boundingBox;
    VectorXd prObject;
    VectorXd objDetScore;
    bool bAccepted;
    VectorXd deltaEnergy;
};

struct objCandCmp {
    bool operator() (objCandT a, objCandT b) {
        for (int i = 0; i < 4; i++) {
            if (a.boundingBox[i] < b.boundingBox[i])
                return true;
            else if (a.boundingBox[i] > b.boundingBox[i])
                return false;
        }

        for (int i = 0; i < a.prObject.rows(); i++) {
            if (a.prObject[i] < b.prObject[i])
                return true;
            else if (a.prObject[i] > b.prObject[i])
                return false;
        }

        for (int i = 0; i < a.objDetScore.rows(); i++) {
            if (a.objDetScore[i] < b.objDetScore[i])
                return true;
            else if (a.objDetScore[i] > b.objDetScore[i])
                return false;
        }

        for (int i = 0; i < a.deltaEnergy.rows(); i++) {
            if (a.deltaEnergy[i] < b.deltaEnergy[i])
                return true;
            else if (a.deltaEnergy[i] > b.deltaEnergy[i])
                return false;
        }

        return false;
    }
};

class InferenceState {
public:
    // search state
    bool bConverged;
    int numMoves;
    SceneAssignment bestAssignment;
    double bestEnergy;

    // visualization options
    bool bVisualize;
    const char *videoDir;
    set<CvPoint> boundaryPixels;
    CvRect boundingBox;

    // closed-loop learning (owned by application)
    SceneFeatureCache *pClosedLoopLearningDataset;
    SceneFeatureCache *pClosedLoopLearningDatasetObject;
    SceneLayersAndLabels *pTrueLabels;

    // for experiment (record the scores for each obj cand move)
    set<objCandT, objCandCmp>* pObjCandidateRecords;


public:
    InferenceState() :
        bConverged(false), numMoves(0), bestEnergy(0.0),
        bVisualize(false), videoDir(NULL),
        pClosedLoopLearningDataset(NULL),
        pClosedLoopLearningDatasetObject(NULL),
        pTrueLabels(NULL), pObjCandidateRecords(NULL) {
        boundingBox = cvRect(0, 0, 0, 0);
    }

    ~InferenceState() {
        // do nothing
    }

    void swap(InferenceState& state) {
        std::swap(bConverged, state.bConverged);
        std::swap(numMoves, state.numMoves);
        bestAssignment.swap(state.bestAssignment);
        std::swap(bestEnergy, state.bestEnergy);
        std::swap(bVisualize, state.bVisualize);
        std::swap(videoDir, state.videoDir);
        std::swap(boundaryPixels, state.boundaryPixels);
        std::swap(boundingBox, state.boundingBox);
        std::swap(pClosedLoopLearningDataset, state.pClosedLoopLearningDataset);
        std::swap(pClosedLoopLearningDatasetObject, state.pClosedLoopLearningDatasetObject);
        std::swap(pTrueLabels, state.pTrueLabels);
        // for experiment
        std::swap(pObjCandidateRecords, state.pObjCandidateRecords);
    }

    InferenceState& operator=(const InferenceState& state) {
        if (this != &state) {
            InferenceState temp(state);
            swap(temp);
        }
        return *this;
    }

    void init(double e, const SceneAssignment& assignment) {
        bConverged = false;
        numMoves = 0;
        bestEnergy = e;
        bestAssignment = assignment;
    }

    // TODO: same as init()?
    void update(double e, const SceneAssignment& assignment) {
        SVL_LOG(SVL_LOG_MESSAGE, "updating configuration: dE = "
            << bestEnergy - e << " (" << e << ") at move " << numMoves);
#if 0
        bool bEqualConfig = (cvNorm(assignment.pixels,
                bestAssignment.pixels, CV_L1) == 0);
        if (bEqualConfig) {
            SVL_LOG(SVL_LOG_WARNING, "accepting move with equal pixel configuration");
        }
#endif
        if (e != bestEnergy) {
            bConverged = false;
        }
        bestEnergy = e;
        bestAssignment = assignment;
    }
};

// TODO: deprecate
class SceneTopDownMoveDatabase {
public:
    CvMat *segmentation;
    multimap<double, vector<int>, greater<double> > moves;

public:
    SceneTopDownMoveDatabase() : segmentation(NULL) {
        // do nothing
    }
    ~SceneTopDownMoveDatabase() {
        clear();
    }

    int numMoves() const { return (int)moves.size(); }

    void clear() {
        if (segmentation != NULL) {
            cvReleaseMat(&segmentation);
            segmentation = NULL;
        }
        moves.clear();
    }

    void load(SceneInstance &instance) {
        clear();
        string segFilename = gSceneConfig.filename("segDir", instance.name(), "combinedSegExt");
        string moveFilename = gSceneConfig.filename("objDetMoveDir", instance.name(), "objDetMoveExt");

        segmentation = cvCreateMat(instance.height(), instance.width(), CV_32SC1);
        bool bSuccess = readMatrix(segmentation, segFilename.c_str());
        if (!bSuccess) {
            SVL_LOG(SVL_LOG_FATAL, "Could not read combined segmentation from " << segFilename);
        }

        ifstream ifs(moveFilename.c_str());
        SVL_ASSERT_MSG(!ifs.fail(), "Could not read obj. det. moves from " << moveFilename);

        // debugging
        map<int, int> segmentSizes;
        const int *p = (const int *)CV_MAT_ELEM_PTR(*segmentation, 0, 0);
        for (int y = 0; y < segmentation->rows; y++) {
            for (int x = 0; x < segmentation->cols; x++, p++) {
                if (segmentSizes.find(*p) == segmentSizes.end())
                    segmentSizes.insert(make_pair(*p, 0));
                segmentSizes[*p] += 1;
            }
        }

        string line;
        while(getline(ifs, line)) {
            vector<int> vSegIndx;
            stringstream ss;
            ss << line;
            int segCount;
            double prob;
            int segIndx;
            ss >> segCount;
            ss >> prob;
            int segSize = 0;
            while (1) {
                ss >> segIndx;
                if (ss.fail()) {
                    break;
                }
                vSegIndx.push_back(segIndx);
                segSize += segmentSizes[segIndx];
            }

            SVL_LOG(SVL_LOG_DEBUG, "...move " << moves.size() << " with probability " << prob << " has " << vSegIndx.size() << " segments with " << segSize << " pixels");
            if (segSize > 32) {
                moves.insert(make_pair(prob, vSegIndx));
            }
        }

        ifs.close();
    }

    CvMat *mask(int moveId) const {
        multimap<double, vector<int>, greater<double> >::const_iterator it = moves.begin();
        while (moveId-- > 0) {
            it++;
        }

        return mask(it->second);
    }

    CvMat *mask(const vector<int>& segIndx) const {
        SVL_ASSERT(segmentation != NULL);
        CvMat *objMove = cvCreateMat(segmentation->rows, segmentation->cols, CV_8UC1);
        cvSetZero(objMove);

        CvMat* segMask = cvCreateMat(segmentation->rows, segmentation->cols, CV_8UC1);
        for (vector<int>::const_iterator it = segIndx.begin(); it != segIndx.end(); it++) {
            cvCmpS(segmentation, *it, segMask, CV_CMP_EQ);
            cvSet(objMove, cvScalar(1), segMask);
        }
        cvReleaseMat(&segMask);

        return objMove;
    }
};

// move making prototypes ---------------------------------------------------

void finalizeMove(InferenceState& state, SceneRegionModel *model, SceneInstance& instance);

// region moves
void mergeMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance);
void segmentDictionaryMoves(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneSegmentDictionary& dictionary);

// object moves
void objectDeleteMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance);
void objectMergeMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance);
void objectSplitMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance);

// top-down candidate moves
void objDetCandidateMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance,
    SceneTopDownMoveDatabase &moves, bool bRecordDeltaEnergyForObjDetMoves, bool bRecordPedestrian);

// visualization prototypes -------------------------------------------------
void viewProgress(SceneInstance& instance, const InferenceState& state);
void dumpVideoFrame(SceneInstance& instance, const InferenceState& state);

// closed-loop learning prototypes ------------------------------------------

void recordLearningExample(InferenceState& state, int regId, bool bAccepted,
    SceneInstance& instance);

void recordLearningExampleForObjectModel(InferenceState& state, int objId, bool bAccepted,
    SceneInstance& instance, SceneRegionModel* model);

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneInference [OPTIONS] (<evalList> | <baseName>)\n";
    cerr << "OPTIONS:\n"
         << "  -m <model>        :: model type: REGION (default), REGIONPAIR, OBJECT,\n"
         << "                       OBJECTCONTEXT\n"
         << "  -inRegions <ext>  :: initialize regions from file (default: none)\n"
         << "  -inDir <dir>      :: directory for region initialization (default: use configXML)\n"
         << "  -outRegions <ext> :: extension for region id output (default: none)\n"
         << "  -outLabels <ext>  :: extension for region label output (default: none)\n"
         << "  -outClasses <ext> :: extension for region class output (default: none)\n"
         << "  -outObjects <ext> :: extension for object output (default: none)\n"
         << "  -outObjectMoves <ext> :: extension for object moves output (default: none) \n"
         << "  -outHorizon <ext> :: extension for horizon output (default: none)\n"
         << "  -outImages <ext>  :: extension for visualization output (default: none)\n"
         << "  -outSingletons <ext> :: extension for outputing region singletons (default: none)\n"
         << "  -skipExisting <ext> :: skip image if file with <ext> exists in output\n"
         << "  -maxMoves <n>     :: maximum number of proposal moves (default: Inf)\n"
         << "  -disableMerges    :: disable merge moves\n"
         << "  -disableSegDict   :: disable segmentation dictionary moves\n"
         << "  -enableObjMoves   :: enable object merges and splits\n"
         << "  -enableObjCandMoves :: enable top-down object candidate moves\n"
         << "  -record <file>    :: record moves to file (for closed-loop learning)\n"
         << "  -recordObject <file>    :: record object moves to file (for closed-loop learning)\n"
         << "  -recordPedestrian :: record pedestrian features for closed-loop learning of pedestrian classifier (default: false) \n"
         << "  -recordDeltaEnergy :: record delta energy for object detection candidate moves (default: false) \n"
         << "  -video <dir>      :: dump frames to <dir> for generating videos\n"
         << "  -x[0]             :: visualize output (0: end of iterations only)\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // inference state
    InferenceState state;

    // read commandline parameters
    const char *modelType = "REGION";
    const char *inRegionsExt = NULL;
    const char *inRegionDir = NULL;
    const char *outRegionsExt = NULL;
    const char *outLabelsExt = NULL;
    const char *outObjectsExt = NULL;
    const char *outClassesExt = NULL;
    const char *outHorizonExt = NULL;
    const char *outImagesExt = NULL;
    const char *outSingletonExt = NULL;
    const char *skipExisting = NULL;
    const char *recordFilename = NULL;
    const char *recordObjectFilename = NULL;
    const char *outObjectMovesExt = NULL;

    int maxMoves = numeric_limits<int>::max();
    bool bDisableMerges = false;
    bool bDisableSegDict = false;
    bool bEnableObjMoves = false;
    bool bEnableObjCandidateMoves = false;
    bool bRecordDeltaEnergyForObjDetMoves = false;
    bool bRecordPedestrian = false;
    bool bVisualizeIterations = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-m", modelType)
        SVL_CMDLINE_STR_OPTION("-inRegions", inRegionsExt)
        SVL_CMDLINE_STR_OPTION("-inDir", inRegionDir)
        SVL_CMDLINE_STR_OPTION("-outRegions", outRegionsExt)
        SVL_CMDLINE_STR_OPTION("-outLabels", outLabelsExt)
        SVL_CMDLINE_STR_OPTION("-outClasses", outClassesExt)
        SVL_CMDLINE_STR_OPTION("-outObjects", outObjectsExt)
        SVL_CMDLINE_STR_OPTION("-outObjectMoves", outObjectMovesExt)
        SVL_CMDLINE_STR_OPTION("-outHorizon", outHorizonExt)
        SVL_CMDLINE_STR_OPTION("-outImages", outImagesExt)
        SVL_CMDLINE_STR_OPTION("-outSingletons", outSingletonExt)
        SVL_CMDLINE_STR_OPTION("-skipExisting", skipExisting)
        SVL_CMDLINE_INT_OPTION("-maxMoves", maxMoves)
        SVL_CMDLINE_BOOL_OPTION("-disableMerges", bDisableMerges)
        SVL_CMDLINE_BOOL_OPTION("-disableSegDict", bDisableSegDict)
        SVL_CMDLINE_BOOL_OPTION("-enableObjMoves", bEnableObjMoves)
        SVL_CMDLINE_BOOL_OPTION("-enableObjCandMoves", bEnableObjCandidateMoves)
        SVL_CMDLINE_STR_OPTION("-record", recordFilename)
        SVL_CMDLINE_STR_OPTION("-recordObject", recordObjectFilename)
        SVL_CMDLINE_BOOL_OPTION("-recordPedestrian", bRecordPedestrian)
        SVL_CMDLINE_BOOL_OPTION("-recordDeltaEnergy", bRecordDeltaEnergyForObjDetMoves)
        SVL_CMDLINE_STR_OPTION("-video", state.videoDir)
        SVL_CMDLINE_BOOL_OPTION("-x", state.bVisualize)
        SVL_CMDLINE_BOOL_OPTION("-x0", bVisualizeIterations)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *evalList = SVL_CMDLINE_ARGV[0];

    // read list of evaluation images
    vector<string> baseNames;
    if (svlFileExists(evalList)) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading evaluation list from " << evalList << "...");
        baseNames = svlReadFile(evalList);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");
    } else {
        baseNames.push_back(string(evalList));
    }

    // randomize basenames
    if (skipExisting != NULL) {
        timeval t;
        gettimeofday(&t, NULL);
        srand(t.tv_usec * t.tv_sec);
        vector<int> order = randomPermutation(baseNames.size());
        vector<string> reorderedBaseNames(baseNames.size());
        for (unsigned i = 0; i < baseNames.size(); i++) {
            reorderedBaseNames[i] = baseNames[order[i]];
        }
        baseNames = reorderedBaseNames;
    }

    // read models
    SceneRegionModel *model = NULL;
    if (!strcasecmp(modelType, "REGION")) {
        model = new SceneRegionModel();
        SVL_ASSERT(!bEnableObjMoves && !bEnableObjCandidateMoves);
    } else if (!strcasecmp(modelType, "REGIONPAIR")) {
        model = new SceneRegionPairModel();
        SVL_ASSERT(!bEnableObjMoves && !bEnableObjCandidateMoves);
    } else if (!strcasecmp(modelType, "OBJECT")) {
        model = new SceneObjectModel();
    } else if (!strcasecmp(modelType, "OBJECTCONTEXT")) {
        SVL_LOG(SVL_LOG_FATAL, "not implemented yet");
    } else {
        SVL_LOG(SVL_LOG_FATAL, "unrecognized model type " << modelType);
    }
    SVL_ASSERT(model != NULL);

    model->loadModels((gSceneConfig.getOption("baseDir") +
            gSceneConfig.getOption("modelsDir")).c_str());
    svlConfusionMatrix regionConfusion(gSceneConfig.numRegionLabels);

    // closed-loop learning datasets
    SceneFeatureCache closedLoopLearningDataset;
    SceneFeatureCache closedLoopLearningDatasetObject;
    SceneLayersAndLabels trueLabels(1, 1);

    // process images
    SVL_LOG(SVL_LOG_MESSAGE, "Processing images...");
    int hProcessImage = svlCodeProfiler::getHandle("processImage");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        // check if we should skip this file
        if (skipExisting != NULL) {
            string checkFilename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(skipExisting);
            if (svlFileExists(checkFilename.c_str())) {
                continue;
            }
        }

        svlCodeProfiler::tic(hProcessImage);
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");

        // load image
        IplImage *img = cvLoadImage(imgFilename.c_str(), CV_LOAD_IMAGE_COLOR);
        SVL_ASSERT_MSG(img != NULL, "path: " << imgFilename);
        SVL_LOG(SVL_LOG_MESSAGE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*img));

        // load labels
        CvMat *classLabels = NULL;
        string lblFilename = gSceneConfig.filename("lblDir", baseNames[i], "regionLblExt");
        classLabels = cvCreateMat(img->height, img->width, CV_32SC1);
        SVL_ASSERT(classLabels != NULL);
        bool bSuccess = readMatrix(classLabels, lblFilename.c_str());
        if (!bSuccess) {
            cvReleaseMat(&classLabels);
            SVL_LOG(SVL_LOG_ERROR, "Could not read class labels for " << baseNames[i]);
        }

        // initialize scene
        SceneInstance instance(img, baseNames[i].c_str());

        if (inRegionsExt != NULL) {
            string filename;
            if (inRegionDir == NULL) {
                filename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(inRegionsExt);
            } else {
                filename = string(inRegionDir) + string("/") + baseNames[i] + string(inRegionsExt);
            }
            SVL_LOG(SVL_LOG_MESSAGE, "initializing regions from " << filename << "...");
            bool bSuccess = readMatrix(instance.assignment.pixels, filename.c_str());
            if (!bSuccess) {
                SVL_LOG(SVL_LOG_FATAL, "could not initialize regions from file for " << baseNames[i]);
            }
            instance.segmentations.addSegmentation(instance.assignment.pixels);
        }

        instance.loadSegmentations(false);

        // true instance for closed-loop learning
        if (recordFilename != NULL) {
            trueLabels.clear(img->width, img->height);
            trueLabels.load(gSceneConfig.filename("lblDir", baseNames[i], "layerLblExt").c_str(),
                gSceneConfig.filename("lblDir", baseNames[i], "layerClassLblExt").c_str());
            trueLabels.getLayerSizes();
            trueLabels.getLayerExtents();
            state.pTrueLabels = &trueLabels;
            state.pClosedLoopLearningDataset = &closedLoopLearningDataset;
        }

        if (recordObjectFilename != NULL) {
            if (recordFilename == NULL) {
                trueLabels.clear(img->width, img->height);
                trueLabels.load(gSceneConfig.filename("lblDir", baseNames[i], "layerLblExt").c_str(),
                    gSceneConfig.filename("lblDir", baseNames[i], "layerClassLblExt").c_str());
                trueLabels.getLayerSizes();
                trueLabels.getLayerExtents();
                state.pTrueLabels = &trueLabels;
            }
            state.pClosedLoopLearningDatasetObject = &closedLoopLearningDatasetObject;
        }

        // first-stage inference
        SVL_LOG(SVL_LOG_MESSAGE, "inferring pixel stage...");
        model->inferPixels(&instance);


        // second-stage inference
        SVL_LOG(SVL_LOG_MESSAGE, "inferring horizon stage...");
        model->inferHorizon(&instance);

        // third-stage inference
        SVL_LOG(SVL_LOG_MESSAGE, "inferring region stage...");
        double e = model->inferRegions(&instance);

        // FOR Richard, only computing the features
        continue; // TODO: remove
        //instance.validateData();

#if 1
        // assign one object per foreground region
        if (!strcasecmp(modelType, "OBJECT")) {
            SVL_LOG(SVL_LOG_MESSAGE, "inferring object stage...");
            for (int k = 0; k < instance.numRegions(); k++) {
                if (instance.assignment.regions[k]->Sk == gSceneConfig.numRegionLabels - 1) {
                    instance.moveRegions(k, -1);
                }
            }

            e = model->inferRegions(&instance);

            if (recordObjectFilename != NULL) {
                for (int k = 0; k < instance.numObjects(); k++) {
                    recordLearningExampleForObjectModel(state, k, 1, instance, model);
                }
            }
        }
#endif

        // initialize search state
        SVL_LOG(SVL_LOG_VERBOSE, "initializing state with " << instance.numRegions() << " regions");
        state.init(e, instance.assignment);

        // generate proposal dictionary
        SceneSegmentDictionary dictionary;
        for (int s = 0; s < instance.segmentations.count(); s++) {
            instance.assignPixels(instance.segmentations.segmentation(s));
            for (int k = 0; k < instance.numRegions(); k++) {
                if (instance.assignment.regions[k]->size() > 1) {
                    dictionary.addSegment(SceneRegionFeatures::regionEntropy(&instance, k),
                        SceneSegment(instance.assignment.pixels, k));
                }
            }
        }

        // initialize object candidate moves
        SceneTopDownMoveDatabase objCandidateMoves;
        if (bEnableObjCandidateMoves) {
            objCandidateMoves.load(instance);
            if (outObjectMovesExt) {
                state.pObjCandidateRecords = new set<objCandT, objCandCmp>;
            }

            //            if (outObjectMovesExt != NULL || recordObjectFilename != NULL) {
            // maxMoves = objCandidateMoves.moves.size();
            // }
        }

        // reset instance to initialize assignment
        instance.assignment = state.bestAssignment;
        SVL_LOG(SVL_LOG_VERBOSE, "...regions: " << instance.numRegions()
            << ", objects: " << instance.numObjects() << ", energy: " << e);

        // show results
        if (state.bVisualize || bVisualizeIterations) {
            viewProgress(instance, state);
        }
        dumpVideoFrame(instance, state);

        // search over proposal moves
        while (!state.bConverged && (state.numMoves < maxMoves)) {
            state.bConverged = true;
            instance.validateData();

            // merge-moves
            if (!bDisableMerges) {
                mergeMoves(state, model, instance);
                if (state.numMoves > maxMoves) break;
            }

            // re-segment moves
            if (!bDisableSegDict) {
                segmentDictionaryMoves(state, model, instance, dictionary);
                if (state.numMoves > maxMoves) break;
            }

            // create objects associated with newly created foreground regions
            if (!strcasecmp(modelType, "OBJECT")) {
                objectDeleteMoves(state, model, instance);

                bool bNewObjectCreated = false;
                for (vector<SceneRegion *>::iterator r = instance.assignment.regions.begin();
                     r != instance.assignment.regions.end(); r++) {
                    if (((*r)->Sk == gSceneConfig.numRegionLabels - 1) && ((*r)->object == NULL)) {
                        instance.moveRegions((*r)->id(), -1);
                        bNewObjectCreated = true;
                    }
                }

                if (bNewObjectCreated) {
                    e = model->inferRegions(&instance);
                    state.update(e, instance.assignment);
                }
            }

            // object merges and splits
            if (bEnableObjMoves) {
                objectSplitMoves(state, model, instance);
                objectMergeMoves(state, model, instance);
                if (state.numMoves > maxMoves) break;
            }

            // top-down object candidate moves
            if (bEnableObjCandidateMoves) {
                objDetCandidateMoves(state, model, instance, objCandidateMoves, bRecordDeltaEnergyForObjDetMoves, bRecordPedestrian);
                if (outObjectMovesExt != NULL || recordObjectFilename != NULL) {
                    break;
                }
            }

            // update horizon
            model->inferHorizon(&instance);

            // show results
            state.boundaryPixels.clear();
            state.boundingBox = cvRect(0, 0, 0, 0);
            if (bVisualizeIterations) {
                viewProgress(instance, state);
            }
            dumpVideoFrame(instance, state);
        }

        // show results
        state.boundaryPixels.clear();
        state.boundingBox = cvRect(0, 0, 0, 0);
        if (state.bVisualize || bVisualizeIterations) {
            viewProgress(instance, state);
        }
        dumpVideoFrame(instance, state);

        SVL_LOG(SVL_LOG_VERBOSE, "...regions: " << instance.numRegions() <<
            ", objects: " << instance.numObjects());
        SVL_LOG(SVL_LOG_MESSAGE, "final energy is " << model->energy(&instance));

        // accumulate predicted classes
        CvMat *predictedClasses = instance.getPixelClasses();
        if (classLabels != NULL) {
            for (int y = 0; y < instance.height(); y++) {
                for (int x = 0; x < instance.width(); x++) {
                    int regionClassLabel = std::min(CV_MAT_ELEM(*classLabels, int, y, x),
                        gSceneConfig.numRegionLabels);
                    regionConfusion.accumulate(regionClassLabel,
                        CV_MAT_ELEM(*predictedClasses, int, y, x));
                }
            }
        }

        // output images
        if (outImagesExt) {
            IplImage *canvas = visualize(&instance);
            cvSaveImage((gSceneConfig.filebase("outputDir", baseNames[i]) +
                    string(outImagesExt)).c_str(), canvas);
            cvReleaseImage(&canvas);
        }

        // output regions
        if (outRegionsExt) {
            writeMatrix(instance.assignment.pixels, (gSceneConfig.filebase("outputDir", baseNames[i]) +
                    string(outRegionsExt)).c_str());
        }

        // output region labels
        if (outLabelsExt) {
            // TODO: extend to objects, etc
            ofstream ofs((gSceneConfig.filebase("outputDir", baseNames[i]) + string(outLabelsExt)).c_str());
            for (vector<SceneRegion *>::const_iterator it = instance.assignment.regions.begin();
                 it != instance.assignment.regions.end(); it++) {
                ofs << (*it)->Sk << "\n";
            }
            ofs.close();
        }

        // output classes
        if (outClassesExt) {
            writeMatrix(predictedClasses, (gSceneConfig.filebase("outputDir", baseNames[i]) +
                    string(outClassesExt)).c_str());
        }

        // output objects
        if (outObjectsExt) {
            ofstream ofs((gSceneConfig.filebase("outputDir", baseNames[i]) + string(outObjectsExt)).c_str());
            for (vector<SceneObject *>::const_iterator it = instance.assignment.objects.begin();
                 it != instance.assignment.objects.end(); it++) {
                ofs << (*it)->extent().transpose()
                    << " " << (*it)->prObject.transpose()
                    << " " << (*it)->objDetScore.transpose() << "\n";
            }
            ofs.close();
        }

        // output object moves
        if (outObjectMovesExt) {
            ofstream ofs((gSceneConfig.filebase("outputDir", baseNames[i]) + string(outObjectMovesExt)).c_str());
            for (set<objCandT, objCandCmp>::const_iterator it = state.pObjCandidateRecords->begin();
                 it != state.pObjCandidateRecords->end(); it++) {
                ofs << it->boundingBox.transpose()
                    << " " << it->prObject.transpose()
                    << " " << it->objDetScore.transpose()
                    << " " << it->bAccepted
                    << " " << it->deltaEnergy.transpose() << "\n";
            }
            ofs.close();
        }

        // output horizon
        if (outHorizonExt) {
            string filename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(outHorizonExt);
            ofstream ofs(filename.c_str());
            SVL_ASSERT(!ofs.fail());
            ofs << instance.assignment.horizon << "\n";
            ofs.close();
        }

        // output singletons
        if (outSingletonExt) {
            string filename = gSceneConfig.filebase("outputDir", baseNames[i]) +
                string(outSingletonExt);
            ofstream ofs(filename.c_str());
            SVL_ASSERT(!ofs.fail());
            for (int k = 0; k < instance.numRegions(); k++) {
                ofs << instance.assignment.regions[k]->phi.transpose() << "\n";
            }
            ofs.close();
        }

        // show final instance
        if (svlLogger::getLogLevel() >= SVL_LOG_DEBUG) {
            instance.printScene();
        }

        // compute score for this image
        if (classLabels != NULL) {
            CvMat *mask = cvCreateMat(predictedClasses->rows, predictedClasses->cols, CV_8UC1);
            cvCmpS(predictedClasses, -1.0, mask, CV_CMP_LE);
            cvSub(classLabels, predictedClasses, predictedClasses);
            int errors = cvCountNonZero(predictedClasses);
            int missing = cvCountNonZero(mask);
            SVL_LOG(SVL_LOG_MESSAGE, "...region class score: "
                << (1.0 - (double)(errors - missing) / (img->width * img->height - missing)));
            cvReleaseMat(&mask);
        }

        // write closed-loop training data
        if (recordFilename != NULL) {
            closedLoopLearningDataset.write(recordFilename);
            trueLabels.clear();
            state.pClosedLoopLearningDataset = NULL;
            state.pTrueLabels = NULL;
        }

        if (recordObjectFilename != NULL) {
            closedLoopLearningDatasetObject.write(recordObjectFilename);
            state.pClosedLoopLearningDatasetObject = NULL;
            if (recordFilename == NULL) {
                trueLabels.clear();
                state.pTrueLabels = NULL;
            }
        }

        // free memory
        state.bestAssignment.clear();
        cvReleaseMat(&predictedClasses);
        if (classLabels != NULL) {
            cvReleaseMat(&classLabels);
        }
        cvReleaseImage(&img);

        if (state.pObjCandidateRecords != NULL) {
            delete state.pObjCandidateRecords;
        }

        svlCodeProfiler::toc(hProcessImage);
    }

    // show accuracy
    if (svlLogger::getLogLevel() >= SVL_LOG_VERBOSE) {
        regionConfusion.printRowNormalized(cout, "--- Class Confusion Matrix ---");
    }
    SVL_LOG(SVL_LOG_MESSAGE, "Overall class accuracy: " << regionConfusion.accuracy());
    SVL_LOG(SVL_LOG_MESSAGE, "Average class accuracy: " << regionConfusion.avgRecall());

    // free memory
    delete model;
    if ((baseNames.size() == 1) && (state.bVisualize || bVisualizeIterations)) {
        cvWaitKey(-1);
    }
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

// move-making functions ---------------------------------------------------------------

// common functionality to call after making moves
void finalizeMove(InferenceState& state, SceneRegionModel *model, SceneInstance& instance)
{
    // reset configuration to best assignment and clean up orphaned and small regions
    instance.assignment.swap(state.bestAssignment);
    instance.splitOrphans();
    instance.removeSmallRegions();
    model->inferRegions(&instance);

    state.bestAssignment = instance.assignment;

    // show results
    if (state.bVisualize) {
        viewProgress(instance, state);
    }
    dumpVideoFrame(instance, state);

    state.boundingBox = cvRect(0, 0, 0, 0);
    state.boundaryPixels.clear();
}

void mergeMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance)
{
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << instance.numRegions() << " merges...");
    int regId = 0;
    while (regId < instance.numRegions()) {
        SceneRegion *region = instance.assignment.regions[regId];
        SVL_ASSERT(region->id() == regId);
        if (region->empty()) {
            regId += 1;
            continue;
        }

        // show merge move
        state.boundaryPixels = set<CvPoint>(region->pixels.begin(), region->pixels.end());

        state.numMoves += 1;
        set<SceneRegion *> nearRegions;
        for (set<SceneRegionBoundary *>::const_iterator it = region->boundaries.begin();
             it != region->boundaries.end(); it++) {
            nearRegions.insert((*it)->neighborOf(region));
        }
        vector<CvPoint> pixels(region->pixels.begin(), region->pixels.end());

        // evaluate neighbors
        for (set<SceneRegion *>::const_iterator r = nearRegions.begin(); r != nearRegions.end(); ++r) {
            if ((*r)->empty()) break;

            instance.movePixels(pixels, (*r)->id());
            double e = model->inferRegions(&instance);

            SVL_LOG(SVL_LOG_DEBUG, "energy for merging " << regId
                << " with " << (*r)->id() << " is " << e);
            bool bAcceptMove = (state.bestEnergy - e > SVL_EPSILON);

            // closed-loop learning
            recordLearningExample(state, (*r)->id(), bAcceptMove, instance);

            // make the move
            if (bAcceptMove) {
                state.update(e, instance.assignment);
            }
        }

        // finalize the move
        state.boundingBox = cvRect(0, 0, 0, 0);
        finalizeMove(state, model, instance);
        regId += 1;
    }
}

void segmentDictionaryMoves(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneSegmentDictionary& dictionary)
{
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << dictionary.size() << " moves...");
    for (map<double, SceneSegment>::const_iterator it = dictionary.entries.begin(); it != dictionary.entries.end(); ++it) {
        state.numMoves += 1;

        // neighbor regions
        set<int> nearRegions;
        for (vector<CvPoint>::const_iterator p = it->second.pixels.begin();
             p != it->second.pixels.end(); ++p) {
            nearRegions.insert(CV_MAT_ELEM(*instance.assignment.pixels, int, p->y, p->x));
        }
        set<CvPoint> const& nbrs = it->second.outsideBoundaryPixels;
        for (set<CvPoint>::const_iterator p = nbrs.begin(); p != nbrs.end(); ++p) {
            nearRegions.insert(CV_MAT_ELEM(*instance.assignment.pixels, int, p->y, p->x));
        }

        // "new" region
        nearRegions.insert(-1);

        SVL_ASSERT(nearRegions.size() > 1);
#if 1
        if ((nearRegions.size() == 2) && (it->second.pixels.size() < 2.0e-3 * instance.size())) {
            SVL_LOG(SVL_LOG_DEBUG, "skipping small internal move of size " << it->second.pixels.size());
            continue;
        }
#endif

        // evaluate neighbors
        for (set<int>::const_iterator s = nearRegions.begin(); s != nearRegions.end(); ++s) {
            int numPixelsMoved = instance.movePixels(it->second.pixels, *s);
            if (numPixelsMoved == 0) continue;
            double e = model->inferRegions(&instance);

            bool bAcceptMove = (state.bestEnergy - e > SVL_EPSILON);

            // closed-loop learning
            if (*s >= 0) recordLearningExample(state, *s, bAcceptMove, instance);

            // make the move
            if (bAcceptMove) {
                state.update(e, instance.assignment);

                // show results
                state.boundaryPixels = it->second.insideBoundaryPixels;
                //state.boundaryPixels = set<CvPoint>(it->second.pixels.begin(), it->second.pixels.end());
                if (state.bVisualize) {
                    viewProgress(instance, state);
                }
                dumpVideoFrame(instance, state);
            }
        }

        // finalize the move
        //state.boundaryPixels = set<CvPoint>(it->second.pixels.begin(), it->second.pixels.end());
        state.boundaryPixels = it->second.insideBoundaryPixels;
        finalizeMove(state, model, instance);
    }
}

void objectDeleteMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance)
{
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << instance.assignment.numObjects() << " object deletions...");
    for (int objId = instance.assignment.numObjects() - 1; objId >= 0; objId--) {
        SVL_LOG(SVL_LOG_DEBUG, "...deleting object " << objId);
        int numPixelsMoved = instance.removeObject(objId);
        if (numPixelsMoved == 0) continue;

        state.numMoves += 1;
        double e = model->inferRegions(&instance);

        // make the move
        if (state.bestEnergy - e > SVL_EPSILON) {
            state.update(e, instance.assignment);

            // show results
            if (state.bVisualize) {
                viewProgress(instance, state);
            }
            dumpVideoFrame(instance, state);
        }

        // finalize the move
        finalizeMove(state, model, instance);
    }
}

void objectMergeMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance)
{
    // for each region neighboring each object, attempt to move the region into the object
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << instance.assignment.numObjects() << " object merge moves...");
    for (vector<SceneObject *>::const_iterator o = instance.assignment.objects.begin();
         o != instance.assignment.objects.end(); o++) {

        // get set of neighbouring regions
        set<SceneRegion *> neighborRegions;
        for (set<SceneObjRegBoundary *>::const_iterator it = (*o)->boundaries.begin();
             it != (*o)->boundaries.end(); it++) {
            if ((*it)->region != NULL)
                neighborRegions.insert((*it)->region);
        }

        // try merge regions greedily
        for (set<SceneRegion *>::const_iterator r = neighborRegions.begin(); r != neighborRegions.end(); r++) {
            state.numMoves += 1;

            SceneObject *oldObject = (*r)->object;
            int numPixelsMoved = instance.moveRegions((*r)->id(), (*o)->id());
            if (numPixelsMoved == 0) continue;

            double e = model->inferRegions(&instance);
            SVL_LOG(SVL_LOG_VERBOSE, "merging region " << (*r)->id() << " with object "
                << (*o)->id() << " has energy " << e);

            bool bAcceptMove = (state.bestEnergy - e > SVL_EPSILON);
            // closed-loop learning
            recordLearningExampleForObjectModel(state, (*o)->id(), bAcceptMove, instance, model);

            // make the move
            if (bAcceptMove) {
                state.update(e, instance.assignment);

                // show results
                state.boundaryPixels.clear();
                state.boundaryPixels.insert((*r)->pixels.begin(), (*r)->pixels.end());
                if (state.bVisualize) {
                    viewProgress(instance, state);
                }
                dumpVideoFrame(instance, state);
            } else {
                // undo the move
                if (oldObject == NULL) {
                    instance.removeRegions((*r)->id(), (*o)->id());
                } else {
                    instance.moveRegions((*r)->id(), oldObject->id());
                }
            }
        }
    }

    // finalize the move
    finalizeMove(state, model, instance);
}

void objectSplitMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance)
{
    // take all regions within an object and create one object for each of them
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << instance.assignment.numObjects() << " object split moves...");
    for (int objId = 0; objId < instance.assignment.numObjects(); objId++) {
        SceneObject *object = instance.assignment.objects[objId];

        // only split objects with 2 or more regions
        if (object->regions.size() < 2) continue;

        state.numMoves += 1;

        list<CvPoint> pixels = object->pixels();
        state.boundaryPixels = set<CvPoint>(pixels.begin(), pixels.end());
        int numPixelsMoved = 0;
        while (object->regions.size() > 1) {
            numPixelsMoved += instance.moveRegions((*object->regions.begin())->id(), -1);
        }

        if (numPixelsMoved == 0) continue;
        double e = model->inferRegions(&instance);

        // make the move
        if (state.bestEnergy - e > SVL_EPSILON) {
            state.update(e, instance.assignment);

            // show results
            if (state.bVisualize) {
                viewProgress(instance, state);
            }
            dumpVideoFrame(instance, state);
        }

        // finalize the move
        finalizeMove(state, model, instance);
    }
}

void objDetCandidateMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance,
    SceneTopDownMoveDatabase &moves, bool bRecordDeltaEnergyForObjDetMoves, bool bRecordPedestrian)
{
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << moves.numMoves() << " object candidate moves...");

    vector<vector<double> > objDetScores(moves.numMoves());
    bool bUseCache = gSceneConfig.getOptionAsBool("useCache");
    bool bLoadFromCache = false;
    if (bUseCache) {
        SceneFeatureCache cache;
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "objDetMoveScoreExt");
        if (svlFileExists(filename.c_str())) {
            cache.read(filename.c_str(), true);
            objDetScores = cache.features;
            bLoadFromCache = true;
            SVL_LOG(SVL_LOG_MESSAGE, "loaded object move scores from cache...");
        }
        SVL_ASSERT(moves.numMoves() == (int)objDetScores.size());
    }

    for (int i = 0; i < moves.numMoves(); i++) {
        // find overlapping region
        CvMat *mask = moves.mask(i);
        int size = cvCountNonZero(mask);
        if (size < 50) {
            if (bUseCache && bLoadFromCache == false) {
                objDetScores[i] = vector<double>(gSceneConfig.numObjectLabels, 0.0);
            }
            cvReleaseMat(&mask);
            continue;
        }

        // record the score and delta energy for this object move
        VectorXd energyBeforeMove;
        if (bRecordDeltaEnergyForObjDetMoves) {
            if (state.pObjCandidateRecords != NULL) {
                ((SceneObjectModel *)model)->inferRegions(&instance, &energyBeforeMove);
            }
        }

        CvMat *internalRegions = cvCreateMat(instance.height(), instance.width(), CV_32SC1);
        cvSet(internalRegions, cvScalar(-1));
        cvCopy(instance.assignment.pixels, internalRegions, mask);

        // get list of regions overlapping mask
        map<int, vector<CvPoint> > newRegions;

        const int *p = (const int *)CV_MAT_ELEM_PTR(*internalRegions, 0, 0);
        for (int y = 0; y < instance.height(); y++) {
            for (int x = 0; x < instance.width(); x++, p++) {
                if (*p == -1) continue;
                if (newRegions.find(*p) == newRegions.end()) {
                    newRegions.insert(make_pair(*p, vector<CvPoint>()));
                }
                newRegions[*p].push_back(cvPoint(x, y));
            }
        }

        // create new regions and add to object
        instance.assignment.objects.push_back(new SceneObject(instance.assignment.objects.size()));
        int totalNumMoved = 0;
        for (map<int, vector<CvPoint> >::const_iterator jt = newRegions.begin(); jt != newRegions.end(); jt++) {
            if (jt->second.empty()) continue;
            int numMoved = instance.movePixels(jt->second, (jt == newRegions.begin()) ? -1 : instance.numRegions() - 1);
            totalNumMoved += numMoved;
            SVL_ASSERT(numMoved != 0);
        }

        int numMovedRegion = instance.moveRegions(instance.numRegions() - 1, instance.numObjects() - 1);
        SVL_LOG(SVL_LOG_DEBUG, "numMovedRegion: " << numMovedRegion << " " << endl);
        Vector4i extent;

        instance.splitOrphans();
        instance.removeSmallRegions();

        state.numMoves += 1;

        // set the score if loaded from the cache
        if (bUseCache && bLoadFromCache) {
            SVL_ASSERT(objDetScores[i].size() == (unsigned)gSceneConfig.numObjectLabels);
            VectorXd objDetScore = Eigen::Map<VectorXd>(&objDetScores[i][0], objDetScores[i].size());
            instance.assignment.objects.back()->setObjDetScoreFromCache(objDetScore);
        }


        // record training samples for pedestrian classifier, only compute features, no inference
        if (state.pClosedLoopLearningDatasetObject != NULL) {
            if (bRecordPedestrian) {
                recordLearningExampleForObjectModel(state, instance.assignment.objects.size()-1, 1.0, instance, model);
                // free memory
                cvReleaseMat(&mask);
                cvReleaseMat(&internalRegions);

                if (state.numMoves >= 40) {
                    break;
                } else {
                    continue;
                }
            }
        }
        // evaluate the move
        VectorXd energyAfterMove;
        double e = ((SceneObjectModel *)model)->inferRegions(&instance, &energyAfterMove);

        if (bUseCache && bLoadFromCache == false) {
            objDetScores[i].resize(gSceneConfig.numObjectLabels);
            SceneObject* object = instance.assignment.objects.back();
            SVL_ASSERT(object->objDetScore.rows() >= objDetScores[i].size());
            Eigen::Map<VectorXd>(&objDetScores[i][0], objDetScores[i].size())
                = Eigen::Map<VectorXd>(&(object->objDetScore[0]), objDetScores[i].size());
        }

        bool bAcceptMove = (state.bestEnergy - e > SVL_EPSILON);
        // closed-loop learning
        recordLearningExampleForObjectModel(state, instance.assignment.objects.size()-1, bAcceptMove, instance, model);

#if 0
        IplImage* maskImageGray = cvCreateImage(cvSize(mask->width, mask->height), 8, 1);
        IplImage* maskImage = cvCreateImage(cvSize(mask->width, mask->height), 8, 3);
        cvConvertScale(mask, maskImageGray, 250);
        cvCvtColor(maskImageGray, maskImage, CV_GRAY2RGB);
        extent = instance.assignment.objects.back()->extent();
        cout << "obj extent: " << extent << endl;
        SceneObject* obj = instance.assignment.objects.back();
        cout << "obj size: " << obj->size() << endl;
        cout << "obj singleton: " << obj->prObject[0] << ", "
             << obj->prObject[1] << ", " << obj->prObject[2] << endl;
        cout << "obj detScore: " << obj->objDetScore[0] << ", "
             << obj->objDetScore[1] << endl;

        cout << "original detection score: " << -log(1/it->first - 1) << endl;

        cvRectangle(maskImage, cvPoint(extent[0], extent[1]),
            cvPoint(extent[0] + extent[2], extent[1] + extent[3]),
            CV_RGB(0,255,0), 3);
        cvNamedWindow("debug", 1);
        cvShowImage("debug", maskImage);
        cvWaitKey(-1);
        cvDestroyWindow("debug");
        cvReleaseImage(&maskImageGray);
        cvReleaseImage(&maskImage);
#endif

        // record the score and delta energy for the object move
        if (state.pObjCandidateRecords != NULL) {
            objCandT objCand;
            SceneObject* obj = instance.assignment.objects.back();
            objCand.boundingBox = obj->extent();
            objCand.prObject = obj->prObject;
            objCand.objDetScore = obj->objDetScore;
            objCand.bAccepted = state.bestEnergy - e > SVL_EPSILON ? true : false;
            if (bRecordDeltaEnergyForObjDetMoves) {
                SVL_ASSERT(energyAfterMove.size() == energyBeforeMove.size());
                objCand.deltaEnergy = energyAfterMove - energyBeforeMove;
            } else {
                objCand.deltaEnergy = VectorXd::Zero(energyAfterMove.rows());
            }
            state.pObjCandidateRecords->insert(objCand);
        }

        // keep the move if better
        if (bAcceptMove) {
            state.update(e, instance.assignment);
            // show results
            //if (state.bVisualize) {
            //viewProgress(instance, state);
            //}
            dumpVideoFrame(instance, state);
        }

        // finalize the move
        if (state.bVisualize) {
            Vector4i e = instance.assignment.objects.back()->extent();
            state.boundingBox = cvRect(e[0], e[1], e[2], e[3]);
        }
        finalizeMove(state, model, instance);
        cvReleaseMat(&mask);
        cvReleaseMat(&internalRegions);
    }

    if (bUseCache && bLoadFromCache == false) {
        SceneFeatureCache cache;
        cache.features = objDetScores;
        cache.labels = vector<int>(objDetScores.size(), 0);
        cache.weights = vector<double>(objDetScores.size(), 1.0);
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "objDetMoveScoreExt");
        cache.write(filename.c_str());
    }
}

// visualization functions -------------------------------------------------------------

void viewProgress(SceneInstance& instance, const InferenceState& state)
{
    IplImage *canvas = visualize(&instance);

    if ((state.boundingBox.width > 0) && (state.boundingBox.height > 0)) {
        cvRectangle(canvas, cvPoint(state.boundingBox.x, state.boundingBox.y),
            cvPoint(state.boundingBox.x + state.boundingBox.width, state.boundingBox.y + state.boundingBox.height),
            CV_RGB(0, 255, 0), 3);
    }

    for (set<CvPoint>::const_iterator p = state.boundaryPixels.begin();
         p != state.boundaryPixels.end(); ++p) {
        cvRectangle(canvas, *p, *p, CV_RGB(255, 255, 255), 5);
    }

    for (set<CvPoint>::const_iterator p = state.boundaryPixels.begin();
         p != state.boundaryPixels.end(); ++p) {
        cvRectangle(canvas, *p, *p, CV_RGB(255, 0, 0), 3);
    }

    cvNamedWindow(WINDOW_NAME, 1);
    cvShowImage(WINDOW_NAME, canvas);
    int ch = cvWaitKey(100);
    if ((char)ch == 'q') {
        if (svlLogger::getLogLevel() >= SVL_LOG_DEBUG) {
            instance.printScene();
        }
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
        svlCodeProfiler::print(cerr);
        exit(0);
    }

    if ((char)ch == 'p') {
        cvWaitKey(-1);
    }
    cvReleaseImage(&canvas);
}

void dumpVideoFrame(SceneInstance& instance, const InferenceState& state)
{
    static int frameCount = 0;
    if (state.videoDir == NULL)
        return;

#if 0
    IplImage *canvas = visualize(&instance);
#else
    vector<IplImage *> views;
    IplImage *grayImage = cvCreateImage(cvGetSize(instance.image()), IPL_DEPTH_8U, 1);
    cvCvtColor(instance.image(), grayImage, CV_BGR2GRAY);
    IplImage *background = cvCreateImage(cvGetSize(instance.image()), IPL_DEPTH_8U, 3);
    for (int c = 1; c <= 3; c++) {
        cvSetImageCOI(background, c);
        cvCopy(grayImage, background);
    }
    cvSetImageCOI(background, 0);
    cvReleaseImage(&grayImage);

    views.push_back(cvCloneImage(instance.image()));

    views.push_back(cvCloneImage(instance.image()));
    cvSet(views.back(), cvScalar(255, 255, 255));
    visualizeRegions(&instance, views.back());
    visualizeClassBoundaries(&instance, views.back(), true);

    views.push_back(cvCloneImage(background));
    visualizeClasses(&instance, views.back(), 0.5);
    visualizeObjects(&instance, views.back());

    views.push_back(cvCloneImage(background));
    visualizeHorizon(&instance, views.back());

    IplImage *canvas = combineImages(views, 2, 2);
    svlReleaseImages(views);
    cvReleaseImage(&background);
#endif

    if ((state.boundingBox.width > 0) && (state.boundingBox.height > 0)) {
        cvRectangle(canvas, cvPoint(state.boundingBox.x, state.boundingBox.y),
            cvPoint(state.boundingBox.x + state.boundingBox.width, state.boundingBox.y + state.boundingBox.height),
            CV_RGB(0, 255, 0), 3);
    }

    for (set<CvPoint>::const_iterator p = state.boundaryPixels.begin();
         p != state.boundaryPixels.end(); ++p) {
        cvRectangle(canvas, *p, *p, CV_RGB(255, 255, 255), 5);
    }

    for (set<CvPoint>::const_iterator p = state.boundaryPixels.begin();
         p != state.boundaryPixels.end(); ++p) {
        cvRectangle(canvas, *p, *p, CV_RGB(255, 0, 0), 3);
    }

    string filename = string(state.videoDir) + string("/frame") +
        padString(toString(frameCount), 4) + string(".png");
    cvSaveImage(filename.c_str(), canvas);
    cvReleaseImage(&canvas);

    frameCount += 1;
}

// closed-loop learning prototypes ------------------------------------------

void recordLearningExample(InferenceState& state, int regId, bool bAccepted,
    SceneInstance& instance)
{
    if ((state.pClosedLoopLearningDataset == NULL) || (state.pTrueLabels == NULL))
        return;

    SceneRegion *region = instance.assignment.regions[regId];

    // don't add if region is too small
    if (region->size() < 1.0e-3 * instance.size())
        return;

    // find groundtruth labels for proposal move
    vector<int> regionLabelCount(state.pTrueLabels->size(), 0);
    vector<int> classLabelCount(gSceneConfig.numRegionLabels, 0);
    int unknownLabelCount = 0;

    for (list<CvPoint>::const_iterator p = region->pixels.begin(); p != region->pixels.end(); p++) {
        int gtRegId = CV_MAT_ELEM(*state.pTrueLabels->layers, int, p->y, p->x);
        SVL_ASSERT((gtRegId >= 0) && (gtRegId < (int)state.pTrueLabels->labels.size()));
        if (state.pTrueLabels->labels[gtRegId] < 0) {
            unknownLabelCount += 1;
        } else {
            regionLabelCount[gtRegId] += 1;
            // any label above numRegionLabels is considered to be the last class
            int semanticClass = std::min(state.pTrueLabels->labels[gtRegId],
                gSceneConfig.numRegionLabels - 1);
            classLabelCount[semanticClass] += 1;
        }
    }

    // get modes
    int maxRegCount = maxElem(regionLabelCount);
    int maxClass = argmax(classLabelCount);
    if ((unknownLabelCount > maxRegCount) || (unknownLabelCount > 0.05 * region->size())) {
        return;
    }

    // if max region has too few pixels then the move should not have been made
    if (bAccepted && (maxRegCount < 0.75 * region->size())) {
        SVL_LOG(SVL_LOG_VERBOSE, "adding poor accept to active learning dataset ("
            << state.pClosedLoopLearningDataset->size() << ")");

        vector<double> x;
        SceneRegionFeatures::regionFeatures(&instance, regId, x);

        state.pClosedLoopLearningDataset->features.push_back(x);
        state.pClosedLoopLearningDataset->labels.push_back(-1);
        state.pClosedLoopLearningDataset->weights.push_back(region->size());
        return;
    }

#if 0
    // if region is okay then move should have been accepted
    if (!bAccepted && (maxRegCount > 0.95 * region->size())) {
        if (rand() % 3 != 0) return;
        SVL_LOG(SVL_LOG_VERBOSE, "adding poor reject to active learning dataset ("
            << state.pClosedLoopLearningDataset->size() << ")");

        vector<double> x;
        SceneRegionFeatures::regionFeatures(&instance, regId, x);

        state.pClosedLoopLearningDataset->features.push_back(x);
        state.pClosedLoopLearningDataset->labels.push_back(maxClass);
        state.pClosedLoopLearningDataset->weights.push_back(region->size());
        return;
    }
#endif

    // if classes don't match and move was accepted then add to training set
    if (bAccepted && (maxClass != region->Sk)) {
        SVL_LOG(SVL_LOG_VERBOSE, "adding incorrect labels to active learning dataset ("
            << state.pClosedLoopLearningDataset->size() << ")");

        vector<double> x;
        SceneRegionFeatures::regionFeatures(&instance, regId, x);

        state.pClosedLoopLearningDataset->features.push_back(x);
        state.pClosedLoopLearningDataset->labels.push_back(maxClass);
        state.pClosedLoopLearningDataset->weights.push_back(region->size());
        return;
    }
}


// closed-loop training for object model
void recordLearningExampleForObjectModel(InferenceState& state, int objId, bool bAccepted,
    SceneInstance& instance, SceneRegionModel* model)
{
    if ((state.pClosedLoopLearningDatasetObject == NULL) || (state.pTrueLabels == NULL))
        return;

    SceneObject *object = instance.assignment.objects[objId];
    // don't add if object is too small
    //    if (region->size() < 1.0e-3 * instance.size())
    //     return;

    // find groundtruth labels for proposal move
    //vector<int> regionLabelCount(state.pTrueLabels->size(), 0);
    //vector<int> classLabelCount(gSceneConfig.numObjectLabels + 1, 0);

    int unknownLabelCount = 0;

    Vector4i objExtent = object->extent();
    CvRect objExtentRect = cvRect(objExtent[0], objExtent[1], objExtent[2], objExtent[3]);
    double bestOverlapScore = -1;
    int bestOverlapIndx = -1;
    for (int i = 0; i < (int)state.pTrueLabels->labels.size(); i++) {
        //if (state.pTrueLabels->labels[i] < gSceneConfig.numRegionLabels) {
        //continue;
        //}

        if (state.pTrueLabels->labels[i] != gSceneConfig.numRegionLabels + 1) {
            continue;
        }
        Vector4i layerExtent = state.pTrueLabels->layerExtents[i];
        CvRect layerExtentRect = cvRect(layerExtent[0], layerExtent[1], layerExtent[2], layerExtent[3]);
        CvRect interExtentRect = svlIntersection(objExtentRect, layerExtentRect);
        double overlapScore = (double)(interExtentRect.width * interExtentRect.height)
            / (double)(objExtentRect.width * objExtentRect.height + layerExtentRect.width * layerExtentRect.height - interExtentRect.width*interExtentRect.height);
        overlapScore = overlapScore > 0 ? overlapScore : 0;
        if (overlapScore > bestOverlapScore) {
            bestOverlapScore = overlapScore;
            bestOverlapIndx = i;
        }
    }

    int trueClass;
    if (bestOverlapIndx == -1) {
        trueClass = 0; // non-object (non-pedestrian)
    } else if (bestOverlapScore > 0.5) {
        trueClass = state.pTrueLabels->labels[bestOverlapIndx] - gSceneConfig.numRegionLabels + 1;
    } else {
        trueClass = 0;
    }
    //    cout << "bestOverlapIndx: " << bestOverlapIndx << "; bestOverlapScore: " << bestOverlapScore
    //         << "; true class: " << trueClass
    //         << "; objectClassifierScore: " << object->objDetScore[object->objDetScore.rows()-1] << endl;

    // if classes don't match and move was accepted then add to training set
#if 0
    if (trueClass == object->Ok + 1) {
        IplImage* maskImage = cvCloneImage(instance.image());
        Vector4i extent = object->extent();
        cout << "obj extent: " << extent << endl;
        cout << "obj singleton: " << object->prObject[0] << ", "
             << object->prObject[1] << ", " << object->prObject[2] << endl;
        cout << "obj detScore: " << object->objDetScore[0] << ", "
             << object->objDetScore[1] << endl;

        cvRectangle(maskImage, cvPoint(extent[0], extent[1]),
            cvPoint(extent[0] + extent[2], extent[1] + extent[3]),
            CV_RGB(255, 0 ,0), 3);

        if (bestOverlapIndx != -1) {
            Vector4i gtObjExtent = state.pTrueLabels->layerExtents[bestOverlapIndx];
            cvRectangle(maskImage, cvPoint(gtObjExtent[0], gtObjExtent[1]),
                cvPoint(gtObjExtent[0] + gtObjExtent[2], gtObjExtent[1] + gtObjExtent[3]),
                CV_RGB(0, 255 ,0), 3);
        }

        list<CvPoint> objectPixels = object->pixels();
        for (list<CvPoint>::const_iterator p = objectPixels.begin();
             p != objectPixels.end(); ++p) {
            cvRectangle(maskImage, *p, *p, CV_RGB(255, 255, 255), 5);
        }

        cvNamedWindow("debugCorrectObj", 1);
        cvShowImage("debugCorrectObj", maskImage);
        cvWaitKey(-1);
        cvDestroyWindow("debugCorrectObj");
        cvReleaseImage(&maskImage);
    }
#endif
    // record high-score errors for pedestrian classifier
    SVL_ASSERT(gSceneConfig.getOptionAsBool("useIndependentPedestrianClassifierAtObjectLevel"));
    if (trueClass == 0 && bestOverlapScore < 0.2 && object->objDetScore[object->objDetScore.rows()-1] > -2) {
        SVL_LOG(SVL_LOG_VERBOSE, "adding incorrect labels to active learning dataset "
            << "predicted class: " << object->Ok+1 << ", true class: " << trueClass << " ("
            << state.pClosedLoopLearningDatasetObject->size() << ")");

        // vector<double> x = ((SceneObjectModel*)model)->computeObjectFeatureVector(&instance, objId);

        // record for pedestrian classifier
        vector<double> x;
        if (object->bComputeObjDetScores) {
            x = ((SceneObjectModel*)model)->computePatchBasedPedestrianFeature(&instance, objId, NULL);
        } else {
            double pedScore = object->objDetScore[1];
            x = ((SceneObjectModel*)model)->computePatchBasedPedestrianFeature(&instance, objId, &pedScore);
        }
        state.pClosedLoopLearningDatasetObject->features.push_back(x);
        state.pClosedLoopLearningDatasetObject->labels.push_back(trueClass);
        state.pClosedLoopLearningDatasetObject->weights.push_back(1.0);

#if 0
        IplImage* maskImage = cvCloneImage(instance.image());
        Vector4i extent = object->extent();
        cout << "obj extent: " << extent << endl;
        cout << "obj singleton: " << object->prObject[0] << ", "
             << object->prObject[1] << ", " << object->prObject[2] << endl;
        cout << "obj detScore: " << object->objDetScore[0] << ", "
             << object->objDetScore[1] << endl;

        cvRectangle(maskImage, cvPoint(extent[0], extent[1]),
        cvPoint(extent[0] + extent[2], extent[1] + extent[3]),
            CV_RGB(255, 0 ,0), 3);

        if (bestOverlapIndx != -1) {
            Vector4i gtObjExtent = state.pTrueLabels->layerExtents[bestOverlapIndx];
            cvRectangle(maskImage, cvPoint(gtObjExtent[0], gtObjExtent[1]),
                cvPoint(gtObjExtent[0] + gtObjExtent[2], gtObjExtent[1] + gtObjExtent[3]),
                CV_RGB(0, 255 ,0), 3);
        }

        list<CvPoint> objectPixels = object->pixels();
        for (list<CvPoint>::const_iterator p = objectPixels.begin();
             p != objectPixels.end(); ++p) {
            cvRectangle(maskImage, *p, *p, CV_RGB(255, 255, 255), 5);
        }

        cvNamedWindow("debug", 1);
        cvShowImage("debug", maskImage);
        cvWaitKey(-1);
        cvDestroyWindow("debug");
        cvReleaseImage(&maskImage);
#endif
        return;
    }


    /*
    for (set<SceneRegion *>::iterator itrRegion = object->regions.begin();
         itrRegion != object->regions.end(); ++itrRegion) {
        if ((*itrRegion)->empty()) continue;
        for (list<CvPoint>::const_iterator p = (*itrRegion)->pixels.begin(); p != (*itrRegion)->pixels.end(); p++) {
            int gtRegId = CV_MAT_ELEM(*state.pTrueLabels->layers, int, p->y, p->x);
            SVL_ASSERT((gtRegId >= 0) && (gtRegId < (int)state.pTrueLabels->labels.size()));
            if (state.pTrueLabels->labels[gtRegId] < 0) {
                unknownLabelCount += 1;
            } else {
                regionLabelCount[gtRegId] += 1;
                // any label above numRegionLabels is considered to be the last class
                int semanticClass = std::max(state.pTrueLabels->labels[gtRegId] - (gSceneConfig.numRegionLabels - 1), 0);
                SVL_ASSERT_MSG(semanticClass <= gSceneConfig.numObjectLabels, semanticClass);
                classLabelCount[semanticClass] += 1;
            }
        }
    }

    // get modes
    int maxRegId = argmax(regionLabelCount);
    int maxRegCount = regionLabelCount[maxRegId];
    int maxRegSize = state.pTrueLabels->layerSizes[maxRegId];
    int maxClass = argmax(classLabelCount);
    int unionSize = maxRegSize + object->size() - maxRegCount;
    double overlapRaito = (double)maxRegCount/(double)unionSize;
    if ((unknownLabelCount > 0.05 * object->size())) {
        return;
    }

    // if classes don't match and move was accepted then add to training set
    //    if (bAccepted && (maxClass != object->Ok)) {
    if (maxClass != object->Ok + 1) {
        SVL_LOG(SVL_LOG_VERBOSE, "adding incorrect labels to active learning dataset "
            << "predicted class: " << object->Ok << ", true class: " << maxClass << " ("
            << state.pClosedLoopLearningDatasetObject->size() << ")");

        vector<double> x = ((SceneObjectModel*)model)->computeObjectFeatureVector(&instance, objId);
        state.pClosedLoopLearningDatasetObject->features.push_back(x);
        if (maxClass == 0) {
            state.pClosedLoopLearningDatasetObject->labels.push_back(maxClass);
        } else {
            if (overlapRaito > 0.85)
                state.pClosedLoopLearningDatasetObject->labels.push_back(maxClass);
            else // how about ranking?
                state.pClosedLoopLearningDatasetObject->labels.push_back(0);
        }
        state.pClosedLoopLearningDatasetObject->weights.push_back(1.0);

#if 0
        IplImage* maskImage = cvCloneImage(instance.image());
        Vector4i extent = object->extent();
        cout << "obj extent: " << extent << endl;
        cout << "obj singleton: " << object->prObject[0] << ", "
             << object->prObject[1] << ", " << object->prObject[2] << endl;
        cout << "obj detScore: " << object->objDetScore[0] << ", "
             << object->objDetScore[1] << endl;

        cvRectangle(maskImage, cvPoint(extent[0], extent[1]),
        cvPoint(extent[0] + extent[2], extent[1] + extent[3]),
        CV_RGB(0,255,0), 3);

        list<CvPoint> objectPixels = object->pixels();
        for (list<CvPoint>::const_iterator p = objectPixels.begin();
             p != objectPixels.end(); ++p) {
            cvRectangle(maskImage, *p, *p, CV_RGB(255, 255, 255), 5);
        }

        cvNamedWindow("debug", 1);
        cvShowImage("debug", maskImage);
        cvWaitKey(-1);
        cvDestroyWindow("debug");
        cvReleaseImage(&maskImage);
#endif

        return;
    }

    // if not enough overlap
    if (maxClass != 0 && overlapRaito < 0.5) {
        SVL_LOG(SVL_LOG_VERBOSE, "adding incorrect label without enough overlap to active learning dataset ("
            << state.pClosedLoopLearningDatasetObject->size() << ")");
        vector<double> x = ((SceneObjectModel*)model)->computeObjectFeatureVector(&instance, objId);
        state.pClosedLoopLearningDatasetObject->features.push_back(x);
        state.pClosedLoopLearningDatasetObject->labels.push_back(0); // nonobject
        state.pClosedLoopLearningDatasetObject->weights.push_back(1.0);

#if 0
        IplImage* maskImage = cvCloneImage(instance.image());
        Vector4i extent = object->extent();
        cout << "obj extent: " << extent << endl;
        cout << "obj singleton: " << object->prObject[0] << ", "
             << object->prObject[1] << ", " << object->prObject[2] << endl;
        cout << "obj detScore: " << object->objDetScore[0] << ", "
             << object->objDetScore[1] << endl;

        cvRectangle(maskImage, cvPoint(extent[0], extent[1]),
            cvPoint(extent[0] + extent[2], extent[1] + extent[3]),
            CV_RGB(0,255,0), 3);

        list<CvPoint> objectPixels = object->pixels();
        for (list<CvPoint>::const_iterator p = objectPixels.begin();
             p != objectPixels.end(); ++p) {
            cvRectangle(maskImage, *p, *p, CV_RGB(255, 255, 255), 5);
        }

        cvNamedWindow("debug", 1);
        cvShowImage("debug", maskImage);
        cvWaitKey(-1);
        cvDestroyWindow("debug");
        cvReleaseImage(&maskImage);
#endif
        return;
    }

#if 0
    // if region is okay then move should have been accepted
    if (!bAccepted && (maxClass == object->Ok && maxClass != 0) && overlapRaito > 0.9) {
        SVL_LOG(SVL_LOG_VERBOSE, "adding poor reject object to active learning dataset ("
            << state.pClosedLoopLearningDatasetObject->size() << ")");

        vector<double> x = ((SceneObjectModel*)model)->computeObjectFeatureVector(&instance, objId);
        state.pClosedLoopLearningDatasetObject->features.push_back(x);
        state.pClosedLoopLearningDatasetObject->labels.push_back(maxClass);
        state.pClosedLoopLearningDatasetObject->weights.push_back(1.0);
        return;
    }
#endif
    */
}
