/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneInstance.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"

using namespace std;

// SceneInstance class ----------------------------------------------------

SceneInstance::SceneInstance(const char *imgFilename, const char *baseName) :
    _labImg(NULL), _grayImg(NULL), _softEdgeImg(NULL), _meanContrast(0.0)
{
    _img = cvLoadImage(imgFilename, CV_LOAD_IMAGE_COLOR);
    SVL_ASSERT_MSG(_img != NULL, imgFilename);
    if (baseName != NULL) {
        _baseName = string(baseName);
    } else {
        _baseName = strBaseName(imgFilename);
    }
    initScene();
}

SceneInstance::SceneInstance(const IplImage *img, const char *baseName) :
    _labImg(NULL), _grayImg(NULL), _softEdgeImg(NULL), _meanContrast(0.0)
{
    SVL_ASSERT(img != NULL);
    _img = cvCloneImage(img);
    if (baseName != NULL)
        _baseName = string(baseName);
    initScene();
}

SceneInstance::SceneInstance(const SceneInstance& scene)
{
    // scene instance should never be copied
    SVL_ASSERT(false);
}

SceneInstance::~SceneInstance()
{
    freeScene();
}

void SceneInstance::resize(int w, int h)
{
    // resize singletons
    vector<CvMat *> singletons;
    if (!pixelSingletons.empty()) {
        for (int i = 0; i < gSceneConfig.numRegionLabels; i++) {
            CvMat *s = cvCreateMat(height(), width(), CV_32FC1);
            float *p = (float *)CV_MAT_ELEM_PTR(*s, 0, 0);
            for (int indx = 0; indx < size(); indx++, p++) {
                *p = (float)pixelSingletons[indx][i];
            }
            resizeInPlace(&s, h, w);
            singletons.push_back(s);
        }        
    }

    // resize image and segmentations
    resizeInPlace(&_img, h, w);

    if (_labImg != NULL) {
        cvReleaseImage(&_labImg);
    }
    if (_grayImg != NULL) {
        cvReleaseImage(&_grayImg);
    }
    if (_softEdgeImg != NULL) {
        cvReleaseImage(&_softEdgeImg);
    }

    svlMultiSeg resizedSegmentations(w, h);
    CvMat *seg = cvCreateMat(h, w, CV_32SC1);
    for (int i = 0; i < segmentations.count(); i++) {
        cvResize(segmentations.segmentation(i), seg, CV_INTER_NN);
        resizedSegmentations.addSegmentation(seg);
    }

    if (assignment.pixels != NULL) {
        cvResize(assignment.pixels, seg, CV_INTER_NN);
    }
    int resizedHorizon = assignment.horizon * h / assignment.height();

    // recompute features and assignments
    assignment.clear();
    _filters.clear();
    _dirFilters.clear();

    initScene();
    segmentations = resizedSegmentations;

    // recompute pixel singletons
    pixelSingletons.clear();
    pixelSingletons.resize(size(), vector<double>(gSceneConfig.numRegionLabels, 0.0));
    for (int i = 0; i < gSceneConfig.numRegionLabels; i++) {
        float *p = (float *)CV_MAT_ELEM_PTR(*singletons[i], 0, 0);
        for (int indx = 0; indx < size(); indx++, p++) {
            pixelSingletons[indx][i] = *p;
        }
    }
    svlReleaseMatrices(singletons);

    // reassign variables
    assignPixels(seg);
    assignment.remap();
    assignment.horizon = resizedHorizon;

    // free temporary storage
    cvReleaseMat(&seg);
}

// i/o
void SceneInstance::loadSegmentations(bool bClearExisting)
{
    SVL_ASSERT(_img != NULL);

    if (bClearExisting) {
        segmentations.clear();
   }

    SVL_LOG(SVL_LOG_VERBOSE, "loading multiple segmentations");
    string fileStem = gSceneConfig.filebase("segDir", _baseName);
    vector<string> fileExts;
    parseString(gSceneConfig.getOption("multiSegExt"), fileExts);

    // initialize seg buffer
    CvMat *seg = cvCreateMat(height(), width(), CV_32SC1);

    for (int i = 0; i < (int)fileExts.size(); i++) {
        svlCodeProfiler::tic(svlCodeProfiler::getHandle("loadSegmentation"));
        // TODO: convert to binary
        readMatrix(seg, (fileStem + fileExts[i]).c_str());

        // TODO: this should be done externally
        svlConnectedComponents(seg, 8);

        svlImageSegmentation imgSeg(seg);
        //        imgSeg.removeSmallSegments(20);

        segmentations.addSegmentation(imgSeg);
        SVL_LOG(SVL_LOG_VERBOSE, "...segmentation " << segmentations.count() << " has "
            << segmentations.numSegments(segmentations.count() - 1) << " segments");
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("loadSegmentation"));
    }

    // free memory
    cvReleaseMat(&seg);
}

void SceneInstance::loadGroundTruth()
{
    // TODO: fix to new labelings

    // TODO: modify to allow loading of regions and objects
    SVL_ASSERT(_img != NULL);

    SVL_LOG(SVL_LOG_VERBOSE, "loading groundtruth assignments");

    // load region labels
    string lblFilename = gSceneConfig.filename("lblDir", _baseName, "regionLblExt");
    CvMat *regionLabels = cvCreateMat(height(), width(), CV_32SC1);
    bool bSuccess = readMatrix(regionLabels, lblFilename.c_str());
    if (!bSuccess) {
        SVL_LOG(SVL_LOG_ERROR, "Could not read region labels for " << _baseName);
        cvSet(regionLabels, cvScalar(-1));
    }

#if 1
    // load region assignments
    lblFilename = gSceneConfig.filename("lblDir", _baseName, "layerLblExt");
    CvMat *layerLabels = cvCreateMat(height(), width(), CV_32SC1);
    bSuccess = readMatrix(layerLabels, lblFilename.c_str());
    if (!bSuccess) {
        SVL_LOG(SVL_LOG_ERROR, "Could not read layers for " << _baseName << " so using regions");
        cvCopy(regionLabels, layerLabels);
    }
#else
    SVL_LOG(SVL_LOG_WARNING, "using region labels for layers");
    CvMat *layerLabels = cvCloneMat(regionLabels);
#endif

    // assign singletons and regions
    pixelSingletons.resize(size());
    int varIndx = 0;
    for (int y = 0; y < height(); y++) {
        for (int x = 0; x < width(); x++, varIndx++) {
            int regLabel = CV_MAT_ELEM(*regionLabels, int, y, x);

            pixelSingletons[varIndx].resize(gSceneConfig.numRegionLabels);
            fill(pixelSingletons[varIndx].begin(), pixelSingletons[varIndx].end(), 0.0);

            if ((regLabel < 0) || (regLabel >= gSceneConfig.numRegionLabels)) {
                SVL_LOG(SVL_LOG_DEBUG, "invalid label " << regLabel
                    << " for pixel (" << x << ", " << y << ")");
                continue;
            }

            pixelSingletons[varIndx][regLabel] = 1.0;
        }
    }

    assignPixelsFromLabels(layerLabels);

    // make sure "unknown" regions remain unknown
    for (int k = 0; k < numRegions(); k++) {
        if (assignment.regions[k]->pixelScores[assignment.regions[k]->Sk] <
            0.5 * assignment.regions[k]->size()) {
            assignment.regions[k]->Sk = -1;
        }
    }

    // load horizon
    if (gSceneConfig.getOptionAsBool("useGeometry")) {
        map<string, int> h = SceneDataLoader::loadHorizonTrainingData();
        if (h.find(_baseName) != h.end()) {
            assignment.horizon = h[_baseName];
        } else {
            SVL_LOG(SVL_LOG_ERROR, "could not find groundtruth horizon for " << _baseName);
        }
    }

    // free memory
    cvReleaseMat(&layerLabels);
    cvReleaseMat(&regionLabels);
}

void SceneInstance::loadAssignments(const char *directory)
{
    // TODO: fix to new labelings

    SVL_ASSERT(_img != NULL);

    SVL_LOG(SVL_LOG_VERBOSE, "loading assignments from file");

    // load region labels
    string lblFilename = string(directory) + _baseName + string(".predictedClass.txt");
    CvMat *regionLabels = cvCreateMat(height(), width(), CV_32SC1);
    bool bSuccess = readMatrix(regionLabels, lblFilename.c_str());
    SVL_ASSERT_MSG(bSuccess, "Could not read region labels for " << _baseName);

    // load region assignments
    lblFilename = string(directory) + _baseName + string(".predictedRegions.txt");
    CvMat *layerLabels = cvCreateMat(height(), width(), CV_32SC1);
    bSuccess = readMatrix(layerLabels, lblFilename.c_str());
    SVL_ASSERT_MSG(bSuccess, "Could not read layers for " << _baseName);

    // load horizon
    if (gSceneConfig.getOptionAsBool("useGeometry")) {
        lblFilename = string(directory) + _baseName + string(".predictedHorizon.txt");
        ifstream ifs(lblFilename.c_str());
        if (ifs.fail()) {
            SVL_LOG(SVL_LOG_ERROR, "Could not read horizon for " << _baseName);
        } else {
            ifs >> assignment.horizon;
        }
        ifs.close();
    }

    // assign singletons and regions
    pixelSingletons.resize(size());
    int varIndx = 0;
    for (int y = 0; y < height(); y++) {
        for (int x = 0; x < width(); x++, varIndx++) {
            int regLabel = CV_MAT_ELEM(*regionLabels, int, y, x);

            pixelSingletons[varIndx].resize(gSceneConfig.numRegionLabels);
            fill(pixelSingletons[varIndx].begin(), pixelSingletons[varIndx].end(), 0.0);

            if ((regLabel < 0) || (regLabel >= gSceneConfig.numRegionLabels)) {
                SVL_LOG(SVL_LOG_WARNING, "invalid label " << regLabel
                    << " for pixel (" << x << ", " << y << ")");
                continue;
            }

            pixelSingletons[varIndx][regLabel] = 1.0;
        }
    }

    assignPixelsFromLabels(layerLabels);

    // free memory
    cvReleaseMat(&layerLabels);
    cvReleaseMat(&regionLabels);
}

void SceneInstance::loadRegions(const char *filename)
{
    SVL_ASSERT(filename != NULL);

    CvMat *r = cvCreateMat(height(), width(), CV_32SC1);
    bool success = readMatrix(r, filename);
    SVL_ASSERT(success);

    assignPixels(r);

    cvReleaseMat(&r);
}

// load pixel assignments as set pixelSingletons accordingly
void SceneInstance::loadPixelMAPSingletons(const char *filename)
{
    SVL_ASSERT(filename != NULL);

    pixelSingletons.clear();
    pixelSingletons.resize(size(), vector<double>(gSceneConfig.numRegionLabels, 0.0));

    CvMat *mapAssignment = cvCreateMat(height(), width(), CV_32SC1);
    bool success = readMatrix(mapAssignment, filename);
    SVL_ASSERT(success);

    int *p = (int *)CV_MAT_ELEM_PTR(*mapAssignment, 0, 0);
    for (int indx = 0; indx < size(); indx++, p++) {
        if (*p < 0) continue;
        SVL_ASSERT(*p < gSceneConfig.numRegionLabels);
        pixelSingletons[indx][*p] = 1.0;
    }

    cvReleaseMat(&mapAssignment);
}

// variable updates/access
CvMat *SceneInstance::getPixelClasses() const
{
    SVL_ASSERT(assignment.pixels != NULL);

    CvMat *m = cvCreateMat(height(), width(), CV_32SC1);
    int *p = (int *)CV_MAT_ELEM_PTR(*assignment.pixels, 0, 0);
    int *q = (int *)CV_MAT_ELEM_PTR(*m, 0, 0);
    for (int y = 0; y < height(); y++) {
        for (int x = 0; x < width(); x++, p++, q++) {
            SVL_ASSERT((*p >= 0) && (*p < numRegions()));
            *q = assignment.regions[*p]->Sk;
        }
    }

    return m;
}

// full assignment update. calculates new sufficient statistics
void SceneInstance::replaceAssignment(const SceneAssignment &a, bool bUpdateStatistics)
{
    assignment = a;

    if (bUpdateStatistics) {
        for (vector<SceneRegion *>::iterator it = assignment.regions.begin();
             it != assignment.regions.end(); it++) {
            (*it)->clear();
            //(*it)->pixels.clear();
            (*it)->Ak.clear();
            (*it)->xAk.clear();
            (*it)->yAk.clear();
            (*it)->centroidSum.setZero();
            (*it)->pixelScores.setZero();
        }

        bool bIncAppMoments = gSceneConfig.getOptionAsBool("includeMomentFeatures");

        int pixIndx = 0;
        int *regId = (int *)CV_MAT_ELEM_PTR(*assignment.pixels, 0, 0);
        for (int y = 0; y < height(); y++) {
            for (int x = 0; x < width(); x++, pixIndx++, regId++) {
                SVL_ASSERT(*regId >= 0);

                //assignment.regions[*regId]->pixels.push_back(cvPoint(x, y));

                // update sufficient statistics
                addRegionStatistics(*regId, pixIndx, x, y, bIncAppMoments);
            }
        }
    }
}


void SceneInstance::resetRegions()
{
    // assign each pixel to its own region
    CvMat *m = cvCreateMat(height(), width(), CV_32SC1);
    int varIndx = 0;
    int *q = (int *)CV_MAT_ELEM_PTR(*m, 0, 0);
    for (int y = 0; y < height(); y++) {
        for (int x = 0; x < width(); x++, varIndx++, q++) {
            *q = varIndx;
        }
    }

    assignPixels(m);
    cvReleaseMat(&m);
}

void SceneInstance::assignPixelsFromLabels(CvMat *regionLabels)
{
    SVL_FCN_TIC;

    svlConnectedComponents(regionLabels, 8);
    assignPixels(regionLabels);
    assignment.remap();

    SVL_FCN_TOC;
}

void SceneInstance::assignPixels(const CvMat *r)
{
    SVL_FCN_TIC;
    SVL_ASSERT((r != NULL) && (r->rows == height()) && (r->cols == width()));
    int nPixelLabels = pixelSingletons.empty() ? 0 : gSceneConfig.numRegionLabels;
    int nAppFeatures = pixelAppearance.empty() ? numFilters() : numAppFeatures();
    bool bIncAppMoments = gSceneConfig.getOptionAsBool("includeMomentFeatures");

    // clear meta-data and random variables
    assignment.clear();
    cvCopy(r, assignment.pixels);

    int pixIndx = 0;
    int *regId = (int *)CV_MAT_ELEM_PTR(*assignment.pixels, 0, 0);
    for (int y = 0; y < height(); y++) {
        for (int x = 0; x < width(); x++, pixIndx++, regId++) {
            if (*regId < 0) {
                SVL_LOG(SVL_LOG_ERROR, "pixel (" << x << ", " << y << ") has region id " << *regId);
                continue;
            }

            while (*regId >= numRegions()) {
                assignment.regions.push_back(new SceneRegion(numRegions()));
                assignment.regions.back()->pixelScores = VectorXd::Zero(nPixelLabels);
                assignment.regions.back()->Ak.clear(nAppFeatures, SVL_PSS_DIAG);
                if (bIncAppMoments) {
                    assignment.regions.back()->xAk.clear(nAppFeatures, SVL_PSS_NONE);
                    assignment.regions.back()->yAk.clear(nAppFeatures, SVL_PSS_NONE);
                }
            }

            assignment.regions[*regId]->pixels.push_back(cvPoint(x, y));

            // update sufficient statistics
            addRegionStatistics(*regId, pixIndx, x, y, bIncAppMoments);
        }
    }

    // inititalize remaining random variables and cached data
    for (int regId = 0; regId < numRegions(); regId++) {
        assignment.regions[regId]->Sk = argmax(assignment.regions[regId]->pixelScores);
    }

    // update region and object neighbors
    assignment.updateNeighborhoods();

    SVL_FCN_TOC;
}

int SceneInstance::movePixels(const CvMat *r)
{
    SVL_FCN_TIC;
    SVL_ASSERT(r != NULL);
    SVL_ASSERT_MSG((r->rows == height()) && (r->cols == width()),
        r->rows << " != " << height() << " or " << r->cols << " != " << width());

    int nPixelLabels = pixelSingletons.empty() ? 0 : gSceneConfig.numRegionLabels;
    int nAppFeatures = pixelAppearance.empty() ? numFilters() : numAppFeatures();
    bool bIncAppMoments = gSceneConfig.getOptionAsBool("includeMomentFeatures");

    set<int> modifiedRegions;
    int numMoved = 0;

    int numExistingRegions = numRegions();
    int pixIndx = 0;
    int *p = (int *)CV_MAT_ELEM_PTR(*r, 0, 0);
    int *q = (int *)CV_MAT_ELEM_PTR(*assignment.pixels, 0, 0);
    for (int y = 0; y < height(); y++) {
        for (int x = 0; x < width(); x++, pixIndx++, p++, q++) {
            int regionId = *p;
            int oldRegId = *q;

            if (regionId == oldRegId)
                continue;
            SVL_ASSERT(*p < numRegions());

            // new region?
            if (regionId < 0) {
                if (numRegions() == numExistingRegions) {
                    assignment.regions.push_back(new SceneRegion(numRegions()));
                    assignment.regions.back()->pixelScores = VectorXd::Zero(nPixelLabels);
                    assignment.regions.back()->Ak.clear(nAppFeatures, SVL_PSS_DIAG);
                    if (bIncAppMoments) {
                        assignment.regions.back()->xAk.clear(nAppFeatures, SVL_PSS_NONE);
                        assignment.regions.back()->yAk.clear(nAppFeatures, SVL_PSS_NONE);
                    }
                }
                regionId = numExistingRegions;
            }

            *q = regionId;
            assignment.regions[regionId]->pixels.push_back(cvPoint(x, y));

            // maintain sufficient statistics
            subRegionStatistics(oldRegId, pixIndx, x, y, bIncAppMoments);
            addRegionStatistics(regionId, pixIndx, x, y, bIncAppMoments);

            // remember modified regions
            modifiedRegions.insert(regionId);
            modifiedRegions.insert(oldRegId);
            numMoved += 1;
        }
    }

    if (numMoved == 0) {
        SVL_FCN_TOC;
        return 0;
    }

    // update region meta-data
    for (set<int>::const_iterator regId = modifiedRegions.begin(); regId != modifiedRegions.end(); regId++) {
        SVL_LOG(SVL_LOG_DEBUG, "inferring class label for region " << *regId);
        assignment.regions[*regId]->Sk = argmax(assignment.regions[*regId]->pixelScores);

        // delete moved pixels
        for (list<CvPoint>::iterator it = assignment.regions[*regId]->pixels.begin();
             it != assignment.regions[*regId]->pixels.end(); ) {
            if (CV_MAT_ELEM(*assignment.pixels, int, it->y, it->x) != *regId) {
                it = assignment.regions[*regId]->pixels.erase(it);
            } else {
                it++;
            }
        }

        assignment.regions[*regId]->clear();
    }

    // update neighbors
    assignment.updateNeighborhoods();

    SVL_FCN_TOC;
    return numMoved;
}

int SceneInstance::movePixels(const vector<CvPoint>& pixels, int regionId)
{
    if (pixels.empty()) return 0;

    SVL_FCN_TIC;
    SVL_ASSERT(regionId < numRegions());

#if 0
    // use other movePixels() function
    CvMat *r = cvCloneMat(assignment.pixels);
    for (vector<CvPoint>::const_iterator p = pixels.begin(); p != pixels.end(); p++) {
        CV_MAT_ELEM(*r, int, p->y, p->x) = regionId;
    }
    int n = movePixels(r);
    cvReleaseMat(&r);
    SVL_FCN_TOC;
    return n;
#endif

    int nPixelLabels = pixelSingletons.empty() ? 0 : gSceneConfig.numRegionLabels;
    int nAppFeatures = pixelAppearance.empty() ? numFilters() : numAppFeatures();
    bool bIncAppMoments = gSceneConfig.getOptionAsBool("includeMomentFeatures");

    set<int> modifiedRegions;
    int numMoved = 0;

    // create a new region
    if (regionId < 0) {
        assignment.regions.push_back(new SceneRegion(numRegions()));
        assignment.regions.back()->pixelScores = VectorXd::Zero(nPixelLabels);
        assignment.regions.back()->Ak.clear(nAppFeatures, SVL_PSS_DIAG);
        if (bIncAppMoments) {
            assignment.regions.back()->xAk.clear(nAppFeatures, SVL_PSS_NONE);
            assignment.regions.back()->yAk.clear(nAppFeatures, SVL_PSS_NONE);
        }
        regionId = (int)assignment.regions.size() - 1;
        SVL_ASSERT(assignment.regions.back()->id() == regionId);
    }

    // update region statistics
    int *q = (int *)CV_MAT_ELEM_PTR(*assignment.pixels, 0, 0);
    for (vector<CvPoint>::const_iterator p = pixels.begin(); p != pixels.end(); ++p) {
        int pixIndx = pixel2Indx(p->x, p->y);
        int oldRegId = q[pixIndx];

        if (oldRegId == regionId)
            continue;

        q[pixIndx] = regionId;
        assignment.regions[regionId]->pixels.push_back(*p);

        subRegionStatistics(oldRegId, pixIndx, p->x, p->y, bIncAppMoments);
        addRegionStatistics(regionId, pixIndx, p->x, p->y, bIncAppMoments);

        modifiedRegions.insert(oldRegId);
        numMoved += 1;
    }

    // return if nothing changed
    if (numMoved == 0) {
        SVL_FCN_TOC;
        return 0;
    }

    // update region meta-data
    SVL_LOG(SVL_LOG_DEBUG, "updating modified regions " << toString(modifiedRegions));
    for (set<int>::const_iterator it = modifiedRegions.begin(); it != modifiedRegions.end(); it++) {
        int oldRegId = *it;

        SVL_LOG(SVL_LOG_DEBUG, "inferring class label for region " << oldRegId);
        assignment.regions[oldRegId]->Sk = argmax(assignment.regions[oldRegId]->pixelScores);

        // delete moved pixels
        for (list<CvPoint>::iterator jt = assignment.regions[oldRegId]->pixels.begin();
             jt != assignment.regions[oldRegId]->pixels.end(); ) {
            if (CV_MAT_ELEM(*assignment.pixels, int, jt->y, jt->x) != oldRegId) {
                jt = assignment.regions[oldRegId]->pixels.erase(jt);
            } else {
                jt++;
            }
        }

        SVL_LOG(SVL_LOG_DEBUG, "clearing region " << oldRegId);
        assignment.regions[oldRegId]->clear();
        //validateData();
    }

    SVL_LOG(SVL_LOG_DEBUG, "inferring class label for region " << regionId);
    assignment.regions[regionId]->Sk = argmax(assignment.regions[regionId]->pixelScores);
    assignment.regions[regionId]->clear();

    // update neighbors
    assignment.updateNeighborhoods();

    SVL_FCN_TOC;
    return numMoved;
}

int SceneInstance::movePixels(const vector<CvPoint>& pixels, const vector<int>& regionId)
{
    SVL_ASSERT(pixels.size() == regionId.size());

    CvMat *m = cvCloneMat(assignment.pixels);
    for (unsigned i = 0; i < pixels.size(); i++) {
        CV_MAT_ELEM(*m, int, pixels[i].y, pixels[i].x) = regionId[i];
    }

    int numMoved = movePixels(m);
    cvReleaseMat(&m);
    return numMoved;
}

int SceneInstance::movePixels(int oldRegionId, int newRegionId)
{
    if ((oldRegionId == newRegionId) || (newRegionId < 0) ||
        (assignment.regions[oldRegionId]->empty()))
        return 0;

#if 0 // slow version
    vector<CvPoint> pixels(assignment.regions[oldRegionId]->pixels.begin(),
        assignment.regions[oldRegionId]->pixels.end());
    return movePixels(pixels, newRegionId);
#else // fast version
    SVL_FCN_TIC;

    bool bIncAppMoments = gSceneConfig.getOptionAsBool("includeMomentFeatures");
    int numMoved = assignment.regions[oldRegionId]->size();

    // udpate region correspondence variables
    int *q = (int *)CV_MAT_ELEM_PTR(*assignment.pixels, 0, 0);
    for (list<CvPoint>::const_iterator p = assignment.regions[oldRegionId]->pixels.begin();
         p != assignment.regions[oldRegionId]->pixels.end(); ++p) {
        q[pixel2Indx(p->x, p->y)] = newRegionId;
    }

    // update region meta-information
    assignment.regions[newRegionId]->pixels.insert(assignment.regions[newRegionId]->pixels.end(),
        assignment.regions[oldRegionId]->pixels.begin(),
        assignment.regions[oldRegionId]->pixels.end());
    assignment.regions[oldRegionId]->pixels.clear();

    assignment.regions[newRegionId]->pixelScores += assignment.regions[oldRegionId]->pixelScores;
    assignment.regions[oldRegionId]->pixelScores.setZero();

    assignment.regions[newRegionId]->Ak.accumulate(assignment.regions[oldRegionId]->Ak);
    assignment.regions[oldRegionId]->Ak.clear();

    if (bIncAppMoments) {
        assignment.regions[newRegionId]->xAk.accumulate(assignment.regions[oldRegionId]->xAk);
        assignment.regions[oldRegionId]->xAk.clear();
        assignment.regions[newRegionId]->yAk.accumulate(assignment.regions[oldRegionId]->yAk);
        assignment.regions[oldRegionId]->yAk.clear();
    }

    assignment.regions[newRegionId]->centroidSum += assignment.regions[oldRegionId]->centroidSum;
    assignment.regions[oldRegionId]->centroidSum.setZero();

    SVL_LOG(SVL_LOG_DEBUG, "inferring class label for region " << newRegionId);
    assignment.regions[newRegionId]->Sk = argmax(assignment.regions[newRegionId]->pixelScores);
    assignment.regions[oldRegionId]->Sk = -1;

    assignment.regions[oldRegionId]->clear();
    assignment.regions[newRegionId]->clear();

    // update neighbors
    assignment.updateNeighborhoods();

    SVL_FCN_TOC;
    return numMoved;
#endif
}

// assign/re-assign regions to objects. returns the number of pixels moves
int SceneInstance::moveRegions(int regId, int objId)
{
    SVL_ASSERT((regId >= 0) && (regId < numRegions()) && (objId < numObjects()));
    if (objId < 0) {
        objId = numObjects();
        assignment.objects.push_back(new SceneObject(numObjects()));
    }

    if (assignment.objects[objId]->hasRegion(assignment.regions[regId])) {
        return 0;
    }

    if (assignment.regions[regId]->object != NULL) {
        assignment.regions[regId]->object->regions.erase(assignment.regions[regId]);
        assignment.regions[regId]->object->clear();
    }
    assignment.objects[objId]->regions.insert(assignment.regions[regId]);
    assignment.regions[regId]->object = assignment.objects[objId];

    assignment.objects[objId]->clear();

    // update neighbors
    assignment.updateNeighborhoods();

    return assignment.regions[regId]->size();
}

int SceneInstance::moveRegions(const vector<int> regIds, int objId)
{
    SVL_ASSERT(objId < numObjects());
    if (regIds.empty()) {
        return 0;
    }

    if (objId < 0) {
        objId = numObjects();
        assignment.objects.push_back(new SceneObject(numObjects()));
    }

    int numPixelsMoved = 0;
    for (vector<int>::const_iterator r = regIds.begin(); r != regIds.end(); r++) {
        SVL_ASSERT((*r >= 0) && (*r < numRegions()));

        if (assignment.objects[objId]->hasRegion(assignment.regions[*r])) {
            continue;
        }

        if (assignment.regions[*r]->object != NULL) {
            assignment.regions[*r]->object->regions.erase(assignment.regions[*r]);
            assignment.regions[*r]->object->clear();
        }
        assignment.objects[objId]->regions.insert(assignment.regions[*r]);
        assignment.regions[*r]->object = assignment.objects[objId];

        assignment.objects[objId]->clear();
        numPixelsMoved += assignment.regions[*r]->size();
    }

    // update neighbors
    assignment.updateNeighborhoods();
    return numPixelsMoved;
}

int SceneInstance::removeRegions(int regId, int objId)
{
    SVL_ASSERT((regId >= 0) && (regId < numRegions()) && (objId >= 0) && (objId < numObjects()));
    SVL_ASSERT(assignment.regions[regId]->object == assignment.objects[objId]);

    assignment.regions[regId]->object = NULL;
    assignment.objects[objId]->regions.erase(assignment.regions[regId]);
    assignment.objects[objId]->clear();

    // update neighbors
    assignment.updateNeighborhoods();

    return assignment.regions[regId]->size();
}

int SceneInstance::removeObject(int objId)
{
    SVL_ASSERT((objId >= 0) && (objId < numObjects()));

    int numPixelsMoved = assignment.objects[objId]->size();
    for (set<SceneRegion *>::iterator it = assignment.objects[objId]->regions.begin();
         it != assignment.objects[objId]->regions.end(); it++) {
        (*it)->object = NULL;
    }
    delete assignment.objects[objId];
    assignment.objects[objId] = NULL;

    // remap object ids and update neighbors
    assignment.remap();
    assignment.updateNeighborhoods();

    return numPixelsMoved;
}

void SceneInstance::removeSmallRegions(int minSize)
{
    for (int i = 0; i < numRegions(); i++) {
        if ((assignment.regions[i]->size() < minSize) && (assignment.regions[i]->size() > 0)) {
            // look for largest neighbor
            // TODO: look for most compatible neighbor
            set<SceneRegionBoundary *>::const_iterator it = assignment.regions[i]->boundaries.begin();
            SceneRegion *mergeRegion = (*it)->neighborOf(assignment.regions[i]);
            while (++it != assignment.regions[i]->boundaries.end()) {
                if ((*it)->neighborOf(assignment.regions[i])->size() >
                    mergeRegion->size()) {
                    mergeRegion = (*it)->neighborOf(assignment.regions[i]);
                }
            }

            // merge the regions
            vector<CvPoint> pixels(assignment.regions[i]->pixels.begin(),
                assignment.regions[i]->pixels.end());
            movePixels(pixels, mergeRegion->id());
        }
    }

    assignment.remap();
}

// split disconnected components into separate regions
void SceneInstance::splitOrphans()
{
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("splitOrphans"));
    // TODO: what should happen to objects?
    CvMat *components = cvCloneMat(assignment.pixels);
    int numComponents = svlConnectedComponents(components, 8);

    map<int, SceneObject *> newRegionObject;

    if (numComponents > numRegions()) {
        SVL_LOG(SVL_LOG_WARNING, "splitting disconnected components " <<
            numComponents << " vs " << numRegions());

        vector<vector<CvPoint> > newRegions;
        vector<int> componentMapping(numComponents, -1);
        vector<int> regionMapping(numComponents, -1);

        for (int y = 0; y < height(); y++) {
            for (int x = 0; x < width(); x++) {
                int compId = CV_MAT_ELEM(*components, int, y, x);
                if (componentMapping[compId] == -2) {
                    continue;
                }

                if (componentMapping[compId] == -1) {
                    int oldRegId = CV_MAT_ELEM(*assignment.pixels, int, y, x);

                    SVL_ASSERT_MSG(oldRegId < numComponents, oldRegId << " !< " << numComponents);

                    if (regionMapping[oldRegId] == -1) {
                        regionMapping[oldRegId] = compId;
                        componentMapping[compId] = -2;
                        continue;
                    } else if (regionMapping[oldRegId] != compId) {
                        componentMapping[compId] = (int)newRegions.size();
                        newRegions.resize(newRegions.size() + 1);
                        if (assignment.regions[oldRegId]->object != NULL) {
                            newRegionObject.insert(make_pair(newRegions.size() - 1, assignment.regions[oldRegId]->object));
                        }
                    }
                }
                newRegions[componentMapping[compId]].push_back(cvPoint(x, y));
            }
        }

        // move pixels
        for (int i = 0; i < (int)newRegions.size(); i++) {
            movePixels(newRegions[i], -1);

            // create split-objects
            map<int, SceneObject *>::iterator regionObjectItr = newRegionObject.find(i);
            if (regionObjectItr != newRegionObject.end()) {
                SVL_LOG(SVL_LOG_MESSAGE, "region " << assignment.regions.size() - 1
                    << " (" << assignment.regions.back()->size() << ") is split from object "
                    << regionObjectItr->second->id());
                regionObjectItr->second->regions.insert(assignment.regions.back());
                assignment.regions.back()->object = regionObjectItr->second;
            }
        }
    }

    cvReleaseMat(&components);

    assignment.remap();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("splitOrphans"));
}

// remove any objects that do not form connected components
int SceneInstance::removeSplitObjects()
{
    if (numObjects() == 0) {
        return 0;
    }

    int numPixelsMoved = 0;
    CvMat *mask = cvCloneMat(assignment.pixels);
    int *pMask = (int *)CV_MAT_ELEM_PTR(*mask, 0, 0);

    // iterate through objects
    for (int objId = 0; objId < numObjects(); objId++) {

        // check for split object
        cvZero(mask);
        list<CvPoint> objectPixels = assignment.objects[objId]->pixels();
        for (list<CvPoint>::const_iterator it = objectPixels.begin(); it != objectPixels.end(); it++) {
            pMask[it->y * mask->cols + it->x] = 1;
        }

        if (!svlIsConnectedComponent(mask, true, 1)) {
            SVL_LOG(SVL_LOG_WARNING, "removing split object " << objId << "...");

            // remove of object
            numPixelsMoved += assignment.objects[objId]->size();
            for (set<SceneRegion *>::iterator it = assignment.objects[objId]->regions.begin();
                 it != assignment.objects[objId]->regions.end(); it++) {
                (*it)->object = NULL;
            }
            delete assignment.objects[objId];
            assignment.objects[objId] = NULL;
        }
    }

    // remap object ids and update neighbors
    if (numPixelsMoved > 0) {
        assignment.remap();
        assignment.updateNeighborhoods();
    }

    cvReleaseMat(&mask);
    return numPixelsMoved;
}

// merge adjacent background regions: eg, sky, road, grass, water, mountain
void SceneInstance::mergeAdjacentBackground(const vector<bool>& mergeClassIndicator)
{
    SVL_ASSERT((int)mergeClassIndicator.size() == gSceneConfig.numRegionLabels);

    int regId = 0;
    while (regId < numRegions()) {
        if ((assignment.regions[regId]->size() == 0) || (assignment.regions[regId]->Sk < 0)) {
            regId += 1;
            continue;
        }

        // prevent lots of little merges at pixel level
        if (assignment.regions[regId]->size() == 1) {
            regId += 1;
            continue;
        }

        // check if we should merge this class
        if (!mergeClassIndicator[assignment.regions[regId]->Sk]) {
            regId += 1;
            continue;
        }

        vector<CvPoint> pixels;
        for (set<SceneRegionBoundary *>::const_iterator it = assignment.regions[regId]->boundaries.begin();
             it != assignment.regions[regId]->boundaries.end(); ++it) {
            SceneRegion *r = (*it)->neighborOf(assignment.regions[regId]);
            if (assignment.regions[regId]->Sk == r->Sk) {
                SVL_LOG(SVL_LOG_VERBOSE, "merging region " << r->id() << " with "
                    << regId << " (" << r->Sk << ")");
                pixels.insert(pixels.end(), r->pixels.begin(), r->pixels.end());
            }
        }

        if (!pixels.empty()) {
            movePixels(pixels, regId);
        }
        regId += 1;
    }

    assignment.remap();
}

//
// image features
//
vector<double> SceneInstance::rawPixelFeatures(int x, int y) const
{
    SVL_ASSERT((_img != NULL) && (x < _img->width) && (y < _img->height));

    vector<double> v(pixelFilterFeatures(x, y));

    // append response at various offsets
    int scales = gSceneConfig.getOptionAsInt("featureGridScales");
    int offset = gSceneConfig.getOptionAsInt("featureGridSpacing");
    if (offset > 0) {
        while (scales-- > 0) {
            v.reserve(v.size() + 18 * numFilters());
            for (int dy = (int)(-1.5 * offset); dy < offset; dy += offset) {
                for (int dx = (int)(-1.5 * offset); dx < offset; dx += offset) {
                    CvRect roi = cvRect(x + dx, y + dy, offset, offset);
                    svlClipRect(roi, _img);

                    if ((roi.width == 0) || (roi.height == 0)) {
                        for (int i = 0; i < numFilters(); i++) {
                            v.push_back(v[i]);
                            v.push_back(v[i]);
                        }
                    } else {
                        VectorXd mu = _filters.mean(roi.x, roi.y, roi.width, roi.height);
                        VectorXd var = _filters.variance(roi.x, roi.y, roi.width, roi.height);
                        for (int i = 0; i < _filters.size(); i++) {
                            v.push_back(mu[i]);
                            v.push_back(var[i]);
                        }
                    }
                }
            }

            offset *= 3;
        }
    }

    // append row feature (relative height)
    if (gSceneConfig.getOptionAsBool("includeRowFeature")) {
        v.push_back((double)y / (double)_img->height);
    }

    return v;
}

double SceneInstance::pixelContrast(const CvPoint &p, const CvPoint& q) const
{
#if 1
    int d = 0;
    unsigned char *pp = (unsigned char *)_img->imageData + p.y * _img->widthStep + 3 * p.x;
    unsigned char *pq = (unsigned char *)_img->imageData + q.y * _img->widthStep + 3 * q.x;
    for (int c = 0; c < 3; c++) {
        d += (int)(pp[c] - pq[c]) * (pp[c] - pq[c]);
    }
    return (double)d;
#else
    double d = 0.0;
    for (int c = 0; c < 3; c++) {
        // CIELab color-space
        double tmp = (double)(CV_IMAGE_ELEM(_labImg, float, p.y, 3 * p.x + c) -
            CV_IMAGE_ELEM(_labImg, float, q.y, 3 * q.x + c));
        d += tmp * tmp;
    }
#endif

    return d;
}

double SceneInstance::regionContrast(int k, int l) const
{
    SVL_ASSERT((k < numRegions()) && (l < numRegions()));

    // assume beginning of appearance vector contains 17-dimensional filters
    SVL_ASSERT(assignment.regions[k]->Ak.size() >= numFilters());
#if 0
    VectorXd z = assignment.regions[k]->appearance().start(numFilters())
        - assignment.regions[l].appearance().start(numFilters());
    return (z.cwise() / _sigmaFilters.start(numFilters())).dot(z);
#else
    VectorXd z = assignment.regions[k]->appearance().start(3) -
        assignment.regions[l]->appearance().start(3);
    return (z.cwise() / _sigmaFilters.start(3)).dot(z);
#endif
}

vector<double> SceneInstance::regionFilterMean(int regId) const
{
    CvMat *mask = cvCreateMat(height(), width(), CV_8UC1);
    cvCmpS(assignment.pixels, (double)regId, mask, CV_CMP_EQ);
    vector<double> v = regionFilterMean(mask);
    cvReleaseMat(&mask);
    return v;
}

vector<double> SceneInstance::regionFilterVariance(int regId) const
{
    CvMat *mask = cvCreateMat(height(), width(), CV_8UC1);
    cvCmpS(assignment.pixels, (double)regId, mask, CV_CMP_EQ);
    vector<double> v = regionFilterVariance(mask);
    cvReleaseMat(&mask);
    return v;
}

vector<double> SceneInstance::regionFilterMean(CvMat *mask) const
{
    vector<double> v(_filters.size());
    Eigen::Map<VectorXd>(&v[0], v.size()) = (mask == NULL) ?
        _filters.mean() : _filters.mean(mask);
    return v;
}

vector<double> SceneInstance::regionFilterVariance(CvMat *mask) const
{
    vector<double> v(_filters.size());
    Eigen::Map<VectorXd>(&v[0], v.size()) = (mask == NULL) ?
        _filters.variance() : _filters.variance(mask);
    return v;
}

// cache's raw pixel appearance features (model may change this)
void SceneInstance::cachePixelAppearance()
{
    int pixIndx = 0;
    pixelAppearance.resize(size());
    for (int y = 0; y < height(); y++) {
        for (int x = 0; x < width(); x++, pixIndx++) {
            pixelAppearance[pixIndx] = pixelFilterFeatures(x, y);
        }
    }
}

void SceneInstance::cachePixelContrast()
{
    horizontalContrast = MatrixXd::Zero(width(), height());
    verticalContrast = MatrixXd::Zero(width(), height());

    double beta = 0.5 / meanContrast();

    for (int y = 0; y < height(); y++) {
        for (int x = 0; x < width(); x++) {
            if (x > 0) {
                horizontalContrast(x, y) =
                    exp(-beta * pixelContrast(cvPoint(x, y), cvPoint(x - 1, y)));
            }

            if (y > 0) {
                verticalContrast(x, y) =
                    exp(-beta * pixelContrast(cvPoint(x, y), cvPoint(x, y - 1)));
            }
        }
    }
}

//
// debugging
//
void SceneInstance::validateData() const
{
    SVL_FCN_TIC;
    SVL_ASSERT(_img != NULL);
    SVL_ASSERT((assignment.pixels != NULL) &&
        (assignment.pixels->rows == height()) &&
        (assignment.pixels->cols == width()));

    // check for valid region id's
    for (int y = 0; y < assignment.pixels->rows; y++) {
        for (int x = 0; x < assignment.pixels->cols; x++) {
            SVL_ASSERT(CV_MAT_ELEM(*assignment.pixels, int, y, x) < numRegions());
        }
    }

    // check for region meta-data is valid
    int totalPixels = 0;
    for (int k = 0; k < numRegions(); k++) {
        totalPixels += assignment.regions[k]->pixels.size();
        for (list<CvPoint>::const_iterator ip = assignment.regions[k]->pixels.begin();
             ip != assignment.regions[k]->pixels.end(); ip++) {
            SVL_ASSERT(CV_MAT_ELEM(*assignment.pixels, int, ip->y, ip->x) == k);
        }
    }
    SVL_ASSERT(totalPixels == size());

    // check multiple-segmentations
    for (int i = 0; i < segmentations.count(); i++) {
        SVL_ASSERT((segmentations[i].height() == height()) &&
            (segmentations[i].width() == width()));

        // check all segments with same id are connected
        CvMat *seg = cvCloneMat(segmentations[i].segmentation());
        for (int segId = 0; ; segId++) {
            CvPoint p = getSeedPixel(seg, segId);
            if (p.x == -1) break;

            CvMat *mask = getConnectedComponent(seg, p);
            cvSet(seg, cvScalar(-1), mask);
            cvReleaseMat(&mask);

            p = getSeedPixel(seg, segId);
            SVL_ASSERT_MSG(p.x == -1, "segment " << i << "." << segId
                << " is disconnected");
        }
        cvReleaseMat(&seg);
    }

    // check all regions with the same id are connected
    CvMat *seg = cvCloneMat(assignment.pixels);
    for (int segId = 0; ; segId++) {
        CvPoint p = getSeedPixel(seg, segId);
        if (p.x == -1) break;

        CvMat *mask = getConnectedComponent(seg, p);
        cvSet(seg, cvScalar(-1), mask);
        cvReleaseMat(&mask);

        p = getSeedPixel(seg, segId);
        if (p.x != -1) {
            SVL_LOG(SVL_LOG_WARNING, "region " << segId << " is disconnected");
        }
    }
    cvReleaseMat(&seg);

    // check objects link to valid and unique regions
    vector<bool> regionCounted(numRegions(), false);
    for (vector<SceneObject *>::const_iterator it = assignment.objects.begin();
         it != assignment.objects.end(); it++) {
        for (set<SceneRegion *>::const_iterator jt = (*it)->regions.begin(); jt != (*it)->regions.end(); jt++) {
            if (((*jt)->id() < 0) || ((*jt)->id() >= numRegions())) {
                printScene();
                SVL_LOG(SVL_LOG_FATAL, "object has invalid region " << *jt);
            }
            if (regionCounted[(*jt)->id()]) {
                printScene();
                SVL_LOG(SVL_LOG_FATAL, "object has repeated region " << *jt);
            }
            regionCounted[(*jt)->id()] = true;
        }
    }

    // check boundaries point to valid objects
    for (vector<SceneObjRegBoundary *>::const_iterator it = assignment.objectBoundaries.begin();
         it != assignment.objectBoundaries.end(); it++) {
        if (!(*it)->valid()) {
            printScene();
            SVL_LOG(SVL_LOG_FATAL, "objectBoundary has missing object or region " << *it);
        }
        if (find(assignment.regions.begin(), assignment.regions.end(), (*it)->region) == assignment.regions.end()) {
            printScene();
            SVL_LOG(SVL_LOG_FATAL, "objectBoundary has invalid region " << *it);
        }
        if (find(assignment.objects.begin(), assignment.objects.end(), (*it)->object) == assignment.objects.end()) {
            printScene();
            SVL_LOG(SVL_LOG_FATAL, "objectBoundary has invalid object " << *it);
        }
    }

    SVL_LOG(SVL_LOG_VERBOSE, "scene data validation: OK");
    SVL_FCN_TOC;
}

void SceneInstance::printScene() const
{
    SVL_LOG(SVL_LOG_MESSAGE, "===== scene has " << numRegions() << " regions =====");
    for (int i = 0; i < numRegions(); i++) {
        if (assignment.regions[i] == NULL) {
            SVL_LOG(SVL_LOG_MESSAGE, "      " << i << ": <null>");
            continue;
        }

        SVL_LOG(SVL_LOG_MESSAGE, "      " << i << ": Sk = "
            << assignment.regions[i]->Sk << ", Nk = " << assignment.regions[i]->size());
        SVL_LOG(SVL_LOG_DEBUG, "       : Ak = " << assignment.regions[i]->appearance().transpose());
        SVL_LOG(SVL_LOG_DEBUG, "       : phi = " << assignment.regions[i]->phi.transpose());
    }

    SVL_LOG(SVL_LOG_MESSAGE, "===== scene has " << numObjects() << " objects =====");
    for (int i = 0; i < numObjects(); i++) {
        if (assignment.objects[i] == NULL) {
            SVL_LOG(SVL_LOG_MESSAGE, "      " << i << ": <null>");
            continue;
        }

        set<int> regionSet;
        for (set<SceneRegion *>::const_iterator it = assignment.objects[i]->regions.begin();
             it != assignment.objects[i]->regions.end(); it++) {
            regionSet.insert((*it)->id());
        }

        if (assignment.objects[i]->prObject.size() == 0) {
            SVL_LOG(SVL_LOG_MESSAGE, "      " << i << ": Ok = " << assignment.objects[i]->Ok
                << " Rk = " << toString(regionSet) << " Pr(Ok) = n/a");
        } else {
            SVL_LOG(SVL_LOG_MESSAGE, "      " << i << ": Ok = " << assignment.objects[i]->Ok
                << " Rk = " << toString(regionSet)
                << " Pr(Ok) = " << assignment.objects[i]->prObject.transpose());
            SVL_LOG(SVL_LOG_DEBUG, "       : phi = " << assignment.objects[i]->phi.transpose());
        }
    }

    SVL_LOG(SVL_LOG_MESSAGE, "===== scene has "
        << assignment.regionBoundaries.size() << " region boundaries =====");
    for (vector<SceneRegionBoundary *>::const_iterator it = assignment.regionBoundaries.begin();
         it != assignment.regionBoundaries.end(); it++) {
        SVL_LOG(SVL_LOG_MESSAGE, "      " << (*it)->id() << ": (k, l) = "
            << (*it)->regionA->id() << ", " << (*it)->regionB->id());
        if ((*it)->phi.size() > 0) {
            SVL_LOG(SVL_LOG_DEBUG, "        : phi = " << (*it)->phi.transpose());
        }
    }

    SVL_LOG(SVL_LOG_MESSAGE, "===== scene has "
        << assignment.objectBoundaries.size() << " object boundaries =====");
    for (vector<SceneObjRegBoundary *>::const_iterator it = assignment.objectBoundaries.begin();
         it != assignment.objectBoundaries.end(); it++) {
        SVL_LOG(SVL_LOG_MESSAGE, "      " << (*it)->id() << ": region = "
            << (*it)->region->id() << ", object = " << (*it)->object->id());
        if ((*it)->phi.size() > 0) {
            SVL_LOG(SVL_LOG_DEBUG, "        : phi = " << (*it)->phi.transpose());
        }
    }
}

// protected member functions --------------------------------------------------------

void SceneInstance::initScene()
{
    SVL_FCN_TIC;
    SVL_ASSERT(_img != NULL);

    // set up automatic regularization of covariance matrices
    svlGaussian::AUTO_RIDGE = true;

    // gaussian smooth image
    if (gSceneConfig.getOptionAsBool("smoothImage")) {
        cvSmooth(_img, _img, CV_GAUSSIAN, 3, 3);
    }

    // convert image to CIELab
    _labImg = cvCreateImage(cvGetSize(_img), IPL_DEPTH_32F, 3);
    cvConvertScale(_img, _labImg, 1.0 / 255.0, 0.0);
    cvCvtColor(_labImg, _labImg, CV_BGR2Lab);

    // convert image to greyscale
    _grayImg = ::greyImage(_img);

    // create edgemap
    svlSoftEdgeMap edgeMapCalculator;
    _softEdgeImg = cvCloneImage(edgeMapCalculator.processImage(_img));

    // initialize random variables
    assignment.resize(height(), width());

    // min/maxDepth of pixels (used in sceneReconstruct3D)
    assignment.minDepth = MatrixXd::Zero(height(), width());
    assignment.maxDepth = MatrixXd::Ones(height(), width());

    // initialize over-segmentations
    segmentations.clear(width(), height());

    // filterbank response images
    svlTextonFilterBank filterBank(gSceneConfig.getOptionAsDouble("filterBandwidth"));
    vector<IplImage *> responses;
    filterBank.filter(_img, responses);
    SVL_LOG(SVL_LOG_DEBUG, "...allocated "  << (responses.size() * size() * sizeof(float) / (1024 * 1024))
        << "MB for filter responses");
    _filters.addResponseImages(responses);
    _sigmaFilters = _filters.variance().cwise() + SVL_DBL_MIN;

    // directional filter derivatives
    SVL_LOG(SVL_LOG_DEBUG, "...allocating "  << (2 * _filters.size() * size() * sizeof(float) / (1024 * 1024))
        << "MB for directional filter responses");
    IplImage *r = cvCreateImage(cvGetSize(_img), IPL_DEPTH_32F, 1);
    for (int i = 0; i < _filters.size(); i++) {
        cvSobel(_filters.getResponseImage(i), r, 1, 0, 3);
        _dirFilters.copyResponseImage(r);
        cvSobel(_filters.getResponseImage(i), r, 0, 1, 3);
        _dirFilters.copyResponseImage(r);
    }
    cvReleaseImage(&r);

#if 0
    cvNamedWindow("debug", 1);
    for (int i = 0; i < (int)_filters.responses.size(); i++) {
        IplImage *tmp = cvCloneImage(_filters.responses[i]);
        scaleToRange(tmp, 0.0, 1.0);
        cvShowImage("debug", tmp);
        cvWaitKey(1000);
        cvReleaseImage(&tmp);
    }
    cvDestroyWindow("debug");
#endif

    // mean contrast
    _meanContrast = 0.0;
    for (int y = 0; y < _img->height - 1; y++) {
        for (int x = 0; x < _img->width - 1; x++) {
            _meanContrast += pixelContrast(cvPoint(x, y), cvPoint(x + 1, y));
            _meanContrast += pixelContrast(cvPoint(x, y), cvPoint(x, y + 1));
        }
    }
    _meanContrast /= 2.0 * (double)((_img->height - 1) * (_img->width - 1));

    SVL_FCN_TOC;
}

void SceneInstance::freeScene()
{
    if (_img != NULL) {
        cvReleaseImage(&_img);
    }

    if (_labImg != NULL) {
        cvReleaseImage(&_labImg);
    }

    if (_grayImg != NULL) {
        cvReleaseImage(&_grayImg);
    }

    if (_softEdgeImg != NULL) {
        cvReleaseImage(&_softEdgeImg);
    }

    segmentations.clear();
    assignment.clear();

    _filters.clear();
    _dirFilters.clear();

    svlReleaseImages(objDetectorChannels);
    objDetectorChannels.clear();
}

void SceneInstance::addRegionStatistics(int regId, int pixIndx, int x, int y, bool bIncAppMoments)
{
    if (!pixelSingletons.empty()) {
        assignment.regions[regId]->pixelScores += Eigen::Map<VectorXd>(&pixelSingletons[pixIndx][0],
            pixelSingletons[pixIndx].size());
    }

    if (pixelAppearance.empty()) {
        assignment.regions[regId]->Ak.accumulate(pixelFilterFeatures(x, y));
        if (bIncAppMoments) {
            assignment.regions[regId]->xAk.accumulate(pixelFilterFeatures(x, y), (double)x);
            assignment.regions[regId]->yAk.accumulate(pixelFilterFeatures(x, y), (double)y);
        }
    } else {
        assignment.regions[regId]->Ak.accumulate(pixelAppearance[pixIndx]);
        if (bIncAppMoments) {
            assignment.regions[regId]->xAk.accumulate(pixelAppearance[pixIndx], (double)x);
            assignment.regions[regId]->yAk.accumulate(pixelAppearance[pixIndx], (double)y);
        }
    }

    assignment.regions[regId]->centroidSum.x() += x;
    assignment.regions[regId]->centroidSum.y() += y;
}

void SceneInstance::subRegionStatistics(int regId, int pixIndx, int x, int y, bool bIncAppMoments)
{
    if (!pixelSingletons.empty()) {
        assignment.regions[regId]->pixelScores -= Eigen::Map<VectorXd>(&pixelSingletons[pixIndx][0],
            pixelSingletons[pixIndx].size());
    }

    if (pixelAppearance.empty()) {
        assignment.regions[regId]->Ak.subtract(pixelFilterFeatures(x, y));
        if (bIncAppMoments) {
            assignment.regions[regId]->xAk.subtract(pixelFilterFeatures(x, y), (double)x);
            assignment.regions[regId]->yAk.subtract(pixelFilterFeatures(x, y), (double)y);
        }
    } else {
        assignment.regions[regId]->Ak.subtract(pixelAppearance[pixIndx]);
        if (bIncAppMoments) {
            assignment.regions[regId]->xAk.subtract(pixelAppearance[pixIndx], (double)x);
            assignment.regions[regId]->yAk.subtract(pixelAppearance[pixIndx], (double)y);
        }
    }

    assignment.regions[regId]->centroidSum.x() -= x;
    assignment.regions[regId]->centroidSum.y() -= y;
}

// returns seed pixel belonging to component
CvPoint SceneInstance::getSeedPixel(const CvMat *m, int componentId) const
{
    SVL_ASSERT((m != NULL) && (cvGetElemType(m) == CV_32SC1));

    int *p = (int *)CV_MAT_ELEM_PTR(*m, 0, 0);
    for (int y = 0; y < m->height; y++) {
        for (int x = 0; x < m->width; x++, p++) {
            if (*p == componentId) {
                return cvPoint(x, y);
            }
        }
    }

    return cvPoint(-1, -1);
}

// returns a mask corresponding to the connected component
CvMat *SceneInstance::getConnectedComponent(const CvMat *m, const CvPoint p) const
{
    SVL_ASSERT((m != NULL) && (cvGetElemType(m) == CV_32SC1));
    const CvPoint NEIGHBORS[8] = {{-1, -1}, {-1, 0}, {-1, 1},
        {0, -1}, {0, 1}, {1, -1}, {1, 0}, {1, 1}};

    CvMat *component = cvCreateMat(m->rows, m->cols, CV_8UC1);
    cvZero(component);

    int componentId = CV_MAT_ELEM(*m, int, p.y, p.x);

    deque<CvPoint> frontier(1, p);
    CV_MAT_ELEM(*component, unsigned char, p.y, p.x) = 1;
    while (!frontier.empty()) {
        CvPoint q = frontier.front();
        frontier.pop_front();
        SVL_ASSERT(CV_MAT_ELEM(*component, unsigned char, q.y, q.x) != 0);

        // 8-connected neighbors
        for (int i = 0; i < 8; i++) {
            CvPoint r = cvPoint(q.x + NEIGHBORS[i].x, q.y + NEIGHBORS[i].y);

            if ((r.x >= 0) && (r.x < m->width) && (r.y >= 0) && (r.y < m->height)) {
                if ((CV_MAT_ELEM(*component, unsigned char, r.y, r.x) == 0) &&
                    (CV_MAT_ELEM(*m, int, r.y, r.x) == componentId)) {
                    frontier.push_back(r);
                    CV_MAT_ELEM(*component, unsigned char, r.y, r.x) = 1;
                }
            }
        }
    }

    return component;
}
