/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneInstance.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**  This file encapsulates a single instance of a scene.
**
** TODO:
**  1. remove pixelSingletons and pixelMarginals
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <list>

#include "Eigen/Core"

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"

using namespace std;
USING_PART_OF_NAMESPACE_EIGEN

// SceneInstance class ----------------------------------------------------

class SceneInstance {
 protected:
    string _baseName;                   // image identifier
    IplImage *_img;                     // 3-channel RGB image of the scene
    IplImage *_labImg;                  // 3-channel CIELab image of the scene
    IplImage *_grayImg;                 // 1-channel greyscale image of the scene
    IplImage *_softEdgeImg;             // 1-channel edgemap image of the scene
    svlFilterBankResponse _filters;     // pixel filter responses
    svlFilterBankResponse _dirFilters;  // filter directional derivatives
    VectorXd _sigmaFilters;             // pixel filter response variance    
    double _meanContrast;               // mean contrast

 public:
    // cached pixel features (initialized externally)
    vector<vector<double> > pixelBoostedResponses; // boosted classifier responses
    vector<vector<double> > pixelSingletons;    // log-space class singletons (Sk)
    vector<vector<double> > pixelMarginals;     // exp-normalized class singletons (Sk)
    vector<vector<double> > pixelAppearance;    // pixel appearance features (for Ak)
    vector<double> hpw;//horizontal pairwise term
    vector<double> vpw;//vertical pairwise term
    MatrixXd horizontalContrast;                // contrast between (x, y) and (x - 1, y)
    MatrixXd verticalContrast;                  // contrast between (x, y) and (x, y - 1)
    vector<IplImage *> objDetectorChannels;     // image channels for object detection
    
    // scene structure, region meta-data and random variables
    SceneAssignment assignment;                // pixels, regions and objects

     // over-segmentation proposal moves during inference
    svlMultiSeg segmentations;

 public:
    SceneInstance(const char *imgFilename, const char *baseName = NULL);
    SceneInstance(const IplImage *img, const char *baseName = NULL);
    SceneInstance(const SceneInstance& instance);
    virtual ~SceneInstance();

    // accessors
    inline IplImage *image() const { return _img; }
    inline IplImage *greyImage() const { return _grayImg; }
    inline IplImage *edgeMap() const { return _softEdgeImg; }
    inline string filename() const;
    inline const string &name() const { return _baseName; }

    inline int width() const { return _img->width; }
    inline int height() const { return _img->height; }
    inline int size() const { return (_img->width * _img->height); }
    void resize(int w, int h);

    inline int numFilters() const { return _filters.size(); }
    inline int numAppFeatures() const { 
        return pixelAppearance.empty() ? 0 : (int)pixelAppearance[0].size();
    }
    inline int numRegions() const { return (int)assignment.regions.size(); }
    inline int numObjects() const { return (int)assignment.objects.size(); }

    // i/o
    void loadSegmentations(bool bClearExisting = true);
    void loadGroundTruth();
    void loadAssignments(const char *directory);
    void loadRegions(const char *filename);
    void loadPixelMAPSingletons(const char *filename);

    // variable/label indices
    inline int pixel2Indx(int x, int y) const { return y * _img->width + x; }
    inline CvPoint indx2Pixel(int indx) const {
        return cvPoint(indx % _img->width, indx / _img->width);
    }

    // variable updated/access
    CvMat *getPixelClasses() const;

    // full assignment update. calculates new sufficient statistics
    void replaceAssignment(const SceneAssignment &a, bool bUpdateStatistics = true);

    // assign/re-assign pixels to regions (neighborhood may change, but
    // region assignments are unaffected). returns number of pixels moved.
    // -1 argument for id indicates create region
    void resetRegions(); 
    void assignPixelsFromLabels(CvMat *regionLabel);
    void assignPixels(const CvMat *r);
    int movePixels(const CvMat *r);
    int movePixels(const vector<CvPoint>& pixels, int regionId);
    int movePixels(const vector<CvPoint>& pixels, const vector<int>& regionId);
    int movePixels(int oldRegionId, int newRegionId);

    // assign/re-assign regions to objects. returns the number of pixels moves
    void resetObjects();
    int moveRegions(int regId, int objId);
    int moveRegions(const vector<int> regIds, int objId);
    int removeRegions(int regId, int objId);
    int removeObject(int objId);
    
    // re-assigns pixels and re-maps regions
    void removeSmallRegions(int minSize = 10);
    void splitOrphans();
    int removeSplitObjects();
    void mergeAdjacentBackground(const vector<bool>& mergeClassIndicator);

    // image features
    inline vector<double> pixelFilterFeatures(int x, int y) const;
    inline vector<double> pixelFilterDerivatives(int x, int y) const;
    vector<double> rawPixelFeatures(int x, int y) const;
    double pixelContrast(const CvPoint &p, const CvPoint& q) const;
    double regionContrast(int k, int l) const;
    inline double meanContrast() const { return _meanContrast; }
    vector<double> regionFilterMean(int regId) const;
    vector<double> regionFilterVariance(int regId) const;
    vector<double> regionFilterMean(CvMat *mask = NULL) const;
    vector<double> regionFilterVariance(CvMat *mask = NULL) const;

    void cachePixelAppearance();
    void cachePixelContrast();

    // debugging
    void validateData() const;
    void printScene() const;

 protected:
    void initScene();
    void freeScene();
    void addRegionStatistics(int regId, int pixIndx, int x, int y, bool bIncAppMoments);
    void subRegionStatistics(int regId, int pixIndx, int x, int y, bool bIncAppMoments);

    // debugging
    CvPoint getSeedPixel(const CvMat *m, int componentId) const;
    CvMat *getConnectedComponent(const CvMat *m, const CvPoint p) const;
    
};

// inline functions --------------------------------------------------

inline string SceneInstance::filename() const
{
    return gSceneConfig.filename("imgDir", _baseName, "imgExt");
}

inline vector<double> SceneInstance::pixelFilterFeatures(int x, int y) const
{
    SVL_ASSERT((_img != NULL) && (x < _img->width) && (y < _img->height));

    vector<double> v(numFilters());
    Eigen::Map<VectorXd>(&v[0], v.size()) = _filters.value(x, y);

    return v;
}

inline vector<double> SceneInstance::pixelFilterDerivatives(int x, int y) const
{
    SVL_ASSERT((_img != NULL) && (x < _img->width) && (y < _img->height));

    vector<double> v(_dirFilters.size());
    Eigen::Map<VectorXd>(&v[0], v.size()) = _dirFilters.value(x, y);

    return v;
}

