/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneLearnAugmentedObjectModel.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Application for learning a "scene"-augmented object model using HOG feature.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string.h>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

#include "sceneDataLoader.h"
#include "sceneConfiguration.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneVisualization.h"

#include "sceneHOGFeatures.h"
#include "sceneSlidingWindowDetector.h"

#define WINDOW_NAME "objectDetector"

using namespace std;

// globals --------------------------------------------------------------------

bool gbVisualize = true;

// helper functions -----------------------------------------------------------

vector<double> extractFeatures(SceneInstance &instance, SceneRegionObjectClassifier *detector, CvRect& r)
{
    int winWidth = detector->getWinWidth();
    int winHeight = detector->getWinHeight();

#if 0
    IplImage *subImage = cvCreateImage(cvSize(r.width, r.height), instance.image()->depth, instance.image()->nChannels);
    cvSetImageROI(instance.image(), r);
    cvCopyImage(instance.image(), subImage);
    cvResetImageROI(instance.image());
    resizeInPlace(&subImage, winHeight, winWidth);

    double score = detector->classifyRegion(subImage);
    cvReleaseImage(&subImage);
#else
    double score = detector->classifyRegion(r, instance.image());
#endif
    SVL_LOG(SVL_LOG_MESSAGE, "score: " << score);

    vector<double> fv(gSceneConfig.numRegionLabels, 0.0);
    for (int y = r.y; y < r.y + r.height; y++) {
        for (int x = r.x; x < r.x + r.width; x++) {
            int indx = instance.pixel2Indx(x, y);
            fv[argmax(instance.pixelMarginals[indx])] += 1.0;
        }
    }
    Eigen::Map<VectorXd>(&fv[0], fv.size()) /= (double)r.width * r.height;
    fv.push_back(score);

    return fv;
}

void test(ScenePixelModel *model, SceneRegionObjectClassifier *detector, svlObject2dSequence &objectList)
{
    cvNamedWindow(WINDOW_NAME, 1);

    for (svlObject2dSequence::iterator it = objectList.begin(); it != objectList.end(); it++) {
        SVL_LOG(SVL_LOG_VERBOSE, "Processing " << it->first << "...");
        if (it->second.empty()) continue;

        string imgFilename = gSceneConfig.filename("imgDir", it->first, "imgExt");
        SceneInstance instance(imgFilename.c_str());
        svlImageSegmentation *superSeg = NULL;

        string superSegFilename = gSceneConfig.filename("segDir", it->first, "combinedSegExt");
        if (svlFileExists(superSegFilename.c_str())) {
            superSeg = new svlImageSegmentation(superSegFilename.c_str(), instance.width(), instance.height());
        } else {
            instance.loadSegmentations();
            superSeg = new svlImageSegmentation(instance.segmentations.intersection());
            writeMatrix(superSeg->segmentation(), superSegFilename.c_str());
        }
        SVL_LOG(SVL_LOG_VERBOSE, "supersegmentation has " << superSeg->numSegments() << " segments");

        model->initializeInstance(&instance);
        model->inferPixels(&instance);
        model->inferRegions(&instance);

        IplImage *canvas = cvCloneImage(instance.image());
        for (svlObject2dFrame::iterator jt = it->second.begin(); jt != it->second.end(); jt++) {
            CvRect r = cvRect((int)jt->x, (int)jt->y, (int)jt->w, (int)jt->h);
            r.x = std::max(0, (int)(r.x - 0.1 * r.width));
            r.y = std::max(0, (int)(r.y - 0.1 * r.height));
            r.width = std::min((int)(1.2 * r.width), instance.width() - r.x);
            r.height = std::min((int)(1.2 * r.height), instance.height() - r.y);

            svlImageSegmentation visSeg(*superSeg);
            set<int> objMove = visSeg.findSegments(r, 0.9);
            CvRect bbMove = visSeg.boundingBox(objMove, 0.0);
#if 0
            visSeg.mergeSegments(objMove);
#else
            set<int> otherSegs;
            for (int i = 0; i < visSeg.numSegments(); i++) {
                if (objMove.find(i) == objMove.end()) {
                    otherSegs.insert(i);
                }
            }
            visSeg.mergeSegments(otherSegs);
#endif

            if (gbVisualize) {
                cvCopyImage(instance.image(), canvas);
                //visualizeClasses(&instance, canvas);
                //visualizeSegmentation(&instance, canvas, superSeg->segmentation());
                visualizeSegmentation(&instance, canvas, visSeg.segmentation(), -1, 0x00, 0xff, 0x00);
                cvRectangle(canvas, cvPoint(r.x, r.y), cvPoint(r.x + r.width, r.y + r.height), CV_RGB(255,255,255), 3);
                cvRectangle(canvas, cvPoint(r.x, r.y), cvPoint(r.x + r.width, r.y + r.height), CV_RGB(0,0,255), 1);
                //cvRectangle(canvas, cvPoint(bbMove.x, bbMove.y), cvPoint(bbMove.x + bbMove.width, bbMove.y + bbMove.height), CV_RGB(255,255,255), 3);
                //cvRectangle(canvas, cvPoint(bbMove.x, bbMove.y), cvPoint(bbMove.x + bbMove.width, bbMove.y + bbMove.height), CV_RGB(255,0,255), 1);

                cvShowImage(WINDOW_NAME, canvas);
                cvWaitKey(100);
            }

            double pOriginal = detector->classifyRegion(r, instance.image());
            double pMove = detector->classifyRegion(bbMove, instance.image());
            SVL_LOG(SVL_LOG_MESSAGE, "p(A), p(B): " << pOriginal << ", " << pMove);
        }
        cvReleaseImage(&canvas);
        delete superSeg;
    }
}

// main -----------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./sceneLearnAugmentedObjectModel [OPTIONS] <objId> <posXML> <negXML>\n";
    cerr << "OPTIONS:" << endl
         << "  -width <num>      :: window width (default: 64)" << endl
         << "  -height <num>     :: window height (default: 128)" << endl
         << "  -posDir <dir>     :: positive images directory" << endl
         << "  -negDir <dir>     :: negative images directory" << endl
         << "  -n <object name>  :: name of object" << endl
         << "  -outModel <file>  :: model to output" << endl
         << "  -o <xml>          :: output object detections" << endl
         << "  -cache <file>     :: feature cache" << endl
         << "  -noFeatures       :: turn off region features" << endl
         << "  -x                :: trun off visualization" << endl
         << "  -TEST             :: testing code" << endl
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    const int NUM_PARAMETERS = 3;

    // read commandline parameters
    int width = 64;
    int height = 128;
    const char *posDir = "";
    const char *negDir = "";
    const char *objectName = NULL;
    const char *outputModel = NULL;
    const char *outputDetections = NULL;
    const char *featureCache = NULL;
    bool bNoFeatures = false;
    bool bTest = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_INT_OPTION("-width", width)
        SVL_CMDLINE_INT_OPTION("-height", height)
        SVL_CMDLINE_STR_OPTION("-posDir", posDir)
        SVL_CMDLINE_STR_OPTION("-negDir", negDir)
        SVL_CMDLINE_STR_OPTION("-n", objectName)
        SVL_CMDLINE_STR_OPTION("-outModel", outputModel)
        SVL_CMDLINE_STR_OPTION("-o", outputDetections)
        SVL_CMDLINE_STR_OPTION("-cache", featureCache)
        SVL_CMDLINE_BOOL_OPTION("-noFeatures", bNoFeatures)
        SVL_CMDLINE_BOOL_OPTION("-TEST", bTest)
        SVL_CMDLINE_BOOL_TOGGLE_OPTION("-x", gbVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != NUM_PARAMETERS) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    int objId = atoi(SVL_CMDLINE_ARGV[0]);
    const char *posXML = SVL_CMDLINE_ARGV[1];
    const char *negXML = SVL_CMDLINE_ARGV[2];

    // create the HOG-based object detector
    SVL_LOG(SVL_LOG_VERBOSE, "creating detector for " << gSceneObjectDefs.name(objId));
    SceneRegionObjectClassifier *detector = gSceneObjectDefs.createDetector(objId);
    SVL_ASSERT(detector != NULL);

    int winWidth = detector->getWinWidth();
    int winHeight = detector->getWinHeight();

    // initialize pixel models
    ScenePixelModel *model = new ScenePixelModel();
    SVL_ASSERT(model != NULL);
    model->loadModels(gSceneConfig.filebase("modelsDir", "").c_str());

    // load positive and negative examples
    svlObject2dSequence posObjects;
    svlObject2dSequence negObjects;

    posObjects.read(posXML);
    negObjects.read(negXML);

    // remove positives from negatives
    for (svlObject2dSequence::iterator it = negObjects.begin(); it != negObjects.end(); ) {
        svlObject2dSequence::iterator jt = it++;
        if (posObjects.find(jt->first) != posObjects.end()) {
            negObjects.erase(jt);
        }
    }

    SVL_LOG(SVL_LOG_MESSAGE, "read " << countObjects(posObjects) << " positive and "
        << countObjects(negObjects) << " negative samples");

    SceneFeatureCache trainingData;

    // DEBUGGING
    if (bTest) {
        test(model, detector, posObjects);
        return 0;
    }

    if ((featureCache != NULL) && (svlFileExists(featureCache))) {
        trainingData.read(featureCache);

        // iterate through positive and negative samples
        int count = 0;
        for (svlObject2dSequence::iterator it = posObjects.begin(); it != posObjects.end(); it++) {
            for (svlObject2dFrame::iterator jt = it->second.begin(); jt != it->second.end(); jt++) {
                jt->index = count++;
            }
        }
        for (svlObject2dSequence::iterator it = negObjects.begin(); it != negObjects.end(); it++) {
            for (svlObject2dFrame::iterator jt = it->second.begin(); jt != it->second.end(); jt++) {
                jt->index = count++;
            }
        }
        SVL_ASSERT(trainingData.size() == count);

    } else {

        // iterate through positive samples
        for (svlObject2dSequence::iterator it = posObjects.begin(); it != posObjects.end(); it++) {
            SVL_LOG(SVL_LOG_VERBOSE, "Processing " << it->first << " (" << trainingData.size() << " samples)...");
            if (it->second.empty()) continue;

            string imgFilename = string(posDir) + string(it->first) + string(".jpg");
            IplImage *img = cvLoadImage(imgFilename.c_str());
            SVL_ASSERT_MSG(img != NULL, imgFilename);

            SceneInstance instance(img, it->first.c_str());
            model->initializeInstance(&instance);
            //model->inferPixels(&instance);
            //model->inferRegions(&instance);

            for (svlObject2dFrame::iterator jt = it->second.begin(); jt != it->second.end(); jt++) {
                CvRect r = cvRect((int)jt->x, (int)jt->y, (int)jt->w, (int)jt->h);
                r.x = std::max(0, (int)(r.x - 0.1 * r.width));
                r.y = std::max(0, (int)(r.y - 0.1 * r.height));
                r.width = std::min((int)(1.2 * r.width), img->width - r.x);
                r.height = std::min((int)(1.2 * r.height), img->height - r.y);

                trainingData.features.push_back(extractFeatures(instance, detector, r));
                trainingData.labels.push_back(0);

                jt->index = trainingData.size() - 1;
            }

            cvReleaseImage(&img);
        }

        // iterate through negative samples
        for (svlObject2dSequence::iterator it = negObjects.begin(); it != negObjects.end(); it++) {
            SVL_LOG(SVL_LOG_VERBOSE, "Processing " << it->first << " (" << trainingData.size() << " samples)...");
            if (it->second.empty()) continue;

            string imgFilename = string(negDir) + string(it->first) + string(".jpg");
            IplImage *img = cvLoadImage(imgFilename.c_str());
            SVL_ASSERT_MSG(img != NULL, imgFilename);

            SceneInstance instance(img, it->first.c_str());
            model->initializeInstance(&instance);
            //model->inferPixels(&instance);
            //model->inferRegions(&instance);

            for (svlObject2dFrame::iterator jt = it->second.begin(); jt != it->second.end(); jt++) {
                CvRect r = cvRect((int)jt->x, (int)jt->y, (int)jt->w, (int)jt->h);

                trainingData.features.push_back(extractFeatures(instance, detector, r));
                trainingData.labels.push_back(1);

                jt->index = trainingData.size() - 1;
            }

            cvReleaseImage(&img);
        }

        // set all training data to have equal weight and write to cache
        trainingData.weights.resize(trainingData.size(), 1.0);
        if (featureCache != NULL) {
            trainingData.write(featureCache);
        }
    }

    // augment feature vectors

#if 1
    // DEBUGGING: no score
    for (int i = 0; i < trainingData.size(); i++) {
        trainingData.features[i].pop_back();
    }
#endif

#if 0
    if (!bNoFeatures) {
        int n = trainingData.numFeatures();
        for (int i = 0; i < trainingData.size(); i++) {
            for (int j = 0; j < n; j++) {
                trainingData.features[i].push_back(trainingData.features[i][j] *
                    trainingData.features[i][j]);
            }
        }
    }
#endif
    if (bNoFeatures) {
        for (int i = 0; i < trainingData.size(); i++) {
            double score = trainingData.features[i].back();
            trainingData.features[i] = vector<double>(1, score);
        }
    }
    for (int i = 0; i < trainingData.size(); i++) {
        trainingData.features[i].push_back(1.0);
    }

    // train classifier
    SVL_LOG(SVL_LOG_MESSAGE, "training classifier with " << trainingData.size() << " samples");
    svlMultiClassLogistic objClassifier(trainingData.numFeatures(), 2);
    objClassifier.train(trainingData.features, trainingData.labels, trainingData.weights);

    if (outputModel != NULL) {
        objClassifier.save(outputModel);
    }

    vector<int> predictedLabels;
    objClassifier.getClassifications(trainingData.features, predictedLabels);
    svlConfusionMatrix confusion(2);
    confusion.accumulate(trainingData.labels, predictedLabels);
    confusion.write(cout);

    // output detections
    if (outputDetections != NULL) {
        svlObject2dSequence outputObjects(posObjects);
        for (svlObject2dSequence::const_iterator it = negObjects.begin(); it != negObjects.end(); it++) {
            outputObjects.insert(*it);
        }
        SVL_ASSERT_MSG(outputObjects.size() == posObjects.size() + negObjects.size(),
            outputObjects.size() << " != " << posObjects.size() << " + " << negObjects.size());

        vector<vector<double> > m;
        objClassifier.getMarginals(trainingData.features, m);
        for (svlObject2dSequence::iterator it = outputObjects.begin(); it != outputObjects.end(); it++) {
            for (svlObject2dFrame::iterator jt = it->second.begin(); jt != it->second.end(); jt++) {
                jt->pr = m[jt->index][0];
            }
        }

        outputObjects.write(outputDetections);
    }

    // free memory
    cvDestroyAllWindows();
    delete detector;
    delete model;
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);

    return 0;
}
