/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneLearnErodedObjectModel.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**   Requires input from sceneCacheErodedObjectFeatures.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <limits>
#include <set>
#include <algorithm>
#include <iterator>
#include <sys/time.h>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneObjectModel.h"
#include "sceneVisualization.h"
#include "sceneSegmentDict.h"
#include "sceneObjDetDict.h"
#include "sceneRegionFeatures.h"

#define WINDOW_NAME "sceneInference"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneLearnErodedObjectModel [OPTIONS] <objectName> <trainList>\n";
    cerr << "OPTIONS:\n"
         << "  -t <type>         :: classifier type (BOOSTING (default), LOGISTIC)\n"
         << "  -i <stem>         :: input filestem (appends <objectName>.cachedFeatures.bin and <objectName>.cachedDetections.xml)\n"
         << "  -o <stem>         :: output filestem (appends <objectName>.objFeatureWhitener.xml and <objectName>.objClassifier.xml)\n"
         << "  -eval <list> <file> :: evaluation list and output filename\n"
         << "  -x                :: visualize output\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();
    svlConfigurationManager::get().configure("svlML.svlBoostedClassifier", "boostMethod", "DISCRETE");
    svlConfigurationManager::get().configure("svlML.svlBoostedClassifier", "pseudoCounts", "0");

    // read commandline parameters
    const char *classifierType = "BOOSTING";
    const char *inStem = "";
    const char *outStem = NULL;
    const char *evalList = NULL;
    const char *evalFilename = NULL;
    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-t", classifierType)
        SVL_CMDLINE_STR_OPTION("-i", inStem)
        SVL_CMDLINE_STR_OPTION("-o", outStem)
        SVL_CMDLINE_OPTION_BEGIN("-eval", p)
            evalList = p[0];
            evalFilename = p[1];
        SVL_CMDLINE_OPTION_END(2)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 2) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *objectName = SVL_CMDLINE_ARGV[0];
    const char *trainList = SVL_CMDLINE_ARGV[1];

    // read list of training images
    vector<string> baseNames;
    SVL_LOG(SVL_LOG_MESSAGE, "Reading training list from " << trainList << "...");
    baseNames = svlReadFile(trainList);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");

    // read cached data
    svlObject2dSequence cachedDetections;
    SceneFeatureCache data;

    string filename = string(inStem) + objectName + string(".cachedFeatures.bin");
    SVL_LOG(SVL_LOG_MESSAGE, "Reading cached features from " << filename << "...");
    data.read(filename.c_str());
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << data.size() << " feature vectors");

    filename = string(inStem) + objectName + string(".cachedDetections.xml");
    SVL_LOG(SVL_LOG_MESSAGE, "Reading cached detections from " << filename << "...");
    cachedDetections.read(filename.c_str());
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << cachedDetections.size() << " image frames");

    // add square features
    if (!strcasecmp(classifierType, "LOGISTIC")) {
        int numFeatures = data.numFeatures() - 1;
        for (int i = 0; i < data.size(); i++) {
#if 1
            data.features[i].reserve(2 * numFeatures + 1);
            for (int j = 0; j < numFeatures; j++) {
                data.features[i].push_back(data.features[i][j] * data.features[i][j]);
            }
#else
            data.features[i].reserve(numFeatures * (numFeatures  + 1) / 2 + 1);
            for (int j = 0; j < numFeatures; j++) {
                for (int k = 0; k <= j; k++) {
                    data.features[i].push_back(data.features[i][j] * data.features[i][k]);
                }
            }
#endif
        }
    }

    // filter non-training images
    SVL_LOG(SVL_LOG_MESSAGE, "Filtering non-training images...");
    SceneFeatureCache trainingData;
    trainingData.reserve(data.size() * (baseNames.size() + 1) / (cachedDetections.size() + 1));
    for (unsigned i = 0; i < baseNames.size(); i++) {
        svlObject2dFrame frame = cachedDetections[baseNames[i]];
        for (svlObject2dFrame::const_iterator it = frame.begin(); it != frame.end(); it++) {
            if ((it->pr > 0.25) && (it->pr < 0.5))
                continue;
            trainingData.features.push_back(data.features[it->index]);
            trainingData.labels.push_back(data.labels[it->index]);
            trainingData.weights.push_back(data.weights[it->index]);
        }
    }

    // train classifier
    SVL_LOG(SVL_LOG_MESSAGE, "whitening " << trainingData.numFeatures() << " features...");
    svlFeatureWhitener featureWhitener;
    featureWhitener.clear(trainingData.numFeatures());
    featureWhitener.train(trainingData.features);
    featureWhitener.evaluate(trainingData.features);

    SVL_LOG(SVL_LOG_MESSAGE, "training classifier with " << trainingData.numFeatures() << " features...");
    svlClassifier *classifier = NULL;
    if (!strcasecmp(classifierType, "LOGISTIC")) {
        double lambda = 1.0e-3;
        classifier = new svlMultiClassLogistic(trainingData.numFeatures(), 2, lambda);
        classifier->train(trainingData.features, trainingData.labels, trainingData.weights);
    } else if (!strcasecmp(classifierType, "BOOSTING")) {
        classifier = new svlBoostedClassifierSet(1);
        classifier->train(trainingData.features, trainingData.labels);
    } else {
        SVL_LOG(SVL_LOG_FATAL, "unsuported classifier type " << classifierType);
    }
    SVL_ASSERT(classifier != NULL);

    // output classifier
    if (outStem != NULL) {
        string filename = string(outStem) + objectName + string(".objFeatureWhitener.xml");
        SVL_LOG(SVL_LOG_VERBOSE, "Writing feature whitener to " << filename << "...");
        featureWhitener.save(filename.c_str());
        filename = string(outStem) + objectName + string(".objClassifier.xml");
        SVL_LOG(SVL_LOG_VERBOSE, "Writing object classifier to " << filename << "...");
        classifier->save(filename.c_str());
    }

    // evaluation
    if (evalList != NULL) {
        // read evaluation list
        SVL_LOG(SVL_LOG_MESSAGE, "Reading evaluation list from " << evalList << "...");
        baseNames = svlReadFile(evalList);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");

        // whiten data correctly
        featureWhitener.evaluate(data.features);

        // evaluate on cached detection candidates
        svlObject2dSequence detections;
        for (unsigned i = 0; i < baseNames.size(); i++) {
            svlObject2dFrame frame = cachedDetections[baseNames[i]];
            for (svlObject2dFrame::iterator it = frame.begin(); it != frame.end(); it++) {
                vector<double> scores;
                classifier->getClassScores(data.features[it->index], scores);
                it->pr = -scores[0];
            }
            detections[baseNames[i]] = frame;
        }

        SVL_LOG(SVL_LOG_MESSAGE, "Writing detections to " << evalFilename << "...");
        detections.write(evalFilename);
    }

    // free memory
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

