/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneLearnObjectDetector.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for learning object detectors.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "scenePatchObjectDetector.h"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneLearnObjectDetector [OPTIONS] <posImgDir> <negImgDir>\n";
    cerr << "OPTIONS:\n"
         << "  -w <n>           :: base width (default 64)\n"
         << "  -h <n>           :: base height (default: 128)\n"
         << "  -o <filestem>    :: output filestem (appends .dict and .model)\n"
         << "  -maxDictSize <n> :: maximum size of dictionary (default: 500)\n"        
         << "  -useDictionary   :: use existing dictionary\n"
         << "  -maxSamples <n>  :: maximum number of pos/neg samples (default: 1000)\n"
         << "  -test <img>      :: tests object detector classes\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // set defaults and read commandline parameters
    svlConfigurationManager::get().configure("svlML.svlBoostedClassifier", "boostMethod", "DISCRETE");
    int width = 64;
    int height = 128;
    const char *outFileStem = NULL;
    bool bUseExistingDict = false;
    int maxDictSize = 500;
    int maxSamples = 1000;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_INT_OPTION("-w", width)
        SVL_CMDLINE_INT_OPTION("-h", height)
        SVL_CMDLINE_STR_OPTION("-o", outFileStem)
        SVL_CMDLINE_INT_OPTION("-maxDictSize", maxDictSize)
        SVL_CMDLINE_BOOL_OPTION("-useDictionary", bUseExistingDict)
        SVL_CMDLINE_INT_OPTION("-maxSamples", maxSamples)
        SVL_CMDLINE_OPTION_BEGIN("-test", p)
            IplImage *img = cvLoadImage(p[0], CV_LOAD_IMAGE_GRAYSCALE);
            ScenePatch::test(img);
            cvReleaseImage(&img);
            svlCodeProfiler::print(cerr);
            return 0;
        SVL_CMDLINE_OPTION_END(1)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 2) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *posList = SVL_CMDLINE_ARGV[0];
    const char *negList = SVL_CMDLINE_ARGV[1];

    // initialize detector
    ScenePatchObjectDetector detector(width, height);

    // learn dictionary
    if (bUseExistingDict) {
        SVL_LOG(SVL_LOG_MESSAGE, "loading existing dictionary...");
        SVL_ASSERT_MSG(outFileStem != NULL, "output filestem must be supplied");       
        string filename = string(outFileStem) + string(".dict");
        detector.readDictionary(filename.c_str());
    } else {
        SVL_LOG(SVL_LOG_MESSAGE, "learning dictionary...");
        detector.learnPatchDictionary(posList, maxDictSize);
        if (outFileStem != NULL) {
            string filename = string(outFileStem) + string(".full.dict");
            detector.writeDictionary(filename.c_str());
        }
    }

    // learn classifier
    SVL_LOG(SVL_LOG_MESSAGE, "learning classifier...");
    detector.learnClassifier(posList, negList, maxSamples);
    if (outFileStem != NULL) {
        string filename = string(outFileStem) + string(".full.model");
        detector.writeClassifier(filename.c_str());
    }

    // trim dictionary
    SVL_LOG(SVL_LOG_MESSAGE, "trimming patch dictionary...");
    detector.trimDictionary();
    if (outFileStem != NULL) {
        string filename = string(outFileStem) + string(".dict");
        detector.writeDictionary(filename.c_str());
        filename = string(outFileStem) + string(".model");
        detector.writeClassifier(filename.c_str());
    }

#if 0
    SVL_LOG(SVL_LOG_MESSAGE, "validating classifier...");
    detector.validateClassifier(posList, negList, maxSamples);
#endif    

    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

