/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneLearnPixelClassifiers.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for learning pixel classifiers based on texton filterbank features.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneModel.h"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneLearnPixelClassifiers [OPTIONS] <trainList>\n";
    cerr << "OPTIONS:\n"
         << "  -classId <n>      :: learn single class only (default: -1 (ALL))\n"
         << "  -rounds <num>     :: number of rounds of boosting (default: 100)\n"
         << "  -splits <num>     :: number of splits in weak learners (default: 2)\n"
         << "  -subSample <n>    :: subsample data by 1/<n>\n"
         << "  -prior <alpha>    :: dirchelet prior on +ve and -ve training sets\n"
         << "  -validate <list>  :: validation file list (default: none)\n"
         << SVL_STANDARD_OPTIONS_USAGE 
         << endl;
}

int main(int argc, char *argv[])
{
    // TODO: rename REGIONS to PIXELS
    int classId = -1;
    int numRounds = 1000;
    int numSplits = 2;    
    int subSample = 0;
    int pseudoCounts = 0;
    const char *validateList = NULL;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_INT_OPTION("-classId", classId)
        SVL_CMDLINE_INT_OPTION("-rounds", numRounds)
        SVL_CMDLINE_INT_OPTION("-splits", numSplits)
        SVL_CMDLINE_INT_OPTION("-subSample", subSample)
        SVL_CMDLINE_INT_OPTION("-prior", pseudoCounts)
        SVL_CMDLINE_STR_OPTION("-validate", validateList)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *trainList = SVL_CMDLINE_ARGV[0];
    
    // read list of training images
    SVL_LOG(SVL_LOG_MESSAGE, "Reading training list from " << trainList << "...");
    vector<string> baseNames = svlReadFile(trainList);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");

    vector<string> validateBaseNames;
    if (validateList != NULL) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading validation list from " << validateList << "...");
        validateBaseNames = svlReadFile(validateList);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << validateBaseNames.size() << " images");        
    }

    // train boosted classifier
    string outputDir = gSceneConfig.filebase("modelsDir", "");
    ScenePixelModel model;
    model.learnBoostedPixelClassModels(baseNames, numRounds, numSplits, subSample, 
        pseudoCounts, classId);            
    model.saveBoostedPixelClassModels(outputDir.c_str(), classId);

    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

