/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneLearnPixelModel.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for learning ScenePixelModel singleton and pairwise potentials.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneModel.h"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneLearnPixelModel [OPTIONS] <trainList>\n";
    cerr << "OPTIONS:\n"
         << "  -component <str>  :: model component to learn SINGLETON (default), CONTRAST\n"
         << "  -i <model>        :: input pixel models directory (default: same as output)\n"
         << "  -subSample <n>    :: subsample data by 1/<n>\n"
         << "  -maxIters <n>     :: maximum training iterations (default: 1000)\n"
         << SVL_STANDARD_OPTIONS_USAGE 
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // read commandline parameters
    const char *modelComponent = "SINGLETON";
    const char *inputDir = NULL;
    int subSample = 0;
    int maxIterations = 1000;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-i", inputDir)
        SVL_CMDLINE_STR_OPTION("-component", modelComponent)            
        SVL_CMDLINE_INT_OPTION("-subSample", subSample)
        SVL_CMDLINE_INT_OPTION("-maxIters", maxIterations)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *trainList = SVL_CMDLINE_ARGV[0];
    string outputModelDir = gSceneConfig.filebase("modelsDir", "");
    string inputModelDir = outputModelDir;
    if (inputDir != NULL) {
        inputModelDir = string(inputDir);
    }

    // read list of training images
    SVL_LOG(SVL_LOG_MESSAGE, "Reading training list from " << trainList << "...");
    vector<string> baseNames = svlReadFile(trainList);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");

    // learn pixel singletons
    if (!strcasecmp(modelComponent, "SINGLETON")) {
        ScenePixelModel model;
        model.SceneBaseModel::loadModels(inputModelDir.c_str());
        model.learnPixelSingletonModel(baseNames, maxIterations, subSample);
        model.savePixelSingletonModel(outputModelDir.c_str());
    }

    // learn pixel pairwise contrast weight
    if (!strcasecmp(modelComponent, "CONTRAST")) {
        ScenePixelModel model;
        model.SceneBaseModel::loadModels(inputModelDir.c_str());
        model.loadPixelSingletonModel(inputModelDir.c_str());
        model.learnPixelContrastWeight(baseNames);
        model.savePixelContrastWeight(outputModelDir.c_str());
    }

    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

