/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneLearnRegionModel.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for learning a SceneRegionModel or SceneObjectModel.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneObjectModel.h"
#include "sceneVisualization.h"

#define WINDOW_NAME "sceneLearnRegionModel"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneLearnRegionModel [OPTIONS] <trainList>\n";
    cerr << "OPTIONS:\n"
         << "  -m <model>        :: model type: REGION (default), REGIONPAIR, OBJECT,\n"
         << "                       OBJECTCONTEXT, OBJECTNEW\n"
         << "  -component <str>  :: model component to learn SINGLETON (default), HORIZON,\n"
         << "                       HORIZONPAIR, PAIRWISE\n"
         << "  -i <model>        :: input pixel models directory (default: -o)\n"
         << "  -o <model>        :: output model directory (default: use configXML)\n"
         << "  -subSample <n>    :: subsample data by 1/<n>\n"
         << "  -maxIters <n>     :: maximum training iterations (default: 1000)\n"
         << "  -resumeTraining   :: initialize training with last parameters\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // read commandline parameters
    const char *modelType = "REGION";
    const char *modelComponent = "SINGLETON";
    string inputModelDir;
    string outputModelDir;
    int subSample = 0;
    int maxIterations = 1000;
    bool bResumeTraining = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_OPTION_BEGIN("-o", p)
            outputModelDir = string(p[0]);
        SVL_CMDLINE_OPTION_END(1)
        SVL_CMDLINE_STR_OPTION("-component", modelComponent)
        SVL_CMDLINE_STR_OPTION("-m", modelType)
        SVL_CMDLINE_OPTION_BEGIN("-i", p)
            inputModelDir = string(p[0]);
        SVL_CMDLINE_OPTION_END(1)
        SVL_CMDLINE_INT_OPTION("-subSample", subSample)
        SVL_CMDLINE_INT_OPTION("-maxIters", maxIterations)
        SVL_CMDLINE_BOOL_OPTION("-resumeTraining", bResumeTraining)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *trainList = SVL_CMDLINE_ARGV[0];
    if (outputModelDir.empty()) {
        outputModelDir = gSceneConfig.filebase("modelsDir", "");
    }
    if (inputModelDir.empty()) {
        inputModelDir = outputModelDir;
    }

    // read list of training images
    SVL_LOG(SVL_LOG_MESSAGE, "Reading training list from " << trainList << "...");
    vector<string> baseNames = svlReadFile(trainList);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");

    // initialize model
    SceneRegionModel *model = NULL;
    if (!strcasecmp(modelType, "REGION")) {
        model = new SceneRegionModel();
    } else if (!strcasecmp(modelType, "REGIONPAIR")) {
        model = new SceneRegionPairModel();
    } else if (!strcasecmp(modelType, "OBJECT")) {
        model = new SceneObjectModel();
    } else if (!strcasecmp(modelType, "OBJECTNEW")) {
        model = new SceneObjectModelNew();
    } else if (!strcasecmp(modelType, "OBJECTCONTEXT")) {
        model = new SceneObjectContextModel();
    } else {
        SVL_LOG(SVL_LOG_FATAL, "unrecognized model type " << modelType);
    }
    SVL_ASSERT(model != NULL);

    model->ScenePixelModel::loadModels(inputModelDir.c_str());


    // learn horizon model
    if (!strcasecmp(modelComponent, "HORIZON")) {
        model->learnHorizonSingletonModel(baseNames);
        model->saveHorizonModel(outputModelDir.c_str());
    }

    // learn horizon model
    if (!strcasecmp(modelComponent, "HORIZONPAIR")) {
        model->learnHorizonPairwiseModel(baseNames);
        model->saveHorizonPairwiseModel(outputModelDir.c_str());
    }

    // learn region/object singleton
    if (!strcasecmp(modelComponent, "SINGLETON")) {
        if (!strcasecmp(modelType, "REGION")) {
            if (bResumeTraining) {
                model->loadRegionSingletonModel(inputModelDir.c_str());
            }            
            model->learnRegionSingletonModel(baseNames, maxIterations, bResumeTraining);
            model->saveRegionSingletonModel(outputModelDir.c_str());

        } else if (!strcasecmp(modelType, "OBJECT")) {
            ((SceneObjectModel *)model)->loadObjectDetectors(inputModelDir.c_str());
            ((SceneObjectModel *)model)->loadPedestrianClassifier(inputModelDir.c_str());
            ((SceneObjectModel *)model)->loadRegionSingletonModel(inputModelDir.c_str());
            if (bResumeTraining) {
                ((SceneObjectModel *)model)->loadObjectSingletonModel(inputModelDir.c_str());
            }
            ((SceneObjectModel *)model)->learnObjectSingletons(baseNames, maxIterations, bResumeTraining);
            ((SceneObjectModel *)model)->saveObjectSingletonModel(outputModelDir.c_str());

        } else if (!strcasecmp(modelType, "OBJECTNEW")) {
            if (bResumeTraining) {
                model->loadModels(inputModelDir.c_str());
                ((SceneObjectModelNew *)model)->updateObjectSingletons(baseNames, maxIterations);                
            } else {
                model->SceneRegionModel::loadModels(inputModelDir.c_str());
                ((SceneObjectModelNew *)model)->loadObjectDetectors(inputModelDir.c_str());
                ((SceneObjectModelNew *)model)->learnObjectSingletons(baseNames, maxIterations);
            }
            ((SceneObjectModelNew *)model)->saveObjectSingletonModel(outputModelDir.c_str());

        } else {
            SVL_LOG(SVL_LOG_FATAL, "model " << modelType << " learning not implemented yet");
        }
    }

    // learn pedestrian feature
    if (!strcasecmp(modelComponent, "PEDFEATURE")) {
        SVL_ASSERT(!strcasecmp(modelType, "OBJECT"));
        ((SceneObjectModel *)model)->loadObjectDetectors(inputModelDir.c_str());
        ((SceneObjectModel *)model)->loadRegionSingletonModel(inputModelDir.c_str());
        ((SceneObjectModel *)model)->learnPedestrianClassifier(baseNames);
        //((SceneObjectModel *)model)->savePedestrianClassifier(outputModelDir.c_str());            
    }
    
    // learn region pairwise/object context models
    if (!strcasecmp(modelComponent, "PAIRWISE")) {
        if (!strcasecmp(modelType, "REGIONPAIR")) {
            if (bResumeTraining) {
                ((SceneRegionPairModel *)model)->loadRegionPairwiseModel(inputModelDir.c_str());
            }            
            ((SceneRegionPairModel *)model)->learnRegionPairwiseModel(baseNames, maxIterations, bResumeTraining);
            ((SceneRegionPairModel *)model)->saveRegionPairwiseModel(outputModelDir.c_str());
        } else if (!strcasecmp(modelType, "OBJECTCONTEXT")) {
            if (bResumeTraining) {
                ((SceneObjectContextModel *)model)->loadObjectContextModel(inputModelDir.c_str());
            }            
            ((SceneObjectContextModel *)model)->learnObjectContextModel(baseNames, maxIterations, bResumeTraining);
            ((SceneObjectContextModel *)model)->saveObjectContextModel(outputModelDir.c_str());
        } else {
            SVL_LOG(SVL_LOG_FATAL, "model " << modelType << " learning not implemented yet");
        }
    }

    delete model;
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}
