/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneModel.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
** TODO:
**  1. allow pixel models to be loaded in the form: "pixel.<name>.model"
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "Eigen/Core"

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionFeatures.h"

#ifdef KOLMOGOROV_MAXFLOW
#include "../../external/maxflow-v3.0.src/graph.h"
#else
#error "Need Kolmogorov's maxflow-v3.0 installed (see external/installMaxFlow.pl)"
#endif

using namespace std;

// SceneBaseModel class ---------------------------------------------------

SceneBaseModel::SceneBaseModel() :
    _pixelClassModels(NULL)
{
    // do nothing
}

SceneBaseModel::~SceneBaseModel()
{
    if (_pixelClassModels != NULL)
        delete _pixelClassModels;
}

// i/o
void SceneBaseModel::loadModels(const char *modelDir)
{
    SVL_ASSERT(modelDir != NULL);
    loadBoostedPixelClassModels(modelDir);
}

void SceneBaseModel::loadBoostedPixelClassModels(const char* modelDir)
{
    if (_pixelClassModels != NULL) {
        delete _pixelClassModels;
    }
    _pixelClassModels = new svlBoostedClassifierSet(gSceneConfig.numRegionLabels);
    SVL_LOG(SVL_LOG_VERBOSE, "...reading pixel class models from " << modelDir);
    string filename = string(modelDir) + string("/pixel.class.xml");
    bool bSuccess = _pixelClassModels->load(filename.c_str());
    SVL_ASSERT(bSuccess);
}

void SceneBaseModel::saveBoostedPixelClassModels(const char* modelDir, int classId)
{
    SVL_ASSERT(_pixelClassModels != NULL);
    SVL_LOG(SVL_LOG_VERBOSE, "...writing pixel class models to " << modelDir);
    string filename = string(modelDir) + string("/pixel.class.xml");
    _pixelClassModels->save(filename.c_str(), ".model", classId);
}

// learning
void SceneBaseModel::learnBoostedPixelClassModels(vector<string>& baseNames, int numRounds,
    int numSplits, int subSample, int pseudoCounts, int classId)
{
    // accumulate training data
    vector<vector<double> > featureVectors;
    vector<int> featureLabels;
    buildBalancedTrainingSet(baseNames, "regionLblExt", gSceneConfig.numRegionLabels,
        featureVectors, featureLabels, subSample);

    // train classifiers
    if (_pixelClassModels != NULL) {
        delete _pixelClassModels;
    }

    // subsample pixels from negative set (saves memory)
    if ((classId != -1) && (gSceneConfig.getOptionAsBool("subSampleAllPixelClass"))) {
        SVL_LOG(SVL_LOG_WARNING, "subsampling pixel classes not equal to " << classId << "...");
        int posClassCount = 0;
        for (int i = 0; i < (int)featureLabels.size(); i++) {
            if (featureLabels[i] == classId)
                posClassCount += 1;
        }
        int subSampleRatio = (int)(featureLabels.size() - posClassCount) / (posClassCount + 1);
        if (subSampleRatio > 2) {        
            vector<vector<double> > newFeatures;
            vector<int> newLabels;
            newFeatures.reserve(2 * posClassCount);
            newLabels.reserve(2 * posClassCount);
            
            for (int i = 0; i < (int)featureLabels.size(); i++) {
                if ((featureLabels[i] != classId) && (rand() % subSampleRatio != 0))
                    continue;

                newFeatures.push_back(featureVectors[i]);
                newLabels.push_back(featureLabels[i]);
            }
            featureVectors = newFeatures;
            featureLabels = newLabels;
        }
        SVL_LOG(SVL_LOG_WARNING, "...done");
    }

    _pixelClassModels = new svlBoostedClassifierSet(gSceneConfig.numRegionLabels);
    _pixelClassModels->setOption("boostingRounds", numRounds);
    _pixelClassModels->setOption("numSplits", numSplits);
    _pixelClassModels->setOption("pseudoCounts", pseudoCounts);

    double avgRecall = _pixelClassModels->train(featureVectors, featureLabels, classId);
    SVL_LOG(SVL_LOG_MESSAGE, "...average recall: " << avgRecall);
}

// initialize instance (cache boosted features)
void SceneBaseModel::initializeInstance(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);
    cacheBoostedPixelResponses(*instance);
}

// inference: given pixels, infers horizon. returns energy.
double SceneBaseModel::inferHorizon(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);

    // default: model doesn't handle horizons
    instance->assignment.horizon = instance->height() / 2;

    return energy(instance);
}

// energy for complete assignment
double SceneBaseModel::energy(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);
    return 0.0;
}

// helper functions
void SceneBaseModel::cacheBoostedPixelResponses(SceneInstance &instance)
{
    SVL_ASSERT(_pixelClassModels != NULL);
    if (!instance.pixelBoostedResponses.empty())
        return;

    SVL_FCN_TIC;

    int nFeatures = _pixelClassModels->size();
    instance.pixelBoostedResponses.reserve(instance.size());
    bool bReadCacheSuccess = true;

    // check if instance is in the cache
    if (gSceneConfig.getOptionAsBool("useCache") && (!instance.name().empty())) {
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "featuresExt");
        if (svlFileExists(filename.c_str())) {
            svlCodeProfiler::tic(svlCodeProfiler::getHandle("readPixelFeatureCache"));
            vector<double> featureVector(nFeatures);
            ifstream ifs(filename.c_str(), ifstream::in | ifstream::binary);
            SVL_ASSERT(!ifs.fail());
            for (int i = 0; i < instance.size(); i++) {
                ifs.read((char *)&featureVector[0], nFeatures * sizeof(double));
                for (int j = 0; j < nFeatures; j++) {
                    SVL_ASSERT_MSG(isfinite(featureVector[j]), filename);
                }
                if (ifs.fail()) {
                    bReadCacheSuccess = false;
                    SVL_LOG(SVL_LOG_WARNING, "load pixel feature cache failed for " 
                        << filename << ". Re-computing pixel features...");
                    break;
                }
                //SVL_ASSERT(!ifs.fail());
                instance.pixelBoostedResponses.push_back(featureVector);
            }
            ifs.close();
            svlCodeProfiler::toc(svlCodeProfiler::getHandle("readPixelFeatureCache"));
            SVL_FCN_TOC;

            if (bReadCacheSuccess == true) {
                return;
            } else {
                instance.pixelBoostedResponses.clear();  
            }
        }
    }

    // not in cache (or cache not enabled), so compute
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("computePixelFeatures"));
    vector<vector<double> > rawFeatures;
    rawFeatures.reserve(instance.size());
    for (int y = 0; y < instance.height(); y++) {
        for (int x = 0; x < instance.width(); x++) {
            rawFeatures.push_back(instance.rawPixelFeatures(x, y));
        }
    }

    _pixelClassModels->getClassScores(rawFeatures, instance.pixelBoostedResponses);
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("computePixelFeatures"));

    // write to cache
    if (gSceneConfig.getOptionAsBool("useCache") && (!instance.name().empty())) {
        svlCodeProfiler::tic(svlCodeProfiler::getHandle("writePixelFeatureCache"));
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "featuresExt");
        ofstream ofs(filename.c_str(), ofstream::out | ofstream::binary);
        SVL_ASSERT(!ofs.fail());
        for (unsigned i = 0; i < instance.pixelBoostedResponses.size(); i++) {
            ofs.write((char *)&instance.pixelBoostedResponses[i][0],
                nFeatures * sizeof(double));
        }
        SVL_ASSERT(!ofs.fail());
        ofs.close();
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("writePixelFeatureCache"));
    }

    SVL_FCN_TOC;
}

// load training data
void SceneBaseModel::buildBalancedTrainingSet(const vector<string>& baseNames, const char *labelExt,
    int nLabels, vector<vector<double> >& featureVectors, vector<int>& featureLabels,
    int subSample) const
{
    // accumulate training data
    vector<int> classCounts(nLabels, 0);
    int maxClassCount = 0;

    SVL_LOG(SVL_LOG_MESSAGE, "Processing image features...");
    int hProcessImage = svlCodeProfiler::getHandle("processImage");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        svlCodeProfiler::tic(hProcessImage);
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        string lblFilename = gSceneConfig.filename("lblDir", baseNames[i], labelExt);

        // load scene
        SceneInstance scene(imgFilename.c_str());
        SVL_LOG(SVL_LOG_VERBOSE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*scene.image()));

        // load labels
        CvMat *labels = cvCreateMat(scene.height(), scene.width(), CV_32SC1);
        SVL_ASSERT(labels != NULL);
        bool bSuccess = readMatrix(labels, lblFilename.c_str());
        SVL_ASSERT_MSG(bSuccess, "Could not read labels for " << baseNames[i]);

        // accumulate "known" labels
        for (int y = 0; y < scene.height(); y++) {
            for (int x = 0; x < scene.width(); x++) {
                int lbl = CV_MAT_ELEM(*labels, int, y, x);
                if (lbl < 0) continue;
                SVL_ASSERT_MSG(lbl < nLabels, lbl << " !< " << nLabels);

                // random skip
                if ((subSample > 1) && (rand() % subSample != 0) &&
                    (classCounts[lbl] > maxClassCount - 500))
                    continue;

                // accumulate
                featureVectors.push_back(scene.rawPixelFeatures(x, y));
                featureLabels.push_back(lbl);
                classCounts[lbl] += 1;
                if (classCounts[lbl] > maxClassCount)
                    maxClassCount = classCounts[lbl];
            }

        }

        SVL_LOG(SVL_LOG_VERBOSE, "..." << featureVectors.size() << " ("
            << toString(classCounts) << ") samples accumulated");

        // free memory
        cvReleaseMat(&labels);
        svlCodeProfiler::toc(hProcessImage);
    }
}

// ScenePixelModel class --------------------------------------------------

ScenePixelModel::ScenePixelModel() : SceneBaseModel(),
    _pixelSingletons(gSceneConfig.numRegionLabels), _pixelPairwiseContrastWeight(1.0) {
    // do nothing
}

ScenePixelModel::~ScenePixelModel()
{
    // do nothing
}

// i/o
void ScenePixelModel::loadModels(const char *modelDir)
{
    SceneBaseModel::loadModels(modelDir);
    loadPixelSingletonModel(modelDir);
    loadPixelContrastWeight(modelDir);
}

void ScenePixelModel::loadPixelSingletonModel(const char* modelDir)
{
    string fileName;
    fileName = string(modelDir) + string("/pixel.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading pixel feature whitening parameters from " << fileName);
    _pixelFeatureWhitener.load(fileName.c_str());

    fileName = string(modelDir) + string("/pixel.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading pixel singleton parameters from " << fileName);
    _pixelSingletons.load(fileName.c_str());
}

void ScenePixelModel::savePixelSingletonModel(const char* modelDir)
{
    string fileName;
    fileName = string(modelDir) + string("/pixel.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing pixel feature whitening parameters to " << fileName);
    _pixelFeatureWhitener.save(fileName.c_str());

    fileName = string(modelDir) + string("/pixel.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing pixel singleton parameters to " << fileName);
    _pixelSingletons.save(fileName.c_str());
}

void ScenePixelModel::loadPixelContrastWeight(const char* modelDir)
{
    string fileName = string(modelDir) + string("/pixel.pairwise.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading pixel pairwise parameters from " << fileName);
    ifstream ifs(fileName.c_str());
    SVL_ASSERT(!ifs.fail());
    ifs >> _pixelPairwiseContrastWeight;
    ifs.close();
}

void ScenePixelModel::savePixelContrastWeight(const char* modelDir)
{
    string fileName = string(modelDir) + string("/pixel.pairwise.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing pixel pairwise parameters to " << fileName);
    ofstream ofs(fileName.c_str());
    SVL_ASSERT(!ofs.fail());
    ofs << _pixelPairwiseContrastWeight << "\n";
    ofs.close();
}

// learning
void ScenePixelModel::learnPixelSingletonModel(vector<string>& baseNames,
    int maxIterations, int subSample)
{
    SVL_ASSERT(_pixelClassModels != NULL);

    // accumulate training data (balanced on class label)
    bool bBalanceSingletons = gSceneConfig.getOptionAsBool("balancedSingletons");
    int nLabels = gSceneConfig.numRegionLabels;
    vector<int> classCounts(nLabels, 0);
    int maxClassCount = 0;
    vector<vector<double> > featureVectors;
    vector<int> featureLabels;

    SVL_LOG(SVL_LOG_MESSAGE, "Processing image features...");
    int hProcessImage = svlCodeProfiler::getHandle("processImage");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        svlCodeProfiler::tic(hProcessImage);
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        string regLblFilename = gSceneConfig.filename("lblDir", baseNames[i], "regionLblExt");

        // load scene
        SceneInstance scene(imgFilename.c_str());
        SVL_LOG(SVL_LOG_VERBOSE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*scene.image()));

        // load labels
        CvMat *regLabels = cvCreateMat(scene.height(), scene.width(), CV_32SC1);
        SVL_ASSERT(regLabels != NULL);

        bool bSuccess = readMatrix(regLabels, regLblFilename.c_str());
        SVL_ASSERT_MSG(bSuccess, "Could not read class labels for " << baseNames[i]);

        // load auxiliary features
        vector<CvMat *> auxFeatures = loadAuxiliaryFeatures(scene);

        // accumulate "known" labels
        for (int y = 0; y < scene.height(); y++) {
            for (int x = 0; x < scene.width(); x++) {
                int lbl = CV_MAT_ELEM(*regLabels, int, y, x);
                if (lbl < 0) continue;
                SVL_ASSERT_MSG(lbl < nLabels, lbl << " !< " << nLabels);

                // random skip
                if ((subSample > 1) && (rand() % subSample != 0) &&
                    (classCounts[lbl] > maxClassCount - 500))
                    continue;

                // accumulate
                vector<double> p;
                _pixelClassModels->getClassScores(scene.rawPixelFeatures(x, y), p);

                // append auxiliary features
                for (int i = 0; i < (int)auxFeatures.size(); i++) {
                    p.push_back(CV_MAT_ELEM(*auxFeatures[i], float, y, x));
                }

                // append bias
                p.push_back(1.0);

                featureVectors.push_back(p);
                featureLabels.push_back(lbl);

                // balanced sampling
                classCounts[lbl] += 1;
                if (bBalanceSingletons) {
                    if (classCounts[lbl] > maxClassCount)
                        maxClassCount = classCounts[lbl];
                }
            }
        }

        SVL_LOG(SVL_LOG_VERBOSE, "..." << featureVectors.size() << " ("
            << toString(classCounts) << ") samples accumulated");

        // free memory
        svlReleaseMatrices(auxFeatures);
        cvReleaseMat(&regLabels);
        svlCodeProfiler::toc(hProcessImage);
    }

    // train feature whitener
    _pixelFeatureWhitener.clear(featureVectors[0].size());
    if (gSceneConfig.getOptionAsBool("whitenPixelFeatures")) {
        _pixelFeatureWhitener.train(featureVectors);
    }
    _pixelFeatureWhitener.evaluate(featureVectors);

    // train model
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("trainPixelSingletons"));
    double lambda = gSceneConfig.getOptionAsDouble("regPixelSingleton");
    _pixelSingletons = svlMultiClassLogistic(featureVectors[0].size(), nLabels, lambda);
    _pixelSingletons.setOption("maxIterations", maxIterations);
    _pixelSingletons.train(featureVectors, featureLabels);
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("trainPixelSingletons"));

    // evaluate (on training data)
    if (svlLogger::getLogLevel() <= SVL_LOG_VERBOSE) {
        vector<int> predictedLabels;
        _pixelSingletons.getClassifications(featureVectors, predictedLabels);

        svlConfusionMatrix confusion(nLabels);
        confusion.accumulate(featureLabels, predictedLabels);
        confusion.write(cout);
        SVL_LOG(SVL_LOG_VERBOSE, "Accuracy: " << confusion.accuracy());
    }
}

void ScenePixelModel::learnPixelContrastWeight(vector<string>& baseNames)
{
    // cross-validate contrast weight
    map<double, double> scores;
    int delta = std::max(1, (int)baseNames.size() / 100);
    for (int i = 0; i < (int)baseNames.size(); i += delta) {
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");

        // load image
        IplImage *img = cvLoadImage(imgFilename.c_str(), CV_LOAD_IMAGE_COLOR);
        SVL_ASSERT(img != NULL);
        SVL_LOG(SVL_LOG_VERBOSE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*img));

        // load labels
        CvMat *labels = cvCreateMat(img->height, img->width, CV_32SC1);
        SVL_ASSERT(labels != NULL);
        string lblFilename = gSceneConfig.filename("lblDir", baseNames[i], "regionLblExt");
        bool bSuccess = readMatrix(labels, lblFilename.c_str());
        SVL_ASSERT_MSG(bSuccess, "Could not read class labels for " << baseNames[i]);

        // predict labels
        SceneInstance scene(img, baseNames[i].c_str());

        for (double w = 1.0; w < 129.0; w *= sqrt(2.0)) {
            _pixelPairwiseContrastWeight = w;
            inferPixels(&scene);

            CvMat *predictedLabels = scene.getPixelClasses();
            cvSub(labels, predictedLabels, predictedLabels);
            double errors = (double)cvCountNonZero(predictedLabels) / scene.size();
            if (scores.find(w) == scores.end()) {
                scores[w] = errors;
            } else {
                scores[w] += errors;
            }
            cvReleaseMat(&predictedLabels);
        }

        // free memory
        cvReleaseMat(&labels);
        cvReleaseImage(&img);

        for (map<double, double>::const_iterator it = scores.begin(); it != scores.end(); it++) {
            SVL_LOG(SVL_LOG_VERBOSE, "CONTRAST WEIGHT " << setw(5) << setprecision(3) << it->first 
                << "\t" << setprecision(5) << it->second);
        }
    }

    double bestScore = numeric_limits<double>::max();
    for (map<double, double>::const_iterator it = scores.begin(); it != scores.end(); it++) {
        if (it->second < bestScore) {
            _pixelPairwiseContrastWeight = it->first;
            bestScore = it->second;
        }
    }
}

// initialize instance (cache boosted features and pixel singletons)
void ScenePixelModel::initializeInstance(SceneInstance *instance)
{
    SceneBaseModel::initializeInstance(instance);
    cachePixelSingletons(*instance);
    instance->cachePixelContrast();
}

// inference
double ScenePixelModel::inferPixels(SceneInstance *instance)
{
    SVL_FCN_TIC;
    SVL_ASSERT(instance != NULL);

    // compute image features
    if (instance->pixelSingletons.empty()) {
        initializeInstance(instance);
    }

    // output folder  string outputFolder =  gSceneConfig.filebase("outputDir", "");
    //    FILE * pFile;
    //    string filename = instance->name().c_str();
    //    string outputFolderhead = "costoutput/";
    //    string outputFolder = outputFolderhead+filename+".txt";

    //    const char * str1 = const_cast<const char*>(outputFolder.c_str());
    //    pFile = fopen(str1, "w");
    // image base name can be got from instance->name()
    
    // loop through instance->pixelSingletons to get the dataterm
    // get the horizontal pairwise w (copy from alphaExpansion)
//     int* p = (int *)CV_MAT_ELEM_PTR(*instance->assignment.pixels, 0, 0);
//     for (int y = 0; y < instance->height(); y++) {
//         p++;
//         for (int x = 1; x < instance->width(); x++, p++) {
//             double w = _pixelPairwiseContrastWeight * instance->horizontalContrast(x, y);
//             instance->hpw.push_back(w);
// 	}
//     }

//     // get the vertical pairwise w (copy from alphaExpansion)
//     p = (int *)CV_MAT_ELEM_PTR(*instance->assignment.pixels, 0, 0);
//     for (int y = 0; y < instance->height(); y++) {
//         p++;
//         for (int x = 1; x < instance->width(); x++, p++) {
//             double w = _pixelPairwiseContrastWeight * instance->verticalContrast(x, y);
//             instance->vpw.push_back(w);
// 	}
//     }


    // run inference
    SVL_LOG(SVL_LOG_VERBOSE, "running inference on pixels...");
    alphaExpansion(instance->pixelSingletons, instance);


//     cout<<"printing singleton in file: "<<str1<<endl;
//     for(int i =0;i<instance->pixelSingletons.size();i++){
//         fprintf(pFile,"%d %f %f %f %f %f %f %f %f %f %f\n",i, instance->pixelSingletons[i][0],instance->pixelSingletons[i][1],instance->pixelSingletons[i][2],instance->pixelSingletons[i][3],instance->pixelSingletons[i][4],instance->pixelSingletons[i][5],instance->pixelSingletons[i][6],instance->pixelSingletons[i][7], instance->hpw[i],instance->vpw[i]);
//     }
//     fclose(pFile);

    CvMat *m = cvCloneMat(instance->assignment.pixels);
    instance->assignPixelsFromLabels(m);
    cvReleaseMat(&m);

    SVL_FCN_TOC;
    return energy(instance);
}

double ScenePixelModel::inferRegions(SceneInstance *instance)
{
    SVL_FCN_TIC;
    SVL_ASSERT(instance != NULL);

    // update region class and appearance models
    for (int regId = 0; regId < instance->numRegions(); regId++) {
        if (instance->assignment.regions[regId]->empty())
            continue;

        SVL_LOG(SVL_LOG_DEBUG, "inferring class/geometry label for region " << regId);

        // construct singleton potential
        instance->assignment.regions[regId]->Sk = 
            argmax(instance->assignment.regions[regId]->pixelScores);
    }

    SVL_FCN_TOC;
    return energy(instance);
}

double ScenePixelModel::energy(SceneInstance *instance)
{
    // TODO
    SVL_LOG(SVL_LOG_WARNING, "energy for pixel model not implemented yet");
    return 0.0;
}

// energy components
double ScenePixelModel::boundaryPenalty(const SceneInstance &instance) const
{
    double contrastPenalty = 0.0;

    const int *p = (int *)CV_MAT_ELEM_PTR(*instance.assignment.pixels, 0, 1);
    for (int x = 1; x < instance.width(); x++, p++) {
        if (*p != *(p - 1)) {
            contrastPenalty += instance.horizontalContrast(x, 0);
        }
    }

    const int *q = p - instance.width();
    for (int y = 1; y < instance.height(); y++) {
        if (*p++ != *q++) {
            contrastPenalty += instance.verticalContrast(0, y);
        }
        for (int x = 1; x < instance.width(); x++, p++, q++) {
            if (*p != *q) {
                contrastPenalty += instance.verticalContrast(x, y);
            }
            if (*p != *(p - 1)) {
                contrastPenalty += instance.horizontalContrast(x, y);
            }
        }
    }

    return _pixelPairwiseContrastWeight * contrastPenalty;
}

// scene features
void ScenePixelModel::cachePixelSingletons(SceneInstance &instance) const
{
    SVL_FCN_TIC;

    instance.pixelSingletons.clear();
    instance.pixelSingletons.reserve(instance.size());

    instance.pixelMarginals.clear();
    instance.pixelMarginals.reserve(instance.size());

    // load auxiliary features
    vector<CvMat *> auxFeatures = loadAuxiliaryFeatures(instance);

    int varIndx = 0;
    for (int y = 0; y < instance.height(); y++) {
        for (int x = 0; x < instance.width(); x++, varIndx++) {
            // boosted pixel features
            vector<double> f(instance.pixelBoostedResponses[varIndx]);

            // append auxliliary features
            for (int i = 0; i < (int)auxFeatures.size(); i++) {
                f.push_back(CV_MAT_ELEM(*auxFeatures[i], float, y, x));
            }

            // append bias
            f.push_back(1.0);

            // whiten features
            _pixelFeatureWhitener.evaluate(f);

            // compute marginal
            vector<double> marginal;
            _pixelSingletons.getMarginals(f, marginal);
            instance.pixelMarginals.push_back(marginal);

            for (int i = 0; i < (int)marginal.size(); i++) {
                marginal[i] = log(marginal[i] + SVL_DBL_MIN);
            }
            instance.pixelSingletons.push_back(marginal);
        }
    }

    // free auxiliary features, if any
    svlReleaseMatrices(auxFeatures);
    SVL_FCN_TOC;
}

vector<CvMat *> ScenePixelModel::loadAuxiliaryFeatures(SceneInstance &instance) const
{
    // caller must free auxFeatures by calling svlReleaseMatrices()
    vector<CvMat *> auxFeatures;
    if (gSceneConfig.getOptionAsBool("useAuxiliaryPixelFeatures")) {
        string fileStem = gSceneConfig.getOption("featuresDir") + instance.name();
        vector<string> fileExts;
        parseString(gSceneConfig.getOption("auxFeatureExt"), fileExts);

        auxFeatures = svlCreateMatrices(fileExts.size(), instance.height(),
            instance.width(), CV_32FC1);
        for (int i = 0; i < (int)fileExts.size(); i++) {
            SVL_LOG(SVL_LOG_VERBOSE, "...loading auxiliary features from " << fileStem
                << fileExts[i]);
            bool success = readMatrix(auxFeatures[i], (fileStem + fileExts[i]).c_str());
            SVL_ASSERT_MSG(success, fileStem << fileExts[i]);
        }
    }

    return auxFeatures;
}

// fast/memory-efficient inference
void ScenePixelModel::alphaExpansion(const vector<vector<double> >& phi, SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);
    SVL_ASSERT(phi.size() == (unsigned)instance->size());

    // check cardinality
    int nVariables = instance->size();
    int maxAlpha = (int)phi[0].size();
    for (int i = 1; i < nVariables; i++) {
        SVL_ASSERT(phi[i].size() == (unsigned)maxAlpha);
    }

    // initialize using singletons
    int varIndx = 0;
    for (int y = 0; y < instance->height(); y++) {
        for (int x = 0; x < instance->width(); x++, varIndx++) {
            CV_MAT_ELEM(*instance->assignment.pixels, int,  y, x) = argmax(phi[varIndx]);
        }
    }

    // return if no pairwise connections
    if (_pixelPairwiseContrastWeight == 0.0)
        return;

    int hConstructGraph = svlCodeProfiler::getHandle("alphaExpansion.constructGraph");
    int hMinCutIteration = svlCodeProfiler::getHandle("alphaExpansion.mincut");

    bool bChanged = true;
    int lastChanged = -1;
    double minEnergy = numeric_limits<double>::max();
    for (int nCycle = 0; bChanged; nCycle += 1) {
        bChanged = false;
	cout<<"nCycle: "<<nCycle<<endl;
        for (int alpha = 0; alpha < maxAlpha; alpha++) {
            if (alpha == lastChanged)
                break;
            svlCodeProfiler::tic(hConstructGraph);
            typedef Graph<double, double, double> GraphType;
            GraphType *g = new GraphType(nVariables, 8 * nVariables);
            g->add_node(nVariables);
            // add singleton terms
            varIndx = 0;
            int *p = (int *)CV_MAT_ELEM_PTR(*instance->assignment.pixels, 0, 0);
            for (int y = 0; y < instance->height(); y++) {
                for (int x = 0; x < instance->width(); x++, varIndx++, p++) {
                    g->add_tweights(varIndx, -phi[varIndx][*p], -phi[varIndx][alpha]);
                }
            }
            // add pairwise terms
            int numNonSubmodular = 0;
            // add horizontal pairwise terms
            p = (int *)CV_MAT_ELEM_PTR(*instance->assignment.pixels, 0, 0);
            vector<double>hw;
 	    for (int y = 0; y < instance->height(); y++) {
                p++;
                for (int x = 1; x < instance->width(); x++, p++) {
                    int u = instance->pixel2Indx(x, y);
                    int v = instance->pixel2Indx(x - 1, y);
                    double w = _pixelPairwiseContrastWeight * instance->horizontalContrast(x, y);
		    hw.push_back(w);	
                    double A = 0.0;
                    double B = *(p - 1) == alpha ? 0.0 : w;
                    double C = *p == alpha ? 0.0 : w;
                    double D = *(p - 1) == *p ? 0.0 : w;
                    // check for submodularity
                    if (A + D > C + B) {
                        // truncate non-submodular functions
                        numNonSubmodular += 1;
                        double delta = A + D - C - B;
                        A -= delta / 3 - SVL_EPSILON;
                        C += delta / 3 + SVL_EPSILON;
                        B = A + D - C + SVL_EPSILON;
                    }
                    g->add_tweights(u, D, A);
                    B -= A; C -= D;
                    SVL_ASSERT_MSG(B + C >= 0, "B = " << B << ", C = " << C);
                    if (B < 0) {
                        g->add_tweights(u, 0, B);
                        g->add_tweights(v, 0, -B);
                        g->add_edge(u, v, 0, B + C);
                    } else if (C < 0) {
                        g->add_tweights(u, 0, -C);
                        g->add_tweights(v, 0, C);
                        g->add_edge(u, v, B + C, 0);
                    } else {
                        g->add_edge(u, v, B, C);
                    }
                }
            }
	    instance->hpw=hw;
            // add vertical pairwise terms
            p = (int *)CV_MAT_ELEM_PTR(*instance->assignment.pixels, 1, 0);
            vector<double> vw;
      	    for (int y = 1; y < instance->height(); y++) {
                for (int x = 0; x < instance->width(); x++, p++) {
                    int u = instance->pixel2Indx(x, y);
                    int v = instance->pixel2Indx(x, y - 1);

                    double w = _pixelPairwiseContrastWeight * instance->verticalContrast(x, y);
		    vw.push_back(w);
                    double A = 0.0;
                    double B = *(p - instance->width()) == alpha ? 0.0 : w;
                    double C = *p == alpha ? 0.0 : w;
                    double D = *(p - instance->width()) == *p ? 0.0 : w;

                    // check for submodularity
                    if (A + D > C + B) {
                        // truncate non-submodular functions
                        numNonSubmodular += 1;
                        double delta = A + D - C - B;
                        A -= delta / 3 - SVL_EPSILON;
                        C += delta / 3 + SVL_EPSILON;
                        B = A + D - C + SVL_EPSILON;
                    }

                    g->add_tweights(u, D, A);

                    B -= A; C -= D;
                    SVL_ASSERT_MSG(B + C >= 0, "B = " << B << ", C = " << C);
                    if (B < 0) {
                        g->add_tweights(u, 0, B);
                        g->add_tweights(v, 0, -B);
                        g->add_edge(u, v, 0, B + C);
                    } else if (C < 0) {
                        g->add_tweights(u, 0, -C);
                        g->add_tweights(v, 0, C);
                        g->add_edge(u, v, B + C, 0);
                    } else {
                        g->add_edge(u, v, B, C);
                    }
                }
            }
	    instance->vpw=vw;
		
            if (numNonSubmodular > 0) {
                SVL_LOG(SVL_LOG_WARNING, numNonSubmodular << " non-submodular potentials");
            }
            svlCodeProfiler::toc(hConstructGraph);
            svlCodeProfiler::tic(hMinCutIteration);
            double e = g->maxflow();

            SVL_LOG(SVL_LOG_DEBUG, "...cycle " << nCycle << ", iteration "
                << alpha << " has energy " << e);
            if (e < minEnergy) {
                minEnergy = e;
                lastChanged = alpha;
                bChanged = true;

                varIndx = 0;
                int *p = (int *)CV_MAT_ELEM_PTR(*instance->assignment.pixels, 0, 0);
                for (int y = 0; y < instance->height(); y++) {
                    for (int x = 0; x < instance->width(); x++, varIndx++, p++) {
                        if (g->what_segment(varIndx) == GraphType::SOURCE) {
                            *p = alpha;
                        }
                    }
                }
            }

            delete g;
            svlCodeProfiler::toc(hMinCutIteration);
        }
    }
}

