/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneModel.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  This file contains common functionality for region-based scene understanding
**  models using texton filterbank based features. SceneRegionModel and
**  SceneRegionPairModel are defined in sceneRegionModel.h. SceneObjectModel and
**  SceneObjectContextModel are defined in sceneObjectModel.h.
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"
#include "sceneEntities.h"
#include "sceneInstance.h"

using namespace std;

// SceneBaseModel class ---------------------------------------------------
// Implements base functionality for learning pixel classifiers. Provides
// abstract interface for higher-level models.

class SceneBaseModel {
 protected:
    // pixelwise singleton features
    svlBoostedClassifierSet *_pixelClassModels;

 public:
    SceneBaseModel();
    virtual ~SceneBaseModel();

    // i/o
    virtual void loadModels(const char *modelDir);

    void loadBoostedPixelClassModels(const char* modelDir);
    void saveBoostedPixelClassModels(const char* modelDir, int classId = -1);

    // learning
    void learnBoostedPixelClassModels(vector<string>& baseNames, int numRounds = 1000,
        int numSplits = 2, int subSample = 0, int pseudoCounts = 0, int classId = -1);

    // initialize instance (cache boosted features)
    virtual void initializeInstance(SceneInstance *instance);

    // first-stage inference: initial assignment of pixels to connected
    // regions and populates instance->Ri. returns energy.
    virtual double inferPixels(SceneInstance *instance) = 0;

    // second-stage inference: given pixels, infers horizon. returns energy.
    virtual double inferHorizon(SceneInstance *instance);

    // third-stage inference: given regions (instance->Ri), infers region and
    // object class labels. returns energy.
    virtual double inferRegions(SceneInstance* instance) = 0;
    
    // energy for complete assignment
    virtual double energy(SceneInstance *instance);

 protected:
    // scene features
    void cacheBoostedPixelResponses(SceneInstance &instance);

    // dataset creation for training
    void buildBalancedTrainingSet(const vector<string>& baseNames, const char *labelExt,
        int nLabels, vector<vector<double> >& featureVectors, vector<int>& featureLabels, 
        int subSample) const;
};

// ScenePixelModel class --------------------------------------------------
// Implements baseline pixel CRF model with pairwise smoothness potential.

class ScenePixelModel : public SceneBaseModel {
 protected:
    // pixelwise singleton and pairwise potentials
    svlFeatureWhitener _pixelFeatureWhitener;
    svlMultiClassLogistic _pixelSingletons;
    double _pixelPairwiseContrastWeight;

 public:
    ScenePixelModel();
    ~ScenePixelModel();

    // i/o
    void loadModels(const char *modelDir);

    void loadPixelSingletonModel(const char* modelDir);
    void savePixelSingletonModel(const char* modelDir);
    void loadPixelContrastWeight(const char* modelDir);
    void savePixelContrastWeight(const char* modelDir);

    // learning
    void learnPixelSingletonModel(vector<string>& baseNames,
        int maxIterations = 1000, int subSample = 0);
    void learnPixelContrastWeight(vector<string>& baseNames);
    
    // initialize instance (cache boosted features and pixel singletons)
    void initializeInstance(SceneInstance *instance);

    // inference
    double inferPixels(SceneInstance *instance);
    double inferRegions(SceneInstance *instance);

    // energy for complete assignment
    double energy(SceneInstance *instance);

 protected:
    // energy components
    virtual double boundaryPenalty(const SceneInstance &instance) const;

    // scene features
    void cachePixelSingletons(SceneInstance &instance) const;
    vector<CvMat *> loadAuxiliaryFeatures(SceneInstance &instance) const;

    // fast/memory-efficient inference
    void alphaExpansion(const vector<vector<double> >& singletons, SceneInstance *instance);
};
