/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneObjDetDict.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <map>
#include <set>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneObjDetDict.h"

#define DICTIONARY_WINDOW_NAME "ObjectDetectionDictionary"

using namespace std;

// SceneObjDetDictEntry ------------------------------------------------------

SceneObjDetDictEntry::SceneObjDetDictEntry() :
    originalScore(-SVL_DBL_MAX), score(-SVL_DBL_MAX)
{
    originalExtent = cvRect(0, 0, 0, 0);
    extent = cvRect(0, 0, 0, 0);
}

SceneObjDetDictEntry::SceneObjDetDictEntry(const SceneObjDetDictEntry& e) :
    originalScore(e.originalScore), move(e.move), score(e.score)
{
    originalExtent = e.originalExtent;
    extent = e.extent;
}

SceneObjDetDictEntry::~SceneObjDetDictEntry()
{
    // do nothing
}

void SceneObjDetDictEntry::read(istream& is)
{
    int nVersion;
    is.read((char *)&nVersion, sizeof(int));
    SVL_ASSERT(nVersion == 1);

    is.read((char *)&originalExtent.x, sizeof(int));
    is.read((char *)&originalExtent.y, sizeof(int));
    is.read((char *)&originalExtent.width, sizeof(int));
    is.read((char *)&originalExtent.height, sizeof(int));

    is.read((char *)&originalScore, sizeof(double));

    unsigned codeLength;
    is.read((char *)&codeLength, sizeof(unsigned));
    move.resize(codeLength);
    is.read((char *)&move[0], codeLength * sizeof(unsigned char));

    is.read((char *)&extent.x, sizeof(int));
    is.read((char *)&extent.y, sizeof(int));
    is.read((char *)&extent.width, sizeof(int));
    is.read((char *)&extent.height, sizeof(int));

    is.read((char *)&score, sizeof(double));
    SVL_ASSERT(!is.fail());
}

void SceneObjDetDictEntry::write(ostream& os) const
{
    const int nVersion = 1;
    os.write((char *)&nVersion, sizeof(int));

    os.write((char *)&originalExtent.x, sizeof(int));
    os.write((char *)&originalExtent.y, sizeof(int));
    os.write((char *)&originalExtent.width, sizeof(int));
    os.write((char *)&originalExtent.height, sizeof(int));

    os.write((char *)&originalScore, sizeof(double));

    unsigned codeLength = move.size();
    os.write((char *)&codeLength, sizeof(unsigned));
    os.write((char *)&move[0], codeLength * sizeof(unsigned char));

    os.write((char *)&extent.x, sizeof(int));
    os.write((char *)&extent.y, sizeof(int));
    os.write((char *)&extent.width, sizeof(int));
    os.write((char *)&extent.height, sizeof(int));

    os.write((char *)&score, sizeof(double));
    SVL_ASSERT(!os.fail());
}

void SceneObjDetDictEntry::decode(CvMat *mask) const
{
    SVL_ASSERT((mask != NULL) && (cvGetElemType(mask) == CV_32SC1));
    SVL_ASSERT(!move.empty());

    int count = 0;
    int *p = (int *)CV_MAT_ELEM_PTR(*mask, 0, 0);
    for (int i = 0; i < (int)move.size(); i++) {
        count += move[i];
        for (int j = 0; j < (int)move[i]; j++) {
            *p++ = (i % 2);
        }
    }

    SVL_ASSERT_MSG(count == mask->rows * mask->cols, "diff = " << (count - mask->rows * mask->cols));
}

void SceneObjDetDictEntry::encode(const CvMat *mask, int segId)
{
    SVL_ASSERT((mask != NULL) && (cvGetElemType(mask) == CV_32SC1));

    move.clear();
    move.reserve(mask->rows);
    extent = cvRect(mask->cols, mask->rows, 0, 0);

    const int *p = (const int *)CV_MAT_ELEM_PTR(*mask, 0, 0);
    int indx = 0; int totalCount = 0;
    while (indx < mask->rows * mask->cols) {
        int count = 0;
        if (move.size() % 2 == 0) {
            while ((count < 255) && (p[indx] != segId)) {
                count += 1;
                indx += 1;
                if (indx == mask->rows * mask->cols)
                    break;
            }
        } else {
            while ((count < 255) && (p[indx] == segId)) {
                int x = indx % mask->cols;
                int y = indx / mask->cols;
                extent.x = std::min(x, extent.x);
                extent.y = std::min(y, extent.y);
                extent.width = std::max(x, extent.width);
                extent.height = std::max(y, extent.height);

                count += 1;
                indx += 1;
                if (indx == mask->rows * mask->cols)
                    break;
            }
        }

        move.push_back((unsigned char)count);
        totalCount += count;
    }

    SVL_ASSERT_MSG(totalCount == mask->rows * mask->cols,
        totalCount << " != " << mask->rows * mask->cols);

    extent.width -= extent.x - 1;
    extent.height -= extent.y - 1;

    SVL_ASSERT_MSG((extent.x + extent.width <= mask->cols) && (extent.y + extent.height <= mask->rows),
        toString(extent) << " exceeds mask size " << mask->rows << "-by-" << mask->cols);

    if ((extent.width < 0) || (extent.height < 0)) {
        extent = cvRect(0, 0, 0, 0);
    }
}

SceneObjDetDictEntry& SceneObjDetDictEntry::operator=(const SceneObjDetDictEntry& e)
{
    if (this != &e) {
        originalExtent = e.originalExtent;
        originalScore = e.originalScore;
        move = e.move;
        extent = e.extent;
        score = e.score;
    }

    return *this;
}

bool SceneObjDetDictEntry::operator<(const SceneObjDetDictEntry& e) const
{
    return (originalScore < e.originalScore);
}

// SceneObjDetDict -----------------------------------------------------------

SceneObjDetDict::SceneObjDetDict(const char *filename) :
    _width(0), _height(0)
{
    if (filename != NULL) {
        load(filename);
    }
}

SceneObjDetDict::SceneObjDetDict(const SceneObjDetDict& d) :
    _width(d._width), _height(d._height), _moves(d._moves) {
    // do nothing
}

SceneObjDetDict::~SceneObjDetDict()
{
    clear();
}

CvMat *SceneObjDetDict::mask(int i) const
{
    CvMat *m = cvCreateMat(_height, _width, CV_32SC1);
    _moves[i].decode(m);
    return m;
}

// i/o
void SceneObjDetDict::clear() {
    _width = 0;
    _height = 0;
    _moves.clear();
}

bool SceneObjDetDict::load(const char *filename)
{
    SVL_ASSERT(filename != NULL);
    clear();

    SVL_LOG(SVL_LOG_DEBUG, "loading object detection dictionary from " << filename);

    // read moves
    ifstream *ifs = new ifstream(filename, ios::binary);
    SVL_ASSERT_MSG(!ifs->fail(), filename);

    ifs->read((char *)&_width, sizeof(int));
    ifs->read((char *)&_height, sizeof(int));
    SVL_ASSERT(!ifs->fail() && (_width * _height >= 0));

    int nMoves;
    ifs->read((char *)&nMoves, sizeof(int));
    SVL_ASSERT(nMoves >= 0);

    SVL_LOG(SVL_LOG_DEBUG, "...loading " << nMoves << " of size " << _width << "x" << _height);
    _moves.resize(nMoves);
    for (int i = 0; i < nMoves; i++) {
        _moves[i].read(*ifs);
    }

    SVL_ASSERT(!ifs->fail());
    ifs->close();
    delete ifs;

    return true;
}

bool SceneObjDetDict::save(const char *filename) const
{
    SVL_ASSERT(filename != NULL);

    ofstream *ofs = new ofstream(filename, ios::binary);
    SVL_ASSERT(!ofs->fail());

    ofs->write((char *)&_width, sizeof(int));
    ofs->write((char *)&_height, sizeof(int));

    int nMoves = (int)_moves.size();
    ofs->write((char *)&nMoves, sizeof(int));

    for (int i = 0; i < nMoves; i++) {
        _moves[i].write(*ofs);
    }

    SVL_ASSERT(!ofs->fail());
    ofs->close();
    delete ofs;

    return true;
}

// create from detections (in sorted order)
void SceneObjDetDict::create(int w, int h, const svlObject2dFrame& detections)
{
    SVL_ASSERT((w > 0) && (h > 0));

    clear();
    _width = w;
    _height = h;
    _moves.reserve(detections.size());

    CvMat *mask = cvCreateMat(_height, _width, CV_32SC1);

    // add moves as bounding boxes
    svlObject2dFrame sortedDetections = sortObjects(detections);
    for (svlObject2dFrame::const_iterator it = sortedDetections.begin();
         it != sortedDetections.end(); it++) {

        SVL_LOG(SVL_LOG_DEBUG, "...extracting object <" << it->x << ", "
            << it->y << ", " << it->w << ", " << it->h << ">");

        cvZero(mask);
        for (int i = 0; i < (int)it->h; i++) {
            int *p = (int *)CV_MAT_ELEM_PTR(*mask, (int)(it->y + i), (int)it->x);
            for (int j = 0; j < (int)it->w; j++) {
                p[j] = 1;
            }
        }

        _moves.push_back(SceneObjDetDictEntry());

        _moves.back().originalExtent = cvRect((int)it->x, (int)it->y, (int)it->w, (int)it->h);
        _moves.back().originalScore = it->pr;
        _moves.back().encode(mask, 1);
    }

    cvReleaseMat(&mask);
}

void SceneObjDetDict::create(int w, int h, const char *detections, const char *frameId)
{
    SVL_ASSERT((w > 0) && (h > 0) && (detections != NULL) && (frameId != NULL));

    svlObject2dSequence objects;
    bool success = objects.read(detections);
    SVL_ASSERT_MSG(success, detections);

    create(w, h, objects[string(frameId)]);
}

// access
int SceneObjDetDict::find(const CvRect& extent) const
{
    int moveId = (int)_moves.size();
    while (moveId-- > 0) {
        if (_moves[moveId].extent == extent)
            break;
    }

    return moveId;
}

const SceneObjDetDictEntry& SceneObjDetDict::operator[](unsigned int moveId) const
{
    return _moves[moveId];
}

SceneObjDetDictEntry& SceneObjDetDict::operator[](unsigned int moveId)
{
    return _moves[moveId];
}
