/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneObjDetDict.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Dictionary of proposal moves for objects derived from a bounding-box-based
**  object detector (e.g., sliding-window).
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <map>
#include <set>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

#include "sceneConfiguration.h"

using namespace std;

// SceneObjDetDictEntry ------------------------------------------------------

class SceneObjDetDictEntry {
 public:
    CvRect originalExtent;                  // original detection extent
    double originalScore;                   // original detection score
    vector<unsigned char> move;             // run-length encoded move
    CvRect extent;                          // bounding box for move
    double score;                           // cached score for move

 public:
    SceneObjDetDictEntry();
    SceneObjDetDictEntry(const SceneObjDetDictEntry& e);
    ~SceneObjDetDictEntry();

    void read(istream& is);
    void write(ostream& os) const;

    void decode(CvMat *mask) const;                // expand move into mask
    void encode(const CvMat *mask, int segId = 1); // extract move from mask

    SceneObjDetDictEntry& operator=(const SceneObjDetDictEntry& e);
    bool operator<(const SceneObjDetDictEntry& e) const;
};

// SceneObjDetDict -----------------------------------------------------------

class SceneObjDetDict {
 protected:
    int _width;                               // segmentation size
    int _height;
    vector<SceneObjDetDictEntry> _moves;      // proposal moves

 public:
    SceneObjDetDict(const char *filename = NULL);
    SceneObjDetDict(const SceneObjDetDict& d);
    virtual ~SceneObjDetDict();

    inline int numMoves() const { return (int)_moves.size(); }
    inline int width() const { return _width; }
    inline int height() const { return _height; }

    CvMat *mask(int i) const;

    // clear dictionary
    void clear();

    // save and load from filename
    bool load(const char *filename);
    bool save(const char *filename) const;

    // create from detections (in sorted order)
    void create(int w, int h, const svlObject2dFrame& detections);
    void create(int w, int h, const char *detections, const char *frameId);
    
    // access
    int find(const CvRect& extent) const;
    
    const SceneObjDetDictEntry& operator[](unsigned int moveId) const;
    SceneObjDetDictEntry& operator[](unsigned int moveId);
};
