/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneObjectDiscovery.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Computes region features over foreground regions and ...
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneVisualization.h"

#define WINDOW_NAME "sceneObjectDiscovery"

using namespace std;

// prototypes ---------------------------------------------------------------

VectorXd regionBoundaryHistogramFeatures(vector<CvMat *> &orientedEdgeMaps,
    Vector4i extent, CvMat *mask = NULL, int nGrid = 4);
void visualizeRegionOrientedEdgeHistogram(VectorXd &f, Vector4i extent,
    IplImage *canvas, int nGrid = 4);
IplImage *visualizeEdgeMaps(const vector<CvMat *> &orientedEdgeMaps);

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneObjectDiscovery [OPTIONS] (<trainList> | <baseName>) (<outputStem>)?\n";
    cerr << "OPTIONS:\n"
         << "  -inDir <dir>      :: directory for input layers (default: <sceneConfig.outputDir>)\n"
         << "  -inRegions <ext>  :: extension for region id input (default: .predictedRegions.txt)\n"
         << "  -inLabels <ext>   :: extention for region class label input (default: .predictedClasses.txt)\n"
         << "  -outDir <dir>     :: directory for output images and features (default: none)\n"
         << "  -x                :: visualize features\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    // read commandline parameters
    captureOpenCVerrors();

    const char *inDir = NULL;
    const char *inRegions = ".predictedRegions.txt";
    const char *inLabels = ".predictedClasses.txt";
    const char *outDir = NULL;
    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-inDir", inDir)
        SVL_CMDLINE_STR_OPTION("-inRegions", inRegions)
        SVL_CMDLINE_STR_OPTION("-inLabels", inLabels)
        SVL_CMDLINE_STR_OPTION("-outDir", outDir)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if ((SVL_CMDLINE_ARGC < 1) || (SVL_CMDLINE_ARGC > 2)) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *imgList = SVL_CMDLINE_ARGV[0];
    const char *outFileStem = (SVL_CMDLINE_ARGC < 2) ? NULL : SVL_CMDLINE_ARGV[1];

    // read list of images
    vector<string> baseNames;
    if (svlFileExists(imgList)) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading image list from " << imgList << "...");
        baseNames = svlReadFile(imgList);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");
    } else {
        baseNames.push_back(string(imgList));
    }

    if (bVisualize) {
        cvNamedWindow(WINDOW_NAME, 1);
    }

    // process images
    int objectCount = 0;
    vector<vector<double> > shapeFeatures;
    vector<vector<double> > appearanceFeatures;
    vector<pair<string, int> > labels;
    SVL_LOG(SVL_LOG_MESSAGE, "Processing images...");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        // load image
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        IplImage *img = cvLoadImage(imgFilename.c_str(), CV_LOAD_IMAGE_COLOR);
        SVL_ASSERT_MSG(img != NULL, imgFilename.c_str());
        SVL_LOG(SVL_LOG_VERBOSE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*img));

        // TODO: loading groundtruth for now
        SceneInstance scene(img, baseNames[i].c_str());
        scene.loadGroundTruth();

        if (bVisualize) {
            IplImage *canvas = cvCloneImage(img);
            visualizeClasses(&scene, canvas, 0.5);
            visualizeBoundaries(&scene, canvas);
            //IplImage *canvas = visualize(&scene);
            cvShowImage(WINDOW_NAME, canvas);
            cvWaitKey(10);
            cvReleaseImage(&canvas);
        }

        // load pb oriented edge maps
        string pbEdgeMapFilename = string("/afs/cs/group/dags/scratch/sceneDatasets/mammals/features/") +
            baseNames[i] + string(".gPb.bin");
        vector<CvMat *> pbEdgeMaps = SceneDataLoader::loadGlobalPbFeatures(pbEdgeMapFilename.c_str(),
            scene.width(), scene.height());

#if 1
        if (bVisualize && (svlLogger::getLogLevel() >= SVL_LOG_DEBUG)) {
            cvNamedWindow("debug.edgemap", 1);
            IplImage *canvas = visualizeEdgeMaps(pbEdgeMaps);        
            cvShowImage("debug.edgemap", canvas);
            cvWaitKey(100);
            cvReleaseImage(&canvas);
        }
#endif

        // loop through foreground regions
        CvMat *regionMask = cvCreateMat(scene.height(), scene.width(), CV_8UC1);
        for (int regId = 0; regId < scene.numRegions(); regId++) {
            SceneRegion *region = scene.assignment.regions[regId];
            if (region->Sk != gSceneConfig.numRegionLabels - 1)
                continue;
            if (region->size() < 1024)
                continue;

            // mask for pixels in region
            cvCmpS(scene.assignment.pixels, regId, regionMask, CV_CMP_EQ);
           
            // output image
            if (outDir != NULL) {
                IplImage *canvas = cvCreateImage(cvSize(scene.width(), scene.height()), IPL_DEPTH_8U, 3);
                cvSet(canvas, cvScalar(255, 255, 255));
                cvCopy(scene.image(), canvas, regionMask);
                string filename = string(outDir) + string("/object") + 
                    padString(toString(objectCount), 5) + string(".jpg");
                cvSaveImage(filename.c_str(), canvas);
                cvReleaseImage(&canvas);
            }

            // TODO
            VectorXd f = regionBoundaryHistogramFeatures(pbEdgeMaps, region->extent(), regionMask);
            SVL_LOG(SVL_LOG_DEBUG, "f_shp(" << regId << ") = " << f.transpose());

            shapeFeatures.push_back(vector<double>(f.size()));
            Eigen::Map<VectorXd>(&shapeFeatures.back()[0], f.size()) = f;

            f = VectorXd(3 * region->Ak.size());
            f << region->appearance(), region->appearanceXMoment(), region->appearanceYMoment();
            SVL_LOG(SVL_LOG_DEBUG, "f_app(" << regId << ") = " << f.transpose());            
            appearanceFeatures.push_back(vector<double>(f.size()));
            Eigen::Map<VectorXd>(&appearanceFeatures.back()[0], f.size()) = f;

            labels.push_back(make_pair(baseNames[i], regId));

            if (bVisualize) {
                IplImage *canvas = cvCloneImage(img);
                visualizeRegionOrientedEdgeHistogram(f, region->extent(), canvas);
                cvShowImage(WINDOW_NAME, canvas);
                cvWaitKey(100);
                cvReleaseImage(&canvas);
            }

            // update object count
            objectCount += 1;
            SVL_LOG(SVL_LOG_VERBOSE, objectCount << " objects accumulated");
        }

        // free memory
        cvReleaseMat(&regionMask);
        svlReleaseMatrices(pbEdgeMaps);
        cvReleaseImage(&img);
    }
   
    if (bVisualize && (baseNames.size() == 1)) {
        cvWaitKey(-1);
    }

    // save features and labels
    if (outFileStem != NULL) {
        // shape
        string filename = string(outFileStem) + string(".shape.txt");
        ofstream ofs(filename.c_str());
        SVL_ASSERT_MSG(!ofs.fail(), filename.c_str());
        for (int i = 0; i < (int)shapeFeatures.size(); i++) {
            ofs << toString(shapeFeatures[i]) << "\n";
        }
        ofs.close();

        // appearance
        filename = string(outFileStem) + string(".appearance.txt");
        ofs.open(filename.c_str());
        SVL_ASSERT_MSG(!ofs.fail(), filename.c_str());
        for (int i = 0; i < (int)appearanceFeatures.size(); i++) {
            ofs << toString(appearanceFeatures[i]) << "\n";
        }
        ofs.close();

        // labels
        filename = string(outFileStem) + string(".labels.txt");
        ofs.open(filename.c_str());
        SVL_ASSERT_MSG(!ofs.fail(), filename.c_str());
        for (int i = 0; i < (int)labels.size(); i++) {
            ofs << labels[i].first << " " << labels[i].second << "\n";
        }
        ofs.close();
    }

    // shutdown
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

// helper functions ------------------------------------------------------

VectorXd regionBoundaryHistogramFeatures(vector<CvMat *> &orientedEdgeMaps,
    Vector4i extent, CvMat *mask, int nGrid)
{
    SVL_ASSERT((nGrid > 0) && (extent[2] > 0) && (extent[3] > 0));
    SVL_ASSERT((mask == NULL) || (cvGetElemType(mask) == CV_8UC1));

    VectorXd featureSum(nGrid * nGrid * orientedEdgeMaps.size());
    VectorXd featureCount(nGrid * nGrid * orientedEdgeMaps.size());
    featureSum.setZero();
    featureCount.setConstant(SVL_DBL_MIN);

    int dx = (extent[2] + nGrid - 1) / nGrid;
    int dy = (extent[3] + nGrid - 1) / nGrid;

    for (int c = 0; c < (int)orientedEdgeMaps.size(); c++) {
        SVL_ASSERT((orientedEdgeMaps[c] != NULL) && (orientedEdgeMaps[c]->cols >= extent[0] + extent[2]) &&
            (orientedEdgeMaps[c]->rows >= extent[1] + extent[3]));
        int baseIndx = c * nGrid * nGrid;
        if (mask == NULL) {
            for (int y = 0; y < extent[3]; y++) {
                const float *p = (const float *)CV_MAT_ELEM_PTR(*orientedEdgeMaps[c], extent[1] + y, extent[0]);
                for (int x = 0; x < extent[2]; x++, p++) {
                    featureSum[baseIndx + (y / dy) * nGrid + x / dx] += (double)((*p) * (*p));
                    featureCount[baseIndx + (y / dy) * nGrid + x / dx] += 1.0;
                }
            }
        } else {
            for (int y = 0; y < extent[3]; y++) {
                const float *p = (const float *)CV_MAT_ELEM_PTR(*orientedEdgeMaps[c], extent[1] + y, extent[0]);
                const unsigned char *q = (const unsigned char *)CV_MAT_ELEM_PTR(*mask, extent[1] + y, extent[0]);
                for (int x = 0; x < extent[2]; x++, p++, q++) {
                    if (*q == 0x00) continue;
                    featureSum[baseIndx + (y / dy) * nGrid + x / dx] += (double)((*p) * (*p));
                    featureCount[baseIndx + (y / dy) * nGrid + x / dx] += 1.0;
                }
            }
        }
    }

    VectorXd f((featureSum.cwise() / featureCount).cwise().sqrt());
    f.cwise() -= f.minCoeff();
    return (f / f.maxCoeff());
}

void visualizeRegionOrientedEdgeHistogram(VectorXd &f, Vector4i extent,
    IplImage *canvas, int nGrid)
{
    SVL_ASSERT((canvas != NULL) && (canvas->height >= extent[1] + extent[3]) &&
        (canvas->width >= extent[0] + extent[2]) && (nGrid > 0));
    SVL_ASSERT((canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    int dx = (extent[2] + nGrid - 1) / nGrid;
    int dy = (extent[3] + nGrid - 1) / nGrid;

    // draw grids
    cvRectangle(canvas, cvPoint(extent[0], extent[1]),
        cvPoint(extent[0] + extent[2], extent[1] + extent[3]), CV_RGB(255, 0, 0));
    for (int y = extent[1]; y < extent[1] + extent[3]; y += dy) {
        cvLine(canvas, cvPoint(extent[0], y), cvPoint(extent[0] + extent[2], y), CV_RGB(255, 0, 0));
    }
    for (int x = extent[0]; x < extent[0] + extent[2]; x += dx) {
        cvLine(canvas, cvPoint(x, extent[1]), cvPoint(x, extent[1] + extent[3]), CV_RGB(255, 0, 0));
    }

    // draw features
    int featureIndx = 0;
    int nChannels = f.size() / (nGrid * nGrid);
    double theta = 0.0;
    double lx = 0.5 * extent[2] / nGrid - 1;
    double ly = 0.5 * extent[3] / nGrid - 1;
    for (int c = 0; c < nChannels; c++) {
        CvPoint midPoint = cvPoint(0, extent[1] + dy / 2);
        for (int y = 0; y < nGrid; y++) {            
            midPoint.x = extent[0] + dx / 2;
            for (int x = 0; x < nGrid; x++, featureIndx++) {
                cvLine(canvas, cvPoint((int)(midPoint.x - lx * sin(theta)), (int)(midPoint.y - ly * cos(theta))),
                    cvPoint((int)(midPoint.x + lx * sin(theta)), (int)(midPoint.y + ly * cos(theta))),
                    CV_RGB(0, (int)(255 * f[featureIndx]), 255 - (int)(255 * f[featureIndx])), 2);
                midPoint.x += dx;
            }
            midPoint.y += dy;
        }
        theta += M_PI / (double)nChannels;
    }
}

IplImage *visualizeEdgeMaps(const vector<CvMat *> &edgeMaps)
{
    vector<IplImage *> views;
    for (vector<CvMat *>::const_iterator it = edgeMaps.begin(); it != edgeMaps.end(); it++) {
#if 0
        views.push_back(cvCreateImage(cvSize((*it)->cols, (*it)->rows), IPL_DEPTH_32F, 1));
        cvConvertScale(*it, views.back());
#else
        CvMat *m = cvCloneMat(*it);
        scaleToRange(m, 0.0, 1.0);
        views.push_back(svlCreateHeatMap(m));
        cvReleaseMat(&m);
#endif
    }
    IplImage *canvas = combineImages(views);
    svlReleaseImages(views);
    return canvas;
}
