/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneObjectInference.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**  Code for running inference on a SceneObjectModel.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <limits>
#include <set>
#include <algorithm>
#include <iterator>
#include <sys/time.h>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneObjectModel.h"
#include "sceneVisualization.h"
#include "sceneSegmentDict.h"
#include "sceneObjDetDict.h"
#include "sceneRegionFeatures.h"

#define WINDOW_NAME "sceneInference"

using namespace std;

// datatypes ----------------------------------------------------------------

class InferenceState {
public:
    // search state
    bool bConverged;
    int numMoves;
    SceneAssignment bestAssignment;
    double bestEnergy;

    // visualization options
    bool bVisualize;
    const char *videoDir;
    set<CvPoint> boundaryPixels;
    CvRect boundingBox;

public:
    InferenceState() :
        bConverged(false), numMoves(0), bestEnergy(0.0),
        bVisualize(false), videoDir(NULL) {
        boundingBox = cvRect(0, 0, 0, 0);
    }

    ~InferenceState() {
        // do nothing
    }

    void swap(InferenceState& state) {
        std::swap(bConverged, state.bConverged);
        std::swap(numMoves, state.numMoves);
        bestAssignment.swap(state.bestAssignment);
        std::swap(bestEnergy, state.bestEnergy);
        std::swap(bVisualize, state.bVisualize);
        std::swap(videoDir, state.videoDir);
        std::swap(boundaryPixels, state.boundaryPixels);
        std::swap(boundingBox, state.boundingBox);
    }

    InferenceState& operator=(const InferenceState& state) {
        if (this != &state) {
            InferenceState temp(state);
            swap(temp);
        }
        return *this;
    }

    void init(double e, const SceneAssignment& assignment) {
        bConverged = false;
        numMoves = 0;
        bestEnergy = e;
        bestAssignment = assignment;
    }

    // TODO: same as init()?
    void update(double e, const SceneAssignment& assignment) {
        SVL_LOG(SVL_LOG_MESSAGE, "updating configuration: dE = "
            << bestEnergy - e << ", E = " << e << " at (sub)move " << numMoves);
        if (e != bestEnergy) {
            bConverged = false;
        }
        bestEnergy = e;
        bestAssignment = assignment;
    }
};

// move making prototypes ---------------------------------------------------

void finalizeMove(InferenceState& state, SceneRegionModel *model, SceneInstance& instance);

// region moves
void mergeMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance);
void segmentDictionaryMoves(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneSegmentDictionary& dictionary);

// object moves
void objectDeleteMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance);
void objectMergeMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance);
void objectSplitMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance);

// top-down candidate moves
void objDetCandidateMoves(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneObjDetDict &moves);

void objectErosionMove(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneSegmentDictionary& dictionary, int objId);
void objDetMovesWithErosion(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneObjDetDict &moves, SceneSegmentDictionary& dictionary);

void evalObjDetCandidateMoves(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneObjDetDict &moves, svlObject2dFrame *frame,
    SceneFeatureCache *data, bool bSaveContext = false);

void evalCachedObjectMoves(InferenceState& state, SceneInstance& instance,
    svlObject2dSequence *detections);

// visualization prototypes -------------------------------------------------

void viewProgress(SceneInstance& instance, const InferenceState& state);
void dumpVideoFrame(SceneInstance& instance, const InferenceState& state);

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneObjectInference [OPTIONS] (<evalList> | <baseName>)\n";
    cerr << "OPTIONS:\n"
         << "  -m <model>        :: model type: REGION (default), REGIONPAIR, OBJECT,\n"
         << "                       OBJECTCONTEXT\n"
         << "  -inRegions <ext>  :: initialize regions from file (default: none)\n"
         << "  -inDir <dir>      :: directory for region initialization (default: use configXML)\n"
         << "  -outRegions <ext> :: extension for region id output (default: none)\n"
         << "  -outLabels <ext>  :: extension for region label output (default: none)\n"
         << "  -outClasses <ext> :: extension for region class output (default: none)\n"
         << "  -outObjects <ext> :: extension for object output (default: none)\n"
         << "  -outHorizon <ext> :: extension for horizon output (default: none)\n"
         << "  -outImages <ext>  :: extension for visualization output (default: none)\n"
         << "  -outSingletons <ext> :: extension for outputing region singletons (default: none)\n"
         << "  -skipExisting <ext> :: skip image if file with <ext> exists in output\n"
         << "  -maxMoves <n>     :: maximum number of proposal moves (default: Inf)\n"
         << "  -maxIters <n>     :: maximum iterations through each set of moves (default: Inf)\n"
         << "  -disableMerges    :: disable merge moves\n"
         << "  -disableSegDict   :: disable segmentation dictionary moves\n"
         << "  -enableObjMoves   :: enable object merges and splits\n"
         << "  -enableObjCandMoves :: enable top-down object candidate moves\n"
         << "  -enableObjErosion :: enable top-down object moves with erosion\n"
         << "  -evalObjectMoves <fn> :: save objects from candidate moves\n"
         << "  -saveObjMoveEnergy <fn> :: save energy components for candidate object moves\n"
         << "  -cachedObjectMoves:: use cached object detections (sole option)\n"
         << "  -video <dir>      :: dump frames to <dir> for generating videos\n"
         << "  -x[0]             :: visualize output (0: end of iterations only)\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // inference state
    InferenceState state;

    // read commandline parameters
    const char *modelType = "REGION";
    const char *inRegionsExt = NULL;
    const char *inRegionDir = NULL;
    const char *outRegionsExt = NULL;
    const char *outLabelsExt = NULL;
    const char *outObjectsExt = NULL;
    const char *outClassesExt = NULL;
    const char *outHorizonExt = NULL;
    const char *outImagesExt = NULL;
    const char *outSingletonExt = NULL;
    const char *skipExisting = NULL;

    const char *evalObjectMovesFilename = NULL;
    const char *saveObjMoveEnergyFile = NULL;

    int maxMoves = numeric_limits<int>::max();
    int maxIters = numeric_limits<int>::max();
    bool bDisableMerges = false;
    bool bDisableSegDict = false;
    bool bEnableObjMoves = false;
    bool bEnableObjCandidateMoves = false;
    bool bEnableObjErosionMoves = false;
    bool bCachedObjectMoves = false;
    bool bVisualizeIterations = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-m", modelType)
        SVL_CMDLINE_STR_OPTION("-inRegions", inRegionsExt)
        SVL_CMDLINE_STR_OPTION("-inDir", inRegionDir)
        SVL_CMDLINE_STR_OPTION("-outRegions", outRegionsExt)
        SVL_CMDLINE_STR_OPTION("-outLabels", outLabelsExt)
        SVL_CMDLINE_STR_OPTION("-outClasses", outClassesExt)
        SVL_CMDLINE_STR_OPTION("-outObjects", outObjectsExt)
        SVL_CMDLINE_STR_OPTION("-outHorizon", outHorizonExt)
        SVL_CMDLINE_STR_OPTION("-outImages", outImagesExt)
        SVL_CMDLINE_STR_OPTION("-outSingletons", outSingletonExt)
        SVL_CMDLINE_STR_OPTION("-skipExisting", skipExisting)
        SVL_CMDLINE_INT_OPTION("-maxMoves", maxMoves)
        SVL_CMDLINE_INT_OPTION("-maxIters", maxIters)
        SVL_CMDLINE_BOOL_OPTION("-disableMerges", bDisableMerges)
        SVL_CMDLINE_BOOL_OPTION("-disableSegDict", bDisableSegDict)
        SVL_CMDLINE_BOOL_OPTION("-enableObjMoves", bEnableObjMoves)
        SVL_CMDLINE_BOOL_OPTION("-enableObjCandMoves", bEnableObjCandidateMoves)
        SVL_CMDLINE_BOOL_OPTION("-enableObjErosion", bEnableObjErosionMoves)
        SVL_CMDLINE_STR_OPTION("-evalObjectMoves", evalObjectMovesFilename)
        SVL_CMDLINE_STR_OPTION("-saveObjMoveEnergy", saveObjMoveEnergyFile)
        SVL_CMDLINE_BOOL_OPTION("-cachedObjectMoves", bCachedObjectMoves)
        SVL_CMDLINE_STR_OPTION("-video", state.videoDir)
        SVL_CMDLINE_BOOL_OPTION("-x", state.bVisualize)
        SVL_CMDLINE_BOOL_OPTION("-x0", bVisualizeIterations)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));
    SVL_ASSERT_MSG((saveObjMoveEnergyFile == NULL) || (evalObjectMovesFilename != NULL),
        "-saveObjMoveEnergyFile requires -evalObjectMoves");
    SVL_ASSERT_MSG(!bCachedObjectMoves || (!bEnableObjCandidateMoves && 
            !bEnableObjErosionMoves && (evalObjectMovesFilename == NULL)),
        "-cachedObjectMoves cannot be used with other options");

    // configuration
    const char *evalList = SVL_CMDLINE_ARGV[0];

    // read list of evaluation images
    vector<string> baseNames;
    if (svlFileExists(evalList)) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading evaluation list from " << evalList << "...");
        baseNames = svlReadFile(evalList);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");
    } else {
        baseNames.push_back(string(evalList));
    }

    // randomize basenames
    if (skipExisting != NULL) {
        timeval t;
        gettimeofday(&t, NULL);
        srand(t.tv_usec * t.tv_sec);
        vector<int> order = randomPermutation(baseNames.size());
        vector<string> reorderedBaseNames(baseNames.size());
        for (unsigned i = 0; i < baseNames.size(); i++) {
            reorderedBaseNames[i] = baseNames[order[i]];
        }
        baseNames = reorderedBaseNames;
    }

    // read models
    SceneRegionModel *model = NULL;
    if (!strcasecmp(modelType, "REGION")) {
        model = new SceneRegionModel();
        SVL_ASSERT(!bEnableObjMoves && !bEnableObjCandidateMoves);
    } else if (!strcasecmp(modelType, "REGIONPAIR")) {
        model = new SceneRegionPairModel();
        SVL_ASSERT(!bEnableObjMoves && !bEnableObjCandidateMoves);
    } else if (!strcasecmp(modelType, "OBJECT")) {
        //model = new SceneObjectModel();
        model = new SceneObjectModelNew();
    } else if (!strcasecmp(modelType, "OBJECTCONTEXT")) {
        //model = new SceneObjectContextModel();
        model = new SceneObjectModelNew();
    } else {
        SVL_LOG(SVL_LOG_FATAL, "unrecognized model type " << modelType);
    }
    SVL_ASSERT(model != NULL);

    model->loadModels((gSceneConfig.getOption("baseDir") +
            gSceneConfig.getOption("modelsDir")).c_str());
    svlConfusionMatrix regionConfusion(gSceneConfig.numRegionLabels);

    // saving object detections and corresponding energies
    svlObject2dSequence detections;
    SceneFeatureCache *detectionsEnergyData = NULL;
    if (saveObjMoveEnergyFile != NULL) {
        detectionsEnergyData = new SceneFeatureCache();
    }

    // load all detections from cache
    if (bCachedObjectMoves) {
        for (int o = 0; o < gSceneConfig.numObjectLabels; o++) {
            SVL_LOG(SVL_LOG_VERBOSE, "loading " << gSceneObjectDefs.name(o) << " detections...");
            svlObject2dSequence *d = gSceneObjectDefs.loadCachedDetections(o);
            if (d == NULL) {
                SVL_LOG(SVL_LOG_WARNING, "...no cached detections for " << gSceneObjectDefs.name(o));
                continue;
            }
            
            detections.merge(*d);
            delete d;
        }
    }

    // process images
    SVL_LOG(SVL_LOG_MESSAGE, "Processing images...");
    int hProcessImage = svlCodeProfiler::getHandle("processImage");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        // check if we should skip this file
        if (skipExisting != NULL) {
            string checkFilename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(skipExisting);
            if (svlFileExists(checkFilename.c_str())) {
                continue;
            }
        }

        svlCodeProfiler::tic(hProcessImage);
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");

        // load image
        IplImage *img = cvLoadImage(imgFilename.c_str(), CV_LOAD_IMAGE_COLOR);
        SVL_ASSERT_MSG(img != NULL, "path: " << imgFilename);
        SVL_LOG(SVL_LOG_MESSAGE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*img));

        // load labels
        CvMat *classLabels = NULL;
        string lblFilename = gSceneConfig.filename("lblDir", baseNames[i], "regionLblExt");
        classLabels = cvCreateMat(img->height, img->width, CV_32SC1);
        SVL_ASSERT(classLabels != NULL);
        bool bSuccess = readMatrix(classLabels, lblFilename.c_str());
        if (!bSuccess) {
            cvReleaseMat(&classLabels);
            SVL_LOG(SVL_LOG_ERROR, "Could not read class labels for " << baseNames[i]);
        }

        // initialize scene
        SceneInstance instance(img, baseNames[i].c_str());

        if (inRegionsExt != NULL) {
            string filename;
            if (inRegionDir == NULL) {
                filename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(inRegionsExt);
            } else {
                filename = string(inRegionDir) + string("/") + baseNames[i] + string(inRegionsExt);
            }
            SVL_LOG(SVL_LOG_MESSAGE, "initializing regions from " << filename << "...");
            bool bSuccess = readMatrix(instance.assignment.pixels, filename.c_str());
            if (!bSuccess) {
                SVL_LOG(SVL_LOG_FATAL, "could not initialize regions from file for " << baseNames[i]);
            }
            instance.segmentations.addSegmentation(instance.assignment.pixels);
        }

        instance.loadSegmentations(false);

        // first-stage inference
        SVL_LOG(SVL_LOG_MESSAGE, "inferring pixel stage...");
        model->inferPixels(&instance);

        // second-stage inference
        SVL_LOG(SVL_LOG_MESSAGE, "inferring horizon stage...");
        model->inferHorizon(&instance);

        // third-stage inference
        SVL_LOG(SVL_LOG_MESSAGE, "inferring region stage...");
        double e = model->inferRegions(&instance);
        //instance.validateData();

#if 1
        // assign one object per foreground region
        if (!strcasecmp(modelType, "OBJECT")) {
            SVL_LOG(SVL_LOG_MESSAGE, "inferring object stage...");
            for (int k = 0; k < instance.numRegions(); k++) {
                if (instance.assignment.regions[k]->Sk == gSceneConfig.numRegionLabels - 1) {
                    instance.moveRegions(k, -1);
                }
            }

            e = model->inferRegions(&instance);
        }
#endif

        // initialize search state
        SVL_LOG(SVL_LOG_VERBOSE, "initializing state with " << instance.numRegions() << " regions");
        state.init(e, instance.assignment);

        // generate proposal dictionary
        SceneSegmentDictionary dictionary;
        for (int s = 0; s < instance.segmentations.count(); s++) {
            instance.assignPixels(instance.segmentations.segmentation(s));
            for (int k = 0; k < instance.numRegions(); k++) {
                if (instance.assignment.regions[k]->size() > 1) {
                    dictionary.addSegment(SceneRegionFeatures::regionEntropy(&instance, k),
                        SceneSegment(instance.assignment.pixels, k));
                }
            }
        }

        // initialize object candidate moves
        SceneObjDetDict objCandidateMoves;
        if (bEnableObjCandidateMoves || bEnableObjErosionMoves || (evalObjectMovesFilename != NULL)) {
            objCandidateMoves.load(gSceneConfig.filename("movesDir", instance.name(), "movesExt").c_str());
            SVL_LOG(SVL_LOG_VERBOSE, "..." << objCandidateMoves.numMoves()
                << " object candidate moves loaded");
        }

        // reset instance to initialize assignment
        instance.assignment = state.bestAssignment;
        SVL_LOG(SVL_LOG_VERBOSE, "...regions: " << instance.numRegions()
            << ", objects: " << instance.numObjects() << ", energy: " << e);

        // show results
        if (state.bVisualize || bVisualizeIterations) {
            viewProgress(instance, state);
        }
        dumpVideoFrame(instance, state);

        // evaluate object moves
        if (evalObjectMovesFilename != NULL) {
            svlObject2dFrame frame;
            evalObjDetCandidateMoves(state, model, instance, objCandidateMoves,
                &frame, detectionsEnergyData, !strcasecmp(modelType, "OBJECTCONTEXT"));
            detections[instance.name()] = frame;
        }

        // evaluate cached object moves
        if (bCachedObjectMoves) {
            evalCachedObjectMoves(state, instance, &detections);
            state.bConverged = true;
        }

        // search over proposal moves
        int nIterations = 0;
        while (!state.bConverged && (state.numMoves < maxMoves) && (nIterations < maxIters)) {
            nIterations += 1;
            state.bConverged = true;
            instance.validateData();

            // merge-moves
            if (!bDisableMerges) {
                mergeMoves(state, model, instance);
                if (state.numMoves > maxMoves) break;
            }

            // re-segment moves
            if (!bDisableSegDict) {
                segmentDictionaryMoves(state, model, instance, dictionary);
                if (state.numMoves > maxMoves) break;
            }

            // create objects associated with newly created foreground regions
            if (!strcasecmp(modelType, "OBJECT")) {
                objectDeleteMoves(state, model, instance);

                bool bNewObjectCreated = false;
                for (vector<SceneRegion *>::iterator r = instance.assignment.regions.begin();
                     r != instance.assignment.regions.end(); r++) {
                    if (((*r)->Sk == gSceneConfig.numRegionLabels - 1) && ((*r)->object == NULL)) {
                        instance.moveRegions((*r)->id(), -1);
                        bNewObjectCreated = true;
                    }
                }

                if (bNewObjectCreated) {
                    e = model->inferRegions(&instance);
                    state.update(e, instance.assignment);
                }
            }

            // object merges and splits
            if (bEnableObjMoves) {
                objectSplitMoves(state, model, instance);
                objectMergeMoves(state, model, instance);
                if (state.numMoves > maxMoves) break;
            }

            // top-down object candidate moves
            if (bEnableObjErosionMoves) {
                objDetMovesWithErosion(state, model, instance, objCandidateMoves, dictionary);
            }
            if (bEnableObjCandidateMoves) {
                objDetCandidateMoves(state, model, instance, objCandidateMoves);
            }

            // update horizon
            model->inferHorizon(&instance);

            // show results
            state.boundaryPixels.clear();
            state.boundingBox = cvRect(0, 0, 0, 0);
            if (bVisualizeIterations) {
                viewProgress(instance, state);
            }
            dumpVideoFrame(instance, state);
        }

        // show results
        state.boundaryPixels.clear();
        state.boundingBox = cvRect(0, 0, 0, 0);
        if (state.bVisualize || bVisualizeIterations) {
            viewProgress(instance, state);
        }
        dumpVideoFrame(instance, state);

        SVL_LOG(SVL_LOG_VERBOSE, "...regions: " << instance.numRegions() <<
            ", objects: " << instance.numObjects());
        SVL_LOG(SVL_LOG_MESSAGE, "final energy is " << model->energy(&instance));

        // output eroded object moves
        if (bEnableObjErosionMoves) {
            objCandidateMoves.save(gSceneConfig.filename("movesDir", instance.name(), "movesExt").c_str());
            SVL_LOG(SVL_LOG_VERBOSE, "..." << objCandidateMoves.numMoves()
                << " object candidate moves saved");
        }

        // accumulate predicted classes
        CvMat *predictedClasses = instance.getPixelClasses();
        if (classLabels != NULL) {
            for (int y = 0; y < instance.height(); y++) {
                for (int x = 0; x < instance.width(); x++) {
                    int regionClassLabel = std::min(CV_MAT_ELEM(*classLabels, int, y, x),
                        gSceneConfig.numRegionLabels);
                    regionConfusion.accumulate(regionClassLabel,
                        CV_MAT_ELEM(*predictedClasses, int, y, x));
                }
            }
        }

        // output images
        if (outImagesExt) {
            IplImage *canvas = visualize(&instance);
            cvSaveImage((gSceneConfig.filebase("outputDir", baseNames[i]) +
                    string(outImagesExt)).c_str(), canvas);
            cvReleaseImage(&canvas);
        }

        // output regions
        if (outRegionsExt) {
            writeMatrix(instance.assignment.pixels, (gSceneConfig.filebase("outputDir", baseNames[i]) +
                    string(outRegionsExt)).c_str());
        }

        // output region labels
        if (outLabelsExt) {
            // TODO: extend to objects, etc
            ofstream ofs((gSceneConfig.filebase("outputDir", baseNames[i]) + string(outLabelsExt)).c_str());
            for (vector<SceneRegion *>::const_iterator it = instance.assignment.regions.begin();
                 it != instance.assignment.regions.end(); it++) {
                ofs << (*it)->Sk << "\n";
            }
            ofs.close();
        }

        // output classes
        if (outClassesExt) {
            writeMatrix(predictedClasses, (gSceneConfig.filebase("outputDir", baseNames[i]) +
                    string(outClassesExt)).c_str());
        }

        // output objects
        if (outObjectsExt) {
            string filename = gSceneConfig.filebase("outputDir", "mapObjects") + string(outObjectsExt);
            svlObject2dSequence objectDetections;
            if (svlFileExists(filename.c_str())) {
                objectDetections.read(filename.c_str());
            }

            svlObject2dFrame frame;
            for (vector<SceneObject *>::const_iterator it = instance.assignment.objects.begin();
                 it != instance.assignment.objects.end(); it++) {

                if ((*it)->Ok == -1) continue;

                VectorXi r = (*it)->extent();
                frame.push_back(svlObject2d(r[0], r[1], r[2], r[3]));
                frame.back().pr = (*it)->objDetScore[(*it)->Ok];
                frame.back().name = gSceneObjectDefs.name((*it)->Ok);
            }

            objectDetections[baseNames[i]] = frame;
            objectDetections.write(filename.c_str());

#if 0
            ofstream ofs((gSceneConfig.filebase("outputDir", baseNames[i]) + string(outObjectsExt)).c_str());
            for (vector<SceneObject *>::const_iterator it = instance.assignment.objects.begin();
                 it != instance.assignment.objects.end(); it++) {
                ofs << (*it)->extent().transpose()
                    << " " << (*it)->prObject.transpose()
                    << " " << (*it)->objDetScore.transpose() << "\n";
            }
            ofs.close();
#endif
        }

        // output horizon
        if (outHorizonExt) {
            string filename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(outHorizonExt);
            ofstream ofs(filename.c_str());
            SVL_ASSERT(!ofs.fail());
            ofs << instance.assignment.horizon << "\n";
            ofs.close();
        }

        // output singletons
        if (outSingletonExt) {
            string filename = gSceneConfig.filebase("outputDir", baseNames[i]) +
                string(outSingletonExt);
            ofstream ofs(filename.c_str());
            SVL_ASSERT(!ofs.fail());
            for (int k = 0; k < instance.numRegions(); k++) {
                ofs << instance.assignment.regions[k]->phi.transpose() << "\n";
            }
            ofs.close();
        }

        // show final instance
        if (svlLogger::getLogLevel() >= SVL_LOG_DEBUG) {
            instance.printScene();
        }

        // compute score for this image
        if (classLabels != NULL) {
            CvMat *mask = cvCreateMat(predictedClasses->rows, predictedClasses->cols, CV_8UC1);
            cvCmpS(predictedClasses, -1.0, mask, CV_CMP_LE);
            cvSub(classLabels, predictedClasses, predictedClasses);
            int errors = cvCountNonZero(predictedClasses);
            int missing = cvCountNonZero(mask);
            SVL_LOG(SVL_LOG_MESSAGE, "...region class score: "
                << (1.0 - (double)(errors - missing) / (img->width * img->height - missing)));
            cvReleaseMat(&mask);
        }

        // free memory
        state.bestAssignment.clear();
        cvReleaseMat(&predictedClasses);
        if (classLabels != NULL) {
            cvReleaseMat(&classLabels);
        }
        cvReleaseImage(&img);

        svlCodeProfiler::toc(hProcessImage);
    }

    // output object detections and energies
    if (evalObjectMovesFilename != NULL) {
        detections.write(evalObjectMovesFilename);
    }

    if (saveObjMoveEnergyFile != NULL) {
        SVL_ASSERT(detectionsEnergyData != NULL);
        detectionsEnergyData->weights.resize(detectionsEnergyData->size(), 1.0);
        detectionsEnergyData->write(saveObjMoveEnergyFile);
        delete detectionsEnergyData;
    }

    // show accuracy
    if (svlLogger::getLogLevel() >= SVL_LOG_VERBOSE) {
        regionConfusion.printRowNormalized(cout, "--- Class Confusion Matrix ---");
    }
    SVL_LOG(SVL_LOG_MESSAGE, "Overall class accuracy: " << regionConfusion.accuracy());
    SVL_LOG(SVL_LOG_MESSAGE, "Average class accuracy: " << regionConfusion.avgRecall());

    // free memory
    delete model;
    if ((baseNames.size() == 1) && (state.bVisualize || bVisualizeIterations)) {
        cvWaitKey(-1);
    }
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

// move-making functions ---------------------------------------------------------------

// common functionality to call after making moves
void finalizeMove(InferenceState& state, SceneRegionModel *model, SceneInstance& instance)
{
    // reset configuration to best assignment and clean up orphaned and small regions
    instance.assignment.swap(state.bestAssignment);
    instance.splitOrphans();
    instance.removeSmallRegions();
    model->inferRegions(&instance);

    state.bestAssignment = instance.assignment;

    // show results
    if (state.bVisualize) {
        viewProgress(instance, state);
    }
    dumpVideoFrame(instance, state);

    state.boundingBox = cvRect(0, 0, 0, 0);
    state.boundaryPixels.clear();
}

void mergeMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance)
{
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << instance.numRegions() << " merges...");
    int regId = 0;
    while (regId < instance.numRegions()) {
        SceneRegion *region = instance.assignment.regions[regId];
        SVL_ASSERT(region->id() == regId);
        if (region->empty()) {
            regId += 1;
            continue;
        }

        // show merge move
        state.boundaryPixels = set<CvPoint>(region->pixels.begin(), region->pixels.end());

        state.numMoves += 1;
        set<SceneRegion *> nearRegions;
        for (set<SceneRegionBoundary *>::const_iterator it = region->boundaries.begin();
             it != region->boundaries.end(); it++) {
            nearRegions.insert((*it)->neighborOf(region));
        }
        vector<CvPoint> pixels(region->pixels.begin(), region->pixels.end());

        // evaluate neighbors
        for (set<SceneRegion *>::const_iterator r = nearRegions.begin(); r != nearRegions.end(); ++r) {
            if ((*r)->empty()) break;

            instance.movePixels(pixels, (*r)->id());
            double e = model->inferRegions(&instance);

            SVL_LOG(SVL_LOG_DEBUG, "energy for merging " << regId
                << " with " << (*r)->id() << " is " << e);
            bool bAcceptMove = (state.bestEnergy - e > SVL_EPSILON);

            // make the move
            if (bAcceptMove) {
                state.update(e, instance.assignment);
            }
        }

        // finalize the move
        state.boundingBox = cvRect(0, 0, 0, 0);
        finalizeMove(state, model, instance);
        regId += 1;
    }
}

void segmentDictionaryMoves(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneSegmentDictionary& dictionary)
{
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << dictionary.size() << " moves...");
    for (map<double, SceneSegment>::const_iterator it = dictionary.entries.begin(); it != dictionary.entries.end(); ++it) {
        state.numMoves += 1;

        // neighbor regions
        set<int> nearRegions;
        for (vector<CvPoint>::const_iterator p = it->second.pixels.begin();
             p != it->second.pixels.end(); ++p) {
            nearRegions.insert(CV_MAT_ELEM(*instance.assignment.pixels, int, p->y, p->x));
        }
        set<CvPoint> const& nbrs = it->second.outsideBoundaryPixels;
        for (set<CvPoint>::const_iterator p = nbrs.begin(); p != nbrs.end(); ++p) {
            nearRegions.insert(CV_MAT_ELEM(*instance.assignment.pixels, int, p->y, p->x));
        }

        // "new" region
        nearRegions.insert(-1);

        SVL_ASSERT(nearRegions.size() > 1);
#if 1
        if ((nearRegions.size() == 2) && (it->second.pixels.size() < 2.0e-3 * instance.size())) {
            SVL_LOG(SVL_LOG_DEBUG, "skipping small internal move of size " << it->second.pixels.size());
            continue;
        }
#endif

        // evaluate neighbors
        for (set<int>::const_iterator s = nearRegions.begin(); s != nearRegions.end(); ++s) {
            int numPixelsMoved = instance.movePixels(it->second.pixels, *s);
            if (numPixelsMoved == 0) continue;
            double e = model->inferRegions(&instance);

            bool bAcceptMove = (state.bestEnergy - e > SVL_EPSILON);

            // make the move
            if (bAcceptMove) {
                state.update(e, instance.assignment);

                // show results
                state.boundaryPixels = it->second.insideBoundaryPixels;
                //state.boundaryPixels = set<CvPoint>(it->second.pixels.begin(), it->second.pixels.end());
                if (state.bVisualize) {
                    viewProgress(instance, state);
                }
                dumpVideoFrame(instance, state);
            }
        }

        // finalize the move
        //state.boundaryPixels = set<CvPoint>(it->second.pixels.begin(), it->second.pixels.end());
        state.boundaryPixels = it->second.insideBoundaryPixels;
        finalizeMove(state, model, instance);
    }
}

void objectDeleteMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance)
{
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << instance.assignment.numObjects() << " object deletions...");
    for (int objId = instance.assignment.numObjects() - 1; objId >= 0; objId--) {
        SVL_LOG(SVL_LOG_DEBUG, "...deleting object " << objId);
        int numPixelsMoved = instance.removeObject(objId);
        if (numPixelsMoved == 0) continue;

        state.numMoves += 1;
        double e = model->inferRegions(&instance);

        // make the move
        if (state.bestEnergy - e > SVL_EPSILON) {
            state.update(e, instance.assignment);

            // show results
            if (state.bVisualize) {
                viewProgress(instance, state);
            }
            dumpVideoFrame(instance, state);
        }

        // finalize the move
        finalizeMove(state, model, instance);
    }
}

void objectMergeMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance)
{
    // for each region neighboring each object, attempt to move the region into the object
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << instance.assignment.numObjects() << " object merge moves...");
    for (vector<SceneObject *>::const_iterator o = instance.assignment.objects.begin();
         o != instance.assignment.objects.end(); o++) {

        // get set of neighbouring regions
        set<SceneRegion *> neighborRegions;
        for (set<SceneObjRegBoundary *>::const_iterator it = (*o)->boundaries.begin();
             it != (*o)->boundaries.end(); it++) {
            if ((*it)->region != NULL)
                neighborRegions.insert((*it)->region);
        }

        // try merge regions greedily
        for (set<SceneRegion *>::const_iterator r = neighborRegions.begin(); r != neighborRegions.end(); r++) {
            state.numMoves += 1;

            SceneObject *oldObject = (*r)->object;
            int numPixelsMoved = instance.moveRegions((*r)->id(), (*o)->id());
            if (numPixelsMoved == 0) continue;

            double e = model->inferRegions(&instance);
            SVL_LOG(SVL_LOG_VERBOSE, "merging region " << (*r)->id() << " with object "
                << (*o)->id() << " has energy " << e);

            bool bAcceptMove = (state.bestEnergy - e > SVL_EPSILON);

            // make the move
            if (bAcceptMove) {
                state.update(e, instance.assignment);

                // show results
                state.boundaryPixels.clear();
                state.boundaryPixels.insert((*r)->pixels.begin(), (*r)->pixels.end());
                if (state.bVisualize) {
                    viewProgress(instance, state);
                }
                dumpVideoFrame(instance, state);
            } else {
                // undo the move
                if (oldObject == NULL) {
                    instance.removeRegions((*r)->id(), (*o)->id());
                } else {
                    instance.moveRegions((*r)->id(), oldObject->id());
                }
            }
        }
    }

    // finalize the move
    finalizeMove(state, model, instance);
}

void objectSplitMoves(InferenceState& state, SceneRegionModel *model, SceneInstance &instance)
{
    // take all regions within an object and create one object for each of them
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << instance.assignment.numObjects() << " object split moves...");
    for (int objId = 0; objId < instance.assignment.numObjects(); objId++) {
        SceneObject *object = instance.assignment.objects[objId];

        // only split objects with 2 or more regions
        if (object->regions.size() < 2) continue;

        state.numMoves += 1;

        list<CvPoint> pixels = object->pixels();
        state.boundaryPixels = set<CvPoint>(pixels.begin(), pixels.end());
        int numPixelsMoved = 0;
        while (object->regions.size() > 1) {
            numPixelsMoved += instance.moveRegions((*object->regions.begin())->id(), -1);
        }

        if (numPixelsMoved == 0) continue;
        double e = model->inferRegions(&instance);

        // make the move
        if (state.bestEnergy - e > SVL_EPSILON) {
            state.update(e, instance.assignment);

            // show results
            if (state.bVisualize) {
                viewProgress(instance, state);
            }
            dumpVideoFrame(instance, state);
        }

        // finalize the move
        finalizeMove(state, model, instance);
    }
}

void objDetCandidateMoves(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneObjDetDict &moves)
{
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << moves.numMoves() << " object candidate moves...");

    for (int i = 0; i < moves.numMoves(); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "...attempting move " << i << " of " << moves.numMoves());

        // create new region out of pixels
        CvMat *mask = moves.mask(i);
        if (!svlIsConnectedComponent(mask, true, 1)) {
            SVL_LOG(SVL_LOG_WARNING, "...object candidate move " << i << " is not connected");
            cvReleaseMat(&mask);
            continue;
        }
        convertInPlace(&mask, CV_8UC1);

        CvMat *pixels = cvCloneMat(instance.assignment.pixels);
        cvSet(pixels, cvScalar(-1), mask);
        cvReleaseMat(&mask);

        instance.movePixels(pixels);
        cvReleaseMat(&pixels);

        // create new object
        instance.moveRegions(instance.numRegions() - 1, -1);
        Vector4i r = instance.assignment.objects.back()->extent();

        // update visualization state here since object may be removed
        // if proposed by a disconnected component
        if (state.bVisualize) {
            state.boundingBox = cvRect(r[0], r[1], r[2], r[3]);
            list<CvPoint> p = instance.assignment.objects.back()->pixels();
            state.boundaryPixels.clear();
            state.boundaryPixels.insert(p.begin(), p.end());
        }

        /*
        instance.splitOrphans();
        instance.removeSmallRegions();
        instance.removeSplitObjects();
        */

        // evaluate the move
        double e = ((SceneObjectModel *)model)->inferRegions(&instance);
        bool bAcceptMove = (state.bestEnergy - e > SVL_EPSILON);

        // keep the move if better
        if (bAcceptMove) {
            state.update(e, instance.assignment);
        }

        // finalize the move
        finalizeMove(state, model, instance);

        state.numMoves += 1;
    }
}

void objectErosionMove(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneSegmentDictionary& dictionary, int objId)
{
    SVL_ASSERT(objId < instance.numObjects());

    bool bConverged = false;
    vector<bool> rejected(dictionary.entries.size(), false);

    while (!bConverged) {
        bConverged = true;

        // look for overlapping segment moves
        int moveIndx = 0;
        for (map<double, SceneSegment>::const_iterator it = dictionary.entries.begin();
             it != dictionary.entries.end(); ++it, moveIndx++) {

            if (rejected[moveIndx]) continue;

            // count overlapping pixels
            const int *pRegId = (const int *)CV_MAT_ELEM_PTR(*instance.assignment.pixels, 0, 0);
            vector<int> overlapCount(instance.numRegions(), 0);
            for (vector<CvPoint>::const_iterator jt = it->second.pixels.begin();
                 jt != it->second.pixels.end(); jt++) {
                //SVL_ASSERT_MSG(jt->y * instance.width() + jt->x < instance.size(), toString(*jt));
                overlapCount[pRegId[jt->y * instance.assignment.pixels->cols + jt->x]] += 1;
            }

            int objectCount = 0;
            for (int regId = 0; regId < instance.numRegions(); regId++) {
                if ((instance.assignment.regions[regId]->object != NULL) &&
                    (instance.assignment.regions[regId]->object->id() == objId)) {
                    objectCount += overlapCount[regId];
                }
            }

            if (objectCount == 0) {
                rejected[moveIndx] = true;
                continue;
            }

            double overlap = (double)objectCount / (double)it->second.size();
            if ((overlap < 0.05) || (overlap > 0.95))
                continue;

            // attempt move to neighbor
            for (int regId = 0; regId < instance.numRegions(); regId++) {
                // skip non-overlapping
                if (overlapCount[regId] == 0)
                    continue;
                /*
                if ((instance.assignment.regions[regId]->object != NULL) &&
                    (instance.assignment.regions[regId]->object->id() == objId))
                    continue;
                */
                int numPixelsMoved = instance.movePixels(it->second.pixels, regId);
                SVL_ASSERT(numPixelsMoved != 0);
                double e = model->inferRegions(&instance);

                // make the move
                if (state.bestEnergy - e > SVL_EPSILON) {
                    bConverged = false;
                    state.update(e, instance.assignment);

                    // show results
                    state.boundaryPixels = it->second.insideBoundaryPixels;
                    if (state.bVisualize) {
                        viewProgress(instance, state);
                    }
                    dumpVideoFrame(instance, state);
                }
            }

            // finalize the move
            CvRect bbox = state.boundingBox;
            state.boundaryPixels = it->second.insideBoundaryPixels;
            finalizeMove(state, model, instance);
            state.numMoves += 1;
            state.boundingBox = bbox;
        }
    }
}

void objDetMovesWithErosion(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneObjDetDict &moves, SceneSegmentDictionary& dictionary)
{
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << moves.numMoves() << " object candidate moves...");

    for (int i = 0; i < moves.numMoves(); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "***** attempting erosion move " << i
            << " of " << moves.numMoves() << " *****");

        // create new region from bounding box
        CvMat *pixels = cvCloneMat(instance.assignment.pixels);
        CvRect bbox = moves[i].originalExtent;
        int *p = (int *)CV_MAT_ELEM_PTR(*pixels, 0, 0);
        for (int y = 0; y < pixels->height; y++) {
            for (int x = 0; x < pixels->width; x++, p++) {
                if ((y >= bbox.y) && (y < bbox.y + bbox.height) &&
                    (x >= bbox.x) && (x < bbox.x + bbox.width)) {
                    *p = -1;
                }
            }
        }

        // force the move
        instance.movePixels(pixels);
        cvReleaseMat(&pixels);

        // create new object
        instance.moveRegions(instance.numRegions() - 1, -1);
        instance.splitOrphans();
        instance.removeSmallRegions();
        instance.removeSplitObjects();

        double e = ((SceneObjectModel *)model)->inferRegions(&instance);
        InferenceState forcedState(state);
        forcedState.init(e, instance.assignment);
        forcedState.boundingBox = bbox;

        // erode bounding box
        objectErosionMove(forcedState, model, instance, dictionary, instance.numObjects() - 1);

        if (instance.numObjects() == 0) {
            SVL_LOG(SVL_LOG_WARNING, "...object from move " << i << " fully eroded");

            CvMat *mask = cvCreateMat(instance.height(), instance.width(), CV_32SC1);
            cvZero(mask);
            moves[i].encode(mask);
            cvReleaseMat(&mask);

        } else {
            // evaluate the move
            e = ((SceneObjectModel *)model)->inferRegions(&instance);
            bool bAcceptMove = (state.bestEnergy - e > SVL_EPSILON);

#if 0
            // print results
            SceneObject *o = instance.assignment.objects.back();
            SVL_LOG(SVL_LOG_VERBOSE, "***** DETECTION: " << toString(bbox) << " "
                << o->extent().transpose() << " " << o->objDetScore.transpose());
#endif

            // save the erosion move
            SceneObject *o = instance.assignment.objects.back();
            CvMat *mask = o->mask(instance.width(), instance.height());
            moves[i].encode(mask);
            cvReleaseMat(&mask);

            // keep the move if better
            if (bAcceptMove) {
                state.update(e, instance.assignment);
            }
        }

        // finalize the move
        finalizeMove(state, model, instance);
        state.numMoves += 1;
    }
}

void evalObjDetCandidateMoves(InferenceState& state, SceneRegionModel *model,
    SceneInstance &instance, SceneObjDetDict &moves, svlObject2dFrame *frame,
    SceneFeatureCache *data, bool bSaveContext)
{
    SVL_ASSERT(frame != NULL);
    SVL_LOG(SVL_LOG_MESSAGE, "attempting " << moves.numMoves() << " object candidate moves...");

    // TODO: add initial set of regions

    for (int i = 0; i < moves.numMoves(); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "...attempting move " << i << " of " << moves.numMoves());

        // create new region out of pixels
        CvMat *mask = moves.mask(i);
        if (!svlIsConnectedComponent(mask, true, 1)) {
            SVL_LOG(SVL_LOG_WARNING, "...object candidate move " << i << " is not connected");
            cvReleaseMat(&mask);
            continue;
        }
        convertInPlace(&mask, CV_8UC1);

        CvMat *pixels = cvCloneMat(instance.assignment.pixels);
        cvSet(pixels, cvScalar(-1), mask);
        cvReleaseMat(&mask);

        instance.movePixels(pixels);
        cvReleaseMat(&pixels);

        // create new object
        instance.moveRegions(instance.numRegions() - 1, -1);
        Vector4i r = instance.assignment.objects.back()->extent();

        // update visualization state here since object may be removed
        // if proposed by a disconnected component
        if (state.bVisualize) {
            state.boundingBox = cvRect(r[0], r[1], r[2], r[3]);
            list<CvPoint> p = instance.assignment.objects.back()->pixels();
            state.boundaryPixels.clear();
            state.boundaryPixels.insert(p.begin(), p.end());
        }

        /*
        instance.splitOrphans();
        instance.removeSmallRegions();
        instance.removeSplitObjects();
        */

        // evaluate the move
        double e = ((SceneObjectModel *)model)->inferRegions(&instance);
        if (!isfinite(e)) {
            instance.printScene();
            SVL_LOG(SVL_LOG_FATAL, toString(((SceneObjectModelNew *)model)->energyComponents(&instance)));
        }

        if (instance.numObjects() > 0) {
            SceneObject *o = instance.assignment.objects.back();
            frame->push_back(svlObject2d((double)r[0], (double)r[1], (double)r[2], (double)r[3]));
            frame->back().pr = o->phi[1];
            //frame->back().pr = log(moves[i].originalScore);
            if (data != NULL) {
                frame->back().index = data->size();
                data->features.push_back(((SceneObjectModelNew *)model)->energyComponents(&instance));
                data->features.back().push_back(frame->back().pr);

                // add context
                // TODO: should be done automatically by context model
                if (bSaveContext) {
                    vector<double> context(gSceneConfig.numRegionLabels, 0.0);
                    for (set<SceneObjRegBoundary *>::const_iterator it = o->boundaries.begin();
                         it != o->boundaries.end(); it++) {
                        context[(*it)->region->Sk] += (double)(*it)->length();                        
                    }
                    Eigen::Map<VectorXd>(&context[0], context.size()) /=
                        Eigen::Map<VectorXd>(&context[0], context.size()).sum() + SVL_DBL_MIN;
                    data->features.back().insert(data->features.back().end(), context.begin(), context.end());
                }

                data->labels.push_back(-1);
            }
        }

        // finalize the move
        finalizeMove(state, model, instance);
        state.numMoves += 1;
    }
}

void evalCachedObjectMoves(InferenceState& state, SceneInstance& instance,
    svlObject2dSequence *detections)
{

    // extract frame for this image
    svlObject2dFrame frame = sortObjects((*detections)[instance.name()]);
    SVL_LOG(SVL_LOG_VERBOSE, "..." << frame.size() << " frames found");
    if (frame.empty()) return;

    // load moves
    vector<SceneObjDetDict *> moves;
    for (int o = 0; o < gSceneConfig.numObjectLabels; o++) {
        moves.push_back(new SceneObjDetDict());
        moves.back()->load(gSceneConfig.filename("movesDir", gSceneObjectDefs.name(o) + 
                string(".") + instance.name(), "movesExt").c_str());
        SVL_LOG(SVL_LOG_VERBOSE, "..." << moves.back()->numMoves() 
            << " candidate moves loaded for " << gSceneObjectDefs.name(o));
    }

    // evaluate moves
    for (int i = 0; i < (int)frame.size(); i++) {
        if (frame[i].pr < 0.0) break;

        // find object
        int o = gSceneObjectDefs.size();
        while (o-- > 0) {
            if (gSceneObjectDefs.name(o) == frame[i].name)
                break;
        }
        if (o == -1) continue;

        // find the move        
        int moveId = moves[o]->find(frame[i].getRect());
        if (moveId < 0) {
            SVL_LOG(SVL_LOG_ERROR, "missing move for " << gSceneObjectDefs.name(o) 
                << " object at " << toString(frame[i].getRect()));
            continue;
        }

        if ((*moves[o])[i].extent.width * (*moves[o])[i].extent.height < 100)
                continue;

        // is move allowed (i.e. will it lower energy?)
        bool bMoveAllowed = true;
        CvMat *mask = moves[o]->mask(moveId);
        const int *p = (int *)CV_MAT_ELEM_PTR(*mask, 0, 0);
        const int *q = (int *)CV_MAT_ELEM_PTR(*instance.assignment.pixels, 0, 0);
        for (int j = 0; j < instance.size(); j++, p++, q++) {
            if (*p == 0) continue;
            SceneObject *object = instance.assignment.regions[*q]->object;
            if (object == NULL) continue;
            if (object->Ok != -1) {
                bMoveAllowed = false;
                break;
            }
        }
        
        // make move
        if (bMoveAllowed) {
            CvMat *pixels = cvCloneMat(instance.assignment.pixels);
            convertInPlace(&mask, CV_8UC1);
            cvSet(pixels, cvScalar(-1), mask);
            instance.movePixels(pixels);
            cvReleaseMat(&pixels);
            
            instance.assignment.regions.back()->Sk = gSceneConfig.numRegionLabels - 1;
            instance.moveRegions(instance.numRegions() - 1, -1);
            instance.assignment.objects.back()->Ok = o;
            instance.assignment.objects.back()->objDetScore = 
                VectorXd::Zero(gSceneConfig.numObjectLabels);
            instance.assignment.objects.back()->objDetScore[o] = frame[i].pr;
        }
        
        cvReleaseMat(&mask);
        
        // show results
        if (state.bVisualize) {
            state.boundingBox = frame[i].getRect();
            viewProgress(instance, state);
        }
    }
}

// visualization functions -------------------------------------------------------------

void viewProgress(SceneInstance& instance, const InferenceState& state)
{
    IplImage *canvas = visualize(&instance);

    if ((state.boundingBox.width > 0) && (state.boundingBox.height > 0)) {
        cvRectangle(canvas, cvPoint(state.boundingBox.x, state.boundingBox.y),
            cvPoint(state.boundingBox.x + state.boundingBox.width, state.boundingBox.y + state.boundingBox.height),
            CV_RGB(0, 255, 0), 3);
    }

    for (set<CvPoint>::const_iterator p = state.boundaryPixels.begin();
         p != state.boundaryPixels.end(); ++p) {
        cvRectangle(canvas, *p, *p, CV_RGB(255, 255, 255), 5);
    }

    for (set<CvPoint>::const_iterator p = state.boundaryPixels.begin();
         p != state.boundaryPixels.end(); ++p) {
        cvRectangle(canvas, *p, *p, CV_RGB(255, 0, 0), 3);
    }

    resizeInPlace(&canvas, 240, 240 * canvas->width / canvas->height);

    cvNamedWindow(WINDOW_NAME, 1);
    cvShowImage(WINDOW_NAME, canvas);
    int ch = cvWaitKey(100);
    if ((char)ch == 'q') {
        if (svlLogger::getLogLevel() >= SVL_LOG_DEBUG) {
            instance.printScene();
        }
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
        svlCodeProfiler::print(cerr);
        exit(0);
    }

    if ((char)ch == 'p') {
        cvWaitKey(-1);
    }
    cvReleaseImage(&canvas);
}

void dumpVideoFrame(SceneInstance& instance, const InferenceState& state)
{
    static int frameCount = 0;
    if (state.videoDir == NULL)
        return;

#if 0
    IplImage *canvas = visualize(&instance);
#else
    vector<IplImage *> views;
    IplImage *grayImage = cvCreateImage(cvGetSize(instance.image()), IPL_DEPTH_8U, 1);
    cvCvtColor(instance.image(), grayImage, CV_BGR2GRAY);
    IplImage *background = cvCreateImage(cvGetSize(instance.image()), IPL_DEPTH_8U, 3);
    for (int c = 1; c <= 3; c++) {
        cvSetImageCOI(background, c);
        cvCopy(grayImage, background);
    }
    cvSetImageCOI(background, 0);
    cvReleaseImage(&grayImage);

    views.push_back(cvCloneImage(instance.image()));

    views.push_back(cvCloneImage(instance.image()));
    cvSet(views.back(), cvScalar(255, 255, 255));
    visualizeRegions(&instance, views.back());
    visualizeClassBoundaries(&instance, views.back(), true);

    views.push_back(cvCloneImage(background));
    visualizeClasses(&instance, views.back(), 0.5);
    visualizeObjects(&instance, views.back());

    views.push_back(cvCloneImage(background));
    visualizeHorizon(&instance, views.back());

    IplImage *canvas = combineImages(views, 2, 2);
    svlReleaseImages(views);
    cvReleaseImage(&background);
#endif

    if ((state.boundingBox.width > 0) && (state.boundingBox.height > 0)) {
        cvRectangle(canvas, cvPoint(state.boundingBox.x, state.boundingBox.y),
            cvPoint(state.boundingBox.x + state.boundingBox.width, state.boundingBox.y + state.boundingBox.height),
            CV_RGB(0, 255, 0), 3);
    }

    for (set<CvPoint>::const_iterator p = state.boundaryPixels.begin();
         p != state.boundaryPixels.end(); ++p) {
        cvRectangle(canvas, *p, *p, CV_RGB(255, 255, 255), 5);
    }

    for (set<CvPoint>::const_iterator p = state.boundaryPixels.begin();
         p != state.boundaryPixels.end(); ++p) {
        cvRectangle(canvas, *p, *p, CV_RGB(255, 0, 0), 3);
    }

    string filename = string(state.videoDir) + string("/frame") +
        padString(toString(frameCount), 4) + string(".png");
    cvSaveImage(filename.c_str(), canvas);
    cvReleaseImage(&canvas);

    frameCount += 1;
}
