/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneObjectModel.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <sys/time.h>

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneObjectModel.h"
#include "sceneRegionFeatures.h"
#include "sceneVisualization.h"

using namespace std;

// SceneObjectModel ------------------------------------------------------

SceneObjectModel::SceneObjectModel() :
    SceneRegionModel(), _boostedObjectClassifiers(NULL)
{
    // do nothing
}

SceneObjectModel::~SceneObjectModel()
{
    // free
    for (vector<SceneRegionObjectClassifier *>::iterator it = _objClassifiers.begin();
         it != _objClassifiers.end(); it++) {
        delete *it;
    }
    _objClassifiers.clear();

    if (_boostedObjectClassifiers != NULL) {
        delete _boostedObjectClassifiers;
    }
}

// i/o
void SceneObjectModel::loadModels(const char *modelDir)
{
    SceneRegionModel::loadModels(modelDir);
    loadObjectDetectors(modelDir);
    loadObjectSingletonModel(modelDir);
}

void SceneObjectModel::loadObjectDetectors(const char* modelDir)
{
    // delete existing object detectors
    for (vector<SceneRegionObjectClassifier *>::iterator it = _objClassifiers.begin();
         it != _objClassifiers.end(); it++) {
        delete *it;
    }
    _objClassifiers.clear();

    // load object detectors
    set<int> objectIds = gSceneObjectDefs.keys();
    for (set<int>::const_iterator it = objectIds.begin(); it != objectIds.end(); it++) {
        _objClassifiers.push_back(gSceneObjectDefs.createDetector(*it));
    }
}

void SceneObjectModel::loadObjectSingletonModel(const char *modelDir)
{
    string fileName = string(modelDir) + string("/object.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading object feature whitening parameters from " << fileName);
    _objectFeatureWhitener.load(fileName.c_str());

    fileName = string(modelDir) + string("/object.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading object singleton parameters from " << fileName);
    _objectSingletons.load(fileName.c_str());

    if (gSceneConfig.getOptionAsBool("useBoostedClassifierAtObjectLevel")) {
        loadObjectBoostedClassifier(modelDir);
    }

    if (gSceneConfig.getOptionAsBool("useIndependentPedestrianClassifierAtObjectLevel")) {
        loadPedestrianClassifier(modelDir);
    }
}

void SceneObjectModel::saveObjectSingletonModel(const char *modelDir)
{
    string fileName = string(modelDir) + string("/object.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing object feature whitening parameters to " << fileName);
    _objectFeatureWhitener.save(fileName.c_str());

    fileName = string(modelDir) + string("/object.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing object singleton parameters to " << fileName);
    _objectSingletons.save(fileName.c_str());

    if (gSceneConfig.getOptionAsBool("useBoostedClassifierAtObjectLevel")) {
        saveObjectBoostedClassifier(modelDir, -1);
    }

    //if (gSceneConfig.getOptionAsBool("useIndependentPedestrianClassifierAtObjectLevel")) {
    //savePedestrianClassifier(modelDir);
    //}
}

void SceneObjectModel::loadObjectBoostedClassifier(const char *modelDir)
{
    if (_boostedObjectClassifiers != NULL) {
        delete _boostedObjectClassifiers;
    }

    string fileName = string(modelDir) + string("/object.boostedFeatures.xml");
    _boostedObjectClassifiers = new svlBoostedClassifierSet(gSceneConfig.numObjectLabels);
    SVL_LOG(SVL_LOG_VERBOSE, "...reading boosted classifier features from " << fileName);
    bool bSuccess = _boostedObjectClassifiers->load(fileName.c_str());
    SVL_ASSERT(bSuccess);
}

void SceneObjectModel::saveObjectBoostedClassifier(const char *modelDir, int classId)
{
    SVL_ASSERT(_boostedObjectClassifiers != NULL);
    string fileName = string(modelDir) + string("/object.boostedFeatures.xml");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing boosted classifier features to " << fileName);
    _boostedObjectClassifiers->save(fileName.c_str(), ".model", classId);
}

void SceneObjectModel::loadPedestrianClassifier(const char *modelDir)
{
    string fileName = string(modelDir) + string("/pedestrian.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading pedestrian feature whitening parameters from " << fileName);
    _pedestrianFeatureWhitener.load(fileName.c_str());

    fileName = string(modelDir) + string("/pedestrian.classifier.SVMLight.model");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading pedestrian singleton parameters from " << fileName);
    _pedestrianClassifier.load(fileName.c_str());
}

void SceneObjectModel::savePedestrianClassifier(const char *modelDir)
{
    string fileName = string(modelDir) + string("/pedestrian.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing pedestrian feature whitening parameters to " << fileName);
    _pedestrianFeatureWhitener.save(fileName.c_str());

    //    fileName = string(modelDir) + string("/pedestrian.singleton.txt");
    //    SVL_LOG(SVL_LOG_VERBOSE, "...writing pedestrian singleton parameters to " << fileName);
    //    _pedestrianClassifier.save(fileName.c_str());
}

// learning
void SceneObjectModel::learnObjectSingletons(const vector<string>& baseNames,
    int maxIterations, bool bResumeTraining, bool bTrainBoostedClassifiers)
{
    // baseDirimgDir/baseName.jpg               [H x W x 3]
    // baseDir/lblDir/baseName.layers.txt       [H x W]
    // baseDir/lblDir/baseName.layerLabels.txt  [K x 1]

    // baseDir/auxLblDir/baseName.layers.txt
    // baseDir/auxLblDir/baseName.layerLabels.txt

    // load horizon groundtruth
    bool bUseGeometry = gSceneConfig.getOptionAsBool("useGeometry");
    map<string, int> horizons;
    if (bUseGeometry) {
        horizons = SceneDataLoader::loadHorizonTrainingData();
    }

    // feature values and labels for training
    SceneFeatureCache trainingData;
    SceneFeatureCache trainingDataPed;
    int nLabels = gSceneConfig.numObjectLabels + 1; // first class is the "unknown" class
    vector<int> labelCounts(nLabels, 0);
    vector<double> labelWeight(nLabels, 0.0);

    // accumulate training data
    for (int i = 0; i < (int)baseNames.size(); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "processing image " << baseNames[i] << " ("
            << (i + 1) << " of " << baseNames.size() << ")...");
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        SceneInstance instance(imgFilename.c_str());

        // locate horizon for this scene
        if (bUseGeometry) {
            if (horizons.find(baseNames[i]) == horizons.end()) {
                SVL_LOG(SVL_LOG_ERROR, "could not find horizon for " << baseNames[i]);
            } else {
                instance.assignment.horizon = horizons[baseNames[i]];
                instance.assignment.horizon = std::min(std::max(instance.assignment.horizon, 0), instance.height() - 1);
            }
        }

        // initialize pixel features
        this->initializeInstance(&instance);

        // compute region features
        int initialCacheSize = trainingData.size();
        computeObjectTrainingFeatures(instance, trainingData);

        if (gSceneConfig.getOptionAsBool("useIndependentPedestrianClassifierAtObjectLevel")) {
            computePedestrianTrainingFeatures(instance, trainingDataPed);
        }

        // update label counts/weights
        for (int k = initialCacheSize; k < trainingData.size(); k++) {
            // adjust "invalid" labels
            //if (trainingData.labels[k] < 0)
            //    trainingData.labels[k] = nLabels - 1;

            // update counts;
            labelCounts[trainingData.labels[k]] += 1;
            labelWeight[trainingData.labels[k]] += trainingData.weights[k];
        }

        SVL_LOG(SVL_LOG_VERBOSE, "...accumulated " << toString(labelCounts) << " samples");
    }

    // add data from closed-loop cache
    // TODO: weight these differently to GT samples?
    if (!gSceneConfig.getOption("closedLoopFeatureCacheObject").empty()) {
        SVL_LOG(SVL_LOG_VERBOSE, "adding closed-loop learning features for object model");
        SceneFeatureCache cache((gSceneConfig.getOption("baseDir") +
                gSceneConfig.getOption("closedLoopFeatureCacheObject")).c_str());
        for (int k = 0; k < cache.size(); k++) {
            SVL_ASSERT_MSG(cache.labels[k] < gSceneConfig.numObjectLabels + 1,
                cache.labels[k] << " !< " << gSceneConfig.numObjectLabels + 1);

            double w = cache.weights[k];
            //if (w < 100) continue;

            trainingData.features.push_back(cache.features[k]);
            trainingData.labels.push_back(cache.labels[k]);
            trainingData.weights.push_back(0.5 * w); // TODO: remove hardcoded weight

            labelCounts[trainingData.labels.back()] += 1;
            labelWeight[trainingData.labels.back()] += trainingData.weights.back();
        }

        SVL_LOG(SVL_LOG_VERBOSE, "...accumulated " << toString(labelCounts) << " samples");
    }

    // train boosted features
    if (gSceneConfig.getOptionAsBool("useBoostedClassifierAtObjectLevel")) {
        if (bTrainBoostedClassifiers) {
            // delete old classifiers
            if (_boostedObjectClassifiers != NULL)
                delete _boostedObjectClassifiers;

            // train
            _boostedObjectClassifiers = new svlBoostedClassifierSet(gSceneConfig.numObjectLabels + 1);
            double avgRecall = _boostedObjectClassifiers->train(trainingData.features, trainingData.labels);
            SVL_LOG(SVL_LOG_MESSAGE, "...average recall: " << avgRecall);
        }

        // evaluate for next stage
        SVL_ASSERT(_boostedObjectClassifiers != NULL);
        vector<vector<double> > boostedFeatures;
        _boostedObjectClassifiers->getClassScores(trainingData.features, boostedFeatures);

        /*
        double posCount = 0, negCount = 0;
        // train pedestrian classifier independently
        if (gSceneConfig.getOptionAsBool("useIndependentPedestrianClassifierAtObjectLevel")) {
            SVL_ASSERT(gSceneConfig.getOptionAsBool("usePedestrianFeaturesAtObjectLevel"));
            SVL_LOG(SVL_LOG_VERBOSE, "training independent pedestrian classifier ...");
            // extract features
            SceneFeatureCache trainingDataPedestrian;
            trainingDataPedestrian.reserve(trainingData.size());
            for (int i = 0; i < (int)trainingData.size(); i++) {
                // TODO: remove hand coded index after experimenting
                vector<double> pedFeature;
                pedFeature.insert(pedFeature.begin(), trainingData.features[i].begin()+1, trainingData.features[i].begin()+8);
                // add bias term
                pedFeature.push_back(1);
                trainingDataPedestrian.features.push_back(pedFeature);
                trainingDataPedestrian.labels.push_back(trainingData.labels[i] == 2 ? 1 : 0);
                trainingDataPedestrian.weights.push_back(1);
                if (trainingData.labels[i] == 2)
                    posCount++;
                else
                    negCount++;
            }

            // for debugging dump the feature matrix for pedestrian (both txt and bin)
            string ofsPedFn = gSceneConfig.getOption("baseDir") + gSceneConfig.getOption("featuresDir") + "featureMatrix.pedestrian.txt";
            ofstream ofsPed(ofsPedFn.c_str());
            SVL_ASSERT(ofsPed.fail() == false);
            for (int i = 0; i < (int)trainingDataPedestrian.features.size(); i++) {
                ofsPed << toString(trainingDataPedestrian.features[i]) << " " << trainingDataPedestrian.labels[i] << endl;
            }
            ofsPed.close();

            // train feature whitener
            _pedestrianFeatureWhitener.clear(trainingDataPedestrian.numFeatures());
            _pedestrianFeatureWhitener.train(trainingDataPedestrian.features);
            _pedestrianFeatureWhitener.evaluate(trainingDataPedestrian.features);

            double posWeight = 1.0;
            // balance training data
            if (gSceneConfig.getOptionAsBool("balancedSingletons")) {
                posWeight = negCount / posCount;
            }

            // train model
            svlCodeProfiler::tic(svlCodeProfiler::getHandle("trainPedestrianSingletons"));
            double lambda = gSceneConfig.getOptionAsDouble("regObjectSingleton");

            _pedestrianClassifier = svlLogistic(trainingDataPedestrian.features[0].size(), posWeight, lambda, false);

            SVL_ASSERT(_pedestrianClassifier.numFeatures() == trainingDataPedestrian.numFeatures());
            SVL_ASSERT(_pedestrianClassifier.numClasses() == 2);
            _pedestrianClassifier.setOption("maxIterations", maxIterations);
            vector<double> pedLabels(trainingDataPedestrian.labels.size());
            for (int i = 0; i < (int)pedLabels.size(); i++) {
                pedLabels[i] = trainingDataPedestrian.labels[i];
            }
            _pedestrianClassifier.train(trainingDataPedestrian.features, pedLabels);

            svlCodeProfiler::toc(svlCodeProfiler::getHandle("trainPedestrianSingletons"));

            // get scores
            SVL_ASSERT(trainingDataPedestrian.size() == boostedFeatures.size());
            vector<double> pedScores(trainingDataPedestrian.size());
            for (int i = 0; i < (int)trainingDataPedestrian.size(); i++) {
                pedScores[i] = _pedestrianClassifier.getScore(trainingDataPedestrian.features[i]);

                // add to the boosted featues
                boostedFeatures[i].push_back(pedScores[i]);
            }
        }
        */


//         if (gSceneConfig.getOptionAsBool("useRegionFeaturesAfterObjectBoosting")) {
//             for (int t = 0; t < (int)boostedFeatures.size(); t++) {
//                 boostedFeatures[t].insert(boostedFeatures[t].end(),
//                     featureVectors[t].begin(), featureVectors[t].end());
//             }
//         }

        if (gSceneConfig.getOptionAsBool("usePedDetectorAfterObjectBoosting")) {
            int pedId = -1;
            for (int i = 0; i < (int)_objClassifiers.size(); i++) {
                if (_objClassifiers[i]->getObjName() == "pedestrian") {
                    pedId = i;
                     break;
                }
            }
            SVL_ASSERT(pedId != -1);
            for (int t = 0; t < (int)boostedFeatures.size(); t++) {
                // assumption is that the feature vector starts from object detectionn scores
                boostedFeatures[t].push_back(trainingData.features[t][pedId]);
            }
        }

        // add bias term
        for (int i = 0; i < trainingData.size(); i++) {
            boostedFeatures[i].push_back(1.0);
        }

        trainingData.features = boostedFeatures;
    }

    if (gSceneConfig.getOptionAsBool("useIndependentPedestrianClassifierAtObjectLevel")) {
        SVL_ASSERT(trainingData.features.size() == trainingDataPed.features.size());
        _pedestrianFeatureWhitener.evaluate(trainingDataPed.features);
        for (int i = 0; i < trainingData.size(); i++) {
            trainingData.features[i].push_back(_pedestrianClassifier.getScore(trainingDataPed.features[i]));
        }
    }

    // train feature whitener
    _objectFeatureWhitener.clear(trainingData.numFeatures());
    _objectFeatureWhitener.train(trainingData.features);
    _objectFeatureWhitener.evaluate(trainingData.features);

    // balance training data
    if (gSceneConfig.getOptionAsBool("balancedSingletons")) {
        for (int i = 0; i < trainingData.size(); i++) {
            trainingData.weights[i] /= (double)(labelWeight[trainingData.labels[i]] + 1.0);
        }
    }

    // train model
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("trainObjectSingletons"));
    double lambda = gSceneConfig.getOptionAsDouble("regObjectSingleton");
    if (!bResumeTraining) {
        _objectSingletons = svlMultiClassLogistic(trainingData.numFeatures(), nLabels, lambda);
    }
    SVL_ASSERT(_objectSingletons.numFeatures() == trainingData.numFeatures());
    SVL_ASSERT(_objectSingletons.numClasses() == nLabels);
    _objectSingletons.setOption("maxIterations", maxIterations);
    _objectSingletons.train(trainingData.features, trainingData.labels, trainingData.weights);
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("trainObjectSingletons"));

    // evaluate (on training data)
    if (svlLogger::getLogLevel() <= SVL_LOG_VERBOSE) {
        vector<int> predictedLabels;
        _objectSingletons.getClassifications(trainingData.features, predictedLabels);

        svlConfusionMatrix confusion(nLabels);
        confusion.accumulate(trainingData.labels, predictedLabels);
        confusion.write(cout);
        SVL_LOG(SVL_LOG_VERBOSE, "Accuracy: " << confusion.accuracy());
    }
}

// learning pedestrian classifier (currently only output features for SVM)
void SceneObjectModel::learnPedestrianClassifier(const vector<string>& baseNames)
{
        // baseDirimgDir/baseName.jpg               [H x W x 3]
    // baseDir/lblDir/baseName.layers.txt       [H x W]
    // baseDir/lblDir/baseName.layerLabels.txt  [K x 1]

    // baseDir/auxLblDir/baseName.layers.txt
    // baseDir/auxLblDir/baseName.layerLabels.txt

    // load horizon groundtruth
    bool bUseGeometry = gSceneConfig.getOptionAsBool("useGeometry");
    map<string, int> horizons;
    if (bUseGeometry) {
        horizons = SceneDataLoader::loadHorizonTrainingData();
    }

    // feature values and labels for training
    SceneFeatureCache trainingDataPedestrian;
    int nLabels = 2;
    vector<int> labelCounts(nLabels, 0);
    vector<double> labelWeight(nLabels, 0.0);

    // randomize baseNames for parallel
    timeval t;
    gettimeofday(&t, NULL);
    srand(t.tv_usec * t.tv_sec);
    vector<int> order = randomPermutation(baseNames.size());
    vector<string> reorderedBaseNames(baseNames.size());
    for (unsigned i = 0; i < baseNames.size(); i++) {
        reorderedBaseNames[i] = baseNames[order[i]];
    }

    // accumulate training data
    for (int i = 0; i < (int)reorderedBaseNames.size(); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "processing image " << reorderedBaseNames[i] << " ("
            << (i + 1) << " of " << reorderedBaseNames.size() << ")...");
        string imgFilename = gSceneConfig.filename("imgDir", reorderedBaseNames[i], "imgExt");
        SceneInstance instance(imgFilename.c_str());

        // locate horizon for this scene
        if (bUseGeometry) {
            if (horizons.find(reorderedBaseNames[i]) == horizons.end()) {
                SVL_LOG(SVL_LOG_ERROR, "could not find horizon for " << reorderedBaseNames[i]);
            } else {
                instance.assignment.horizon = horizons[reorderedBaseNames[i]];
                instance.assignment.horizon = std::min(std::max(instance.assignment.horizon, 0), instance.height() - 1);
            }
        }

        // initialize pixel features
        this->initializeInstance(&instance);

        // compute features for pedestrian classifier
        int initialCacheSize = trainingDataPedestrian.size();
        computePedestrianTrainingFeatures(instance, trainingDataPedestrian);

        // update label counts/weights
        for (int k = initialCacheSize; k < trainingDataPedestrian.size(); k++) {
            // adjust "invalid" labels
            //if (trainingData.labels[k] < 0)
            //    trainingData.labels[k] = nLabels - 1;

            // update counts;
            labelCounts[trainingDataPedestrian.labels[k]] += 1;
            labelWeight[trainingDataPedestrian.labels[k]] += trainingDataPedestrian.weights[k];
        }

        SVL_LOG(SVL_LOG_VERBOSE, "...accumulated " << toString(labelCounts) << " samples");
    }

    /*
    string ofsPedFn = gSceneConfig.getOption("baseDir") + gSceneConfig.getOption("featuresDir") + "featureMatrix.pedestrian.txt";
    ofstream ofsPed(ofsPedFn.c_str());
    SVL_ASSERT(ofsPed.fail() == false);
    for (int i = 0; i < (int)trainingDataPedestrian.features.size(); i++) {
        ofsPed << toString(trainingDataPedestrian.features[i]) << " " << trainingDataPedestrian.labels[i] << endl;
    }
    ofsPed.close();
    */

    /*
    // train feature whitener
    _pedestrianFeatureWhitener.clear(trainingDataPedestrian.numFeatures());
    _pedestrianFeatureWhitener.train(trainingDataPedestrian.features);
    _pedestrianFeatureWhitener.evaluate(trainingDataPedestrian.features);
    */
}

// inference
double SceneObjectModel::inferRegions(SceneInstance *instance, VectorXd *energyRecord)
{
    SVL_FCN_TIC;
    SVL_ASSERT(instance != NULL);

    double e = 0.0;
    double eRegion = 0.0;
    double eObject = 0.0;
    double eFgRegion = 0.0;
    double eBoundary = 0.0;
    if (energyRecord != NULL) {
        *energyRecord = VectorXd::Zero(5);
    }

    // update cache with singleton potentials
    updateInstanceCache(instance);

    // maximize each region term
    for (int k = 0; k < instance->numRegions(); k++) {
        // check for empty regions
        if (instance->assignment.regions[k]->empty())
            continue;

        int indx = argmax(instance->assignment.regions[k]->phi);
        instance->assignment.regions[k]->Sk = indx;

        // update energy
        e -= (double)instance->assignment.regions[k]->size() *
            instance->assignment.regions[k]->phi[indx];
    }

    eRegion = e;

    // maximize each object term
    bool bMultiplyBySize = gSceneConfig.getOptionAsBool("multiplyObjectScoreBySize");
    for (int k = 0; k < (int)instance->assignment.objects.size(); k++) {
        // check for empty objects
        if (instance->assignment.objects[k]->empty())
            continue;

        SVL_ASSERT_MSG(instance->assignment.objects[k]->phi.size() == gSceneConfig.numObjectLabels + 1,
            "object " << k << " : " << toString(instance->assignment.objects[k]->regions) << " : "
            << instance->assignment.objects[k]->phi.transpose());
        int indx = argmax(instance->assignment.objects[k]->phi);
        instance->assignment.objects[k]->Ok = indx - 1;

        double objWeight = gSceneConfig.getOptionAsDouble("objSingletonWeight");
        // update energy
        if (bMultiplyBySize) {
            eObject -= (double)instance->assignment.objects[k]->size() * objWeight *
                instance->assignment.objects[k]->phi[indx];
        } else {
            eObject -= objWeight * instance->assignment.objects[k]->phi[indx];
        }

        // ensure regions are foreground
        // TODO: do we want to force this even if object is "other"?
        for (set<SceneRegion *>::const_iterator it = instance->assignment.objects[k]->regions.begin();
             it != instance->assignment.objects[k]->regions.end(); it++) {
            if ((*it)->empty()) continue;
            eFgRegion -= (double)(*it)->size() *
                ((*it)->phi[gSceneConfig.numRegionLabels - 1] - (*it)->phi[(*it)->Sk]);
            (*it)->Sk = gSceneConfig.numRegionLabels - 1;
        }
    }

    e += eObject + eFgRegion;

    // add boundary penalty to energy
    eBoundary = boundaryPenalty(*instance);
    e += eBoundary;

    // record energies
    if (energyRecord != NULL) {
        (*energyRecord)[0] = eRegion;
        (*energyRecord)[1] = eFgRegion;
        (*energyRecord)[2] = eObject;
        (*energyRecord)[3] = eBoundary;
        (*energyRecord)[4] = e;
    }

    SVL_FCN_TOC;
    return e;
}

// energy for complete assignment
double SceneObjectModel::energy(SceneInstance *instance)
{
    SVL_FCN_TIC;
    SVL_ASSERT(instance != NULL);

    // update cache with singleton potentials
    updateInstanceCache(instance);
    double e = 0.0;

    // accumulate energy for each region term
    for (int k = 0; k < instance->numRegions(); k++) {
        // check for empty regions
        if (instance->assignment.regions[k]->empty())
            continue;

        // update energy
        int label = instance->assignment.regions[k]->Sk;
        SVL_ASSERT((label >= 0) && (label < gSceneConfig.numRegionLabels));

        // update energy
        e -= (double)instance->assignment.regions[k]->size() *
            instance->assignment.regions[k]->phi[label];
    }

    // accumulate each object term
    bool bMultiplyBySize = gSceneConfig.getOptionAsBool("multiplyObjectScoreBySize");
    for (int k = 0; k < (int)instance->assignment.objects.size(); k++) {
        // check for empty objects
        if (instance->assignment.objects[k]->empty())
            continue;

        // update energy
        int label = instance->assignment.objects[k]->Ok;
        SVL_ASSERT((label >= -1) && (label < gSceneConfig.numObjectLabels));
        if (bMultiplyBySize) {
            e -= (double)instance->assignment.objects[k]->size() *
                instance->assignment.objects[k]->phi[label + 1];
        } else {
            e -= instance->assignment.objects[k]->phi[label + 1];
        }
    }

    // add boundary penalty term to energy
    e += boundaryPenalty(*instance);

    SVL_FCN_TOC;
    return e;
}


// updating instance cache with singletons
void SceneObjectModel::updateInstanceCache(SceneInstance *instance)
{
    SVL_FCN_TIC;
    double pixelScoreWeight = gSceneConfig.getOptionAsDouble("addWeightedPixelScore");

    // update each region singleton
    for (int k = 0; k < instance->numRegions(); k++) {
        // check for empty regions
        if (instance->assignment.regions[k]->empty())
            continue;

        // update unary potential
        if (instance->assignment.regions[k]->phi.size() == 0) {
            // compute region feature vector
            vector<double> f = computeRegionFeatureVector(instance, k);

            // whiten features
            _featureWhitener.evaluate(f);

            // compute marginals (removing "invalid" class)
            vector<double> y;
            _regionSingletons.getMarginals(f, y);

            instance->assignment.regions[k]->phi = Eigen::Map<VectorXd>(&y[0], y.size() - 1).cwise().log();
            SVL_ASSERT(instance->assignment.regions[k]->phi.rows() == gSceneConfig.numRegionLabels);

            // add multiple of pixel scores
            if (pixelScoreWeight > 0.0) {
                SVL_ASSERT(instance->assignment.regions[k]->pixelScores.size() == gSceneConfig.numRegionLabels);
                instance->assignment.regions[k]->phi += (pixelScoreWeight / instance->assignment.regions[k]->size()) *
                    Eigen::Map<VectorXd>(&instance->assignment.regions[k]->pixelScores[0], gSceneConfig.numRegionLabels);
            }
        }
    }

    // update each object singleton
    for (int k = 0; k < (int)instance->assignment.objects.size(); k++) {
        // check for empty objects
        if (instance->assignment.objects[k]->empty())
            continue;

        // update unary potential
        if (instance->assignment.objects[k]->phi.size() == 0) {
            // record object related scores
            vector<double> objDetScores;

            // compute region feature vector
            // (scores from the object detector are stored in the feature vector)
            vector<double> f = computeObjectFeatureVector(instance, k);

            // scores from object detector
            objDetScores.resize(instance->assignment.objects[k]->objDetScore.rows());
            Eigen::Map<VectorXd>(&objDetScores[0], objDetScores.size()) = instance->assignment.objects[k]->objDetScore;


            // using boosted features?
            if (_boostedObjectClassifiers != NULL) {
                vector<double> y;
                _boostedObjectClassifiers->getClassScores(f, y);

                /*
                double pedScore = 0;
                if (gSceneConfig.getOptionAsBool("useIndependentPedestrianClassifierAtObjectLevel")) {
                    SVL_ASSERT(gSceneConfig.getOptionAsBool("usePedestrianFeaturesAtObjectLevel"));
                    vector<double> pedFeature;
                    pedFeature.insert(pedFeature.begin(), f.begin()+1, f.begin()+7);
                    // add bias term
                    pedFeature.push_back(1);
                    // whiten features
                    _pedestrianFeatureWhitener.evaluate(pedFeature);
                    pedScore = _pedestrianClassifier.getScore(pedFeature);
                    y.push_back(pedScore);
                }
                */

                // scores from boosting
                objDetScores.insert(objDetScores.end(), y.begin(), y.end());



                if (gSceneConfig.getOptionAsBool("useCrossTermsAfterObjectBoosting")) {
                    unsigned numFeatures = y.size();
                    for (unsigned i = 0; i < numFeatures; i++) {
                        for (unsigned j = 0; j <= i; j++) {
                            y.push_back(y[i] * y[j]);
                        }
                    }
                }

                if (gSceneConfig.getOptionAsBool("useRegionFeaturesAfterObjectBoosting")) {
                    y.insert(y.end(), f.begin(), f.end());
                }

                if (gSceneConfig.getOptionAsBool("usePedDetectorAfterObjectBoosting")) {
                    int pedId = -1;
                    for (int i = 0; i < (int)_objClassifiers.size(); i++) {
                        if (_objClassifiers[i]->getObjName() == "pedestrian") {
                            pedId = i;
                            break;
                        }
                    }
                    SVL_ASSERT(pedId != -1);
                    y.push_back(f[pedId]);
                }

                y.push_back(1.0);
                f = y;
            }

            if (gSceneConfig.getOptionAsBool("useIndependentPedestrianClassifierAtObjectLevel")) {
                // TODO: remove hand coded index for pedestrian
                double pedDetScore = objDetScores[1];
                vector<double> pedFeature = computePatchBasedPedestrianFeature(instance, k, &pedDetScore);
                _pedestrianFeatureWhitener.evaluate(pedFeature);
                double pedClassifierScore = _pedestrianClassifier.getScore(pedFeature);
                objDetScores.push_back(pedClassifierScore);
                f.push_back(pedClassifierScore);
            }

            instance->assignment.objects[k]->objDetScore = Eigen::Map<VectorXd>(&objDetScores[0], objDetScores.size());


            // whiten features
            _objectFeatureWhitener.evaluate(f);

            // compute marginals
            vector<double> y;
            _objectSingletons.getMarginals(f, y);

            instance->assignment.objects[k]->prObject = Eigen::Map<VectorXd>(&y[0], y.size());
            instance->assignment.objects[k]->phi = Eigen::Map<VectorXd>(&y[0], y.size()).cwise().log();

            // normalize so that non-recognizable objects have score 0
            instance->assignment.objects[k]->phi.cwise() -= instance->assignment.objects[k]->phi[0];
        }
    }

    SVL_FCN_TOC;
}

// compute region/object feature vectors
vector<double> SceneObjectModel::computeObjectFeatureVector(SceneInstance *instance, int objId)
{
    SVL_FCN_TIC;

    // object detetor scores
    vector<double> f;
    SceneObject *object = instance->assignment.objects[objId];

    if (object->bComputeObjDetScores) {
        bool bPatchBased = gSceneConfig.getOptionAsBool("patchBasedObjectDetector");
        if (bPatchBased) {
            f = computePatchBasedObjectFeature(instance, objId);
        } else {
            f = computeRegionBasedObjectFeature(instance, objId);
        }
    } else {
        int numObjectLabels = gSceneConfig.numObjectLabels;
        SVL_ASSERT(object->objDetScore.rows() >= numObjectLabels);
        f.resize(numObjectLabels, 0);
        Eigen::Map<VectorXd>(&f[0], numObjectLabels) = Eigen::Map<VectorXd>(&(object->objDetScore[0]), numObjectLabels);
    }

    // "cache" the det. scores
    object->bComputeObjDetScores = false;

    // record scores from region-based object detector alone
    instance->assignment.objects[objId]->objDetScore = Eigen::Map<VectorXd>(&f[0], f.size());

    SVL_LOG(SVL_LOG_DEBUG, "score(Ok[" << objId << "]) = " << toString(f));

    // include cross terms
//     if (gSceneConfig.getOptionAsBool("includeObjectCrossTerms")) {
//         unsigned numFeatures = f.size();
//         f.reserve(numFeatures * (numFeatures + 1) / 2);
//         for (unsigned i = 0; i < numFeatures; i++) {
//             for (unsigned j = 0; j <= i; j++) {
//                 f.push_back(f[i] * f[j]);
//             }
//         }
//     }

    // add region features
    if (gSceneConfig.getOptionAsBool("useRegionFeaturesAtObjectLevel")) {
        vector<double> x;
        SceneRegionFeatures::objectFeatures(instance, objId, x);

        unsigned numStartFeatures = f.size();
        f.insert(f.end(), x.begin(), x.end());

        // include square terms
        unsigned numFeatures = f.size();
        f.reserve(2 * numFeatures - numStartFeatures);
        for (unsigned i = numStartFeatures; i < numFeatures; i++) {
            f.push_back(f[i] * f[i]);
        }
    }

    // add bias term
    f.push_back(1.0);

    SVL_FCN_TOC;
    return f;
}

vector<double> SceneObjectModel::computeRegionBasedObjectFeature(SceneInstance *instance, int k)
{
    if (instance->objDetectorChannels.empty()) {
        svlImageLoader loader;
        vector<string> files;
        SVL_LOG(SVL_LOG_DEBUG, "loading image channels from " << instance->filename());
        files.push_back(instance->filename());
        files.push_back(instance->filename());
        bool bSuccess = loader.getFrameFromFiles(files, instance->objDetectorChannels);
        SVL_ASSERT(bSuccess == true);

        // initialize HOG feature extractor
        for (int i = 0; i < (int)_objClassifiers.size(); i++) {
            SVL_ASSERT(_objClassifiers[i] != NULL);
            _objClassifiers[i]->loadImage(instance->objDetectorChannels[0]);
        }
    }
    SVL_ASSERT(!instance->objDetectorChannels.empty());

    vector<double> regionObjFeature(_objClassifiers.size(), 0);
    double score = 0;
    list<CvPoint> objectPixels = instance->assignment.objects[k]->pixels();
    for (int i = 0; i < (int)_objClassifiers.size(); i++) {
        // create the region mask
        CvMat *regionMask = cvCreateMat(instance->height(), instance->width(), CV_32SC1);
        cvSetZero(regionMask);
        for (list<CvPoint>::const_iterator p = objectPixels.begin(); p != objectPixels.end(); p++) {
            CV_MAT_ELEM(*regionMask, int, p->y, p->x) = 1;
        }
        // get the obj classification score
        score = _objClassifiers[i]->classifyRegion(regionMask);
        // free memory
        cvReleaseMat(&regionMask);
        regionObjFeature[i] = score;
    }
    return regionObjFeature;
}

vector<double> SceneObjectModel::computePatchBasedObjectFeature(SceneInstance *instance, int k)
{
    //
    // TODO: remove hardcoding
    //
    if (instance->objDetectorChannels.empty()) {
        svlImageLoader loader(false, true);
        vector<string> files;
        string filename = string("/afs/cs/u/sgould/scratch2/streetScenes/data/working/Original/") +
            instance->name() + string(".jpg");
        //string filename = instance->filename();
        SVL_LOG(SVL_LOG_DEBUG, "loading image channels from " << filename);
        //        files.push_back(filename);
        files.push_back(filename);
        bool bSuccess = loader.getFrameFromFiles(files, instance->objDetectorChannels);
        SVL_ASSERT(bSuccess == true);
    }
    SVL_ASSERT(!instance->objDetectorChannels.empty());

    int objDetScale = instance->objDetectorChannels[0]->width / instance->width();
    vector<double> regionObjFeature(_objClassifiers.size(), 0);
    double score = 0;

    // regionExt[0-3]: minX, minY, width, height
    Vector4i regionExt = instance->assignment.objects[k]->extent();
    //SVL_LOG(SVL_LOG_DEBUG, "object " << k << "'s extent is (" << regionExt.transpose() << ")");

    for (int i = 0; i < (int)_objClassifiers.size(); i++) {
        double winRatio = (double)_objClassifiers[i]->getWinWidth() / (double)_objClassifiers[i]->getWinHeight();

        //SVL_LOG(SVL_LOG_DEBUG, "scoring detector " << i << " with winRatio " << winRatio);

        if (_objClassifiers[i]->getObjName() == "pedestrian") {
            score = getPedestrianDetectorScore(instance, k, i);
        }
        else if (_objClassifiers[i]->getObjName() == "car") {
            CvPoint shift = cvPoint(0, 0);
            IplImage* patchImg = getRegionPatch(regionExt, instance, 0.1, shift, winRatio, objDetScale);
#if 0
            cvNamedWindow("debug", 1);
            cvShowImage("debug", patchImg);
            cvWaitKey(-1);
#endif
            score = patchImg == NULL ? 0 : _objClassifiers[i]->classifyRegion(patchImg);
            cvReleaseImage(&patchImg);
        }
        regionObjFeature[i] = score;
    }
    return regionObjFeature;
}

void SceneObjectModel::computeObjectTrainingFeatures(SceneInstance &instance, SceneFeatureCache &cache)
{
    SVL_FCN_TIC;

    // check if instance is in the cache
    if (gSceneConfig.getOptionAsBool("useCache") && (!instance.name().empty())) {
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "objectFeaturesExt");
        if (svlFileExists(filename.c_str())) {
            svlCodeProfiler::tic(svlCodeProfiler::getHandle("readObjectFeatureCache"));
            cache.read(filename.c_str(), true);
            svlCodeProfiler::toc(svlCodeProfiler::getHandle("readObjectFeatureCache"));
            SVL_FCN_TOC;
            return;
        }
    }

    // not in cache (or cache not enabled), so compute
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("computeObjectFeatures"));
    int initialCacheSize = cache.size();

    // groundtruth layer file and labels
    SceneLayersAndLabels layers(instance.width(), instance.height());
    layers.load(gSceneConfig.filename("lblDir", instance.name(), "layerLblExt").c_str(),
        gSceneConfig.filename("lblDir", instance.name(), "layerClassLblExt").c_str());

    instance.assignPixels(layers.layers);

    // accumulate features and labels
    cache.reserve(cache.size() + instance.numRegions());
    for (int k = 0; k < instance.numRegions(); k++) {
        instance.moveRegions(k, -1); // FIXME: iterate above over objects. calling function should define objects

        if (instance.assignment.regions[k]->size() < 10) continue;
        if (layers.labels[k] < gSceneConfig.numRegionLabels - 1) continue; // ignore non-foreground regions

        int objectLabel = layers.labels[k] - gSceneConfig.numRegionLabels + 1;
        SVL_ASSERT(objectLabel <= gSceneConfig.numObjectLabels);

        int objId = instance.assignment.objects.size() - 1;
        instance.assignment.objects[objId]->Ok = objectLabel;
        cache.features.push_back(computeObjectFeatureVector(&instance, objId));
        cache.labels.push_back(objectLabel);
        //cache.weights.push_back((double)instance.assignment.objects[k]->size());
        cache.weights.push_back(1.0);
    }

    svlCodeProfiler::toc(svlCodeProfiler::getHandle("computeObjectFeatures"));

    // write to cache
    if (gSceneConfig.getOptionAsBool("useCache") && (!instance.name().empty())) {
        svlCodeProfiler::tic(svlCodeProfiler::getHandle("writeObjectFeatureCache"));
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "objectFeaturesExt");
        if (initialCacheSize < cache.size()) {
            cache.write(filename.c_str(), initialCacheSize, cache.size() - 1, false);
        }
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("writeObjectFeatureCache"));
    }

    SVL_FCN_TOC;
}

IplImage* SceneObjectModel::getRegionPatch(Vector4i& regionExt, SceneInstance *instance, double enlargeFactor, CvPoint& shift, double winRatio, int objDetScale)
{
    int newWidth = regionExt[2];
    int newHeight = regionExt[3];
    // calculate the region patch size according to detector's ratio
    if (winRatio != 0) {
        if ( (double)regionExt[2]/(double)regionExt[3] >= winRatio) {
            newWidth = (int)(regionExt[2] * (enlargeFactor*2 + 1));
            newHeight = (int)(newWidth / winRatio);
        } else {
            newHeight = (int)(regionExt[3] * (enlargeFactor*2 + 1));
            newWidth = (int)(newHeight * winRatio);
        }
    }

    // origins in the enlarged region patch
    int origX = regionExt[0] + shift.x - (newWidth - regionExt[2]) / 2;
    int origY = regionExt[1] + shift.y - (newHeight - regionExt[3]) / 2;

    if ( (origX >= instance->width() - 1)
        || (origY >= instance->height() - 1)
        || (origX + newWidth - 1 <= 0)
        || (origY + newHeight - 1 <= 0) ) {
        return NULL;
    }

    //SVL_LOG(SVL_LOG_DEBUG, "newWidth = " << newWidth << ", newHeight = " << newHeight);

    // resize to image resolution
    newHeight *= objDetScale;
    newWidth *= objDetScale;
    origX *= objDetScale;
    origY *= objDetScale;
    int imgWidth = objDetScale * instance->width();
    int imgHeight = objDetScale * instance->height();

    // prepare to copy from the original image to the enlarged region patch
    int newMinX = origX;
    int newMinY = origY;
    int newMaxX = newMinX + newWidth - 1;
    int newMaxY = newMinY + newHeight - 1;

    int inImgMinX = std::max(newMinX, 0);
    int inImgMinY = std::max(newMinY, 0);
    int inImgMaxX = std::min(newMaxX, imgWidth-1);
    int inImgMaxY = std::min(newMaxY, imgHeight-1);

    int inImgWidth = inImgMaxX - inImgMinX + 1;
    int inImgHeight = inImgMaxY - inImgMinY + 1;

    // if too little in image then return
    if ((double) inImgWidth / (double) newWidth < 1/4
        || (double) inImgHeight / (double) newHeight < 1/4) {
        return NULL;
    }

    int patchOrigX = inImgMinX - newMinX;
    int patchOrigY = inImgMinY - newMinY;

    // create the region patch
    IplImage* patchImg = cvCreateImage(cvSize(newWidth, newHeight),
        instance->objDetectorChannels[0]->depth,
        instance->objDetectorChannels[0]->nChannels);
    SVL_ASSERT(patchImg != NULL);

    // fill in the region patch
    SVL_ASSERT(patchOrigX >=0 && patchOrigY >= 0);
    SVL_ASSERT(patchOrigX + inImgWidth -1 < newWidth);
    SVL_ASSERT(patchOrigY + inImgHeight -1 < newHeight);
    SVL_ASSERT(inImgMinX + inImgWidth -1 < imgWidth && inImgMinX >= 0);
    SVL_ASSERT(inImgMinY + inImgHeight - 1 < imgHeight && inImgMinY >= 0);

    // set ROI
    cvSetImageROI(patchImg, cvRect(patchOrigX, patchOrigY, inImgWidth, inImgHeight));
    cvSetImageROI(instance->objDetectorChannels[0],
        cvRect(inImgMinX, inImgMinY, inImgWidth, inImgHeight));

    // copy
    cvCopy(instance->objDetectorChannels[0], patchImg);

    // reset ROI
    cvResetImageROI(patchImg);
    cvResetImageROI(instance->objDetectorChannels[0]);

    // If enlarged patch is outside of the image, flip to fill in the zeros
    if (newMinX < 0 || newMinY < 0 || newMaxX > imgWidth - 1 || newMaxY > imgHeight - 1) {
        IplImage* regionImgCopy = cvCloneImage(patchImg);
        if (newMinX < 0) {
            //SVL_ASSERT(2*patchOrigX-1 < newWidth && patchOrigX >= 0);
            SVL_ASSERT(patchOrigX >= 0);

            if ( 2*patchOrigX-1 < newWidth ) {
                cvSetImageROI(patchImg, cvRect(0, 0, patchOrigX, newHeight));
                cvSetImageROI(regionImgCopy, cvRect(patchOrigX, 0, patchOrigX, newHeight));
            } else {
                int flipWidth = newWidth - 1 - patchOrigX;
                SVL_ASSERT(patchOrigX - flipWidth >= 0);
                cvSetImageROI(patchImg, cvRect(patchOrigX - flipWidth, 0, flipWidth, newHeight));
                cvSetImageROI(regionImgCopy, cvRect(patchOrigX, 0, flipWidth, newHeight));
            }

            cvFlip(regionImgCopy, patchImg, 1);

            cvResetImageROI(patchImg);
            cvResetImageROI(regionImgCopy);
        }

        if (newMinY < 0) {
            //SVL_ASSERT(2*patchOrigY-1 < newHeight && patchOrigY >= 0);
            SVL_ASSERT(patchOrigY >= 0);

            if ( 2*patchOrigY-1 < newHeight ) {
                cvSetImageROI(patchImg, cvRect(0, 0, newWidth, patchOrigY));
                cvSetImageROI(regionImgCopy, cvRect(0, patchOrigY, newWidth, patchOrigY));
            } else {
                int flipHeight = newHeight - 1 - patchOrigY;
                SVL_ASSERT(patchOrigY - flipHeight >= 0);
                cvSetImageROI(patchImg, cvRect(0, patchOrigY - flipHeight, newWidth, flipHeight));
                cvSetImageROI(regionImgCopy, cvRect(0, patchOrigY, newWidth, flipHeight));
            }


            cvFlip(regionImgCopy, patchImg, 0);

            cvResetImageROI(patchImg);
            cvResetImageROI(regionImgCopy);
        }

        if (newMaxX > imgWidth - 1) {
                int tmpWidth = newWidth - (patchOrigX + inImgWidth);
                tmpWidth = std::min(tmpWidth, patchOrigX + inImgWidth - 1);
                SVL_ASSERT(tmpWidth >= 0);
                SVL_ASSERT(patchOrigX+inImgWidth+tmpWidth-1 < newWidth && patchOrigX+inImgWidth >= 0);
                SVL_ASSERT(patchOrigX+inImgWidth-1-tmpWidth+1 >= 0);

                cvSetImageROI(patchImg, cvRect(patchOrigX+inImgWidth, 0, tmpWidth, newHeight));
                cvSetImageROI(regionImgCopy, cvRect(patchOrigX+inImgWidth-1-tmpWidth+1, 0, tmpWidth, newHeight));

                cvFlip(regionImgCopy, patchImg, 1);

                cvResetImageROI(patchImg);
                cvResetImageROI(regionImgCopy);
        }

        if (newMaxY > imgHeight - 1) {
            int tmpHeight = newHeight - (patchOrigY + inImgHeight);
            tmpHeight = std::min(tmpHeight, patchOrigY+inImgHeight-1);
            SVL_ASSERT(tmpHeight >= 0);
            SVL_ASSERT(patchOrigY+inImgHeight+tmpHeight-1 < newHeight && patchOrigY+inImgHeight >= 0);
            SVL_ASSERT(patchOrigY+inImgHeight-1-tmpHeight+1 >= 0);

            cvSetImageROI(patchImg, cvRect(0, patchOrigY+inImgHeight, newWidth, tmpHeight));
            cvSetImageROI(regionImgCopy, cvRect(0, patchOrigY+inImgHeight-1-tmpHeight+1, newWidth, tmpHeight));

            cvFlip(regionImgCopy, patchImg, 0);

            cvResetImageROI(patchImg);
            cvResetImageROI(regionImgCopy);
        }

        // free the image region copy
        cvReleaseImage(&regionImgCopy);
    }

    return patchImg;
}

double SceneObjectModel::getPedestrianDetectorScore(SceneInstance *instance, int k, int detectorId)
{
    //
    // TODO: remove hardcoding
    //
    if (instance->objDetectorChannels.empty()) {
        svlImageLoader loader(false, true);
        vector<string> files;
        string filename = string("/afs/cs/u/sgould/scratch2/streetScenes/data/working/Original/") +
            instance->name() + string(".jpg");
        //string filename = instance->filename();
        SVL_LOG(SVL_LOG_DEBUG, "loading image channels from " << filename);
        //        files.push_back(filename);
        files.push_back(filename);
        bool bSuccess = loader.getFrameFromFiles(files, instance->objDetectorChannels);
        SVL_ASSERT(bSuccess == true);
    }
    SVL_ASSERT(!instance->objDetectorChannels.empty());

    int objDetScale = instance->objDetectorChannels[0]->width / instance->width();
    double score = 0;
    double winRatio = -1;
    if (detectorId != -1) {
        winRatio = (double)_objClassifiers[detectorId]->getWinWidth() / (double)_objClassifiers[detectorId]->getWinHeight();
    } else {
        for (int i = 0; i < (int)_objClassifiers.size(); i++) {
            if (_objClassifiers[i]->getObjName() == "pedestrian") {
                winRatio = (double)_objClassifiers[i]->getWinWidth() / (double)_objClassifiers[i]->getWinHeight();
                detectorId = i;
                break;
            }
        }
    }
    SVL_ASSERT(winRatio != -1);

    // regionExt[0-3]: minX, minY, width, height
    Vector4i regionExt = instance->assignment.objects[k]->extent();
    if (gSceneConfig.getOptionAsBool("searchRegionsForPedestrianDetector") ) {
        double bestScore = -10000.0;
        double step = 0.05;
        double bestFactor = 0;
        int deltaX = 4, deltaY = 4;
        int halfSteps = 2;
        vector<CvPoint> shifts;
        CvPoint bestShift;
        IplImage* resizedPatchImage = cvCreateImage(cvSize(_objClassifiers[detectorId]->getWinWidth(),_objClassifiers[detectorId]->getWinHeight()),
            instance->objDetectorChannels[0]->depth,
            instance->objDetectorChannels[0]->nChannels);
        for (int j = 2; j < 10; j++ ) {
            // new extent
            Vector4i newRegionExt = regionExt;
            int newWidth, newHeight;
            if ( (double)regionExt[2]/(double)regionExt[3] >= winRatio) {
                newWidth = (int)(regionExt[2] * (2*j*step + 1));
                newHeight = (int)(newWidth / winRatio);
            } else {
                newHeight = (int)(regionExt[3] * (2*j*step + 1));
                newWidth = (int)(newHeight * winRatio);
            }


            newRegionExt[0] = newRegionExt[0] - (newWidth - newRegionExt[2]) / 2 -halfSteps*deltaX;
            newRegionExt[1] = newRegionExt[1] - (newHeight - newRegionExt[3]) / 2 -halfSteps*deltaY;
            newRegionExt[2] = newWidth + 2*halfSteps*deltaX;
            newRegionExt[3] = newHeight + 2*halfSteps*deltaY;
            CvPoint tmpShift = cvPoint(0, 0);
            // get the image patch at a given scale
            IplImage* patchImg = getRegionPatch(newRegionExt, instance, 0.0, tmpShift, 0, objDetScale);
            if (patchImg == NULL) {
                continue;
            }

#if 0
            IplImage* patchImgCopy = NULL;
            if (k == instance->assignment.objects.size() -1) {
                patchImgCopy = cvCloneImage(patchImg);
                cvNamedWindow("debugObject", 1);
                cvShowImage("debugObject", patchImgCopy);
            }
#endif
            // candidates by shifting
            shifts.clear();
            for (int ii = 0; ii <= 2*halfSteps; ii++ ) {
                for (int jj = 0; jj <= 2*halfSteps; jj++) {
                    shifts.push_back(cvPoint(ii*deltaX*objDetScale, jj*deltaY*objDetScale));
                }
            }
            // get the score
            for (vector<CvPoint>::iterator itr = shifts.begin(); itr != shifts.end(); ++itr ) {
                SVL_ASSERT(itr->x >= 0 && itr->y >=0
                    && itr->x + newWidth*objDetScale - 1 < patchImg->width
                    && itr->y + newHeight*objDetScale - 1 < patchImg->height);
                cvSetImageROI(patchImg, cvRect(itr->x, itr->y, newWidth*objDetScale, newHeight*objDetScale));

                // resize to the detector's window size
                cvResize(patchImg, resizedPatchImage);

#if 0
                if (patchImgCopy != NULL) {
                    SVL_ASSERT(itr->x >= 0 && itr->y >=0
                        && itr->x + newWidth*objDetScale - 1 < patchImgCopy->width
                        && itr->y + newHeight*objDetScale - 1 < patchImgCopy->height);
                    cvRectangle(patchImgCopy, cvPoint(itr->x, itr->y), cvPoint(itr->x + newWidth*objDetScale-1, itr->y + newHeight*objDetScale - 1), CV_RGB(0, 255, 0));
                            cvShowImage("debugObject", patchImgCopy);
                            cvWaitKey(-1);
                }
#endif
                //resizeInPlace(&patchImg, _objClassifiers[i]->getWinHeight(), _objClassifiers[i]->getWinWidth());

                // get the obj classification score
                score = _objClassifiers[detectorId]->classifyRegion(resizedPatchImage);
                cvResetImageROI(patchImg);
                if (score > bestScore) {
                    bestScore = score;
                    bestFactor = j*step;
                    bestShift = *itr;
                }
                    }

            // free memory
            if (patchImg != NULL)
                cvReleaseImage(&patchImg);
        }
        score = bestScore;
        cvReleaseImage(&resizedPatchImage);
        //             cout << "bese shift: (" << bestShift.x << ", " << bestShift.y << ")"
                //                  << "; best scale: " << bestFactor/step << endl;
    } else {
        CvPoint shift = cvPoint(0, 0);
        IplImage* patchImg = getRegionPatch(regionExt, instance, 0.1, shift, winRatio, objDetScale);
        score = patchImg == NULL ? 0 : _objClassifiers[detectorId]->classifyRegion(patchImg);
        cvReleaseImage(&patchImg);
    }

    return score;
}

// for pedestrian classifier
vector<double> SceneObjectModel::computePatchBasedPedestrianFeature(SceneInstance *instance, int k, double* pPedDetScore) {
    vector<double> f;
    if (pPedDetScore != NULL) {
        f.push_back(*pPedDetScore);
    } else {
        f.push_back(getPedestrianDetectorScore(instance, k));
    }

    vector<double> x;
    SceneRegionFeatures::pedestrianFeatures(instance, k, x);
    f.insert(f.end(), x.begin(), x.end());

    // add the bias term
    f.push_back(1);

    return f;
}

void SceneObjectModel::computePedestrianTrainingFeatures(SceneInstance& instance, SceneFeatureCache &cache)
{
    SVL_FCN_TIC;

    // check if instance is in the cache
    if (gSceneConfig.getOptionAsBool("useCache") && (!instance.name().empty())) {
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "pedestrianFeaturesExt");
        if (svlFileExists(filename.c_str())) {
            svlCodeProfiler::tic(svlCodeProfiler::getHandle("readPedestrianFeatureCache"));
            cache.read(filename.c_str(), true);
            svlCodeProfiler::toc(svlCodeProfiler::getHandle("readPedestrianFeatureCache"));
            SVL_FCN_TOC;
            return;
        }
    }

    // not in cache (or cache not enabled), so compute
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("computePedestrianFeatures"));
    int initialCacheSize = cache.size();

    // groundtruth layer file and labels
    SceneLayersAndLabels layers(instance.width(), instance.height());
    layers.load(gSceneConfig.filename("lblDir", instance.name(), "layerLblExt").c_str(),
        gSceneConfig.filename("lblDir", instance.name(), "layerClassLblExt").c_str());

    instance.assignPixels(layers.layers);

    SceneFeatureCache objDetScoreCache;
    // load in the object detection scores
    if (gSceneConfig.getOptionAsBool("useCache") && (!instance.name().empty())) {
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "objDetScoreCacheExt");
        if (svlFileExists(filename.c_str())) {
            objDetScoreCache.read(filename.c_str(), true);
        }
    }

    // accumulate features and labels
    cache.reserve(cache.size() + instance.numRegions());
    int sampleCount = 0;
    for (int k = 0; k < instance.numRegions(); k++) {
        instance.moveRegions(k, -1); // FIXME: iterate above over objects. calling function should define objects

        if (instance.assignment.regions[k]->size() < 10) continue;
        if (layers.labels[k] < gSceneConfig.numRegionLabels - 1) continue; // ignore non-foreground regions

        int objectLabel = layers.labels[k] - gSceneConfig.numRegionLabels + 1;

        SVL_ASSERT(objectLabel <= gSceneConfig.numObjectLabels);

        int objId = instance.assignment.objects.size() - 1;
        instance.assignment.objects[objId]->Ok = objectLabel;

        if (objDetScoreCache.empty()) {
            cache.features.push_back(computePatchBasedPedestrianFeature(&instance, objId, NULL));
        } else {
            double pedScore = objDetScoreCache.features[sampleCount][0];
            cache.features.push_back(computePatchBasedPedestrianFeature(&instance, objId, &pedScore));
        }

        // TODO: remove hand coded pedestrian id
        objectLabel = objectLabel == 2 ? 1 : 0;
        cache.labels.push_back(objectLabel);

        cache.weights.push_back(1.0);

        sampleCount++;
    }

    if (!objDetScoreCache.empty()) {
        SVL_ASSERT(sampleCount == objDetScoreCache.size());
    }

    svlCodeProfiler::toc(svlCodeProfiler::getHandle("computePedestrianFeatures"));

    // write to cache
    if (gSceneConfig.getOptionAsBool("useCache") && (!instance.name().empty())) {
        svlCodeProfiler::tic(svlCodeProfiler::getHandle("writePedestrianFeatureCache"));
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "pedestrianFeaturesExt");
        if (initialCacheSize < cache.size()) {
            cache.write(filename.c_str(), initialCacheSize, cache.size() - 1, false);
        }
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("writePedestrianFeatureCache"));
    }

    SVL_FCN_TOC;
}


// // SceneObjectContextModel -----------------------------------------------------

SceneObjectContextModel::SceneObjectContextModel() :
    SceneObjectModel(), _objectContextWeight(1.0)
{
    // do nothing
}

SceneObjectContextModel::~SceneObjectContextModel()
{
    // do nothing
}

// i/o
void SceneObjectContextModel::loadModels(const char *modelDir)
{
    SceneObjectModel::loadModels(modelDir);

    loadObjectContextModel(modelDir);
    loadObjectContextWeight(modelDir);
}

void SceneObjectContextModel::loadObjectContextModel(const char *modelDir)
{
    string fileName = string(modelDir) + string("/objectContext.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading object context feature whitening parameters from " << fileName);
    _objectContextFeatureWhitener.load(fileName.c_str());

    fileName = string(modelDir) + string("/objectContext.potential.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading object context parameters from " << fileName);
    _objectContextPotential.load(fileName.c_str());
}

void SceneObjectContextModel::saveObjectContextModel(const char *modelDir)
{
    string fileName = string(modelDir) + string("/objectContext.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing object context feature whitening parameters to " << fileName);
    _objectContextFeatureWhitener.save(fileName.c_str());

    fileName = string(modelDir) + string("/objectContext.potential.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing object context parameters to " << fileName);
    _objectContextPotential.save(fileName.c_str());
}

void SceneObjectContextModel::loadObjectContextWeight(const char *modelDir)
{
    string fileName = string(modelDir) + string("/objectContext.weight.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading object context weight from " << fileName);
    ifstream ifs(fileName.c_str());
    SVL_ASSERT(!ifs.fail());
    ifs >> _objectContextWeight;
    SVL_ASSERT(!ifs.fail());
    ifs.close();
}

void SceneObjectContextModel::saveObjectContextWeight(const char *modelDir)
{
    string fileName = string(modelDir) + string("/objectContext.weight.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing object context weight to " << fileName);
    ofstream ofs(fileName.c_str());
    ofs << _objectContextWeight;
    SVL_ASSERT(!ofs.fail());
    ofs.close();
}

// learning
void SceneObjectContextModel::learnObjectContextModel(const vector<string>& baseNames,
    int maxIterations, bool bResumeTraining)
{
    // baseDirimgDir/baseName.jpg               [H x W x 3]
    // baseDir/lblDir/baseName.layers.txt       [H x W]
    // baseDir/lblDir/baseName.layerLabels.txt  [K x 1]

    // feature values and labels for training
    vector<vector<double> > featureVectors;
    vector<int> featureLabels;
    vector<double> featureWeights;
    int nLabels = gSceneConfig.numRegionLabels * (gSceneConfig.numObjectLabels + 1);
    vector<int> labelCounts(nLabels, 0);

    // accumulate training data
    for (int i = 0; i < (int)baseNames.size(); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "processing image " << baseNames[i] << " ("
            << (i + 1) << " of " << baseNames.size() << ")...");
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        SceneInstance instance(imgFilename.c_str());

        // load layers
        string layerFilename = gSceneConfig.filename("lblDir", baseNames[i], "layerLblExt");
        CvMat *layers = cvCreateMat(instance.height(), instance.width(), CV_32SC1);
        bool bSuccess = readMatrix(layers, layerFilename.c_str());
        if (!bSuccess) {
            cvReleaseMat(&layers);
            SVL_LOG(SVL_LOG_FATAL, "Could not read layers for " << baseNames[i]);
        }

        instance.assignPixels(layers);

        // load layer labels
        string layerClassFilename = gSceneConfig.filename("lblDir", baseNames[i], "layerClassLblExt");
        ifstream ifs(layerClassFilename.c_str());
        SVL_ASSERT_MSG(!ifs.fail(), "could not read layer classes from " << layerClassFilename);
        vector<int> layerLabels(instance.numRegions(), -1);
        for (int k = 0; k < instance.numRegions(); k++) {
            ifs >> layerLabels[k];
            SVL_ASSERT_MSG(!ifs.fail(), "Layer class labels file " << layerClassFilename
                << " is too short. Should have " << instance.numRegions() << " entries.");
        }
        ifs.close();

        // introduce objects
        for (int k = 0; k < instance.numRegions(); k++) {
            instance.assignment.regions[k]->Sk = std::min(layerLabels[k], gSceneConfig.numRegionLabels - 1);
            if (layerLabels[k] >= gSceneConfig.numRegionLabels - 1) {
                instance.moveRegions(k, -1);
                instance.assignment.objects.back()->Ok = layerLabels[k] - gSceneConfig.numRegionLabels;
            }
        }

        // accumulate features and labels for all region-object pairs
        for (vector<SceneObjRegBoundary *>::const_iterator ib = instance.assignment.objectBoundaries.begin();
             ib != instance.assignment.objectBoundaries.end(); ib++) {
            SceneObject *o = (*ib)->object;
            SceneRegion *r = (*ib)->region;
            if (r->Sk < 0) continue; // skip unknown

            int lbl = (o->Ok + 1) * gSceneConfig.numRegionLabels + r->Sk;

            featureVectors.push_back(computeObjectRegionFeatureVector(&instance, o->id(), r->id()));
            featureLabels.push_back(lbl);
            featureWeights.push_back(r->size());
            labelCounts[featureLabels.back()] += 1;
        }

        SVL_LOG(SVL_LOG_VERBOSE, "...accumulated " << toString(labelCounts) << " samples");

        // free memory
        cvReleaseMat(&layers);
    }

    // train feature whitener
    _objectContextFeatureWhitener.clear(featureVectors[0].size());
    _objectContextFeatureWhitener.train(featureVectors);
    _objectContextFeatureWhitener.evaluate(featureVectors);

#if 1
    // balance training data
    for (int i = 0; i < (int)featureWeights.size(); i++) {
        featureWeights[i] /= (double)(labelCounts[featureLabels[i]] + 1.0);
    }
#endif

    // train model
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("trainObjectContext"));
    double lambda = gSceneConfig.getOptionAsDouble("regObjectContext");
    if (!bResumeTraining) {
        _objectContextPotential = svlMultiClassLogistic(featureVectors[0].size(), nLabels, lambda);
    }
    SVL_ASSERT(_objectContextPotential.numFeatures() == (int)featureVectors[0].size());
    SVL_ASSERT(_objectContextPotential.numClasses() == nLabels);
    _objectContextPotential.setOption("maxIterations", maxIterations);
    _objectContextPotential.train(featureVectors, featureLabels, featureWeights);
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("trainObjectContext"));

    // evaluate (on training data)
    if (svlLogger::getLogLevel() <= SVL_LOG_VERBOSE) {
        vector<int> predictedLabels;
        _objectContextPotential.getClassifications(featureVectors, predictedLabels);

        svlConfusionMatrix confusion(nLabels);
        confusion.accumulate(featureLabels, predictedLabels);
        confusion.write(cout);
        SVL_LOG(SVL_LOG_VERBOSE, "Accuracy: " << confusion.accuracy());
    }
}

// inference
double SceneObjectContextModel::inferRegions(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);

    // construct cluster graph
    SVL_LOG(SVL_LOG_VERBOSE, "building graph on regions...");
    svlClusterGraph *graph = buildGraph(*instance);

    // run inference
    SVL_LOG(SVL_LOG_VERBOSE, "running inference on regions...");
    svlFactorOperation::CACHE_INDEX_MAPPING = true;
    svlFactorOperation::USE_SHARED_INDEX_CACHE = false;

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("inference"));
    svlMessagePassingInference infObject(*graph);
    infObject.inference(SVL_MP_ASYNCLOGMAXPROD, 250);
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("inference"));

    // decode MAP
    for (int regId = 0; regId < instance->numRegions(); regId++) {
        SVL_ASSERT(infObject[regId].size() == gSceneConfig.numRegionLabels);
        instance->assignment.regions[regId]->Sk = infObject[regId].indexOfMax();
    }

    bool bMultiplyBySize = gSceneConfig.getOptionAsBool("multiplyObjectScoreBySize");
    for (int objId = 0; objId < instance->numObjects(); objId++) {
        SVL_ASSERT(infObject[instance->numRegions() + objId].size() == gSceneConfig.numObjectLabels + 1);
        instance->assignment.objects[objId]->Ok = infObject[instance->numRegions() + objId].indexOfMax() - 1;
        instance->assignment.objects[objId]->prObject = VectorXd::Zero(gSceneConfig.numObjectLabels + 1);
        for (int i = 0; i < gSceneConfig.numObjectLabels + 1; i++) {
            instance->assignment.objects[objId]->prObject[i] = infObject[objId + instance->numRegions()][i];
        }
        if (bMultiplyBySize) {
            instance->assignment.objects[objId]->prObject /= (double)(instance->assignment.objects[objId]->size() + 1.0);
        }
        SVL_LOG(SVL_LOG_DEBUG, "log p(On) = [" << instance->assignment.objects[objId]->prObject.transpose() << "]");
        instance->assignment.objects[objId]->prObject = (instance->assignment.objects[objId]->prObject.cwise() -
            instance->assignment.objects[objId]->prObject.maxCoeff()).cwise().exp();
        instance->assignment.objects[objId]->prObject /= instance->assignment.objects[objId]->prObject.sum();
        SVL_LOG(SVL_LOG_DEBUG, "p(On) = [" << instance->assignment.objects[objId]->prObject.transpose() << "]");
    }

    // get energy
    double e = clusterGraphEnergy(instance, graph);

    delete graph;
    return e + boundaryPenalty(*instance);
}

// energy for complete assignment
double SceneObjectContextModel::energy(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);

    svlClusterGraph *graph = buildGraph(*instance);
    double e = clusterGraphEnergy(instance, graph);

    // check energy
    if (!isfinite(e)) {
        instance->validateData();
        instance->printScene();
        graph->write(cout);
        SVL_LOG(SVL_LOG_FATAL, "energy is not finite");
    }

    delete graph;
    return e + boundaryPenalty(*instance);
}

// updating instance cache with region singletons and pairwise features
void SceneObjectContextModel::updateInstanceCache(SceneInstance *instance)
{
    // update singleton terms
    SceneObjectModel::updateInstanceCache(instance);

    // update pairwise terms
    // HANDLED BY buildGraph
}

// energy for cluster graph potentials
double SceneObjectContextModel::clusterGraphEnergy(SceneInstance *instance, svlClusterGraph *graph) const
{
    SVL_ASSERT(graph != NULL);

    // extract assignment
    vector<int> assignment(instance->numRegions() + instance->numObjects());
    for (int i = 0; i < instance->numRegions(); i++) {
        assignment[i] = instance->assignment.regions[i]->Sk;
    }
    for (int i = 0; i < instance->numObjects(); i++) {
        assignment[i + instance->numRegions()] = instance->assignment.objects[i]->Ok + 1;
    }

    return graph->getEnergy(assignment);
}

// general CRF functions
svlClusterGraph *SceneObjectContextModel::buildGraph(SceneInstance &instance)
{
    SVL_FCN_TIC;

    // update cached singleton data
    updateInstanceCache(&instance);

    // initialize graph
    vector<int> cards;
    cards.resize(instance.numRegions(), gSceneConfig.numRegionLabels);
    cards.resize(cards.size() + instance.numObjects(), gSceneConfig.numObjectLabels + 1);

    svlClusterGraph *graph = new svlClusterGraph(cards.size(), cards);

    // add unary terms
    for (int k = 0; k < instance.numRegions(); k++) {
        svlClique c; c.insert(k);
        svlFactor phi(k, gSceneConfig.numRegionLabels);

        SceneRegion *region = instance.assignment.regions[k];
        for (int i = 0; i < phi.size(); i++) {
            phi[i] = (double)region->size() * region->phi[i];
        }

        graph->addClique(c, phi);
    }

    bool bMultiplyBySize = gSceneConfig.getOptionAsBool("multiplyObjectScoreBySize");
    for (int o = 0; o < instance.numObjects(); o++) {
        svlClique c; c.insert(o + instance.numRegions());
        svlFactor phi(o + instance.numRegions(), gSceneConfig.numObjectLabels + 1);

        SceneObject *object = instance.assignment.objects[o];
        for (int i = 0; i < phi.size(); i++) {
            phi[i] = object->phi[i];
        }
        if (bMultiplyBySize) {
            phi.scale((double)object->size());
        }

        graph->addClique(c, phi);
    }

    vector<pair<int, int> > edges;
    edges.reserve(4 * instance.numRegions());

    // add pairwise context terms
    double w = gSceneConfig.getOptionAsDouble("objectContextWeight") * _objectContextWeight;
    for (vector<SceneObjRegBoundary *>::const_iterator it = instance.assignment.objectBoundaries.begin();
         it != instance.assignment.objectBoundaries.end(); it++) {

        // don't add terms between object and object
        if ((*it)->region->object == NULL) {
            continue;
        }

        vector<double> f(computeObjectRegionFeatureVector(&instance, (*it)->object->id(), (*it)->region->id()));
        _objectContextFeatureWhitener.evaluate(f);
        vector<double> y;
        //_objectContextPotential.getClassScores(f, y);
        _objectContextPotential.getMarginals(f, y);
        Eigen::Map<VectorXd>(&y[0], y.size()) =
            Eigen::Map<VectorXd>(&y[0], y.size()).cwise().log();

        svlClique c;
        c.insert((*it)->object->id() + instance.numRegions());
        c.insert((*it)->region->id());

        svlFactor phi;
        phi.addVariable((*it)->region->id(), gSceneConfig.numRegionLabels);
        phi.addVariable((*it)->object->id() + instance.numRegions(),
            gSceneConfig.numObjectLabels + 1);

        SVL_ASSERT_MSG(phi.size() == (int)y.size(), phi.size() << " != " << y.size());
        for (int i = 0; i < (int)y.size(); i++) {
            phi[i] = w * y[i];
        }

        edges.push_back(make_pair(graph->numCliques(), phi.variableId(0)));
        edges.push_back(make_pair(graph->numCliques(), phi.variableId(1)));
        graph->addClique(c, phi);
    }

    // connect graph
    graph->connectGraph(edges);

    SVL_FCN_TOC;
    return graph;
}

// compute pairwise region feature
vector<double> SceneObjectContextModel::computeObjectRegionFeatureVector(SceneInstance *instance,
    int objId, int regId)
{
    SVL_ASSERT((instance != NULL) && (objId < instance->numObjects()) &&
        (regId < instance->numRegions()));

    Vector2d r = (instance->assignment.objects[objId]->centroid() -
        instance->assignment.regions[regId]->centroid()).normalized();

    vector<double> f;
    f.push_back(r.x());
    f.push_back(r.y());
    f.push_back(1.0);
    return f;
}

// SceneObjectModelNew -------------------------------------------------

SceneObjectModelNew::SceneObjectModelNew() :
    SceneRegionModel()
{
    // do nothing
}

SceneObjectModelNew::~SceneObjectModelNew()
{
    // free
    for (vector<ScenePatchObjectDetector *>::iterator it = _objClassifiers.begin();
         it != _objClassifiers.end(); it++) {
        delete *it;
    }
    _objClassifiers.clear();
}

// i/o
void SceneObjectModelNew::loadModels(const char *modelDir)
{
    SceneRegionModel::loadModels(modelDir);
    loadObjectDetectors(modelDir);
    loadObjectSingletonModel(modelDir);
}

void SceneObjectModelNew::loadObjectDetectors(const char* modelDir)
{
    // delete existing object detectors
    for (vector<ScenePatchObjectDetector *>::iterator it = _objClassifiers.begin();
         it != _objClassifiers.end(); it++) {
        delete *it;
    }
    _objClassifiers.clear();

    // load object detectors
    set<int> objectIds = gSceneObjectDefs.keys();
    for (set<int>::const_iterator it = objectIds.begin(); it != objectIds.end(); it++) {
        _objClassifiers.push_back(gSceneObjectDefs.createPatchDetector(*it));
    }
}

void SceneObjectModelNew::loadObjectSingletonModel(const char *modelDir)
{
    string fileName = string(modelDir) + string("/object.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading object feature whitening parameters from " << fileName);
    _objectFeatureWhitener.load(fileName.c_str());

    fileName = string(modelDir) + string("/object.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading object singleton parameters from " << fileName);
    _objectSingletons.load(fileName.c_str());
}

void SceneObjectModelNew::saveObjectSingletonModel(const char *modelDir)
{
    string fileName = string(modelDir) + string("/object.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing object feature whitening parameters to " << fileName);
    _objectFeatureWhitener.save(fileName.c_str());

    fileName = string(modelDir) + string("/object.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing object singleton parameters to " << fileName);
    _objectSingletons.save(fileName.c_str());
}

// learning
void SceneObjectModelNew::learnObjectSingletons(const vector<string>& baseNames, int maxIterations)
{
    SVL_ASSERT(false); // TODO
}

void SceneObjectModelNew::updateObjectSingletons(const vector<string>& baseNames, int maxIterations)
{
    SVL_ASSERT(false); // TODO
}

// inference
double SceneObjectModelNew::inferRegions(SceneInstance *instance)
{
    SVL_FCN_TIC;
    SVL_ASSERT(instance != NULL);

    double e = 0.0;

    // update cache with singleton potentials
    updateInstanceCache(instance);

    // maximize each region term
    for (int k = 0; k < instance->numRegions(); k++) {
        // check for empty regions
        if (instance->assignment.regions[k]->empty())
            continue;

        int indx = argmax(instance->assignment.regions[k]->phi);
        // ensure objects are foreground
        if (instance->assignment.regions[k]->object != NULL) {
            indx = gSceneConfig.numObjectLabels - 1;
        }
        instance->assignment.regions[k]->Sk = indx;

        // update energy
        e -= (double)instance->assignment.regions[k]->size() *
            instance->assignment.regions[k]->phi[indx];
    }

    // maximize each object term
    bool bMultiplyBySize = gSceneConfig.getOptionAsBool("multiplyObjectScoreBySize");
    for (int k = 0; k < (int)instance->assignment.objects.size(); k++) {
        // check for empty objects
        if (instance->assignment.objects[k]->empty())
            continue;

        SVL_ASSERT_MSG(instance->assignment.objects[k]->phi.size() == gSceneConfig.numObjectLabels + 1,
            "object " << k << " : " << toString(instance->assignment.objects[k]->regions) << " : "
            << instance->assignment.objects[k]->phi.transpose());
        int indx = argmax(instance->assignment.objects[k]->phi);
        instance->assignment.objects[k]->Ok = indx - 1;

        // update energy
        if (bMultiplyBySize) {
            e -= (double)instance->assignment.objects[k]->size() *
                instance->assignment.objects[k]->phi[indx];
        } else {
            e -= instance->assignment.objects[k]->phi[indx];
        }
    }

    // add boundary penalty to energy
    e += boundaryPenalty(*instance);

    SVL_FCN_TOC;
    return e;
}

// energy for complete assignment
double SceneObjectModelNew::energy(SceneInstance *instance)
{
    vector<double> ec = energyComponents(instance);
    double e = 0.0;
    for (unsigned i = 0; i < ec.size(); i++) {
        e += ec[i];
    }

    return e;
}

vector<double> SceneObjectModelNew::energyComponents(SceneInstance *instance)
{
    SVL_FCN_TIC;
    SVL_ASSERT(instance != NULL);

    // update cache with singleton potentials
    updateInstanceCache(instance);

    // accumulate energy for each region term
    double eRegion = 0.0;
    for (int k = 0; k < instance->numRegions(); k++) {
        // check for empty regions
        if (instance->assignment.regions[k]->empty())
            continue;

        // update energy
        int label = instance->assignment.regions[k]->Sk;
        SVL_ASSERT((label >= 0) && (label < gSceneConfig.numRegionLabels));

        // update energy
        eRegion -= (double)instance->assignment.regions[k]->size() *
            instance->assignment.regions[k]->phi[label];
    }

    // accumulate each object term
    double eObject = 0.0;
    bool bMultiplyBySize = gSceneConfig.getOptionAsBool("multiplyObjectScoreBySize");
    for (int k = 0; k < (int)instance->assignment.objects.size(); k++) {
        // check for empty objects
        if (instance->assignment.objects[k]->empty())
            continue;

        // update energy
        int label = instance->assignment.objects[k]->Ok;
        SVL_ASSERT((label >= -1) && (label < gSceneConfig.numObjectLabels));
        if (bMultiplyBySize) {
            eObject -= (double)instance->assignment.objects[k]->size() *
                instance->assignment.objects[k]->phi[label + 1];
        } else {
            eObject -= instance->assignment.objects[k]->phi[label + 1];
        }
    }

    // add boundary penalty term to energy
    double eBoundary = boundaryPenalty(*instance);

    // assemble energy components
    vector<double> e(3);
    e[0] = eRegion;
    e[1] = eObject;
    e[2] = eBoundary;

    SVL_FCN_TOC;
    return e;
}


// updating instance cache with singletons
void SceneObjectModelNew::updateInstanceCache(SceneInstance *instance)
{
    SVL_FCN_TIC;
    SceneRegionModel::updateInstanceCache(instance);

#if 0
    // DEBUGGING WITH CCM CANDIDATES
    if (_objectFeatureWhitener.size() == 1) {
        for (int k = 0; k < (int)instance->assignment.objects.size(); k++) {
            if (instance->assignment.objects[k]->empty())
                continue;

            instance->assignment.objects[k]->objDetScore = VectorXd::Zero(gSceneConfig.numObjectLabels + 1);
            instance->assignment.objects[k]->prObject = VectorXd::Zero(gSceneConfig.numObjectLabels + 1);
            instance->assignment.objects[k]->phi = VectorXd::Zero(gSceneConfig.numObjectLabels + 1);
        }

        SVL_FCN_TOC;
        return;
    }
#endif


    // update each object singleton
    for (int k = 0; k < (int)instance->assignment.objects.size(); k++) {
        // check for empty objects
        if (instance->assignment.objects[k]->empty())
            continue;

        // update unary potential
        if (instance->assignment.objects[k]->phi.size() == 0) {
            // compute region feature vector
            // (scores from the object detector are stored in the feature vector)
            vector<double> f = computeObjectFeatureVector(instance, k);

            SVL_LOG(SVL_LOG_DEBUG, "...features for candidate window "
                << instance->assignment.objects[k]->extent().transpose()
                << " with " << instance->assignment.objects[k]->size() << " pixels");
            SVL_LOG(SVL_LOG_DEBUG, "f^obj = " << toString(f));

    // add square features
#ifdef LOGISTICSINGLETONMODEL
            int numFeatures = (int)f.size() - 1;
            f.reserve(2 * numFeatures + 1);
            for (int j = 0; j < numFeatures; j++) {
                f.push_back(f[j] * f[j]);
            }
#endif

#if 1
            // whiten features
            _objectFeatureWhitener.evaluate(f);

            // compute marginals
            vector<double> y;
            _objectSingletons.getClassScores(f, y);
#ifndef LOGISTICSINGLETONMODEL
            y.push_back(0.0);
#endif

            instance->assignment.objects[k]->phi = Eigen::Map<VectorXd>(&y[0], y.size());
            instance->assignment.objects[k]->prObject = (Eigen::Map<VectorXd>(&y[0], y.size()).cwise() -
                Eigen::Map<VectorXd>(&y[0], y.size()).maxCoeff()).cwise().exp();
            instance->assignment.objects[k]->prObject /= instance->assignment.objects[k]->prObject.sum();

            // normalize so that non-recognizable objects have score 0
            instance->assignment.objects[k]->phi.cwise() -= instance->assignment.objects[k]->phi[0];
#else
            // DEBUGGING
            instance->assignment.objects[k]->objDetScore = Eigen::Map<VectorXd>(&f[0], f.size());
            instance->assignment.objects[k]->prObject = VectorXd::Zero(gSceneConfig.numObjectLabels + 1);
            instance->assignment.objects[k]->phi = VectorXd::Zero(gSceneConfig.numObjectLabels + 1);
#endif
        }
    }

    SVL_FCN_TOC;
}

// compute region/object feature vectors
vector<double> SceneObjectModelNew::computeObjectFeatureVector(SceneInstance *instance, int objId)
{
    SVL_FCN_TIC;
    vector<double> f;

    // extract image patch
    SceneObject *o = instance->assignment.objects[objId];
    Vector4i e = o->extent();
    CvRect r = cvRect(e[0], e[1], e[2], e[3]);

    list<CvPoint> p = o->pixels();
    CvMat *m = o->mask(instance->width(), instance->height());
    convertInPlace(&m, CV_8UC1);

    // object detector features
    o->objDetScore = VectorXd::Zero(_objClassifiers.size());
    for (int i = 0; i < (int)_objClassifiers.size(); i++) {
        double softWindowScore = _objClassifiers[i]->evaluateRegion(instance->greyImage(), p);
        double fullWindowScore = _objClassifiers[i]->evaluateWindow(instance->greyImage(), r);
        f.push_back(softWindowScore);
        f.push_back(fullWindowScore);

        o->objDetScore[i] = softWindowScore;
    }

    // add appearance features
#if 0
    svlSufficientStats stats(instance->numFilters(), SVL_PSS_DIAG);
    for (list<CvPoint>::const_iterator ip = p.begin(); ip != p.end(); ip++) {
        stats.accumulate(instance->pixelFilterFeatures(ip->x, ip->y));
    }

    for (int i = 0; i < instance->numFilters(); i++) {
        f.push_back(stats.sum(i) / stats.count());
        f.push_back(stats.sum2(i, i) / stats.count() - f.back() * f.back());
    }
#else
    vector<double> v = instance->regionFilterMean(m);
    f.insert(f.end(), v.begin(), v.end());
    v = instance->regionFilterVariance(m);
    f.insert(f.end(), v.begin(), v.end());
#endif

    // appearance features for entire image
    v = instance->regionFilterMean();
    f.insert(f.end(), v.begin(), v.end());
    v = instance->regionFilterVariance();
    f.insert(f.end(), v.begin(), v.end());

    // add "shape" features
    f.push_back((double)r.width / instance->width());
    f.push_back((double)r.height / instance->height());
    f.push_back((double)p.size() / r.width * r.height);
    f.push_back((double)r.height / (double)r.width);
    f.push_back((double)(r.x + r.width) / (2.0 * instance->width()));
    f.push_back((double)(r.y + r.height) / (2.0 * instance->height()));

    int nTop = 0;
    int nBottom = 0;
    int nLeft = 0;
    int nRight = 0;
    for (list<CvPoint>::const_iterator ip = p.begin(); ip != p.end(); ip++) {
        if (ip->x == r.x) nLeft += 1;
        if (ip->x == r.x + r.width - 1) nRight += 1;
        if (ip->y == r.y) nTop += 1;
        if (ip->y == r.y + r.height - 1) nBottom += 1;        
    }
    f.push_back((double)nTop / r.height);
    f.push_back((double)nBottom / r.height);
    f.push_back((double)nLeft / r.width);
    f.push_back((double)nRight / r.width);

    // add bias term
    f.push_back(1.0);

    cvReleaseMat(&m);

    SVL_FCN_TOC;
    return f;
}
