/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneObjectModel.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"
#include "sceneEntities.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneHOGFeatures.h"
#include "sceneRegionObjectClassifier.h"

using namespace std;

// SceneObjectModel ----------------------------------------------------
// models regions comprised of pixels and objects comprised of regions
//

class SceneObjectModel : public SceneRegionModel {
 protected:
    // object detectors
    vector<SceneRegionObjectClassifier *> _objClassifiers;
    svlFeatureWhitener _objectFeatureWhitener;
    svlMultiClassLogistic _objectSingletons;
    
    svlBoostedClassifierSet *_boostedObjectClassifiers;

    // for pedestrian
    //svlLogistic _pedestrianClassifier;
    SceneLinearClassifier _pedestrianClassifier;
    svlFeatureWhitener _pedestrianFeatureWhitener;

 public:
    SceneObjectModel();
    ~SceneObjectModel();

    // i/o
    void loadModels(const char *modelDir);
    void loadObjectDetectors(const char* modelDir);
    void loadObjectSingletonModel(const char *modelDir);
    void saveObjectSingletonModel(const char *modelDir);
    void loadObjectBoostedClassifier(const char *modelDir);
    void saveObjectBoostedClassifier(const char *modelDir, int classId);
    void loadPedestrianClassifier(const char *modelDir);
    void savePedestrianClassifier(const char *modelDir);
    
    // learning
    void learnObjectSingletons(const vector<string>& baseNames, int maxIterations,
        bool bResumeTraining = false, bool bTrainBoostedClassifiers = true);

    // currently only output feature cache (then train a SVM outside)
    void learnPedestrianClassifier(const vector<string>& baseNames);

    //    void learnPedestrianSingleton(const vector<string>& baseNames, int maxIterations);
    
    // inference
    double inferRegions(SceneInstance *instance) { return inferRegions(instance, NULL); }
    double inferRegions(SceneInstance *instance, VectorXd *energyRecord);

    // compute object feature vectors
    vector<double> computeObjectFeatureVector(SceneInstance *instance, int objId);

    // compute pedestrian feature vector for pedestrian classifier
    vector<double> computePatchBasedPedestrianFeature(SceneInstance *instance, int k, double* pPedDetScore = NULL);

    // energy for complete assignment
    double energy(SceneInstance *instance);

 protected:
    // updating instance cache with region singletons
    void updateInstanceCache(SceneInstance *instance);

    // compute region and object feature vectors
    vector<double> computeRegionBasedObjectFeature(SceneInstance *instance, int k);
    vector<double> computePatchBasedObjectFeature(SceneInstance *instance, int k);
    void computeObjectTrainingFeatures(SceneInstance &instance, SceneFeatureCache &cache);

    IplImage* getRegionPatch(Vector4i& regionExt, SceneInstance *instance, double enlargeFactor, 
        CvPoint& shift, double winRatio, int objDetScale);

    // region's pedestrian score based on the pedestrian detector (search around the region)
    double getPedestrianDetectorScore(SceneInstance *instance, int k, int detectorId = -1);

    // for pedestrian classifier learning
    void computePedestrianTrainingFeatures(SceneInstance& instance, SceneFeatureCache &cache);
};

// SceneObjectContextModel ------------------------------------------------
// includes context into above model
//

class SceneObjectContextModel : public SceneObjectModel {
 protected:
    // object context potential
    svlFeatureWhitener _objectContextFeatureWhitener;
    svlMultiClassLogistic _objectContextPotential;
    double _objectContextWeight;

 public:
    SceneObjectContextModel();
    ~SceneObjectContextModel();

    // i/o
    void loadModels(const char *modelDir);

    void loadObjectContextModel(const char *modelDir);
    void saveObjectContextModel(const char *modelDir);
    void loadObjectContextWeight(const char *modelDir);
    void saveObjectContextWeight(const char *modelDir);

    // learning
    void learnObjectContextModel(const vector<string>& baseNames, int maxIterations,
        bool bResumeTraining = false);

    // inference
    double inferRegions(SceneInstance *instance);

    // energy for complete assignment
    double energy(SceneInstance *instance);

 protected:
    // updating instance cache with region singletons and pairwise features
    void updateInstanceCache(SceneInstance *instance);

    // energy for cluster graph potentials
    double clusterGraphEnergy(SceneInstance *instance, svlClusterGraph *graph) const;

    // general CRF functions
    svlClusterGraph *buildGraph(SceneInstance &instance);

    // compute pairwise object-to-region feature
    vector<double> computeObjectRegionFeatureVector(SceneInstance *instance, int objId, int regId);
};

// SceneObjectModelNew -------------------------------------------------
// models regions comprised of pixels and objects comprised of regions
//

#define LOGISTICSINGLETONMODEL

class SceneObjectModelNew : public SceneRegionModel {
 protected:
    // object detectors
    vector<ScenePatchObjectDetector *> _objClassifiers;
    svlFeatureWhitener _objectFeatureWhitener;
#ifdef LOGISTICSINGLETONMODEL
    svlMultiClassLogistic _objectSingletons;
#else
    svlBoostedClassifierSet _objectSingletons;
#endif

 public:
    SceneObjectModelNew();
    ~SceneObjectModelNew();

    // i/o
    void loadModels(const char *modelDir);
    void loadObjectDetectors(const char* modelDir);
    void loadObjectSingletonModel(const char *modelDir);
    void saveObjectSingletonModel(const char *modelDir);
    
    // learning
    void learnObjectSingletons(const vector<string>& baseNames, int maxIterations);
    void updateObjectSingletons(const vector<string>& baseNames, int maxIterations);

    // inference
    double inferRegions(SceneInstance *instance);

    // energy for complete assignment
    double energy(SceneInstance *instance);

    // TODO: make this a virtual function of SceneBaseModel
    vector<double> energyComponents(SceneInstance *instance);

 protected:
    // updating instance cache with region singletons
    void updateInstanceCache(SceneInstance *instance);

    // compute region and object feature vectors
    vector<double> computeObjectFeatureVector(SceneInstance *instance, int objId);
};
