/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    scenePatchObjectDetector.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <map>
#include <set>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "scenePatchObjectDetector.h"

using namespace std;

// Threading -----------------------------------------------------------------

typedef struct _ScenePatchValueArgs {
    // input
    const ScenePatch *patch;
    const IplImage *img;
    const IplImage *imgSum;
    const IplImage *imgSum2;
    // output
    double *featureValue;
} ScenePatchValueArgs;

void *ScenePatchValueThread(void *arg, unsigned threadId)
{
    const ScenePatchValueArgs *p = (ScenePatchValueArgs *)arg;
    *p->featureValue = p->patch->value(p->img, p->imgSum, p->imgSum2);

    return NULL;
}

typedef struct _ScenePatchResponseArgs {
    // input
    const ScenePatch *patch;
    const IplImage *img;
    const IplImage *imgSum;
    const IplImage *imgSum2;
    // output
    CvMat **resp;
} ScenePatchResponseArgs;

void *ScenePatchResponseThread(void *arg, unsigned threadId)
{
    const ScenePatchResponseArgs *p = (ScenePatchResponseArgs *)arg;
    if (*p->resp != NULL) {
        cvReleaseMat(p->resp);
    }
    *p->resp = p->patch->response(p->img, p->imgSum, p->imgSum2);

    return NULL;
}

// ScenePatch ----------------------------------------------------------------

ScenePatch::ScenePatch() :
    _patch(NULL)
{
    _location = cvRect(0, 0, 0, 0);
}

ScenePatch::ScenePatch(IplImage *img, const CvRect &region) :
    _patch(NULL)
{
    SVL_ASSERT(img->nChannels == 1);

    // clip out region
    cvSetImageROI(img, region);
    IplImage *clip = cvCreateImage(cvSize(region.width, region.height), img->depth, 1);
    cvCopyImage(img, clip);
    cvResetImageROI(img);

    // convert to patch
    _patch = cvCreateMat(clip->height, clip->width, CV_32FC1);
    cvConvert(clip, _patch);
    scaleToRange(_patch, 0.0, 1.0);

    // free clipped region
    cvReleaseImage(&clip);

    // assign location
    _location.x = std::max(region.x - 3, 0);
    _location.y = std::max(region.y - 3, 0);
    _location.width = 7;
    if (_location.x + _location.width > img->width - region.width) {
        _location.width = img->width - region.width - _location.x;
    }
    _location.height = 7;
    if (_location.y + _location.height > img->height - region.height) {
        _location.height = img->height - region.height - _location.y;
    }
}

ScenePatch::ScenePatch(const ScenePatch& p) :
    _patch(NULL), _location(p._location)
{
    if (p._patch != NULL) {
        _patch = cvCloneMat(p._patch);
    }
}

ScenePatch::~ScenePatch()
{
    clear();
}

double ScenePatch::value(const IplImage *img, CvPoint offset) const
{
    /*
    SVL_ASSERT((img != NULL) && (img->nChannels == 1) && (img->depth == IPL_DEPTH_32F));
    SVL_ASSERT_MSG(img->height >= _location.y + _location.height + _patch->rows,
        img->height << " !>= " << _location.y << " + " << _location.height << " + " << _patch->rows);
    SVL_ASSERT_MSG(img->width >= _location.x + _location.width + _patch->cols,
        img->width << " !>= " << _location.x << " + " << _location.width << " + " << _patch->cols);
    */

    float maxValue = -SVL_FLT_MAX;

    for (int y = _location.y; y < _location.y + _location.height; y++) {
        for (int x = _location.x; x < _location.x + _location.width; x++) {
            const float *p = (const float *)&CV_IMAGE_ELEM(img, float, y + offset.y, x + offset.x);
            const float *q = (const float *)_patch->data.ptr;

            float crossCorrelation = 0.0f;
            float energy = 0.01f;
            for (int i = 0; i < _patch->height; i++) {
#if 0
                for (int j = 0; j < _patch->width; j++) {
                    crossCorrelation += q[j] * p[j];
                    energy += p[j] * p[j];
                }

                q += _patch->width;
                p += img->widthStep;
#else
                for (int j = 0; j < _patch->width - 1; j += 2) {
                    crossCorrelation += q[j] * p[j] + q[j + 1] * p[j + 1];
                    energy += p[j] * p[j] + p[j + 1] * p[j + 1];
                }
                if (_patch->width % 2 == 1) {
                    crossCorrelation += q[_patch->width - 1] * p[_patch->width - 1];
                    energy += p[_patch->width - 1] * p[_patch->width - 1];
                }

                q += _patch->width;
                p += img->widthStep / sizeof(float);
#endif
            }

            float v = crossCorrelation / sqrt(energy);
            if (v > maxValue) maxValue = v;
        }
    }

    SVL_ASSERT(maxValue > -SVL_FLT_MAX);
    return (double)maxValue;
}

double ScenePatch::value(const IplImage *img, const IplImage *imgSum,
    const IplImage *imgSum2, CvPoint offset) const
{
    float maxValue = -SVL_FLT_MAX;

    for (int y = _location.y; y < _location.y + _location.height; y++) {
        for (int x = _location.x; x < _location.x + _location.width; x++) {
            const float *p = (const float *)&CV_IMAGE_ELEM(img, float, y + offset.y, x + offset.x);
            const float *q = (const float *)_patch->data.ptr;

            float crossCorrelation = 0.0f;
            float energy = (float)(CV_IMAGE_ELEM(imgSum2, double, y + offset.y, x + offset.x) +
                CV_IMAGE_ELEM(imgSum2, double, y + offset.y + _patch->height, x + offset.x + _patch->width) -
                CV_IMAGE_ELEM(imgSum2, double, y + offset.y + _patch->height, x + offset.x) -
                CV_IMAGE_ELEM(imgSum2, double, y + offset.y, x + offset.x + _patch->width));

            for (int i = 0; i < _patch->height; i++) {
                for (int j = 0; j < _patch->width - 1; j += 2) {
                    crossCorrelation += q[j] * p[j] + q[j + 1] * p[j + 1];
                }
                if (_patch->width % 2 == 1) {
                    crossCorrelation += q[_patch->width - 1] * p[_patch->width - 1];
                }

                q += _patch->width;
                p += img->widthStep / sizeof(float);
            }

            float v = crossCorrelation / sqrt(energy + 0.01f);
            if (v > maxValue) maxValue = v;
        }
    }

    SVL_ASSERT(maxValue > -SVL_FLT_MAX);
    return (double)maxValue;
}

double ScenePatch::value(const CvMat *resp, CvPoint offset) const
{
    float maxValue = -SVL_FLT_MAX;

    for (int y = _location.y; y < _location.y + _location.height; y++) {
        float *r = (float *)CV_MAT_ELEM_PTR(*resp, y + offset.y, offset.x);
        for (int x = _location.x; x < _location.x + _location.width; x++) {
            if (r[x] > maxValue)
                maxValue = r[x];
        }
    }

    SVL_ASSERT(maxValue > -SVL_FLT_MAX);
    return maxValue;
}

CvMat *ScenePatch::response(const IplImage *img) const
{
    CvMat *resp = cvCreateMat(img->height - _patch->height + 1,
        img->width - _patch->width + 1, CV_32FC1);
    cvZero(resp);

    float *r = (float *)CV_MAT_ELEM_PTR(*resp, 0, 0);
    for (int y = 0; y < resp->height; y++) {
        for (int x = 0; x < resp->width; x++, r++) {

            const float *p = (const float *)&CV_IMAGE_ELEM(img, float, y, x);
            const float *q = (const float *)_patch->data.ptr;

            float crossCorrelation = 0.0f;
            float energy = 0.01f;

            for (int i = 0; i < _patch->height; i++) {
                for (int j = 0; j < _patch->width - 1; j += 2) {
                    crossCorrelation += q[j] * p[j] + q[j + 1] * p[j + 1];
                    energy += p[j] * p[j] + p[j + 1] * p[j + 1];
                }
                if (_patch->width % 2 == 1) {
                    crossCorrelation += q[_patch->width - 1] * p[_patch->width - 1];
                    energy += p[_patch->width - 1] * p[_patch->width - 1];
                }

                q += _patch->width;
                p += img->widthStep / sizeof(float);
            }

            *r = crossCorrelation / sqrt(energy);
        }
    }

    return resp;
}

CvMat *ScenePatch::response(const IplImage *img, const IplImage *imgSum,
    const IplImage *imgSum2) const
{
    CvMat *resp = cvCreateMat(img->height - _patch->height + 1,
        img->width - _patch->width + 1, CV_32FC1);
    cvZero(resp);

    float *r = (float *)CV_MAT_ELEM_PTR(*resp, 0, 0);
    for (int y = 0; y < resp->height; y++) {
        for (int x = 0; x < resp->width; x++, r++) {

            const float *p = (const float *)&CV_IMAGE_ELEM(img, float, y, x);
            const float *q = (const float *)_patch->data.ptr;

            float crossCorrelation = 0.0f;
            float energy = (float)(CV_IMAGE_ELEM(imgSum2, double, y, x) +
                CV_IMAGE_ELEM(imgSum2, double, y + _patch->height, x + _patch->width) -
                CV_IMAGE_ELEM(imgSum2, double, y + _patch->height, x) -
                CV_IMAGE_ELEM(imgSum2, double, y, x + _patch->width));

            for (int i = 0; i < _patch->height; i++) {
                for (int j = 0; j < _patch->width - 1; j += 2) {
                    crossCorrelation += q[j] * p[j] + q[j + 1] * p[j + 1];
                }
                if (_patch->width % 2 == 1) {
                    crossCorrelation += q[_patch->width - 1] * p[_patch->width - 1];
                }

                q += _patch->width;
                p += img->widthStep / sizeof(float);
            }

            *r = crossCorrelation / sqrt(energy + 0.01f);
        }
    }

    return resp;
}

// i/o
void ScenePatch::read(istream& is)
{
    // clear existing
    clear();

    // read patch
    int width, height;
    is.read((char *)&width, sizeof(int));
    is.read((char *)&height, sizeof(int));
    _patch = cvCreateMat(height, width, CV_32FC1);

    is.read((char *)_patch->data.ptr, width * height * sizeof(float));

    // read location
    is.read((char *)&_location.x, sizeof(int));
    is.read((char *)&_location.y, sizeof(int));
    is.read((char *)&_location.width, sizeof(int));
    is.read((char *)&_location.height, sizeof(int));
}

void ScenePatch::write(ostream& os) const
{
    // write patch
    int width = _patch->cols;
    int height = _patch->rows;
    os.write((char *)&width, sizeof(int));
    os.write((char *)&height, sizeof(int));
    os.write((char *)_patch->data.ptr, width * height * sizeof(float));

    // write location
    os.write((char *)&_location.x, sizeof(int));
    os.write((char *)&_location.y, sizeof(int));
    os.write((char *)&_location.width, sizeof(int));
    os.write((char *)&_location.height, sizeof(int));
}

IplImage *ScenePatch::visualize() const
{
    IplImage *canvas = cvCreateImage(cvSize(_patch->cols, _patch->rows), IPL_DEPTH_8U, 1);
    cvConvertScale(_patch, canvas, 255.0);
    return canvas;
}

ScenePatch& ScenePatch::operator=(const ScenePatch& p)
{
    if (&p != this) {
        clear();
        _location = p._location;
        _patch = cvCloneMat(p._patch);
    }

    return *this;
}

// testing
void ScenePatch::test(IplImage *img)
{
    SVL_ASSERT(img != NULL);
    SVL_LOG(SVL_LOG_MESSAGE, "testing ScenePatch with image " << toString(*img) << "...");

    // convert image to float
    IplImage *fltImg = cvCloneImage(img);
    convertInPlace(&fltImg, IPL_DEPTH_32F);
    SVL_LOG(SVL_LOG_MESSAGE, "...image converted to " << toString(*fltImg) << "...");

    // create patch
    static const int minWidth = 4;
    static const int minHeight = 4;
    const int maxWidth = fltImg->width / 2;
    const int maxHeight = fltImg->height / 2;

    CvRect r;
    r.width = rand() % (maxWidth - minWidth) + minWidth;
    r.height = rand() % (maxHeight - minHeight) + minHeight;
    r.x = rand() % (fltImg->width - r.width);
    r.y = rand() % (fltImg->height - r.height);

    r = cvRect(0, 0, 4, 4);

    SVL_LOG(SVL_LOG_MESSAGE, "...taking patch as location " << toString(r));
    ScenePatch patch(fltImg, r);

    // test value functions
    IplImage *imgSum = cvCreateImage(cvSize(fltImg->width + 1, fltImg->height + 1),
        IPL_DEPTH_64F, fltImg->nChannels);
    IplImage *imgSumSq = cvCreateImage(cvSize(fltImg->width + 1, fltImg->height + 1),
        IPL_DEPTH_64F, fltImg->nChannels);
    cvIntegral(fltImg, imgSum, imgSumSq);

    CvMat *resp = patch.response(fltImg);
    SVL_LOG(SVL_LOG_MESSAGE, "...first response map is " << resp->rows << "x" << resp->cols);
    CvMat *resp2 = patch.response(fltImg, imgSum, imgSumSq);
    SVL_LOG(SVL_LOG_MESSAGE, "...second response map is " << resp2->rows << "x" << resp2->cols);
    double respError = cvNorm(resp, resp2, CV_L1);
    SVL_LOG(SVL_LOG_MESSAGE, "...avg. response difference: "
        << respError / (resp->rows * resp->cols));

    r.height += 7;
    r.width += 7;

    double v1 = 0.0, v2 = 0.0, v3 = 0.0;
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("patch.value(fltImg)"));
    for (int y = 0; y < fltImg->height - r.y - r.height; y++) {
        for (int x = 0; x < fltImg->width - r.x - r.width; x++) {
            v1 += patch.value(fltImg, cvPoint(x, y));
        }
    }
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("patch.value(fltImg)"));
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("patch.value(fltImg, imgSum, imgSumSq)"));
    for (int y = 0; y < fltImg->height - r.y - r.height; y++) {
        for (int x = 0; x < fltImg->width - r.x - r.width; x++) {
            v2 += patch.value(fltImg, imgSum, imgSumSq, cvPoint(x, y));
        }
    }
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("patch.value(fltImg, imgSum, imgSumSq)"));
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("patch.value(resp)"));
    for (int y = 0; y < fltImg->height - r.y - r.height; y++) {
        for (int x = 0; x < fltImg->width - r.x - r.width; x++) {
            v3 += patch.value(resp, cvPoint(x, y));
        }
    }
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("patch.value(resp)"));

    SVL_LOG(SVL_LOG_MESSAGE, "...value results " << v1 << ", " << v2 << ", " << v3);

    // free memory
    cvReleaseMat(&resp2);
    cvReleaseMat(&resp);
    cvReleaseImage(&imgSumSq);
    cvReleaseImage(&imgSum);
    cvReleaseImage(&fltImg);
}

void ScenePatch::clear()
{
    if (_patch != NULL) {
        cvReleaseMat(&_patch);
    }
}


// ScenePatchObjectDetector --------------------------------------------------

ScenePatchObjectDetector::ScenePatchObjectDetector(int w, int h) :
    _baseWidth(w), _baseHeight(h), _classifier(NULL), _objectName("unknown")
{
    // do nothing
}

ScenePatchObjectDetector::ScenePatchObjectDetector(const char *dictFilename,
    const char *classifierFilename) :
    _baseWidth(0), _baseHeight(0), _classifier(NULL), _objectName("unknown")
{
    // load models
    SVL_ASSERT(dictFilename != NULL);
    readDictionary(dictFilename);

    if (classifierFilename != NULL)
        readClassifier(classifierFilename);
}

ScenePatchObjectDetector::ScenePatchObjectDetector(const ScenePatchObjectDetector& d)
{
    SVL_ASSERT(false);
}

ScenePatchObjectDetector::~ScenePatchObjectDetector()
{
    if (_classifier != NULL)
        delete _classifier;
}

// i/o
void ScenePatchObjectDetector::clear()
{
    if (_classifier != NULL) {
        delete _classifier;
        _classifier = NULL;
    }

    _patchDictionary.clear();
}

void ScenePatchObjectDetector::readDictionary(const char *filename)
{
    _patchDictionary.clear();

    ifstream ifs(filename, ios::binary);
    SVL_ASSERT(!ifs.fail());

    ifs.read((char *)&_baseWidth, sizeof(int));
    ifs.read((char *)&_baseHeight, sizeof(int));

    int nPatches;
    ifs.read((char *)&nPatches, sizeof(int));
    _patchDictionary.resize(nPatches);
    for (int i = 0; i < nPatches; i++) {
        ifs.read((char *)&_patchDictionary[i].second, sizeof(int));
        _patchDictionary[i].first.read(ifs);
    }

    SVL_ASSERT(!ifs.fail());
    ifs.close();
}

void ScenePatchObjectDetector::readClassifier(const char *filename)
{
    if (_classifier == NULL) {
        _classifier = new svlBoostedClassifier(filename);
    } else {
        _classifier->load(filename);
    }
}

void ScenePatchObjectDetector::writeDictionary(const char *filename) const
{
    ofstream ofs(filename, ios::binary);
    SVL_ASSERT(!ofs.fail());

    ofs.write((char *)&_baseWidth, sizeof(int));
    ofs.write((char *)&_baseHeight, sizeof(int));

    int nPatches = (int)_patchDictionary.size();
    ofs.write((char *)&nPatches, sizeof(int));
    for (int i = 0; i < nPatches; i++) {
        ofs.write((char *)&_patchDictionary[i].second, sizeof(int));
        _patchDictionary[i].first.write(ofs);
    }

    SVL_ASSERT(!ofs.fail());
    ofs.close();
}

void ScenePatchObjectDetector::writeClassifier(const char *filename) const
{
    SVL_ASSERT(_classifier != NULL);
    _classifier->save(filename);
}

void ScenePatchObjectDetector::trimDictionary()
{
    SVL_ASSERT(_classifier != NULL);

    vector<int> activeVars = ((svlBoostedClassifier *)_classifier)->reorderActiveVars();
    SVL_LOG(SVL_LOG_DEBUG, "active patches: " << toString(activeVars));

    SVL_LOG(SVL_LOG_MESSAGE, "reducing patch dictionary from " << _patchDictionary.size()
        << " to " << activeVars.size() << " entries");
    vector<pair<ScenePatch, int> > reorderedDictionary;
    reorderedDictionary.reserve(activeVars.size());
    for (vector<int>::const_iterator it = activeVars.begin(); it != activeVars.end(); it++) {
        reorderedDictionary.push_back(_patchDictionary[*it]);
    }

    _patchDictionary = reorderedDictionary;
}

IplImage *ScenePatchObjectDetector::visualizeDictionary() const
{
    vector<IplImage *> patches(_patchDictionary.size(), NULL);
    for (unsigned i = 0; i < _patchDictionary.size(); i++) {
        // draw patch
        IplImage *tmp = _patchDictionary[i].first.visualize();
        resizeInPlace(&tmp, _baseHeight, _baseWidth);
        patches[i] = colorImage(tmp);
        cvReleaseImage(&tmp);

        // indicate channel
        for (int j = 0; j <= _patchDictionary[i].second; j++) {
            cvLine(patches[i], cvPoint(_baseWidth - 2 * j - 2, _baseHeight - 2),
                cvPoint(_baseWidth - 2 * j - 2, _baseHeight - 3), CV_RGB(255, 0, 0));
        }
    }

    IplImage *canvas = combineImages(patches);
    svlReleaseImages(patches);

    return canvas;
}

// learning
void ScenePatchObjectDetector::learnPatchDictionary(const char *imgDir, int maxSize)
{
    SVL_FCN_TIC;
    vector<string> trainingImages = svlDirectoryListing(imgDir, NULL, true, true);
    SVL_LOG(SVL_LOG_VERBOSE, "found " << trainingImages.size() << " training images");
    int nSamples = std::max(2, (int)trainingImages.size() / (2 * maxSize));

    // build dictionary
    for (int i = 0; i < (int)trainingImages.size(); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "processing image " << i << " of " << trainingImages.size()
            << " (" << _patchDictionary.size() << " entries in dictionary)");

        // check maximum size
        if (_patchDictionary.size() >= (unsigned)maxSize)
            break;

        // load and resize image
        IplImage *img = cvLoadImage(trainingImages[i].c_str(), CV_LOAD_IMAGE_GRAYSCALE);
        if (img == NULL) {
            SVL_LOG(SVL_LOG_ERROR, "could not load image " << trainingImages[i]);
            continue;
        }

        if ((img->width != _baseWidth) || (img->height != _baseHeight)) {
            SVL_LOG(SVL_LOG_DEBUG, "... resizing to " << _baseWidth << " x " << _baseHeight);
            resizeInPlace(&img, _baseHeight, _baseWidth);
        }

        vector<IplImage *> channels = createImageChannels(img);

        // extract patches
        for (int n = 0; n < nSamples; n++) {
            for (int c = 0; c < (int)channels.size(); c++) {
                CvRect region = randomRectangle();
                ScenePatch patch(channels[c], region);
                _patchDictionary.push_back(make_pair(patch, c));
            }
        }

        // free memory
        svlReleaseImages(channels);
        cvReleaseImage(&img);
    }

    SVL_FCN_TOC;
}

void ScenePatchObjectDetector::learnClassifier(const char *posImgDir,
    const char *negImgDir, int maxSamples)
{
    SVL_FCN_TIC;

    vector<string> posImages = svlDirectoryListing(posImgDir, NULL, true, true);
    vector<string> negImages = svlDirectoryListing(negImgDir, NULL, true, true);
    SVL_LOG(SVL_LOG_MESSAGE, "found " << posImages.size() << " positive and "
        << negImages.size() << " negative training images (limiting to " << maxSamples << ")");
    SVL_ASSERT(posImages.size() && negImages.size());

    vector<vector<double> > posSamples;
    vector<vector<double> > negSamples;

    // process positive images
    for (int i = 0; i < std::min((int)posImages.size(), maxSamples); i++) {
        cerr << "--- " << i << " (+ve)...\r";
        SVL_LOG(SVL_LOG_VERBOSE, "processing positive image " << i << " of " << posImages.size());

        // load and resize image
        IplImage *img = cvLoadImage(posImages[i].c_str(), CV_LOAD_IMAGE_GRAYSCALE);
        if (img == NULL) {
            SVL_LOG(SVL_LOG_ERROR, "could not load image " << posImages[i]);
            continue;
        }

        if ((img->width != _baseWidth) || (img->height != _baseHeight)) {
            SVL_LOG(SVL_LOG_DEBUG, "... resizing to " << _baseWidth << " x " << _baseHeight);
            resizeInPlace(&img, _baseHeight, _baseWidth);
        }

        vector<IplImage *> channels = createImageChannels(img);

        posSamples.push_back(computeFeatureVector(channels));

        // free memory
        svlReleaseImages(channels);
        cvReleaseImage(&img);
    }

    // process negative images
    for (int i = 0; i < std::min((int)negImages.size(), maxSamples); i++) {
        cerr << "--- " << i << " (-ve)...\r";
        SVL_LOG(SVL_LOG_VERBOSE, "processing negative image " << i << " of " << negImages.size());

        // load and resize image
        IplImage *img = cvLoadImage(negImages[i].c_str(), CV_LOAD_IMAGE_GRAYSCALE);
        if (img == NULL) {
            SVL_LOG(SVL_LOG_ERROR, "could not load image " << negImages[i]);
            continue;
        }

        if ((img->width != _baseWidth) || (img->height != _baseHeight)) {
            SVL_LOG(SVL_LOG_DEBUG, "... resizing to " << _baseWidth << " x " << _baseHeight);
            resizeInPlace(&img, _baseHeight, _baseWidth);
        }

        vector<IplImage *> channels = createImageChannels(img);

        negSamples.push_back(computeFeatureVector(channels));

        // free memory
        svlReleaseImages(channels);
        cvReleaseImage(&img);
    }

    // learn classifier
    if (_classifier != NULL) {
        delete _classifier;
    }

    _classifier = new svlBoostedClassifier();
    _classifier->train(posSamples, negSamples);

    SVL_FCN_TOC;
}

void ScenePatchObjectDetector::validateClassifier(const char *posImgDir,
    const char *negImgDir, int maxSamples)
{
    SVL_FCN_TIC;
    SVL_ASSERT(_classifier != NULL);

    vector<string> posImages = svlDirectoryListing(posImgDir, NULL, true, true);
    vector<string> negImages = svlDirectoryListing(negImgDir, NULL, true, true);
    SVL_LOG(SVL_LOG_MESSAGE, "found " << posImages.size() << " positive and "
        << negImages.size() << " negative validation images (limiting to " << maxSamples << ")");
    SVL_ASSERT(posImages.size() && negImages.size());

    vector<vector<double> > posSamples;
    vector<vector<double> > negSamples;

    // process positive images
    for (int i = 0; i < std::min((int)posImages.size(), maxSamples); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "processing positive image " << i << " of " << posImages.size());

        // load and resize image
        IplImage *img = cvLoadImage(posImages[i].c_str(), CV_LOAD_IMAGE_GRAYSCALE);
        if (img == NULL) {
            SVL_LOG(SVL_LOG_ERROR, "could not load image " << posImages[i]);
            continue;
        }

        if ((img->width != _baseWidth) || (img->height != _baseHeight)) {
            SVL_LOG(SVL_LOG_DEBUG, "... resizing to " << _baseWidth << " x " << _baseHeight);
            resizeInPlace(&img, _baseHeight, _baseWidth);
        }

        vector<IplImage *> channels = createImageChannels(img);

        posSamples.push_back(computeFeatureVector(channels));

        // free memory
        svlReleaseImages(channels);
        cvReleaseImage(&img);
    }

    // process negative images
    for (int i = 0; i < std::min((int)negImages.size(), maxSamples); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "processing negative image " << i << " of " << negImages.size());

        // load and resize image
        IplImage *img = cvLoadImage(negImages[i].c_str(), CV_LOAD_IMAGE_GRAYSCALE);
        if (img == NULL) {
            SVL_LOG(SVL_LOG_ERROR, "could not load image " << negImages[i]);
            continue;
        }

        if ((img->width != _baseWidth) || (img->height != _baseHeight)) {
            SVL_LOG(SVL_LOG_DEBUG, "... resizing to " << _baseWidth << " x " << _baseHeight);
            resizeInPlace(&img, _baseHeight, _baseWidth);
        }

        vector<IplImage *> channels = createImageChannels(img);

        negSamples.push_back(computeFeatureVector(channels));

        // free memory
        svlReleaseImages(channels);
        cvReleaseImage(&img);
    }

    // validate classifier
    _classifier->validate(posSamples, negSamples);

    SVL_FCN_TOC;
}


// inference
double ScenePatchObjectDetector::evaluateWindow(const IplImage *img) const
{
    SVL_ASSERT((img != NULL) && (_classifier != NULL));
    SVL_FCN_TIC;

    vector<IplImage *> channels;
    IplImage *resizedImg = NULL;

    if ((img->width != _baseWidth) || (img->height != _baseHeight)) {
        resizedImg = cvCreateImage(cvSize(_baseWidth, _baseHeight), IPL_DEPTH_8U, img->nChannels);
        cvResize(img, resizedImg);
        channels = createImageChannels(resizedImg);
    } else {
        channels = createImageChannels(img);
    }

    double score = _classifier->getScore(computeFeatureVector(channels));

    svlReleaseImages(channels);
    if (resizedImg != NULL)
        cvReleaseImage(&resizedImg);

    SVL_FCN_TOC;
    return score;
}

double ScenePatchObjectDetector::evaluateWindow(IplImage *img, const CvRect& region) const
{
    SVL_ASSERT((img != NULL) && (_classifier != NULL));
    SVL_FCN_TIC;

    IplImage *croppedImage = cvCreateImage(cvSize(region.width, region.height), IPL_DEPTH_8U, img->nChannels);
    cvSetImageROI(img, region);
    cvCopyImage(img, croppedImage);
    cvResetImageROI(img);

    resizeInPlace(&croppedImage, _baseHeight, _baseWidth);

    vector<IplImage *> channels = createImageChannels(croppedImage);
    double score = _classifier->getScore(computeFeatureVector(channels));
    svlReleaseImages(channels);

    cvReleaseImage(&croppedImage);

    SVL_FCN_TOC;
    return score;
}

vector<double> ScenePatchObjectDetector::evaluateWindows(IplImage *img, const vector<CvRect>& regions) const
{
    vector<double> scores(regions.size());
    for (int i = 0; i < (int)regions.size(); i++) {
        scores[i] = evaluateWindow(img, regions[i]);
    }

    return scores;
}

vector<double> ScenePatchObjectDetector::evaluateWindows(IplImage *img, const svlObject2dFrame& regions) const
{
    vector<double> scores(regions.size());
    for (int i = 0; i < (int)regions.size(); i++) {
        CvRect r = cvRect((int)(regions[i].x + 0.5), (int)(regions[i].y + 0.5),
            (int)(regions[i].w + 0.5), (int)(regions[i].h + 0.5));
        scores[i] = evaluateWindow(img, r);
    }

    return scores;
}

svlObject2dFrame ScenePatchObjectDetector::evaluateImage(IplImage *img) const
{
    SVL_ASSERT(img != NULL);
    svlObject2dFrame detections;

    double scale = 1.0;
    while ((scale * img->width >= _baseWidth) && (scale * img->height >= _baseHeight)) {
        svlObject2dFrame d = evaluateImage(img, scale);
        detections.insert(detections.end(), d.begin(), d.end());
        scale *= 0.8;
    }

    return detections;
}

svlObject2dFrame ScenePatchObjectDetector::evaluateImage(IplImage *img, double scale) const
{
    SVL_FCN_TIC;
    svlObject2dFrame detections;

    int width = (int)(scale * img->width);
    int height = (int)(scale * img->height);
    if ((width < _baseWidth) || (height < _baseHeight)) {
        SVL_LOG(SVL_LOG_WARNING, "image " << toString(img) << " at scale " << scale <<
            " is smaller than base size " << _baseWidth << "-by-" << _baseHeight);
        return detections;
    }

    SVL_LOG(SVL_LOG_DEBUG, "...evaluating on " << width << "-by-" << height << " (scale: " << scale << ")");

    // compute channels
    IplImage *scaledImage = cvCreateImage(cvSize(width, height), img->depth, img->nChannels);
    cvResize(img, scaledImage, CV_INTER_LINEAR);
    vector<IplImage *> channels = createImageChannels(scaledImage);
    cvReleaseImage(&scaledImage);

    vector<IplImage *> channelSums;
    vector<IplImage *> channelSumSq;

    for (int c = 0; c < (int)channels.size(); c++) {
	channelSums.push_back(cvCreateImage(cvSize(channels[c]->width + 1, channels[c]->height + 1),
                IPL_DEPTH_64F, channels[c]->nChannels));
        channelSumSq.push_back(cvCreateImage(cvSize(channels[c]->width + 1, channels[c]->height + 1),
                IPL_DEPTH_64F, channels[c]->nChannels));
	cvIntegral(channels[c], channelSums[c], channelSumSq[c]);
    }

    // compute response images
    vector<CvMat *> responses(_patchDictionary.size(), NULL);
#if 1
    // unthreaded
    for (int i = 0; i < (int)_patchDictionary.size(); i++) {
        int ch = _patchDictionary[i].second;
        responses[i] = _patchDictionary[i].first.response(channels[ch],
            channelSums[ch], channelSumSq[ch]);
    }
#else
    // threaded
    svlThreadPool threadPool;
    vector<ScenePatchResponseArgs> threadArgs;
    threadArgs.resize(_patchDictionary.size());
    threadPool.start();
    for (int i = 0; i < (int)_patchDictionary.size(); i++) {
        int ch = _patchDictionary[i].second;
        threadArgs[i].patch = &_patchDictionary[i].first;
        threadArgs[i].img = channels[ch];
        threadArgs[i].imgSum = channelSums[ch];
        threadArgs[i].imgSum2 = channelSumSq[ch];
        threadArgs[i].resp = &responses[i];
        threadPool.addJob(&ScenePatchResponseThread, &threadArgs[i]);
    }
    threadPool.finish();
#endif

    // scan image
    vector<double> f(_patchDictionary.size(), 0.0);
    for (int y = 0; y < height - _baseHeight + 1; y += 4) {
        for (int x = 0; x < width - _baseWidth + 1; x += 4) {
            for (int i = 0; i < (int)_patchDictionary.size(); i++) {
                f[i] = _patchDictionary[i].first.value(responses[i], cvPoint(x, y));
            }

            double score = _classifier->getScore(f);
            detections.push_back(svlObject2d(x, y, _baseWidth, _baseHeight, score));
            detections.back().scale(1.0 / scale);
            detections.back().name = _objectName;
        }
    }

    // free memory
    svlReleaseMatrices(responses);
    svlReleaseImages(channelSumSq);
    svlReleaseImages(channelSums);
    svlReleaseImages(channels);

    SVL_FCN_TOC;
    return detections;
}

double ScenePatchObjectDetector::evaluateRegion(IplImage *img, const CvMat *mask) const
{
    SVL_ASSERT((img != NULL) && (mask != NULL));
    SVL_ASSERT((img->width == mask->cols) && (img->height == mask->rows));
    SVL_FCN_TIC;

    // create soft region mask
    CvMat *softMask = cvCreateMat(mask->rows, mask->cols, CV_32FC1);
#if 0
    cvConvertScale(mask, softMask);
    IplConvKernel* kernel = cvCreateStructuringElementEx(5, 5, 2, 2, CV_SHAPE_RECT);
    cvDilate(softMask, softMask, kernel);
    cvReleaseStructuringElement(&kernel);
    cvSmooth(softMask, softMask, CV_GAUSSIAN, 3);
#else
    CvMat *invMask = cvCreateMat(mask->rows, mask->cols, CV_8UC1);
    cvConvertScale(mask, invMask, -1, 1);
    cvDistTransform(invMask, softMask, CV_DIST_L1);
    cvReleaseMat(&invMask);
    cvConvertScale(softMask, softMask, -0.1);
    cvExp(softMask, softMask);
#endif

    // get extent of region
    CvRect r = cvRect(img->width, img->height, 0, 0);
    const float *p = (const float *)CV_MAT_ELEM_PTR(*softMask, 0, 0);
    for (int y = 0; y < img->height; y++) {
        for (int x = 0; x <img->width; x++, p++) {
            if (*p > 0.1) {
                r.x = std::min(r.x, x);
                r.width = std::max(r.width, x);
                r.y = std::min(r.y, y);
                r.height = std::max(r.height, y);
            }
        }
    }

    r.width = r.width - r.x + 1;
    r.height = r.height - r.y + 1;
    SVL_ASSERT((r.width > 0) && (r.height > 0));

    // correct aspect ratio
    if (r.width * _baseHeight > r.height * _baseWidth) {
        int oldHeight = r.height;
        r.height = r.width * _baseHeight / _baseWidth;
        r.y = std::max(0, r.y - (r.height - oldHeight) / 2);
        if (r.y + r.height > img->height) {
            r.height = img->height - r.y;
        }
    } else {
        int oldWidth = r.width;
        r.width = r.height * _baseWidth / _baseHeight;
        r.x = std::max(0, r.x - (r.width - oldWidth) / 2);
        if (r.x + r.width > img->width) {
            r.width = img->width - r.x;
        }
    }

    // extract image patch
    IplImage *croppedImage = cvCreateImage(cvSize(r.width, r.height), IPL_DEPTH_8U, img->nChannels);
    cvSetImageROI(img, r);
    cvCopyImage(img, croppedImage);
    cvResetImageROI(img);

    // convert to greyscale
    if (img->nChannels != 1) {
        IplImage *tmp = greyImage(croppedImage);
        std::swap(tmp, croppedImage);
        cvReleaseImage(&tmp);
    }

    // apply soft mask
    p = (const float *)CV_MAT_ELEM_PTR(*softMask, r.y, r.x);
    for (int y = 0; y < r.height; y++) {
        unsigned char * const q =
            (unsigned char *)&croppedImage->imageData[y * croppedImage->widthStep];
        for (int x = 0; x < r.width; x++) {
            q[x] = (unsigned char)(p[x] * q[x]);
        }
        p += mask->cols;
    }

#if 0
    // show patch being classified
    IplImage *canvas = cvCreateImage(cvSize(320, 240), IPL_DEPTH_8U, croppedImage->nChannels);
    cvResize(croppedImage, canvas);
    cvNamedWindow("softMaskedWindow", 1);
    cvShowImage("softMaskedWindow", canvas);
    cvWaitKey(-1);
    cvReleaseImage(&canvas);
#endif

    // classify patch
    double score = evaluateWindow(croppedImage);

    // free memory
    cvReleaseImage(&croppedImage);
    cvReleaseMat(&softMask);

    SVL_FCN_TOC;
    return score;
}

double ScenePatchObjectDetector::evaluateRegion(IplImage *img, const list<CvPoint>& pixels) const
{
    SVL_ASSERT(img != NULL);
    CvMat *mask = cvCreateMat(img->height, img->width, CV_32SC1);
    cvZero(mask);

    int *p = (int *)CV_MAT_ELEM_PTR(*mask, 0, 0);
    for (list<CvPoint>::const_iterator it = pixels.begin(); it != pixels.end(); it++) {
        p[it->y * img->width + it->x] = 1;
    }

    double score = evaluateRegion(img, mask);

    cvReleaseMat(&mask);
    return score;
}

// features
vector<double> ScenePatchObjectDetector::computeFeatureVector(vector<IplImage *>& channels) const
{
    SVL_FCN_TIC;
    vector<double> f(_patchDictionary.size());

#if 0
    for (int i = 0; i < (int)_patchDictionary.size(); i++) {
        f[i] = _patchDictionary[i].first.value(channels[_patchDictionary[i].second]);
    }
#else
    vector<IplImage *> channelSums;
    vector<IplImage *> channelSumSq;

    for (int c = 0; c < (int)channels.size(); c++) {
	channelSums.push_back(cvCreateImage(cvSize(channels[c]->width + 1, channels[c]->height + 1),
                IPL_DEPTH_64F, channels[c]->nChannels));
        channelSumSq.push_back(cvCreateImage(cvSize(channels[c]->width + 1, channels[c]->height + 1),
                IPL_DEPTH_64F, channels[c]->nChannels));
	cvIntegral(channels[c], channelSums[c], channelSumSq[c]);
    }

#if 0
    // unthreaded
    for (int i = 0; i < (int)_patchDictionary.size(); i++) {
        int ch = _patchDictionary[i].second;
        f[i] = _patchDictionary[i].first.value(channels[ch], channelSums[ch], channelSumSq[ch]);
    }
#else
    // threaded
    svlThreadPool threadPool;
    vector<ScenePatchValueArgs> threadArgs;
    threadArgs.resize(_patchDictionary.size());
    threadPool.start();
    for (int i = 0; i < (int)_patchDictionary.size(); i++) {
        int ch = _patchDictionary[i].second;
        threadArgs[i].patch = &_patchDictionary[i].first;
        threadArgs[i].img = channels[ch];
        threadArgs[i].imgSum = channelSums[ch];
        threadArgs[i].imgSum2 = channelSumSq[ch];
        threadArgs[i].featureValue = &f[i];
        threadPool.addJob(&ScenePatchValueThread, &threadArgs[i]);
    }
    threadPool.finish();
#endif

    svlReleaseImages(channelSums);
    svlReleaseImages(channelSumSq);
#endif

    SVL_FCN_TOC;
    return f;
}

vector<IplImage *> ScenePatchObjectDetector::createImageChannels(const IplImage *img) const
{
    vector<IplImage *> channels;

    svlSoftEdgeMap edgeMapCalculator;

    if (img->nChannels == 1) {
        channels.push_back(cvCloneImage(img));
    } else {
        channels.push_back(greyImage(img));
    }
    channels.push_back(cvCloneImage(edgeMapCalculator.processImage(channels.back())));

    convertInPlace(&channels[0], IPL_DEPTH_32F);
    convertInPlace(&channels[1], IPL_DEPTH_32F);

    return channels;
}

CvRect ScenePatchObjectDetector::randomRectangle() const
{
    static const int minWidth = 4;
    static const int minHeight = 4;
    const int maxWidth = _baseWidth / 2;
    const int maxHeight = _baseHeight / 2;

    CvRect r;

    r.width = rand() % (maxWidth - minWidth) + minWidth;
    r.height = rand() % (maxHeight - minHeight) + minHeight;
    r.x = rand() % (_baseWidth - r.width);
    r.y = rand() % (_baseHeight - r.height);

    return r;
}
