/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    scenePatchObjectDetector.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <map>
#include <set>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

#include "sceneConfiguration.h"

using namespace std;

// ScenePatch ----------------------------------------------------------------

class ScenePatch {
 protected:
    CvMat *_patch;
    CvRect _location;

 public:
    ScenePatch();
    ScenePatch(IplImage *img, const CvRect &region);
    ScenePatch(const ScenePatch& p);
    virtual ~ScenePatch();

    // feature value calculations
    double value(const IplImage *img, CvPoint offset = cvPoint(0, 0)) const;
    double value(const IplImage *img, const IplImage *imgSum,
        const IplImage *imgSum2, CvPoint offset = cvPoint(0, 0)) const;
    double value(const CvMat *resp, CvPoint offset = cvPoint(0, 0)) const;

    CvMat *response(const IplImage *img) const;
    CvMat *response(const IplImage *img, const IplImage *imgSum,
        const IplImage *imgSum2) const;

    // i/o
    void read(istream& is);
    void write(ostream& os) const;
    IplImage *visualize() const;

    ScenePatch& operator=(const ScenePatch& p);

    // testing
    static void test(IplImage *img);

 protected:
    void clear();
};

// ScenePatchObjectDetector --------------------------------------------------

class ScenePatchObjectDetector {
 protected:
    int _baseWidth;
    int _baseHeight;
    vector<pair<ScenePatch, int> > _patchDictionary;
    svlBinaryClassifier *_classifier;
    string _objectName;

 public:
    ScenePatchObjectDetector(int w, int h);
    ScenePatchObjectDetector(const char *dictFilename,
        const char *classifierFilename = NULL);
    ScenePatchObjectDetector(const ScenePatchObjectDetector& d);
    virtual ~ScenePatchObjectDetector();

    // i/o
    void clear();
    void readDictionary(const char *filename);
    void readClassifier(const char *filename);
    void writeDictionary(const char *filename) const;
    void writeClassifier(const char *filename) const;

    void trimDictionary();

    int baseWidth() const { return _baseWidth; }
    int baseHeight() const { return _baseHeight; }

    IplImage *visualizeDictionary() const;

    void setObjectName(const char *name) { _objectName = string(name); }
    const char *getObjectName() const { return _objectName.c_str(); }

    // learning
    void learnPatchDictionary(const char *imgDir, int maxSize = 1000);
    void learnClassifier(const char *posImgDir,
        const char *negImgDir, int maxSamples = SVL_INT_MAX);
    void validateClassifier(const char *posImgDir,
        const char *negImgDir, int maxSamples = SVL_INT_MAX);

    // inference
    double evaluateWindow(const IplImage *img) const;
    double evaluateWindow(IplImage *img, const CvRect& region) const;
    vector<double> evaluateWindows(IplImage *img, const vector<CvRect>& regions) const;
    vector<double> evaluateWindows(IplImage *img, const svlObject2dFrame& regions) const;

    svlObject2dFrame evaluateImage(IplImage *img) const;
    svlObject2dFrame evaluateImage(IplImage *img, double scale) const;

    double evaluateRegion(IplImage *img, const CvMat *mask) const;
    double evaluateRegion(IplImage *img, const list<CvPoint>& pixels) const;

 protected:
    vector<double> computeFeatureVector(vector<IplImage *>& channels) const;
    vector<IplImage *> createImageChannels(const IplImage *img) const;
    CvRect randomRectangle() const;
};
