/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneReconstruct3D.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for 3d scene reconstruction.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <limits>

#include "Eigen/Core"
#include "Eigen/SVD"

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneVisualization.h"
#include "sceneSegmentDict.h"
#include "sceneRegionFeatures.h"
#include "sceneEntities.h"

#define WINDOW_NAME "sceneReconstruct3D"

using namespace std;
USING_PART_OF_NAMESPACE_EIGEN

// prototypes ---------------------------------------------------------------

void computeSceneDepths(SceneInstance &instance, bool bVisualize);

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneReconstruct3D [OPTIONS] (<baseName> | <evalList>)\n";
    cerr << "OPTIONS:\n"
         << "  -groundtruth      :: groundtruth visualization\n"
         << "  -outPoints <ext>  :: extension for reconstructed point cloud\n"
         << "  -outMesh <ext>    :: extension for reconstructed mesh\n"
         << "  -mergeClasses     :: merge classes into connected components\n"
         << "  -focalLength <n>  :: camera focal length (default: 300)\n"
         << "  -x                :: visualize reconstruction\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

//helper function (used to be in sceneRegionFeatures.h)
double lineFit(const vector<CvPoint>& points, Vector2d& normal)
{
    if (points.empty()) {
        SVL_LOG(SVL_LOG_ERROR, "no points for SceneRegionFeatures::lineFit()");
        normal = Vector2d::Zero();
        return 0.0;
    }

    Matrix2d C = Matrix2d::Zero();
    Vector2d mu = Vector2d::Zero();

    for (unsigned i = 0; i < points.size(); i++) {
        mu.x() += (double)points[i].x;
        mu.y() += (double)points[i].y;
        C(0, 0) += (double)(points[i].x * points[i].x);
        C(0, 1) += (double)(points[i].x * points[i].y);
        C(1, 1) += (double)(points[i].y * points[i].y);
    }
    C(1, 0) = C(0, 1);
    C -= mu * mu.transpose() / (double)points.size();

    Eigen::SVD<Matrix2d> S(C);

    double residual = S.singularValues()(1) / (double)points.size();

    normal.x() = S.matrixU()(1, 0);
    normal.y() = S.matrixU()(1, 1);

    return residual;
}

//helper function (used to be in sceneRegionFeatures.h)
Vector2d baseline(const list<CvPoint>& pixels, int height)
{
    if (pixels.size() < 2) {
        return Vector2d::Zero();
    }

    // compute centroid
    double meanY = 0.0;
    for (list<CvPoint>::const_iterator p = pixels.begin(); p != pixels.end(); p++) {
        meanY += p->y;
    }
    meanY /= (double)pixels.size();

    // find base points
    map<int, int> basePoints;
    int bottomRowCount = 0;

    for (list<CvPoint>::const_iterator p = pixels.begin(); p != pixels.end(); p++) {
        if (p->y < meanY)
            continue;

        if (basePoints.find(p->x) == basePoints.end()) {
            if (basePoints[p->x] < p->y) {
                basePoints[p->x] = p->y;
            }
        } else {
            basePoints[p->x] = p->y;
        }

        if (basePoints[p->x] == (height - 1)) {
            bottomRowCount += 1;
        }
    }

    // remove points touching bottom of image
    if (bottomRowCount < 0.5 * basePoints.size()) {
        map<int, int> newBasePoints;
        for (map<int, int>::const_iterator p = basePoints.begin(); p != basePoints.end(); p++) {
            if (p->second != (height - 1)) {
                newBasePoints[p->first] = p->second;
            }
        }    
        basePoints = newBasePoints;
    }

    // keep bottom-most 50%
    double yMedian = 0.0;
    if (basePoints.size() > 3) {
        vector<double> yValues;
        yValues.reserve(basePoints.size());
        for (map<int, int>::const_iterator p = basePoints.begin(); p != basePoints.end(); p++) {
            yValues.push_back((double)p->second);
        }
        
        yMedian = median<double>(yValues);
    }

    vector<CvPoint> points;
    for (map<int, int>::const_iterator p = basePoints.begin(); p != basePoints.end(); p++) {
        if (p->second >= yMedian) {
            points.push_back(cvPoint(p->first, p->second));
        }
    }
         
    Vector2d n;
    lineFit(points, n);
    
    Vector2d a(-n.x() / n.y(), 1.0 / n.y());
    for (map<int, int>::const_iterator p = basePoints.begin(); p != basePoints.end(); p++) {
        double delta = a.x() * p->first + a.y() - p->second;
        if (delta < 0.0) {
            a.y() -= delta;
        }
    }
    
    return a;
}



int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // constants
    const double CAMERA_HEIGHT = 1.8;

    // read commandline parameters
    const char *outPointCloudExt = NULL;
    const char *outMeshExt = NULL;
    double cameraFocalLength = 300;
    bool bGroundtruth = false;
    bool bMergeClasses = false;
    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-outPoints", outPointCloudExt)
        SVL_CMDLINE_STR_OPTION("-outMesh", outMeshExt)
        SVL_CMDLINE_BOOL_OPTION("-groundtruth", bGroundtruth)
        SVL_CMDLINE_BOOL_OPTION("-mergeClasses", bMergeClasses)
        SVL_CMDLINE_INT_OPTION("-focalLength", cameraFocalLength)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *evalList = SVL_CMDLINE_ARGV[0];

    // read list of evaluation images
    vector<string> baseNames;
    if (svlFileExists(evalList)) {
        SVL_LOG(SVL_LOG_MESSAGE, "Reading evaluation list from " << evalList << "...");
        baseNames = svlReadFile(evalList);
        SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");
    } else {
        baseNames.push_back(string(evalList));
    }

    // process images
    SVL_LOG(SVL_LOG_MESSAGE, "Processing images...");
    int hProcessImage = svlCodeProfiler::getHandle("processImage");
    for (int i = 0; i < (int)baseNames.size(); i++) {
        svlCodeProfiler::tic(hProcessImage);
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");

        // load image
        IplImage *img = cvLoadImage(imgFilename.c_str(), CV_LOAD_IMAGE_COLOR);
        SVL_ASSERT_MSG(img != NULL, "path: " << imgFilename);
        SVL_LOG(SVL_LOG_MESSAGE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*img));

        // initialize scene
        SceneInstance instance(img, baseNames[i].c_str());
        if (bGroundtruth) {
            instance.loadGroundTruth();
        } else {
            instance.loadAssignments(gSceneConfig.getOption("outputDir").c_str());
        }

        // merge adjacent background regions
        vector<bool> mergeIndex(gSceneConfig.numRegionLabels, false);
        if (bMergeClasses) {
            fill(mergeIndex.begin(), mergeIndex.end(), true);
        } else {
            for (int Si = 0; Si < gSceneConfig.numRegionLabels; Si++) {
                if ((gSceneRegionDefs.surface(Si) == ST_SKY) ||
                    (gSceneRegionDefs.surface(Si) == ST_GROUNDPLANE)) {
                    mergeIndex[Si] = true;
                }
            }
        }
        instance.mergeAdjacentBackground(mergeIndex);

        // determine camera rotation matrix
        double theta = atan2(instance.assignment.horizon - 0.5 * instance.height(),
            cameraFocalLength);
        SVL_LOG(SVL_LOG_VERBOSE, "...estimated camera angle: " << 180 / M_PI * theta);

        Matrix3d R;
        R << 1.0, 0.0, 0.0, 0.0, cos(theta), -sin(theta), 0.0, sin(theta), cos(theta);

        Matrix3d K;
        K << cameraFocalLength, 0.0, 0.5 * instance.width(),
            0.0, cameraFocalLength, 0.5 * instance.height(),
            0.0, 0.0, 1.0;

        K = R * K.inverse();

        // determine pixel 3d locations and region plane parameters
        MatrixXd X(instance.height(), instance.width());
        MatrixXd Y(instance.height(), instance.width());
        MatrixXd Z(instance.height(), instance.width());
        X.setZero(); Y.setZero(); Z.setConstant(-1.0);


        // first, place groundplane regions
	vector<SceneRegion *> &regions = instance.assignment.regions;
        for (unsigned int k = 0; k < regions.size(); k++) {
            if (gSceneRegionDefs.surface(regions[k]->Sk) != ST_GROUNDPLANE)
                continue;

            for (list<CvPoint>::const_iterator p = regions[k]->pixels.begin();
                 p != regions[k]->pixels.end(); p++) {
                if (p->y < instance.assignment.horizon) {
                    SVL_LOG(SVL_LOG_WARNING, "ground plane point " << toString(*p) <<
                        " is above the horizon");
                    continue;
                }
                Vector3d r_p = K * Vector3d((double)p->x, (double)p->y, 1.0);
                Vector3d x_p = (CAMERA_HEIGHT / (r_p.y() + SVL_DBL_MIN)) * r_p;

                X(p->y, p->x) = x_p.x();
                Y(p->y, p->x) = x_p.y();
                Z(p->y, p->x) = x_p.z();
            }
        }

        // next, place vertical regions
        // TODO: fix me
        for (unsigned int k = 0; k < regions.size(); k++) {
            if (gSceneRegionDefs.surface(regions[k]->Sk) != ST_VERTICAL)
                continue;

#if 0
            // TODO: fix me
            int yBase = instance.assignment.horizon;
            for (list<CvPoint>::const_iterator p = regions[k]->pixels.begin();
                 p != regions[k]->pixels.end(); p++) {
                if (p->y > yBase) yBase = p->y;
            }

            for (list<CvPoint>::const_iterator p = regions[k]->pixels.begin();
                 p != regions[k]->pixels.end(); p++) {

                Vector3d r_p = K * Vector3d((double)p->x, (double)p->y, 1.0);
                Vector3d r_q = K * Vector3d((double)p->x, (double)yBase, 1.0);
                Vector3d x_p = (CAMERA_HEIGHT * r_q.z() / (r_q.y() * r_p.z())) * r_p;

                X(p->y, p->x) = x_p.x();
                Y(p->y, p->x) = x_p.y();
                Z(p->y, p->x) = x_p.z();
            }
#else
            VectorXd a = baseline(regions[k]->pixels, instance.height());

            int yDelta = 0;
            for (list<CvPoint>::const_iterator p = regions[k]->pixels.begin();
                 p != regions[k]->pixels.end(); p++) {
                double yBase = a.x() * p->x + a.y() + yDelta;
                if (p->y > yBase) yDelta += (int)(p->y - yBase);
            }

            for (list<CvPoint>::const_iterator p = regions[k]->pixels.begin();
                 p != regions[k]->pixels.end(); p++) {

                double yBase = a.x() * p->x + a.y();
                yBase = std::max(yBase, (double)p->y);
                Vector3d r_p = K * Vector3d((double)p->x, (double)p->y, 1.0);
                Vector3d r_q = K * Vector3d((double)p->x, (double)yBase, 1.0);
                Vector3d x_p = (CAMERA_HEIGHT * r_q.z() / (r_q.y() * r_p.z())) * r_p;

                X(p->y, p->x) = x_p.x();
                Y(p->y, p->x) = x_p.y();
                Z(p->y, p->x) = x_p.z();
            }
#endif
        }

        // last, place sky regions
        double maxDepth = Z.maxCoeff();

        for (unsigned int k = 0; k < regions.size(); k++) {
            if (gSceneRegionDefs.surface(regions[k]->Sk) != ST_SKY)
                continue;

            for (list<CvPoint>::const_iterator p = regions[k]->pixels.begin();
                 p != regions[k]->pixels.end(); p++) {
                if (p->y > instance.assignment.horizon) {
                    SVL_LOG(SVL_LOG_WARNING, "sky point " << toString(*p) << " is below the horizon");
                    continue;
                }
                Vector3d r_p = K * Vector3d((double)p->x, (double)p->y, 1.0);
                Vector3d x_p = (1.1 * maxDepth / r_p.z()) * r_p;

                X(p->y, p->x) = x_p.x();
                Y(p->y, p->x) = x_p.y();
                Z(p->y, p->x) = x_p.z();
            }

        }

#if 0
        if (bVisualize) {
            IplImage *canvas = visualizeDepths(Z);
            visualizeClasses(&instance, canvas);
            visualizeHorizon(&instance, canvas);
            cvNamedWindow(WINDOW_NAME, 1);
            cvShowImage(WINDOW_NAME, canvas);
            cvWaitKey(10);
            cvReleaseImage(&canvas);
        }
#endif

        // sort point cloud by depth
        multimap<double, pair<int, int> > depthBuffer;
        for (int y = 0; y < instance.height(); y++) {
            for (int x = 0; x < instance.width(); x++) {
                depthBuffer.insert(make_pair(Z(y, x), make_pair(x, y)));
            }
        }

        // output point cloud
        if (outPointCloudExt != NULL) {
            SVL_LOG(SVL_LOG_MESSAGE, "writing point cloud...");
            string filename = gSceneConfig.filebase("outputDir", baseNames[i]) +
                string(outPointCloudExt);
            ofstream ofs(filename.c_str());
            SVL_ASSERT(!ofs.fail());

            for (multimap<double, pair<int, int> >::const_iterator it = depthBuffer.begin();
                 it != depthBuffer.end(); it++) {
                int x = it->second.first;
                int y = it->second.second;
                ofs << X(y, x) << " " << Y(y, x) << " " << Z(y, x) << " 0.0 0.0 0.0"
                    << " " << (double)CV_IMAGE_ELEM(img, unsigned char, y, 3 * x + 2) / 255.0
                    << " " << (double)CV_IMAGE_ELEM(img, unsigned char, y, 3 * x + 1) / 255.0
                    << " " << (double)CV_IMAGE_ELEM(img, unsigned char, y, 3 * x + 0) / 255.0
                    << " 1.0\n";
            }
            ofs.close();
            SVL_LOG(SVL_LOG_MESSAGE, "...done");
        }

        // output mesh
        if (outMeshExt != NULL) {
            SVL_LOG(SVL_LOG_MESSAGE, "writing point cloud...");
            string filename = gSceneConfig.filebase("outputDir", baseNames[i]) + string(outMeshExt);
            ofstream ofs(filename.c_str());
            SVL_ASSERT(!ofs.fail());

            // build reverse index for points
            MatrixXi reverseIndex(instance.height(), instance.width());
            int indx = 0;
            for (multimap<double, pair<int, int> >::const_iterator it = depthBuffer.begin();
                 it != depthBuffer.end(); it++) {
                reverseIndex(it->second.second, it->second.first) = indx++;
            }

            // output mesh
            int *p = (int *)CV_MAT_ELEM_PTR(*instance.assignment.pixels, 0, 0);
            int *q = p + instance.width();
            for (int y = 0; y < instance.height(); y++) {
                for (int x = 0; x < instance.width(); x++, p++, q++) {
                    if ((y < instance.height() - 1) && (x < instance.width() - 1)) {
                        if ((Z(y, x) > SVL_DBL_MIN) && (Z(y + 1, x) > SVL_DBL_MIN) && (Z(y, x + 1) > SVL_DBL_MIN)) {
                            if ((*(p + 1) == *p) && (*q == *p)) {
                                ofs << reverseIndex(y, x) << " " << reverseIndex(y + 1, x)
                                    << " " << reverseIndex(y, x + 1) << "\n";
                            }
                        }
                    }

                    if ((y < instance.height() - 1) && (x > 0)) {
                        if ((Z(y, x) > SVL_DBL_MIN) && (Z(y + 1, x) > SVL_DBL_MIN) && (Z(y + 1, x - 1) > SVL_DBL_MIN)) {
                            if ((*(q - 1) == *q) && (*p == *q)) {
                                ofs << reverseIndex(y, x) << " " << reverseIndex(y + 1, x)
                                    << " " << reverseIndex(y + 1, x - 1) << "\n";
                            }
                        }
                    }
                }
            }
	    
            ofs.close();
            SVL_LOG(SVL_LOG_MESSAGE, "...done");
        }

        // min/max depths
        computeSceneDepths(instance, bVisualize);

        // free memory
        cvReleaseImage(&img);
        svlCodeProfiler::toc(hProcessImage);
    }

    // shutdown
    if ((baseNames.size() == 1) && bVisualize) {
        cvWaitKey(-1);
    }
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

// helper functions ---------------------------------------------------------

void computeSceneDepths(SceneInstance &instance, bool bVisualize)
{
    SVL_FCN_TIC;

    // compute maximum and minimum depths
    // maximum:
    //   sky: horizon
    //   ground: ground point or horizon
    //   vertical: horizon or bottom-most region pixel in column
    // minimum:
    //   sky: horizon
    //   ground: ground point or horizon
    //   vertical: zero or depth of pixel below bottom-most pixel in region
    //             (could be wrong if region changes shape behind occlusion)
    vector<SceneRegion *> &regions = instance.assignment.regions;


    instance.assignment.maxDepth.setConstant(instance.assignment.horizon);
    instance.assignment.minDepth.setConstant(instance.height());
    vector<double> minRegionDepth(regions.size(), (double)instance.assignment.horizon);
    MatrixXd bottomMostRegion =
        MatrixXd::Constant(regions.size(), instance.width(), instance.assignment.horizon);
    int *p = (int *)CV_MAT_ELEM_PTR(*instance.assignment.pixels,
        instance.height() - 1, instance.width() - 1);
    for (int y = instance.height() - 1; y >= 0; y--) {
        for (int x = instance.width() - 1; x >= 0; x--, p--) {
            // update bottom-most pixel for region k
            bottomMostRegion(*p, x) = std::max(bottomMostRegion(*p, x), (double)y);

            // update min/max depths for pixel
            switch (gSceneRegionDefs.surface(regions[*p]->Sk)) {
            case ST_SKY:
                instance.assignment.minDepth(y, x) = instance.assignment.horizon;
                break;
            case ST_GROUNDPLANE:
            case ST_UNKNOWN:
                instance.assignment.minDepth(y, x) = (double)std::max(y, instance.assignment.horizon);
                instance.assignment.maxDepth(y, x) = (double)std::max(y, instance.assignment.horizon);
                break;
            case ST_VERTICAL:
                if (y == instance.height() - 1) {
                    minRegionDepth[*p] = instance.height();
                } else {
                    minRegionDepth[*p] = std::max(minRegionDepth[*p],
                        instance.assignment.minDepth(y + 1, x));
                }
                instance.assignment.maxDepth(y, x) = bottomMostRegion(*p, x);
                instance.assignment.minDepth(y, x) = minRegionDepth[*p];
                break;
            default:
                SVL_LOG(SVL_LOG_FATAL, "illegal surface type");
            }
        }
    }

    // second pass to make all vertical regions at minimum depth
    for (unsigned int k = 0; k < regions.size(); k++) {
        if (gSceneRegionDefs.surface(regions[k]->Sk) != ST_VERTICAL)
            continue;
        for (list<CvPoint>::const_iterator it = regions[k]->pixels.begin();
             it != regions[k]->pixels.end(); it++) {
            instance.assignment.minDepth(it->y, it->x) = minRegionDepth[k];
        }
    }

    // normalize and correct sense
    double scale = -1.0 / instance.height();
    instance.assignment.maxDepth = (scale * instance.assignment.maxDepth).cwise() + 1.0;
    instance.assignment.minDepth = (scale * instance.assignment.minDepth).cwise() + 1.0;

#if 0
    if (bVisualize) {
        vector<IplImage *> views;
        views.push_back(cvCloneImage(instance.image()));
        visualizeClasses(&instance, views.back());
        visualizeBoundaries(&instance, views.back());
        visualizeHorizon(&instance, views.back());
        views.push_back(visualizeDepths(instance.assignment.minDepth, instance.assignment.maxDepth.maxCoeff()));
        visualizeHorizon(&instance, views.back());
        views.push_back(visualizeDepths(instance.assignment.maxDepth, instance.assignment.maxDepth.maxCoeff()));
        visualizeHorizon(&instance, views.back());
        IplImage *canvas = combineImages(views, 1, 3);
        cvNamedWindow(WINDOW_NAME, 1);
        cvShowImage(WINDOW_NAME, canvas);
        cvWaitKey(-1);
        svlReleaseImages(views);
        cvReleaseImage(&canvas);
    }
#endif

    // third pass to ensure regions are planar
    for (unsigned int k = 0; k < regions.size(); k++) {        
        if (gSceneRegionDefs.surface(regions[k]->Sk) != ST_VERTICAL)
            continue;

        list<CvPoint>::const_iterator it = regions[k]->pixels.begin();
        Vector2d leftPoint((double)it->x, instance.assignment.maxDepth(it->y, it->x)); 
        Vector2d rightPoint = leftPoint;

        while (++it != regions[k]->pixels.end()) {
            if ((double)it->x < leftPoint.x()) {
                leftPoint.x() = (double)it->x;
                leftPoint.y() = instance.assignment.maxDepth(it->y, it->x);
            }
            if ((double)it->x > rightPoint.x()) {
                rightPoint.x() = (double)it->x;
                rightPoint.y() = instance.assignment.maxDepth(it->y, it->x);
            }
        }

        double dx = rightPoint.x() - leftPoint.x();
        if (dx == 0.0)
            continue;
        double dy = rightPoint.y() - leftPoint.y();
        for (it = regions[k]->pixels.begin();
             it != regions[k]->pixels.end(); it++) {
            double t = ((double)it->x - leftPoint.x()) / dx;
            double d = leftPoint.y() + t * dy;
            if (d < instance.assignment.maxDepth(it->y, it->x)) {
                instance.assignment.maxDepth(it->y, it->x) = d;
            }
        }
    }
  
    SVL_FCN_TOC;

    if (bVisualize) {
        vector<IplImage *> views;
        views.push_back(cvCloneImage(instance.image()));
        visualizeClasses(&instance, views.back());
        visualizeBoundaries(&instance, views.back());
        visualizeHorizon(&instance, views.back());
        views.push_back(visualizeDepths(instance.assignment.minDepth, instance.assignment.maxDepth.maxCoeff()));
        visualizeHorizon(&instance, views.back());
        views.push_back(visualizeDepths(instance.assignment.maxDepth, instance.assignment.maxDepth.maxCoeff()));
        visualizeHorizon(&instance, views.back());
        IplImage *canvas = combineImages(views, 1, 3);
        cvNamedWindow(WINDOW_NAME, 1);
        cvShowImage(WINDOW_NAME, canvas);
        cvWaitKey(50);
        svlReleaseImages(views);
        cvReleaseImage(&canvas);
    }
}
