/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneRegionFeatures.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Tianshi Gao   <tianshig@stanford.edu>
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "Eigen/SVD"

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneRegionFeatures.h"

using namespace std;


// SceneRegionFeatures -----------------------------------------------------

double SceneRegionFeatures::regionEntropy(const SceneInstance *instance, int k)
{
    vector<double> q(gSceneConfig.numRegionLabels, 0.0);

    int varIndx = 0;
    int *regId = (int *)CV_MAT_ELEM_PTR(*instance->assignment.pixels, 0, 0);
    for (int y = 0; y < instance->height(); y++) {
        for (int x = 0; x < instance->width(); x++, varIndx++, regId++) {
            if (*regId == k) {
                vector<double> m(instance->pixelMarginals[varIndx]);
                for (int lbl = 0; lbl < (int)m.size(); lbl++) {
                    q[lbl] += m[lbl];
                }
            }
        }
    }

    return entropy(q);
}

double SceneRegionFeatures::lineFit(const vector<CvPoint>& points, Vector2d *normal)
{
    if (points.empty()) {
        SVL_LOG(SVL_LOG_ERROR, "no points for SceneRegionFeatures::lineFit()");
        if (normal != NULL)
            *normal = Vector2d::Zero();
        return 0.0;
    }

    Matrix2d C = Matrix2d::Zero();
    Vector2d mu = Vector2d::Zero();

    for (unsigned i = 0; i < points.size(); i++) {
        mu.x() += (double)points[i].x;
        mu.y() += (double)points[i].y;
        C(0, 0) += (double)(points[i].x * points[i].x);
        C(0, 1) += (double)(points[i].x * points[i].y);
        C(1, 1) += (double)(points[i].y * points[i].y);
    }
    C(1, 0) = C(0, 1);
    C -= mu * mu.transpose() / (double)points.size();

    Eigen::SVD<Matrix2d> S(C);

    double residual = S.singularValues()(1) / (double)points.size();

    if (normal != NULL) {
        normal->x() = S.matrixU()(1, 0);
        normal->y() = S.matrixU()(1, 1);
    }

    return residual;
}

void SceneRegionFeatures::regionFeatures(SceneInstance *instance, int regId, vector<double>& f)
{
    SVL_FCN_TIC;
    SVL_ASSERT((instance != NULL) && (regId < instance->numRegions()));
    SVL_ASSERT(!instance->assignment.regions[regId]->empty());

    if (gSceneConfig.getOptionAsBool("debug:boostedPixelFeaturesOnly")) {
#if 0
        f.resize(instance->assignment.regions[regId]->pixelScores.size());
        Eigen::Map<VectorXd>(&f[0], f.size()) = instance->assignment.regions[regId]->pixelScores /
            (double)instance->assignment.regions[regId]->size();
#else
        f.resize(instance->assignment.regions[regId]->appearance().size());
        Eigen::Map<VectorXd>(&f[0], f.size()) = instance->assignment.regions[regId]->appearance();
#endif
        SVL_FCN_TOC;
        return;
    }

    f.clear();
    f.reserve(256);

    //
    // pre-process
    //

    const SceneRegion *region = instance->assignment.regions[regId];
    const int instanceWidth = instance->width();
    const int instanceHeight = instance->height();
    const double imageArea = (double)instance->size();
    const double sqrtImgArea = sqrt(imageArea);
    const double regionArea = (double)region->size();
    const double sqrtRegArea = sqrt(regionArea);

    // boundary contrast
    double boundaryContrast = 0.0;
    double internalContrast = 0.0;
    double boundaryCount = 0.0;

    // compute some shape statistics
    double muX2 = 0.0;
    double muY2 = 0.0;
    double muXY = 0.0;
    double muX = 0.0;
    double muY = 0.0;
    int minX = instanceWidth;
    int minY = instanceHeight;
    int maxX = 0;
    int maxY = 0;
    int perimeterLength = 0;

    // find lowest and highest pixel in region
    vector<int> minYx(instanceWidth, instanceHeight);
    vector<int> maxYx(instanceWidth, -1);

    // horizon counts
    double muAbove = 0.0;
    double muBelow = 0.0;
    double stdAbove = 0.0;
    double stdBelow = 0.0;
    int countAbove = 0;
    int countBelow = 0;

    for (list<CvPoint>::const_iterator ip = region->pixels.begin(); ip != region->pixels.end(); ip++) {
        int y = ip->y;
        int x = ip->x;
        SVL_ASSERT_MSG((x >= 0) && (y >= 0) && (x < instanceWidth) && (y < instanceHeight),
            x << " !< " << instanceWidth << ", " << y << " !< " << instanceHeight);
        const int *p = (int *)CV_MAT_ELEM_PTR(*instance->assignment.pixels, y, x);
        if (*p != regId) {
            instance->printScene();
            SVL_LOG(SVL_LOG_FATAL, "pixel (" << x << ", " << y << ") " << regId << " != " << *p);
        }
        double deltaHorizon = y - instance->assignment.horizon;

        // boundary
        if (y > 0) {
            if (*(p - instanceWidth) != regId) {
                boundaryContrast += instance->verticalContrast(x, y);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->verticalContrast(x, y);
            }
        }

        if (x > 0) {
            if (*(p - 1) != regId) {
                boundaryContrast += instance->horizontalContrast(x, y);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->horizontalContrast(x, y);
            }
        }

        if (y < instanceHeight - 1) {
            if (*(p + instanceWidth) != regId) {
                boundaryContrast += instance->verticalContrast(x, y + 1);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->verticalContrast(x, y + 1);
            }
        }

        if (x < instanceWidth - 1) {
            if (*(p + 1) != regId) {
                boundaryContrast += instance->horizontalContrast(x + 1, y);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->horizontalContrast(x + 1, y);
            }
        }

        // shape
        muX += x;
        muY += y;
        muX2 += x * x;
        muY2 += y * y;
        muXY += x * y;
        if (x < minX) minX = x;
        if (y < minY) minY = y;
        if (x > maxX) maxX = x;
        if (y > maxY) maxY = y;

        // 4-neighbors
        if ((x == 0) || (x == instanceWidth - 1) ||
            (y == 0) || (y == instanceHeight - 1) ||
            (*(p - instanceWidth) != regId) ||
            (*(p + instanceWidth) != regId) ||
            (*(p - 1) != regId) ||
            (*(p + 1) != regId)) {
            perimeterLength += 1;
        }

        // upper/lower boundary
        if (y < minYx[x]) minYx[x] = y;
        if (y > maxYx[x]) maxYx[x] = y;

        // geometry
        if (y >= instance->assignment.horizon) {
            countAbove += 1;
            muAbove += deltaHorizon;
            stdAbove += deltaHorizon * deltaHorizon;
        }
        if (y < instance->assignment.horizon) {
            countBelow += 1;
            muBelow -= deltaHorizon;
            stdBelow += deltaHorizon * deltaHorizon;
        }
    }

    //
    // appearance
    //

    // mean and variance appearance features
    svlSufficientStats const& suffStats = region->Ak;
    for (int i = 0; i < suffStats.size(); i++) {
        double mu = suffStats.sum(i) / suffStats.count();
        f.push_back(mu);
        double sigma = suffStats.sum2(i, i) / suffStats.count() - mu * mu;
        f.push_back(sigma);
    }

    f.push_back(boundaryContrast / (boundaryCount + SVL_DBL_MIN));
    f.push_back(boundaryContrast / sqrtRegArea);
    f.push_back(internalContrast / regionArea);
    f.push_back(internalContrast / (boundaryContrast + internalContrast + SVL_DBL_MIN));

    // appearance moments
    if (gSceneConfig.getOptionAsBool("includeMomentFeatures")) {
        VectorXd ax = region->appearanceXMoment();
        VectorXd ay = region->appearanceYMoment();
        for (int i = 0; i < ax.size(); i++) {
            f.push_back(ax[i]);
            f.push_back(ay[i]);
        }
    }

    //
    // shape
    //

    // normalize
    muX /= regionArea;
    muY /= regionArea;
    muX2 /= regionArea; muX2 -= muX * muX;
    muY2 /= regionArea; muY2 -= muY * muY;
    muXY /= regionArea; muXY -= muX * muY;

    double extentX = (double)(maxX - minX + 1);
    double extentY = (double)(maxY - minY + 1);

    // shape features (size, extent, perimeter, moments)
    f.push_back(regionArea / imageArea);
    f.push_back((double)perimeterLength / sqrtImgArea);
    f.push_back((double)perimeterLength / sqrtRegArea);
    f.push_back(extentX / (double)instanceWidth);
    f.push_back(extentY / (double)instanceHeight);
    f.push_back(regionArea / (extentX * extentY));
    f.push_back(0.5 * (double)perimeterLength / (extentX + extentY));
    f.push_back(muX / (double)instanceWidth);
    f.push_back(muY / (double)instanceHeight);
    f.push_back(muX2 / regionArea);
    f.push_back(muY2 / regionArea);
    f.push_back(muXY / regionArea);
    f.push_back((minX == 0) || (maxX == instanceWidth - 1) ? 1.0 : 0.0);
    f.push_back(minY == 0 ? 1.0 : 0.0);
    f.push_back(maxY == instanceHeight - 1 ? 1.0 : 0.0);

#if 0
    // find bottom-most and top-most pixel in each column
    multimap<int, CvPoint> orderedBottomRow;
    multimap<int, CvPoint> orderedTopRow;

    for (int x = 0; x < instanceWidth; x++) {
        if (minYx[x] != instanceHeight) {
            orderedTopRow.insert(make_pair(minYx[x], cvPoint(x, minYx[x])));
            orderedBottomRow.insert(make_pair(-maxYx[x], cvPoint(x, maxYx[x])));
        }
    }

    // remove "highest" 50% of bottom "lowest" 50% of top points
    unsigned keepSize = (unsigned)(0.5 * orderedBottomRow.size()) + 2;
    vector<CvPoint> bottomRow;
    vector<CvPoint> topRow;
    bottomRow.reserve(keepSize);
    topRow.reserve(keepSize);
    for (multimap<int, CvPoint>::const_iterator p = orderedBottomRow.begin();
         p != orderedBottomRow.end(); ++p) {
        bottomRow.push_back(p->second);
        if (bottomRow.size() >= keepSize) break;
    }
    for (multimap<int, CvPoint>::const_iterator p = orderedTopRow.begin();
         p != orderedTopRow.end(); ++p) {
        topRow.push_back(p->second);
        if (topRow.size() >= keepSize) break;
    }

    // find line-of-best fit along top and bottom row
    Vector2d normal;
    double residual = lineFit(topRow, &normal);
    f.push_back(residual);
    f.push_back(atan2(normal.y(), normal.x()));

    residual = lineFit(bottomRow, &normal);
    f.push_back(residual);
    f.push_back(atan2(normal.y(), normal.x()));
#endif

    //
    // geometry
    //

    if (gSceneConfig.getOptionAsBool("useGeometry")) {
        f.push_back((double)instance->assignment.horizon / instanceHeight);

        f.push_back((double)countAbove / (countAbove + countBelow + SVL_EPSILON));
        f.push_back((double)countBelow / (countAbove + countBelow + SVL_EPSILON));
        muAbove /= (double)(countAbove + SVL_EPSILON);
        muBelow /= (double)(countBelow + SVL_EPSILON);
        f.push_back(muAbove / instanceHeight);
        f.push_back(muBelow / instanceHeight);
        stdAbove /= (double)(countAbove + SVL_EPSILON);
        stdAbove -= muAbove * muAbove;
        stdBelow /= (double)(countBelow + SVL_EPSILON);
        stdBelow -= muBelow * muBelow;
        f.push_back(stdAbove / instanceHeight);
        f.push_back(stdBelow / instanceHeight);
        f.push_back((double)(maxY - instance->assignment.horizon) / instanceHeight);
        f.push_back((double)(instance->assignment.horizon - minY) / instanceHeight);
    }

    SVL_FCN_TOC;
}

// Compute region-like features over an object.
void SceneRegionFeatures::objectFeatures(SceneInstance *instance, int objId, vector<double>& f)
{
    SVL_FCN_TIC;
    SVL_ASSERT((instance != NULL) && (objId < instance->numObjects()));
    SVL_ASSERT(!instance->assignment.objects[objId]->empty());

    f.clear();
    f.reserve(256);

    //
    // pre-process
    //

    const SceneObject *object = instance->assignment.objects[objId];
    const int instanceWidth = instance->width();
    const int instanceHeight = instance->height();
    const double imageArea = (double)instance->size();
    const double sqrtImgArea = sqrt(imageArea);
    const double objectArea = (double)object->size();
    const double sqrtObjArea = sqrt(objectArea);

    // boundary contrast
    double boundaryContrast = 0.0;
    double internalContrast = 0.0;
    double boundaryCount = 0.0;

    // compute some shape statistics
    double muX2 = 0.0;
    double muY2 = 0.0;
    double muXY = 0.0;
    double muX = 0.0;
    double muY = 0.0;
    int minX = instanceWidth;
    int minY = instanceHeight;
    int maxX = 0;
    int maxY = 0;
    int perimeterLength = 0;

    // find lowest and highest pixel in region
    vector<int> minYx(instanceWidth, instanceHeight);
    vector<int> maxYx(instanceWidth, -1);

    // horizon counts
    double muAbove = 0.0;
    double muBelow = 0.0;
    double stdAbove = 0.0;
    double stdBelow = 0.0;
    int countAbove = 0;
    int countBelow = 0;

    list<CvPoint> objectPixels = object->pixels();
    set<int> objectRegions;
    for (set<SceneRegion *>::const_iterator it = object->regions.begin(); it != object->regions.end(); it++) {
        objectRegions.insert((*it)->id());
    }

    CvMat *objectMask = cvCreateMat(instanceHeight, instanceWidth, CV_32SC1);
    cvZero(objectMask);
    for (set<SceneRegion *>::const_iterator it = object->regions.begin(); it != object->regions.end(); it++) {
        for (list<CvPoint>::const_iterator ip = (*it)->pixels.begin(); ip != (*it)->pixels.end(); ip++) {
            CV_MAT_ELEM(*objectMask, int, ip->y, ip->x) = 1;
        }
    }

    for (list<CvPoint>::const_iterator ip = objectPixels.begin(); ip != objectPixels.end(); ip++) {
        int y = ip->y;
        int x = ip->x;

        const int *p = (int *)CV_MAT_ELEM_PTR(*objectMask, y, x);
        double deltaHorizon = y - instance->assignment.horizon;

        // boundary
        if (y > 0) {
            if (*(p - instanceWidth) == 0) {
                boundaryContrast += instance->verticalContrast(x, y);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->verticalContrast(x, y);
            }
        }

        if (x > 0) {
            if (*(p - 1) == 0) {
                boundaryContrast += instance->horizontalContrast(x, y);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->horizontalContrast(x, y);
            }
        }

        if (y < instanceHeight - 1) {
            if (*(p + instanceWidth) == 0) {
                boundaryContrast += instance->verticalContrast(x, y + 1);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->verticalContrast(x, y + 1);
            }
        }

        if (x < instanceWidth - 1) {
            if (*(p + 1) == 0) {
                boundaryContrast += instance->horizontalContrast(x + 1, y);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->horizontalContrast(x + 1, y);
            }
        }

        // shape
        muX += x;
        muY += y;
        muX2 += x * x;
        muY2 += y * y;
        muXY += x * y;
        if (x < minX) minX = x;
        if (y < minY) minY = y;
        if (x > maxX) maxX = x;
        if (y > maxY) maxY = y;

        // 4-neighbors
        if ((x == 0) || (x == instanceWidth - 1) ||
            (y == 0) || (y == instanceHeight - 1) ||
            (*(p - instanceWidth) == 0) ||
            (*(p + instanceWidth) == 0) ||
            (*(p - 1) == 0) ||
            (*(p + 1) == 0)) {
            perimeterLength += 1;
        }

        // upper/lower boundary
        if (y < minYx[x]) minYx[x] = y;
        if (y > maxYx[x]) maxYx[x] = y;

        // geometry
        if (y >= instance->assignment.horizon) {
            countAbove += 1;
            muAbove += deltaHorizon;
            stdAbove += deltaHorizon * deltaHorizon;
        }
        if (y < instance->assignment.horizon) {
            countBelow += 1;
            muBelow -= deltaHorizon;
            stdBelow += deltaHorizon * deltaHorizon;
        }
    }

    //
    // appearance
    //

    // mean and variance appearance features
    set<SceneRegion *>::const_iterator it = object->regions.begin();
    svlSufficientStats suffStats = (*it)->Ak;
    while (++it != object->regions.end()) {
        suffStats.accumulate((*it)->Ak);
    }
    for (int i = 0; i < suffStats.size(); i++) {
        double mu = suffStats.sum(i) / suffStats.count();
        f.push_back(mu);
        double sigma = suffStats.sum2(i, i) / suffStats.count() - mu * mu;
        f.push_back(sigma);
    }

    f.push_back(boundaryContrast / (boundaryCount + SVL_DBL_MIN));
    f.push_back(boundaryContrast / sqrtObjArea);
    f.push_back(internalContrast / objectArea);
    f.push_back(internalContrast / (boundaryContrast + internalContrast + SVL_DBL_MIN));

    // appearance moments
    if (gSceneConfig.getOptionAsBool("includeMomentFeatures")) {
        VectorXd ax = object->appearanceXMoment();
        VectorXd ay = object->appearanceYMoment();
        for (int i = 0; i < ax.size(); i++) {
            f.push_back(ax[i]);
            f.push_back(ay[i]);
        }
    }

    //
    // shape
    //

    // normalize
    muX /= objectArea;
    muY /= objectArea;
    muX2 /= objectArea; muX2 -= muX * muX;
    muY2 /= objectArea; muY2 -= muY * muY;
    muXY /= objectArea; muXY -= muX * muY;

    double extentX = (double)(maxX - minX + 1);
    double extentY = (double)(maxY - minY + 1);

    // shape features (size, extent, perimeter, moments)
    f.push_back(objectArea / imageArea);
    f.push_back((double)perimeterLength / sqrtImgArea);
    f.push_back((double)perimeterLength / sqrtObjArea);
    f.push_back(extentX / (double)instanceWidth);
    f.push_back(extentY / (double)instanceHeight);
    f.push_back(objectArea / (extentX * extentY));
    f.push_back(0.5 * (double)perimeterLength / (extentX + extentY));
    f.push_back(muX / (double)instanceWidth);
    f.push_back(muY / (double)instanceHeight);
    f.push_back(muX2 / objectArea);
    f.push_back(muY2 / objectArea);
    f.push_back(muXY / objectArea);
    f.push_back((minX == 0) || (maxX == instanceWidth - 1) ? 1.0 : 0.0);
    f.push_back(minY == 0 ? 1.0 : 0.0);
    f.push_back(maxY == instanceHeight - 1 ? 1.0 : 0.0);

    //
    // geometry
    //

    if (gSceneConfig.getOptionAsBool("useGeometry")) {
        f.push_back((double)instance->assignment.horizon / instanceHeight);

        f.push_back((double)countAbove / (countAbove + countBelow + SVL_EPSILON));
        f.push_back((double)countBelow / (countAbove + countBelow + SVL_EPSILON));
        muAbove /= (double)(countAbove + SVL_EPSILON);
        muBelow /= (double)(countBelow + SVL_EPSILON);
        f.push_back(muAbove / instanceHeight);
        f.push_back(muBelow / instanceHeight);
        stdAbove /= (double)(countAbove + SVL_EPSILON);
        stdAbove -= muAbove * muAbove;
        stdBelow /= (double)(countBelow + SVL_EPSILON);
        stdBelow -= muBelow * muBelow;
        f.push_back(stdAbove / instanceHeight);
        f.push_back(stdBelow / instanceHeight);
        f.push_back((double)(maxY - instance->assignment.horizon) / instanceHeight);
        f.push_back((double)(instance->assignment.horizon - minY) / instanceHeight);
    }

    SVL_FCN_TOC;
}

void SceneRegionFeatures::pedestrianFeatures(SceneInstance *instance, int objId, vector<double>& f)
{
    SVL_FCN_TIC;
    SVL_ASSERT((instance != NULL) && (objId < instance->numObjects()));
    SVL_ASSERT(!instance->assignment.objects[objId]->empty());

    f.clear();
    f.reserve(25);

    //
    // Pre-process
    //

    const SceneObject *object = instance->assignment.objects[objId];
    const int instanceWidth = instance->width();
    const int instanceHeight = instance->height();
    const double imageArea = (double)instance->size();
    const double sqrtImgArea = sqrt(imageArea);
    const double objectArea = (double)object->size();
    const double sqrtObjArea = sqrt(objectArea);

    // boundary contrast
    double boundaryContrast = 0.0;
    double internalContrast = 0.0;
    double boundaryCount = 0.0;

    // compute some shape statistics
    double muX2 = 0.0;
    double muY2 = 0.0;
    double muXY = 0.0;
    double muX = 0.0;
    double muY = 0.0;
    int minX = instanceWidth;
    int minY = instanceHeight;
    int maxX = 0; 
    int maxY = 0; int xMaxY = 0;
    int perimeterLength = 0;

    // find lowest and highest pixel in region
    vector<int> minYx(instanceWidth, instanceHeight);
    vector<int> maxYx(instanceWidth, -1);

    // horizon counts
    double muAbove = 0.0;
    double muBelow = 0.0;
    double stdAbove = 0.0;
    double stdBelow = 0.0;
    int countAbove = 0;
    int countBelow = 0;

    list<CvPoint> objectPixels = object->pixels();
    set<int> objectRegions;
    for (set<SceneRegion *>::const_iterator it = object->regions.begin(); it != object->regions.end(); it++) {
        objectRegions.insert((*it)->id());
    }

    CvMat *objectMask = cvCreateMat(instanceHeight, instanceWidth, CV_32SC1);
    cvZero(objectMask);
    for (set<SceneRegion *>::const_iterator it = object->regions.begin(); it != object->regions.end(); it++) {
        for (list<CvPoint>::const_iterator ip = (*it)->pixels.begin(); ip != (*it)->pixels.end(); ip++) {
            CV_MAT_ELEM(*objectMask, int, ip->y, ip->x) = 1;
        }
    }

    for (list<CvPoint>::const_iterator ip = objectPixels.begin(); ip != objectPixels.end(); ip++) {
        int y = ip->y;
        int x = ip->x;

        const int *p = (int *)CV_MAT_ELEM_PTR(*objectMask, y, x);
        double deltaHorizon = y - instance->assignment.horizon;

        // boundary
        if (y > 0) {
            if (*(p - instanceWidth) == 0) {
                boundaryContrast += instance->verticalContrast(x, y);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->verticalContrast(x, y);
            }
        }

        if (x > 0) {
            if (*(p - 1) == 0) {
                boundaryContrast += instance->horizontalContrast(x, y);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->horizontalContrast(x, y);
            }
        }

        if (y < instanceHeight - 1) {
            if (*(p + instanceWidth) == 0) {
                boundaryContrast += instance->verticalContrast(x, y + 1);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->verticalContrast(x, y + 1);
            }
        }

        if (x < instanceWidth - 1) {
            if (*(p + 1) == 0) {
                boundaryContrast += instance->horizontalContrast(x + 1, y);
                boundaryCount += 1.0;
            } else {
                internalContrast += instance->horizontalContrast(x + 1, y);
            }
        }

        // shape
        muX += x;
        muY += y;
        muX2 += x * x;
        muY2 += y * y;
        muXY += x * y;
        if (x < minX) minX = x;
        if (y < minY) minY = y;
        if (x > maxX) maxX = x;
        if (y > maxY) {maxY = y; xMaxY = x;}

        // 4-neighbors
        if ((x == 0) || (x == instanceWidth - 1) ||
            (y == 0) || (y == instanceHeight - 1) ||
            (*(p - instanceWidth) == 0) ||
            (*(p + instanceWidth) == 0) ||
            (*(p - 1) == 0) ||
            (*(p + 1) == 0)) {
            perimeterLength += 1;
        }

        // upper/lower boundary
        if (y < minYx[x]) minYx[x] = y;
        if (y > maxYx[x]) maxYx[x] = y;

        // geometry
        if (y >= instance->assignment.horizon) {
            countAbove += 1;
            muAbove += deltaHorizon;
            stdAbove += deltaHorizon * deltaHorizon;
        }
        if (y < instance->assignment.horizon) {
            countBelow += 1;
            muBelow -= deltaHorizon;
            stdBelow += deltaHorizon * deltaHorizon;
        }
    }

    //
    // shape
    //

    // normalize
    muX /= objectArea;
    muY /= objectArea;
    muX2 /= objectArea; muX2 -= muX * muX;
    muY2 /= objectArea; muY2 -= muY * muY;
    muXY /= objectArea; muXY -= muX * muY;

    double extentX = (double)(maxX - minX + 1);
    double extentY = (double)(maxY - minY + 1);

    // shape features 
    f.push_back(extentX /  (extentY + 1.0));
    f.push_back(objectArea / imageArea);
    f.push_back((double)perimeterLength / sqrtImgArea);
    f.push_back((double)perimeterLength / sqrtObjArea);
    f.push_back(extentX / (double)instanceWidth);
    f.push_back(extentY / (double)instanceHeight);
    f.push_back(objectArea / (extentX * extentY));
    f.push_back(0.5 * (double)perimeterLength / (extentX + extentY));
    f.push_back(muX / (double)instanceWidth);
    f.push_back(muY / (double)instanceHeight);
    f.push_back(muX2 / objectArea);
    f.push_back(muY2 / objectArea);
    f.push_back(muXY / objectArea);
    f.push_back((minX == 0) || (maxX == instanceWidth - 1) ? 1.0 : 0.0);
    f.push_back(minY == 0 ? 1.0 : 0.0);
    f.push_back(maxY == instanceHeight - 1 ? 1.0 : 0.0);

    //
    // context
    //
    int lowestPoint = std::min(maxY+5, instanceHeight - 1);

    // how far away from the road
    double distToRoad = 0;

    bool bFoundRd = false;
    int y = maxY;
    for (int x = minX; x <= maxX; x++) {
        for (y = maxY+1; y < instanceHeight; y++) {
            int regionId =  instance->assignment.regionId(x, y);
            SceneRegion* region = instance->assignment.regions[regionId];
            if (gSceneRegionDefs.name(region->Sk) == "road") {
                bFoundRd = true;
                break;
            }
        }
    }
    distToRoad = (double)(y - maxY + 1) / extentY;
    f.push_back(distToRoad);
    f.push_back(bFoundRd ? 1.0 : 0.0);
    

    // percentage of road in the neighbor
    double roadBeneath = 0;
    int count = 0;
    for (int x = 0; x < (int)maxYx.size(); x++) {
        if (maxYx[x]  == -1 || maxYx[x] == instanceHeight - 1 || maxYx[x] < minY + extentY/4*3)
            continue;
        for (int y = maxYx[x]+1; y < lowestPoint; y++) {
            int regionId =  instance->assignment.regionId(x, y);
            SceneRegion* region = instance->assignment.regions[regionId];
            if (gSceneRegionDefs.name(region->Sk) == "road") {
                roadBeneath++;
            }
            count++;
        }
    }
    if (count != 0) {
        roadBeneath /= (double)count;
    }
    f.push_back(roadBeneath);
    
    // surrounding regions
    // # of surrounding regions
    double numSurroundingRegions = (double)object->boundaries.size();
    f.push_back(numSurroundingRegions);

    // surrounding region class distribution
    VectorXd regionClassDistribution = VectorXd::Zero(gSceneConfig.numRegionLabels);
    for (set<SceneObjRegBoundary *>::iterator itr = object->boundaries.begin();
         itr != object->boundaries.end(); ++itr) {
        regionClassDistribution[(*itr)->region->Sk] += 1;
    }
    
    regionClassDistribution /= regionClassDistribution.sum();
    for (int i = 0; i < (int)regionClassDistribution.rows(); i++) {
        f.push_back(regionClassDistribution[i]);
    }
    
    //
    // geometry
    //
    if (gSceneConfig.getOptionAsBool("useGeometry")) {
        f.push_back((double)countAbove / (countAbove + countBelow + SVL_EPSILON));
        f.push_back((double)instance->assignment.horizon / instanceHeight);
        muAbove /= (double)(countAbove + SVL_EPSILON);
        muBelow /= (double)(countBelow + SVL_EPSILON); 
        f.push_back(muAbove / instanceHeight);
        f.push_back(muBelow / instanceHeight);
        stdAbove /= (double)(countAbove + SVL_EPSILON);
        stdAbove -= muAbove * muAbove;
        stdBelow /= (double)(countBelow + SVL_EPSILON);
        stdBelow -= muBelow * muBelow;
        f.push_back(stdAbove / instanceHeight);
        f.push_back(stdBelow / instanceHeight);

        // marginal of the prob. of being a ped. (should be more robust to a single point)
        // joint: (prob{h | cam height, horizon}* prob{cam height} * prob{horizon})
        // TODO: if it works, then do not hard coded
        //double worldHeight = extentY ./ (double)(maxY - instance->assignment.horizon + SVL_EPSILON);
        map<double, double> camHeightPrior;
        loadCameraHeightPrior(camHeightPrior);
        // normalize
        double h = extentY / (double)instanceHeight;
        double v = 1.0 - (double)maxY / (double)instanceHeight;
        double v0 = 1.0 - (double)instance->assignment.horizon / (double)instanceHeight;
        double probHeight = getHeightMarginal(h, v, v0, camHeightPrior, instanceHeight);
        f.push_back(probHeight);

        double hw = h / (v0 - v + SVL_EPSILON);
        f.push_back(hw);
        f.push_back(hw * hw);

        f.push_back((double)(maxY - instance->assignment.horizon) / instanceHeight / h);
        f.push_back((double)(instance->assignment.horizon - minY) / instanceHeight / h);
    }
    
    //SVL_LOG(SVL_LOG_DEBUG, "pedestrian feature: " << toString(f) << "; class: " <<  gSceneObjectDefs.name(object->Ok - 1) << endl);
    SVL_FCN_TOC;
}

// Compute features for adjacent pair of regions. The feature computation
// is directed, so regionPairFeatures(k, l) != regionPairFeatures(l, k).
void SceneRegionFeatures::regionPairFeatures(SceneInstance *instance, int k, int l,
    vector<double>& f)
{
    SVL_FCN_TIC;
    SVL_ASSERT((instance != NULL) && (k < instance->numRegions()) && (l < instance->numRegions()));

    f.clear();
    f.reserve(8);

    // centroid difference
    double Yk = instance->assignment.regions[k]->centroid().y();
    double Yl = instance->assignment.regions[l]->centroid().y();
    f.push_back((Yk - Yl) / instance->height());

    // find boundary pixels
    int countBelow = 0;
    int countAbove = 0;
    vector<CvPoint> boundaryPixels;

    // TODO: use smaller of k and l
    for (list<CvPoint>::const_iterator it = instance->assignment.regions[k]->pixels.begin();
         it != instance->assignment.regions[k]->pixels.end(); it++) {
        const int *p = (int *)CV_MAT_ELEM_PTR(*instance->assignment.pixels, it->y, it->x);

        if ((it->y > 0) && (*(p - instance->width()) == l)) {
            countAbove += 1;
        }
        if ((it->y < instance->height() - 1) && (*(p + instance->width()) == l)) {
            countBelow += 1;
        }

        if (((it->y > 0) && (*(p - instance->width()) == l)) ||
            ((it->y < instance->height() - 1) && (*(p + instance->width()) == l)) ||
            ((it->x > 0) && (*(p - 1) == l)) ||
            ((it->x < instance->width() - 1) && (*(p + 1) == l))) {
            boundaryPixels.push_back(*it);
        }

    }

    // amount above/below
    f.push_back((double)countAbove / (double)(countAbove + countBelow + SVL_DBL_MIN));
    f.push_back((double)countBelow / (double)(countAbove + countBelow + SVL_DBL_MIN));

    // boundary line of fit
    Vector2d n;
    double r = lineFit(boundaryPixels, &n);
    f.push_back(r);
    f.push_back(n.y());

    SVL_FCN_TOC;
}


void SceneRegionFeatures::loadCameraHeightPrior(map<double, double>& camHeightPrior)
{
    camHeightPrior.clear();
    string cameraHeightFilename = gSceneConfig.getOption("baseDir") + gSceneConfig.getOption("modelsDir") + "camera.height.prior.txt";
    ifstream ifs(cameraHeightFilename.c_str());
    SVL_ASSERT_MSG(!ifs.fail(), "Could not read camera height prior from " << cameraHeightFilename);

    string line;
    vector<double> h;
    vector<double> f;
    int count = 0;
    while(getline(ifs, line)) {
        stringstream ss;
        double tmp;
        ss << line;
        while (1) {
            ss >> tmp;
            if (ss.fail()) {
                break;
            }
            if (count == 0) {
                h.push_back(tmp);
            } else {
                f.push_back(tmp);
            }
        }
        count++;
    }
    
    SVL_ASSERT_MSG(f.size() == h.size(), f.size() << " != " << h.size());
    for (int i = 0; i < (int)f.size(); i++) {
        camHeightPrior.insert(make_pair(h[i], f[i]));
    }

    ifs.close();
}

double SceneRegionFeatures::getHeightMarginal(double h, double v, double v0, map<double, double>& camHeightPrior, int imgHeight)
{
    if (v0 >= 1.0) {
        return 0.5;
    }
    
    double pedMu = 1.7, pedSigma = 0.203;

    double Vf[] = {0.0145, 0.0179, 0.0221, 0.0272, 0.0335, 0.0412, 
                  0.0508, 0.0625, 0.0770, 0.0949, 0.1168, 0.0949, 
                  0.0770, 0.0652, 0.0508, 0.0412, 0.0335, 0.0272, 
                  0.0221, 0.0179, 0.0145};

    double Vx[] = {-10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0,
                  1, 2, 3, 4, 5, 6, 7, 8, 9, 10};

    double prob = 0;
    for (int i = 0; i < 20; i++) {
        for (map<double, double>::const_iterator itr = camHeightPrior.begin();
             itr != camHeightPrior.end(); ++itr) {
            double worldHeight = h / (v0+Vx[i]/(double)imgHeight - v + SVL_EPSILON) * itr->first;
            prob += 1.0 / (sqrt(2*3.1415926) * pedSigma) * exp(-0.5 * (pedMu - worldHeight)*(pedMu-worldHeight) / (pedSigma*pedSigma)) * Vf[i] * itr->second;
        }
    }
    return prob;
}
