/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneRegionModel.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "Eigen/Core"

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneRegionFeatures.h"
#include "sceneVisualization.h"

using namespace std;

// SceneRegionModel ---------------------------------------------------------

SceneRegionModel::SceneRegionModel() :
    ScenePixelModel(),
    _regionClassModels(NULL),
    _regionSingletons(gSceneConfig.numRegionLabels + 1),
    _horizonSingleton(1),
    _regionPairwiseContrastWeight(1.0)
{
    _horizonRegionPairwise = MatrixXd::Ones(gSceneConfig.numRegionLabels, 2);
}

SceneRegionModel::~SceneRegionModel()
{
    if (_regionClassModels != NULL) {
        delete _regionClassModels;
    }
}

// i/o
void SceneRegionModel::loadModels(const char *modelDir)
{
    ScenePixelModel::loadModels(modelDir);

    loadRegionSingletonModel(modelDir);
    if (gSceneConfig.getOptionAsBool("useGeometry")) {
        loadHorizonModel(modelDir);
        loadHorizonPairwiseModel(modelDir);
    }
    loadRegionContrastWeight(modelDir);
}

void SceneRegionModel::loadHorizonModel(const char* modelDir)
{
    string fileName = string(modelDir) + string("/horizon.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading horizon singleton parameters from " << fileName);
    _horizonSingleton.load(fileName.c_str());
}

void SceneRegionModel::saveHorizonModel(const char* modelDir)
{
    string fileName = string(modelDir) + string("/horizon.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing horizon singleton parameters to " << fileName);
    _horizonSingleton.save(fileName.c_str());
}

void SceneRegionModel::loadHorizonPairwiseModel(const char* modelDir)
{
    SVL_LOG(SVL_LOG_VERBOSE, "...reading horizon-region pairwise parameters from " << modelDir);
    string fileName = string(modelDir) + string("/region.horizon.pairwise.txt");
    ifstream ifs(fileName.c_str());
    for (int i = 0; i < _horizonRegionPairwise.rows(); i++) {
        for (int j = 0; j < _horizonRegionPairwise.cols(); j++) {
            ifs >> _horizonRegionPairwise(i, j);
        }
    }
    SVL_ASSERT_MSG(!ifs.fail(), "could read from " << fileName);
    ifs.close();
}

void SceneRegionModel::saveHorizonPairwiseModel(const char* modelDir)
{
    SVL_LOG(SVL_LOG_VERBOSE, "...writing horizon-region pairwise parameters to " << modelDir);
    string fileName = string(modelDir) + string("/region.horizon.pairwise.txt");
    ofstream ofs(fileName.c_str());
    ofs << _horizonRegionPairwise << "\n";
    SVL_ASSERT_MSG(!ofs.fail(), "could write from " << fileName);
    ofs.close();
}

void SceneRegionModel::loadRegionSingletonModel(const char *modelDir)
{
    if (gSceneConfig.getOptionAsBool("useBoostedRegionFeatures")) {
        if (_regionClassModels != NULL)
            delete _regionClassModels;

        _regionClassModels = new svlBoostedClassifierSet(gSceneConfig.numRegionLabels);
        SVL_LOG(SVL_LOG_VERBOSE, "...reading region class models from " << modelDir);
        string filename = string(modelDir) + string("/region.class.xml");
        bool bSuccess = _regionClassModels->load(filename.c_str());
        SVL_ASSERT(bSuccess);
    }

    string fileName = string(modelDir) + string("/region.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading region feature whitening parameters from " << fileName);
    _featureWhitener.load(fileName.c_str());

    fileName = string(modelDir) + string("/region.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading region singleton parameters from " << fileName);
    _regionSingletons.load(fileName.c_str());
}

void SceneRegionModel::saveRegionSingletonModel(const char *modelDir)
{
    if (_regionClassModels != NULL) {
        SVL_LOG(SVL_LOG_VERBOSE, "...writing region class models to " << modelDir);
        string filename = string(modelDir) + string("/region.class.xml");
        bool bSuccess = _regionClassModels->save(filename.c_str());
        SVL_ASSERT(bSuccess);
    }

    string fileName = string(modelDir) + string("/region.featureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing region feature whitening parameters to " << fileName);
    _featureWhitener.save(fileName.c_str());

    fileName = string(modelDir) + string("/region.singleton.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing region singleton parameters to " << fileName);
    _regionSingletons.save(fileName.c_str());
}

void SceneRegionModel::loadRegionContrastWeight(const char *modelDir)
{
    string fileName = string(modelDir) + string("/region.pairwiseContrastWeight.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading pairwise contrast weight from " << fileName);
    ifstream ifs(fileName.c_str());
    SVL_ASSERT(!ifs.fail());
    ifs >> _regionPairwiseContrastWeight;
    SVL_ASSERT(!ifs.fail());
    ifs.close();
}

void SceneRegionModel::saveRegionContrastWeight(const char *modelDir)
{
    string fileName = string(modelDir) + string("/region.pairwiseContrastWeight.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing pairwise contrast weight to " << fileName);
    ofstream ofs(fileName.c_str());
    ofs << _regionPairwiseContrastWeight;
    SVL_ASSERT(!ofs.fail());
    ofs.close();
}

// learning
void SceneRegionModel::learnHorizonSingletonModel(vector<string>& baseNames)
{
    // load horizon groundtruth
    map<string, int> horizons = SceneDataLoader::loadHorizonTrainingData();

    // accumulate training data
    vector<vector<double> > x;
    x.reserve(baseNames.size());

    SVL_LOG(SVL_LOG_MESSAGE, "Processing images...");
    int hProcessImage = svlCodeProfiler::getHandle("processImage");
    for (int n = 0; n < (int)baseNames.size(); n++) {
        // locate horizon for this scene
        if (horizons.find(baseNames[n]) == horizons.end()) {
            SVL_LOG(SVL_LOG_WARNING, "could not find horizon for " << baseNames[n]);
            continue;
        }

        svlCodeProfiler::tic(hProcessImage);
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[n], "imgExt");

        // load scene
        SceneInstance scene(imgFilename.c_str(), baseNames[n].c_str());
        SVL_LOG(SVL_LOG_VERBOSE, baseNames[n] << " (" << (n + 1) << " of "
            << baseNames.size() << "): " << toString(*scene.image()));

        // accumulate horizon location
        x.push_back(vector<double>(1, (double)horizons[baseNames[n]] / scene.height()));
        svlCodeProfiler::toc(hProcessImage);
    }

    // train model
    _horizonSingleton.train(x);
}

void SceneRegionModel::learnHorizonPairwiseModel(vector<string>& baseNames, int subSample)
{
    // load horizon groundtruth
    map<string, int> horizons = SceneDataLoader::loadHorizonTrainingData();

    // accumulate training data
    _horizonRegionPairwise.setOnes();
    SVL_LOG(SVL_LOG_MESSAGE, "Processing images...");
    int hProcessImage = svlCodeProfiler::getHandle("processImage");
    for (int n = 0; n < (int)baseNames.size(); n++) {
        // locate horizon for this scene
        if (horizons.find(baseNames[n]) == horizons.end()) {
            SVL_LOG(SVL_LOG_WARNING, "could not find horizon for " << baseNames[n]);
            continue;
        }

        SVL_LOG(SVL_LOG_VERBOSE, "...using horizon location " << horizons[baseNames[n]]);

        svlCodeProfiler::tic(hProcessImage);
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[n], "imgExt");

        // load scene and groundtruth assignments
        SceneInstance scene(imgFilename.c_str(), baseNames[n].c_str());
        SVL_LOG(SVL_LOG_VERBOSE, baseNames[n] << " (" << (n + 1) << " of "
            << baseNames.size() << "): " << toString(*scene.image()));
        scene.loadGroundTruth();

        int h = horizons[baseNames[n]];
        for (int y = 0; y < scene.height(); y++) {
            for (int x = 0; x < scene.width(); x++) {
                int regId = CV_MAT_ELEM(*scene.assignment.pixels, int, y, x);
                if (scene.assignment.regions[regId]->Sk < 0) continue; // skip unlabeled
                SVL_ASSERT(scene.assignment.regions[regId]->Sk < gSceneConfig.numRegionLabels);

                _horizonRegionPairwise(scene.assignment.regions[regId]->Sk, y < h ? 0 : 1) += 1.0;
            }
        }

        svlCodeProfiler::toc(hProcessImage);
    }

    // train model
    if (svlLogger::getLogLevel() >= SVL_LOG_VERBOSE) {
        SVL_LOG(SVL_LOG_VERBOSE, "pixel counts:");
        cout << _horizonRegionPairwise << "\n";
    }

    for (int i = 0; i < gSceneConfig.numRegionLabels; i++) {
        _horizonRegionPairwise.row(i) /= _horizonRegionPairwise.row(i).sum();
    }
}

void SceneRegionModel::learnRegionSingletonModel(const vector<string>& baseNames, 
        int maxIterations, bool bResume)
{
    // baseDir/imgDir/baseName.jpg              [H x W x 3]
    // baseDir/lblDir/baseName.layers.txt       [H x W]
    // baseDir/lblDir/baseName.layerLabels.txt  [K x 1]

    // baseDir/closedLoopFeatureCache

    // TODO: all training data should come from a training cache file (make similar to pixel cache)

    // load horizon groundtruth
    bool bUseGeometry = gSceneConfig.getOptionAsBool("useGeometry");
    map<string, int> horizons;
    if (bUseGeometry) {
        horizons = SceneDataLoader::loadHorizonTrainingData();
    }

    // feature values and labels for training
    SceneFeatureCache trainingData;
    int nLabels = gSceneConfig.numRegionLabels + 1; // last class in the "invalid" class
    vector<int> labelCounts(nLabels, 0);
    vector<double> labelWeight(nLabels, 0.0);

    // accumulate training data
    for (int i = 0; i < (int)baseNames.size(); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "processing image " << baseNames[i] << " ("
            << (i + 1) << " of " << baseNames.size() << ")...");
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        SceneInstance instance(imgFilename.c_str());

        // locate horizon for this scene
        if (bUseGeometry) {
            if (horizons.find(baseNames[i]) == horizons.end()) {
                SVL_LOG(SVL_LOG_ERROR, "could not find horizon for " << baseNames[i]);
            } else {
                instance.assignment.horizon = horizons[baseNames[i]];
                instance.assignment.horizon = std::min(std::max(instance.assignment.horizon, 0), instance.height() - 1);
            }
        }

        // initialize pixel features
        this->initializeInstance(&instance);

        // compute region features
        int initialCacheSize = trainingData.size();
        computeRegionTrainingFeatures(instance, trainingData);
        
        // update label counts/weights
        for (int k = initialCacheSize; k < trainingData.size(); k++) {
            // adjust "invalid" labels
            if (trainingData.labels[k] < 0) 
                trainingData.labels[k] = nLabels - 1;

            // update counts;
            labelCounts[trainingData.labels[k]] += 1;
            labelWeight[trainingData.labels[k]] += trainingData.weights[k];
        }

        SVL_LOG(SVL_LOG_VERBOSE, "...accumulated " << toString(labelCounts) << " samples");
    }

    // add data from closed-loop cache
    // TODO: weight these differently to GT samples
    if (!gSceneConfig.getOption("closedLoopFeatureCache").empty()) {
        SVL_LOG(SVL_LOG_VERBOSE, "adding closed-loop learning features");
        SceneFeatureCache cache((gSceneConfig.getOption("baseDir") +
                gSceneConfig.getOption("closedLoopFeatureCache")).c_str());
        for (int k = 0; k < cache.size(); k++) {
            SVL_ASSERT_MSG(cache.labels[k] < gSceneConfig.numRegionLabels, 
                cache.labels[k] << " !< " << gSceneConfig.numRegionLabels);

            double w = cache.weights[k];
            if (w < 100) continue;

            trainingData.features.push_back(cache.features[k]);
            trainingData.labels.push_back(cache.labels[k] < 0 ? nLabels - 1 : cache.labels[k]);
            trainingData.weights.push_back(0.1 * w); // TODO: remove hardcoded weight

            labelCounts[trainingData.labels.back()] += 1;
            labelWeight[trainingData.labels.back()] += trainingData.weights.back();
        }

        SVL_LOG(SVL_LOG_VERBOSE, "...accumulated " << toString(labelCounts) << " samples");
    }

    // train boosted region features (TODO: move to another function)
    if (gSceneConfig.getOptionAsBool("useBoostedRegionFeatures")) {
        SVL_LOG(SVL_LOG_VERBOSE, "...learning boosted region features");
        if (_regionClassModels != NULL)
            delete _regionClassModels;

        _regionClassModels = new svlBoostedClassifierSet(gSceneConfig.numRegionLabels);
        double avgRecall = _regionClassModels->train(trainingData.features, trainingData.labels);
        SVL_LOG(SVL_LOG_MESSAGE, "...average recall: " << avgRecall);

        vector<vector<double> > boostedFeatures;
        _regionClassModels->getClassScores(trainingData.features, boostedFeatures);
        for (int i = 0; i < trainingData.size(); i++) {
            trainingData.features[i].insert(trainingData.features[i].end(), 
                boostedFeatures[i].begin(), boostedFeatures[i].end());
        }
    }

    // apply logistic feature kernel
    for (int k = 0; k < trainingData.size(); k++) {
        // include square terms
        int numFeatures = (int)trainingData.features[k].size();
        trainingData.features[k].reserve(2 * numFeatures + 1);
        for (int t = 0; t < numFeatures; t++) {
            trainingData.features[k].push_back(trainingData.features[k][t] * trainingData.features[k][t]);
        }
            
        // add bias term
        trainingData.features[k].push_back(1.0);
    }

    // train feature whitener
    SVL_LOG(SVL_LOG_VERBOSE, "whitening features...");
    _featureWhitener.clear(trainingData.numFeatures());
    _featureWhitener.train(trainingData.features);
    _featureWhitener.evaluate(trainingData.features);

    // balance training data
    if (gSceneConfig.getOptionAsBool("balancedSingletons")) {
        for (int i = 0; i < trainingData.size(); i++) {
            //featureWeights[i] /= (double)(labelCounts[featureLabels[i]] + 1.0);
            trainingData.weights[i] /= (double)(labelWeight[trainingData.labels[i]] + 1.0);
        }
    }

    // train model
    SVL_LOG(SVL_LOG_VERBOSE, "learning region singletons...");
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("trainRegionSingletons"));
    double lambda = gSceneConfig.getOptionAsDouble("regRegionSingleton");
    if (!bResume) {
        _regionSingletons = svlMultiClassLogistic(trainingData.numFeatures(), nLabels, lambda);
    }
    SVL_ASSERT(_regionSingletons.numFeatures() == trainingData.numFeatures());
    SVL_ASSERT(_regionSingletons.numClasses() == nLabels);
    _regionSingletons.setOption("maxIterations", maxIterations);
    _regionSingletons.setOption("regStrength", lambda);
    _regionSingletons.train(trainingData.features, trainingData.labels, trainingData.weights);
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("trainRegionSingletons"));

    // evaluate (on training data)
    if (svlLogger::getLogLevel() <= SVL_LOG_VERBOSE) {
        vector<int> predictedLabels;
        _regionSingletons.getClassifications(trainingData.features, predictedLabels);

        svlConfusionMatrix confusion(nLabels);
        confusion.accumulate(trainingData.labels, predictedLabels);
        confusion.write(cout);
        SVL_LOG(SVL_LOG_VERBOSE, "Accuracy: " << confusion.accuracy());
    }
}

void SceneRegionModel::learnRegionPairwiseContrastWeight(const vector<string>& baseNames)
{
    SVL_ASSERT_MSG(false, "TODO");
}

// initialize instance (cache boosted features and pixel appearances)
void SceneRegionModel::initializeInstance(SceneInstance *instance)
{
    SVL_LOG(SVL_LOG_VERBOSE, "initializing instance " << instance->name() << "...");

    // boosted pixel features
    ScenePixelModel::initializeInstance(instance);

    // load auxiliary features
    vector<CvMat *> auxFeatures = loadAuxiliaryFeatures(*instance);

    // appearance vector for each pixel
    int pixIndx = 0;
    instance->pixelAppearance.resize(instance->size());
    for (int y = 0; y < instance->height(); y++) {
        for (int x = 0; x < instance->width(); x++, pixIndx++) {
            // filter features
            instance->pixelAppearance[pixIndx] = instance->pixelFilterFeatures(x, y);

            // pixel boosted classifiers
#if 0
            instance->pixelAppearance[pixIndx].insert(instance->pixelAppearance[pixIndx].end(),
                instance->pixelBoostedResponses[pixIndx].begin(),
                instance->pixelBoostedResponses[pixIndx].end());
#else
            instance->pixelAppearance[pixIndx].insert(instance->pixelAppearance[pixIndx].end(),
                instance->pixelSingletons[pixIndx].begin(),
                instance->pixelSingletons[pixIndx].end());
#endif

            // auxiliary features
            for (int i = 0; i < (int)auxFeatures.size(); i++) {
                instance->pixelAppearance[pixIndx].push_back(
                    CV_MAT_ELEM(*auxFeatures[i], float, y, x));
            }
        }
    }

    // free auxiliary features
    svlReleaseMatrices(auxFeatures);
}

// inference
double SceneRegionModel::inferPixels(SceneInstance *instance)
{
    SVL_FCN_TIC;
    SVL_ASSERT(instance != NULL);

    // cache pixel appearance
    if (instance->pixelAppearance.empty()) {
        initializeInstance(instance);
    }


    // output region/superpixel features for richard
    string outputFolder =  gSceneConfig.filebase("featuresDir", "");
    outputFolder = outputFolder + "/richard";
    for (int i = 0; i < 3; i++) {

        string outputFn;
        string segFn;
        if (i == 0) {
            outputFn = outputFolder + "/" + instance->name() + ".0.txt";
            segFn = outputFolder + "/" + instance->name() + ".0.seg";
        } else if (i == 1) {
            outputFn = outputFolder + "/" + instance->name() + ".1.txt";
            segFn = outputFolder + "/" + instance->name() + ".1.seg";
        } else {
            outputFn = outputFolder + "/" + instance->name() + ".2.txt";
            segFn = outputFolder + "/" + instance->name() + ".2.seg";
        }

        cout << "output file: " << outputFn << endl;

        ofstream ofs(outputFn.c_str());
        SVL_ASSERT_MSG(ofs.fail() == false, "cannot open file " << outputFn);
        instance->assignPixels(instance->segmentations.segmentation(i));

        cout << "num of regions: " << instance->numRegions() << endl;
        // update cache with singleton potentials and assign regions
        //        updateInstanceCache(instance);

        for (int regId = 0; regId < instance->numRegions(); regId++) {
            // check for empty regions
            if (instance->assignment.regions[regId]->empty()) {
                cout << "skipped region..." << endl; 
                continue;
            }
            
            // assign labels
            instance->assignment.regions[regId]->Sk = argmax(instance->assignment.regions[regId]->phi);
            
            // output region/superpixel features for richard
            vector<double> f;
            SceneRegionFeatures::regionFeatures(instance, regId, f);
            // can be removed
            if (_regionClassModels != NULL) {
                vector<double> x;
                _regionClassModels->getClassScores(f, x);
                cout << "x size: " << x.size() << endl;
                f.insert(f.end(), x.begin(), x.end());
            }

            for (int j = 0; j < f.size(); j++) {
                ofs << f[j] << " ";
            }
            ofs << endl;


        }

        ofs.close();
        writeMatrix(instance->assignment.pixels, segFn.c_str());
    }

    SVL_FCN_TOC;
    return 0;

    // use first over-segmentation
    SVL_ASSERT(instance->segmentations.count() != 0);
    instance->assignPixels(instance->segmentations.segmentation(0));

    // update cache with singleton potentials and assign regions
    updateInstanceCache(instance);



    for (int regId = 0; regId < instance->numRegions(); regId++) {
        // check for empty regions
        if (instance->assignment.regions[regId]->empty())
            continue;

        // assign labels
        instance->assignment.regions[regId]->Sk = argmax(instance->assignment.regions[regId]->phi);

        // output region/superpixel features for richard

    
    }    

    // compute energy of assignment
    double e = energy(instance);

    SVL_FCN_TOC;
    return e;
}

double SceneRegionModel::inferHorizon(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);

    // no horizon model
    if (!gSceneConfig.getOptionAsBool("useGeometry")) {
        return energy(instance);
    }

    vector<vector<int> > aboveCount(instance->numRegions(),
        vector<int>(instance->height(), 0));

    for (int y = 1; y < instance->height(); y++) {
        for (int i = 0; i < instance->numRegions(); i++) {
            aboveCount[i][y] = aboveCount[i][y - 1];
        }
        for (int x = 0; x < instance->width(); x++) {
            int regId = CV_MAT_ELEM(*instance->assignment.pixels, int, y - 1, x);
            SVL_ASSERT(regId >= 0);
            aboveCount[regId][y] += 1;
        }
    }

    double bestScore = -SVL_DBL_MAX;
    for (int vHorizon = 0; vHorizon < instance->height(); vHorizon++) {
        double score = 0.0;
        for (int regId = 0; regId < instance->numRegions(); regId++) {
            score += aboveCount[regId][vHorizon] *
                _horizonRegionPairwise(instance->assignment.regions[regId]->Sk, 0) +
                (double)(instance->assignment.regions[regId]->size() - aboveCount[regId][vHorizon]) *
                _horizonRegionPairwise(instance->assignment.regions[regId]->Sk, 1);
        }

        if (score > bestScore) {
            bestScore = score;
            instance->assignment.horizon = vHorizon;
        }
    }

    // clear feature cache since horizon has changed
    instance->assignment.clearCachedModelData();

    return energy(instance);
}

double SceneRegionModel::inferRegions(SceneInstance *instance)
{
    SVL_FCN_TIC;
    SVL_ASSERT(instance != NULL);

    double e = 0.0;

    // update cache with singleton potentials
    updateInstanceCache(instance);

    // maximize each singleton
    for (int k = 0; k < instance->numRegions(); k++) {
        // check for empty regions
        if (instance->assignment.regions[k]->empty())
            continue;

        // assign labels
        SVL_ASSERT(instance->assignment.regions[k]->phi.size() == gSceneConfig.numRegionLabels);
        instance->assignment.regions[k]->Sk = argmax(instance->assignment.regions[k]->phi);

        // update energy
        e -= (double)instance->assignment.regions[k]->size() *
            instance->assignment.regions[k]->phi[instance->assignment.regions[k]->Sk];
    }

    // add boundary term to energy
    e += boundaryPenalty(*instance);

    SVL_FCN_TOC;
    return e;
}

// energy for complete assignment
double SceneRegionModel::energy(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);
    return singletonEnergy(instance) + boundaryPenalty(*instance);
}

// energy components
double SceneRegionModel::boundaryPenalty(const SceneInstance &instance) const
{
    double pixelwisePenalty = ScenePixelModel::boundaryPenalty(instance);

    // compute appearance similarity
    double regionwisePenalty = 0.0;
    if (_regionPairwiseContrastWeight > 0.0) {
        // boundary penalty is cached in contrast field
        for (vector<SceneRegionBoundary *>::const_iterator it = instance.assignment.regionBoundaries.begin();
             it != instance.assignment.regionBoundaries.end(); it++) {
            if ((*it)->contrast == 0.0) {
                SVL_ASSERT((*it)->length > 0);
                double c = instance.regionContrast((*it)->regionA->id(), (*it)->regionB->id());
                (*it)->contrast = (double)(*it)->length * exp(-0.5 * c);
            }
            regionwisePenalty += (*it)->contrast;
        }
        
        regionwisePenalty *= _regionPairwiseContrastWeight;
    }

    return regionwisePenalty + pixelwisePenalty;
}

double SceneRegionModel::singletonEnergy(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);
    SVL_FCN_TIC;

    double e = 0.0;

    // update cache with singleton potentials
    updateInstanceCache(instance);

    // maximize each singleton
    for (int k = 0; k < instance->numRegions(); k++) {
        // check for empty regions
        if (instance->assignment.regions[k]->empty())
            continue;

        //SVL_ASSERT(instance->assignment.regions[k].phi.size() == gSceneConfig.numRegionLabels);
        int label = instance->assignment.regions[k]->Sk;
        SVL_ASSERT((label >= 0) && (label < gSceneConfig.numRegionLabels));

        // update energy
        e -= (double)instance->assignment.regions[k]->size() *
            instance->assignment.regions[k]->phi[label];
    }

    //SVL_LOG(SVL_LOG_VERBOSE, "singletonEnergy: " << e);
    SVL_FCN_TOC;
    return e;
}

void SceneRegionModel::updateInstanceCache(SceneInstance *instance)
{
    double pixelScoreWeight = gSceneConfig.getOptionAsDouble("addWeightedPixelScore");

    // update each singleton
    for (int k = 0; k < instance->numRegions(); k++) {
        // check for empty regions
        if (instance->assignment.regions[k]->empty())
            continue;

        // update unary potential
        if (instance->assignment.regions[k]->phi.size() == 0) {
            // whiten features
            vector<double> f;
            SceneRegionFeatures::regionFeatures(instance, k, f);
            if (_regionClassModels != NULL) {
                vector<double> x;
                _regionClassModels->getClassScores(f, x);
                f.insert(f.end(), x.begin(), x.end());
            }

            // include square terms
            int numFeatures = (int)f.size();
            f.reserve(2 * numFeatures + 1);
            for (int t = 0; t < numFeatures; t++) {
                f.push_back(f[t] * f[t]);
            }
            
            // add bias term
            f.push_back(1.0);
           
            // whiten features
            _featureWhitener.evaluate(f);

            // singletons
            vector<double> y;
#if 1
            _regionSingletons.getMarginals(f, y);

            // remove "invalid" region label (assignment then becomes second highest)
            instance->assignment.regions[k]->phi = 
                (Eigen::Map<VectorXd>(&y[0], y.size() - 1).cwise() + SVL_DBL_MIN).cwise().log();
#else
            _regionSingletons.getClassScores(f, y);
            // remove "invalid" region label (assignment then becomes second highest)
            instance->assignment.regions[k]->phi = Eigen::Map<VectorXd>(&y[0], y.size() - 1);
#endif

            // add multiple of pixel scores
            if (pixelScoreWeight > 0.0) {
                instance->assignment.regions[k]->phi +=
                    (pixelScoreWeight / instance->assignment.regions[k]->size()) *
                    instance->assignment.regions[k]->pixelScores;
            }

#if 0
            // DEBUGGING
            SVL_LOG(SVL_LOG_DEBUG, instance->assignment.regions[k]->appearance().transpose());
            SVL_LOG(SVL_LOG_DEBUG, (instance->assignment.regions[k]->pixelScores /
                    instance->assignment.regions[k]->size()).transpose());
#endif

        }
    }
}

vector<double> SceneRegionModel::computeRegionFeatureVector(SceneInstance *instance, int regId)
{
    vector<double> f;    
    SceneRegionFeatures::regionFeatures(instance, regId, f);
     
    // include square terms
    f.reserve(2 * f.size() + 1);
    int numFeatures = (int)f.size();
    for (int i = 0; i < numFeatures; i++) {
        f.push_back(f[i] * f[i]);
    }

    // add bias term
    f.push_back(1.0);

    return f;
}

void SceneRegionModel::computeRegionTrainingFeatures(SceneInstance &instance, SceneFeatureCache &cache)
{
    SVL_FCN_TIC;

    // check if instance is in the cache
    if (gSceneConfig.getOptionAsBool("useCache") && (!instance.name().empty())) {
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "regionFeaturesExt");
        if (svlFileExists(filename.c_str())) {
            svlCodeProfiler::tic(svlCodeProfiler::getHandle("readRegionFeatureCache"));
            cache.read(filename.c_str(), true);
            svlCodeProfiler::toc(svlCodeProfiler::getHandle("readRegionFeatureCache"));
            SVL_FCN_TOC;
            return;
        }
    }

    // not in cache (or cache not enabled), so compute
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("computeRegionFeatures"));
    int initialCacheSize = cache.size();

    // groundtruth layer file and labels
    SceneLayersAndLabels layers(instance.width(), instance.height());
    layers.load(gSceneConfig.filename("lblDir", instance.name(), "layerLblExt").c_str(),
        gSceneConfig.filename("lblDir", instance.name(), "layerClassLblExt").c_str());

    instance.assignPixels(layers.layers);

#if 0
    // visualize
    for (int k = 0; k < instance.numRegions(); k++) {
        instance.assignment.regions[k]->Sk = layers.labels[k];
    }
    IplImage *canvas = cvCloneImage(instance.image());
    visualizeClasses(&instance, canvas, 0.5);
    visualizeBoundaries(&instance, canvas);
    cvNamedWindow("SceneRegionModel::debug", 1);
    cvShowImage("SceneRegionModel::debug", canvas);
    cvWaitKey(100);
    cvReleaseImage(&canvas);        
#endif

    // accumulate features and labels
    cache.reserve(cache.size() + instance.numRegions());
    for (int k = 0; k < instance.numRegions(); k++) {
        if (instance.assignment.regions[k]->size() < 10) continue;
            
        if (layers.labels[k] < 0) continue; // layer labels < 0 => unknown (e.g. MSRC)
        //SVL_ASSERT_MSG(layers.labels[k] < gSceneConfig.numRegionLabels, 
        //layers.labels[k] << " !< " << gSceneConfig.numRegionLabels);

        if (layers.labels[k] >= gSceneConfig.numRegionLabels) {
            layers.labels[k] = gSceneConfig.numRegionLabels - 1;
        }
        

        cache.features.push_back(vector<double>());
        SceneRegionFeatures::regionFeatures(&instance, k, cache.features.back());
        cache.labels.push_back(layers.labels[k]);
        cache.weights.push_back((double)instance.assignment.regions[k]->size());
    }

    // generate "invalid" labels (by merging groundtruth regions)
    list<pair<int, int> > regionPairs;
    for (vector<SceneRegionBoundary *>::const_iterator it = instance.assignment.regionBoundaries.begin();
         it != instance.assignment.regionBoundaries.end(); it++) {
        // sub-sample
        //if (rand() % instance.numRegions() != 0) continue;
        if (((*it)->regionA->size() < 10) || ((*it)->regionB->size() < 10))
            continue;
        regionPairs.push_back(make_pair((*it)->regionA->id(), (*it)->regionB->id()));
    }
        
    for (list<pair<int, int> >::const_iterator it = regionPairs.begin();
         it != regionPairs.end(); it++) {
        vector<CvPoint> pixels(instance.assignment.regions[it->first]->pixels.begin(),
            instance.assignment.regions[it->first]->pixels.end());
        instance.movePixels(pixels, it->second);
            
        cache.features.push_back(vector<double>());
        SceneRegionFeatures::regionFeatures(&instance, it->second, cache.features.back());
        cache.labels.push_back(-1);
        cache.weights.push_back((double)instance.assignment.regions[it->second]->size());
    
        instance.movePixels(pixels, it->first);
    }

    // generate additional labels from over-segmentations
    instance.loadSegmentations();
    for (int t = 0; t < instance.segmentations.count(); t++) {
        SceneLayersAndLabels segLayers(layers);
        segLayers.resegment(instance.segmentations.segmentation(t));
            
        instance.assignPixels(segLayers.layers);

#if 0
        // visualize
        bool bStop = false;
        for (int k = 0; k < instance.numRegions(); k++) {
            instance.assignment.regions[k]->Sk = segLayers.labels[k];
            if (segLayers.labels[k] < 0) bStop = true;
        }
        IplImage *canvas = cvCloneImage(instance.image());
        visualizeClasses(&instance, canvas, 1.0);
        visualizeBoundaries(&instance, canvas);
        cvNamedWindow("SceneRegionModel::debug2", 1);
        cvShowImage("SceneRegionModel::debug2", canvas);
        cvWaitKey(bStop ? -1 : 100);
        cvReleaseImage(&canvas);        
#endif

        // accumulate features and labels
        cache.reserve(cache.size() + instance.numRegions());
        for (int k = 0; k < instance.numRegions(); k++) {
            if (instance.assignment.regions[k]->size() < 100) continue;
                
            if (segLayers.labels[k] < 0) continue; // layer labels < 0 => unknown (e.g. MSRC)
            SVL_ASSERT_MSG(segLayers.labels[k] < gSceneConfig.numRegionLabels, 
                segLayers.labels[k] << " !< " << gSceneConfig.numRegionLabels);

            cache.features.push_back(vector<double>());
            SceneRegionFeatures::regionFeatures(&instance, k, cache.features.back());
            cache.labels.push_back(segLayers.labels[k]);
            cache.weights.push_back((double)instance.assignment.regions[k]->size());
        }
    }

    svlCodeProfiler::toc(svlCodeProfiler::getHandle("computeRegionFeatures"));

    // write to cache
    if (gSceneConfig.getOptionAsBool("useCache") && (!instance.name().empty())) {
        svlCodeProfiler::tic(svlCodeProfiler::getHandle("writeRegionFeatureCache"));
        string filename = gSceneConfig.filename("featuresDir",
            instance.name(), "regionFeaturesExt");
        if (initialCacheSize < cache.size()) {
            cache.write(filename.c_str(), initialCacheSize, cache.size() - 1, false);
        }
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("writeRegionFeatureCache"));
    }

    SVL_FCN_TOC;    
}

// SceneRegionPairModel -----------------------------------------------------

SceneRegionPairModel::SceneRegionPairModel() : SceneRegionModel(),
    _regionPairwiseWeight(1.0)
{
    // do nothing
}

SceneRegionPairModel::~SceneRegionPairModel()
{
    // do nothing
}

// i/o
void SceneRegionPairModel::loadModels(const char *modelDir)
{
    SceneRegionModel::loadModels(modelDir);

    loadRegionPairwiseModel(modelDir);
    loadRegionPairwiseWeight(modelDir);
}

void SceneRegionPairModel::loadRegionPairwiseModel(const char *modelDir)
{
    string fileName = string(modelDir) + string("/region.pairwiseFeatureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading region pairwise feature whitening parameters from " << fileName);
    _regionPairwiseFeatureWhitener.load(fileName.c_str());

    fileName = string(modelDir) + string("/region.pairwise.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading region pairwise parameters from " << fileName);
    _regionPairwisePotential.load(fileName.c_str());
}

void SceneRegionPairModel::saveRegionPairwiseModel(const char *modelDir)
{
    string fileName = string(modelDir) + string("/region.pairwiseFeatureWhitener.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing region pairwise feature whitening parameters to " << fileName);
    _regionPairwiseFeatureWhitener.save(fileName.c_str());

    fileName = string(modelDir) + string("/region.pairwise.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing region pairwise parameters to " << fileName);
    _regionPairwisePotential.save(fileName.c_str());
}

void SceneRegionPairModel::loadRegionPairwiseWeight(const char *modelDir)
{
    string fileName = string(modelDir) + string("/region.pairwiseWeight.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...reading region pairwise weight from " << fileName);
    ifstream ifs(fileName.c_str());
    SVL_ASSERT(!ifs.fail());
    ifs >> _regionPairwiseWeight;
    SVL_ASSERT(!ifs.fail());
    ifs.close();
}

void SceneRegionPairModel::saveRegionPairwiseWeight(const char *modelDir)
{
    string fileName = string(modelDir) + string("/region.pairwiseWeight.txt");
    SVL_LOG(SVL_LOG_VERBOSE, "...writing region pairwise weight to " << fileName);
    ofstream ofs(fileName.c_str());
    ofs << _regionPairwiseWeight;
    SVL_ASSERT(!ofs.fail());
    ofs.close();
}

// learning
void SceneRegionPairModel::learnRegionPairwiseModel(const vector<string>& baseNames, 
    int maxIterations, bool bResume)
{
    // load horizon groundtruth
    map<string, int> horizons = SceneDataLoader::loadHorizonTrainingData();

    // feature values and labels for training
    vector<vector<double> > featureVectors;
    vector<int> featureLabels;
    vector<double> featureWeights;
    int nLabels = gSceneConfig.numRegionLabels * gSceneConfig.numRegionLabels;
    vector<int> labelCounts(nLabels, 0);

    // accumulate training data
    for (int i = 0; i < (int)baseNames.size(); i++) {
        SVL_LOG(SVL_LOG_VERBOSE, "processing image " << baseNames[i] << " ("
            << (i + 1) << " of " << baseNames.size() << ")...");
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        SceneInstance instance(imgFilename.c_str());

        // locate horizon for this scene
        if (horizons.find(baseNames[i]) == horizons.end()) {
            SVL_LOG(SVL_LOG_ERROR, "could not find horizon for " << baseNames[i]);
        } else {
            instance.assignment.horizon = horizons[baseNames[i]];
            instance.assignment.horizon = std::min(std::max(instance.assignment.horizon, 0), instance.height() - 1);
        }

        // initialize pixel features
        this->initializeInstance(&instance);

        // load training labels for this image
        SceneLayersAndLabels layers(instance.width(), instance.height());
        layers.load(gSceneConfig.filename("lblDir", baseNames[i], "layerLblExt").c_str(),
            gSceneConfig.filename("lblDir", baseNames[i], "layerClassLblExt").c_str());

        instance.assignPixels(layers.layers);

        // accumulate features and labels
        for (vector<SceneRegionBoundary *>::const_iterator it = instance.assignment.regionBoundaries.begin();
                 it != instance.assignment.regionBoundaries.end(); it++) {
            
            if (layers.labels[(*it)->regionA->id()] < 0) continue;
            if (layers.labels[(*it)->regionB->id()] < 0) continue;
            
            // A->B
            int lbl = layers.labels[(*it)->regionA->id()] * gSceneConfig.numRegionLabels +
                layers.labels[(*it)->regionB->id()];
            SVL_ASSERT(lbl < nLabels);
            
            featureVectors.push_back(computeRegionBoundaryFeatureVector(&instance, *it));
            featureLabels.push_back(lbl);
            featureWeights.push_back((double)((*it)->regionA->size() + (*it)->regionB->size()));
            labelCounts[featureLabels.back()] += 1;
            
            // B->A
            lbl = layers.labels[(*it)->regionB->id()] * gSceneConfig.numRegionLabels +
                layers.labels[(*it)->regionA->id()];
            SVL_ASSERT(lbl < nLabels);
            
            SceneRegionBoundary b(*(*it));
            std::swap(b.regionA, b.regionB);
            featureVectors.push_back(computeRegionBoundaryFeatureVector(&instance, &b));
            featureLabels.push_back(lbl);
            featureWeights.push_back((double)(b.regionA->size() + b.regionB->size()));
            labelCounts[featureLabels.back()] += 1;
            
        }

        SVL_LOG(SVL_LOG_VERBOSE, "...accumulated " << toString(labelCounts) << " samples");
    }

    // train feature whitener
    SVL_LOG(SVL_LOG_VERBOSE, "whitening features...");
    _regionPairwiseFeatureWhitener.clear(featureVectors[0].size());
    _regionPairwiseFeatureWhitener.train(featureVectors);
    _regionPairwiseFeatureWhitener.evaluate(featureVectors);

    // balance training data
    if (gSceneConfig.getOptionAsBool("balancedSingletons")) {
        for (int i = 0; i < (int)featureWeights.size(); i++) {
            featureWeights[i] /= (double)(labelCounts[featureLabels[i]] + 1.0);
        }
    }

    // train model
    SVL_LOG(SVL_LOG_VERBOSE, "learning region pairwise...");
    svlCodeProfiler::tic(svlCodeProfiler::getHandle("trainRegionPairwise"));
    double lambda = gSceneConfig.getOptionAsDouble("regRegionPairwise");
    if (!bResume) {
        _regionPairwisePotential = svlMultiClassLogistic(featureVectors[0].size(), nLabels, lambda);
    }
    SVL_ASSERT(_regionPairwisePotential.numFeatures() == (int)featureVectors[0].size());
    SVL_ASSERT(_regionPairwisePotential.numClasses() == nLabels);
    _regionPairwisePotential.setOption("maxIterations", maxIterations);
    _regionPairwisePotential.train(featureVectors, featureLabels, featureWeights);
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("trainRegionPairwise"));

    // evaluate (on training data)
    if (svlLogger::getLogLevel() <= SVL_LOG_VERBOSE) {
        vector<int> predictedLabels;
        _regionPairwisePotential.getClassifications(featureVectors, predictedLabels);

        svlConfusionMatrix confusion(nLabels);
        confusion.accumulate(featureLabels, predictedLabels);
        confusion.write(cout);
        SVL_LOG(SVL_LOG_VERBOSE, "Accuracy: " << confusion.accuracy());
    }
}

void SceneRegionPairModel::learnRegionPairwiseWeight(const vector<string>& baseNames)
{
    // TODO
    SVL_ASSERT_MSG(false, "TODO");
}

// inference
double SceneRegionPairModel::inferRegions(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);

    // construct cluster graph
    SVL_LOG(SVL_LOG_VERBOSE, "building graph on regions...");
    svlClusterGraph *graph = buildGraph(*instance);

    // run inference
    SVL_LOG(SVL_LOG_VERBOSE, "running inference on regions...");
    svlFactorOperation::CACHE_INDEX_MAPPING = true;
    svlFactorOperation::USE_SHARED_INDEX_CACHE = false;

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("inference"));
    svlMessagePassingInference infObject(*graph);
    infObject.inference(SVL_MP_ASYNCLOGMAXPROD, 250);
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("inference"));

    // decode MAP
    for (int regId = 0; regId < instance->numRegions(); regId++) {
        SVL_ASSERT(infObject[regId].size() == gSceneConfig.numRegionLabels);
        instance->assignment.regions[regId]->Sk = infObject[regId].indexOfMax();
    }

    // get energy and delete graph
    double e = clusterGraphEnergy(instance, graph);
    delete graph;
    return e + boundaryPenalty(*instance);
}

// energy for complete assignment
double SceneRegionPairModel::energy(SceneInstance *instance)
{
    SVL_ASSERT(instance != NULL);

    svlClusterGraph *graph = buildGraph(*instance);
    double e = clusterGraphEnergy(instance, graph);

    // check energy
    if (!isfinite(e)) {
        instance->validateData();
        instance->printScene();
        graph->write(cout);
        SVL_LOG(SVL_LOG_FATAL, "energy is not finite");
    }

    delete graph;
    return e + boundaryPenalty(*instance);
}

// updating instance cache with region unary and pairwise potentials
void SceneRegionPairModel::updateInstanceCache(SceneInstance *instance)
{
    // update singleton terms
    SceneRegionModel::updateInstanceCache(instance);

    // update pairwise terms
    for (vector<SceneRegionBoundary *>::const_iterator it = instance->assignment.regionBoundaries.begin();
         it != instance->assignment.regionBoundaries.end(); it++) {
        if (!(*it)->dirty()) continue;

        vector<double> f = computeRegionBoundaryFeatureVector(instance, *it);
        _regionPairwiseFeatureWhitener.evaluate(f);
        
        vector<double> y;
        _regionPairwisePotential.getMarginals(f, y);
        (*it)->phi = Eigen::Map<VectorXd>(&y[0], y.size()).cwise().log();               
    }
}

// energy for cluster graph potentials
double SceneRegionPairModel::clusterGraphEnergy(SceneInstance *instance, svlClusterGraph *graph) const
{
    SVL_ASSERT((graph != NULL) && (graph->numVariables() == instance->numRegions()));

    // extract assignment
    vector<int> assignment(instance->numRegions());
    for (int i = 0; i < instance->numRegions(); i++) {
        assignment[i] = instance->assignment.regions[i]->Sk;
    }

    return graph->getEnergy(assignment);
}

// general CRF functions
svlClusterGraph *SceneRegionPairModel::buildGraph(SceneInstance &instance)
{
    SVL_FCN_TIC;

    // update cached singleton data
    updateInstanceCache(&instance);

    // initialize graph
    vector<int> cards(instance.numRegions(), gSceneConfig.numRegionLabels);

    svlClusterGraph *graph = new svlClusterGraph(cards.size(), cards);

    // add region unary terms
    for (vector<SceneRegion *>::const_iterator it = instance.assignment.regions.begin();
         it != instance.assignment.regions.end(); it++) {
        svlClique c; c.insert((*it)->id());
        svlFactor phi((*it)->id(), gSceneConfig.numRegionLabels);

        for (int i = 0; i < phi.size(); i++) {
            phi[i] = (double)(*it)->size() * (*it)->phi[i];
        }

        graph->addClique(c, phi);
    }

    vector<pair<int, int> > edges;
    edges.reserve(4 * instance.numRegions());

    // add region pairwise terms
    for (vector<SceneRegionBoundary *>::const_iterator it = instance.assignment.regionBoundaries.begin();
         it != instance.assignment.regionBoundaries.end(); it++) {
        svlClique c; 
        c.insert((*it)->regionA->id());
        c.insert((*it)->regionB->id());

        svlFactor phi;
        phi.addVariable((*it)->regionB->id(), gSceneConfig.numRegionLabels);
        phi.addVariable((*it)->regionA->id(), gSceneConfig.numRegionLabels);

        SVL_ASSERT_MSG(phi.size() == (*it)->phi.size(), phi.size() << " != " << (*it)->phi.size());
        double nu = _regionPairwiseWeight *
            ((double)(*it)->regionA->size() / (double)(*it)->regionA->boundaries.size() +
                (double)(*it)->regionB->size() / (double)(*it)->regionB->boundaries.size());
        for (int i = 0; i < phi.size(); i++) {
            phi[i] = nu * (*it)->phi[i];
        }

        edges.push_back(make_pair(graph->numCliques(), phi.variableId(0)));
        edges.push_back(make_pair(graph->numCliques(), phi.variableId(1)));
        graph->addClique(c, phi);
    }

    // connect graph
    graph->connectGraph(edges);

    SVL_FCN_TOC;
    return graph;
}

// feature vector calculations
vector<double> SceneRegionPairModel::computeRegionBoundaryFeatureVector(SceneInstance *instance, 
    SceneRegionBoundary *boundary)
{
    SVL_ASSERT((instance != NULL) && (boundary != NULL));
 
    vector<double> f;
    SceneRegionFeatures::regionPairFeatures(instance, boundary->regionA->id(),
        boundary->regionB->id(), f);

    return f;
}
