/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneRegionModel.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  This file contains common functionality for region-based scene understanding
**  models using texton filterbank based features.
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"
#include "sceneEntities.h"
#include "sceneInstance.h"
#include "sceneModel.h"

using namespace std;

// SceneRegionModel class -----------------------------------------------
// Implements region-based scene understanding model. Energy function has
// one multi-class logistic regression term per region and one class-independent
// pairwise boundary term per pair of adjacent regions.

class SceneRegionModel : public ScenePixelModel {
 protected:
    // singleton potetials (class predictors + "invalid" class)
    svlBoostedClassifierSet *_regionClassModels;
    svlFeatureWhitener _featureWhitener;
    svlMultiClassLogistic _regionSingletons;

    // horizon potentials
    svlGaussian _horizonSingleton;
    double _horizonSingletonWeight;
    MatrixXd _horizonRegionPairwise;

    // pairwise potential
    double _regionPairwiseContrastWeight;

 public:
    SceneRegionModel();
    ~SceneRegionModel();

    // i/o
    void loadModels(const char *modelDir);

    void loadHorizonModel(const char* modelDir);
    void saveHorizonModel(const char* modelDir);
    void loadHorizonPairwiseModel(const char* modelDir);
    void saveHorizonPairwiseModel(const char* modelDir);
    void loadRegionSingletonModel(const char *modelDir);
    void saveRegionSingletonModel(const char *modelDir);
    void loadRegionContrastWeight(const char *modelDir);
    void saveRegionContrastWeight(const char *modelDir);

    // learning
    void learnHorizonSingletonModel(vector<string>& baseNames);
    void learnHorizonPairwiseModel(vector<string>& baseNames, int subSample = 0);
    void learnRegionSingletonModel(const vector<string>& baseNames, 
        int maxIterations = 1000, bool bResume = false);
    void learnRegionPairwiseContrastWeight(const vector<string>& baseNames);

    // initialize instance (cache boosted features and pixel appearances)
    void initializeInstance(SceneInstance *instance);

    // inference
    double inferPixels(SceneInstance *instance);
    double inferHorizon(SceneInstance *instance);
    double inferRegions(SceneInstance *instance);

    // energy for complete assignment
    double energy(SceneInstance *instance);

 protected:
    // energy components
    double boundaryPenalty(const SceneInstance &instance) const;
    double singletonEnergy(SceneInstance *instance);

    // updating instance cache with region singletons
    virtual void updateInstanceCache(SceneInstance *instance);

    // feature vector calculations
    virtual vector<double> computeRegionFeatureVector(SceneInstance *instance, int regId);
    void computeRegionTrainingFeatures(SceneInstance &instance, SceneFeatureCache &cache);
};

// SceneRegionPairModel class -------------------------------------------
// Extends SceneRegionModel by adding a class-specific pairwise term.

class SceneRegionPairModel : public SceneRegionModel {
 protected:
    // pairwise potential
    svlFeatureWhitener _regionPairwiseFeatureWhitener;
    svlMultiClassLogistic _regionPairwisePotential;
    double _regionPairwiseWeight;

 public:
    SceneRegionPairModel();
    ~SceneRegionPairModel();

    // i/o
    void loadModels(const char *modelDir);

    void loadRegionPairwiseModel(const char *modelDir);
    void saveRegionPairwiseModel(const char *modelDir);
    void loadRegionPairwiseWeight(const char *modelDir);
    void saveRegionPairwiseWeight(const char *modelDir);

    // learning
    void learnRegionPairwiseModel(const vector<string>& baseNames, 
        int maxIterations = 1000, bool bResume = false);
    void learnRegionPairwiseWeight(const vector<string>& baseNames);

    // inference
    double inferRegions(SceneInstance *instance);

    // energy for complete assignment
    double energy(SceneInstance *instance);

 protected:
    // updating instance cache with region unary and pairwise potentials
    void updateInstanceCache(SceneInstance *instance);

    // energy for cluster graph potentials
    double clusterGraphEnergy(SceneInstance *instance, svlClusterGraph *graph) const;

    // general CRF functions
    svlClusterGraph *buildGraph(SceneInstance &instance);

    // feature vector calculations
    virtual vector<double> computeRegionBoundaryFeatureVector(SceneInstance *instance, 
        SceneRegionBoundary *boundary);
};
