/***************           **************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneRegionObjectClassifier.cpp
** AUTHOR(S):   Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**
*****************************************************************************/

#include "sceneRegionObjectClassifier.h"
#include "sceneHOGFeatures.h"

using namespace std;

SceneRegionObjectClassifier::SceneRegionObjectClassifier(const char *name)
{
    init(name);
}

SceneRegionObjectClassifier::~SceneRegionObjectClassifier()
{
    if (_owns_classifier && _classifier)
        delete _classifier;

    if (_owns_patchClassifier && _patchClassifier)
        delete _patchClassifier;
    
    if (_owns_extractor && _extractor)
        delete _extractor;
}

bool SceneRegionObjectClassifier::load(const char* modelFilename, const char* extractorFilename, bool bPatchBased)
{
    if (_owns_extractor && _extractor)
        delete _extractor;

    if (extractorFilename != NULL) {
        _extractor = new SceneHOGFeatures();
        bool bSucess = _extractor->loadExtractor(extractorFilename);
        if (bSucess == false) return false;
        _owns_extractor = true;
    }

    if ( modelFilename ) {
        if (bPatchBased) { // patch based classifier (svmlight linear classifier)
            SceneLinearClassifier* c = new SceneLinearClassifier(modelFilename);
            if (!c)  return false;
            if (_owns_patchClassifier && _patchClassifier)
                delete _patchClassifier;
            _patchClassifier = c;
            _owns_patchClassifier = true;
        } else { 
            svlBinaryClassifier *c = (svlBinaryClassifier *) svlClassifierFactory().load(modelFilename);
            if (!c) return false;
            if (_owns_classifier && _classifier)
                delete _classifier;
            _classifier = c;
            _owns_classifier = true;
        }
    }
    return true;
}

void SceneRegionObjectClassifier::loadImage(const char* imgFilname)
{
    SVL_ASSERT(_extractor != NULL);
    _extractor->load(imgFilname);
    _extractor->prePocessing();
}

void SceneRegionObjectClassifier::loadImage(IplImage* img)
{
    SVL_ASSERT(_extractor != NULL);
    _extractor->load(img);
    _extractor->prePocessing();
}

double SceneRegionObjectClassifier::classifyRegion(const CvMat* regionMask, IplImage* regionImg)
{
    if (regionImg != NULL) {
        loadImage(regionImg);
    }
    SVL_ASSERT(_extractor != NULL);
    SVL_ASSERT(_classifier != NULL);
    VectorXd f = _extractor->extractFeatures(regionMask);
    vector<double> fd(f.rows(), 0);
    Eigen::Map<VectorXd>(&fd[0], fd.size()) = f;
    float score = _classifier->getScore(fd);
    return _classifier->scoreToProbability(score);
}

double SceneRegionObjectClassifier::classifyRegion(const CvRect& r, IplImage *regionImg)
{
    SVL_ASSERT(regionImg != NULL);
    SVL_ASSERT_MSG((r.x >= 0) && (r.x + r.width <= regionImg->width) &&
        (r.y >= 0) && (r.y + r.height <= regionImg->height), toString(r));

    IplImage *subImage = cvCreateImage(cvSize(r.width, r.height), regionImg->depth, regionImg->nChannels);
    cvSetImageROI(regionImg, r);
    cvCopyImage(regionImg, subImage);
    cvResetImageROI(regionImg);

    double pr = classifyRegion(subImage);
    cvReleaseImage(&subImage);
    return pr;
}

double SceneRegionObjectClassifier::classifyRegion(IplImage* patchImg)
{
    SVL_ASSERT(patchImg != NULL);
    SVL_ASSERT(_extractor != NULL);
    SVL_ASSERT(_patchClassifier != NULL);
    
    CvRect r = cvRect(0, 0, _extractor->getWinWidth(), _extractor->getWinHeight());
    VectorXd f;
    if (patchImg->width != r.width  || patchImg->height != r.height) {
        IplImage* resizedPatchImg = cvCloneImage(patchImg);
        resizeInPlace(&resizedPatchImg, r.height, r.width);
        loadImage(resizedPatchImg);
        f = _extractor->extractFeatures(r);
        cvReleaseImage(&resizedPatchImg);
    } else {
        loadImage(patchImg);
        f = _extractor->extractFeatures(r);
    }
    return _patchClassifier->getScore(f);
}

void SceneRegionObjectClassifier::init(const char *name)
{
    _classifier = NULL;
    _patchClassifier = NULL;
    _extractor = NULL;
    _owns_classifier = false;
    _owns_patchClassifier = false;
    _owns_extractor = false;
    
    if (name == NULL) {
        _objectName = "[unknown]";
    } else {
        _objectName = string(name);
    }
}
