/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneScoreResults.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Scores results for scene models.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneDataLoader.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneVisualization.h"

#define WINDOW_NAME "sceneResults"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneScoreResults [OPTIONS] <evalList>\n";
    cerr << "OPTIONS:\n"
         << "  -regions <ext>    :: extension for region id output (default: none)\n"
         << "  -classes <ext>    :: extension for region class output (default: none)\n"
         << "  -horizon <ext>    :: extension for horizon output (default: none)\n"
         << "  -confusion        :: show confusion matrix\n"
         << "  -jaccard          :: use intersection over union metric (as in PASCAL)\n"
         << "  -x                :: visualize output\n"
         << SVL_STANDARD_OPTIONS_USAGE 
         << endl;
}

int main(int argc, char *argv[])
{
    // read commandline parameters
    captureOpenCVerrors();
    
    const char *extRegions = NULL;
    const char *extClasses = NULL;
    const char *extHorizon = NULL;
    bool bShowConfusion = false;
    bool bJaccardMetric = false;
    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-regions", extRegions)
        SVL_CMDLINE_STR_OPTION("-classes", extClasses)
        SVL_CMDLINE_STR_OPTION("-horizon", extHorizon)
        SVL_CMDLINE_BOOL_OPTION("-confusion", bShowConfusion)
        SVL_CMDLINE_BOOL_OPTION("-jaccard", bJaccardMetric)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // configuration
    const char *evalList = SVL_CMDLINE_ARGV[0];

    // read list of evaluation images
    SVL_LOG(SVL_LOG_MESSAGE, "Reading evaluation list from " << evalList << "...");
    vector<string> baseNames = svlReadFile(evalList);
    SVL_LOG(SVL_LOG_MESSAGE, "...read " << baseNames.size() << " images");

    // clear statistics
    svlConfusionMatrix regionConfusion(gSceneConfig.numRegionLabels);
    double regionScore = 0.0;
    double horizonScore = 0.0;
    vector<double> horizonErrors;

    // load horizon groundtruth
    map<string, int> horizons;
    if (extHorizon != NULL) {
        horizons = SceneDataLoader::loadHorizonTrainingData();
    }

    // process images
    SVL_LOG(SVL_LOG_MESSAGE, "Reading results...");
    for (int i = 0; i < (int)baseNames.size(); i++) {

        // load image
        string imgFilename = gSceneConfig.filename("imgDir", baseNames[i], "imgExt");
        IplImage *img = cvLoadImage(imgFilename.c_str(), CV_LOAD_IMAGE_COLOR);
        SVL_ASSERT(img != NULL);
        SVL_LOG(SVL_LOG_VERBOSE, baseNames[i] << " (" << (i + 1) << " of "
            << baseNames.size() << "): " << toString(*img));

        // load groundtruth and predicted class labels
        CvMat *classLabels = NULL;
        CvMat *predictedClasses = NULL;
        if (extClasses != NULL) {
            string lblFilename = gSceneConfig.filename("lblDir", baseNames[i], "regionLblExt");
            classLabels = cvCreateMat(img->height, img->width, CV_32SC1);
            SVL_ASSERT(classLabels != NULL);
            bool bSuccess = readMatrix(classLabels, lblFilename.c_str());
            SVL_ASSERT_MSG(bSuccess, "Could not read class groundtruth for " << baseNames[i]);

            predictedClasses = cvCreateMat(img->height, img->width, CV_32SC1);
            SVL_ASSERT(predictedClasses != NULL);
            bSuccess = readMatrix(predictedClasses, (gSceneConfig.filebase("outputDir", baseNames[i]) + 
                    string(extClasses)).c_str());
            SVL_ASSERT_MSG(bSuccess, "Could not read class predictions for " << baseNames[i]);            

            // accumulate confusion matrix
            for (int y = 0; y < img->height; y++) {
                for (int x = 0; x < img->width; x++) {
                    SVL_ASSERT_MSG(CV_MAT_ELEM(*predictedClasses, int, y, x) < 
                        gSceneConfig.numRegionLabels, CV_MAT_ELEM(*predictedClasses, int, y, x)
                        << " !< " << gSceneConfig.numRegionLabels);
                    regionConfusion.accumulate(CV_MAT_ELEM(*classLabels, int, y, x), 
                        CV_MAT_ELEM(*predictedClasses, int, y, x));
                }
            }
        }

        // horizon
        if (extHorizon != NULL) {
            int hPredicted;
            ifstream ifs((gSceneConfig.filebase("outputDir", baseNames[i]) +
                    string(extHorizon)).c_str());
            SVL_ASSERT_MSG(!ifs.fail(), "could not read horizon for " << baseNames[i]);
            ifs >> hPredicted;
            ifs.close();

            // locate horizon for this scene
            if (horizons.find(baseNames[i]) == horizons.end()) {
                SVL_LOG(SVL_LOG_WARNING, "could not find horizon groundtruth for " << baseNames[i]);
            } else {
                int hActual = horizons[baseNames[i]];
                double err = (double)(hActual - hPredicted) / (double)img->height;
                horizonErrors.push_back(fabs(err));
                horizonScore += fabs(err);
            }
        }

        // load groundtruth and predicted regions
        CvMat *regionLabels = NULL;
        CvMat *predictedRegions = NULL;
        if (extRegions != NULL) {
            string lblFilename = gSceneConfig.filename("lblDir", baseNames[i], "layerLblExt");
            regionLabels = cvCreateMat(img->height, img->width, CV_32SC1);
            SVL_ASSERT(regionLabels != NULL);
            bool bSuccess = readMatrix(regionLabels, lblFilename.c_str());
            SVL_ASSERT_MSG(bSuccess, "Could not read region groundtruth for " << baseNames[i]);
            
            predictedRegions = cvCreateMat(img->height, img->width, CV_32SC1);
            SVL_ASSERT(predictedRegions != NULL);
            bSuccess = readMatrix(predictedRegions, (gSceneConfig.filebase("outputDir", baseNames[i]) +
                    string(extRegions)).c_str());
            SVL_ASSERT_MSG(bSuccess, "Could not read region predictions for " << baseNames[i]);

            // if we have class labels then mask out non-foreground
            if (classLabels != NULL) {
                for (int y = 0; y < img->height; y++) {
                    for (int x = 0; x < img->width; x++) {
                        if (CV_MAT_ELEM(*classLabels, int, y, x) != 7) {
                            CV_MAT_ELEM(*regionLabels, int, y, x) = 0;
                        }
                        if (CV_MAT_ELEM(*predictedClasses, int, y, x) != 7) {
                            CV_MAT_ELEM(*predictedRegions, int, y, x) = 0;
                        }
                    }
                }
            }

            // find boundaries
            convertInPlace(&regionLabels, CV_32FC1);
            convertInPlace(&predictedRegions, CV_32FC1);
            CvMat *tmp = cvCreateMat(img->height, img->width, CV_32FC1);
            
            cvSobel(regionLabels, tmp, 1, 0, 1);
            cvAbs(tmp, tmp);
            cvSobel(regionLabels, regionLabels, 0, 1, 1);
            cvAbs(regionLabels, regionLabels);
            cvAdd(regionLabels, tmp, regionLabels);
            cvMinS(regionLabels, 1.0, regionLabels);

            cvSobel(predictedRegions, tmp, 1, 0, 1);
            cvAbs(tmp, tmp);
            cvSobel(predictedRegions, predictedRegions, 0, 1, 1);
            cvAbs(predictedRegions, predictedRegions);
            cvAdd(predictedRegions, tmp, predictedRegions);
            cvMinS(predictedRegions, 1.0, predictedRegions);

            // blur boundaries
            IplConvKernel* kernel = cvCreateStructuringElementEx(5, 5, 2, 2, CV_SHAPE_RECT);
            cvDilate(regionLabels, regionLabels, kernel);
            cvDilate(predictedRegions, predictedRegions, kernel);
            cvReleaseStructuringElement(&kernel);

            cvMul(regionLabels, predictedRegions, tmp);
            regionScore += 2.0 * cvSum(tmp).val[0] / (cvSum(regionLabels).val[0] + 
                cvSum(predictedRegions).val[0] + SVL_DBL_MIN);
                       
            cvReleaseMat(&tmp);            
        }
        
        // show results
        if (bVisualize) {
            SceneInstance instance(img);
            if (predictedClasses != NULL) {
                instance.assignPixelsFromLabels(predictedClasses);
            }

            IplImage *canvas = visualize(&instance);
            cvNamedWindow(WINDOW_NAME, 1);
            cvShowImage(WINDOW_NAME, canvas);
            cvWaitKey(100);

            cvReleaseImage(&canvas);
        }

        // free memory
        if (classLabels != NULL) {
            cvReleaseMat(&predictedClasses);
            cvReleaseMat(&classLabels);
        }
        if (regionLabels != NULL) {
            cvReleaseMat(&predictedRegions);
            cvReleaseMat(&regionLabels);
        }
        cvReleaseImage(&img);
    }

    regionScore /= (double)baseNames.size();
    horizonScore /= (double)baseNames.size();
    sort(horizonErrors.begin(), horizonErrors.end());

    if (bShowConfusion) {
        regionConfusion.printRowNormalized(cout, "--- Class Confusion Matrix ---");
        regionConfusion.printPrecisionRecall(cout, "--- Recall/Precision (by Class) ---");
        regionConfusion.printF1Score(cout, "--- F1-Score (by Class) ---");
    }
    if (bJaccardMetric) {
        regionConfusion.printJaccard(cout, "--- Intersection/Union Metric (by Class) ---");
    }

    SVL_LOG(SVL_LOG_MESSAGE, "Overall class accuracy: " << regionConfusion.accuracy());
    SVL_LOG(SVL_LOG_MESSAGE, "Average class accuracy: " << regionConfusion.avgRecall());
    SVL_LOG(SVL_LOG_MESSAGE, "Average Jaccard score: " << regionConfusion.avgJaccard());
    SVL_LOG(SVL_LOG_MESSAGE, "Region boundary score: " << regionScore);
    SVL_LOG(SVL_LOG_MESSAGE, "Horizon mean error: " << horizonScore);
    if (!horizonErrors.empty()) {
        SVL_LOG(SVL_LOG_MESSAGE, "Horizon median error: " << horizonErrors[horizonErrors.size() / 2]);
    }

    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

