/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneSegmentDict.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Dictionary of proposal segments.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <map>
#include <set>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneSegmentDict.h"

#define DICTIONARY_WINDOW_NAME "SegmentDictionary"

using namespace std;

// SceneSegment --------------------------------------------------------------

SceneSegment::SceneSegment()
{
    // do  nothing
}

SceneSegment::SceneSegment(const CvMat *regions, int n)
{
    initialize(regions, n);
}

SceneSegment::SceneSegment(const CvRect& r)
{
    initialize(r);
}

SceneSegment::SceneSegment(const svlMultiSeg& multiSegs, const vector<pair<int, int> >& segv)
{
    initialize(multiSegs, segv);
}

SceneSegment::SceneSegment(XMLNode& node)
{
    read(node);
}

SceneSegment::~SceneSegment()
{
    // do nothing
}

void SceneSegment::read(XMLNode &node)
{
    SVL_ASSERT(!node.isEmpty());
    clear();

    XMLNode childNode = node.getChildNode("Pixels"); 
    vector<int> v;
    parseString(childNode.getText(), v);
    SVL_ASSERT(v.size() % 2 == 0);
    for (int i = 0; i < (int)v.size(); i += 2) {
        pixels.push_back(cvPoint(v[i], v[i] + 1));
    }
    
    childNode = node.getChildNode("InsideBoundary"); 
    v.clear();
    parseString(childNode.getText(), v);
    SVL_ASSERT(v.size() % 2 == 0);
    for (int i = 0; i < (int)v.size(); i += 2) {
        insideBoundaryPixels.insert(cvPoint(v[i], v[i] + 1));
    }

    childNode = node.getChildNode("OutsideBoundary"); 
    v.clear();
    parseString(childNode.getText(), v);
    SVL_ASSERT(v.size() % 2 == 0);
    for (int i = 0; i < (int)v.size(); i += 2) {
        outsideBoundaryPixels.insert(cvPoint(v[i], v[i] + 1));
    }
 
    extent = cvRect(numeric_limits<int>::max(), numeric_limits<int>::max(), 0, 0);
    for (vector<CvPoint>::const_iterator p = pixels.begin(); p != pixels.end(); ++p) {
        if (p->x < extent.x) extent.x = p->x;
        if (p->y < extent.y) extent.y = p->y;
        if (p->x > extent.width) extent.width = p->x;
        if (p->y > extent.height) extent.height = p->y;
    }
    extent.width -= extent.x - 1;
    extent.height -= extent.y - 1;   
}

void SceneSegment::write(ostream& os) const
{
    os << "    <Pixels>\n";
    for (int i = 0; i < (int)pixels.size(); i++) {
        os << "      " << pixels[i].x << " " << pixels[i].y << "\n";
    }
    os << "    </Pixels>\n";

    os << "    <InsideBoundary>\n";
    for (set<CvPoint>::const_iterator p = insideBoundaryPixels.begin();
         p != insideBoundaryPixels.end(); p++) {
        os << "      " << p->x << " " << p->y << "\n";
    }
    os << "    </InsideBoundary>\n";

    os << "    <OutsideBoundary>\n";
    for (set<CvPoint>::const_iterator p = outsideBoundaryPixels.begin();
         p != outsideBoundaryPixels.end(); p++) {
        os << "      " << p->x << " " << p->y << "\n";
    }
    os << "    </OutsideBoundary>\n";
}

void SceneSegment::clear()
{
    outsideBoundaryPixels.clear();
    insideBoundaryPixels.clear();
    pixels.clear();
}

void SceneSegment::initialize(const CvMat *regions, int n)
{
    SVL_ASSERT(regions != NULL);
    clear();

    for (int y = 0; y < regions->rows; y++) {
        for (int x = 0; x < regions->cols; x++) {
            if (CV_MAT_ELEM(*regions, int, y, x) != n) {
                continue;
            }

            // pixel is part of the region
            pixels.push_back(cvPoint(x, y));

            // check 4-neighbours
            if (x == 0) {
                insideBoundaryPixels.insert(cvPoint(x, y));
            } else if (CV_MAT_ELEM(*regions, int, y, x - 1) != n) {
                insideBoundaryPixels.insert(cvPoint(x, y));
                outsideBoundaryPixels.insert(cvPoint(x - 1, y));
            }

            if (x == regions->cols - 1) {
                insideBoundaryPixels.insert(cvPoint(x, y));
            } else if (CV_MAT_ELEM(*regions, int, y, x + 1) != n) {
                insideBoundaryPixels.insert(cvPoint(x, y));
                outsideBoundaryPixels.insert(cvPoint(x + 1, y));
            }

            if (y == 0) {
                insideBoundaryPixels.insert(cvPoint(x, y));
            } else if (CV_MAT_ELEM(*regions, int, y - 1, x) != n) {
                insideBoundaryPixels.insert(cvPoint(x, y));
                outsideBoundaryPixels.insert(cvPoint(x, y - 1));
            }

            if (y == regions->rows - 1) {
                insideBoundaryPixels.insert(cvPoint(x, y));
            } else if (CV_MAT_ELEM(*regions, int, y + 1, x) != n) {
                insideBoundaryPixels.insert(cvPoint(x, y));
                outsideBoundaryPixels.insert(cvPoint(x, y + 1));
            }
        }
    }

    extent = cvRect(numeric_limits<int>::max(), numeric_limits<int>::max(), 0, 0);
    for (vector<CvPoint>::const_iterator p = pixels.begin(); p != pixels.end(); ++p) {
        if (p->x < extent.x) extent.x = p->x;
        if (p->y < extent.y) extent.y = p->y;
        if (p->x > extent.width) extent.width = p->x;
        if (p->y > extent.height) extent.height = p->y;
    }
    extent.width -= extent.x - 1;
    extent.height -= extent.y - 1;

    if (pixels.empty()) {
        SVL_LOG(SVL_LOG_ERROR, "SceneSegment is empty for region " << n);
    }
}

void SceneSegment::initialize(const CvRect& r)
{
    SVL_ASSERT((r.width > 0) && (r.height > 0) && (r.x >= 0) && (r.y >= 0));
    clear();
    
    pixels.reserve(r.width * r.height);
    for (int y = 0; y < r.height; y++) {
        for (int x = 0; x < r.width; x++) {
            pixels.push_back(cvPoint(r.x + x, r.y + y));
            if ((y == 0) || (y == r.height - 1)) {
                insideBoundaryPixels.insert(cvPoint(r.x + x, r.y + y));
                outsideBoundaryPixels.insert(cvPoint(r.x + x, r.y + y + (y == 0 ? -1 : 1)));
            } else if ((x == 0) || (x == r.width - 1)) {
                insideBoundaryPixels.insert(cvPoint(r.x + x, r.y + y));
                outsideBoundaryPixels.insert(cvPoint(r.x + x + (x == 0 ? -1 : 1), r.y + y));
            }
        }
    }

    extent = r;
}

void SceneSegment::initialize(const svlMultiSeg& multiSegs, const vector<pair<int, int> >& segv)
{
    clear();
    
    for (vector<pair<int, int> >::const_iterator itr = segv.begin();
         itr != segv.end(); ++itr) {
        const CvMat* region = multiSegs[itr->first].segmentation();
        if (size() == 0)
            initialize(region, itr->second);
        else {
            merge(SceneSegment(region, itr->second));
        }
    }
}

void SceneSegment::merge(const SceneSegment& seg)
{
    int width = std::max(extent.x + extent.width, seg.extent.x + seg.extent.width);
    int height = std::max(extent.y + extent.height, seg.extent.y + seg.extent.height);

    for (set<CvPoint>::const_iterator it = outsideBoundaryPixels.begin();
         it != outsideBoundaryPixels.end(); it++) {
        if (it->x >= width) width = it->x + 1;
        if (it->y >= height) height = it->y + 1;
    }
    for (set<CvPoint>::const_iterator it = seg.outsideBoundaryPixels.begin();
         it != seg.outsideBoundaryPixels.end(); it++) {
        if (it->x >= width) width = it->x + 1;
        if (it->y >= height) height = it->y + 1;
    }

    CvMat *m = cvCreateMat(height, width, CV_32SC1);
    cvZero(m);
    for (vector<CvPoint>::const_iterator p = pixels.begin(); p != pixels.end(); ++p) {
        CV_MAT_ELEM(*m, int, p->y, p->x) = 1;
    }
    for (vector<CvPoint>::const_iterator p = seg.pixels.begin(); p != seg.pixels.end(); ++p) {
        CV_MAT_ELEM(*m, int, p->y, p->x) = 1;
    }

    initialize(m, 1);

    cvReleaseMat(&m);
}

bool SceneSegment::isAdjacent(const SceneSegment& seg) const
{
    if (seg.size() < size()) {
        return seg.isAdjacent(*this);
    }

    for (vector<CvPoint>::const_iterator p = pixels.begin(); p != pixels.end(); ++p) {
        if (seg.outsideBoundaryPixels.find(*p) != seg.outsideBoundaryPixels.end()) {
            return true;
        }
    }

    return false;
}

double SceneSegment::areaOfOverlap(const SceneSegment& seg) const
{
    // TODO: check this---slow and memory intensive
    set<CvPoint> pixelUnion;

    pixelUnion.insert(pixels.begin(), pixels.end());
    pixelUnion.insert(seg.pixels.begin(), seg.pixels.end());

    double areaSum = (double)(seg.size() + this->size());
    double areaUnion = (double)pixelUnion.size();

    return (areaSum - areaUnion) / areaSum;
}


// SceneSegmentDictionary ----------------------------------------------------

SceneSegmentDictionary::SceneSegmentDictionary()
{
    // do nothing
}

SceneSegmentDictionary::SceneSegmentDictionary(const svlMultiSeg& multiSeg)
{
    for (int i = 0; i < multiSeg.count(); i++) {
        for (int n = 0; n < multiSeg.numSegments(i); n++) {
            entries.insert(make_pair(0.0, SceneSegment(multiSeg.segmentation(i), n)));
        }
    }
}

SceneSegmentDictionary::SceneSegmentDictionary(const svlImageSegmentation& seg)
{
    for (int n = 0; n < seg.numSegments(); n++) {
        entries.insert(make_pair(0.0, SceneSegment(seg.segmentation(), n)));
    }
}

SceneSegmentDictionary::SceneSegmentDictionary(const CvMat *regions)
{
    SVL_ASSERT(regions != NULL);
    double maxVal;
    cvMinMaxLoc(regions, NULL, &maxVal);

    for (int n = 0; n < (int)maxVal; n++) {
        entries.insert(make_pair(0.0, SceneSegment(regions, n)));
    }
}

SceneSegmentDictionary::~SceneSegmentDictionary()
{
    // do nothing
}

void SceneSegmentDictionary::read(const char *filename)
{
    SVL_ASSERT(filename != NULL);

    XMLNode root = XMLNode::parseFile(filename, "SegmentDictionary");

    for (int i = 0; i < root.nChildNode("Segment"); i++) {
        XMLNode node = root.getChildNode("Segment", i);
        double score = atof(node.getAttribute("score"));
        addSegment(score, SceneSegment(node));
    }
}

void SceneSegmentDictionary::write(const char *filename) const
{
    SVL_ASSERT(filename != NULL);

    ofstream ofs(filename);
    SVL_ASSERT(!ofs.fail());
    
    ofs << "<SegmentDictionary>\n";
    
    for (multimap<double, SceneSegment>::const_iterator it = entries.begin();
         it != entries.end(); ++it) {
        ofs << "  <Segment score=\"" << it->first << "\">\n";
        it->second.write(ofs);
        ofs << "  </Segment>\n";
    }

    ofs << "</SegmentDictionary>\n";
    ofs.close();
}

bool SceneSegmentDictionary::addSegment(double score, const SceneSegment& seg)
{
    if (seg.size() < 10)
        return false;

    typedef multimap<double, SceneSegment> T;
    pair<T::iterator, T::iterator> r = entries.equal_range(score);

    for (T::iterator it = r.first; it != r.second; it++) {
        if ((it->first == score) && (it->second.extent == seg.extent)) {
            //SVL_LOG(SVL_LOG_VERBOSE, "not adding repeated segment");
            return false;
        }
    }

    entries.insert(make_pair(score, seg));
    //SVL_LOG(SVL_LOG_VERBOSE, "...dictionary has " << entries.size() << " entries");

    return true;
}

bool SceneSegmentDictionary::addSegment(SceneInstance& instance, int k)
{
    if (instance.assignment.regions[k]->empty())
        return false;

    double score = scoreDictionarySegment(instance, k);
    return addSegment(score, SceneSegment(instance.assignment.pixels, k));
}

void SceneSegmentDictionary::buildDictionary(SceneInstance& instance)
{
    SVL_ASSERT_MSG(!instance.pixelSingletons.empty(),
        "instance must be initialized by model before passing to dictionary");

    // add all individual segments and merge using HAC
    for (int s = 0; s < instance.segmentations.count(); s++) {
        instance.assignPixels(instance.segmentations.segmentation(s));
        for (int k = 0; k < instance.numRegions(); k++) {
            addSegment(instance, k);
        }

#if 0
        // evaluate all edges
        while (instance.numRegions > 2) {
            double bestEdgeScore = numeric_limits<double>::max();
            pair<int, int> edge;
            for (int k = 0; k < instance.numRegions; k++) {
                if (instance.regionData[k].size > instance.size() / 8)
                    continue;
                for (set<int>::const_iterator n = instance.regionData[k].regionNbrs.begin();
                     n != instance.regionData[k].regionNbrs.end(); ++n) {
                    if (*n < k) continue;
                    if (instance.regionData[*n].size > instance.size() / 8)
                        continue;

                    double score = scoreDictionarySegments(instance, k, *n);
                    SVL_LOG(SVL_LOG_VERBOSE, "score of <" << k << "," << *n << "> is " << score);
                    if (score < bestEdgeScore) {
                        bestEdgeScore = score;
                        edge.first = k;
                        edge.second = *n;
                    }
                }
            }

            if (bestEdgeScore == numeric_limits<double>::max()) {
                break;
            }

            SVL_LOG(SVL_LOG_VERBOSE, "merging region " << edge.second << " with " << edge.first);
            instance.movePixels(edge.second, edge.first);
            instance.remapRegions();

            addSegment(instance, edge.first);

#if 1
            showDictionary();
#endif
        }
#endif
    }
}

void SceneSegmentDictionary::prune(int n)
{
    if ((int)entries.size() <= n) {
        return;
    }

    multimap<double, SceneSegment>::iterator it = entries.begin();
    while (--n > 0) {
        it++;
    }
    entries.erase(it, entries.end());
}

void SceneSegmentDictionary::removeRepeated()
{
    for (multimap<double, SceneSegment>::iterator it = entries.begin();
         it != entries.end(); ++it) {
        double area = it->second.extent.width * it->second.extent.height;
        multimap<double, SceneSegment>::iterator jt = it;
        while (++jt != entries.end()) {
            if (jt->first - it->first > SVL_EPSILON) {
                break;
            }

            CvRect r = svlIntersection(it->second.extent, jt->second.extent);
            if ((r.width > 0) && (r.height > 0) &&
                ((double)(r.width * r.height) / area > 0.95)) {
                SVL_LOG(SVL_LOG_VERBOSE, "erasing dictionary entry");
                entries.erase(jt);
            }
        }
    }
}

void SceneSegmentDictionary::showDictionary(SceneInstance &instance) const
{
    IplImage *canvas = visualize(instance.image());
    cvNamedWindow(DICTIONARY_WINDOW_NAME);
    cvShowImage(DICTIONARY_WINDOW_NAME, canvas);
    cvWaitKey(10);
    cvReleaseImage(&canvas);
}

void SceneSegmentDictionary::hideDictionary() const
{
    cvDestroyWindow(DICTIONARY_WINDOW_NAME);
}

IplImage *SceneSegmentDictionary::visualize(const IplImage *baseImage) const
{
    SVL_ASSERT(baseImage != NULL);

    vector<IplImage *> views(180);
    multimap<double, SceneSegment>::const_iterator it = entries.begin();
    for (unsigned i = 0; i < views.size(); i++) {
        views[i] = cvCreateImage(cvGetSize(baseImage), IPL_DEPTH_8U, 3);
        cvZero(views[i]);

        if (it != entries.end()) {
            cvConvertScale(baseImage, views[i], 0.3);

            CvRect roi = cvRect(baseImage->width, baseImage->height, 0, 0);
            IplImage *mask = cvCreateImage(cvGetSize(baseImage), IPL_DEPTH_8U, 1);
            cvZero(mask);
            for (vector<CvPoint>::const_iterator p = it->second.pixels.begin();
                 p != it->second.pixels.end(); ++p) {
                CV_IMAGE_ELEM(mask, unsigned char, p->y, p->x) = 1;
                if (p->x < roi.x) roi.x = p->x;
                if (p->y < roi.y) roi.y = p->y;
                if (p->x > roi.width) roi.width = p->x;
                if (p->y > roi.height) roi.height = p->y;
            }
            cvCopy(baseImage, views[i], mask);
            cvReleaseImage(&mask);

            cvRectangle(views[i], cvPoint(0, 0),
                cvPoint(baseImage->width - 1, baseImage->height - 1), CV_RGB(255, 0, 0));

            for (set<CvPoint>::const_iterator p = it->second.outsideBoundaryPixels.begin();
                 p != it->second.outsideBoundaryPixels.end(); ++p) {
                CV_IMAGE_ELEM(views[i], unsigned char, p->y, 3 * p->x + 2) = 0xff;
                if (p->x < roi.x) roi.x = p->x;
                if (p->y < roi.y) roi.y = p->y;
                if (p->x > roi.width) roi.width = p->x;
                if (p->y > roi.height) roi.height = p->y;
            }

#if 0
            roi.width -= roi.x - 1;
            roi.height -= roi.y - 1;
            if (roi.height * baseImage->width > roi.width * baseImage->height) {
                int delta = baseImage->width * roi.height / baseImage->height - roi.width;
                roi.width += delta;
                roi.x -= delta / 2;
                if (roi.x < 0) { roi.x = 0; }
            } else {
                int delta = baseImage->height * roi.width / baseImage->width - roi.height;
                roi.height += delta;
                roi.y -= delta / 2;
                if (roi.y < 0) { roi.y = 0; }
            }
            cropInPlace(&views[i], roi);
            resizeInPlace(&views[i], baseImage->height, baseImage->width);
#endif
            ++it;
        }

        resizeInPlace(&views[i], 96, 96 * baseImage->width / baseImage->height);
    }

    IplImage *canvas = combineImages(views, 10, 18);
    //resizeInPlace(&canvas, 768, 768 * canvas->width / canvas->height);
    svlReleaseImages(views);

    return canvas;
}

void SceneSegmentDictionary::print(ostream& os) const
{
    for (multimap<double, SceneSegment>::const_iterator it = entries.begin();
         it != entries.end(); ++it) {
        os << "\t" << it->first << "\t: " << it->second.size()
           << "\t" << toString(it->second.extent) << "\n";
    }
}

// protected functions ------------------------------------------------------------

double SceneSegmentDictionary::scoreDictionarySegment(SceneInstance& instance, int k)
{
    return -svlGaussian(instance.assignment.regions[k]->Ak).logPartitionFunction();
}

double SceneSegmentDictionary::scoreDictionarySegments(SceneInstance& instance, int k, int l)
{
    svlSufficientStats stats(instance.assignment.regions[k]->Ak);
    stats.accumulate(instance.assignment.regions[l]->Ak);
    svlGaussian g(stats.size());
    g.train(stats, 1.0e-3);

    return -g.logPartitionFunction();
}
