/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneSegmentDict.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Dictionary of proposal segmentations.
**
*****************************************************************************/

#pragma once

#include <cstdlib>
#include <cassert>
#include <map>
#include <set>
#include <vector>

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"

using namespace std;

// SceneSegment --------------------------------------------------------------

class SceneSegment {
 public:
    set<CvPoint> outsideBoundaryPixels;
    set<CvPoint> insideBoundaryPixels;
    vector<CvPoint> pixels;
    CvRect extent;

 public:
    SceneSegment();
    SceneSegment(const CvMat *regions, int n);
    SceneSegment(const CvRect& r);
    SceneSegment(const svlMultiSeg& multiSegs, const vector<pair<int, int> >& segv);
    SceneSegment(XMLNode& node);
    ~SceneSegment();

    inline int size() const { return (int)pixels.size(); }

    void read(XMLNode &node);
    void write(ostream& os) const;

    void clear();
    void initialize(const CvMat *regions, int n);
    void initialize(const CvRect& r);
    void initialize(const svlMultiSeg& multiSegs, const vector<pair<int, int> >& segv);
    void merge(const SceneSegment& seg);

    // overlapping or touching (not fully contained in)
    bool isAdjacent(const SceneSegment& seg) const;

    // area of overlap
    double areaOfOverlap(const SceneSegment& seg) const;
};

// SceneSegmentDictionary ----------------------------------------------------

class SceneSegmentDictionary {
 public:
    multimap<double, SceneSegment> entries;
    
 public:
    SceneSegmentDictionary();
    SceneSegmentDictionary(const svlMultiSeg& multiSeg);
    SceneSegmentDictionary(const svlImageSegmentation& seg);
    SceneSegmentDictionary(const CvMat *regions);
    ~SceneSegmentDictionary();

    inline int size() const { return (int)entries.size(); }    

    void read(const char *filename);
    void write(const char *filename) const;

    bool addSegment(double score, const SceneSegment& seg);
    bool addSegment(SceneInstance& instance, int k);
    void buildDictionary(SceneInstance& instance);

    void prune(int n);
    void removeRepeated();

    void showDictionary(SceneInstance& instance) const;
    void hideDictionary() const;
    IplImage *visualize(const IplImage *baseImage) const;
    void print(ostream& os) const;

 protected:
    double scoreDictionarySegment(SceneInstance& instance, int k);
    double scoreDictionarySegments(SceneInstance& instance, int k, int l);
};
