/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneSlidingWindowDetector.cpp
** AUTHOR(S):   Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**
*****************************************************************************/

#include "cv.h"
#include "cxcore.h"
#include "ml.h"

#include "sceneSlidingWindowDetector.h"

/* --------------- linear classifier ------------------- */
SceneLinearClassifier::SceneLinearClassifier(const char* modelFilename) : 
    _length(0), _linearBias(0)
{
    load(modelFilename);
}

SceneLinearClassifier::SceneLinearClassifier(const SceneLinearClassifier& classifier) :
    _length(classifier._length), _linearBias(classifier._linearBias)
{
    _linearWeights.resize(classifier._linearWeights.size(), 0);
    copy(classifier._linearWeights.begin(), classifier._linearWeights.end(), _linearWeights.begin());
}

SceneLinearClassifier& SceneLinearClassifier::operator=(const SceneLinearClassifier& classifier)
{
    if (this != &classifier) {
        _length = classifier._length;
        _linearBias = classifier._linearBias;
        _linearWeights.resize(classifier._linearWeights.size(), 0);
        copy(classifier._linearWeights.begin(), classifier._linearWeights.end(), _linearWeights.begin());
    }
    return *this;
}

void SceneLinearClassifier::load(const char* modelFilename)
{
    SVL_ASSERT(modelFilename != NULL);
    SVL_LOG(SVL_LOG_MESSAGE, "Reading model file: " << modelFilename);

    FILE *modelfl;
    modelfl = fopen (modelFilename, "rb");
    SVL_ASSERT_MSG(modelfl != NULL, "cannot open pedestrian classifier " << modelFilename);

    char version_buffer[10];
    if (!fread (&version_buffer,sizeof(char),10,modelfl)) {
        SVL_ASSERT_MSG(0, "unable to read version");
    }

    //if(strcmp(version_buffer,"V6.01")) {
    //    throw Exception("LinearClassify::LinearClassify", "Version of model-file does not match version of svm_classify!"); 
    //}

    /* read version number */
    int version = 0;
    if (!fread (&version,sizeof(int),1,modelfl)) { 
        SVL_ASSERT_MSG(0, "Unable to read version number"); 
    }
    if (version < 200) {
        SVL_ASSERT_MSG(0, "Does not support model file compiled for light version"); 
    }

    long kernel_type;
    fread(&(kernel_type),sizeof(long),1,modelfl);

    {// ignore these
        long poly_degree;
        fread(&(poly_degree),sizeof(long),1,modelfl);

        double rbf_gamma;
        fread(&(rbf_gamma),sizeof(double),1,modelfl);

        double  coef_lin;
        fread(&(coef_lin),sizeof(double),1,modelfl); 
        double coef_const;
        fread(&(coef_const),sizeof(double),1,modelfl);

        long l;
        fread(&l,sizeof(long),1,modelfl);
        char* custom = new char[l];
        fread(custom,sizeof(char),l,modelfl);
        delete[] custom;
    }

    long totwords;
    fread(&(totwords),sizeof(long),1,modelfl);

    {// ignore these
        long totdoc;
        fread(&(totdoc),sizeof(long),1,modelfl);

        long sv_num;
        fread(&(sv_num), sizeof(long),1,modelfl);
    }

    fread(&_linearBias, sizeof(double),1,modelfl);

    
    if(kernel_type == 0) { /* linear kernel */
        /* save linear wts also */
        double* linearwt = new double[totwords+1];
        _length = totwords;
        fread(linearwt, sizeof(double),totwords+1,modelfl);
        _linearWeights.clear();
        for (int i = 0; i < _length; i++) {
            _linearWeights.push_back(linearwt[i]);
        }
    } else {
        SVL_ASSERT_MSG(0, "only support linear kernel!");
    }

    fclose(modelfl);
}


float SceneLinearClassifier::getScore(const vector<double>& vec) const
{
    SVL_ASSERT_MSG(vec.size() == _length, "feature vector size: " << vec.size() << "; classifier weight length: " << _linearWeights.size());
    double sum = 0;
    int i = 0;
    
    for (i = 0; i < vec.size(); i+=2) {
        sum += vec[i]*_linearWeights[i] + vec[i+1]*_linearWeights[i+1];
    }

    if (vec.size()%2 == 1) {
        sum += vec[i-1] * _linearWeights[i-1];
    }
    return (float)(sum - _linearBias);
}


float SceneLinearClassifier::getScore(const VectorXd& vec) const
{
    SVL_ASSERT_MSG(vec.rows() == _length, "feature vector size: " << vec.rows() << "; classifier weight length: " << _linearWeights.size());
    double sum = 0;
    int i = 0;
    
    for (i = 0; i < vec.rows(); i+=2) {
        sum += vec[i]*_linearWeights[i] + vec[i+1]*_linearWeights[i+1];
    }

    if (vec.rows()%2 == 1) {
        sum += vec[i-1] * _linearWeights[i-1];
    }
    return (float)(sum - _linearBias);
}

double SceneLinearClassifier::scoreToProbability(float score) const
{
    return 1 / (1 + exp(-score));
}

/* --------------- sliding window detector -------------------------- */

SceneSlidingWindowDetector::SceneSlidingWindowDetector(const char *name, int w, int h) :
    svlSlidingWindowDetector(name, w, h)
{
    // do nothing
}

SceneSlidingWindowDetector::~SceneSlidingWindowDetector()
{
    // do nothing

}

bool SceneSlidingWindowDetector::load(const char* modelFilename, SceneHOGFeatures* extractorHOG)
{
	if (!extractorHOG) return false;

        _extractorHOG = extractorHOG;
	_windowWidth = _extractorHOG->getWinWidth(); 
        _windowHeight = _extractorHOG->getWinHeight();
        
	for ( vector<svlWindowDiscarder*>::iterator it=_discarders.begin(); it!=_discarders.end(); ++it )
		(*it)->setSize(_windowWidth, _windowHeight);

	if ( modelFilename ) {
            //svlBinaryClassifier *c = (svlBinaryClassifier *) svlClassifierFactory().load(modelFilename);
            SceneLinearClassifier* c = new SceneLinearClassifier(modelFilename);
            if (!c) return false;
            if (_owns_classifier && _svmLinearClassifier)
                delete _svmLinearClassifier;
            _svmLinearClassifier = c;
            _owns_classifier = true;
	}
	return true;
}

bool SceneSlidingWindowDetector::load(const char* modelFilename, const char* extractorFilename)
{
        if (extractorFilename != NULL) {
            _extractorHOG = new SceneHOGFeatures();
            bool bSucess = _extractorHOG->loadExtractor(extractorFilename);
            if (bSucess == false) return false;
            _owns_extractor = true;
        }

	if ( modelFilename ) {
            //svlBinaryClassifier *c = (svlBinaryClassifier *) svlClassifierFactory().load(modelFilename);
            SceneLinearClassifier* c = new SceneLinearClassifier(modelFilename);
            if (!c) return false;
            if (_owns_classifier && _svmLinearClassifier)
                delete _svmLinearClassifier;
            _svmLinearClassifier = c;
            _owns_classifier = true;
	}
	return true;
}

/* -----------------------  version of image pyramid ----------------------*/
void SceneSlidingWindowDetector::classifyImage(const vector<IplImage*> &images, 
                                               svlObject2dFrame *objects,
                                               svlFeatureVectors *vecs)
{
    if (images.empty() == true) {
        SVL_LOG(SVL_LOG_ERROR, "images are empty!");
        return;
    }
    static int h = svlCodeProfiler::getHandle("classifyImage::getScore");
    svlCodeProfiler::tic(h);                

    SVL_ASSERT(images[0] != NULL);

    SVL_ASSERT_MSG(objects || vecs, "classifyImage not asked to return anything");
    SVL_ASSERT_MSG(_extractorHOG, "Need to provide an extractor first");
    //    SVL_ASSERT_MSG(!objects || _classifier, "Need to provide a classifier first");
    SVL_ASSERT_MSG(!objects || _svmLinearClassifier, "Need to provide a classifier first");
    
    SVL_ASSERT(!images.empty());

    // each image in the vector must not be NULL and all images must be the same size
    for ( unsigned i = 0; i < images.size(); i++ ) {
        SVL_ASSERT(images[i] != NULL);
        SVL_ASSERT(images[i]->width == images[0]->width);
        SVL_ASSERT(images[i]->height == images[0]->height);
    }
    
    double scale = 1.0;
    //float scoreThreshold = _classifier->probabilityToScore((float)_threshold);
    // prob. = 1 / (1+exp(-score)) <--> score = -ln(1/p - 1) 
    float scoreThreshold = -logf(1/(float)_threshold - 1);
    int numChannels = (int)images.size();
    
    SVL_LOG(SVL_LOG_MESSAGE, "Running classifier on " << numChannels << "-channel image");
    
    // copy of original images for color conversion and rescaling
    //vector <IplImage *> originalImages(numChannels);
    vector <IplImage *> scaledImages(numChannels);

    for (int i = 0; i < numChannels; i++) {
        scaledImages[i] = cvCloneImage(images[i]);
    }
    
    // Iterate over all scales
    while ((scaledImages[0]->width >= _windowWidth) && (scaledImages[0]->height >= _windowHeight)) {
        // iterate over all locations
        vector<CvPoint> locations;
        createWindowLocations(scaledImages[0]->width, scaledImages[0]->height, locations);
        svlSlidingWindowDetector::discardUninteresting(scaledImages, scale, locations);
        
        SVL_LOG(SVL_LOG_VERBOSE, "Computing features over " << locations.size() << " windows for image size "
            << scaledImages[0]->width << " x " << scaledImages[0]->height << "...");

        SVL_LOG(SVL_LOG_VERBOSE, "Window size: " << _windowWidth << "x" << _windowHeight << "...");
        
        vector<vector<double> > features;
        features.reserve(locations.size());
        _extractorHOG->extract(locations, scaledImages, features, false);
        
        SVL_LOG(SVL_LOG_VERBOSE, "...done");
        
        if ( vecs ) {
            vecs->push(scaledImages[0], locations, features);
        }
        if ( objects ) {
            SVL_LOG(SVL_LOG_VERBOSE, "Running multi-channel classifier on " << locations.size() << " windows...");
            for (unsigned i = 0; i < locations.size(); i++) {
                //if(containsInvalidEntries(features[i])) continue;
                //float score = _classifier->getScore(features[i]);
                float score = _svmLinearClassifier->getScore(features[i]);
                if (score > scoreThreshold) {
                    objects->push_back(svlObject2d(scale * locations[i].x, scale * locations[i].y,
                            scale * _windowWidth, scale * _windowHeight));
                    objects->back().name = _objectName;
                    //objects->back().pr = _classifier->scoreToProbability(score);
                    objects->back().pr = 1 / (1+exp(-score));
                }
            }
            SVL_LOG(SVL_LOG_VERBOSE, "...done");
        }
        
        // scale down each image channel
        scale *= _delta_scale;
        
        for (int i = 0; i < numChannels; i++) {
            cvReleaseImage(&(scaledImages[i]));
            scaledImages[i] = cvCreateImage(cvSize((int)(images[i]->width / scale),
                    (int)(images[i]->height / scale)), images[i]->depth, images[i]->nChannels);
            cvResize(images[i], scaledImages[i]);
        }
    }
    
    // free memory
    for (int i = 0; i < numChannels ; i++) {
        cvReleaseImage(&(scaledImages[i]));
    }
    
    svlCodeProfiler::tic(h);
}


/* ------------------------- version of chaning sliding window size ------------------------- */
/*
void SceneSlidingWindowDetector::classifyImage(const vector<IplImage*> &images, 
                                               svlObject2dFrame *objects,
                                               svlFeatureVectors *vecs)
{
    if (images.empty() == true) {
        SVL_LOG(SVL_LOG_ERROR, "images are empty!");
        return;
    }
    SVL_ASSERT(images[0] != NULL);

    svlObject2dFrame windows;
    getAllWindows(images[0]->width, images[0]->height, windows);
    classifyImage(images, windows, objects, vecs);
}
*/

void SceneSlidingWindowDetector::classifyImage(const vector<IplImage*>& images, 
                                               const svlObject2dFrame& windows,
                                               svlObject2dFrame* objects,
                                               svlFeatureVectors* vecs)
{
	SVL_ASSERT_MSG(objects || vecs, "classifyImage not asked to return anything");
	SVL_ASSERT_MSG(_extractorHOG, "Need to provide an extractor first");
        //	SVL_ASSERT_MSG(!objects || _classifier, "Need to provide a classifier first");
        SVL_ASSERT_MSG(!objects || _svmLinearClassifier, "Need to provide a classifier first");

	SVL_ASSERT(!images.empty());
	// each image in the vector must not be NULL and all images must be the same size
	for ( unsigned i = 0; i < images.size(); i++ ) {
		SVL_ASSERT(images[i] != NULL);
		SVL_ASSERT(images[i]->width == images[0]->width);
		SVL_ASSERT(images[i]->height == images[0]->height);
	}

	double scale = 1.0;
	float scoreThreshold = _classifier->probabilityToScore((float)_threshold);

        SVL_ASSERT(_extractorHOG != NULL);
        // only run detector on images[0] which is assumed to be the grayscale image
        _extractorHOG->load(images[0]);
        _extractorHOG->prePocessing();
        
        // filter the windows
        svlObject2dFrame windowsFiltered = windows;
        discardInvalid(images[0]->width, images[0]->height, windowsFiltered);
        discardUninteresting(images[0], windowsFiltered);
        
        // get the feature for each window
        vector<vector<double> > features;
        features.reserve(windowsFiltered.size());
        SVL_LOG(SVL_LOG_VERBOSE, "Computing features over " << windowsFiltered.size() << " windows for image size "
                << images[0]->width << " x " << images[0]->height << "...");        
        _extractorHOG->extractFeatures(windowsFiltered, features);

        // TODO: implement output of the features
        SVL_LOG(SVL_LOG_VERBOSE, "...done");

        // run classifier
        if (objects) {
            
            SVL_LOG(SVL_LOG_VERBOSE, "Running multi-channel classifier on " << windowsFiltered.size() << " windows...");                


//             static int h = svlCodeProfiler::getHandle("classifyImage::getScore");
//             svlCodeProfiler::tic(h);                
//             vector<double> scores;
//             (svlSVM *)_classifier->getScores(features, scores);

//             For (int i = 0; i < (int)windowsFiltered.size(); i++) {
//                 // if(containsInvalidEntries(features[i])) continue;

//                 float score = scores[i];

//                 SVL_LOG(SVL_LOG_DEBUG, "Classify on window: " << toString(windowsFiltered[i].getRect()) 
//                     << " score: " << _classifier->scoreToProbability(score)
//                     << " feature: " << features[i][0]);
                
//                 if (score > scoreThreshold) {
//                     objects->push_back(windowsFiltered[i]);
//                     objects->back().name = _objectName;
//                     objects->back().pr = _classifier->scoreToProbability(score);
//                 }
//             }
//             svlCodeProfiler::toc(h);


            for (int i = 0; i < (int)windowsFiltered.size(); i++) {
                // if(containsInvalidEntries(features[i])) continue;

                static int h = svlCodeProfiler::getHandle("classifyImage::getScore");
                svlCodeProfiler::tic(h);                

                float score = _classifier->getScore(features[i]);

                svlCodeProfiler::toc(h);                

                SVL_LOG(SVL_LOG_DEBUG, "Classify on window: " << toString(windowsFiltered[i].getRect()) 
                    << " score: " << _classifier->scoreToProbability(score)
                    << " feature: " << features[i][0]);
                
                if (score > scoreThreshold) {
                    objects->push_back(windowsFiltered[i]);
                    objects->back().name = _objectName;
                    objects->back().pr = _classifier->scoreToProbability(score);
                }
            }

            SVL_LOG(SVL_LOG_VERBOSE, "...done");
        }
}


void SceneSlidingWindowDetector::classifySubImage(const vector<IplImage *>& images,
				                  CvRect region,
				                  svlObject2dFrame *objects,
				                  svlFeatureVectors *vecs)
{
    	SVL_ASSERT_MSG(objects || vecs, "classifyImage not asked to return anything");
	SVL_ASSERT_MSG(_extractor, "Need to provide an extractor first");
	//SVL_ASSERT_MSG(!objects || _classifier, "Need to provide a classifier first");
        SVL_ASSERT_MSG(!objects || _svmLinearClassifier, "Need to provide a classifier first");

	SVL_ASSERT(!images.empty());
	// each image in the vector must not be NULL and all images must be the same size
	for (unsigned i = 0; i < images.size(); i++) {
		SVL_ASSERT(images[i] != NULL);
		SVL_ASSERT(images[i]->width == images[0]->width);
		SVL_ASSERT(images[i]->height == images[0]->height);
	}

	// if entire image, run full detector
	if ((region.x == 0) && (region.width == images[0]->width) &&
		(region.y == 0) && (region.height == images[0]->height)) {
		classifyImage(images, objects, vecs);
		return;
	}

	int numChannels = (int)images.size();

	// create subimages
	vector <IplImage *> subImages(numChannels);
	svlClipRect(region, images[0]);
	for (int i = 0; i < numChannels; i++) {
		IplImage *imgCopy = cvCloneImage(images[i]);

		subImages[i] = cvCreateImage(cvSize(region.width, region.height),
			images[i]->depth, images[i]->nChannels);

		SVL_ASSERT(subImages[i] != NULL);

		cvSetImageROI(imgCopy, region);
		cvCopyImage(imgCopy, subImages[i]);
		cvReleaseImage(&imgCopy);
	}

	// run full detector on sub-images
	classifyImage(subImages, objects, vecs);

	for (unsigned i = 0; i < objects->size(); i++) {
		(*objects)[i].x += region.x;
		(*objects)[i].y += region.y;
	}

	// free memory
	for (int i = 0; i < numChannels; i++) {
		cvReleaseImage(&subImages[i]);
	}

}


float SceneSlidingWindowDetector::evalModel(const IplImage* image) 
{
    SVL_ASSERT(image != NULL);

    SVL_ASSERT_MSG(_extractorHOG, "Need to provide an extractor first");
    SVL_ASSERT_MSG(_svmLinearClassifier, "Need to provide a classifier first");

    SVL_ASSERT_MSG(image->width == _extractorHOG->getWinWidth() && image->height == _extractorHOG->getWinHeight(),
        "image width: " << image->width << ", height: " << image->height
        << "; extractor winWidth: " << _extractorHOG->getWinWidth() << ", winHeight: " << _extractorHOG->getWinHeight());

    static int h = svlCodeProfiler::getHandle("SceneSlidingWindowDetector::evalModel");
    svlCodeProfiler::tic(h); 
    // extract the feature vector
    _extractorHOG->load(image);
    _extractorHOG->prePocessing();
    CvRect r = cvRect(0, 0, image->width, image->height);
    VectorXd f = _extractorHOG->extractFeatures(r);
    vector<double> feature(f.rows(), 0);
    Eigen::Map<VectorXd>(&feature[0], feature.size()) = f;
    
    // get the classification score
    float score = _svmLinearClassifier->getScore(feature);

    svlCodeProfiler::tic(h);
    return score;
}

void SceneSlidingWindowDetector::discardInvalid(int imageWidth, int imageHeight, svlObject2dFrame& windows)
{
    int origWinSize = windows.size();
    for (int i = (int)windows.size() - 1; i >= 0; i--) {
        CvRect r = windows[i].getRect();
        if (r.x < 0 || r.y < 0 || r.x + r.width - 1 > imageWidth - 1 || r.y + r.height - 1> imageHeight - 1) {
            windows.erase(windows.begin() + i);
        }
    }
    SVL_LOG(SVL_LOG_VERBOSE, origWinSize - windows.size() << " out of " << origWinSize << " windows are invalid.");
}

void SceneSlidingWindowDetector::discardUninteresting(IplImage* image, svlObject2dFrame& windows)
{
    if (image == NULL) return;
    int origWinSize = windows.size();
    for (int i = (int)windows.size() - 1; i >= 0; i--)  {
        cvSetImageROI(image, windows[i].getRect());
        
        if (svlImageAlmostUniform(image)) {
            windows.erase(windows.begin() + i);
        }
        cvResetImageROI(image);
    }
    SVL_LOG(SVL_LOG_VERBOSE, origWinSize - windows.size() << " out of " << origWinSize << " windows are uninteresting.");
}
