/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneSlidingWindowDetector.h
** AUTHOR(S):   Tianshi Gao <tianshig@stanford.edu>
** DESCRIPTION:
**  Defines a multi-channel interface for sliding-window object detectors
**  using HOG features. 
**  Define a linear classifier based on svmdense. 
**
*****************************************************************************/

#pragma once

#include "cv.h"
#include "cxcore.h"
#include "ml.h"

#include "Eigen/Core"

#include "svlBase.h"
#include "svlVision.h"
#include "svlML.h"
#include "sceneHOGFeatures.h"

using namespace std;

class SceneLinearClassifier {
 public:
    SceneLinearClassifier() : _length(0), _linearBias(0) {};
    ~SceneLinearClassifier() {/* do nothing */}
    
    SceneLinearClassifier(const char* modelFilename);
    SceneLinearClassifier(const SceneLinearClassifier& classifier);
    SceneLinearClassifier& operator=(const SceneLinearClassifier& classifier);
    
    // load model
    void load(const char* modelFilename);

    int length() const {return _length;}
    float getScore(const vector<double>& vec) const;
    float getScore(const VectorXd& vec) const;
    double scoreToProbability(float score) const;

 private:
    int _length;
    vector<double> _linearWeights;
    double _linearBias;
};


class SceneSlidingWindowDetector : public svlSlidingWindowDetector {
 public:
    SceneHOGFeatures* _extractorHOG;
    SceneLinearClassifier* _svmLinearClassifier;
 public:
    SceneSlidingWindowDetector(const char *name = NULL, int w = 32, int h = 32);
    virtual ~SceneSlidingWindowDetector();

    bool load(const char* modelFilename, SceneHOGFeatures* extractorHOG);
    bool load(const char* modelFilename, const char* extractorFilename);

    // Classifies a multi-channel image, optionally returning objects or features.
    // objects can be NULL, in which case classification is not performed.
    virtual void classifyImage(const vector<IplImage*> &images, 
			       svlObject2dFrame *objects,
			       svlFeatureVectors *vecs = NULL);

    virtual void classifyImage(const vector<IplImage*> &images, 
                               const svlObject2dFrame &windows,
                               svlObject2dFrame *objects,
			       svlFeatureVectors *vecs = NULL);

    // Evaluate classifier on given region (and subwindows)
    virtual void classifySubImage(const vector<IplImage *>& images,
				  CvRect region,
				  svlObject2dFrame *objects,
				  svlFeatureVectors *vecs = NULL);
    
    virtual float evalModel(const IplImage* image);

    void discardInvalid(int imageWidth, int imageHeight, svlObject2dFrame& windows);
    void discardUninteresting(IplImage* image, svlObject2dFrame& windows);
};
