/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneTestMoves.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for testing pixel and region moves.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneModel.h"
#include "sceneRegionModel.h"
#include "sceneObjectModel.h"
#include "sceneVisualization.h"

#define WINDOW_NAME "sceneTestMoves"

using namespace std;

// prototypes ---------------------------------------------------------------

void visualize(SceneInstance &scene);

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneTestMoves [OPTIONS]\n";
    cerr << "OPTIONS:\n"
         << "  -x                :: visualize\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    // read commandline parameters
    bool bVisualize = false;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_BOOL_OPTION("-x", bVisualize)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 0) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // allocate test scene
    SVL_LOG(SVL_LOG_MESSAGE, "Creating scene...");
    IplImage *testImage = cvCreateImage(cvSize(320, 240), IPL_DEPTH_8U, 3);
    cvZero(testImage);
    for (int y = 0; y < testImage->height; y++) {
        for (int x = 0; x < testImage->width; x++) {
            CV_IMAGE_ELEM(testImage, unsigned char, y, 3 * x) = (x + y) % 256;
            CV_IMAGE_ELEM(testImage, unsigned char, y, 3 * x + 1) = x % 256;
            CV_IMAGE_ELEM(testImage, unsigned char, y, 3 * x + 2) = y % 256;
        }
    }

    SceneInstance scene(testImage, "testScene");
    scene.printScene();
    
    // create four regions
    CvMat *testSeg = cvCreateMat(240, 320, CV_32SC1);
    for (int y = 0; y < testSeg->rows; y++) {
        for (int x = 0; x < testSeg->cols; x++) {
            CV_MAT_ELEM(*testSeg, int, y, x) = (y < 120) ?
                ((x < 80) ? 0 : 1) : ((x < 240) ? 2 : 3);
        }
    }

    SVL_LOG(SVL_LOG_MESSAGE, "Assigning pixels...");
    scene.assignPixels(testSeg);
    scene.printScene();

    if (bVisualize) {
        visualize(scene);
    }

    // try merge moves    
    SVL_LOG(SVL_LOG_MESSAGE, "Testing merge moves...");
    SceneAssignment assignment = scene.assignment;
    for (vector<SceneRegionBoundary *>::const_iterator it = assignment.regionBoundaries.begin();
         it != assignment.regionBoundaries.end(); it++) {
        SVL_LOG(SVL_LOG_VERBOSE, "...merging " << (*it)->regionA->id() << " with " <<
            (*it)->regionB->id());
        scene.movePixels((*it)->regionA->id(), (*it)->regionB->id());
        scene.assignment.remap();
        scene.printScene();

        if (bVisualize) {
            visualize(scene);
        }
        scene.assignment = assignment;
        
        SVL_LOG(SVL_LOG_VERBOSE, "...merging " << (*it)->regionB->id() << " with " <<
            (*it)->regionA->id());
        scene.movePixels((*it)->regionB->id(), (*it)->regionA->id());
        scene.assignment.remap();
        scene.printScene();

        if (bVisualize) {
            visualize(scene);
        }
        scene.assignment = assignment;
    }

    // try segment move
    SVL_LOG(SVL_LOG_MESSAGE, "Testing re-segment moves...");
    vector<CvPoint> segment;
    for (int y = 60; y < 180; y++) {
        for (int x = 80; x < 240; x++) {
            segment.push_back(cvPoint(x, y));
        }
    }

    scene.movePixels(segment, -1);
    scene.assignment.remap();
    scene.printScene();

    if (bVisualize) {
        visualize(scene);
    }
    scene.assignment = assignment;

    // allocate an object
    SVL_LOG(SVL_LOG_MESSAGE, "Allocating objects...");
    scene.moveRegions(0, -1);
    scene.assignment.remap();
    scene.printScene();

    scene.moveRegions(2, -1);
    scene.assignment.remap();
    scene.printScene();

    assignment = scene.assignment;
    
    // merge regions
    SVL_LOG(SVL_LOG_MESSAGE, "Merging regions...");
    scene.movePixels(0, 1);
    scene.assignment.remap();
    scene.printScene();    

    scene.assignment = assignment;

    SVL_LOG(SVL_LOG_MESSAGE, "Merging regions...");
    scene.movePixels(1, 0);
    scene.assignment.remap();
    scene.printScene();    

    scene.assignment = assignment;

    SVL_LOG(SVL_LOG_MESSAGE, "Adding region to object...");
    scene.moveRegions(1, 0);
    scene.assignment.remap();
    scene.printScene();    

    scene.assignment = assignment;

    // merge objects
    SVL_LOG(SVL_LOG_MESSAGE, "Merging regions with objects...");
    scene.movePixels(2, 0);
    scene.assignment.remap();
    scene.printScene();    

    scene.assignment = assignment;

    SVL_LOG(SVL_LOG_MESSAGE, "Moving regions with objects...");
    scene.moveRegions(2, 0);
    scene.assignment.remap();
    scene.printScene();    

    scene.assignment = assignment;

    // free memory
    if (bVisualize) {
        cvDestroyAllWindows();
    }

    cvReleaseMat(&testSeg);
    cvReleaseImage(&testImage);

    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);

    return 0;
}

// helper functions ------------------------------------------------------------------

void visualize(SceneInstance &scene)
{
    IplImage *canvas = cvCloneImage(scene.image());
    
    visualizeRegions(&scene, canvas);
    visualizeBoundaries(&scene, canvas);
    cvNamedWindow(WINDOW_NAME, 1);
    cvShowImage(WINDOW_NAME, canvas);
    cvWaitKey(-1);
    
    cvReleaseImage(&canvas);
}
