/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneViewCachedObjectMoves.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Code for viewing caching object detection moves.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include "svlBase.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneVisualization.h"
#include "sceneSegmentDict.h"
#include "sceneObjDetDict.h"

#define WINDOW_NAME "sceneCachedObjectMoves"

using namespace std;

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./sceneViewCachedObjectMoves [OPTIONS] <filename>\n";
    cerr << "OPTIONS:\n"
         << "  -i <filename>    :: image filename\n"
         << "  -o <filename>    :: output image filename\n"
         << "  -s <w> <h>       :: rescale moves\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    captureOpenCVerrors();

    const char *imgFilename = NULL;
    const char *outImgFilename = NULL;
    int rescaleWidth = 160;
    int rescaleHeight = 120;

    SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
        SVL_CMDLINE_STR_OPTION("-i", imgFilename)
        SVL_CMDLINE_STR_OPTION("-o", outImgFilename)
        SVL_CMDLINE_OPTION_BEGIN("-s", p)
            rescaleWidth = atoi(p[0]);
            rescaleHeight = atoi(p[1]);
        SVL_CMDLINE_OPTION_END(2)
    SVL_END_CMDLINE_PROCESSING(usage());

    if (SVL_CMDLINE_ARGC != 1) {
        usage();
        return -1;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));

    // load image
    IplImage *scene = NULL;
    if (imgFilename != NULL) {
        scene = cvLoadImage(imgFilename, CV_LOAD_IMAGE_COLOR);
        SVL_ASSERT(scene != NULL);
    }

    // load dictionary
    const char *filename = SVL_CMDLINE_ARGV[0];

    SceneObjDetDict moveDict;
    moveDict.load(filename);
    SVL_LOG(SVL_LOG_VERBOSE, "..." << moveDict.numMoves() << " object candidate moves loaded");

    vector<IplImage *> views;
    cvNamedWindow(WINDOW_NAME, 1);
    for (int i = 0; i < moveDict.numMoves(); i++) {
        CvMat *m = moveDict.mask(i);
        bool bConnected = svlIsConnectedComponent(m, true, 1);

        IplImage *greyImg = cvCreateImage(cvSize(m->cols, m->rows), IPL_DEPTH_8U, 1);
        cvConvertScale(m, greyImg, 255, 0);
        IplImage *img = colorImage(greyImg);
        cvReleaseImage(&greyImg);
        cvReleaseMat(&m);

        if (scene != NULL) {
            cvAddWeighted(scene, 0.5, img, 0.5, 0.0, img);
        }

        cvRectangle(img, cvPoint(0, 0), cvPoint(img->width - 1, img->height - 1), CV_RGB(0, 0, 255), 2);

        cvRectangle(img, cvPoint(moveDict[i].originalExtent.x, moveDict[i].originalExtent.y),
            cvPoint(moveDict[i].originalExtent.x + moveDict[i].originalExtent.width,
                moveDict[i].originalExtent.y + moveDict[i].originalExtent.height), CV_RGB(255, 0, 0), 2);

        cvRectangle(img, cvPoint(moveDict[i].extent.x, moveDict[i].extent.y),
            cvPoint(moveDict[i].extent.x + moveDict[i].extent.width,
                moveDict[i].extent.y + moveDict[i].extent.height), CV_RGB(0, 255, 0), 2);
        if (!bConnected) {
            cvLine(img, cvPoint(moveDict[i].extent.x, moveDict[i].extent.y),
                cvPoint(moveDict[i].extent.x + moveDict[i].extent.width,
                    moveDict[i].extent.y + moveDict[i].extent.height), CV_RGB(0, 255, 0), 2);
            cvLine(img, cvPoint(moveDict[i].extent.x, moveDict[i].extent.y + moveDict[i].extent.height),
                cvPoint(moveDict[i].extent.x + moveDict[i].extent.width, moveDict[i].extent.y), CV_RGB(0, 255, 0), 2);
        }

        resizeInPlace(&img, rescaleHeight, rescaleWidth);
        views.push_back(img);
    }

    // show image
    IplImage *canvas = combineImages(views);
    cvShowImage(WINDOW_NAME, canvas);
    cvWaitKey(outImgFilename == NULL ? -1 : 100);
    
    // save image
    if (outImgFilename != NULL) {
        cvSaveImage(outImgFilename, canvas);
    }
    cvReleaseImage(&canvas);    

    // free memory    
    cvDestroyAllWindows();
    svlReleaseImages(views);
    if (scene != NULL) {
        cvReleaseImage(&scene);
    }

    // print profile
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

