/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    sceneVisualization.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>

#include "svlBase.h"
#include "svlML.h"
#include "svlPGM.h"
#include "svlVision.h"

#include "sceneConfiguration.h"
#include "sceneEntities.h"
#include "sceneAssignment.h"
#include "sceneInstance.h"
#include "sceneVisualization.h"

using namespace std;

// visualization routines
IplImage *visualize(const SceneInstance *scene, bool bBlackOutSingletonRegions)
{
    SVL_ASSERT(scene != NULL);

    vector<IplImage *> views;
    IplImage *grayImage = cvCreateImage(cvGetSize(scene->image()), IPL_DEPTH_8U, 1);
    cvCvtColor(scene->image(), grayImage, CV_BGR2GRAY);
    IplImage *background = cvCreateImage(cvGetSize(scene->image()), IPL_DEPTH_8U, 3);
    for (int c = 1; c <= 3; c++) {
        cvSetImageCOI(background, c);
        cvCopy(grayImage, background);
    }
    cvSetImageCOI(background, 0);
    cvReleaseImage(&grayImage);

    views.push_back(cvCloneImage(scene->image()));

    views.push_back(cvCloneImage(background));
    visualizeClasses(scene, views.back(), 0.5);
    if (bBlackOutSingletonRegions)
        maskSingletonRegions(scene, views.back());
    visualizeObjects(scene, views.back());

    views.push_back(cvCloneImage(scene->image()));
    cvSet(views.back(), cvScalar(255, 255, 255));
#if 0
    visualizeRegions(scene, views.back());
    visualizeClassBoundaries(scene, views.back(), true);
#else
    visualizeClasses(scene, views.back());
    visualizeClassBoundaries(scene, views.back(), false);
#endif
    if (bBlackOutSingletonRegions)
        maskSingletonRegions(scene, views.back());

    if (gSceneConfig.getOptionAsBool("useGeometry")) {
        views.push_back(cvCloneImage(background));
        visualizeHorizon(scene, views.back());
    }

    IplImage *canvas = NULL;
    switch (views.size()) {
    case 1:
        canvas = cvCloneImage(scene->image());
        break;
    case 3:
        canvas = combineImages(views, 1, 3);
        break;
    case 4:
        canvas = combineImages(views, 1, 4);
        break;
    }

    svlReleaseImages(views);
    cvReleaseImage(&background);

    return canvas;
}


void visualizeClasses(const SceneInstance *scene, IplImage *canvas, double alpha)
{
    SVL_ASSERT((scene != NULL) && (canvas != NULL) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    // resize pixel assignments
    CvMat *pixels = cvCreateMat(canvas->height, canvas->width, CV_32SC1);
    cvResize(scene->assignment.pixels, pixels, CV_INTER_NN);

    // overlay regions
    const int *q = (const int *)CV_MAT_ELEM_PTR(*pixels, 0, 0);
    for (int y = 0; y < canvas->height; y++) {
        unsigned char * const p = (unsigned char *)&canvas->imageData[y * canvas->widthStep];
        for (int x = 0; x < canvas->width; x++, q++) {
            int regId = *q; //CV_MAT_ELEM(*pixels, int, y, x);
            int classId = scene->assignment.regions[regId]->Sk;
            unsigned int c = gSceneRegionDefs.color(classId);
            unsigned char red = gSceneRegionDefs.red(c);
            unsigned char green = gSceneRegionDefs.green(c);
            unsigned char blue = gSceneRegionDefs.blue(c);

            p[3 * x + 2] = (unsigned char)((1.0 - alpha) * p[3 * x + 2] + alpha * red);
            p[3 * x + 1] = (unsigned char)((1.0 - alpha) * p[3 * x + 1] + alpha * green);
            p[3 * x + 0] = (unsigned char)((1.0 - alpha) * p[3 * x + 0] + alpha * blue);
        }
    }

    // release memory
    cvReleaseMat(&pixels);
}

void visualizeObjects(const SceneInstance *scene, IplImage *canvas, int thickness)
{
    SVL_ASSERT((scene != NULL) && (canvas != NULL) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    // resize pixel assignments
    CvMat *objectPixels = cvCreateMat(scene->height(), scene->width(), CV_32SC1);
    CvMat *canvasPixels = cvCreateMat(canvas->height, canvas->width, CV_32SC1);

    // overlay boundary
    for (int k = 0; k < scene->numObjects(); k++) {
        const SceneObject *object = scene->assignment.objects[k];
        if ((object == NULL) || object->empty()) continue;
        SVL_ASSERT(object->Ok < gSceneConfig.numObjectLabels);

        double pr = 1.0;
        CvScalar color = CV_RGB(127, 127, 127);

#if 1
        if (object->Ok == -1)
            continue;
#endif

        if (object->prObject.size() == gSceneConfig.numObjectLabels) {
            pr = 0.5 + 0.5 * object->prObject[object->Ok + 1];
        }
        unsigned c = gSceneObjectDefs.color(object->Ok);
        color = CV_RGB(pr * gSceneObjectDefs.red(c),
            pr * gSceneObjectDefs.green(c), pr * gSceneObjectDefs.blue(c));

        list<CvPoint> allPixels = object->pixels();
        cvZero(objectPixels);
        for (list<CvPoint>::const_iterator p = allPixels.begin(); p != allPixels.end(); p++) {
            CV_MAT_ELEM(*objectPixels, int, p->y, p->x) = 1;
        }
        cvResize(objectPixels, canvasPixels, CV_INTER_NN);

        list<CvPoint> boundary;
        const int *q = (const int *)CV_MAT_ELEM_PTR(*canvasPixels, 0, 0);
        for (int y = 0; y < canvasPixels->height; y++) {
            for (int x = 0; x < canvasPixels->width; x++, q++) {
                if (*q != 1) continue;
                if ((x == 0) || (x == canvasPixels->width - 1) ||
                    (y == 0) || (y == canvasPixels->height - 1) ||
                    (*(q + 1) == 0) || (*(q - 1) == 0) ||
                    (*(q + canvasPixels->width) == 0) || (*(q - canvasPixels->width) == 0))
                    boundary.push_back(cvPoint(x, y));
            }
        }

        for (list<CvPoint>::const_iterator p = boundary.begin(); p != boundary.end(); p++) {
            cvCircle(canvas, *p, 1, CV_RGB(0xff, 0xff, 0xff), thickness + 4);
        }

        for (list<CvPoint>::const_iterator p = boundary.begin(); p != boundary.end(); p++) {
            cvCircle(canvas, *p, 1, color, thickness);
        }
    }

    // release memory
    cvReleaseMat(&canvasPixels);
    cvReleaseMat(&objectPixels);
}

void visualizeRegions(const SceneInstance *scene, IplImage *canvas, bool bAvgColour, double alpha)
{
    SVL_ASSERT((scene != NULL) && (scene->numRegions() > 0) && (canvas != NULL) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    // populate color table
    vector<unsigned> colors(scene->numRegions(), 0x00000000);
    if (bAvgColour) {
        vector<double> r(scene->numRegions(), 0.0);
        vector<double> g(scene->numRegions(), 0.0);
        vector<double> b(scene->numRegions(), 0.0);
        vector<double> count(scene->numRegions(), SVL_DBL_MIN);

        const int *q = (const int *)CV_MAT_ELEM_PTR(*scene->assignment.pixels, 0, 0);
        for (int y = 0; y < scene->height(); y++) {
            unsigned char * const p = (unsigned char *)&scene->image()->imageData[y * canvas->widthStep];
            for (int x = 0; x < scene->width(); x++, q++) {
                if (*q < 0) continue;
                r[*q] += (double)p[3 * x + 2];
                g[*q] += (double)p[3 * x + 1];
                b[*q] += (double)p[3 * x + 0];
                count[*q] += 1.0;
            }
        }

        for (int i = 0; i < scene->numRegions(); i++) {
            unsigned char red = (unsigned char)(r[i] / count[i]);
            unsigned char green = (unsigned char)(g[i] / count[i]);
            unsigned char blue = (unsigned char)(b[i] / count[i]);
            colors[i] = (red << 16) | (green << 8) | blue;
        }
    } else {
        vector<int> randIndex = randomPermutation(scene->numRegions());
        unsigned delta = (0x01 << 24) / (scene->numRegions() + 1);
        colors[randIndex[0]] = delta;
        for (unsigned i = 1; i < colors.size(); i++) {
            colors[randIndex[i]] = colors[randIndex[i - 1]] + delta;
        }
    }

    // resize pixel assignments
    CvMat *pixels = cvCreateMat(canvas->height, canvas->width, CV_32SC1);
    cvResize(scene->assignment.pixels, pixels, CV_INTER_NN);

    // overlay regions
    const int *q = (const int *)CV_MAT_ELEM_PTR(*pixels, 0, 0);
    for (int y = 0; y < canvas->height; y++) {
        unsigned char * const p = (unsigned char *)&canvas->imageData[y * canvas->widthStep];
        for (int x = 0; x < canvas->width; x++, q++) {
            unsigned char red = (colors[*q] >> 16) & 0x000000ff;
            unsigned char green = (colors[*q] >> 8) & 0x000000ff;
            unsigned char blue = (colors[*q] >> 0) & 0x000000ff;

            p[3 * x + 2] = (unsigned char)((1.0 - alpha) * p[3 * x + 2] + alpha * red);
            p[3 * x + 1] = (unsigned char)((1.0 - alpha) * p[3 * x + 1] + alpha * green);
            p[3 * x + 0] = (unsigned char)((1.0 - alpha) * p[3 * x + 0] + alpha * blue);
            //p[3 * x + 2] = red;
            //p[3 * x + 1] = green;
            //p[3 * x + 0] = blue;
        }
    }

    // release memory
    cvReleaseMat(&pixels);
}

void visualizeHorizon(const SceneInstance *scene, IplImage *canvas, int thickness)
{
    SVL_ASSERT((scene != NULL) && (canvas != NULL) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

#if 0
    // resize pixel assignments
    CvMat *pixels = cvCreateMat(canvas->height, canvas->width, CV_32SC1);
    cvResize(scene->assignment.pixels, pixels, CV_INTER_NN);

    // overlay geometry
    const double alpha = 0.5;
    const int *q = (const int *)CV_MAT_ELEM_PTR(*pixels, 0, 0);
    for (int y = 0; y < canvas->height; y++) {
        unsigned char * const p = (unsigned char *)&canvas->imageData[y * canvas->widthStep];
        for (int x = 0; x < canvas->width; x++, q++) {
            int regId = *q; //CV_MAT_ELEM(*pixels, int, y, x);
            int classId = scene->assignment.regions[regId]->Sk;

            unsigned char red = 0x00;
            unsigned char green = 0x00;
            unsigned char blue = 0x00;

            switch (gSceneRegionDefs.surface(classId)) {
            case ST_SKY:
                red = 0x7f; green = 0x7f; blue = 0x7f; break;
            case ST_GROUNDPLANE:
                red = 0x00; green = 0x7f; blue = 0x00; break;
            case ST_VERTICAL:
                red = 0x7f; green = 0x00; blue = 0x00; break;
            default:
                // do nothing
                break;
            }

            p[3 * x + 2] = (unsigned char)((1.0 - alpha) * p[3 * x + 2] + alpha * red);
            p[3 * x + 1] = (unsigned char)((1.0 - alpha) * p[3 * x + 1] + alpha * green);
            p[3 * x + 0] = (unsigned char)((1.0 - alpha) * p[3 * x + 0] + alpha * blue);
        }
    }

    // release memory
    cvReleaseMat(&pixels);
#endif

    int horizon = scene->assignment.horizon * canvas->height / scene->height();

    cvLine(canvas, cvPoint(0, horizon), cvPoint(canvas->width - 1, horizon),
        CV_RGB(255, 255, 255), 2 * thickness + 1, 8);
    cvLine(canvas, cvPoint(0, horizon), cvPoint(canvas->width - 1, horizon),
        CV_RGB(0, 128, 255), thickness, 8);
}

void visualizeContrast(const SceneInstance *scene, IplImage *canvas, double alpha)
{
    SVL_ASSERT((scene != NULL) && (canvas != NULL) &&
        (canvas->width == scene->width()) && (canvas->height == scene->height()) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    // dull image
    for (int y = 0; y < canvas->height; y++) {
        unsigned char * const p = (unsigned char *)&canvas->imageData[y * canvas->widthStep];
        for (int x = 0; x < canvas->width; x++) {
            p[3 * x + 2] = (unsigned char)((1.0 - alpha) * p[3 * x + 2]);
            p[3 * x + 1] = (unsigned char)((1.0 - alpha) * p[3 * x + 1]);
            p[3 * x + 0] = (unsigned char)((1.0 - alpha) * p[3 * x + 0]);
        }
    }

    // show contrast terms
    double beta = 0.5 / scene->meanContrast();
    for (int y = 1; y < scene->height() - 1; y++) {
        unsigned char * const p = (unsigned char *)&canvas->imageData[y * canvas->widthStep];
        for (int x = 1; x < scene->width() - 1; x++) {
            double d = exp(-beta * scene->pixelContrast(cvPoint(x, y), cvPoint(x - 1, y)));
            d += exp(-beta * scene->pixelContrast(cvPoint(x, y), cvPoint(x + 1, y)));
            d += exp(-beta * scene->pixelContrast(cvPoint(x, y), cvPoint(x, y - 1)));
            d += exp(-beta * scene->pixelContrast(cvPoint(x, y), cvPoint(x, y + 1)));

            d *= 255.0 / 4.0; if (d > 255.0) d = 255.0;
            p[3 * x + 2] += (unsigned char)(alpha * d);
            p[3 * x + 1] += (unsigned char)(alpha * d);
            p[3 * x + 0] += (unsigned char)(alpha * d);
        }
    }
}

void visualizeBoundaries(const SceneInstance *scene, IplImage *canvas,
    unsigned char red, unsigned char green, unsigned char blue)
{
    SVL_ASSERT((scene != NULL) && (canvas != NULL) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    // resize pixel assignments
    CvMat *pixels = cvCreateMat(canvas->height, canvas->width, CV_32SC1);
    cvResize(scene->assignment.pixels, pixels, CV_INTER_NN);

    // draw boundaries
    for (int y = 0; y < canvas->height; y++) {
        unsigned char * const p = (unsigned char *)&canvas->imageData[y * canvas->widthStep];
        for (int x = 0; x < canvas->width; x++) {
            bool bBoundary = false;

            int regId = CV_MAT_ELEM(*pixels, int, y, x);
            bBoundary |= (y < canvas->height - 1) && (CV_MAT_ELEM(*pixels, int, y + 1, x) != regId);
            bBoundary |= (y > 0) && (CV_MAT_ELEM(*pixels, int, y - 1, x) != regId);
            bBoundary |= (x < canvas->width - 1) && (CV_MAT_ELEM(*pixels, int, y, x + 1) != regId);
            bBoundary |= (x > 0) && (CV_MAT_ELEM(*pixels, int, y, x - 1) != regId);

            if (bBoundary) {
                p[3 * x + 2] = red;
                p[3 * x + 1] = green;
                p[3 * x + 0] = blue;
            }
        }
    }

    // release memory
    cvReleaseMat(&pixels);
}

void visualizeClassBoundaries(const SceneInstance *scene, IplImage *canvas, bool bFancy)
{
    SVL_ASSERT((scene != NULL) && (canvas != NULL) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    // resize pixel assignments
    CvMat *pixels = cvCreateMat(canvas->height, canvas->width, CV_32SC1);
    cvResize(scene->assignment.pixels, pixels, CV_INTER_NN);

    if (bFancy) {
        // find boundaries
        convertInPlace(&pixels, CV_32FC1);
        CvMat *boundaries = cvCreateMat(pixels->rows, pixels->cols, CV_32FC1);
        CvMat *tmp = cvCreateMat(pixels->rows, pixels->cols, CV_32FC1);
        cvSobel(pixels, tmp, 1, 0, 1);
        cvAbs(tmp, tmp);
        cvSobel(pixels, boundaries, 0, 1, 1);
        cvAbs(boundaries, boundaries);
        cvAdd(boundaries, tmp, boundaries);
        cvMinS(boundaries, 1.0, boundaries);
        cvReleaseMat(&tmp);


        // create soft mask
        CvMat *softMask = cvCreateMat(canvas->height, canvas->width, CV_32FC1);
        cvConvertScale(boundaries, softMask);

        // dilate boundaries
        IplConvKernel* kernel = cvCreateStructuringElementEx(5, 5, 2, 2, CV_SHAPE_RECT);
        cvDilate(boundaries, boundaries, kernel);
        cvReleaseStructuringElement(&kernel);

        cvAddWeighted(softMask, 0.75, boundaries, 0.25, 0.0, softMask);

        // color using softMask as an alpha channel
        const float *alpha = (const float *)CV_MAT_ELEM_PTR(*softMask, 0, 0);
        const float *regId = (const float *)CV_MAT_ELEM_PTR(*pixels, 0, 0);
        for (int y = 0; y < canvas->height; y++) {
            unsigned char * const p = (unsigned char *)&canvas->imageData[y * canvas->widthStep];
            for (int x = 0; x < canvas->width; x++, alpha++, regId++) {
                int classId = scene->assignment.regions[(int)*regId]->Sk;
                unsigned int c = gSceneRegionDefs.color(classId);

                p[3 * x + 2] = (unsigned char)((1.0f - *alpha) * p[3 * x + 2] +
                    (*alpha) * gSceneRegionDefs.red(c));
                p[3 * x + 1] = (unsigned char)((1.0f - *alpha) * p[3 * x + 1] +
                    (*alpha) * gSceneRegionDefs.green(c));
                p[3 * x + 0] = (unsigned char)((1.0f - *alpha) * p[3 * x + 0] +
                    (*alpha) * gSceneRegionDefs.blue(c));
            }
        }

        // free memory
        cvReleaseMat(&softMask);
        cvReleaseMat(&boundaries);

    } else {
        // draw thin colored boundaries
        for (int y = 0; y < canvas->height; y++) {
            unsigned char * const p = (unsigned char *)&canvas->imageData[y * canvas->widthStep];
            for (int x = 0; x < canvas->width; x++) {
                bool bBoundary = false;

                int regId = CV_MAT_ELEM(*pixels, int, y, x);
                bBoundary |= (y < canvas->height - 1) && (CV_MAT_ELEM(*pixels, int, y + 1, x) != regId);
                bBoundary |= (y > 0) && (CV_MAT_ELEM(*pixels, int, y - 1, x) != regId);
                bBoundary |= (x < canvas->width - 1) && (CV_MAT_ELEM(*pixels, int, y, x + 1) != regId);
                bBoundary |= (x > 0) && (CV_MAT_ELEM(*pixels, int, y, x - 1) != regId);

                if (bBoundary) {
                    int classId = scene->assignment.regions[regId]->Sk;
                    unsigned int c = gSceneRegionDefs.color(classId);
                    p[3 * x + 2] = gSceneRegionDefs.red(c);
                    p[3 * x + 1] = gSceneRegionDefs.green(c);
                    p[3 * x + 0] = gSceneRegionDefs.blue(c);
                }
            }
        }
    }

    // release memory
    cvReleaseMat(&pixels);
}

void visualizeSegmentation(const SceneInstance *scene, IplImage *canvas, const CvMat *seg, int activeSeg,
    unsigned char red,  unsigned char green, unsigned char blue,
    unsigned char activeRed, unsigned char activeGreen, unsigned char activeBlue)
{
    SVL_ASSERT((scene != NULL) && (canvas != NULL) &&
        (canvas->width == scene->width()) && (canvas->height == scene->height()) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));
    SVL_ASSERT((seg != NULL) && (seg->cols == scene->width()) && (seg->rows == scene->height()));

    for (int y = 0; y < scene->height(); y++) {
        unsigned char * const p = (unsigned char *)&canvas->imageData[y * canvas->widthStep];
        for (int x = 0; x < scene->width(); x++) {
            bool bBoundary = false;

            bBoundary |= (y < scene->height() - 1) &&
                (CV_MAT_ELEM(*seg, int, y, x) != CV_MAT_ELEM(*seg, int, y + 1, x));
            bBoundary |= (y > 0) &&
                (CV_MAT_ELEM(*seg, int, y, x) != CV_MAT_ELEM(*seg, int, y - 1, x));
            bBoundary |= (x < scene->width() - 1) &&
                (CV_MAT_ELEM(*seg, int, y, x) != CV_MAT_ELEM(*seg, int, y, x + 1));
            bBoundary |= (x > 0) &&
                (CV_MAT_ELEM(*seg, int, y, x) != CV_MAT_ELEM(*seg, int, y, x - 1));

            if (bBoundary) {
                if (CV_MAT_ELEM(*seg, int, y, x) == activeSeg) {
                    p[3 * x + 2] = activeRed;
                    p[3 * x + 1] = activeGreen;
                    p[3 * x + 0] = activeBlue;
                } else {
                    p[3 * x + 2] = red;
                    p[3 * x + 1] = green;
                    p[3 * x + 0] = blue;
                }
            }
        }
    }
}

void maskSingletonRegions(const SceneInstance *scene, IplImage *canvas,
    unsigned char red, unsigned char green, unsigned char blue)
{
    SVL_ASSERT((scene != NULL) && (canvas != NULL) &&
        (canvas->width == scene->width()) && (canvas->height == scene->height()) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    for (int y = 0; y < scene->height(); y++) {
        for (int x = 0; x < scene->width(); x++) {
            int regId = CV_MAT_ELEM(*scene->assignment.pixels, int, y, x);
            if (scene->assignment.regions[regId]->size() == 1) {
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 0) = blue;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 1) = green;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 2) = red;
            }
        }
    }
}

void maskNonMatchingRegions(const SceneInstance *scene, IplImage *canvas, int regionId,
    unsigned char red, unsigned char green, unsigned char blue)
{
    SVL_ASSERT((scene != NULL) && (canvas != NULL) &&
        (canvas->width == scene->width()) && (canvas->height == scene->height()) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    for (int y = 0; y < scene->height(); y++) {
        for (int x = 0; x < scene->width(); x++) {
            if (CV_MAT_ELEM(*scene->assignment.pixels, int, y, x) != regionId) {
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 0) = blue;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 1) = green;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 2) = red;
            }
        }
    }
}

void maskNonMatchingRegions(IplImage *canvas, const CvMat *mask, int regionId,
    unsigned char red, unsigned char green, unsigned char blue)
{
    SVL_ASSERT((canvas != NULL) && (mask != NULL) &&
        (canvas->width == mask->cols) && (canvas->height == mask->rows) &&
        (canvas->nChannels == 3) && (canvas->depth == IPL_DEPTH_8U));

    for (int y = 0; y < canvas->height; y++) {
        for (int x = 0; x < canvas->width; x++) {
            if (CV_MAT_ELEM(*mask, int, y, x) != regionId) {
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 0) = blue;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 1) = green;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 2) = red;
            }
        }
    }       
}

IplImage *visualizeRegionArray(const SceneInstance *scene, int minSize, int classId)
{
    SVL_ASSERT(scene != NULL);

    vector<IplImage *> views;
    for (int k = 0; k < scene->numRegions(); k++) {
        if (scene->assignment.regions[k]->size() < minSize)
            continue;
        if ((classId >= 0) && (scene->assignment.regions[k]->Sk != classId))
            continue;

        views.push_back(cvCloneImage(scene->image()));
        for (int y = 0; y < scene->height(); y++) {
            for (int x = 0; x < scene->width(); x++) {
                if (CV_MAT_ELEM(*scene->assignment.pixels, int, y, x) == k)
                    continue;
                CV_IMAGE_ELEM(views.back(), unsigned char, y, 3 * x + 0) /= 4;
                CV_IMAGE_ELEM(views.back(), unsigned char, y, 3 * x + 1) /= 4;
                CV_IMAGE_ELEM(views.back(), unsigned char, y, 3 * x + 2) /= 4;

                if (((x > 0) && (CV_MAT_ELEM(*scene->assignment.pixels, int, y, x - 1) == k)) ||
                    ((x < scene->width() - 1) && (CV_MAT_ELEM(*scene->assignment.pixels, int, y, x + 1) == k)) ||
                    ((y > 0) && (CV_MAT_ELEM(*scene->assignment.pixels, int, y - 1, x) == k)) ||
                    ((y < scene->height() - 1) && (CV_MAT_ELEM(*scene->assignment.pixels, int, y + 1, x) == k))) {
                    CV_IMAGE_ELEM(views.back(), unsigned char, y, 3 * x + 0) = 0x00;
                    CV_IMAGE_ELEM(views.back(), unsigned char, y, 3 * x + 1) = 0x00;
                    CV_IMAGE_ELEM(views.back(), unsigned char, y, 3 * x + 2) = 0xff;
                }
            }
        }

        cvRectangle(views.back(), cvPoint(0, 0),
            cvPoint(scene->width() - 1, scene->height() - 1), CV_RGB(255, 0, 0));
        resizeInPlace(&views.back(), 96, 96 * scene->width() / scene->height());
    }

    if (views.empty()) {
        SVL_LOG(SVL_LOG_WARNING, "no regions matching criteria (minSize: " << minSize
            << ", classId:" << classId << ")");
        return NULL;
    }

    IplImage *canvas = combineImages(views);
    svlReleaseImages(views);

    return canvas;
}

IplImage *visualizeRegionSingletons(const SceneInstance *scene)
{
    SVL_ASSERT(scene != NULL);

    vector<IplImage *> views;

    for (int k = 0; k < gSceneConfig.numRegionLabels; k++) {
        views.push_back(cvCreateImage(cvGetSize(scene->image()), IPL_DEPTH_32F, 1));
        cvZero(views.back());
    }

    int varIndx = 0;
    for (int y = 0; y < scene->height(); y++) {
        for (int x = 0; x < scene->width(); x++, varIndx++) {
            vector<double> m = scene->pixelSingletons[varIndx];
            SVL_ASSERT((int)m.size() == gSceneConfig.numRegionLabels);
            expAndNormalize(m);
            for (int k = 0; k < (int)m.size(); k++) {
                CV_IMAGE_ELEM(views[k], float, y, x) += (float)m[k];
            }
        }
    }

    for (int k = 0; k < gSceneConfig.numRegionLabels; k++) {
        scaleToRange(views[k], 0.0, 255.0);
        resizeInPlace(&views[k], 96, 96 * scene->width() / scene->height());
    }

    IplImage *canvas = combineImages(views, 1, gSceneConfig.numRegionLabels);
    convertInPlace(&canvas, IPL_DEPTH_8U);
    svlReleaseImages(views);

    return canvas;
}

IplImage *visualizePixelEntropy(const SceneInstance *scene)
{
    SVL_ASSERT(scene != NULL);

    IplImage *canvas = cvCreateImage(cvGetSize(scene->image()), IPL_DEPTH_32F, 1);
    int varIndx = 0;
    for (int y = 0; y < scene->height(); y++) {
        for (int x = 0; x < scene->width(); x++, varIndx++) {
            vector<double> m = scene->pixelSingletons[varIndx];
            expAndNormalize(m);
            CV_IMAGE_ELEM(canvas, float, y, x) = (float)entropy(m);
        }
    }

    scaleToRange(canvas, 0.0, 255.0);
    convertInPlace(&canvas, IPL_DEPTH_8U);

    return canvas;
}

IplImage *visualizeOverSegmentation(const CvMat *seg)
{
    SVL_ASSERT(seg != NULL);

    IplImage *canvas = cvCreateImage(cvSize(seg->cols, seg->rows), IPL_DEPTH_8U, 3);
    cvZero(canvas);

    int *p = (int *)CV_MAT_ELEM_PTR(*seg, 0, 0);
    int *q = p - seg->cols;
    for (int y = 0; y < seg->rows; y++) {
        for (int x = 0; x < seg->cols; x++, p++, q++) {
            if ((y > 0) && (*q != *p)) {
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 0) = 0x00;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 1) = 0x00;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 2) = 0xff;
            }
            if ((x > 0) && (*(p - 1) != *p)) {
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 0) = 0x00;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 1) = 0x00;
                CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 2) = 0xff;
            }
        }
    }

    return canvas;
}

IplImage *visualizeDepths(const MatrixXd &depths, double maxDepth)
{
    IplImage *canvas = cvCreateImage(cvSize(depths.cols(), depths.rows()), IPL_DEPTH_8U, 3);
    cvZero(canvas);

    if (maxDepth < 0.0) {
        maxDepth = depths.maxCoeff();
    }

    for (int y = 0; y < depths.rows(); y++) {
        for (int x = 0; x < depths.cols(); x++) {
            if (depths(y, x) < 0)
                continue;

            unsigned char v = 0xff - (unsigned char)(255.0 * depths(y, x) / (maxDepth + SVL_DBL_MIN));
            CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 0) = v;
            CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 1) = v;
            CV_IMAGE_ELEM(canvas, unsigned char, y, 3 * x + 2) = v;
        }
    }

    return canvas;
}
