#!/usr/bin/perl
#
# STAIR VISION LIBRARY
# Copyright (c) 2007-2009, Stanford University
#
# FILENAME:    setupSceneExperiment.pl
# AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
# DESCRIPTION:
#   Sets up a scene understanding experiment.
#

use Getopt::Std;
use File::Path;
use strict;

# set defaults
my $BIN_DIR = "$ENV{HOME}/lasik/bin";
my $NUM_CLASSES = 8;
my $NUM_OBJECTS = 0;
my $USE_GEOMETRY = "true";

# check commandline options
my %opts = ();
getopts("b:d:gk:o:ux", \%opts);
if ($#ARGV != 0) {
    print STDERR "USAGE: ./setupSceneExperiment.pl [options] <full dir path>\n";
    print STDERR "OPTIONS:\n";
    print STDERR "  -b <bin>  :: applications (bin) directory (default: $BIN_DIR)\n";
    print STDERR "  -d <data> :: data directory (if one exists)\n";
    print STDERR "  -g        :: no geometry\n";
    print STDERR "  -k <n>    :: number of classes (default: $NUM_CLASSES)\n";
    print STDERR "            :: populates standard labels for (2, 3, 8 and 21)\n";
    print STDERR "  -o <n>    :: number of objects (default: $NUM_OBJECTS)\n";
    print STDERR "  -u        :: update scripts (and config) only\n";
    print STDERR "  -x        :: override existing directory\n";
    exit -1;
}

my $exprDir = $ARGV[0];

$BIN_DIR = $opts{'b'} if defined($opts{'b'});
$NUM_CLASSES = $opts{'k'} if defined($opts{'k'});
$NUM_OBJECTS = $opts{'o'} if defined($opts{'o'});
$USE_GEOMETRY = "false" if defined($opts{'g'});

my $maxClass = $NUM_CLASSES - 1;

# create experiment directory and subdirectories
print STDOUT "*** setting up experiment ***\n";
die "ERROR: experiment directory does not exist" if (defined($opts{'u'}) && (!-d $exprDir));
if (defined($opts{'x'}) && (-d $exprDir)) {
    rmtree($exprDir, {keep_root => 1, safe => 1, verbose => 1});
}

if (!defined($opts{'u'})) {
    die "ERROR: experiment directory already exists" if (-e $exprDir);
    mkpath($exprDir, {verbose => 1}) if (!-e $exprDir);
    mkpath(["${exprDir}/models", "${exprDir}/features", "${exprDir}/output",
            "${exprDir}/scripts", "${exprDir}/moves"], {verbose => 1});
    if (!defined($opts{'d'})) {
        mkpath(["${exprDir}/data/images", "${exprDir}/data/labels",
                "${exprDir}/data/multisegmentations"], {verbose => 1});
    } else {
        symlink($opts{'d'}, "${exprDir}/data");
    }
}

# create configuration file --------------------------------------------------------

open FILE, ">${exprDir}/defaultConfig.xml";
print FILE <<END_OF_CONFIG;
<svl>
  <sceneConfig numRegionLabels="${NUM_CLASSES}"
               numObjectLabels="${NUM_OBJECTS}" >

    <!-- data options -->
    <option name="baseDir" value="${exprDir}/" />
    <option name="imgDir" value="data/images/" />
    <option name="segDir" value="data/multisegmentations/" />
    <option name="lblDir" value="data/labels/" />
    <option name="featuresDir" value="features/" />
    <option name="modelsDir" value="models/" />
    <option name="outputDir" value="output/" />

    <option name="imgExt" value=".jpg" />
    <option name="segExt" value=".seg.int" />
    <option name="multiSegExt" value=".S11.R6.seg.int .S19.R6.seg.int .S31.R6.seg.int" />
    <option name="regionLblExt" value=".regions.txt" />
    <option name="layerLblExt" value=".layers.txt" />
    <option name="layerClassLblExt" value=".layerLabels.txt" />
    <option name="featuresExt" value=".features.bin" />
    <option name="regionFeaturesExt" value=".regions.features.bin" />
    <option name="auxFeatureExt" value=".aux1.txt .aux2.txt" />
    <option name="horizons" value="../horizons.txt" />

    <!-- feature options -->
    <option name="filterBandwidth" value="1" />
    <option name="featureGridSpacing" value="5" />
    <option name="featureGridScales" value = "1" />
    <option name="includeRowFeature" value="true" />
    <option name="useAuxiliaryPixelFeatures" value="false" />
    <option name="includeMomentFeatures" value="true" />

    <!-- experiment options -->
    <option name="useGeometry" value="$USE_GEOMETRY" />
    <option name="addWeightedPixelScore" value="1.0" />
    <option name="debug:boostedPixelFeaturesOnly" value="false" />

    <!-- regularization parameters / training options -->
    <option name="balancedSingletons" value="true" />
    <option name="regRegionSingleton" value="1.0e-6" />
    <option name="regHorizonRegionPair" value="1.0e-1" />

    <!-- region definitions -->
    <regionDefinitions>
      <region id="-1" name="void"       color="0 0 0"        surface="-1" />
END_OF_CONFIG

if ($NUM_CLASSES == 21) {
    print FILE <<END_OF_CONFIG;
      <region id="0"  name="building" color="128 0 0"     surface="-1" />
      <region id="1"  name="grass"    color="0 128 0"     surface="-1" />
      <region id="2"  name="tree"     color="128 128 0"   surface="-1" />
      <region id="3"  name="cow"      color="0 0 128"     surface="-1" />
      <region id="4"  name="sheep"    color="0 128 128"   surface="-1" />
      <region id="5"  name="sky"      color="128 128 128" surface="-1" />
      <region id="6"  name="airplane" color="192 0 0"     surface="-1" />
      <region id="7"  name="water"    color="64 128 0"    surface="-1" />
      <region id="8"  name="face"     color="192 128 0"   surface="-1" />
      <region id="9"  name="car"      color="64 0 128"    surface="-1" />
      <region id="10" name="bicycle"  color="192 0 128"   surface="-1" />
      <region id="11" name="flower"   color="64 128 128"  surface="-1" />
      <region id="12" name="sign"     color="192 128 128" surface="-1" />
      <region id="13" name="bird"     color="0 64 0"      surface="-1" />
      <region id="14" name="book"     color="128 64 0"    surface="-1" />
      <region id="15" name="chair"    color="0 192 0"     surface="-1" />
      <region id="16" name="road"     color="128 64 128"  surface="-1" />
      <region id="17" name="cat"      color="0 192 128"   surface="-1" />
      <region id="18" name="dog"      color="128 192 128" surface="-1" />
      <region id="19" name="body"     color="64 64 0"     surface="-1" />
      <region id="20" name="boat"     color="192 64 0"    surface="-1" />
END_OF_CONFIG
} elsif ($NUM_CLASSES == 8) {
    print FILE <<END_OF_CONFIG;
      <region id="0"  name="sky"        color="128 128 128"  surface="0" />
      <region id="1"  name="tree"       color="128 128 0"    surface="2" />
      <region id="2"  name="road"       color="128 64 128"   surface="1" />
      <region id="3"  name="grass"      color="0 128 0"      surface="1" />
      <region id="4"  name="water"      color="0 0 128"      surface="1" />
      <region id="5"  name="building"   color="128 0 0"      surface="2" />
      <region id="6"  name="mountain"   color="128 50 0"     surface="2" />
      <region id="7"  name="foreground" color="255 128 0"    surface="2" />
END_OF_CONFIG
} elsif ($NUM_CLASSES == 3) {
    print FILE <<END_OF_CONFIG;
      <region id="0"  name="sky"        color="128 128 128"  surface="0" />
      <region id="1"  name="horizontal" color="0 128 0"      surface="1" />
      <region id="2"  name="vertical"   color="128 0 0"      surface="2" />
END_OF_CONFIG
} elsif ($NUM_CLASSES == 2) {
    print FILE <<END_OF_CONFIG;
      <region id="0"  name="background" color="0 0 0"        surface="1" />
      <region id="1"  name="foreground" color="255 0 0"      surface="2" />
END_OF_CONFIG
} else {
    for (my $i = 0; $i < $NUM_CLASSES; $i++) {
        print FILE "      <region id=\"${i}\"  name=\"class ${i}\"    color=\"128 128 128\"  surface=\"-1\" />\n";
    }
}

print FILE <<END_OF_CONFIG;
    </regionDefinitions>

    <!-- object definitions -->
    <objectDefinitions>
END_OF_CONFIG

    for (my $i = 0; $i < $NUM_OBJECTS; $i++) {
        print FILE "      <object id=\"${i}\"  name=\"class ${i}\"    color=\"128 128 128\"\n";
        print FILE "              detector=\"\"\n";
        print FILE "              extractor=\"\" />\n";
    }

print FILE <<END_OF_CONFIG;
    </objectDefinitions>
  </sceneConfig>

  <svlBase.svlCodeProfiler enabled="true" />
  <svlBase.svlLogger logLevel="VERBOSE" />
  <svlBase.svlThreadPool threads="0" />
  <svlML.svlConfusionMatrix colSep=" || " rowBegin="    || " rowEnd=" \\" />
</svl>
END_OF_CONFIG

close FILE;

# create pixel classifier training script -----------------------------------------

open FILE, ">${exprDir}/trainPixelClassifiers.sh";
print FILE <<END_OF_SCRIPT;
#!/bin/csh
# TRAIN BOOSTED PIXEL CLASSIFIERS
# Stephen Gould <sgould\@stanford.edu>
#

set BASEDIR = "${exprDir}"
cd \${BASEDIR}

set FLAGS = "NOQSUB"

if ((`hostname` == "ai.Stanford.EDU") || (`hostname` == "robo.Stanford.EDU")) then
    set FLAGS = "-q quicksail"
    #set FLAGS = "-l arch=x86_64"
endif

set BINDIR = "$BIN_DIR"

set TRAINLIST = "\${BASEDIR}/trainList.txt"
set EVALLIST = "\${BASEDIR}/evalList.txt"

# regions
if (1) then
    set CONFIG = "\${BASEDIR}/defaultConfig.xml"

    set SUBSAMPLE = 250
    set ROUNDS = 500
    set SPLITS = 2

    set OPTIONS = "-subSample \$SUBSAMPLE -rounds \$ROUNDS -splits \$SPLITS"
    set OPTIONS = "-set svlML.svlBoostedClassifier boostMethod DISCRETE \${OPTIONS}"
    set OPTIONS = "-verbose -profile \${OPTIONS}"

    foreach i (`seq 0 ${maxClass}`)
        set SCRIPT = "scripts/__trainBoostedClass\${i}.sh"
        set CMD = "\$BINDIR/sceneLearnPixelClassifiers -config \$CONFIG \$OPTIONS -classId \${i} \$TRAINLIST"
        echo \$CMD

        if ("\$FLAGS" == "NOQSUB") then
            \$CMD
        else
            echo 'cd \$PBS_O_WORKDIR' > \$SCRIPT
            echo "\${CMD}" >>  \$SCRIPT

            waitforQs.sh 10
            qsub \$FLAGS \$SCRIPT
        endif
    end
endif
END_OF_SCRIPT

close FILE;

# create pipeline script ----------------------------------------------------------

open FILE, ">${exprDir}/pipeline.sh";
print FILE <<END_OF_SCRIPT;
#!/bin/csh
# SCENE UNDERSTANDING PIPELINE
# Stephen Gould <sgould\@stanford.edu>
#

set BASEDIR = "${exprDir}"
set BINDIR = "${BIN_DIR}"

set TRAINLIST = "\${BASEDIR}/trainList.txt"
set EVALLIST = "\${BASEDIR}/evalList.txt"

set CONFIG = "\${BASEDIR}/defaultConfig.xml"

# train pixel classifiers
if (! -e \$BASEDIR/models/pixel.class.xml) then
    \$BASEDIR/trainPixelClassifiers.sh
    rm -f \${BASEDIR}/features/*.bin
endif

# train pixel singletons
\$BINDIR/sceneLearnPixelModel -config \$CONFIG -component SINGLETON -subSample 250 \$TRAINLIST

# train pixel pairwise contrast
\$BINDIR/sceneLearnPixelModel -config \$CONFIG -component CONTRAST \$TRAINLIST

# evaluate the pixel model
\$BASEDIR/evalPixelModel.sh
END_OF_SCRIPT

if ($USE_GEOMETRY eq "true") {
    print FILE <<END_OF_SCRIPT;

# learn horizon models
\$BINDIR/sceneLearnRegionModel -config \$CONFIG -component HORIZON \$TRAINLIST
\$BINDIR/sceneLearnRegionModel -config \$CONFIG -component HORIZONPAIR \$TRAINLIST
END_OF_SCRIPT
}

print FILE <<END_OF_SCRIPT;

# learn region models
\$BINDIR/sceneLearnRegionModel -config \$CONFIG -component SINGLETON \$TRAINLIST

# learn pairwise contrast weight
echo 1.0 > \${BASEDIR}/models/region.pairwiseContrastWeight.txt

# evaluate on training data
\$BASEDIR/evalRegionModel.sh 1 1

# harvest training examples
rm -f \${BASEDIR}/output/trainingCache.bin
\$BINDIR/sceneHarvestTrainingExamples -config \$CONFIG -m FEATURES \\
    -outCache output/trainingCache.bin \$TRAINLIST

# re-learn region models
\$BINDIR/sceneLearnRegionModel -config \$CONFIG -component SINGLETON \\
    -set sceneConfig balancedSingletons false \\
    -set sceneConfig closedLoopFeatureCache output/trainingCache.bin \\
    \$TRAINLIST

# evaluate region model
\$BASEDIR/evalRegionModel.sh

exit

# learn pairwise region model
\$BINDIR/sceneLearnRegionModel -config \$CONFIG -m REGIONPAIR -component PAIRWISE \\
    -set sceneConfig auxLblDir output/ \\
    -set sceneConfig auxLayerLblExt .predictedRegions.txt \\
    -set sceneConfig auxLayerClassLblExt .correctedLabels.txt \\
    \$TRAINLIST

# learn pairwise weight
echo 1.0 > \${BASEDIR}/models/region.pairwiseWeight.txt

# learn object context model
\$BINDIR/sceneLearnRegionModel -config \$CONFIG -m OBJECTCONTEXT -component PAIRWISE \\
    \$TRAINLIST

# learn object context weight
echo 1.0 > \${BASEDIR}/models/objectContext.weight.txt

END_OF_SCRIPT

close FILE;

# create pixel evaluation script ---------------------------------------------------

open FILE, ">${exprDir}/evalPixelModel.sh";
print FILE <<END_OF_SCRIPT;
#!/bin/csh
# EVALUATE PIXEL MODEL
# Stephen Gould <sgould\@stanford.edu>
#

set BASEDIR = "${exprDir}"
cd \${BASEDIR}

set FLAGS = "NOQSUB"

if ((`hostname` == "ai.Stanford.EDU") || (`hostname` == "robo.Stanford.EDU")) then
    #set FLAGS = "-q quicksail"
    set FLAGS = "-l arch=x86_64"
endif

set BINDIR = "${BIN_DIR}"

set FEATURESDIR = "\${BASEDIR}/features/"
set MODELDIR = "\${BASEDIR}/models/"

set TRAINLIST = "\${BASEDIR}/trainList.txt"
set EVALLIST = "\${BASEDIR}/evalList.txt"
set ALLLIST = "\${BASEDIR}/allList.txt"

set CONFIG = "\${BASEDIR}/defaultConfig.xml"

# evaluate
set OPTIONS = "-outClasses .pixelClasses.txt -outImages .pixelResults.png"
set OPTIONS = "\${OPTIONS} -verbose -profile"

@ counter = 0
foreach i (`cat \$EVALLIST | sort`)
    set CMD = "\$BINDIR/sceneEvalPixelModel -config \${CONFIG} \$OPTIONS \$i"
    echo \$CMD

    if (! -e \${BASEDIR}/output/\${i}.pixelClasses.txt) then
        if ("\$FLAGS" == "NOQSUB") then
            \$CMD
        else
            set SCRIPT = "scripts/__evalPixelModel.\${i}.sh"
            echo "cd \${BASEDIR}/scripts" > \$SCRIPT
            echo "\${CMD}" >>  \$SCRIPT

            if (\$counter % 5 == 0) then
                waitforQs.sh 50
                sleep 1
            endif
            qsub \$FLAGS \$SCRIPT
            @ counter += 1
        endif
    endif
end

# process training
set OPTIONS = ""

foreach i (`cat \$TRAINLIST | sort`)
    set CMD = "\$BINDIR/sceneEvalPixelModel -config \${CONFIG} \$OPTIONS \$i"
    echo \$CMD

    if (! -e \${BASEDIR}/features/\${i}.features.bin) then
        if (\$FLAGS == "NOQSUB") then
            \$CMD
        else
            set SCRIPT = "scripts/__evalPixelModel.\${i}.sh"
            echo "cd \${BASEDIR}/scripts" > \$SCRIPT
            echo "\${CMD}" >>  \$SCRIPT

            if (\$counter % 5 == 0) then
                waitforQs.sh 50
                sleep 1
            endif
            qsub \$FLAGS \$SCRIPT
            @ counter += 1
        endif
    endif
end
END_OF_SCRIPT

close FILE;

# create region evaluation script ---------------------------------------------------

open FILE, ">${exprDir}/evalRegionModel.sh";
print FILE <<END_OF_SCRIPT;
#!/bin/csh
# EVALUATE REGION MODEL
# Stephen Gould <sgould\@stanford.edu>
#
# USAGE: ./evalRegionModel.sh (<clear>) (<train>)
#

set BASEDIR = "${exprDir}"
cd \${BASEDIR}

set FLAGS = "NOQSUB"

if ((`hostname` == "ai.Stanford.EDU") || (`hostname` == "robo.Stanford.EDU")) then
    #set FLAGS = "-q quicksail"
    set FLAGS = "-l arch=x86_64"
endif

set BINDIR = "${BIN_DIR}"

set FEATURESDIR = "\${BASEDIR}/features/"
set MODELDIR = "\${BASEDIR}/models/"
set OUTPUTDIR = "\${BASEDIR}/output/"

set TRAINLIST = "\${BASEDIR}/trainList.txt"
set EVALLIST = "\${BASEDIR}/evalList.txt"
set ALLLIST = "\${BASEDIR}/allList.txt"

set CONFIG = "\${BASEDIR}/defaultConfig.xml"

set LIST = \$EVALLIST
if (\$2 == 0) then
    set LIST = \$EVALLIST
endif
if (\$2 == 1) then
    set LIST = \$TRAINLIST
endif

if (\$1 == 1) then
    if (\$LIST == \$EVALLIST) then
        rm -f \${OUTPUTDIR}/*.predictedRegions.txt
        rm -f \${OUTPUTDIR}/*.predictedClasses.txt
        rm -f \${OUTPUTDIR}/*.predictedHorizon.txt
        rm -f \${OUTPUTDIR}/*.predictedLabels.txt
        rm -f \${OUTPUTDIR}/*.results.png
    else
        rm -f \${OUTPUTDIR}/*.mistakes.bin
    endif
endif

# evaluate on either training or testing sets
set OPTIONS = "-m REGION"
if (\$LIST == \$EVALLIST) then
    set OPTIONS = "\${OPTIONS} -outRegions .predictedRegions.txt"
    set OPTIONS = "\${OPTIONS} -outClasses .predictedClasses.txt"
    set OPTIONS = "\${OPTIONS} -outHorizon .predictedHorizon.txt"
    set OPTIONS = "\${OPTIONS} -outImages .results.png"
else
#    set OPTIONS = "\${OPTIONS} -outRegions .predictedRegions.txt"
#    set OPTIONS = "\${OPTIONS} -outLabels .predictedLabels.txt"
    set OPTIONS = "\${OPTIONS} -set svlBase.svlLogger logLevel MESSAGE"
endif

@ counter = 0
foreach i (`cat \$LIST | sort`)
    if (\$LIST == \$EVALLIST) then
        set CMD = "\$BINDIR/sceneInference -config \${CONFIG} \$OPTIONS \$i"
        set FILECHECK = "\${OUTPUTDIR}/\${i}.predictedRegions.txt"
    else
        set CMD = "\$BINDIR/sceneInference -config \${CONFIG} -record \${OUTPUTDIR}/\${i}.mistakes.bin \$OPTIONS \$i"
        set FILECHECK = "\${OUTPUTDIR}/\${i}.mistakes.bin"
    endif
    echo \$CMD

    if (! -e \$FILECHECK) then
        if ("\$FLAGS" == "NOQSUB") then
            \$CMD
        else
            set SCRIPT = "scripts/__evalRegionModel.\${i}.sh"
            echo "cd \${BASEDIR}/scripts" > \$SCRIPT
            echo "\${CMD}" >> \$SCRIPT

            if (\$counter % 5 == 0) then
                waitforQs.sh 50
                sleep 1
            endif
            qsub \$FLAGS \$SCRIPT
            @ counter += 1
        endif
    endif
end

END_OF_SCRIPT

close FILE;

# create cross-validatation script --------------------------------------------------

# TODO

# create clean script ---------------------------------------------------------------

open FILE, ">${exprDir}/clean.sh";
print FILE <<END_OF_SCRIPT;
#!/bin/csh
# CLEAN SCRIPT
# Stephen Gould <sgould\@stanford.edu>
#
# USAGE: ./clean.sh
#

rm -f __*.sh.*
rm -f *~

END_OF_SCRIPT

close FILE;

# script execute permissions --------------------------------------------------------

chmod 0755, "${exprDir}/trainPixelClassifiers.sh";
chmod 0755, "${exprDir}/pipeline.sh";
chmod 0755, "${exprDir}/evalPixelModel.sh";
chmod 0755, "${exprDir}/evalRegionModel.sh";
chmod 0755, "${exprDir}/clean.sh";

# give final instructions -----------------------------------------------------------

print STDOUT <<END_OF_INSTRUCTIONS;
*** experiment setup complete ***

You should now prepare the data if it does not already exist. Then
create files "allList.txt", "trainList.txt" and "evalList.txt" in the
experiment directory. Then submit the pipeline script to one of the
cluster machines.  You may also want to edit the "defaultConfig.xml"
file.

Knobs to tune at the pixel level: discrete vs gentle boost, number of
rounds, subsampling rate, balanced singletons, pairwise strength.

Knobs to tune at the region level: balanced singletons, pairwise
strengths, inclusion of pixel terms.

END_OF_INSTRUCTIONS

